c
c    Copyright (C) 1997  Peter Vernon Lanspeary
c
c    This program is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 2 of the License, or
c    (at your option) any later version.
c
c    This program is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program; if not, write to the Free Software
c    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
c


c
c
c
c               ###############################################
c               ##                                           ##
c               ##     ILGLIB - interactive line graphics    ##
c               ##                                           ##
c               ##           Subroutine library              ##
c               ##                                           ##
c               ##        Plotting multiple x-y line         ##
c               ##        graphs :- interactive and          ##
c               ##        hardcopy                           ##
c               ##                                           ##
c               ###############################################
c
c LIST OF SUBROUTINES
c    ILGOPEN( CCAPFILE )
c       Initializes screen graphics and open hard copy 'capture' file
c    ILGHADJ( XSPOFF, YSPOFF )
c       Specify a special hard plot special origin offset .
c    ILGCHAR( RCHRSZ, RSYMSZ )
c       Set the size of labelling characters and plotted special symbols .
c    ILGPAGE( CPAGSIZ, CORIENT, NXPLTS, NYPLTS )
c       Specifies the page size, page orientation, and the number of plotting
c       cells on the page .
c    ILGRAPH( NXPOS, NYPOS, NXAXIS, NYAXIS, NBORDER )
c       Specifies the plotting cell to be used, axis types and border type .
c    ILGSLAW( CXSLAW, CYSLAW )
c       Specifies the scaling law to be used - linear or logarithmic .
c    ILGIMIT( XMIN, YMIN, XMAX, YMAX )
c       Specifies the lower left and upper right data limits
c    ILGLINE( NSYMB, NDASH, NPEN )
c       Specifies the plotting symbol for data points, dashed line type and
c       pen number ( colour / thickness )
c    ILGCLIP( FCLIP )
c       Switch clipping at the graph boundary on or off .
c    ILGFSET( XOFFSET, YOFFSET )
c       Adds offset to any coordinates plotted later .
c    ILGWILD( CUNITS, XMIN, YMIN, XMAX, YMAX )
c       Set limits or sizes to arbitrary values .
c    ILGDATA( X, Y, NINCR, NPTS, NEW )
c       Plots the (x,y) data in the next or current plotting cell .
c    ILGXLAB( CXLABEL )
c       Labels the x axis
c    ILGYLAB( CYLABEL )
c       Labels the y axis
c    ILGTITL( CTITLE  )
c       Puts a title at the top of the graph
c    ILGNOTE( CNOTE )
c       Writes a note onto the graph .
c    ILGREAD( X, Y, CHAR )
c       Read a coordinate from the VDU screen using cross-hairs .
c    ILGWAIT
c       Pauses until the user presses the <carriage return> key
c    ILGCLOS
c       Closes any capture file, clears the vdu screen and terminates graphics
c
c
c
c     HOW TO USE THE SUBROUTINES
c    ############################
c
c    Conventions
c    ===========
c  1) User accessible routines have names beginning with ILG followed by 4
c       letters . Other routines begin with ILG followed by 3 letters .
c  2) The data type of each argument is specified by its initial letter
c         C = character
c         N = integer *4 or integer *2
c               -except in ILGDATA where they must be integer *4
c         X, Y = real *4
c
c    Errors
c    ======
c  Subroutine arguments are checked before action is taken . Most errors
c  cause the called routine to abort without doing anything . The error
c  channel can be changed by a call to ILGERRR .
c
c    Modes
c    =====
c  Some routine can be made to behave in a manner not indicated by their
c  arguments , ie. they can be put into a `wild mode' . This behaviour is
c  triggered by a call to ILGWILD . ILGWILD does this by setting the
c  `wild mode' flag . All other ILGLIB routines reset the `wild mode' flag
c  if they execute successfully . This means that after calling ILGWILD ,
c  only the very next ILGLIB routine used can be made `wild' .
c
c    NOTE :  The conventional behaviour of ILGLIB is quite robust, but if
c    `wild mode' is invoked , ILGLIB can become fragile and MAY produce
c    messy results . Plotting is an inherently complicated process , and
c    there are many implicit rules . By using ILGWILD , you can break some
c    of these rules , but YOU must take the responsibility for the
c    consequences ; so BE CAREFUL ! In most cases , ILGLIB can do whatever
c    you want without using ILGWILD . However , adventurous users will find
c    that ILGWILD can introduce considerable flexibility and freedom to
c    their plotting . Only ILGCHAR, ILGPAGE, ILGRAPH, ILGFSET, ILGCLIP and
c    ILGIMIT can be executed in `wild mode' .
c
c     PAGE MARGIN AND CHARACTER SIZES
c    #################################
c  Dimensions are in centimeters
c
c              A4         B4         A3         A2    Computer Paper
c          ----------------------------------------------------------
c sizes     21.0x29.7  25.5x36.3  29.7x42.0  42.0x59.4  28.0x38.1
c
c wide orientation             <---Margins--->                 tall orientation
c    top       3.5        3.5        3.5        3.5        3.5      LHS
c   other      2.0        2.2        2.6        3.2        2.4     other
c
c Character sizes - heights
c   This table shows character height for NXPLT = NYPLT = 1 ie. apparent
c character height .
c   {actual character height} = {apparent height} * [ NXPLT * NYPLT ] ** -0.25
c labels       0.30       0.36       0.40       0.60       0.40
c titles       0.50       0.60       0.70       0.90       0.70
c minimum      0.16       0.19       0.22       0.30       0.22
c
c     MAXIMUM NUMBER OF PLOTTING CELLS
c  short side   10         11         12         14         13
c  long side    17         19         20         24         21
c


      BLOCK DATA ILGBLK
c
c           initialize common block variables
c
      implicit none
      include 'ilglib.cmn'
      data fgraph, fslaw, fline, flogx, flogy / 5* .false. /
      data idefpen / 4*1 /
      data rpenth   /  0.10 /
      end


      SUBROUTINE ILGOPEN( CCAPFILE )
c
c       Initializes screen graphics and open hard copy 'capture' file
c  Parameters :
c    CCAPFILE - VMS name of the file on which the hard copy information is to
c               be stored .
c               If the value of CCAPFILE is '?' , the user is prompted
c               for the capture file name
c
c  Defaults set :
c    page - paper size = 'A4', paper orientation = 'wide', single plot cell on
c           the page
c    graph - plot cell used is (1,1) ie lower left , axes are drawn and
c           labelled, no border or grid is drawn
c    line - the point '.' symbol is used, no line is drawn to join points,
c           the first pen is used
c    scaling law - is 'linear' along both axes
c
c constants
c  integer *2 Hkbdlu          # keyboard logical unit number
c             Hvdulu          # v.d.u. logical unit number
c  character Capfile*64       # capture file name used if user is prompted
c
      implicit none
      integer *2 Hkbdlu, Hvdulu
      parameter( Hkbdlu=5, Hvdulu=6 )
c
      character Ccapfile*(*), Capfile*64
      include 'ilglib.cmn'
c
c
c           close graphics if already open
c
      if ( fopen )
     1  call ilgclos
      fopen = .true.
c
c           Initialize graphics
c
      call grstrt( 0, 0 )
      fcapt = .false.
      if ( Ccapfile(1:1) .eq. '?' ) then
        write( Hvdulu, 1 )
1       format( ' Specify capture file name ' )
        read( Hkbdlu, 2 ) Capfile
2       format( a )
        if ( Capfile(1:1) .ne. ' ' ) then
          call opncap( Capfile )
          fcapt = .true.
          end if
      else if ( Ccapfile(1:1) .ne. ' ' ) then
        call opncap( Ccapfile )
        fcapt = .true.
        end if
c
c           special plotter origin offset for H/P plotters
c
      if ( ifrom .ne. Hilghadj )
     1  call ilghadj( -2.0, -2.0 )
      call plots( 0, 0, 6 )
c
c           page defaults
c
      if ( ifrom .ne. Hilgchar )
     1  call ilgchar( 1.0, 1.0 )
      if ( ifrom .ne. Hilgpage )
     1  call ilgpage( 'a4', 'wide', 1, 1 )
c
c           graph defaults
c
      if ( ifrom .ne. Hilgraph )
     1  call ilgraph( 1, 1, 2, 2, 0 )
      ixpos = 0
      iypos = 1
c
c           line defaults
c
      if ( ifrom .ne. Hilgline )
     1  call ilgline( 0, 1, 1 )
c
c           scaling defaults
c
      rdatfacx = 0.1
      rdatfacy = 0.1
      rxmin = 0.0
      rxmax = 0.0
c
c           flags
c
      ierlun = 6
      fwait = .false.
      fklip = .true.
      fwklip = .false.
      fgraph = .false.
      return
      end



      SUBROUTINE ILGHADJ( XSPOFF, YSPOFF )
c
c       Specifies a special hard plotter origin offset ( XSPOFF, YSPOFF ) .
c    These are the coordinates of the plotter firmware origin relative to the
c    bottom left hand corner of the page - in centimetres .
c       The default value , which is set by a call to ILGOPEN, is
c    ( XSPOFF, YSPOFF ) = ( -2.0,- 2.0 ) . The default offset is suitable for
c    the Hewlett/Packard 7470A , 7475A and 7550 plotters .
c
c arguments
c  real xspoff, yspoff         # coordinates of plotter firmware origin from
c                              # the bottom left corner of the page .
c
      implicit none
      real xspoff, yspoff
      include 'ilglib.cmn'
c
      ifrom = Hilghadj
      if ( .not. fopen )
     1  call ilgopen( ' ' )
      rxspoff = xspoff
      ryspoff = yspoff
c
c           specify new window
c
      if ( fpage )
     1  call window( rpagzerx - rxspoff,
     2   rpagzerx + rpagsizx + rxspoff + 0.003,
     3   rpagzery - ryspoff,
     4   rpagzery + rpagsizy + ryspoff + 0.003 )
      fwild = .false.
c
      return
      end



      SUBROUTINE ILGCHAR( RCHRSZ, RSYMSZ )
c
c       Set the size of labelling characters and plotted special symbols
c       relative to the default size .
c   Parameters :
c     RCHRSZ - size factor for labelling ( x-axis, y-axis, note, title )
c               characters .
c     RSYMSZ - size factor for special symbols used when plotting
c
c       e.g. CALL ILGCHAR( 2.0, 0.5 ) doubles the height of all labelling
c   characters and halve the height of all plotted special symbols .
c
c   N.B. The character sizes specified by ILGCHAR only become active when the
c        ILGPAGE routine is executed . The relative size limits are 0.2 & 5.0
c
c   When in `wild mode' ( ie. set ) , the character height is
c   IMMEDIATETLY set to the distance between the two points in the wild
c   registers . Plotted symbol height is two thirds of the character height .
c   The values of the arguments RCHRSZ and RSYMSZ are ignored .
c
      implicit none
      real rchrsz, rsymsz, rhi
      logical *1 ilguuc
      include 'ilglib.cmn'
c
      ifrom = Hilgchar
      if ( .not. fopen )
     1  call ilgopen( ' ' )
      ifrom = Hilgnone
c
c         wild action
c
      if ( fwild ) then
        if ( fuser ) then
          if ( .not. ilguuc( .true. ) )
     1      return
          end if
        if ( rwild(1) .ne. rwild(3) .or. rwild(2) .ne. rwild(4) ) then
          rhi = min( sqrt( (rwild(1)-rwild(3))**2 +
     1      (rwild(2)-rwild(4))**2 ), min(rcelsizx,rcelsizy) )
          rchrsiz = rhi / rchrhi
          rsymsiz = rchrsiz
          rchrhi = rhi
          rchrti = rhi
          rsymhi = 0.667 * rhi
        end if
      else
c
c         conventional action
c
        if ( rchrsz .lt. 0.2 .or. rchrsz .gt. 10.0 ) then
          if ( ierlun .gt. 0 ) write( ierlun, 1 ) rchrsz
1         format( ' ILGCHAR : ABORT : Silly character size : ', g13.6 /
     1      11x, 'Must be between 0.1 and 10.0 .' )
          if ( fwait ) call ilgwait( .false. )
        else if ( rsymsz .lt. 0.2 .or. rsymsz .gt. 10.0 ) then
          if ( ierlun .gt. 0 ) write( ierlun, 2 ) rsymsz
2         format( ' ILGCHAR : ABORT : Silly symbol size : ', g13.6 /
     1      11x, 'Must be between 0.1 and 10.0 .' )
          if ( fwait ) call ilgwait( .false. )
        else
          rchrsiz = rchrsz
          rsymsiz = rsymsz
          end if
        end if
      fwild = .false.
      return
      end


      LOGICAL *1 FUNCTION ILGUUC( ISTART )
c
c  Arguments:
c    integer *2 ISTART       # +ve  - forward transformation
c                            # -ve  - backward transformation
c  Forward transformation :-
c       If the wild registers contain coordinates in User Units , convert
c       them into centimetres .
c  Backward transformation :-
c       If the wild registers contain coordinates in centimetres , convert
c       them into User Units .
c  ISTART is the first `wild' register to be converted . Its legal values
c       are -5, -1, 1, and 5 .
c
c  integer *2 ii             # abs( istart )
c
      implicit none
      integer *2 istart, ii
      real rxbase, rybase
      include 'ilglib.cmn'
c
c         forward transformation
c
      ilguuc = .false.
      ii = abs( istart )
      if ( ii .ne. 1 .and. ii .ne. 5 ) return
      ilguuc = .true.
      call ilgrof
      rxbase = rgroffx + ( rxofset - rxmin ) * rdatfacx
      rybase = rgroffy + ( ryofset - rymin ) * rdatfacy
      if ( istart .gt. 0 .and. fuser ) then                               1b
c
c           check for illegal values in wild registers
c
        if ( flogx .and. ( rwild(1) .le. 0 .or. rwild(2) .le. 0 ) .or.
     1      flogy .and. ( rwild(3) .le. 0 .or. rwild(4) .le. 0 ) ) then   2b
          ilguuc = .false.
          if ( ierlun .gt. 0 )
     1      write( ierlun, 3 )
3           format( ' ILG**** : ABORT : Negative value on log axis' )
          if ( fwait ) call ilgwait( .false. )
        else                                                              2m
c
c             convert 1st coordinate to cm
c
          if ( flogx )
     1      rwild(ii) = log10( rwild(ii) )
          rwild(ii) = rwild(ii) * rdatfacx + rxbase
c
          ii = ii + 1
          if ( flogy )
     1      rwild(ii) = log10( rwild(ii) )
          rwild(ii) = rwild(ii) * rdatfacy + rybase
c
c           convert 2nd coordinate to cm
c
          ii = ii + 1
          if ( flogx )
     1      rwild(ii) = log10( rwild(ii) )
          rwild(ii) = rwild(ii) * rdatfacx + rxbase
c
          ii = ii + 1
          if ( flogy )
     1      rwild(ii) = log10( rwild(ii) )
          rwild(ii) = rwild(ii) * rdatfacy + rybase
          end if                                                          2e
c
c         backward transformation
c
      else if ( .not. fuser ) then                                        1m
c
c             convert 1st coordinate
c
        rwild(ii) = ( rwild(ii) - rxbase ) / rdatfacx
        if ( flogx )
     1    rwild(ii) = 10.0 ** rwild(ii)
c
        ii = ii + 1
        rwild(ii) = ( rwild(ii) - rybase ) / rdatfacy
        if ( flogy )
     1    rwild(ii) = 10.0 ** rwild(ii)
c
c           convert 2nd coordinate to cm
c
        ii = ii + 1
        rwild(ii) = ( rwild(ii) - rxbase ) / rdatfacx
        if ( flogx )
     1    rwild(ii) = 10.0 ** rwild(ii)
c
        ii = ii + 1
        rwild(ii) = ( rwild(ii) - rybase ) / rdatfacy
        if ( flogy )
     1    rwild(ii) = 10.0 ** rwild(ii)
        end if                                                           1e
      return
      end


      SUBROUTINE ILGPAGE( CPAGSIZ, CORIENT, NXPLTS, NYPLTS )
c
c       Set page size, orientation and number of plotting cells .
c  Calculates the page margins ( origin and size ) hardclip limits,
c  character heights , and plotting cell size and offset relative to the
c hardclip limits
c
c  Parameters :
c    CPAGSIZ - Size of the paper for hardcopy use .
c         Values :- 'A2', 'A3', 'A4', 'B4', 'CP'
c            'CP' is 15 inch computer paper . Lower case letters are allowed .
c    CORIENT - Orientation of paper in the plotter .
c         Values :- 'wide', 'tall', 'WIDE', 'TALL' . If UPPER case letters are
c         the screen is NOT cleared .
c    NXPLTS - Number of plotting cells in the x direction .
c         Values :- minimum = 1, maximum depends on page size
c    NYPLTS - Number of plotting cells in the y direction .
c         Values :- minimum = 1, maximum depends on page size
c
c  When in `wild mode' , the page corners are set at the lower
c  left and upper right coordinates in the wild registers . The page size
c  and orientation parameters CPAGSIZ and CORIENT are ignored , and the
c  character size and margins are deduced from the previously defined page.
c
c N.B. Total number of plotting cells on the page is   NXPLTS * NYPLTS
c  Defaults set :
c    graph - plot cell used is (1,1) ie lower left
c
c constants
c  integer *2 Maxxplts(5)      # Max no. of cells allowed in x direction
c             Maxyplts(5)      # Max no. of cells allowed in y direction
c             Hv500w,          # screen width in pixels
c             Hv500h           # screen height in pixels
c  real Upagsizw(5)            # page width in cm of each recognized page size
c                              #  A4, B4, A3, A2, CP
c       Upagsizh(5)            # page height in cm of each recognized page size
c       Ulabchr(5)             # apparent height of axis label characters
c       Utitchr(5)             # apparent height of title characters
c       Uminchr(5)             # minimum character height
c       Upagmarn(5)            # narrow page margins
c       Upagmarw(5)            # wide page margins
c       xlo, ylo,              # lower left corner of page
c       xhi, yhi,              # upper right corner of page
c       scrh, scrw             # viewport height & width of the screen
c  character Gpagsel*20        # valid page selections
c            Gorisel*16        # valid orientation selections
c variables
c  real rchrfac                # factor by which char size is reduced
c  integer *2 Imaxxplts,       # max no. of cells in x direction
c             Imaxyplts        # max no. of cells in y direction
c  logical *1 ilgins           # function : "IN String"
c             ilguuc           #          : User units conversion
c
      implicit none
      integer *2 Maxxplts(5), Maxyplts(5)
      real Upagsizw(5), Upagsizh(5), Ulabchr(5), Utitchr(5), Uminchr(5),
     1  Upagmarn(5), Upagmarw(5), xlo, ylo, xhi, yhi, scrw, scrh
      real Hv500w, Hv500h
      character Gpagsel*20, Gorisel*16
      parameter( Gpagsel='A4B4A3A2CPa4b4a3a2cp',
     1           Gorisel='WIDETALLwidetall',
     2           Hv500w=1023.0, Hv500h=779.0 )
c
      character *(*) cpagsiz, corient
      integer *2 nxplts, nyplts, Imaxxplts, Imaxyplts
      logical *1 ilgins, ilguuc
      real rchrfac
      include 'ilglib.cmn'
c
c       page size dependent data
c                      a4     b4     a3     a2     cp
      data Upagsizw / 29.70, 36.30, 42.00, 59.40, 38.10 /
      data Upagsizh / 21.00, 25.50, 29.70, 42.00, 28.00 /
      data Ulabchr  /  0.40,  0.48,  0.56,  0.72,  0.56 /
      data Utitchr  /  0.50,  0.60,  0.70,  0.90,  0.70 /
      data Uminchr  /  0.20,  0.24,  0.28,  0.36,  0.28 /
      data Upagmarw /  3.50,  3.50,  3.50,  3.50,  3.50 /
      data Upagmarn /  2.00,  2.20,  2.60,  3.20,  2.40 /
      data Maxyplts /    10,    11,    12,    14,    13 /
      data Maxxplts /    17,    19,    20,    24,    21 /
c
c           safety catch
c
      ifrom = Hilgpage
      if ( .not. fopen )
     1  call ilgopen( ' ' )
      ifrom = Hilgnone
c
c           wild action
c
      if ( fwild ) then                                                    1b
        if ( nxplts .le. 0 .or. nyplts .le. 0 ) then                       2b
          write( ierlun, 3 ) nxplts, nyplts
          return
          end if                                                           2e
        if ( fuser ) then                                                  2b
          if ( .not. ilguuc( 1 ) )
     1      return
          end if                                                           2e
        rpagzerx = min( rwild(1), rwild(3) ) - rxspoff
        rpagzery = min( rwild(2), rwild(4) ) - ryspoff
        rpagsizx = abs( rwild(3) - rwild(1) )
        rpagsizy = abs( rwild(4) - rwild(2) )
      else                                                                 1m
c
c           check arguments
c
        if ( .not. ilgins( cpagsiz, 2, Gpagsel ) ) then                    2b
          if ( ierlun .gt. 0 ) write( ierlun, 1 ) cpagsiz
1         format( ' ILGPAGE : ABORT error : invalid paper size - ', a /
     1     11x, 'Choose from ''A4'', ''B4'', ''A3'', ''A2'' and ''CP''')
          if ( fwait ) call ilgwait( .false. )
          return
        else if ( .not. ilgins( corient, 4, Gorisel ) ) then               2m
          if ( ierlun .gt. 0 ) write( ierlun, 2 ) corient
2         format(' ILGPAGE : ABORT error : invalid paper orientation- ',
     1      a / 11x, 'Choose from ''WIDE'' and ''TALL''' )
          if ( fwait ) call ilgwait( .false. )
          return
        else if ( nxplts .le. 0 .or. nyplts .le. 0 ) then                  2m
          if ( ierlun .gt. 0 ) write( ierlun, 3 ) nxplts, nyplts
3         format( ' ILGPAGE : ABORT error : number of plotting cells- ',
     1      i5, ' by ', i5 / 11x, 'Must be positive integers' )
          if ( fwait ) call ilgwait( .false. )
          return
          end if                                                           2e
        Ipagsiz = mod( ( index( Gpagsel, cpagsiz(1:2) ) - 1 ) / 2, 5 )+1
        Iorient = mod( ( index( Gorisel, corient(1:4) ) - 1 ) / 4, 2 )+1
        Imaxxplts = Maxxplts( Ipagsiz )
        Imaxyplts = Maxyplts( Ipagsiz )
        if ( Iorient .eq. 2 )
     1    call ilgsw2( Imaxxplts, Imaxyplts )
        if ( Imaxxplts .lt. nxplts .or. Imaxyplts .lt. nyplts ) then       2b
          if ( ierlun .gt. 0 )
     1      write( ierlun, 4 ) nxplts, nyplts, Imaxxplts, Imaxyplts
4         format( ' ILGPAGE : ABORT error : too many plotting cells - ',
     1      i5, ' by ', i5 / 11x, ' Maximum is ', i3, ' by ', i3 )
          if ( fwait ) call ilgwait( .false. )
          return
          end if                                                           2e
c
c             page origin and size limits
c
        Rpagzerx = Upagmarn( Ipagsiz )
        Rpagzery = Rpagzerx
        if ( Iorient .eq. 2 )
     1    Rpagzerx = Upagmarw( Ipagsiz )
        Rpagsizx = Upagsizw( Ipagsiz ) - Rpagzery * 2
        Rpagsizy = Upagsizh( Ipagsiz ) - Rpagzery - Upagmarw( Ipagsiz )
        if ( Iorient .eq. 2 )
     1    call ilgsw4( Rpagsizx, Rpagsizy )
        end if                                                             1e
c
c           plotting cell size
c
      Rcelsizy = Rpagsizy / Nyplts
      Rcelsizx = Rpagsizx / Nxplts
c
c           character heights
c
      rchrfac = ( nxplts * nyplts ) ** 0.25
      Rchrhi = min( max( Uminchr(Ipagsiz), Ulabchr(Ipagsiz)/rchrfac )
     1          * rchrsiz, min( rcelsizx, rcelsizy ) / 7.0 )
      Rchrti = min( max( Uminchr(Ipagsiz), Utitchr(Ipagsiz)/rchrfac )
     1          * rchrsiz, min( rcelsizx, rcelsizy ) / 7.0 )
      rsymhi = rchrhi * 0.6667 * rsymsiz / rchrsiz
c
c           graph area size & offset
c
      Rgrzerx = 3.5 * Rchrhi + 2.0 * Rpenth
      Rgrzery = Rgrzerx
      Rgrsizx = Rcelsizx - Rgrzerx - Rchrhi
      Rgrsizy = Rcelsizy - Rgrzerx - Rchrti - 2. * Rpenth - .5 * Rchrhi
      if ( .not. fwild ) then                                              1b
c
c           specify region of screen to be used
c
        scrh = min( Hv500h, Hv500h * 1.2 * Rpagsizy / Rpagsizx )
        scrw = min( Hv500w, Hv500w * 0.8 * Rpagsizx / Rpagsizy )
        call vwport( ( Hv500w - scrw ) / 2.0, ( Hv500w + scrw ) / 2.0,
     1               ( Hv500h - scrh ) / 2.0, ( Hv500h + scrh ) / 2.0 )
c
c           specify window -- cm to screen mapping
c
        xlo = Rpagzerx + rxspoff
        xhi = xlo + Rpagsizx
        ylo = Rpagzery + ryspoff
        yhi = ylo + Rpagsizy
        call window( xlo, xhi + 0.003, ylo, yhi + 0.003 )
        end if                                                            1e
c
c           put logo
c
      call vecmod
      call psecap
      if ( Corient .gt. 'aaaa' )
     1  call newpag
      call lintyp( 0 )
      call move( rpagzerx + rxspoff + rpenth,
     1           rpagzery + ryspoff + rpenth )
      call txtsiz( 3.0 * rchrhi - rpenth )
      call txtang( 0.0 )
      call greek( 1, 'A' )
c
c           draw border around page
c
      if ( .not. fwild ) then
        call newpen( -2 )
        call move( xlo, ylo )
        call draw( xlo, yhi )
        call draw( xhi, yhi )
        call draw( xhi, ylo )
        call draw( xlo, ylo )
        call newpen( -1 )
        end if
      call rstcap
      call anmode
c
      ixplts = nxplts
      iyplts = nyplts
      ixpos = 0
      iypos = 1
      rxofset = 0.0
      ryofset = 0.0
      fpage = .true.
      flimit = .false.
      fdata = .false.
      fxlab = .false.
      fylab = .false.
      ftitl = .false.
      fbordg = .false.
      fbordc = .false.
      fwild = .false.
      return
      end



      SUBROUTINE ILGRAPH( NXPOS, NYPOS, NXAXIS, NYAXIS, NBORDER )
c
c       Specifies the plotting cell to be used, axis types and border type .
c  Parameters :
c    NXPOS - horizontal position of the plotting cell to be used, from the left
c        Values :- 1 to NXPLTS
c    NYPOS - vertical position of the plotting cell to be used, from the bottom
c        Values :- 1 to NYPLTS
c    NXAXIS - specifies the type of x axis to be drawn
c        Values :- can be the sum of any of the following
c                  0 = not drawn
c                  1 = drawn
c                  2 = drawn and numbered
c                  4 = ticks on the axis are extended to form a grid
c                  8 = put small ticks between the numbered ticks
c    NYAXIS - specifies the type of y axis to be drawn
c        Values :- same as for NXAXIS
c    NBORDER - Specifies the type of border to be drawn around the graph
c        Values :- 0 = not drawn
c                  1 = around the scaled region
c                  2 = around the plotting cell
c
c    When a wild flag is active , the graph corners are specified by the
c  lower left and upper right coordinates in the wild registers . The values
c  of the parameters NXPOS and NYPOS are ignored , and set to NXPOS=1 and
c  NYPOS=NXPLTS to prevent problems with ILGDATA . Consequences of a wild
c  call to ILGRAPH can be resetby a normal call to ILGRAPH .
c
c variables
c  rgroffx, rgroffy            # graph lower left offset
c  logical *1 ilguuc           #          : User units conversion
c
      implicit none
      integer *2 nxpos, nypos, nxaxis, nyaxis, nborder
      logical *1 ilguuc
      include 'ilglib.cmn'
c
      ifrom = Hilgraph
      if ( .not. fopen )
     1  call ilgopen( ' ' )
      ifrom = Hilgnone
c
c           wild active
c
      if ( fwild ) then
        if ( fuser ) then
          if ( .not. ilguuc( 1 ) )
     1      return
          end if
        ixpos = 1
        iypos = iyplts
        rgrzerx = min( rwild(1), rwild(3) ) - rpagzerx - rxspoff
        rgrzery = min( rwild(2), rwild(4) ) - rpagzery - ryspoff
        if ( abs( rwild(3) - rwild(1) ) .le. 0.1 ) then
          Rgrsizx = Rcelsizx - Rgrzerx - Rchrhi
        else
          rgrsizx = abs( rwild(3) - rwild(1) )
          end if
        if ( abs( rwild(4) - rwild(2) ) .le. 0.1 ) then
          Rgrsizy = Rcelsizy - Rgrzerx - Rchrti - 2.0*Rpenth -0.5*Rchrhi
        else
          rgrsizy = abs( rwild(4) - rwild(2) )
          end if
c
c           conventional behaviour
c
      else
        if ( nxpos .le. 0 .or. nxpos .gt. ixplts ) then
          if ( ierlun .gt. 0 ) write( ierlun, 1) nxpos, ixplts
1         format( ' ILGRAPH : ABORT error : x plot position ', i5,
     1      ' out of bounds' / 11x, '1 <= nxpos <= ', i3 )
          if ( fwait ) call ilgwait( .false. )
          return
        else if ( nypos .le. 0 .or. nypos .gt. iyplts ) then
          if ( ierlun .gt. 0 ) write( ierlun, 2 ) nypos, iyplts
2         format( ' ILGRAPH : ABORT error : y plot position ', i5,
     1      ' out of bounds' / 11x, '1 <= nypos <= ', i3 )
          if ( fwait ) call ilgwait( .false. )
          return
          end if
c
c           graph area size & offset
c
        Rgrzerx = 3.5 * Rchrhi + 2.0 * Rpenth
        Rgrzery = Rgrzerx
        Rgrsizx = Rcelsizx - Rgrzerx - Rchrhi
        Rgrsizy = Rcelsizy - Rgrzerx - Rchrti - 2.0*Rpenth -0.5*Rchrhi
        ixpos = nxpos
        iypos = nypos
        end if
c
c         check axes & border
c
      if ( nxaxis .lt. 0 .or. nxaxis .gt. 15 ) then
        if ( ierlun .gt. 0 ) write( ierlun, 3 ) nxaxis
3       format( ' ILGRAPH : ABORT error : illegal x axis selection ',i3
     1    / 11x, '0 <= nxaxis <= 7' )
        if ( fwait ) call ilgwait( .false. )
        return
      else if ( nyaxis .lt. 0 .or. nyaxis .gt. 15 ) then
        if ( ierlun .gt. 0 ) write( ierlun, 4 ) nyaxis
4       format( ' ILGRAPH : ABORT error : illegal y axis selection ',i3
     1    / 11x, '0 <= nyaxis <= 7' )
        if ( fwait ) call ilgwait( .false. )
        return
      else if ( nborder .lt. 0 .or. nborder .gt. 3 ) then
        if ( ierlun .gt. 0 ) write( ierlun, 5 ) nborder
5       format( ' ILGRAPH : ABORT error : illegal border selection ',i3
     1    / 11x, '0 <= nborder <= 3' )
        if ( fwait ) call ilgwait( .false. )
        return
        end if
c
c           set values
c
      ixaxis = nxaxis
      iyaxis = nyaxis
      iborder = nborder
      fgraph = .true.
      flimit = .false.
      fdata  = .false.
      fxlab  = .false.
      fylab  = .false.
      ftitl  = .false.
      fbordg = .false.
      fbordc = .false.
      fwild = .false.
      return
      end



      SUBROUTINE ILGLINE( NSYMB, NDASH, NPEN )
c
c       Specify symbol type to be placed at each data point , dashed line type
c joining data points and pen number
c
c  Parameters :-
c    NSYMB - Specifies the special symbol to be plotted at each data point .
c       Values :- 0 <= NSYMB <= 15 . The symbol is that specified by NSYMB-1
c                 in a call to SYMBOL in the H/P ISPP User's Manual .
c                 NSYMB = 0  is a dot .
c    NDASH - Dashed line type for the line joining data points .
c       Values :- 0 = no line
c                 1 = continuous line
c                 2 to 8 = line type specified by NDASH-2 in a call to BUFF
c                          in the H/P ISPP User's Manual
c    NPEN - Pen number . The colour and thickness depend on what pens are
c           loaded into the plotter . Screen graphics ignores this argument .
c       Values :-  0 <= NPEN <= 8 . NPEN = 0 is no pen
c
c     When in `wild mode', the contents of the wild registers sets default
c pen values :--
c   1) wild register 1 = axes and grids      The only acceptable values
c   2) wild register 2 = border              are 1.0 to less than 9.0 .
c   3) wild register 3 = labels & titles     The initial defaults are 1 .
c   4) wild register 4 = notes               The UNITS must be CM .
c
      implicit none
      integer *2 nsymb, ndash, npen, i
      include 'ilglib.cmn'
c
      ifrom = Hilgline
      if ( .not. fopen )
     1  call ilgopen( ' ' )
      ifrom = Hilgnone
c
c                 wild mode - set defaults
c
      if ( fwild ) then
        do 4 i = 1, 4
4         idefpen(i) = int( max( 1.0, min( 8.99999, rwild(i) ) ) )
        end if
c
c                 check arguments
c
      if ( 0 .gt. nsymb .or. nsymb .gt. 15 ) then
        if ( ierlun .gt. 0 ) write( ierlun, 1 ) nsymb
1       format( ' ILGLINE : ABORT ERROR : Invalid symbol type ', i6 /
     1    11x, '0 = dot, 1 to 15 = H/P ISPP special symbol + 1' )
        if ( fwait ) call ilgwait( .false. )
        return
      else if ( 0 .gt. ndash .or. ndash .gt. 7 ) then
        if ( ierlun .gt. 0 ) write( ierlun, 2 ) ndash
2       format( ' ILGLINE : ABORT ERROR : Invalid dashed line type ',
     1    i6 / 11x, '0 = no line, 1 to 7 = H/P ISPP line type + 2' )
        if ( fwait ) call ilgwait( .false. )
        return
      else if ( 0 .gt. npen .or. npen .gt. 8 ) then
        if ( ierlun .gt. 0 ) write( ierlun, 3 ) npen
3       format( ' ILGLINE : ABORT ERROR : Invalid pen number ',i6 /
     1    11x, '0 = no pen. Must have  0 <= npen <= 8' )
        if ( fwait ) call ilgwait( .false. )
        return
        end if
      isymb = nsymb
      idash = ndash
      ipen  = npen
      fline = .true.
      fwild = .false.
      return
      end



      SUBROUTINE ILGSLAW( CXSLAW, CYSLAW )
c
c       Specify the scaling law type - only two are available :-
c  linear & logarithmic
c
c  Parameters :
c    CXSLAW - Scaling law for abicissa .
c        Values :- 'LIN', 'lin', 'LOG', 'log'  - for linear and logarithmic
c    CYSLAW - Scaling law for ordinate
c        Values :- 'LIN', 'lin', 'LOG', 'log', - for linear and logarithmic
c                  'RAD', 'rad', 'DEG', 'deg', - for polar plots
c
c  N.B. The offset ( see ILGFSET ) is reset to its default value .
c
c variables
c  integer *2 ix, iy           # where is CYSLAW in GLAWSEL ?
c  logical *1 ilgins           # function : "IN String"
c  Glawsel                     # legal scaling law selection
c
      implicit none
      character Glawsel*30
c
      integer *2 ix, iy
      logical *1 ilgins
      character *(*) cxslaw, cyslaw
      include 'ilglib.cmn'
c
      Glawsel = 'LOGlogLINlinANGangRADradDEGdeg'
      if ( .not. fopen )
     1  call ilgopen( ' ' )
      if ( .not. ilgins( cxslaw, 3, Glawsel ) ) then
        if ( ierlun .gt. 0 ) write( ierlun, 1 ) cxslaw
1       format( ' ILGSLAW : ABORT ERROR : Invalid X scaling law ', a /
     1    11x, '- must be `LIN'', `LOG'', `ANG'', `RAD'' or `DEG''' )
        if ( fwait ) call ilgwait( .false. )
        return
      else if ( .not. ilgins( cyslaw, 3, Glawsel(1:18) ) ) then
        if ( ierlun .gt. 0 ) write( ierlun, 2 ) cyslaw
2       format( ' ILGSLAW : ABORT ERROR : Invalid Y scaling law ', a /
     1    11x, '- must be `LIN'', `LOG'' or `ANG''' )
        if ( fwait ) call ilgwait( .false. )
        return
        end if
      rxofset = 0.0
      ryofset = 0.0
      ix = index( Glawsel, cxslaw )
      flogx = ix .lt. 6
      fpolar = ix .gt. 18
      fangx = 12 .lt. ix .and. ix .lt. 18
      iy = index( Glawsel, cyslaw )
      flogy = iy .lt. 6
      fangy = 12 .lt. iy .and. iy .lt. 18
      if ( iy .lt. 24 ) then
        rdatfacx = 1.0 / 0.017453293
      else
        rdatfacx = 1.0
        end if
      flimit = .false.
      fslaw = .true.
      fwild = .false.
      return
      end



      SUBROUTINE ILGIMIT( XMIN, YMIN, XMAX, YMAX )
c
c       Select lower left and upper right scaled graph coordinates
c given the limits of the data  (XMIN,YMIN) and (XMAX,YMAX)
c
c  Parameters :-
c    XMIN - minimum abicissa (ie x) value
c    XMAX - maximum abicissa (ie x) value
c    YMIN - minimum ordinate (ie y) value
c    YMAX - maximum ordinate (ie y) value
c
c  Function :-
c     ILGIMIT selects 'nice' values of coordinate extrema, so that the
c  numbering of axes is 'sensible and reasonable' . It also selects the
c  tick increment along each axis and calculates the mapping constants used
c  by ILGDATA, ILGWILD, ILGNOTE & ILGREAD
c
c     If in `wild mode', the lower left and the upper right data limits
c  are taken from the wild registers . These values are NOT adjusted to
c  the `nice' values as is done by the conventional use of this routine.
c  NOTE that the previously defined `user defined' coordinate system is
c  used when is `wild mode' . The values of the arguments XMIN, YMIN,
c  XMAX and YMAX are ignored .
c
      implicit none
      logical *1 ilguuc
      real xmin, ymin, xmax, ymax, raxmax, raymax
      include 'ilglib.cmn'
c
      if ( .not. fopen )
     1  call ilgopen( ' ' )
c*
c*        wild mode
c*
      if ( fwild ) then
c
c         if not user coordinates, convert to them
c
        if ( .not. fuser ) then
          if ( .not. ilguuc( -1 ) )
     1      return
          end if
c
c         abicissa
c
        raxmin = min( rwild(1), rwild(3) )
        rxmin = raxmin
        raxmax = max( rwild(1), rwild(3) )
        rxmax = raxmax
        call ilglm2( raxmin, raxmax, raxincx, rsmticx, flogx, fangx )
        if ( flogx ) then                                                 1b
          rxmin = alog10( rxmin )
          rxmax = alog10( rxmax )
          end if                                                          1e
c
c         ordinate
c
        raymin = min( rwild(2), rwild(4) )
        rymin = raymin
        raymax = max( rwild(2), rwild(4) )
        rymax = raymax
        call ilglm2( raymin, raymax, raxincy, rsmticy, flogy, fangy )
        if ( flogy ) then                                                 1b
          rymin = alog10( rymin )
          rymax = alog10( rymax )
          end if                                                          1e
c*
c*          conventional behaviour
c*
      else
c
c           abicissa
c
        rxmin = xmin
        rxmax = xmax
        if ( rxmin .gt. rxmax ) call ilgsw4( rxmin, rxmax )
        call ilglm2( rxmin, rxmax, raxincx, rsmticx, flogx, fangx )
        raxmin = rxmin
c
c           ordinate
c
        rymin = ymin
        rymax = ymax
        if ( rymin .gt. rymax ) call ilgsw4( rymin, rymax )
        call ilglm2( rymin, rymax, raxincy, rsmticy, flogy, fangy )
        raymin = rymin
        end if
c*
c*          calculate scaling factors
c*
      rdatfacx = rgrsizx / ( rxmax - rxmin )
      rdatfacy = rgrsizy / ( rymax - rymin )
      flimit = .true.
      fxlab = .false.
      fylab = .false.
      ftitl = .false.
      fwild = .false.
      return
      end



      SUBROUTINE ILGLM1( RANGE, RAXINC, RSMTIC, FANG, FLOGLAW )
c
c       Determine the decade scale of the range of ordinates , and determine
c a suitable axis labelling increment - for linear scaling law .
c
c arguments
c  real range,                 # range of ordinates
c       raxinc,                # axis labelling increment
c       rsmtic                 # initial distance between small ticks on axes
c  logical *1 fang,            # set axis increment as for degrees of arc
c             floglaw          # .true. if log scaling is used
c constants
c  real rscinc(10)             # axis scaling increments
c variables
c  real rdecsc                 # the decade scale
c  integer *1 ilgint           # function : integer part of - round down
c
      implicit none
      logical *1 fang, floglaw
      integer *2 ilgint
      real range, rdecsc, raxinc, rsmtic, rscinc(10), raninc(8)
      data rscinc / 2* 0.5, 3* 1.0, 5* 2.0 /
      data raninc / 15.0, 30.0, 45.0, 4*90.0, 180.0 /
c
      include 'ilglib.cmn'
c
c           increments for ANG or DEG scaling
c
      if ( ( fang .and. range .le. 720 .or. fpolar .and.
     1     rdatfacx .eq. 1.0 ) .and. range .gt. 45 ) then
        raxinc = raninc( min( int( range / 60.0 + 0.99999 ), 7 ) )
        if ( fpolar .and. raxinc .gt. 45.0 )
     1    raxinc = 45.0
        rsmtic = min( raxinc / 3.0, 45.0 )
c
c           increments for LIN and LOG scaling
c
      else
        rdecsc = 10.0 ** ilgint( log10( range ) )
        raxinc = rscinc( int( range / rdecsc + 0.00001 ) )
        if ( floglaw ) then
          rsmtic = rdecsc
        else if ( raxinc .le. 1 ) then
          rsmtic = raxinc * rdecsc / 5.0
        else
          rsmtic = raxinc * rdecsc / 4.0
          end if
        raxinc = raxinc * rdecsc
        end if
      return
      end



      SUBROUTINE ILGLM2( RMIN, RMAX, RAXINC, RSMTIC, FLOGLAW, FANG )
c
c       Determine suitable graph scaling limits for a linear scaling law .
c
c arguments
c  real rmin,                  # lower plotting limit
c       rmax,                  # upper plotting limit
c       raxinc,                # axis labelling increment
c       rsmtic                 # initial space between small ticks on the axes
c  logical *1 floglaw,         # .true. if log scaling is used
c             fang             # number axis as angle
c variables
c  real t                      # temporary
c
      implicit none
      real rmin, rmax, raxinc, rsmtic, range, t
      logical *1 floglaw, fang
      include 'ilglib.cmn'
c
c
      if ( rmin .gt. rmax )
     1  call ilgsw4( rmin, rmax )
c
c             for log scaling law
c
      if ( floglaw ) then
        if ( rmin .le. 0.0 ) then
          if ( ierlun .gt. 0 ) write( ierlun, 1 ) rmin
1         format( ' ILGIMIT : =>LIN ERROR : Negative value on log scal'
     1      , 'ed axis ',g14.7 / 11x, 'All values must be positive .' )
          if ( fwait ) call ilgwait( .false. )
          floglaw = .false.
        else
          rmin = log10( rmin )
          rmax = log10( rmax )
          end if
        end if
c
c             calculate range
c
      if ( rmax .le. rmin ) then
        if ( rmax .eq. 0.0 ) then
          rmin = -0.5
          rmax = 0.5
        else
          rmax = 1.5 * rmax
          rmin = rmin * 0.667
          if ( rmin .gt. rmax )
     1      call ilgsw4( rmin, rmax )
          end if
        end if
      range = rmax - rmin
      call ilglm1( range, raxinc, rsmtic, fang, floglaw )
c
c             calculate lower plotting limit
c
      t = int( rmin / raxinc ) * raxinc
      if ( rmin .lt. t - range / 1000.0 )
     1  t = t - raxinc
      rmin = t
c
c             calculate upper plotting limit
c
      t = int( rmax / raxinc ) * raxinc
      if ( rmax .gt. t + range / 1000.0 )
     1  t = t + raxinc
      rmax = t
c
c     range = rmax - rmin
c     call ilglm1( range, raxinc, rsmtic, fang, floglaw  )
      return
      end


      SUBROUTINE ILGCLIP( FCLIP )
c
c     Switch clipping at the graph boundary on or off .
c  With clipping off , data and notes can be placed anywhere in the
c  current plotting cell .
c Parameters
c  logical *1 fclip    # .true. iff  clipping is on
c
c     When in `wild mode' , a flag is set which forces the
c  ILGDATA or ILGNOTE routine to use the lower left and upper right
c  coordinates in the wild registers as clipping limits . This flag
c  reset after use . NOTE that the clipping limits are taken from
c  the wild registers WHEN ILGDATA OR ILGNOTE IS EXECUTED . The value
c  of the argument FCLIP is ignored .
c
      implicit none
      logical *1 fclip
      include 'ilglib.cmn'
c
      if ( .not. fopen )
     1  call ilgopen( ' ' )
      if ( .not. fwild ) then
        fklip = fclip
        fwklip = .false.
      else
        fwklip = fwild
        fklip = .false.
        rwild(5) = min( rwild(1), rwild(3) )
        rwild(6) = min( rwild(2), rwild(4) )
        rwild(7) = max( rwild(1), rwild(3) )
        rwild(8) = max( rwild(2), rwild(4) )
        end if
      fwild = .false.
      return
      end


      SUBROUTINE ILGFSET( XOFFSET, YOFFSET )
c
c         coordinate oFfSET
c     If the axes are scaled linearly , add the offset ( XOFFSET, YOFFSET )
c  in user defined units to any coordinates plotted after the call to this
c  routine . The default offset is ( 0.0, 0.0 ). Each call to ILGOPEN and
c  ILGPAGE resets the offset to its default value .
c     If any axis is scaled logarithmically , the offset is a gain factor,
c  with the default gain being 1.0 .
c     NOTE : a call to ILGSLAW wil reset the offset to its default value.
c
c Parameters
c  real xoffset, yoffset       # offset in user defined units to be added
c
c  If in `wild mode', the offset is taken to be the displacement of the
c  upper right ( 2nd ) coordinate in the wild registers relative to the
c  position of the lower left ( 1st ) coordinate . The values of the
c  arguments XOFFSET and YOFFSET are ignored .
c
      implicit none
      integer *2 Hvdulu
      parameter( Hvdulu=6 )
c
      real xoffset, yoffset
      logical *1 ilguuc
      include 'ilglib.cmn'
c
      if ( .not. fopen )
     1  call ilgopen( ' ' )
      if ( fwild ) then
        if ( fuser ) then
          if ( .not. ilguuc( 1 ) )
     1      return
          end if
        rxofset = ( rwild(3) - rwild(1) ) / rdatfacx
        ryofset = ( rwild(4) - rwild(2) ) / rdatfacy
      else
        if ( flogx .and. xoffset .le. 0 .or.
     1      flogy .and. yoffset.le.0 ) then
          write( ierlun, 1 )
1         format( ' ILGFSET : ABORT : Negative gain not allowed' )
          return
          end if
        if ( flogx ) then
          rxofset = log10( xoffset )
        else
          rxofset = xoffset
          end if
        if ( flogy ) then
          ryofset = log10( yoffset )
        else
          ryofset = yoffset
          end if
        end if
      fwild = .false.
      return
      end


      SUBROUTINE ILGWILD( CUNITS, XMIN, YMIN, XMAX, YMAX )
c
c         Wild card
c       Set limits or sizes to arbitrary values .
c  Moves the lower left coordinate (XMIN,YMIN) and the upper right coordinate
c  (XMAX,YMAX) into the wild registers , and then set the `wild active' flag.
c  The next call to ANY other ILG#### routine will reset this flag , ie
c  deactivate it . If that other routine is ILGCHAR, ILGPAGE, ILGRAPH, ILGIMIT,
c  ILGFSET, ILGCLIP or ILGLINE, the contents of the wild registers are used to
c  set the appropriate limits for that routine . Consult the documentation for
c  these routines for details of the action taken .
c    NOTE that this is a fragile routine , and the results might not be what
c  you expect .
c
c Parameters
c  character *2 Cunits         # units of the given coordinates
c  real xmin, ymin             # lower left coordinate
c       xmax, ymax             # upper right coordinate
c
c  The choice of units is :-
c    'CM' or 'cm' -    centimetres     this is the default
c    'MM' or 'mm' -    millimetres
c    'M ' or 'm ' -    metres
c    'IN' or 'in' -    inches
c    'FT' or 'ft' -    feet
c    'UU' or 'uu' -    user defined units
c    'XX' or 'xx' -    coordinate parameters are ignored and the limit
c                      coordinates are taken from the crosshairs .
c    'ZZ' or 'zz' -    same as 'XX', but the resulting limit coordinates
c                      are returned in 'CM' through the argument list .
c
c    NOTE :  The conventional behaviour of ILGLIB is quite robust, but if
c    `wild mode' is invoked , ILGLIB can become fragile and MAY produce
c    messy results . Plotting is an inherently complicated process , and
c    there are many implicit rules . By using ILGWILD , you can break some
c    of these rules , but YOU must take the responsibility for the
c    consequences ; so BE CAREFUL ! In most cases , ILGLIB can do whatever
c    you want without using ILGWILD . However , adventurous users will find
c    that ILGWILD can introduce considerable flexibility and freedom to
c    their plotting . Only ILGCHAR, ILGPAGE, ILGRAPH, ILGFSET, ILGCLIP, ILGLINE
c    and ILGIMIT can be executed in `wild mode' .
c
      implicit none
      integer *2 Hvdulu
      parameter( Hvdulu= 6 )
c
      character Cunits*(*), chr*1
      integer *2 i, j
      real xmin, ymin, xmax, ymax, rcvf
      include 'ilglib.cmn'
c
      if ( .not. fopen )
     1  call ilgopen( ' ' )
      fuser = .false.
c
c           use cross hairs
c
      if ( Cunits .eq. 'XX' .or. Cunits .eq. 'xx' .or.
     1     Cunits .eq. 'ZZ' .or. Cunits .eq. 'zz' ) then                  1b
c
c           get coordinates
c
        call vecmod
        do 3 j = 0, 2, 2                                                  2b
c         start of loop                                                   3b
1           call locate( 1, rwild(j+1), rwild(j+2), chr, i )
            if ( chr .ne. 'X' .and. chr .ne. 'x' .and. i .gt. 0 )
     1        goto 3
            write( Hvdulu, 2 ) 7
2           format( '+', a )
c           end of loop                                                   3e
3         continue                                                        2e
        call anmode
c
        if ( Cunits .eq. 'ZZ' .or. Cunits .eq. 'zz' ) then                2b
          xmin = rwild(1)
          ymin = rwild(2)
          xmax = rwild(3)
          ymax = rwild(4)
          end if                                                          2e
c
c         user defined coordinate system
c
      else if ( Cunits .eq. 'UU' .or. Cunits .eq. 'uu' ) then             1m
        if ( flogx .and. ( xmin .le. 0 .or. xmax .le. 0 ) .or.
     1       flogy .and. ( ymin .le. 0 .or. ymax .le. 0 ) ) then
          if ( ierlun .gt. 0 )
     1      write( ierlun, 4 )
4           format( ' ILGWILD : ABORT : Negative value on log axis' )
          if ( fwait ) call ilgwait( .false. )
          return
          end if
        fuser = .true.
        rwild(1) = xmin
        rwild(2) = ymin
        rwild(3) = xmax
        rwild(4) = ymax
c
c           other units systems
c
      else                                                                1m
        if ( Cunits .eq. 'MM' .or. Cunits .eq. 'mm' ) then                2b
          rcvf = 0.1
        else if ( Cunits .eq. 'IN' .or. Cunits .eq. 'in' ) then           2m
          rcvf = 2.54
        else if ( Cunits .eq. 'FT' .or. Cunits .eq. 'ft' ) then           2m
          rcvf = 4.724409
        else if ( Cunits .eq. 'M ' .or. Cunits .eq. 'm ' ) then           2m
          rcvf = 100.0
        else                                                              2m
          rcvf = 1.0
          end if                                                          2e
        rwild(1) = xmin * rcvf
        rwild(2) = ymin * rcvf
        rwild(3) = xmax * rcvf
        rwild(4) = ymax * rcvf
        end if                                                            1e
      fwild = .true.
      return
      end


      SUBROUTINE ILGDATA( X, Y, NINCR, NPTS, NEW )
c
c       Plots the (x,y) data in the next or current plotting cell .
c
c  Parameters :-
c    X - Array of abicissa values
c    Y - Array of ordinate values
c    NINCR - Points to be plotted increment . Ie. starting with ( x(1), y(1) ),
c            every NINCRth point is plotted .
c       Values :- greater than zero
c    NPTS - Number of data points to be plotted .
c    NEW  - Plot the data in the same or new plotting cell .
c       Values 0 =  Use the same plotting cell as the previous call to ILGDATA
c              non-zero  = Use a new plotting cell
c N.B. The X and Y arrays must each have at least  ( NPTS - 1 ) * NINCR + 1
c      elements
c
c constants
c  Hradcon=57.2957795          # radians to degrees conversion factor
c variables
c  logical *1 fnewgr           # new graph cell required flag
c             foutc(4)         # = iouttc
c             ferrbar          # error bars are to be drawn
c  integer *2 i                # data array index
c             incr             # actual data indexing increment
c             ierr             # no. of attempts to plot -ve value on log axis
c             ioutoc, ioutn    # clipping sectors for old & new points -iterated
c             iouttc, ioutnr   # clipping sectors for old & new points - kept
c             ilgout           # function : find clipping sector
c  real xmin, ymin             # minima of x & y data to be plotted
c       xmax, ymax             # maxima of x & y data to be plotted
c       rxtemp, rytemp         # temporary (clipped) end point
c       rxold, ryold           # previous coordinate
c       rxbase, rybase         # base coordinate in cm
c       rxoldc, ryoldc         # clipped previous coordinate
c       reslen                 # length resolution for polar plots
c       rsign                  # sign of the angle change ( -1.0, 0.0, 1.0 )
c       rslope, rofset         # slope & offset for linear interpolation in
c                              # polar coordinates
c
      implicit none
      real Hradcon
      parameter( Hradcon=57.2957795 )
c
      real rxtemp, rytemp, rxold, ryold, x(*), xmax, xmin,
     1  y(*), ymax, ymin, rxbase, rybase, rxoldc, ryoldc,
     2  reslen, rsign, rslope, rofset
      logical *1 fnewgr, foutc(4), ferrbar
      integer *2 new, i, nincr, npts, ierr, incr
      integer *4 ioutoc, iouto, ioutn, ioutnr, iouttc
c
      include 'ilglib.cmn'
      equivalence ( foutc(1), iouttc )
c
c           set increment
c
      if ( nincr .eq. 0 )
     1  return
      incr = abs( nincr )
      ferrbar = nincr .lt. 0
c
c           calculate limits
c
      if ( .not. fopen )
     1  call ilgopen( ' ' )
      fnewgr = .not. ( flimit .or. fdata ) .or. new .ne. 0
      if ( fnewgr ) then
        ymin = y(1)
        ymax = ymin
        xmin = x(1)
        xmax = xmin
        do 10 i = incr + 1, incr * ( npts - 1 ) + 1, incr
          ymin = amin1( ymin, y(i) )
          ymax = amax1( ymax, y(i) )
          xmin = amin1( xmin, x(i) )
          xmax = amax1( xmax, x(i) )
10        continue
        call ilgimit( xmin, ymin, xmax, ymax )
        end if
      call vecmod
      if ( ( fnewgr .or. flimit ) .and. .not. fgraph ) then
c
c           increment plotting cell
c
        ixpos = ixpos + 1
        if ( ixpos .gt. ixplts ) then
          ixpos = 1
          iypos = iypos + 1
          if ( iypos .gt. iyplts ) then
            iypos = 1
            if ( fcapt ) then
              call clocap
              fcapt = .false.
              write( ierlun, 1 )
1               format( ' ILGDATA : WARNING : Page full : closing',
     1            '  capture file .' )
              end if
            call newpag
            end if
          end if
        fxlab = .false.
        fylab = .false.
        ftitl = .false.
        fbordg = .false.
        fbordc = .false.
        end if
c
c           set cell origin and overall graph offset
c
      ierr = 0
      fgraph = .false.
      if ( finvis )
     1  call invis
      call ilgrof
      call newpen( ipen )
      if ( idash .gt. 1 )
     1  call newpen( -idash )
c
c           set graph plotting limits
c
      call ilgclp
      if ( fpolar ) then
        rxbase = 0.0
        rybase = 0.0
        if ( flogy )
     1    rybase = ( ryofset - rymin ) * rdatfacy
      else
        rxbase = rgroffx + ( rxofset - rxmin ) * rdatfacx
        rybase = rgroffy + ( ryofset - rymin ) * rdatfacy
        end if
c
c           plot data points
c
      if ( idash .ne. 0 ) then                                           1b
        if ( idash .gt. 1 )
     1    call newpen( -idash )
        reslen = rchrhi * 0.7 * Hradcon
        do 13 i = 1, ( npts - 1 ) * incr + 1, incr                       2b
          if ( x(i) .le. 0.0 .and. flogx .or. y(i) .le. 0.0 .and. flogy
     1        ) then                                                     3b
            ierr = ierr + 1
          else                                                           3m
c
c           convert data coordinate to cm
c
            rxnew = x(i)
            if ( flogx )
     1        rxnew = log10( rxnew )
            rxnew = rxnew * rdatfacx + rxbase
c
            rynew = y(i)
            if ( flogy )
     1        rynew = log10( rynew )
            rynew = rynew * rdatfacy + rybase
c
c           COHEN-SUTHERLAND CLIPPING ALGORITHM
c             initialize
c
            rxoldc = rxold
            ryoldc = ryold
            rxold = rxnew
            ryold = rynew
            ioutoc = iouto
            call ilgout( ioutn, rxnew, rynew )
            ioutnr = ioutn
c
c               for the first point
c
            if ( i .eq. 1 ) then                                         4b
              if ( ioutnr .eq. 0 ) then                                  5b
                if ( fpolar ) then                                       6b
                  call move( rxpole + rynew * cos( rxnew / Hradcon ),
     1                       rypole + rynew * sin( rxnew / Hradcon ) )
                else                                                     6m
                  call move( rxnew, rynew )
                  end if                                                 6e
                end if                                                   5e
            else                                                         4m
c
c               for the second and following points
c
c           iterate the clipped segments
c
11            if ( ioutoc .ne. 0 .or. ioutn .ne. 0 ) then                5b
                if ( iand( ioutoc, ioutn ) .ne. 0 ) goto 3
                iouttc = ioutoc
                if ( ioutoc .eq. 0 )
     1            iouttc = ioutn
c
c                 interpolate to the clip limits
c
c   crosses left edge
                if ( foutc(1) ) then                                     6b
                  rxtemp = rxllim
                  rytemp = ( rxllim - rxnew ) * ( rynew - ryoldc ) /
     1                     ( rxnew - rxoldc ) + rynew
c   crosses right edge
                else if ( foutc(2) ) then                                6m
                  rxtemp = rxulim
                  rytemp = ( rxulim - rxnew ) * ( rynew - ryoldc ) /
     1                     ( rxnew - rxoldc ) + rynew
c   crosses bottom edge
                else if ( foutc(3) ) then
                  rytemp = ryllim
                  rxtemp = ( ryllim - rynew ) * ( rxnew - rxoldc ) /
     1                     ( rynew - ryoldc ) + rxnew
c   crosses top edge
                else if ( foutc(4) ) then
                  rytemp = ryulim
                  rxtemp = ( ryulim - rynew ) * ( rxnew - rxoldc ) /
     1                     ( rynew - ryoldc ) + rxnew
                  end if                                                 6e
                if ( iouttc .eq. ioutn ) then                            6b
                  rxnew = rxtemp
                  rynew = rytemp
                  call ilgout( ioutn, rxnew, rynew )
                else                                                     6m
                  rxoldc = rxtemp
                  ryoldc = rytemp
                  call ilgout( ioutoc, rxoldc, ryoldc )
                  end if                                                 6e
                goto 11
                end if                                                   5e
c
c             plot points in polar coordinates
c
              if ( fpolar ) then                                         5b
                if ( iouto .ne. 0 )
     1            call move( rxpole + ryoldc * cos( rxoldc / Hradcon ),
     2              rypole + ryoldc * sin( rxoldc / Hradcon ) )
c
c             decide if interpolation is required
c
                xmin = abs( rxnew - rxoldc )
                if ( xmin .gt. 1.0 .and. xmin * abs( max( rynew,
     1              ryoldc ) ) .gt. reslen ) then                        6b
c
c             do interpolation
c
                  rslope = ( rynew - ryoldc ) / ( rxnew - rxoldc )
                  rofset = rynew - rslope * rxnew
                  rsign = sign( 1.0, rxnew - rxoldc )
                  rxtemp = rxoldc
                  rytemp = ryoldc
                  xmax = ( rxnew - rxoldc ) * rsign
12                if ( ( rxtemp - rxoldc ) * rsign .le. xmax ) then      7b
                    rxtemp = rxtemp + rsign *
     1                min( reslen / rytemp, abs( rxnew - rxtemp ) )
                    rytemp = rslope * rxtemp + rofset
                    call draw( rxpole + rytemp * cos( rxtemp /Hradcon ),
     1                rypole + rytemp * sin( rxtemp / Hradcon ) )
                    goto 12
                    end if                                               7e
                else                                                     6m
c
c             don't do interpolation
c
                  call draw( rxpole + rynew * cos( rxnew / Hradcon ),
     1                       rypole + rynew * sin( rxnew / Hradcon ) )
                end if                                                   6e
c
c           plot points in Cartesian coordinates
c
              else                                                       5m
                if ( iouto .ne. 0 )
     1            call move( rxoldc, ryoldc )
                call draw( rxnew, rynew )
                end if                                                   5e
              end if                                                     4e
3           iouto = ioutnr
            end if                                                       3e
13        continue                                                       2e
        if ( idash .gt. 1 )
     1    call newpen( -1 )
        end if                                                           1e
c
c           plot special centred symbols
c
      if ( isymb .gt. 0 .or. isymb .eq. 0 .and. idash .eq. 0 .or.
     1      ferrbar ) then                                               1b
        do 14 i = 1, ( npts - 1 ) * incr + 1, incr                       2b
          if ( .not. ( x(i) .le. 0.0 .and. flogx .or. y(i) .le. 0.0
     1        .and. flogy ) ) then                                       3b
c
c             convert data coordinate to cm
c
            rxnew = x(i)
            if ( flogx )
     1        rxnew = log10( rxnew )
            rxnew = rxnew * rdatfacx + rxbase
c
            rynew = y(i)
            if ( flogy )
     1        rynew = log10( rynew )
            rynew = rynew * rdatfacy + rybase
c
c             if the point is on the graph
c
            if ( rxllim .le. rxnew .and. rxnew .le. rxulim .and.
     1          ryllim .le. rynew .and. rynew .le. ryulim ) then         4b
              if ( fpolar ) then                                         5b
                rxtemp = rxpole + rynew * cos( rxnew / Hradcon )
                rynew  = rypole + rynew * sin( rynew / Hradcon )
                rxnew = rxtemp
                end if                                                   5e
              if ( isymb .gt. 0 )
     1          call symbol( rxnew, rynew, rsymhi, isymb-1, 0.0, -1 )
              if ( ferrbar ) then                                        5b
                call move( rxnew, rynew )
                call ilgerb( x(i+1), y(i+1) )
              else                                                       5m
                call plot( rxnew, rynew, 3 )
                call plot( rxnew, rynew, 2 )
                end if                                                   5e
              end if                                                     4e
            end if                                                       3e
14        continue                                                       2e
        end if                                                           1e
c
c           plot axes
c
      if ( fnewgr .or. flimit ) then                                     1b
        if ( ixaxis .ne. 0 ) then                                        2b
c   peripheral axis
          if ( fpolar ) then                                             3b
            call ilgaxs( 0.0, 1, 4 )
c   x axis
          else                                                           3m
            call ilgaxs( 0.0, -1, 1 )
            end if                                                       3e
          end if                                                         2e
        if ( iyaxis .ne. 0 ) then                                        2b
c   radial axis
          if ( fpolar ) then                                             3b
            rxtemp = min( rxmax, max( rxmin, 0.0 ) )
            call ilgaxs( rxtemp, -1, 3 )
c   y axis
          else                                                           3m
            call ilgaxs( 90.0, 1, 2 )
            end if                                                       3e
          end if                                                         2e
c
c           draw borders if required
c
        if ( iborder .gt. 0 ) call ilgbor
        end if                                                           1e
c
      if ( finvis ) then                                                 1b
        call visibl
        finvis = .false.
        end if                                                           1e
      call anmode
      if ( ierr .gt. 0 ) then
        if ( ierlun .gt. 0 ) write( ierlun, 2 ) ierr
2       format( ' ILGDATA : WARNING : Attempted to plot ', i6,
     1    ' negative values on log axes' )
        if ( fwait ) call ilgwait( .false. )
        end if
      flimit = .false.
      fdata = .true.
      fwild = .false.
      return
      end


      SUBROUTINE ILGERB( DELX, DELY )
c
c       Draw an error bar of size (DELX,DELY) at the current pen position
c  in user units . For linear scaling, the size of the error bar can only
c  have magnitude . For log scaling, the size of the error bar is expressed
c  as a fraction of the data value . Note that scaling of the graph does
c  not include an allowance for the size of the error bars .
c
c Arguments
c  real delx, rdely            # size of error bar in user units
c
      implicit none
      real delx, dely, rxsiz, rysiz
      include 'ilglib.cmn'
c
      if ( flogx ) then                                                  1b
        rxsiz = abs( delx / 2.0 )
        rxsiz = 2.0 * rdatfacx *
     1          log10( rxsiz + sqrt( 1 + rxsiz **2 ) )
      else                                                               1m
        rxsiz = rdatfacx * abs( delx )
        end if                                                           1e
      if ( flogy ) then                                                  1b
        rysiz = abs( dely / 2.0 )
        rysiz = 2.0 * rdatfacy *
     1      log10( rysiz + sqrt( 1 + rysiz **2 ) )
      else                                                               1m
        rysiz = rdatfacy * abs( dely )
        end if                                                           1e
      call txtang( 0.0 )
      call errbar( rxsiz, rysiz, rchrhi / 3.0 )
      return
      end


      SUBROUTINE ILGOUT( FOUT, X, Y )
c
c    Determine if the point (X,Y) is within the current plotting
c    boundaries (RXLLIM,RYLLIM) and (RXULIM,RYULIM) .
c Arguments
c  logical *1 fout      # crosses left, right, bottom & top borders
c  real  x, y           # coordinate to be tested
c
      implicit none
      logical *1 fout(4)
      real x, y
      include 'ilglib.cmn'
c
c                               crosses left edge
      fout(1) = rxllim .gt. x
c                               crosses right edge
      fout(2) = rxulim .lt. x
c                               crosses bottom edge
      fout(3) = ryllim .gt. y
c                               crosses top edge
      fout(4) = ryulim .lt. y
      return
      end


      SUBROUTINE ILGBOR
c
c         Draw a border around the current plotting cell or graph according
c  to the call to ilgraph .
c
      implicit none
      real rxcelmax, rycelmax
      include 'ilglib.cmn'
c
c           border around graph
c
      if ( .not. fbordg .and. iand( iborder, 1 ) .gt. 0 ) then
        call newpen( -1 )
        call newpen( idefpen(2) )
        call move( rgroffx, rgroffy )
        rxnew = rgroffx + rgrsizx
        call draw( rxnew, rgroffy )
        rynew = rgroffy + rgrsizy
        call draw( rxnew, rynew )
        call draw( rgroffx, rynew )
        call draw( rgroffx, rgroffy )
        fbordg = .true.
        end if
c
c           border around cell
c
      if ( .not. fbordc .and. iand( iborder , 2 ) .gt. 0 ) then
        call newpen( -1 )
        call newpen( idefpen(2) )
        rxnew = rpagzerx + rcelzerx + rxspoff
        rynew = rpagzery + rcelzery + ryspoff
        call move( rxnew, rynew )
        rxcelmax = rxnew + rcelsizx
        call draw( rxcelmax, rynew )
        rycelmax = rynew + rcelsizy
        call draw( rxcelmax, rycelmax )
        call draw( rxnew, rycelmax )
        call draw( rxnew, rynew )
        fbordc = .true.
        end if
      return
      end


      SUBROUTINE ILGAXS( RANG, NPOS, ITYPE )
c
c       Draw one of the axes .
c
c  RANG     # direction of axis, degrees anticlockwise from the horizontal
c  NPOS     # position of the labels,  -1 = below the axis
c           #                           1 = above the axis
c           #                           0 = curved ( polar angle ) axis
c  ITYPE    # type of axis 1 = x axis
c           #              2 = y axis
c           #              3 = radial axis on polar plot
c           #              4 = peripheral axis on polar plot
c
c allowed values of arguments
c  ITYPE  RANG   NPOS
c    1    0.0     -1
c    2    90.0     1
c    3    any     1 -1
c    4  ignored   1 -1
c
c   If iaxis > 1  an axis is drawn .
c            & 2  the axis is numbered
c            & 4  the ticks are extended to form a grid
c            & 8  small ticks are drawn between numbered ticks
c
c variables
c  logical *1 flog(2)          # log. scaled axes ?
c  integer *2 nfigs            # no. of sig figs for number label
c             iaxis(2)         # axis labelling flags
c             ndirn            # axis direction ; 1=x ; 2=y
c             nperp            # the other axis direction
c             ldec             # characteristic of log of number label value
c             idecpl           # no. of decimal places
c             ieexp            # no. of exponent places
c             nsign            # space occupied by a - sign
c             npos             # position of labels -1=below 1=above
c             itype            # type of axis
c             ilgint           # function : integer part - round down
c             ilgrmb           #          : remove excess blanks
c             lennum           # length of number label string
c  real r                      # distance of tick from origin in cm
c       rang, rangle           # angle of axis rotation
c       raxlen                 # axis length
c       ramin(2)               # min axis label value
c       raxinc(2)              # axis label increment
c       rsmtic(2)              # initial distance between small ticks on axis
c       raxint                 # interval for ticks in cm
c       raxpos                 # position of axis
c       rcos, rsin             # cosine & sine of axis rotation
c       rctemp, rstemp         # temporary offset
c       rchrwi                 # character width
c       rdatfac(2)             # data conversion factor
c       rgrsiz(2)              # size of graph
c       rmax(2), rmin(2)       # corners of the graph
c       rnumlen                # length of number label string in cm
c       rnum                   # number label value for tick on axis
c       rposn                  # starting position for a number label
c       rprevpos               # position after printing previous number label
c       rtic1, rtic2           # top & bottom vertical position of axis tick
c       rtic3, rtic4           # top & bottom vertical position of lower tick
c       rtic5, rtic6           # top & bottom vertical position of upper tick
c       rtic7, rtic8           # top & bottom vertical position of small tick
c       rx                     # position of tick
c       rypos                  # number label vertical position
c       roldtic                # original small tick position
c       rticinc                # space between small ticks
c  character gnum*14           # number label string
c            Cformat*12        # format specifier for labels
c  subroutine ilgrnd           # round to nfigs significant figures
c
      implicit none
      real Hradcon
      parameter( Hradcon = 0.017453293 )
c
      logical *1 flog(2)
      integer *2 nfigs, ndirn, nperp, ldec, idecpl, ieexp, iflip,
     1  iaxis(2), nsign, ilgint, ilgrmb, lennum, npos, itype
      real r, rang, rangle, raxlen, ramin(2), raxinc(2), raxint, raxpos,
     1  rcos, rsin, rctemp, rstemp, rchrwi, rdatfac(2), rgrsiz(2),
     2  rmax(2), rmin(2), rnumlen, rnum, rposn, rprevpos, rtic1, rtic2,
     3  rtic3, rtic4, rtic5, rtic6, rtic7, rtic8, rx, rypos, roldtic,
     4  rticinc, rsmtic(2)
      character gnum*14, Cformat*12
c
      include 'ilglib.cmn'
      equivalence ( flog, flogx ), ( rmin, rxmin ), ( rmax, rxmax ),
     1  ( rdatfac, rdatfacx ), ( raxinc, raxincx ), ( ramin, raxmin ),
     2  ( rsmtic, rsmticx ), ( rgrsiz, rgrsizx ), ( iaxis, ixaxis )
c
      call newpen( -1 )
      call newpen( idefpen(1) )
c
c         STRAIGHT AXES
c         #############
c
      if ( itype .eq. 1 .or. itype .eq. 2 .or. itype .eq. 3 ) then         1b
        rypos = -0.5 * rchrhi + npos * ( rchrhi + rpenth )
        rchrwi = rchrhi * 0.45
        rprevpos = -2.0 * rchrwi
        if ( itype .eq. 1 ) then                                           2b
          rangle = 0.0
          nperp = 2
          ndirn = 1
          raxlen = rgrsizx
          iflip = 1
        else if ( itype .eq. 2 ) then                                      2m
          rangle = 90.0
          nperp = 1
          ndirn = 2
          raxlen = rgrsizy
          iflip = -1
        else                                                               2m
          rangle = mod( rang, 360.0 )
          if ( rangle .lt. 0.0 )
     1      rangle = 360.0 - rangle
          if ( rangle .gt. 120.0 )
     1      rprevpos = raxlen
          nperp = 1
          ndirn = 2
          raxlen = rgrsizy
          iflip = 1
          end if                                                           2e
        rsin = sin( rangle * Hradcon )
        rcos = cos( rangle * Hradcon )
c
c           determine position of axis on page
c
        if ( itype .eq. 3 ) then                                           2b
          raxpos = 0.0
        else                                                               2m
          raxpos = ( amin1( amax1( rmin( nperp ), 0.0 ), rmax( nperp ) )
     1      - rmin( nperp ) ) * rdatfac( nperp ) * iflip
          end if                                                           2e
        raxint = raxinc( ndirn )
c
c           calculate tick start & stop position
c
        rtic6 = - rchrhi / 3.0 * iflip
        rtic5 = - rtic6
        rtic4 = rgrsiz( nperp ) * iflip + rtic6
        rtic3 = rgrsiz( nperp ) * iflip - rtic6
        rtic2 = raxpos - rtic6
        rtic1 = raxpos + rtic6
        rtic7 = raxpos - rtic6 * 0.4
        rtic8 = raxpos + rtic6 * 0.4
c
c           draw axis
c
        call move( rgroffx - raxpos*rsin, rgroffy + raxpos*rcos )
        call draw( rgroffx + raxlen * rcos - raxpos * rsin,
     1               rgroffy + raxlen * rsin + raxpos * rcos )
c
c           ticks, grids & numbering etc
c
        do 3 rx = ramin( ndirn ), rmax( ndirn ) * 1.00001, raxint          2b
c
c           calculate distance along axis, value of number & no. of sig. figs
c
          if ( flog( ndirn ) ) then                                        3b
            call ilgrnd( rx, raxint, rnum, r, nfigs )
          else                                                             3m
            nfigs =  1
            if ( abs( rx / raxint ) .gt. 0.001 )
     1        nfigs = max0( ilgint( log10( abs( rx ) ) ) -
     2          ilgint( log10( raxint ) ) + 1 , 1 )
            rnum = rx
            r = rx
            end if                                                         3e
          r = ( r - rmin( ndirn ) ) * rdatfac( ndirn )
c
c           draw grid line
c
          rctemp = rgroffx + r * rcos
          rstemp = rgroffy + r * rsin
          if ( 0.0 .le. r .and. r .le. raxlen + 0.005 ) then               3b
            if ( iand( iaxis( ndirn ), 4 ) .gt. 0 ) then                   4b
              if ( itype .eq. 1 .or. itype .eq. 2 ) then                   5b
                call move( rctemp, rstemp )
                call draw( rctemp - rgrsiz(nperp) * iflip * rsin,
     1                     rstemp + rgrsiz(nperp) * iflip * rcos )
              else if ( itype .eq. 3 .and. r .gt. 0 ) then                 5m
                call move( rgroffx, rgroffy )
                call arctol( 6.0 / sqrt( raxlen ) )
                call arc( r, rxmin, rxmax )
                end if                                                     5e
            else                                                           4m
c
c           draw standard tick on the axis
c
              call move( rctemp - rtic1*rsin, rstemp + rtic1*rcos )
              call draw( rctemp - rtic2*rsin, rstemp + rtic2*rcos )
              end if                                                       4e
c
c           number the axis
c           ===============
c
            if ( iand( iaxis( ndirn ), 2 ) .gt. 0 .and. itype .le. 2 )
     1          then                                                       4b
c
c           put standard ticks at sides of graph
c
c
              call move( rctemp - rtic3*rsin, rstemp + rtic3*rcos )
              call draw( rctemp - rtic4*rsin, rstemp + rtic4*rcos )
              call move( rctemp - rtic5*rsin, rstemp + rtic5*rcos )
              call draw( rctemp - rtic6*rsin, rstemp + rtic6*rcos )

c               convert to formatted chars
c               remove leading & trailing blanks, length of number
c
              nsign = 0
              if ( rnum .lt. 0 )
     1          nsign = 1
              if ( .not. flog(ndirn) .and. abs( rnum / raxint ) .lt.
     1            0.001 ) then                                             5b
                gnum = '0.'
              else                                                         5m
                ldec = ilgint( log10( abs( rnum * 1.0000001 ) ) )
                if ( 2 + nfigs .ge. ldec .and. ldec .ge. -3 ) then         6b
                  idecpl = max( nfigs - 1 - ldec, 0 )
                  write( Cformat, 1 ) max( ldec + idecpl + 2 + nsign,
     1              idecpl + nsign + 2 ), idecpl
1                 format( '(f', i2, '.', i2, ')    ' )
                  write( gnum, Cformat ) rnum
                else                                                       6m
                  ieexp = 1
                  if ( abs( ldec ) .ge. 10 )
     1              ieexp = 2
                  write( Cformat, 2  ) ieexp + nfigs + nsign + 3,
     1              nfigs - 1, ieexp
2                 format( '(1pe', i2, '.', i2, 'e', i1, ')' )
                  write( gnum, Cformat ) rnum
                  end if                                                   6e
                end if                                                     5e
              lennum = ilgrmb( gnum )
              rnumlen = lennum * 2 * rchrwi
c
c               only print  the number if there is enough space
c
              rposn = min( max( r - rnumlen / 2.0, - rchrwi ),
     1          raxlen - rnumlen + 3.0 * rchrwi - rpenth )
              if ( rposn .ge. rprevpos ) then                              5b
                call symbol( rgroffx + rposn * rcos - rypos * rsin,
     1                       rgroffy + rposn * rsin + rypos * rcos,
     2                       rchrhi, glabel, rangle, lennum )
                rprevpos = rposn + rnumlen + rchrwi
                end if                                                     5e
              end if                                                       4e
c end of numbering axis
            end if                                                         3e
3         continue                                                         2e
c end of loop
c
c               put small ticks on the axis
c               ***************************
c
        if ( iand( iaxis(ndirn), 8 ) .gt. 0 ) then                         2b
          rticinc = rsmtic( ndirn )
          roldtic = rticinc
c
c               find starting point
c
          if ( flog(ndirn) ) then                                          3b
            rnum = 10 ** rmin(ndirn)
            rx = 10 ** ramin(ndirn)
          else                                                             3m
            rnum = rmin(ndirn)
            rx = ramin(ndirn)
            end if                                                         3e
4         if ( rx - rticinc .gt. rnum ) then                               3b
            rx = rx - rticinc
            goto 4
            end if                                                         3e
          if ( flog(ndirn) ) then                                          3b
            rnum = 10 ** rmax(ndirn)
          else                                                             3m
            rnum = rmax(ndirn)
            end if                                                         3e
5         if ( rx .le. rnum ) then                                         3b
c
c               convert point to cm from bottom LHS of graph
c
            r = rx
            if ( flog(ndirn) )
     1        r = log10( rx )
            r = ( r - rmin(ndirn) ) * rdatfac(ndirn)
c
c               draw small tick
c
            if ( 0.0 .le. r .and. r .le. raxlen + 0.005 ) then            4b
              rctemp = rgroffx + r * rcos
              rstemp = rgroffy + r * rsin
              call move( rctemp - rtic7 * rsin, rstemp + rtic7 * rcos )
              call draw( rctemp - rtic8 * rsin, rstemp + rtic8 * rcos )
              end if                                                      4e
c
c               increment tick position
c
            rx = rx + rticinc
            if ( flog(ndirn) .and. rx .ge. roldtic * 9.99999 ) then       4b
              roldtic = roldtic * 10.0
              rticinc = roldtic
              rx = roldtic
              end if                                                      4e
            goto 5
            end if                                                        3e
          end if                                                          2e
c
c            CURVED AXIS FOR POLAR PLOTS
c            ###########################
        else if ( itype .eq. 4 ) then                                     1m
c plot circumferential arc
c put small ticks on it
c put radial grid - almost from the centre
c edge ticks
c numbering
c
        end if                                                            1e
      return
      end



      SUBROUTINE ILGCLP
c
c           set graph clipping limits
c
      implicit none
      logical *4 ilguuc
      include 'ilglib.cmn'
c
      if ( fwklip ) then
        if ( fuser ) then
          if ( .not. ilguuc( 5 ) )
     1      return
          end if
        rxllim = rwild(5)
        ryllim = rwild(6)
        rxulim = rwild(7)
        ryulim = rwild(8)
        fwklip = .false.
      else if ( fklip ) then
        rxllim = rgroffx - rgrsizx * 0.004
        ryllim = rgroffy - rgrsizy * 0.004
        rxulim = rgroffx + rgrsizx * 1.004
        ryulim = rgroffy + rgrsizy * 1.004
      else
        rxllim = rgroffx - rgrzerx
        ryllim = rgroffy - rgrzery
        rxulim = rxllim + rcelsizx
        ryulim = ryllim + rcelsizy
        end if
      return
      end


      SUBROUTINE ILGROF
c
c       Determine cell origin and overall graph offset
c
      implicit none
      include 'ilglib.cmn'
c
      Rcelzerx = Rcelsizx * ( max( 1, ixpos ) - 1 )
      Rcelzery = Rcelsizy * ( iyplts - iypos )
      Rgroffx = Rpagzerx + Rcelzerx + Rgrzerx + Rxspoff
      Rgroffy = Rpagzery + Rcelzery + Rgrzery + Ryspoff
      return
      end




      SUBROUTINE ILGXLAB( CXLABEL )
c
c       Labels the x axis
c   Labels can be up to 60 characters long .
c       The label is put half way along the axis .
c Group A ( see ILGNOTE ) special functions can be included in the label
c string . Note that the label may longer be correctly centred .
c  Parameters :-
c    CXLABEL - ASCII character x axis label
c variables
c  integer *2 ilen             # no. of chars in CXLABEL
c             ilgrmb           # function : removes blanks from string
c
      implicit none
      integer *2 ilen, ilgrmb
      character *(*) cxlabel
      include 'ilglib.cmn'
c
      if ( .not. fopen )
     1  call ilgopen( ' ' )
c
      if ( .not. fxlab .or. finvis ) then
        call ilgrof
        ilen = ilgrmb( cxlabel )
        fxlab = .not. finvis
        call ilglab( rgroffx + ( rgrsizx - 0.9 *  rchrhi * ilen ) / 2.0,
     1    rgroffy - rchrhi * 3.0 - rpenth, rchrhi, 0.0, ilen )
        fxlab = .true.
        fwild = .false.
        end if
      return
      end


      SUBROUTINE ILGLAB( RX, RY, RCHRSZ, RANG, ILEN )
c
c    Parses and writes an axis label or a graph title
c
c Arguments :
c  real rx,ry       # starting coordinate of string
c       rchrsz      # character height
c       rang        # angle of text ( 0 or 90 degrees )
c       ilen        # length of character string
c
c  Group A special functions are available -- as for ilgnote
c  Hascoff = 48                # offset of character `0' in ASCII character set
c  Hpvsiz = 0.6                # size of subscript & superscript characters
c                              # expressed as a fraction of normal char size
c variables
c  logical *1 gchar            # current character being parsed
c             gspsym           # special centred symbol
c             fchar            # standard ASCII character found flag
c             full             # line is full
c  integer *2 ioptr            # output buffer pointer ; no of chars to symbol
c             ipd              # pen & dashed line value
c             iiptr            # input character string pointer
c             imult            # function multiplier
c             ihdir            # horizontal direction index
c             ivdir            # vertical direction index
c             i                # multiplier loop index
c             npv              # number of sub/superscripted characters
c             kpen             # -ve line style number
c             ispsym           # special centred symbol
c
c  real rxy(2)                 # current position of lower left of cursor
c       rxss, ryss             # position of special centred symbol
c       rchrwi                 # character width in cm
c       rleft                  # starting position for text
c       rllim                  # left limit for text
c       rrlim                  # right limit for text
c
      implicit none
      integer *2 Hascoff
      real Hpvsiz
      parameter( Hpvsiz=0.6, Hascoff=48 )
c
      logical *1 fchar, full, fdigit
      integer *2 ioptr, ilen, iiptr, imult, ihdir, ivdir, i, npv,
     1  ispsym, idigit
      real rx, ry, rchrsz, rang, rxy(2), rxss, ryss, rchrwi,
     1  rllim, rrlim, rysss
      character charf*1, gchar*1
      equivalence ( idigit, fdigit )
c
      include 'ilglib.cmn'
c
c         draw borders if required
c
      if ( iborder .gt. 0 ) call ilgbor
c
c             initialize pointers and page limits
c
      iiptr = 1
      ioptr = 0
      idigit = 0
      rchrwi = 0.9 * rchrsz
      if ( rang .eq. 0.0 ) then
        ihdir = 1
        ivdir = 2
        rllim = rgroffx
        rrlim = rllim + rgrsizx
        rxy(1) = max( rx, rllim )
        rxy(2) = ry
      else
        ihdir = 2
        ivdir = 1
        rllim = rgroffy
        rrlim = rllim + rgrsizy
        rxy(1) = rx
        rxy(2) = max( ry, rllim )
        end if
c
      npv = 0
      call vecmod
      if ( finvis )
     1  call invis
      call newpen( idefpen(3) )
      call newpen( -1 )
c
c             character parsing loop
c
1     if ( iiptr .le. ilen ) then                                          1b
        gchar = charf( glabel(iiptr) )
        iiptr = iiptr + 1
        if ( iiptr .gt. ilen ) then                                        2b
          if ( gchar .eq. '|' ) goto 3
          fchar = gchar .ne. '|'
        else                                                               2m
          fchar = gchar .ne. '|' .or. charf( glabel(iiptr) ) .eq. '|'
          end if                                                           2e
        if ( fchar ) then                                                  2b
c
c             found ordinary character, transfer it to O/P line buffer
c
          npv = 0
          if ( gchar .eq. '|' )
     1      iiptr = iiptr + 1
          ioptr = ioptr + 1
          idigit = ichar( gchar )
          glabel(ioptr) = fdigit
c
c             if O/P buffer is full or line is full, print line
c                                                                          3b
          full = ioptr * rchrwi + rxy(ihdir) .ge. rrlim
          if ( ioptr .ge. 60 .or. full ) then
            call symbol( rxy(1), rxy(2), rchrsz, glabel, rang, ioptr )
            ioptr = 0
            rxy(ihdir) = ioptr * rchrwi + rxy(ihdir)
            end if                                                         3e
          if ( full ) then
            call anmode
            return
            end if
        else                                                               2m
c
c             start of escape sequence
c
c             print contents of buffer
c
          if ( ioptr .gt. 0 ) then
            call symbol( rxy(1), rxy(2), rchrsz, glabel, rang, ioptr )
            rxy(ihdir) = rxy(ihdir) + ioptr * rchrwi
            end if
          ioptr = 0
c
c             get function multiplier or function
c
          imult = 1
          if ( iiptr .gt. ilen ) goto 3
          fdigit = glabel(iiptr)
          gchar = char( idigit )
          iiptr = iiptr + 1
          if ( idigit .ge. 0 + Hascoff .and.
     1         idigit .le. 9 + Hascoff ) then                              3b
            imult = idigit - Hascoff
            if ( iiptr .gt. ilen ) goto 3
            fdigit = glabel(iiptr)
            gchar = char( idigit )
            iiptr = iiptr + 1
            if ( idigit .ge. 0 + Hascoff .and.
     1           idigit .le. 9 + Hascoff ) then                            4b
              imult = imult * 10 + idigit - Hascoff
              if ( iiptr .gt. ilen ) goto 3
              gchar = charf( glabel(iiptr) )
              iiptr = iiptr + 1
              end if                                                       4e
            end if                                                         3e
          if ( imult .gt. 0 ) then                                         3b
c
c             execute special function
c
            do 2 i = 1, imult                                              4b
c
c       Group A functions
c
c        ***  back space
c
              if ( gchar .eq. 'B' .or. gchar .eq. 'b' ) then               5b
                if ( npv .le. 0 ) then                                     6b
                  rxy(ihdir) = max( rxy(ihdir) - rchrwi, rllim )
                else                                                       6m
                  rxy(ihdir) = max( rxy(ihdir) - rchrwi * Hpvsiz,
     1              rllim )
                  npv = npv - 1
                  end if                                                   6e
c
c        ***  forward space
c
              else if ( gchar .eq. 'F' .or. gchar .eq. 'f' .and.
     1            i .eq. 1 ) then                                          5m
                rxy(ihdir) = amin1( rxy(ihdir) + imult * rchrwi, rrlim )
                npv = 0
c
c        *** Greek & other special characters
c
              else if ( gchar .eq. 'K' .or. gchar .eq. 'k' ) then
                call move( rxy(1), rxy(2) )
                call txtang( rang )
                call txtsiz( rchrsz )
                call greek( 1, charf( glabel(iiptr) ) )
                iiptr = iiptr + 1
                rxy(ihdir)  = rxy(ihdir) + rchrwi
                npv = 0
c
c        ***  Pen select
c
              else if ( gchar .eq. 'P' .or. gchar .eq. 'p' .and.
     1            i .eq. 1 ) then
                fdigit = glabel(iiptr)
                if ( idigit .ge. 0 + Hascoff .and.
     1              idigit .le. 9 + Hascoff ) then                         6b
                  call newpen( idigit - Hascoff )
                  iiptr = iiptr + 1
                  end if                                                   6e
c
c        *** horizontal line
c
              else if ( ( gchar .eq. '-'.or. gchar .eq. '~' .or.
     1            gchar .eq. '_' ) .and. i .eq. 1 ) then
                npv = 0
                fdigit = glabel(iiptr)
                if ( idigit .gt. 0 + Hascoff .and.
     1              idigit .le. 9 + Hascoff )
     2            iiptr = iiptr + 1
                call newpen( Hascoff - idigit )
c
                rxss = min( rxy(ihdir), rrlim )
                if ( gchar .eq. '~' ) then
                  rysss = 1.06 * rchrsz + rpenth / 2.0
                else if ( gchar .eq. '_' ) then
                  rysss = -0.06 * rchrsz - rpenth / 2.0
                else
                  rysss = 0.5 * rchrsz
                  end if
                rysss = rxy(ivdir) + rysss * ( ivdir - ihdir )
                ryss = rysss
                if ( rang .ne. 0.0 )
     1            call ilgsw4( rxss, ryss )
                call move( rxss, ryss )
c
                rxss = max( rxy(ihdir), min( rrlim, rxy(ihdir) + imult
     1            * rchrwi ) )
                if ( gchar .eq. '-' )
     1            rxy(ihdir) = rxss
                ryss = rysss
                if ( rang .ne. 0.0 )
     1            call ilgsw4( rxss, ryss )
                call draw( rxss, ryss )
                call newpen( -1 )
c
c        *** superscript and subscript
c
              else if ( gchar .eq. 'T' .or. gchar .eq. 't' .or.
     1                  gchar .eq. 'V' .or. gchar .eq. 'v' ) then          5m
                if ( iiptr .gt. ilen ) goto 2
                rxss = rxy(ivdir)
                rysss = -0.25 * rchrsz * (ivdir-ihdir)
                if ( gchar .eq. 'T' .or. gchar .eq. 't' )
     1            rysss = -2.2 * rysss
                rxy(ivdir) = rxy(ivdir) + rysss
                if ( charf( glabel(iiptr+1) ) .ne. 'K' .and.
     1               charf( glabel(iiptr+1) ) .ne. 'k' .or.
     2               charf( glabel(iiptr) )   .ne. '|' ) then                     6b
                  call symbol( rxy(1), rxy(2), rchrsz * Hpvsiz,
     1              glabel(iiptr), rang, 1 )
                else                                                       6e
                  call move( rxy(1), rxy(2) )
                  call txtang( rang )
                  call txtsiz( rchrsz * Hpvsiz )
                  iiptr = iiptr + 2
                  call greek( 1, charf( glabel(iiptr) ) )
                  end if                                                   6m
                rxy(ivdir) = rxss
                rxy(ihdir) = rxy(ihdir) + Hpvsiz * rchrwi
                iiptr = iiptr + 1
                npv = npv + 1
c
c        ***  special symbol
c
              else if ( gchar .eq. 'S' .or. gchar .eq. 's' ) then          5m
c
c        ***  get special character
c
                if ( i .eq. 1 ) then                                       6b
                  npv = 0
                  fdigit = glabel(iiptr)
                  if ( iiptr .gt. ilen ) goto 3
                  ispsym = 0
                  if ( idigit .ge. 0 + Hascoff .and.
     1                idigit .le. 9 + Hascoff ) then                       7b
                    ispsym = idigit - Hascoff
                    iiptr = iiptr + 1
                    if ( iiptr .le. ilen ) then                            8b
                      fdigit = glabel(iiptr)
                      if ( idigit .ge. 0 + Hascoff .and.
     1                     idigit .le. 9 + Hascoff ) then                  9b
                        ispsym = ispsym * 10 + idigit - Hascoff
                        iiptr = iiptr + 1
                        end if                                             9e
                      end if                                               8e
                    end if                                                 7e
                  end if                                                   6e
                if ( ispsym .gt. 15 )
     1            ispsym = 0
c
c             print special character
c
                rxss = rxy(1) + 0.5 * rchrwi * (ivdir-ihdir)
                ryss = rxy(2) + 0.5 * rchrsz
                if ( ispsym .eq. 0 ) then                                  6b
                  call move( rxss, ryss )
                  call draw( rxss, ryss )
                else                                                       6m
                  call symbol( rxss, ryss, rchrsz, ispsym - 1,
     1              rang, -1 )
                  end if                                                   6e
                rxy(ihdir) = rxy(ihdir) + rchrwi
                end if                                                     5e
2             continue                                                     4e
            end if                                                         3e
          end if                                                           2e
          goto 1
        end if                                                             1e
c
c             flush character buffer
c
3     if ( ioptr .gt. 0 )
     1  call symbol( rxy(1), rxy(2), rchrsz, glabel, rang, ioptr )
      if ( finvis ) then                                                   1b
        call visibl
        finvis = .false.
        end if                                                             1e
      call anmode
      return
      end


      CHARACTER *1 FUNCTION CHARF( FCHAR )
c
c       converts a `character' in a logical *1 variable to a character
c
      implicit none
      logical *1 fchar, ftemp
      integer *2 itemp
      equivalence( itemp, ftemp )
c
      itemp = 0
      ftemp = fchar
      charf = char( itemp )
      return
      end


      INTEGER *2 FUNCTION ILGRMB( GSTR )
c
c       Remove leading & trailing blanks .
c  Removes leading and trailing blanks from the string GSTR and return with
c ILGRMB = the length of the resulting string . Transfers CSTR to the
c logical *1 array GLABEL .
c
c argument
c  character *(*) Gstr         # operand string
c variables
c  integer *2 i                # integer of first non-blank char
c             j                # integer of last non-blank char
c             itemp            # temporary character address
c  logical *1 ftemp            # = itemp
c
      implicit none
      logical *1 ftemp
      integer *2 i,j, itemp
      character *(*) Gstr
      include 'ilglib.cmn'
      equivalence ( itemp, ftemp )
c
      ilgrmb = len( Gstr )
      if ( ilgrmb .le. 0 ) return
c
      i = 1
1     if ( Gstr(i:i) .eq. ' ' .and. i .lt. ilgrmb ) then
        i = i + 1
        goto 1
        end if
      j = ilgrmb
      if ( j .gt. 1 ) then
2       if ( Gstr(j:j) .eq. ' ' .and. j .ge. i ) then
          j = j - 1
          goto 2
          end if
        end if
      ilgrmb = j - i + 1
      if ( j .ge. i ) then
        do 3 j = 1, ilgrmb
          itemp = ichar( Gstr(i:i ) )
          Glabel(j) = ftemp
3         i = i + 1
        end if
      return
      end


      SUBROUTINE ILGYLAB( CYLABEL )
c
c       Labels the y axis
c  Parameters :-
c    CYLABEL - ASCII character y axis label
c       The label is put half way along the axis .
c Group A ( see ILGNOTE ) special functions can be included in the label
c string . Note that the label may longer be correctly centred .
c
c variables
c  integer *2 ilen             # no. of chars in CYLABEL
c             ilgrmb           # function : removes blanks from string
c
      implicit none
      character *(*) cylabel
      integer *2 ilen,ilgrmb
      include 'ilglib.cmn'
c
      if ( .not. fopen )
     1  call ilgopen( ' ' )
c
      if ( .not. fylab .or. finvis ) then
        call ilgrof
        ilen = ilgrmb( cylabel )
        fylab = .not. finvis
        call ilglab(  rgroffx - rchrhi * 2.0 - rpenth,
     1    rgroffy + ( rgrsizy - 0.9 * rchrhi * ilen ) / 2.0,
     2    rchrhi, 90.0, ilen )
        fylab = .true.
        fwild = .false.
        end if
      return
      end



      SUBROUTINE ILGTITL( CTITLE )
c
c       Puts a title at the top of the graph
c  Parameters :-
c    CTITLE - ASCII character title
c       The label is put half way along the top of the graph .
c Group A ( see ILGNOTE ) special functions can be included in the label
c string . Note that the label may longer be correctly centred .
c
c variables
c  integer *2 ilen             # no. of chars in CXLABEL
c             ilgrmb           # function : removes blanks from string
c
      implicit none
      character *(*) ctitle
      integer *2 ilen, ilgrmb
      include 'ilglib.cmn'
c
      if ( .not. fopen )
     1  call ilgopen( ' ' )
c
      if ( .not. ftitl .or. finvis ) then
        call ilgrof
        ilen = ilgrmb( ctitle )
        ftitl = .not. finvis
        call ilglab( rgroffx + ( rgrsizx - 0.9 * rchrti * ilen ) / 2.0,
     1    rgroffy + rgrsizy + 0.5 * rchrhi + rpenth, rchrti, 0.0, ilen )
        ftitl = .true.
        fwild = .false.
        end if
      return
      end


      LOGICAL *1 FUNCTION ILGREAD( X, Y, CHAR )
c
c       Read a coordinate from the VDU screen using cross-hairs .
c  The returned value of ILGREAD is .true. or .false. to indicate success or
c  failure respectively . Failure occurs if :-
c    a) reading of a point outside the current graph cell was attempted more
c       than 8 times .
c    b ) or the identifying character is 'x' or 'X' .
c  On failure , the veturned coordinate is at the lower left corner of the
c  graph .
c  Parameters :
c   X,Y   - returned coordinate value in the user's coordinates
c   CHAR  - returned identifying character value
c
c constants
c  Hvdulu              # vdu logical unit number
c variables
c  integer *2 ntry     # number of attempts
c             i        # number of points read
c  real x, y           # coordinate
c  character *1 char   # identifying character
c
      implicit none
      integer *2 Hvdulu
      parameter( Hvdulu=6 )
c
      integer *2 ntry, i
      real x, y
      character *1 char
c
      include 'ilglib.cmn'
c
      if ( .not. fopen )
     1  call ilgopen( ' ' )
      call vecmod
c
c         read coordinate
c
      do 1 ntry = 1, 9
        ilgread = .true.
        call locate( 1, x, y, char, i )
        if ( char .eq. 'x' .or. char .eq. 'X' ) then
          ilgread = .false.
          x = rxllim
          y = ryllim
          end if
        if ( rxllim .le. x .and. x .le. rxulim .and.
     1       ryllim .le. y .and. y .le. ryulim ) goto 3
        ilgread = .false.
        write( Hvdulu, 2 ) 7
2       format( '+', a1 )
1       continue
c
c         convert to user coordinate system
c
3     x = ( x - rgroffx ) / rdatfacx + rxmin - rxofset
      y = ( y - rgroffy ) / rdatfacy + rymin - ryofset
      if ( flogx )
     1  x = 10.0 ** x
      if ( flogy )
     1  y = 10.0 ** y
c
      call anmode
      fwild = .false.
      return
      end


      SUBROUTINE ILGNOTE( CNOTE )
c
c       Write notes into a plotting cell , starting at the top left hand corner
c . '|' is an escape character for a special function denoted by an upper or
c lower case letter .
c
c  These functions fall into 2 groups . Group A functions are :
c    1) '||'                   '|' character
c    2) '|F' or '|f'           forward space
c    3) '|B' or '|b'           back space
c    4) '|T' or '|t'           The next character is superscripted .
c    5) '|V' or '|v'           The next character is subscripted .
c    6) '|K' or '|k'           The next letter is a Greek character equivalent
c    7) '|P' or '|p'           Next character is pen number to be selected
c    8) '|S' or '|s'           a one or two digit integer follows . This
c                               integer is interpreted as a special symbol in
c                               the same way as the NSYMB argument to ILGLINE .
c    9) '|-'                   a single digit may follow. This digit is inter-
c                              preted as a dashed line style . Horizontal line
c                              of the specified dashed line style, spanning the
c                              character width is drawn . Default dashed line
c                              style is 'continuous'
c    10) '|_'                  underline - same syntax as '|-'
c    11) '|~'                  overline  - same syntax as '|-'
c
c  Group B functions are :
c    1) '|R' or '|r'           cursor carriage return and line feed
c    2) '|H' or '|h'           cursor home  ( ie top LH corner )
c    3) '|L' or '|l'           cursor line feed
c    4) '|U' or '|u'           cursor up
c    5) '|G' or '|g'           Place cursor in next plotting cell
c    6) '|XC' or '|xc'         Place the cursor centre at the cross hairs .
c    7) '|XM' or '|xm'         Move to the cross hairs with pen up .
c    8) '|XD' or '|xd'         Move to the cross hairs with pen down .
c    9) '|M' or '|m'           a coordinate in parenthesis follows . Move to
c                              that coordinate with pen up .
c   10) '|D' or '|d'           a coordinate in parenthesis follows . Draw to
c                              that coordinate with pen down .
c   11) '|C' or '|c'           a coordinate in parentheses follows . Move the
c                              cursor to that coordinate .
c   12) '|E' or '|e'           2 values in parentheses follow . An error bar
c                              of this size is drawn at the current cursor
c                              location .
c
c               The (x,y) coordinate must be enclosed by parentheses . The
c           parentheses must enclose 2 numbers which can be understood by
c           a list directed read statement . The 2 numbers must be separated
c           by a space or a comma . N.B. the M and D functions do not move
c           the cursor .
c
c    Leading Parameters
c  A one or two digit decimal number may be inserted between the escape
c  character and the function character . This is the function multiplier .
c
c    Example CNOTE = '4|4fspaces |3fbetween |f  words'   produces
c     '4    spaces    between    words'
c
c constants
c  Hvdulu                      # vdu logical unit number
c  Hpvsiz                      # size of subscript & superscript characters
c                              # expressed as a fraction of normal char size
c variables
c  character *(*) cnote        # note character string
c            *1   gchar        # current character being parsed
c            *1   gchr2        # 2nd char for a crosshair function
c            *1   gspsym       # special centred symbol
c  logical *1 fchar            # standard ASCII character found flag
c             fplot            # point is on page flag
c             ftemp            # temporary  - for SYMBOL
c  integer *2 ispsym           # special centred symbol
c             ioptr            # output buffer pointer ; no of chars to symbol
c             ipd              # pen & dashed line value
c             jpen             # current cursor pen number
c             kpen             # -ve line style
c             imult            # function multiplier
c             iiptr            # input character string pointer
c             ilen             # length of input string
c             ii               # end of (x,y) coord pointer
c             i                # multiplier loop index
c             npv              # number of sub/superscripted characters
c             itemp            # temporary -  for SYMBOL
c
c  real rx, ry                 # current position of lower left of cursor
c       rxss, ryss             # position of special centred symbol
c       rxx, ryy               # pen position for 'move' and 'draw'
c       rchrwi                 # character width in cm
c       rlfhi                  # line spacing in cm
c       rxleft                 # position of left margin
c       rxrite                 # position of right margin
c       rytop                  # position of first line on page
c       rybot                  # position of last line on page
c
      implicit none
      integer *2 Hvdulu
      real Hpvsiz
      parameter( Hvdulu=6, Hpvsiz=0.6 )
c
      character cnote*(*), gchar*1, gspsym*1, gchr2*1
      logical *1 fchar, ftemp
      integer *2 ioptr, ipd, jpen, kpen, imult, i, iiptr, ilen, ii,
     1  npv, itemp, ispsym
      real rx, ry, rxss, ryss, rchrwi, rlfhi, rxleft, rxrite, rytop,
     1  rybot
c
      include 'ilglib.cmn'
      equivalence ( ftemp, itemp )
c
      if ( .not. fopen )
     1  call ilgopen( ' ' )
c
c             find length of input string
c
      ilen = len( cnote )
1     if ( Cnote(ilen:ilen) .eq. ' ' .and. ilen .gt. 1 ) then              1b
        ilen = ilen - 1
        goto 1
        end if                                                             1e
c
c             initialize pointers and page limits
c
      jpen = idefpen(4)
      iiptr = 1
      ioptr = 0
      rchrwi = 0.9 * rchrhi
      rlfhi  = 1.25 * rchrhi + rpenth
c
2     call ilgrof
c
c             draw border if required
c
      if ( iborder .gt. 0 ) call ilgbor
      call ilgclp
      rxrite = rxulim - rchrwi - 2.0 * rpenth
      rybot  = ryllim + rpenth
      rytop  = ryulim - rlfhi - rpenth
      rxleft = rxllim + 2.0 * rpenth
      rx = rxleft
      ry = rytop
      rxx = rgroffx
      ryy = rgroffy
      npv = 0
      call vecmod
      if ( finvis )
     1  call invis
      call newpen( jpen )
      call newpen( -1 )
c
c             character parsing loop
c
3      if ( iiptr .le. ilen ) then                                         1b
        gchar = cnote( iiptr: iiptr )
        iiptr = iiptr + 1
        if ( iiptr .gt. ilen ) then                                        2b
          if ( gchar .eq. '|' ) goto 10
          fchar = gchar .ne. '|'
        else                                                               2m
          fchar = gchar .ne. '|' .or. cnote( iiptr: iiptr ) .eq. '|'
          end if                                                           2e
        if ( fchar ) then                                                  2b
c
c             found ordinary character, transfer it to O/P line buffer
c
          npv = 0
          if ( gchar .eq. '|' )
     1      iiptr = iiptr + 1
          ioptr = ioptr + 1
          itemp = ichar( gchar )
          glabel( ioptr ) = ftemp
c
c             if O/P buffer is full , print line
c                                                                          3b
          if ( ioptr .ge. 60 ) then
            if ( ry .le. rytop .and. ry .ge. rybot )
     1        call symbol( rx, ry, rchrhi, glabel, 0.0, 60 )
            ioptr = 0
            rx = 60.0 * rchrwi + rx
            end if                                                         3e
c
c             if line is full, print it
c
          if ( ioptr * rchrwi + rx .ge. rxrite ) then                      3b
            if ( ry .le. rytop .and. ry .ge. rybot )
     1        call symbol( rx, ry, rchrhi, glabel, 0.0, ioptr )
            ioptr = 0
            rx = rxleft
            ry = ry - rlfhi
            end if                                                         3e
        else                                                               2m
c
c             start of escape sequence
c
c             print contents of buffer
c
          if ( ioptr .gt. 0 .and. ry .le. rytop .and. ry .ge. rybot )
     1      call symbol( rx, ry, rchrhi, glabel, 0.0, ioptr )
          rx = rx + ioptr * rchrwi
          ioptr = 0
c
c             get function multiplier or function
c
          imult = 1
          if ( iiptr .gt. ilen ) goto 10
          gchar = cnote( iiptr: iiptr )
          iiptr = iiptr + 1
          if ( gchar .ge. '0' .and. gchar .le. '9' ) then                  3b
            imult = ichar( gchar ) - 48
            if ( iiptr .gt. ilen ) goto 10
            gchar = cnote( iiptr: iiptr )
            iiptr = iiptr + 1
            if ( gchar .ge. '0' .and. gchar .le. '9' ) then                4b
              imult = imult * 10 + ichar( gchar ) - 48
              if ( iiptr .gt. ilen ) goto 10
              gchar = cnote( iiptr: iiptr )
              iiptr = iiptr + 1
              end if                                                       4e
            end if                                                         3e
          if ( imult .ne. 0 ) then                                         3b
            if ( 'a' .le. gchar .and. gchar .le. 'z' )
     1        gchar = char( ichar( gchar ) - 32 )
c
c             execute special function
c
            do 9 i = 1, imult                                              4b
c
c        ***  back space, Group A
c
              if ( gchar .eq. 'B' ) then                                   5m
                if ( npv .le. 0 ) then                                     6b
                  rx = rx - rchrwi
                  rxss = 0.0
                else                                                       6m
                  rx = rx - rchrwi * Hpvsiz
                  npv = npv - 1
                  rxss = npv * Hpvsiz * rchrwi
                  end if                                                   6e
                if ( rx .lt. rxleft ) then                                 6b
                  rx = int( ( rxrite - rxleft - rxss ) / rchrwi + 0.1  )
     1              * rchrwi + rxleft + rxss
                  ry = ry + rlfhi
                  end if                                                   6e
c
c        ***  forward space, Group A
c
              else if ( gchar .eq. 'F' ) then                              5m
                rx = rx + rchrwi
                npv = 0
                if ( rx .gt. rxrite ) then                                 6b
                  rx = rxleft
                  ry = ry - rlfhi
                  end if                                                   6e
c
c        ***  new graph / plotting cell, Group B
c
              else if ( gchar .eq. 'G' ) then                              5m
                ixpos = ixpos + 1
                if ( ixpos .gt. ixplts ) then                              6b
                  ixpos = 1
                  iypos = iypos + 1
                  if ( iypos .gt. iyplts )
     1              iypos = 1
                  end if                                                   6e
                npv = 0
c
c       ***  cross-hair functions, Group B
c
              else if ( gchar .eq. 'X' ) then                              5m
                gchr2 = cnote( iiptr: iiptr )
                if ( 'a' .le. gchr2 .and. gchr2 .le. 'z' )
     1            gchr2 = char( ichar( gchr2 ) - 32 )
                if ( i .eq. imult )
     1            iiptr = iiptr + 1
                if ( i .eq. 1 .and. ( gchr2 .eq. 'M' .or.
     1              gchr2 .eq. 'C' ) .or. gchr2 .eq. 'D' ) then            6b
c
c         digitize point
c
c                 repeat                                                   7b
4                   call locate( 1, rxss, ryss, gspsym, ii )
                    if ( ii .eq. 0 ) then                                  8b
                      rxss = rgroffx
                      ryss = rgroffy
                      end if                                               8e
                    if ( rxllim .le. rxss .and. rxss .le. rxulim .and.
     1                ryllim .le. ryss .and. ryss .le. ryulim ) goto 6
                    write( Hvdulu, 5 ) 7
5                   format( '+', a1 )
                    goto 4
c                   until ( point is in clip region )
c
c         draw to cross hairs
c
6                 if ( gchr2 .eq. 'D' ) then                               7b
                    ipd = ipen
                    call newpen( ipd )
                    ipd = -idash
                    if ( idash .gt. 1 )
     1                call newpen( ipd )
                    call move( rxx, ryy )
                    call draw( rxss, ryss )
                    call newpen( -1 )
                    call newpen( jpen )
                    rxx = rxss
                    ryy = ryss
c
c        move to cross hairs
c
                  else if ( gchr2 .eq. 'M' ) then                          7m
                    rxx = rxss
                    ryy = ryss
c
c        cursor to cross hairs
c
                  else if ( gchr2 .eq. 'C' ) then                          7m
                    rx = rxss
                    ry = ryss
                    end if                                                 7e
                  end if                                                   6e
c
c         *** only execute these if on the first pass through the loop
c
              else if ( i .eq. 1 ) then                                    5m
c
c        ***  cursor down  or  linefeed, Group B
c
                if ( gchar .eq. 'L' ) then                                 6b
                  ry = ry - rlfhi * imult
                  npv = 0
c
c        ***  carriage return and linefeed, Group B
c
                else if ( gchar .eq. 'R' ) then                            6m
                  ry = ry - rlfhi * imult
                  rx = rxleft
                  npv = 0
c
c        ***  cursor up, Group B
c
                else if ( gchar .eq. 'U' ) then                            6m
                  ry = ry + rlfhi * imult
                  npv = 0
c
c        ***  home or top left corner, Group B
c
                else if ( gchar .eq. 'H' ) then                            6m
                  rx = rxleft
                  ry = rytop
                  npv = 0
c
c        ***  Move pen, Draw and move Cursor, Group B
c
                else if ( gchar .eq. 'M' .or. gchar .eq. 'D' .or.
     1                    gchar .eq. 'C' .or. gchar .eq. 'E' ) then        6m
                  if ( iiptr .gt. ilen ) goto 10
                  if ( cnote(iiptr:iiptr) .eq.'(' ) then                   7b
                    ii = iiptr + 1
7                   if ( cnote(ii:ii) .ne. ')' .and. ii .le. ilen) then    8b
                      ii = ii + 1
                      goto 7
                      end if                                               8e
                    if ( ii .gt. ilen ) goto 10
                    if ( ii - iiptr - 3 .gt. 0 ) then                      8b
                      read( cnote(iiptr+1:ii-1), *, ERR=8 ) rxss, ryss
                      if ( gchar .eq. 'E' ) then                           9b
                        call move( rx + 0.5* rchrwi, ry + 0.5 * rchrhi )
                        call txtang( 0.0 )
                        call ilgerb( rxss, ryss, rchrhi / 3.0 )
                        rx = rx + rchrwi
                      else                                                 9m
                        if ( flogx )
     1                    rxss = log10( rxss )
                        rxss = ( rxss - rxmin + rxofset ) * rdatfacx +
     1                         rgroffx
                        if ( flogy )
     1                    ryss = log10( ryss )
                        ryss = ( ryss - rymin + ryofset ) * rdatfacy +
     1                         rgroffy
                        if ( gchar .eq. 'D' ) then                         10b
                          ipd = ipen
                          call newpen( ipd )
                          ipd = -idash
                          if ( idash .gt. 1 )
     1                      call newpen( ipd )
                          call move( rxx, ryy )
                          call draw( rxss, ryss )
                          call newpen( -1 )
                          call newpen( jpen )
                          rxx = rxss
                          ryy = ryss
                        else if ( gchar .eq. 'M' ) then                    10m
                          rxx = rxss
                          ryy = ryss
                        else                                               10m
                          rx = rxss
                          ry = ryss
                          end if                                           10e
                        end if                                             9e
                      end if                                               8e
8                   iiptr = ii + 1
                    end if                                                 7e
c
c        ***  Pen select, Group B
c
                else if ( gchar .eq. 'P' ) then                            6m
                  jpen = 1
                  gspsym = cnote( iiptr: iiptr )
                  if ( gspsym .ge. '0' .and. gspsym .le. '9' ) then        7b
                    jpen = ichar( gspsym) - 48
                    iiptr = iiptr + 1
                    end if                                                 7e
                  call newpen( jpen )
                  end if                                                   6e
                end if                                                     5e
c
c       *** check that the cursor is on the page
c
              if ( rxleft .le. rx .and. rx .le. rxrite .and.
     1             rybot .le. ry .and. ry .le. rytop ) then                5b
c
c        *** Greek & other special characters, Group A
c
                if ( gchar .eq. 'K' ) then                                 6b
                  call move( rx, ry )
                  call txtang( 0.0 )
                  call txtsiz( rchrhi )
                  call greek( 1, cnote(iiptr:iiptr) )
                  iiptr = iiptr + 1
                  rx = rx + rchrwi
c
c        *** horizontal line , overline or underline, Group A
c
                else if ( gchar .eq. '-'.or. gchar .eq. '~' .or.           6m
     1              gchar .eq. '_'  ) then
                  if ( i .eq. 1 ) then                                     7b
                    kpen = -1
                    npv = 0
c
c                      decide line style
c
                    gspsym = cnote( iiptr: iiptr )
                    if ( gspsym .ge. '1' .and. gspsym .le. '8' ) then      8b
                      kpen = 48 - ichar( gspsym )
                      iiptr = iiptr + 1
                      end if                                               8e
                    call newpen( kpen )
c
c                      coordinate of line
c
                    if ( gchar .eq. '~' ) then                             8b
                      ryss = 1.06 * rchrhi + rpenth / 2.0
                    else if ( gchar .eq. '_' ) then                        8m
                      ryss = -0.06 * rchrhi - rpenth / 2.0
                    else                                                   8m
                      ryss = 0.5 * rchrhi
                      end if                                               8e
                    ryss = ry + ryss
                    rxss = rx
                    end if                                                 7e
                  if ( i .eq. 1 .or. rxss .le. rxleft )
     1              call move( rxss, ryss)
                  rxss = rxss + rchrwi
                  if ( i .eq. imult .or. rxss .ge. rxrite )
     1              call draw( rxss, ryss )
                  if ( i .eq. imult )
     1              call newpen( -1 )
                  if ( gchar .eq. '-' )
     1              rx = rx + rchrwi
                  if ( rxss .ge. rxrite ) then                             7b
                    rxss = rxleft
                    ryss = ryss - rlfhi
                    end if                                                 7e
c
c        *** superscript and subscript, Group A
c
                else if ( gchar .eq. 'T' .or. gchar .eq. 'V' ) then        6m
                  if ( iiptr .gt. ilen ) goto 10
                  if ( gchar .eq. 'T' ) then                               7b
                    ryss = ry + rlfhi - rpenth - Hpvsiz* rchrhi
                  else                                                     7m
                    ryss = ry - 0.25 * rchrhi
                    end if                                                 7e
                  if ( cnote(iiptr+1:iiptr+1) .ne. 'K' .and.
     1                 cnote(iiptr+1:iiptr+1) .ne. 'k' .or.
     2                 cnote(iiptr:iiptr) .ne. '|' ) then                  7b
                    call symbol( rx, ryss, rchrhi * Hpvsiz,
     1                ichar(cnote(iiptr:iiptr)), 0.0, 1 )
                  else                                                     7m
                    call move( rx, ryss )
                    call txtang( 0.0 )
                    call txtsiz( rchrhi * Hpvsiz )
                    iiptr = iiptr + 2
                    call greek( 1, cnote(iiptr:iiptr) )
                    end if                                                 7e
                  rx = rx + rchrwi * Hpvsiz
                  iiptr = iiptr + 1
                  npv = npv + 1
c
c        ***  special symbol, Group A
c
                else if ( gchar .eq. 'S' ) then                            6m
c
c        get special character
c
                  if ( i .eq. 1 ) then                                     7b
                    npv = 0
                    gspsym = cnote( iiptr: iiptr )
                    if ( iiptr .gt. ilen ) goto 10
                    ispsym = 0
                    if ( gspsym .ge. '0' .and. gspsym .le. '9' ) then      8b
                      ispsym = ichar( gspsym ) - 48
                      iiptr = iiptr + 1
                      if ( iiptr .le. ilen ) then                          9b
                        gspsym = cnote( iiptr: iiptr )
                        if ( gspsym .ge.'0' .and. gspsym .le.'9' ) then    10b
                          ispsym = ispsym * 10 + ichar( gspsym ) - 48
                          iiptr = iiptr + 1
                          end if                                           10e
                        end if                                             9e
                      end if                                               8e
                    end if                                                 7e
                  if ( ispsym .gt. 15 )
     1              ispsym = 0
c
c             print special character
c
                  rxss = rx + 0.5 * rchrwi
                  ryss = ry + 0.5 * rchrhi
                  if ( ispsym .eq. 0 ) then                                7b
                    call move( rxss, ryss )
                    call draw( rxss, ryss )
                  else                                                     7m
                    call symbol( rxss, ryss, rchrhi, ispsym - 1,0.0,-1 )
                    end if                                                 7e
                  rx = rx + rchrwi
                  end if                                                   6e
c
c           *** new line if required
c
                if ( rx .ge. rxrite ) then                                 6b
                  rx = rxleft
                  ry = ry - rlfhi
                  end if                                                   6e
                end if                                                     5e
9             continue                                                     4e
c
c        ***  new Graph
c
            if ( gchar .eq. 'G' ) goto 2
            end if                                                         3e
          end if                                                           2e
          goto 3
        end if                                                             1e
c
c             flush character buffer
c
10    if ( ioptr .gt. 0 .and. rxleft .le. rx .and. rx .le. rxrite .and.
     1     rybot .le. ry .and. ry .le. rytop )
     2  call symbol( rx, ry, rchrhi, glabel, 0.0, ioptr )
      if ( finvis ) then                                                   1b
        call visibl
        finvis = .false.
        end if                                                             1e
      call anmode
      fwild = .false.
      return
      end


      SUBROUTINE ILGWAIT( GRAPHICS )
c
c       Prompt and Wait for <carriage return> from the keyboard .
c Argument
c  logical *1 graphics      # wait in graphics mode ( else test mode )
c
c Constants
c  Hkbdlu=5         # keyboard logical unit number
c  Hvdulu=6         # vdu logical unit number
c
      implicit none
      integer *2 Hkbdlu, Hvdulu, Hrow, Hcol
      parameter( Hkbdlu=5, Hvdulu=6, Hrow=42, Hcol=27 )
      character Creturn*26, Cblank*26
      data Creturn / 'Press <RETURN> to continue' /
      data Cblank  / '                          ' /
c
      logical *1 graphics
c
      if ( graphics ) then                                                  1b
        call vecmod
        call HG_puts( Creturn, Hrow, Hcol )
        read( Hkbdlu, 1 )
1       format( )
        call HG_puts( Cblank, Hrow, Hcol )
        call anmode
      else                                                                  1m
        write( Hvdulu, 2 ) Creturn
2       format( ' ', a )
        read( Hkbdlu, 1 )
        end if                                                              1e
      return
      end


      SUBROUTINE ILGCLOS
c
c       Closes any capture file, clears the vdu screen and terminates graphics
c
      include 'ilglib.cmn'
c
c
      if ( fopen ) then
        call plot( 0.0, 0.0, 999 )
        if ( fcapt )
     1    call clocap
        call grstop
        fopen = .false.
        fwild = .false.
        end if
      return
      end


      SUBROUTINE ILGINVS
c
c       The next ILGLIB call which writes onto the screen writes invisibly
c       ( i.e. in black ), and so anything which is written over is rubbed
c       out . After that, ILGLIB routines return to writing visibly .
c
      implicit none
      include 'ilglib.cmn'
c
      finvis = .true.
      return
      end


      SUBROUTINE ILGERRR( IERRLU )
c
c      Set the logical unit number of the error channel .
c   Parameters :
c integer *2 ierrlu   # logical unit number for error messages
c                     #  if ierrlu < 0 , a <carriage return> must be typed
c                     #     to continue after an error message is displayed
c                     #  if ierrlu = 0 or ierrlu > 999 or ierrlu < -999 ,
c                     #     the error message is suppressed
c                     # default value of ierrlu is 6 .
c
      implicit none
      integer *2 ierrlu
      include 'ilglib.cmn'
c
      fwait = ierrlu .lt. 0
      ierlun = 6
      if ( -999 .le. ierrlu .and. ierrlu .le. 999 )
     1  ierlun = abs( ierrlu )
      return
      end


      INTEGER *2 FUNCTION ILGINT( X )
c
c       Calculates rounded down integer part of x . Ie ILGINT(X) < X
c
      implicit none
      real x
c
      if ( x .ge. 0 ) then
        ilgint = int ( x )
      else
        ilgint = int( 0.999999 * x - 1.0 )
        end if
      return
      end


      SUBROUTINE ILGRND( RLOGIN, RAXINC, ROUT, RLOGOUT, NFIGS )
c
c       Nicely round the value 10**rlogin for numbering logarithmically
c scaled axes .
c
c arguments
c  real rlogin                 # log of input ordinate
c       raxinc                 # axis increment of log values
c       rout                   # rounded value of input ordinate
c       rlogout                # log of rounded value of input ordinate
c  integer *2 nfigs            # number of significant figures of result
c
c constants
c  Hln10                       # reciprocal of ln 10
c variables
c  integer *2 ilgint           # function : integer part of - rounded down
c
      implicit none
      real Hln10, rlogin, raxinc, rout, rlogout
      parameter( Hln10=0.4342945 )
      integer *2 nfigs, ilgint
c
      nfigs = max( int( log10( Hln10 / raxinc ) - rlogin +
     1  real( ilgint( rlogin ) + 2 ) ), 1 )
      rout = 10.0 ** ilgint( rlogin - real( nfigs - 1 ) )
      rout = int( 10.0 ** rlogin / rout + 0.5 ) * rout
      rlogout = log10( rout )
      return
      end



      LOGICAL *1 FUNCTION ILGINS( GSUBSET, MATCHLEN, GSET )
c
c       ILGINS states whether the group (string) GSUBSET is contained in the
c  string GSET . GSET is divided into character groups, each of MATCHLEN
c  characters . Ie MATCHLEN is the matching length .
c
      character Gsubset*(*), Gset*(*)
      integer *2 Matchlen, i
c
      i = index( Gset, Gsubset )
      ilgins = mod( i-1, Matchlen ) .eq. 0 .and. i .le. len( Gset )
      return
      end



      SUBROUTINE ILGSW2( I, J )
c
c       Swaps the values of 2 2 byte values
c
      implicit none
      integer *2 i, j, k
c
      k = i
      i = j
      j = k
      return
      end



      SUBROUTINE ILGSW4( I, J )
c
c       Swaps the values of 2 4 byte values
c
      implicit none
      integer *4 i, j, k
c
      k = i
      i = j
      j = k
      return
      end
