/*
    Copyright (C) 1997  Peter Vernon Lanspeary

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* ************************************************************************ */
/*                                                                          */
/*                                CLIPPING                                  */
/*                                                                          */
/* Code for two dimensional rectangular clipping and 3 dimensional          */
/* rectangular clipping .                                                   */
/*                                                                          */
/* External functions used : fprintf(), free(), malloc()                    */
/*                                                                          */
/* ************************************************************************ */

#include "system.h"     /* system dependent stuff */
#include POINTER_HEADER
#include ERROR_HEADER
#include CLIP_HEADER
#include <stdio.h>

/* ************************************************************************ */
/*                                                                          */
/*                               Definitions                                */
/*                                                                          */
/* ************************************************************************ */

/*
 *    secret structure identifiers
 *  The purpose of the structure identifier is to prevent the programmer from
 *  calling a clipping routine with an inappropriate clip structure pointer .
 *  These are formed by concatenating the least significant hexadecimal digit
 *  from each character of the structure keyword .
 */
#define CLIP2R_ID  0x3c22  /* CL2R two dimensional rectangular clipping     */
#define CLIP3R_ID  0x3c32  /* CL3R three dimensional rectangular clipping   */
#define CLIP3P_ID  0x3c30  /* CL3P three dimensional perspective clipping   */

/*     bit position definitions for the clipping edge-exceeded flag         */
#define LEFT    0   /* set bit 0 if the left clipping limit is exceeded     */
#define RIGHT   1   /* set bit 1 if the right clipping limit is exceeded    */
#define BOTTOM  2   /* set bit 2 if the bottom clipping limit is exceeded   */
#define TOP     3   /* set bit 3 if the top clipping limit is exceeded      */
#define BACK    4   /* set bit 5 if the back clipping limit is exceeded     */
#define FRONT   5   /* set bit 4 if the front clipping limit is exceeded    */


/* ************************************************************************ */
/*                                                                          */
/*                                  Macros                                  */
/*                                                                          */
/* ************************************************************************ */

/*
 *  CATCH_NULL(pointer,caller,err_type)
 ****************************************
 *
 *      If "pointer" has a NULL value, an "Arg is NULL pointer" error message
 * is generated and the function is aborted . Otherwise, nothing is done .
*/
#define CATCH_NULL(pointer,caller,err_type) \
  if ( ! (pointer) )    \
      return (err_type) make_error( ERR_ABORT, caller, err_null )

/*
 *  CLIP2R_EDGE( x, y )
 ************************
 *
 *    Determine which edges of the two dimensional rectangular clipping
 *    boundary are crossed . The returned result is a bit switch .
*/
#define CLIP2R_EDGE(x,y)   ( ( x < clip_ptr->cl_min._x ) << LEFT   ) |  \
                           ( ( x > clip_ptr->cl_max._x ) << RIGHT  ) |  \
                           ( ( y < clip_ptr->cl_min._y ) << BOTTOM ) |  \
                           ( ( y > clip_ptr->cl_max._y ) << TOP    )
/*
 *  CLIP3R_EDGE( x, y, z )
 ***************************
 *
 *    Determine which edges of the two dimensional rectangular clipping
 *    boundary are crossed . The returned result is a bit switch .
*/
#define CLIP3R_EDGE(x,y,z)  ( ( x < clip_ptr->cl_min._x ) << LEFT   ) |   \
                            ( ( x > clip_ptr->cl_max._x ) << RIGHT  ) |   \
                            ( ( y < clip_ptr->cl_min._y ) << BOTTOM ) |   \
                            ( ( y > clip_ptr->cl_max._y ) << TOP    ) |   \
                            ( ( z < clip_ptr->cl_min._z ) << BACK   ) |   \
                            ( ( z > clip_ptr->cl_max._z ) << FRONT  )


/*
 *  R_INTERSECT( subj, src, lim )
 **********************************
 *
 *    Calculates `subj' ( = _x, _y or _z ) the intersection of the line
 *    connecting the clipped endpoints and the rectanglular clipping plane
 *    `lim'.`src', where `lim' = cl_min or cl_max and `src' = _x, _y or _z .
*/
#define R_INTERSECT(subj,src,lim)       clip_ptr->cl_out1.subj +              \
                         ( clip_ptr->lim.src - clip_ptr->cl_out1.src ) *      \
                         ( clip_ptr->cl_out2.subj - clip_ptr->cl_out1.subj )/ \
                         ( clip_ptr->cl_out2.src - clip_ptr->cl_out1.src )

/*
 *  SWAPF( a, b )
 *****************
 *
 *    swaps the two double point values `a' and `b'
*/
#define SWAPF(a,b)       { double temp = b; b = a; a = temp; }


/* ************************************************************************ */
/*                                                                          */
/*                            Clipping routines                             */
/*                                                                          */
/* ************************************************************************ */

static char
    VAR_PTR err_alloc = "Not enough memory",
    VAR_PTR err_null  = "Arg is NULL pointer",
    VAR_PTR err_notc  = "Arg is not a clip_area",
    VAR_PTR lab_clip  = "clip",
    VAR_PTR lab_def2  = "def_2rclip",
    VAR_PTR lab_def3  = "def_3rclip",
    VAR_PTR lab_clear = "clear_clip",
    VAR_PTR lab_off   = "off_clip",
    VAR_PTR lab_on    = "on_clip",
    VAR_PTR lab_what  = "lab_what";
char *malloc();
int free();

boolean clip( coord, clip_ptr )
/*
 *     clip the line drawn to the coordinate `coord' from the previous point
 *     passed through this clipping routine with the same pointer `clip_ptr'.
 *     `clip_ptr' is a pointer to a structure which contains the clipped
 *     output coordinates and all the information used by `clip' .
 *     The returned result is TRUE if a part of the line segment falls in
 *     the clipped region . If the line segment is entirely outside the
 *     clipped region, or `clip_ptr' is a NUL pointer, the returned result is
 *     FALSE .
 *
 *  Arguments
*/
coord_triple VAR_PTR coord;     /* data point to be clipped                 */
clip_area    VAR_PTR clip_ptr;  /* pointer to be clipped                    */
  {
    int edge_1, edge_2, edge;   /* specifies which clipping are exceeded    */
                                /*   each of the line segment endpoints     */
    coord_triple new;           /* new calculated segment endpoint          */
    CATCH_NULL( clip_ptr && coord, lab_clip, boolean );
    if ( clip_ptr->id_struct == CLIP2R_ID )
      {
        if ( clip_ptr->cl_on )
          {
            clip_ptr->cl_out1._x = clip_ptr->cl_prev._x;
            clip_ptr->cl_out1._y = clip_ptr->cl_prev._y;
            clip_ptr->cl_prev._x = clip_ptr->cl_out2._x = coord->_x;
            clip_ptr->cl_prev._y = clip_ptr->cl_out2._y = coord->_y;
            edge_1 = CLIP2R_EDGE( clip_ptr->cl_out1._x,
                                  clip_ptr->cl_out1._y );
            edge_2 = CLIP2R_EDGE( coord->_x, coord->_y );
            while( edge_1 | edge_2 )
              {
                if ( edge_1 & edge_2 )
                    return FALSE;
                if ( edge_1 )
                  edge = edge_1;
                else
                    edge = edge_2;
                if ( edge & ( 1 << LEFT ) )
                  {
                    new._x = clip_ptr->cl_min._x;
                    new._y = R_INTERSECT( _y, _x, cl_min );
                  }
                else if ( edge & ( 1 << RIGHT ) )
                  {
                    new._x = clip_ptr->cl_max._x;
                    new._y = R_INTERSECT( _y, _x, cl_max );
                  }
                else if ( edge & ( 1 << BOTTOM ) )
                  {
                    new._x = R_INTERSECT( _x, _y, cl_min );
                    new._y = clip_ptr->cl_min._y;
                  }
                else if ( edge & ( 1 << TOP ) )
                  {
                    new._x = R_INTERSECT( _x, _y, cl_max );
                    new._y = clip_ptr->cl_max._y;
                  }
                if ( edge_1 == edge )
                  {
                    clip_ptr->cl_out1._x = new._x;
                    clip_ptr->cl_out1._y = new._y;
                    edge_1 = CLIP2R_EDGE( new._x, new._y );
                  }
                else
                  {
                    clip_ptr->cl_out2._x = new._x;
                    clip_ptr->cl_out2._y = new._y;
                    edge_2 = CLIP2R_EDGE( new._x, new._y );
                  }
              }
          }
        else
          {
            clip_ptr->cl_out1._x = clip_ptr->cl_prev._x;
            clip_ptr->cl_out1._y = clip_ptr->cl_prev._y;
            clip_ptr->cl_prev._x = clip_ptr->cl_out2._x = coord->_x;
            clip_ptr->cl_prev._y = clip_ptr->cl_out2._y = coord->_y;
          }
        return TRUE;
      }
    else if ( clip_ptr->id_struct == CLIP3R_ID )
      {
        if ( clip_ptr->cl_on )
          {
            clip_ptr->cl_out1 = clip_ptr->cl_prev;
            clip_ptr->cl_prev = clip_ptr->cl_out2 = *coord;
            edge_1 = CLIP3R_EDGE( clip_ptr->cl_out1._x,
                         clip_ptr->cl_out1._y, clip_ptr->cl_out1._z );
            edge_2 = CLIP3R_EDGE( coord->_x, coord->_y, coord->_z );
            while( edge_1 | edge_2 )
              {
                if ( edge_1 & edge_2 )
                    return FALSE;
                if ( edge_1 )
                  edge = edge_1;
                else
                    edge = edge_2;
                if ( edge & ( 1 << LEFT ) )
                  {
                    new._x = clip_ptr->cl_min._x;
                    new._y = R_INTERSECT( _y, _x, cl_min );
                    new._z = R_INTERSECT( _z, _x, cl_min );
                  }
                else if ( edge & ( 1 << RIGHT ) )
                  {
                    new._x = clip_ptr->cl_max._x;
                    new._y = R_INTERSECT( _y, _x, cl_max );
                    new._z = R_INTERSECT( _z, _x, cl_max );
                  }
                else if ( edge & ( 1 << BOTTOM ) )
                  {
                    new._x = R_INTERSECT( _x, _y, cl_min );
                    new._y = clip_ptr->cl_min._y;
                    new._z = R_INTERSECT( _z, _y, cl_min );
                  }
                else if ( edge & ( 1 << TOP ) )
                  {
                    new._x = R_INTERSECT( _x, _y, cl_max );
                    new._y = clip_ptr->cl_max._y;
                    new._z = R_INTERSECT( _z, _y, cl_max );
                  }
                else if ( edge & ( 1 << BACK ) )
                  {
                    new._x = R_INTERSECT( _x, _z, cl_min );
                    new._y = R_INTERSECT( _y, _z, cl_min );
                    new._z = clip_ptr->cl_min._z;
                  }
                else if ( edge & ( 1 << FRONT ) )
                  {
                    new._x = R_INTERSECT( _x, _z, cl_max );
                    new._y = R_INTERSECT( _y, _z, cl_max );
                    new._z = clip_ptr->cl_max._z;
                  }
                if ( edge_1 == edge )
                  {
                    clip_ptr->cl_out1 = new;
                    edge_1 = CLIP3R_EDGE( new._x, new._y, new._z );
                  }
                else
                  {
                    clip_ptr->cl_out2 = new;
                    edge_2 = CLIP3R_EDGE( new._x, new._y, new._z );
                  }
              }
          }
        else
          {
            clip_ptr->cl_out1 = clip_ptr->cl_prev;
            clip_ptr->cl_prev = clip_ptr->cl_out2 = *coord;
          }
        return TRUE;
      }
    return make_error( ERR_ABORT, lab_clip, err_notc );
  }


boolean move_clip( coord, clip_ptr )
/*
 *     Update the starting point of the line segment to be clipped .
 *     `clip_ptr' is a pointer to a structure which contains the clipped
 *     output coordinates and all the information used by `g_use_clip' .
 *
 *  Arguments
*/
coord_triple VAR_PTR coord;     /* new starting point of line segment    */
clip_area    VAR_PTR clip_ptr;  /* pointer to be clipped                 */
  {
    CATCH_NULL( clip_ptr && coord, lab_clip, boolean );
    if ( clip_ptr->id_struct == CLIP2R_ID )
      {
        clip_ptr->cl_prev._x = coord->_x;
        clip_ptr->cl_prev._y = coord->_y;
        return TRUE;
      }
    else if ( clip_ptr->id_struct == CLIP3R_ID )
      {
        clip_ptr->cl_prev = *coord;
        return TRUE;
      }
    return make_error( ERR_ABORT, (char VAR_PTR) "move_clip", err_notc );
  }

clip_area VAR_PTR def_2rclip( lower_left, upper_right, first )
/*
 * Set the initial and boundary conditions for two dimensional rectangular
 * clipping . The returned value is the pointer to the `clip_area' data
 * structure .
 *
 *   Arguments
*/
coord_pair VAR_PTR lower_left,  /* lower left clipping limit coordinate  */
           VAR_PTR upper_right, /* upper right clipping limit coordinate */
           VAR_PTR first;       /* first coordinate position             */
  {
    clip_area VAR_PTR clip_ptr;
    CATCH_NULL( lower_left && upper_right && first, lab_def2,
        clip_area VAR_PTR );
    clip_ptr = (clip_area VAR_PTR) malloc( sizeof( clip_area ) );
    if ( clip_ptr  )
      {
        if ( lower_left->_x > upper_right->_x )
            SWAPF( lower_left->_x, upper_right->_x );
        if ( lower_left->_y > upper_right->_y )
            SWAPF( lower_left->_y, upper_right->_y );
        clip_ptr->id_struct = CLIP2R_ID;
        clip_ptr->cl_out1._x = clip_ptr->cl_out2._x = clip_ptr->cl_min._x =
                               lower_left->_x;
        clip_ptr->cl_out1._y = clip_ptr->cl_out2._y = clip_ptr->cl_min._y =
                               lower_left->_y;
        clip_ptr->cl_max._x  = upper_right->_x;
        clip_ptr->cl_max._y  = upper_right->_y;
        clip_ptr->cl_prev._x = first->_x;
        clip_ptr->cl_prev._y = first->_y;
        clip_ptr->cl_on  = TRUE;
        return clip_ptr;
      }
    return (clip_area VAR_PTR) make_error( ERR_ABORT, lab_def2, err_alloc );
  }


clip_area VAR_PTR def_3rclip( lower_left, upper_right, first )
/*
 * Set the initial and boundary conditions for three dimensional rectangular
 * clipping . The returned value is the pointer to the `clip_area' data
 * structure .
 *
 *   Arguments
*/
coord_triple VAR_PTR lower_left,    /* lower left clipping limit coordinate  */
             VAR_PTR upper_right,   /* upper right clipping limit coordinate */
             VAR_PTR first;         /* first coordinate position             */
  {
    clip_area VAR_PTR clip_ptr;
    CATCH_NULL( lower_left && upper_right && first, lab_def3,
        clip_area VAR_PTR );
    clip_ptr = (clip_area VAR_PTR) malloc( sizeof( clip_area ) );
    if ( clip_ptr  )
      {
        if ( lower_left->_x > upper_right->_x )
            SWAPF( lower_left->_x, upper_right->_x );
        if ( lower_left->_y > upper_right->_y )
            SWAPF( lower_left->_y, upper_right->_y );
        if ( lower_left->_z > upper_right->_z )
            SWAPF( lower_left->_z, upper_right->_z );
        clip_ptr->id_struct = CLIP3R_ID;
        clip_ptr->cl_out1 = clip_ptr->cl_out2 = clip_ptr->cl_min = *lower_left;
        clip_ptr->cl_max = *upper_right;
        clip_ptr->cl_prev = *first;
        clip_ptr->cl_on  = TRUE;
        return clip_ptr;
      }
    return (clip_area VAR_PTR) make_error( ERR_ABORT, "def3r_clip", err_alloc );
  }


boolean clear_clip( clip_ptr )
/*
 * Erase the data structure pointed at by `clip_ptr'
 *
 *  Argument
*/
clip_area VAR_PTR clip_ptr; /* pointer to clip data structure */
  {
    CATCH_NULL( clip_ptr, lab_clear, boolean );
    if ( clip_ptr->id_struct == CLIP2R_ID ||
         clip_ptr->id_struct == CLIP3R_ID )
      {
        clip_ptr->id_struct = 0;
        (void) free( (char VAR_PTR) clip_ptr );
        return TRUE;
      }
    return make_error( ERR_ABORT, lab_clear, err_alloc );
  }


boolean off_clip( clip_ptr )
/*
 * Temporarily disable clipping by the structure `clip_ptr'
 *
 * Argument
*/
clip_area VAR_PTR clip_ptr; /* pointer to clip data structure */
  {
    CATCH_NULL( clip_ptr, lab_off, boolean );
    if ( clip_ptr->id_struct == CLIP2R_ID ||
         clip_ptr->id_struct == CLIP3R_ID )
        return ! ( clip_ptr->cl_on = FALSE );
    return make_error( ERR_WARN, lab_off, err_alloc );
  }


boolean on_clip( clip_ptr )
/*
 * Restore clipping by the structure `clip_ptr'
 *
 * Argument
*/
clip_area VAR_PTR clip_ptr; /* pointer to clip data structure */
  {
    CATCH_NULL( clip_ptr, lab_on, boolean );
    if ( clip_ptr->id_struct == CLIP2R_ID ||
         clip_ptr->id_struct == CLIP3R_ID )
       return clip_ptr->cl_on = TRUE;
    return make_error( ERR_ABORT, lab_on, err_alloc );
  }


boolean what_clip( file_ptr, clip_ptr )
/*
 * dumps the contents of the clip structure `clip_ptr'
 *
 * Arguments
*/
FILE *file_ptr; /* pointer to output file */
clip_area VAR_PTR clip_ptr; /* pointer to clip data structure */
  {
    CATCH_NULL( file_ptr && clip_ptr, lab_what, boolean );
    switch( clip_ptr->id_struct )
      {
        case CLIP2R_ID:
            (void) fprintf( file_ptr, "Contents of the Clip Structure\n" );
            (void) fprintf( file_ptr, "   two dimensional rectangular clip\n" );
            (void) fprintf( file_ptr, "   first clipped point (%g,%g)\n",
                clip_ptr->cl_out1._x, clip_ptr->cl_out1._y );
            (void) fprintf( file_ptr, "   second clipped point (%g,%g)\n",
                clip_ptr->cl_out2._x, clip_ptr->cl_out2._y );
            (void) fprintf( file_ptr,
                "   clipping limits for abscissa %g and %g\n",
                clip_ptr->cl_min._x, clip_ptr->cl_max._x );
            (void) fprintf( file_ptr,
                "   clipping limits for ordinate %g and %g\n",
                clip_ptr->cl_min._y, clip_ptr->cl_max._y );
            (void) fprintf( file_ptr, "   previous position (%g,%g)\n",
                clip_ptr->cl_prev._x, clip_ptr->cl_prev._y );
            switch( clip_ptr->cl_on )
              {
                case FALSE: (void) fprintf( file_ptr,
                                            "   clipping is switched off\n" );
                            break;
                default:    (void) fprintf( file_ptr,
                                            "   clipping is switched on\n" );
                            break;
              }
            return TRUE;
          case CLIP3R_ID:
            (void) fprintf( file_ptr, "contents of the clip structure\n" );
            (void) fprintf( file_ptr,
                            "   three dimensional rectangular clip\n" );
            (void) fprintf( file_ptr, "   first clipped point (%g,%g,%g)\n",
                clip_ptr->cl_out1._x, clip_ptr->cl_out1._y,
                clip_ptr->cl_out1._z );
            (void) fprintf( file_ptr, "   second clipped point (%g,%g,%g)\n",
                clip_ptr->cl_out2._x, clip_ptr->cl_out2._y,
                clip_ptr->cl_out2._z );
            (void) fprintf( file_ptr,
                "   clipping limits for abscissa %g and %g\n",
                clip_ptr->cl_min._x, clip_ptr->cl_max._x );
            (void) fprintf( file_ptr,
                "   clipping limits for ordinate %g and %g\n",
                clip_ptr->cl_min._y, clip_ptr->cl_max._y );
            (void) fprintf( file_ptr,
                "   clipping limits for depth %g and %g\n",
                clip_ptr->cl_min._z, clip_ptr->cl_max._z );
            (void) fprintf( file_ptr, "   previous position (%g,%g,%g)\n",
                clip_ptr->cl_prev._x, clip_ptr->cl_prev._y,
                clip_ptr->cl_prev._z );
            switch( clip_ptr->cl_on )
              {
                case FALSE: (void) fprintf( file_ptr,
                                            "   clipping is switched off\n" );
                            break;
                default:    (void) fprintf( file_ptr,
                                            "   clipping is switched on\n" );
                            break;
              }
            return TRUE;
      }
    return make_error( ERR_ABORT, lab_what, err_notc );
  }


