/*
    Copyright (C) 1997  Peter Vernon Lanspeary

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* ************************************************************************ */
/*                                                                          */
/*                   Common code for graphics device drivers                */
/*                                                                          */
/* This source file contains the character font handling functions, and     */
/* string processing functions and other miscellaneous functions required   */
/* by every device driver.                                                  */
/*                                                                          */
/* ************************************************************************ */

#include "system.h"
#include <stdio.h>
#include <stdlib.h>        /* required by getenv() */
#include DEVICE_G_HEADER
#include FONT_SOURCE_HEADER

#define MAX_WORD_LEN 15     /* max length of word in the options list   */
#define NO_CHARS     96     /* maximum number of ASCII font characters  */
#define FONT_HEIGHT 10.0    /* height of characters in font directory   */
#define FONT_WIDTH   6.0    /* width of characters in font directory    */
#define FONT_GAP     3.0    /* gap between characters */
#define PI          3.141592653589793238462643

/*
 *  CATCH_NULL(pointer,caller,err_type)
 ****************************************
 *      If "pointer" has a NULL value, an "Arg is NULL pointer" error message
 * is generated and the function is aborted . Otherwise, nothing is done .
 *      "pointer"   is the pointer to be tested,
 *      "caller"    is a character string naming the function
 *      "err_type"  is the type of the returned zero value which indicates an
 *                    error condition
*/
#define CATCH_NULL(pointer,caller,err_type) \
  if ( ! (pointer) )    \
      return (err_type) make_error( ERR_ABORT, caller, err_null )

/*
 *  MAX( x, y )
 ****************
 *      The returned value is the maximum of `x' and `y' .
*/
#define MAX(x,y)        ( (x) > (y) ? (x) : (y) )

/*
 *  LIMIT( x, min, max )
 *************************
 *      Limits the value of `x' to between `min' and `max'
*/
#define LIMIT(x,min,max)  (x) <= (min) ? (min) : ( (x) >= (max) ? (max) : (x) )


/*
 *      Fonts -- declared as external in each driver source file
*/
StrokeFont asc_font,  /* ASCII font */
           sym_font;  /* symbol font */

static char
  VAR_PTR err_null   = "Arg is NULL pointer";
static int
    old_arc = -1,           /* old arc number for bold t_ch_def      */
    old_font_index = -1,    /* old character index for bold t_ch_def */
    wob_passes = 1;         /* number of passes for bold t_ch_def    */
static coord_pair
    wob_offset;             /* wobble offset for bold t_ch_def       */


/* char *malloc(); */
/* int free(); */
double fabs(), cos(), sin();


/* ************************************************************************ */
/*                                                                          */
/*                      Device registration handlers                        */
/*                                                                          */
/* ************************************************************************ */

/*
 * Mechanism for registering drivers with the font machinery . This should be
 * done with an infinitely extensible linked list, but I can't be bothered
 * just yet .
*/
#define MAX_DEVICES 16      /* maximum number of devices which can   */
                            /* be used simultaneously                */

static Register_Elem d_regist[MAX_DEVICES]; /* register of devices in use */
static int           n_devices = 0;         /* number of devices in use   */


static Register_Elem VAR_PTR find_regist( file_ptr, type, data )
/*
 *  Find the registration entry of the device with ``file_ptr'' or ``type'' .
 * If the ``file_ptr'' argument is not a NULL pointer, the register is searched
 * for the ``file_ptr'' file pointer . If ``file_ptr'' is a NULL pointer, the
 * register is searched for the first device of type ``type'' which was
 * registered . The returned value is
 *  a) the pointer to the registration entry of the device -- if it is found,
 *  b) the special value T_NOT_REGIST if the device is not registered,
 *  c) a NULL pointer if an error occurs .
 * If an entry is found in register, the other available arguments are checked
 * against the entry, and an error is generated if one or more of the arguments
 * are not matched by the register entry .
*/
FILE *file_ptr;
              /* pointer to the file to which the graphics is written and   */
              /* which identifies the graphics device .                     */
int  type;    /* defines the set of driver functions used; is one of        */
              /* VISUAL500_TERMINAL, HERCULES_GRAPHICS, SUN_3_WORKSTATION   */
              /* HEWLETT_GL_PLOTTER, POSTSCRIPT_PRINTER, DUMMY_GRAPHICS_DEV */
              /* The following drivers are not yet available                */
              /* ENHANCED_GRAPHICS,  X11_WINDOWS, DOT_MATRIX_PRINTER .      */
char VAR_PTR data;
              /* pointer to a data structure which completely describes the */
              /* state of the graphics device .                             */
  {
    char VAR_PTR name = "t_register";
    Register_Elem VAR_PTR reg_ptr,
                  VAR_PTR end_ptr;
/*
 *        Use ``file_ptr'' to identify the device . Examine the output file
 *        pointer of every registered device to see if it matches ``file_ptr'' .
 *        If the search is successful, check that the device type and pointer to
 *        the graphics status data are the same as the argument values .
*/
    if ( file_ptr )
        for ( reg_ptr = d_regist, end_ptr = d_regist + n_devices;
              reg_ptr < end_ptr; reg_ptr++ )
          {
            if ( reg_ptr->dev_file == file_ptr )
              {
                if ( type && reg_ptr->dev_type != type )
                    return (Register_Elem VAR_PTR) make_error( ERR_ABORT, name,
                       (char VAR_PTR) "Already registered, wrong device type" );
                if ( data && reg_ptr->dev_data != data )
                    return (Register_Elem VAR_PTR) make_error( ERR_ABORT, name,
                      (char VAR_PTR) "Already registered, wrong data pointer" );
                return reg_ptr;
              }
          }
/*
 *        ``file_ptr'' is not supplied, so use the device ``type'' instead .
 *        If the search is successful, check that the pointer to the graphics
 *        status data is the same as the argument value ``data''.
*/
    else if ( type )
        for ( reg_ptr = d_regist, end_ptr = d_regist + n_devices;
              reg_ptr < end_ptr; reg_ptr++ )
          {
            if ( reg_ptr->dev_type == type )
              {
                if ( data && reg_ptr->dev_data != data )
                    return (Register_Elem  VAR_PTR) make_error( ERR_ABORT, name,
                      (char VAR_PTR) "Already registered, wrong data pointer" );
                return reg_ptr;
              }
          }
/*
 *        Too many NULL arguments
*/
    else
        return (Register_Elem VAR_PTR) make_error( ERR_ABORT, name, err_null );
/*
 *        Special return value if the device is not found .
*/
    return T_NOT_REGIST;
  }


char VAR_PTR t_register( action, file_ptr, type, data )
/*
 *     Performs the registration function ``action'' on the device driver with
 * output file *``file_ptr'' . The registration mechanism is provided so that
 * more than one device and more than one device of any one kind may be used .
 * When ``t_begin()'' is used to reselect a device which is already in use,
 * the registration mechanism is used to retrieve the graphics state of the
 * device .
 *
 *     The available ``actions'' are
 * T_REGISTER  : Register the device .
 *     All arguments are used and required . The returned value is TRUE if
 *      a) the device has been registered successfully, or
 *      b) the device is already registered and the ``type'' and ``data''
 *         arguments are consistent with the existing registration .
 *     If the device register is full, or the arguments are inconsistent with
 *     an existing registration, the returned result is FALSE . A device can
 *     registered only once .
 * T_DEREGIST : Deregister the device .
 *     Only the ``file_ptr'' argument is used . The registration entry is
 *     removed . The returned value is TRUE if the arguments supplied (i.e. non
 *     NULL) match a registered device . One of ``file_ptr'' and ``type'' must
 *     be supplied (i.e. non NULL) .
 * T_ISREGIST : Is the device registered ?
 *     If the ``file_ptr'' argument is not a NULL pointer, the register is
 *     searched for the ``file_ptr'' file pointer . If ``file_ptr'' is a
 *     NULL pointer, the register is searched for the first device of type
 *     ``type'' which was registered . If an entry is found in register, the
 *     other available arguments are checked against the entry, and an error
 *     is generated if one or more of the arguments are not matched by the
 *     register entry .
 * T_REG_FILE : Returns file pointer or device handle
 *     Only the ``type'' argument is used to identify the device . Other
 *     arguments are ignored . The returned value identifies the first
 *     registered device of type ``type'' .
 * T_REG_TYPE : Return the device type .
 *     Only the ``file_ptr'' argument is used to identify the device . Other
 *     arguments are ignored .
 * T_REG_DATA : Return a pointer to the graphics status data .
 *     Only the ``file_ptr'' argument is used to identify the device . Other
 *     arguments are ignored . The graphics status data is private to the
 *     driver which generated it, is opaque and so not useful for any other
 *     purpose ( i.e. "no user servicable components" ) .
 * T_REG_COUNT : Returns the total number of registered devices .
*/
int action;   /* what the function is to do -- is one of        */
              /* T_REGISTER, T_DEREGIST, T_REG_FILE, T_REG_TYPE,*/
              /* T_REG_DATA, T_REG_COUNT, T_ISREGIST            */
FILE *file_ptr;
              /* pointer to the file to which the graphics is written and   */
              /* which identifies the graphics device .                     */
int  type;    /* defines the set of driver functions used; is one of        */
              /* VISUAL500_TERMINAL, HERCULES_GRAPHICS, SUN_3_WORKSTATION   */
              /* HEWLETT_GL_PLOTTER, POSTSCRIPT_PRINTER, DUMMY_GRAPHICS_DEV */
              /* The following drivers are not yet available                */
              /* ENHANCED_GRAPHICS,  X11_WINDOWS, DOT_MATRIX_PRINTER .      */
char VAR_PTR data;
              /* pointer to a data structure which completely describes the */
              /* state of the graphics device .                             */
  {
    char VAR_PTR name = "t_register";
    Register_Elem VAR_PTR reg_ptr;
    if ( n_devices < 0 )
        n_devices = 0;
/*
 *      T_REGISTER : REGISTER THE DEVICE
 *      ================================
*/
    if ( action == T_REGISTER )
      {
        CATCH_NULL( file_ptr && data && type, name, char VAR_PTR );
        reg_ptr = find_regist( file_ptr, type, data );
        if ( reg_ptr == T_NOT_REGIST )
          {
/*
 *      Abort if the register is full .
*/
            if ( n_devices >= MAX_DEVICES )
                return (char VAR_PTR) make_error( ERR_ABORT, name,
                    (char VAR_PTR) "Too many drivers" );
/*
 *      Complete registration; put the new device on the end of the list
*/
            reg_ptr = d_regist + n_devices++;
            reg_ptr->dev_file = file_ptr;
            reg_ptr->dev_type = type;
            reg_ptr->dev_data = data;
            return (char VAR_PTR) TRUE;
          }
        else if ( reg_ptr )
            return (char VAR_PTR) TRUE;
      }
    else
/*
 *      For other actions, check arguments to ``t_register()'' and mask
 *      out arguments to ``find_regist()'' if required .
*/
      {
        switch ( action )
          {
            case T_DEREGIST :
            case T_ISREGIST :
                break;
            case T_REG_FILE :
                file_ptr = (FILE *) 0;
                data = (char VAR_PTR) 0;
                break;
            case T_REG_TYPE :
            case T_REG_DATA :
                type = NONEXISTENT_DEVICE;
                data = (char VAR_PTR) 0;
                break;
            case T_REG_COUNT :
                return (char VAR_PTR) n_devices;
            default :
                return (char VAR_PTR) make_error( ERR_ABORT, name,
                    (char VAR_PTR) "Unknown action argument" );
          }
        reg_ptr = find_regist( file_ptr, type, data );
        if ( reg_ptr != FALSE && reg_ptr != T_NOT_REGIST )
            switch ( action )
              {
                case T_DEREGIST :               /* deregister the device */
                  {
                    Register_Elem VAR_PTR next_ptr = reg_ptr + 1,
                                  VAR_PTR last_ptr = reg_ptr + n_devices;
                    while ( next_ptr <= last_ptr )
                        *reg_ptr++ = *next_ptr++;
                    n_devices--;
                    return (char VAR_PTR) TRUE;
                  }
                case T_ISREGIST :               /* is the device registered ? */
                    return (char VAR_PTR) TRUE;
                case T_REG_FILE :               /* return the file pointer */
                    return (char VAR_PTR) reg_ptr->dev_file;
                case T_REG_TYPE :               /* return the device type  */
                    return (char VAR_PTR) reg_ptr->dev_type;
                case T_REG_DATA :               /* return graphics status ptr */
                    return reg_ptr->dev_data;
              }
      }
    return (char VAR_PTR) FALSE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                   Character font handling functions                      */
/*                                                                          */
/* ************************************************************************ */


boolean reset_font( stroke, default_height )
/*
 *  Resets the text attributes of stroke-based characters to the default
 * values .
*/
Text_Attribute VAR_PTR stroke;  /* pointer to structure of text attributes */
double default_height;          /* default character height */
  {
    CATCH_NULL( stroke, (char VAR_PTR) "reset_font", boolean );
    stroke->height = default_height;
    stroke->aspect = 0.6;
    stroke->slant  = stroke->angle = 0.0;
    stroke->gap    = 0.3;
    (stroke->align).horiz = 'l';
    (stroke->align).vert  = 'b';
    stroke->bold   = FALSE;
    stroke->changed= TRUE;
    stroke->font_ptr  = & asc_font;
    stroke->b_radius = default_height / 20.0;
    stroke->b_passes = PI / 10.0 * default_height;
    return TRUE;
  }


#define NEG_ZERO -1.0e-4    /* smallest negative number allowed in the matrix */
#define POS_ZERO  1.0e-4    /* smallest positive number allowed in the matrix */

boolean map_font( stroke, pen_thick )
/*
 *      calculate the transformation matrix for drawing stroke characters
*/
Text_Attribute VAR_PTR stroke;  /* pointer to structure of text attributes */
double pen_thick;               /* pen thickness in pixels */
  {
    double sin_theta, cos_theta, st_aspect, st_height, st_slant;
    CATCH_NULL( stroke, (char VAR_PTR) "map_font", boolean );
/*
 *          preliminaries
*/
    st_aspect = stroke->aspect * stroke->font_ptr->font_hei /
        stroke->font_ptr->font_wid;
    st_height = stroke->height / stroke->font_ptr->font_hei;
    st_slant  = sin( stroke->slant ) / cos( stroke->slant );
/*
 *          calculate transformation matrix
*/
    stroke->map[0][0] = st_height * st_aspect *
        ( cos_theta = cos( stroke->angle ) );
    stroke->map[0][1] = st_height * ( st_slant * cos_theta -
        ( sin_theta = sin( stroke->angle ) ) );
    stroke->map[1][0] = st_height * st_aspect * sin_theta;
    stroke->map[1][1] = st_height * ( st_slant * sin_theta + cos_theta );
/*
 *          If the matrix coefficient is small, round it to zero .
 *          Graphics has limited resolution, you know .
*/
    if ( NEG_ZERO <= stroke->map[0][0] && stroke->map[0][0] <= POS_ZERO )
        stroke->map[0][0] = 0.0;
    if ( NEG_ZERO <= stroke->map[0][1] && stroke->map[0][1] <= POS_ZERO )
        stroke->map[0][1] = 0.0;
    if ( NEG_ZERO <= stroke->map[1][0] && stroke->map[1][0] <= POS_ZERO )
        stroke->map[1][0] = 0.0;
    if ( NEG_ZERO <= stroke->map[1][1] && stroke->map[1][1] <= POS_ZERO )
        stroke->map[1][1] = 0.0;
    stroke->changed = FALSE;
/*
 *          set wobble radius and number of passes for bold characters
*/
    stroke->b_radius =
        MAX( (double) ( pen_thick / 2 ), stroke->height / 20.0 );
    stroke->b_passes = MAX( 3.0, 2.0 * PI / pen_thick * stroke->b_radius );
    return TRUE;
  }

#undef NEG_ZERO
#undef POS_ZERO


int bold_font( stroke, point, index, txt_func )
/*
 *      draws a single character with `txt_func' t_stkchr or a character
 *  label string with `txt_func' t_txtstr specified by index `index' in the
 *  current selected font at position `point' according to the selected
 *  stroke character attributes .
*/
Text_Attribute VAR_PTR stroke;  /* pointer to structure of text attributes */
coord_pair     VAR_PTR point;   /* location of character */
char           VAR_PTR index;   /* index of character in font directory */
int            (* txt_func )(); /* pointer to function which draws the text */
  {
    int ok = TRUE;  /* success flag */
    CATCH_NULL( stroke && point && index, (char VAR_PTR) "bold_font", int );
/*
 *      set up wobble factor for bold characters
*/
    if ( stroke->bold )
      {
        double sin_wobble, cos_wobble,  /* sine & cosine of wobble angle   */
               temp;             /* temporary variable for wobble rotation */
        coord_pair bold_point,   /* adjusted location for boldface chars   */
                   wobble;       /* wobble (offset) for boldface           */
        int no_passes = stroke->b_passes;
                                 /* number of passes required for boldface */
/*
 *          set initial wobble angle & rotation angle
*/
        wobble._y = ( sin_wobble = 2.0 * PI / no_passes ) /
            no_passes;
/*
 *          initial wobble coordinate
*/
        wobble._x = stroke->b_radius * cos( wobble._y );
        wobble._y = stroke->b_radius * sin( wobble._y );
/*
 *          coefficients of wobble rotation matrix
*/
        cos_wobble = cos( sin_wobble );
        sin_wobble = sin( sin_wobble );
/*
 *      loop for generating boldface characters
*/
        while( no_passes-- )
          {
/*
 *                  rotate the wobble coordinate
*/
            temp = cos_wobble * wobble._x - sin_wobble * wobble._y;
            wobble._y = sin_wobble * wobble._x + cos_wobble * wobble._y;
            wobble._x = temp;
/*
 *                  generate new character location
*/
            bold_point._x = point->_x + wobble._x;
            bold_point._y = point->_y + wobble._y;
            ok &= (* txt_func)( index, index+1,
                                (coord_pair VAR_PTR) &bold_point );
          }
      }
    else
        ok = (* txt_func)( index, index+1, point );
    return ok;
  }


char VAR_PTR file_font( font )
/*
 *  Return the name of the file containing the font definition.
 *  The name of the files are defined by the environment variables
 *  G_ASCFNT : for the ascii character font
 *  G_SYMFNT : for the symbol character font
*/
StrokeFont VAR_PTR font;
  {
    char VAR_PTR name = "file_font",
         VAR_PTR file_name;
    CATCH_NULL( font, name, char VAR_PTR );
    if ( font == &asc_font )
      {
        if ( ( file_name = getenv( "G_ASCFNT" ) ) )
            return file_name;
        else
            return (char VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "No environment name G_ASCFNT" );
      }
    else if ( font == &sym_font )
      {
        if ( ( file_name = getenv( "G_SYMFNT" ) ) )
            return file_name;
        else
            return (char VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "No environment name G_SYMFNT" );
      }
    return (char VAR_PTR) make_error( ERR_ABORT, name,
        (char VAR_PTR) "Unknown font" );
  }


boolean clear_font( font )
/*
 * If no devices are registered, the memory used by the character font
 * ``font'' is de-allocated (freed) .
*/
StrokeFont VAR_PTR font;    /* (pointer to) character font */
  {
    char VAR_PTR name = "clear_font";
    StrokeChar VAR_PTR char_ptr,  /* moving pointer to character descriptions */
               VAR_PTR last_char; /* pointer to last character in the font    */
    CATCH_NULL( font, name, boolean );
    if ( n_devices == 0 )
      {
/*
 *      deallocate each character description
*/
        for ( char_ptr = font->font_dir,
              last_char = font->font_dir + font->font_siz;
              char_ptr < last_char; char_ptr++ )
          {
            if ( char_ptr->font_npts > 0 && char_ptr->font_coord )
                (void) free( (char VAR_PTR) char_ptr->font_coord );
          }
/*
 *      deallocate the array of pointers to character descriptions
 *      and clear every value in the font description
*/
        (void) free( (char VAR_PTR) font->font_dir );
        font->font_dir = (StrokeChar VAR_PTR) 0;
        font->font_got = FALSE;
        font->font_siz = font->font_hei = font->font_wid = 0;
      }
    return TRUE;
  }


int read_font( font, file_name, caller )
/*
 *       Reads a character font specification from the text file ``file_name'',
 * into memory and compiles a font description ``font'' which can contain a
 * maximum of ``NO_CHARS'' characters .
 * The stroke coordinates for each character are stored as an array of
 * x1, y1, x2, y2, x3, y3, ...  , which is pointed at by
 * ``font->font_dir[char_no].font_coord' .
*/
StrokeFont VAR_PTR font;    /* (pointer to) character font */
char VAR_PTR file_name,     /* name of file containing character fonts  */
     VAR_PTR caller;        /* name of calling function                 */
  {
    FILE *font_file;        /* file pointer for character font file     */
    char char_label;        /* character defined or ASCII equivalent    */
    int char_no,            /* character font directory index           */
        no_chars,           /* number of characters read                */
        no_coords,          /* number of coordinates defining the char  */
        nitems;             /* number of items obtained by fscanf       */
    float VAR_PTR coord_ptr,/* pointer to coordinates                   */
          displ;            /* vertical displacement of the font coords */
    double y_max, y_min,    /* top & bottom of the character            */
           y;               /* ordinate of point                        */
    StrokeChar
        VAR_PTR init_ptr,   /* moving pointer for initializing directory*/
        VAR_PTR last_ptr;   /* last pointer for initializing directory  */
    static StrokeChar init_char = { '\000', 0, 0.0, 0.0, (float VAR_PTR) 0 };
                            /* initial value of character description   */
    char VAR_PTR err_coord = "Cannot read font coordinate for  ",
         VAR_PTR err_alloc = "Cannot allocate memory",
         VAR_PTR name = "read_font";
    CATCH_NULL( font && file_name && caller, name, int );
/*
 *              Don't read it if you have already done it
*/
    if ( font->font_got )
        return make_error( ERR_ABORT, name,
            (char VAR_PTR) "clear_font before reading it again" );
    if ( ( font_file = fopen( (char *) file_name, "r" ) ) )
      {
/*
 *              Allocate character description directory,
 *              initialize font parameters
*/
        if ( ( font->font_dir = (StrokeChar VAR_PTR)
            malloc( (unsigned int) sizeof( StrokeChar ) * NO_CHARS ) ) )
          {
            font->font_siz = NO_CHARS;
            font->font_hei = FONT_HEIGHT;   /* height of characters */
            font->font_wid = FONT_WIDTH;    /* width of characters  */
            font->font_gap = FONT_GAP;      /* gap between characters */
/*
 *              initialize the character description directory
*/
            for ( init_ptr = font->font_dir,
                  last_ptr = font->font_dir + font->font_siz;
                  init_ptr < last_ptr; init_ptr++ )
                *init_ptr = init_char;
/*
 *              read character, number of coordinates and vertical offset
*/
            for( no_chars = 0;
                 3 == ( nitems = fscanf( font_file, "%c%*c%d%f", &char_label,
                     &no_coords, &displ ) ) && no_chars < font->font_siz;
                 no_chars++ )
              {
                y_min = 1.0e6;
                y_max = -1.0e6;
                char_no = LIMIT( char_label - 32, 0, font->font_siz - 1 );
                (font->font_dir)[ char_no ].font_char  = char_label;
                (font->font_dir)[ char_no ].font_npts  = no_coords;
                (font->font_dir)[ char_no ].font_shift = displ;
                if ( ! ( (font->font_dir)[ char_no ].font_coord = coord_ptr =
                      (float VAR_PTR) malloc( (unsigned int)
                          ( no_coords * 2 * sizeof( float ) ) ) ) )
                    break;
                for( ; no_coords--; coord_ptr += 2 )
                  {
                    if ( 2 != fscanf( font_file, "%f%lf", coord_ptr, &y ) )
                      {
                        err_coord[ 32 ] = char_label;
                        (void) fclose( font_file );
                        return (int) make_error( ERR_FATAL, caller, err_coord );
                      }
                    *( coord_ptr + 1 ) = y;
                    if ( ( y = fabs( y ) ) < y_min )
                        y_min = y;
                    if ( y > y_max )
                        y_max = y;
                  }
                (void) fscanf( font_file, "%*c" );
                (font->font_dir)[ char_no ].font_midpt =
                    ( y_min + y_max ) / 2.0 -
                    (font->font_dir)[ char_no ].font_shift;
              }
            if ( feof(font_file) )
              {
                (void) fclose( font_file );
                font->font_got = TRUE;
                return no_chars;
              }
            else if ( nitems != 3 )
                (void) make_error( ERR_FATAL, caller, (char VAR_PTR)
                    "Cannot read font character or number of coords" );
            else if ( ! coord_ptr )
                (void) make_error( ERR_FATAL, caller, err_alloc );
            else
                (void) make_error( ERR_FATAL, caller,
                    (char VAR_PTR) "Font directory full" );
            (void) fclose( font_file );
          }
        else
            (void) make_error( ERR_ABORT, name, err_alloc );
      }
    else
        (void) make_error( ERR_FATAL, caller,
            (char VAR_PTR) "Cannot open font file" );
    return 0;
  }


/* symbols for indexing from the character definition coordinate pointer ptr */
#define THIS_X 0    /* coordinate at ptr */
#define THIS_Y 1
#define NEXT_X 2    /* next coordinate after ptr */
#define NEXT_Y 3

char VAR_PTR define_font( stroke, pen_thick, what, character, index, point )
/*
 * This function allows the user to obtain the shape of any character defined
 * in the ASCII and SYMBOL fonts. With the exception of string alignment, the
 * transformations implied by t_txtset are applied to the returned coordinates.
*/
Text_Attribute
    VAR_PTR stroke; /* pointer to stroke-character attributes */
double pen_thick;   /* pen thickness -- in pixels */
int what;           /* specifies what to do (T_FT_* defined values) */
char character;     /* character to fetch from the font */
int index;          /* coordinate number   */
coord_pair
    VAR_PTR point;  /* returned coordinate */
  {
    char VAR_PTR name = "t_ch_def";
    double radius;
    int new_arc, font_index = character - 32;
    CATCH_NULL( stroke, (char VAR_PTR) "define_font", char VAR_PTR );
/*
 *           if necessary read character fonts
*/
    if ( ! stroke->font_ptr->font_got )
        if ( ! read_font( stroke->font_ptr, file_font( stroke->font_ptr),
                          name ) )
            return (char VAR_PTR) 0;
    switch( what )
      {
/*
 *      return the width & height of characters
*/
        case T_FT_SIZE:
            CATCH_NULL( point, name, char VAR_PTR );
            point->_x = ( point->_y = stroke->height ) * stroke->aspect;
            return (char VAR_PTR) point;
/*
 *      definition exists for specified character
*/
        case T_FT_EXIST:
            return (char VAR_PTR)
              ( 0 <= font_index && font_index < stroke->font_ptr->font_siz &&
              (stroke->font_ptr->font_dir)[font_index].font_char == character &&
              (stroke->font_ptr->font_dir)[font_index].font_npts > 0 );
/*
 *      number of coordinate pairs in character definition
*/
        case T_FT_LENGTH:
            if ( 0 <= font_index && font_index < stroke->font_ptr->font_siz )
                if ( (stroke->font_ptr->font_dir)[font_index].font_char ==
                         character &&
                       (stroke->font_ptr->font_dir)[font_index].font_npts > 0 )
                  {
                    if ( stroke->changed )
                        (void) map_font( stroke, pen_thick );
                    return (char VAR_PTR)
                      ( (stroke->font_ptr->font_dir)[font_index].font_npts
                      * ( stroke->bold ? stroke->b_passes : 1 ) );
                  }
            return (char VAR_PTR ) make_error( ERR_ABORT, name,
                (char VAR_PTR) "character unavailable" );
/*
 *      all remaining options require that the character shape be
 *      properly  defined
*/
        case T_FT_FIRST:
        case T_FT_LAST:
        case T_FT_MIDDLE:
        case T_FT_POINT:
            if ( 0 <= font_index && font_index < stroke->font_ptr->font_siz )
                if ( (stroke->font_ptr->font_dir)[font_index].font_char ==
                         character &&
                     (stroke->font_ptr->font_dir)[font_index].font_npts > 0 )
                  {
                    if ( stroke->bold )
                      {
                        new_arc = index /
                            (stroke->font_ptr->font_dir)[font_index].font_npts;
                        index %=
                            (stroke->font_ptr->font_dir)[font_index].font_npts;
                        if ( new_arc !=old_arc || old_font_index != font_index )
                          {
                            double angle;
                            wob_passes = stroke->b_passes;
                            angle = 2.0 * PI / wob_passes * new_arc;
                            wob_offset._x = stroke->b_radius * cos( angle );
                            wob_offset._y = stroke->b_radius * sin( angle );
                            old_arc = new_arc;
                            old_font_index = font_index;
                          }
                      }
                    else
                      {
                        wob_passes = 1;
                        new_arc   = 0;
                      }
/*
 *      all remaining options require a legal coordinate index
*/
      if ( 0 <= index &&
           index < (stroke->font_ptr->font_dir)[font_index].font_npts &&
           0 <= new_arc && new_arc < wob_passes )
        {
          float VAR_PTR ptr =
              (stroke->font_ptr->font_dir)[font_index].font_coord + index * 2;
          switch ( what )
            {
/*
 *      point is the first in a line segment
*/
              case T_FT_FIRST:
                  return (char VAR_PTR) ( index == 0 ||
                      *(ptr+THIS_X) < 0.0 || *(ptr+THIS_Y) < 0.0 );
/*
 *      point is the last in a line segment
 *          ( take care to avoid segmentation violation )
*/
              case T_FT_LAST:
                  if ( (stroke->font_ptr->font_dir)[font_index].font_npts - 1
                       == index )
                      return (char VAR_PTR) TRUE;
                  else
                      return (char VAR_PTR)
                          ( *(ptr+NEXT_X) < 0.0 || *(ptr+NEXT_Y) < 0.0 );
/*
 *      point is in the middle of a line segment
*/
              case T_FT_MIDDLE:
                  if ( (stroke->font_ptr->font_dir)[font_index].font_npts - 1
                       == index )
                      return (char VAR_PTR) FALSE;
                  else
                      return (char VAR_PTR) ( index != 0 &&
                          *(ptr+THIS_X) >= 0.0 && *(ptr+THIS_Y) >= 0.0 &&
                          *(ptr+NEXT_X) >= 0.0 && *(ptr+NEXT_Y) >= 0.0 );
/*
 *      coordinate of a point
*/
              case T_FT_POINT:
                  CATCH_NULL( point, name, char VAR_PTR );
                  if ( stroke->changed )
                      (void) map_font( stroke, pen_thick );
/*
 *              calculate coordinate
*/
                  point->_x = fabs( (double) *(ptr+THIS_X) );
                  point->_y = fabs( (double) *(ptr+THIS_Y) )
                      - (stroke->font_ptr->font_dir)[font_index].font_shift;
                  radius = stroke->map[0][0] * point->_x +
                           stroke->map[0][1] * point->_y;
                  point->_y = stroke->map[1][0] * point->_x +
                              stroke->map[1][1] * point->_y;
                  point->_x = radius;
                  if ( stroke->bold )  /* add the wobble offset for bold font */
                    {
                      point->_x += wob_offset._x;
                      point->_y += wob_offset._y;
                    }
                  return (char VAR_PTR) point;
            }
        }
      else
        return (char VAR_PTR) make_error( ERR_ABORT,
            name, (char VAR_PTR) "coordinate index out of range" );
/*
 *              return to old indentation
*/
                  }
                else
                    return (char VAR_PTR ) make_error( ERR_ABORT, name,
                         (char VAR_PTR) "character unavailable" );
            else
                return (char VAR_PTR) make_error( ERR_ABORT, name,
                     (char VAR_PTR) "not printable character" );
        default:
            return (char VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "Unknown attribute" );
      }
  }
#undef THIS_X
#undef THIS_Y
#undef NEXT_X
#undef NEXT_Y


/* ************************************************************************ */
/*                                                                          */
/*                      String handling functions                           */
/*                                                                          */
/* ************************************************************************ */

int t_strlen( string )
/*
 *      Measures the length of a `string' .
*/
char VAR_PTR string;
  {
    char VAR_PTR ptr;
    CATCH_NULL( string, (char VAR_PTR) "t_str_len", int );
    for( ptr = string; *ptr; ptr++ );
    return ptr - string;
  }


char VAR_PTR t_strcpy( str_out, str_in )
/*
 *        String Copy
 *    Copies string `str_in' to string `str_out' . The result is guarranteed
 *    to be NUL terminated .
*/
char VAR_PTR str_out,   /* destination string */
     VAR_PTR str_in;    /* source string */
  {
    char VAR_PTR dest = str_out;
    CATCH_NULL( str_out && str_in, (char VAR_PTR) "t_strcpy", char VAR_PTR );
    while( ( *dest++ = *str_in++ ) );
    return str_out;
  }


int t_stncmp( str_1, str_2, max_chars )
/*
 *      String Compare
 *   The function `strncmp' lexicographically compares a maximum of `n'
 * characters in string `str_1' with string `str_2', starting with the first
 * character and continuing until a difference is found or the end of a string
 * is reached. The returned value of `strncmp' is
 *   < 0  if `str_1' <  `str_2'
 *   = 0  if `str_1' == `str_2'
 *   > 0  if `str_1' >  `str_2'
*/
char VAR_PTR str_1, VAR_PTR str_2;  /* character strings to be compared */
int max_chars;                      /* maximum number of characters compared */
  {
    char VAR_PTR name = "t_strcmp";
    CATCH_NULL( str_1 && str_2, name, int );
    if ( max_chars <= 0 )
        max_chars = make_error( ERR_WARN, name,
            (char VAR_PTR) "Max. string length <= 0" );
    for( ; *str_1 == *str_2 && --max_chars; str_1++, str_2++ )
        if ( ! *str_1 )
          return 0;
    return *str_1 - *str_2;
  }


boolean t_isabbr( cstr, ctest, minlen )
/*
 *    The function `isabbr' tests if the character string `cstr' is a
 * legal abbreviation of the character string `ctest' .
*/
  char VAR_PTR cstr,  /* operand character string */
       VAR_PTR ctest; /* test string */
  int minlen;  /* minimum number of characters allowed */
               /* in the abbreviation */
  {
    boolean same;  /* strings are the same so far */

    if ( t_strlen( cstr ) <= t_strlen( ctest ) )
      {
        for( same = TRUE; *cstr && ( same &= *cstr++ == *ctest++ ); minlen-- );
        return( same && minlen <= 0 );
      }
    return( FALSE );
  }


boolean t_isspace( c )
/*
 * Is character 'c' SPACE, Horiz. Tab., Vert. Tab., newline, return,
 * or form-feed ?
*/
char c;
  {
    return c==' ' || c=='\t' || c=='\v' || c=='\n'  || c=='\r' || c=='\f';
  }


boolean t_parse_options( string, option_table, caller )
/*
 * This function parses ``string'' and assigns values to variables in
 * accordance with the rules specified in the ``option_table'' . Each word
 * in ``string'' is examined to see if it is an abbreviation of an element
 * in the options table. The value of the option is set to the required value .
 * If the caller attempts to use a word not in the ``option_table'', a WARNING
 * error message is generated .
 *   ``option_table'' is an array of data type ``option_elem''. The last element
 * of the array must be filled with NULL pointers and zeros.
*/
char VAR_PTR string;
Option_Elem VAR_PTR option_table;
char VAR_PTR caller;
  {
    int space_left;
    char VAR_PTR next_ptr,
         VAR_PTR str_ptr,
         VAR_PTR name = "t_parse_options",
         word_str[MAX_WORD_LEN+1];
    Option_Elem VAR_PTR table_ptr;
    static char VAR_PTR format[] =
        { "%c", /* OPT_CHAR  */     "%h",  /* OPT_SHORT */
          "%d", /* OPT_INT   */     "%ld", /* OPT_LONG  */
          "%f", /* OPT_FLOAT */     "%lf"  /* OPT_DOUBLE*/ };
    CATCH_NULL( string && option_table && caller, name, boolean );
/*
 *          skip initial white space
*/
    while( *string && t_isspace( *string ) )
        string++;
    for( next_ptr = string; *next_ptr; )
      {
/*
 *          copy next word to word_str,
*/
        for( str_ptr = word_str, space_left = MAX_WORD_LEN;
                *next_ptr && ! t_isspace( *next_ptr );
                next_ptr++ )
            if ( space_left )
              {
                *str_ptr++ = *next_ptr;
                space_left--;
              }
        *str_ptr = '\000';
/*
 *          ensure that ``next_ptr'' points to start of next word
*/
        while( *next_ptr && t_isspace( *next_ptr ) )
            next_ptr++;
/*
 *          change the selected word to upper case
*/
        for( str_ptr = word_str; *str_ptr; str_ptr++ )
            if ( 'a' <= *str_ptr && *str_ptr <= 'z' )
                *str_ptr += 'A' - 'a';
/*
 *          parse the selected word
*/
        for( table_ptr = option_table; table_ptr->opt_string; table_ptr++ )
          {
            if ( t_isabbr( (char VAR_PTR) word_str, table_ptr->opt_string,
                         table_ptr->opt_minlen ) )
              {
/*
 *          At this point, `word_str' is an abbreviation of an element
 *          in the options table. The value of the option is set to the
 *          required value. Then go on to parse the next word in the option
 *          string.
 *
 *          If the required value is OPT_NUM_ARG, the next word in the options
 *          string is decoded and assigned to the variable . Be careful with
 *          the floating point comparison .
*/
                if ( fabs( table_ptr->opt_value / OPT_NUM_ARG - 1.0 ) < 2.0e-6 )
                  {
/*
 *          next_ptr should point to a number which can be decoded and
 *          assigned to the variable . first, decide the format string
*/
                    if ( 1 == sscanf( (char *) next_ptr,
                            (char *) format[ table_ptr->opt_vartyp ],
                            table_ptr->opt_varptr ) )
                      {
/*
 *          move next_ptr to the start of the next word
*/
                        while( *next_ptr && ! t_isspace( *next_ptr ) )
                            next_ptr++;
                        while( *next_ptr && t_isspace( *next_ptr ) )
                            next_ptr++;
                      }
                  }
                else
                  {
/*                  Assign preset value to the variable . Check for floating
 *                  point values first . If the data is an integer type,
 *                  prevent possible type-conversion rounding error . By
 *                  isolating the ``int_value'' from the floating-point stuff,
 *                  we avoid the possibility of an integer overflow on a
 *                  redundant calculation . It is the users responsibility to
 *                  ensure that the specified option value will actually fit
 *                  into where it is supposed to go .
*/
                    if ( (int) table_ptr->opt_vartyp == OPT_FLOAT )
                        *( (float VAR_PTR) table_ptr->opt_varptr) =
                            table_ptr->opt_value;
                    else if ( (int) table_ptr->opt_vartyp == OPT_DOUBLE )
                        *( (double VAR_PTR) table_ptr->opt_varptr) =
                            table_ptr->opt_value;
                    else
                      {
                        long int int_value = table_ptr->opt_value +
                            ( table_ptr->opt_value > 0.0 ? 0.5 : -0.5 );
                        if (      (int) table_ptr->opt_vartyp == OPT_CHAR )
                            *( (char VAR_PTR) table_ptr->opt_varptr) =
                                int_value;
                        else if ( (int) table_ptr->opt_vartyp == OPT_SHORT )
                            *( (short int VAR_PTR) table_ptr->opt_varptr) =
                                int_value;
                        else if ( (int) table_ptr->opt_vartyp == OPT_INT )
                            *( (int VAR_PTR) table_ptr->opt_varptr) =
                                int_value;
                        else if ( (int) table_ptr->opt_vartyp == OPT_LONG )
                            *( (long int VAR_PTR) table_ptr->opt_varptr) =
                                int_value;
                        else
                            return make_error( ERR_ABORT, name,
                                (char VAR_PTR) "Illegal data type" );
                      }
                  }
                break;
              }
          }
/*
 *          Report an error if the end of the options table is reached
*/
        if ( ! table_ptr->opt_string )
          {
            (void) make_error( ERR_WARN + ERR_PRINT, caller,
                (char VAR_PTR) "Illegal option" );
            (void)  make_error( ERR_WARN + ERR_PRINT, caller,
                (char VAR_PTR) word_str );
          }
      }
    return TRUE;
  }



