/*
    Copyright (C) 1997  Peter Vernon Lanspeary

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* ************************************************************************ */
/*                                                                          */
/*                Dummy Basic Graphics Terminal Graphics Driver             */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/*   This is a dummy graphics driver . It has the same calls as all the     */
/*   other basic graphics drivers . The purpose of this dummy driver is to  */
/*   facilitate higher-level source code debugging . It does not actually   */
/*   generate any graphics, but apart from the graphics it does everything  */
/*   else that a real driver does, and in addition it describes what it is  */
/*   doing as each function is called .                                     */
/*                                                                          */
/*      Most functions are of type `boolean' and return a TRUE or FALSE     */
/*   result to indicate success or failure . Error messages are placed on   */
/*   the error queue -- which can be examined at the discretion of the user.*/
/*                                                                          */
/*   Any programs which use this driver must #include DEVICE_G_HEADER .     */
/*   These functions must be linked with "clip.c" and "error.c"             */
/*   The documentation on "clip.c" and "error.c" should be read before      */
/*   continuing .                                                           */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/* The function `t_begin()' has two arguments . The first `device' argument */
/* is the FILE* pointer to the output file . The second argument is a       */
/* character string of selected options . The available options are the     */
/* words                                                                    */
/*   "SCREEN"         (default) Emulate an interactive screen device        */
/*   "HARDCOPY"                 Emulate a (non-interactive) hardcopy device */
/*   "COLOUR"         (default) Device is colour                            */
/*   "MONOCHROME"               Device is monochrome                        */
/*   "SILENT"                   Generates no output                         */
/*   "NO_MESSAGES"              Generates no output                         */
/*   "FUNCTION_USED"            Generates function used messages            */
/*   "NOTE_MESSAGES"  (default) Generates note and function used messages   */
/*   "INFORMATION_MESSAGES"     Generates information & all other messages  */
/*      setting the lower left hardclip limit; default is (0.0,0.0)         */
/*   "X_MINIMUM x"              x is the abscissa of the lower left corner  */
/*   "Y_MINIMUM y"              y is the ordinate of the lower left corner  */
/*      setting the upper right hardclip limit; default is (1000.0,1000.0)  */
/*   "X_MAXIMUM x"              x is the abscissa of the upper right corner */
/*   "Y_MAXIMUM y"              y is the ordinate of the upper right corner */
/*   "X_TTY x"     (default=10) width of teletype characters                */
/*   "Y_TTY y"     (default=15) height of teletype characters               */
/*   "LINE_WIDTH"  (default=1)  width or thickness of a line                */
/*   "RESOLUTION"  (default=1)  resolution of the graphics device           */
/*                                                                          */
/* Words may be abbreviated and in upper, lower or mixed case . For example */
/* "hard mono x_max 1023.0 y_max 780.0" specifies hard-copy emulation for a */
/* monochrome device with abscissa and ordinate maxima of 1023.0 and 780.0  */
/* respectively .                                                           */
/*                                                                          */
/* ************************************************************************ */


#include "system.h"             /* system dependent stuff */
#include <stdio.h>
#include DEVICE_G_HEADER
#include FONT_SOURCE_HEADER
#include "dev_udef.h"

/*
 *      Machine constants
*/
#define TRUE          1
#define FALSE         0
/*#define NULL        0       NULL pointer - already defined in stdio.h   */
/*#define EOF         (-1)    end of file  - already defined in stdio.h   */

/*
 *    non printable ASCII characters
*/
#define NUL  '\000'   /* Nul       */
#define BEL  '\007'   /* Bell      */
#define BS   '\010'   /* Backspace */
#define HT   '\011'   /* Horizontal Tabulation */
#define LF   '\012'   /* Line Feed      */
#define VT   '\013'   /* Vertical Tabulation   */
#define CR   '\015'   /* Carriage Return*/
#define EM   '\031'   /* End of Medium  */
#define SP   '\040'   /* Space     */
#define DEL  '\177'   /* Delete    */

#define PI            3.141592653589793238462643

#define MAX_LIMIT     100000.0  /* maximum/minimum permitted hardclip value */
#define DEF_XUR_LIM   1000.0    /* hardclip upper right abscissa            */
#define DEF_YUR_LIM   1000.0    /* hardclip upper right ordinate            */
#define DEF_XLL_LIM   0.0       /* hardclip lower left abscissa             */
#define DEF_YLL_LIM   0.0       /* hardclip lower left ordinate             */
#define DEF_TTY_HEI   15.0      /* height in pixels of alphagraphics char   */
#define DEF_TTY_WID   10.0      /* width in pixels of alphagraphics char    */
#define PEN_THICKNESS 1.0       /* pen or line thickness                    */
#define DEF_RESOL     1.0       /* default device resolution                */
#define PIXEL_SIZE    0.3       /* horiz. and vertical size of a pixel (mm) */
#define SOLID_LINE    1
#define POINTS_ONLY   0       /* do not draw lines between points         */
#define SOLID_LINE    1       /* dashed line type                         */
#define NO_PEN        0       /* pen 0 - no pen is selected               */
#define BLACK         1       /* pen 1 - black pen                        */
#define MAX_LINE_LEN  80      /* max no. of chars on line sent to plotter */
#define BUFFER_LEN  MAX_LINE_LEN - 5 /* maximum echo buffer length        */


#define   CLIP2R_ID 0x3c22      /* CL2R two dimensional rectangular clipping*/
/*
 *          definition of available options
*/
#define MAX_WORD_LEN   20   /* max length of word in the options list     */
#define NOT_SET        -1   /* option not specified in options string     */
#define SCREEN          0   /* screen device emulation                    */
#define HARDCOPY        1   /* hardcopy device emulation                  */
#define SILENT          3   /* write nothing to output file               */
#define USE_FUNC        2   /* using function message                     */
#define NOTE            1   /* write note, warning & abort messages       */
#define INFORMATION     0   /* write info, note, warning & abort messages */
#define CONTINUED       8   /* this message is continued in the next call */

/* ************************************************************************ */
/*                                                                          */
/*                          Graphics Macros                                 */
/*                                                                          */
/* ************************************************************************ */
/*
 *  ABBREVIATIONS
 ******************
*/
#define  PLOTTER    dummy->plotter
#define  STROKE     dummy->stroke
#define  VIEW_CLIP  dummy->view_clip
#define  OP_FILE    dummy->op_file
#define  SETUP      dummy->setup

/*
 *  CATCH_NULL(pointer,caller,ret_type)
 ****************************************
 *      If "pointer" has a NULL value, an "Arg is NULL pointer" error message
 * is generated and the function is aborted . Otherwise, nothing is done .
 *      "pointer"   is the pointer to be tested,
 *      "caller"    is a character string naming the function
 *      "ret_type"  is the type of the returned zero value which indicates an
 *                    error condition
*/
#define CATCH_NULL(pointer,caller,ret_type) \
  if ( ! (pointer) )    \
      return (ret_type) make_error( ERR_ABORT + ERR_PRINT, caller, err_null )
/*
 *  G_ERROR()
 **************
 *      Determines the error status of the graphics device (terminal) .
 *      The returned value is TRUE if an error has occurred, otherwise FALSE .
*/
#define G_ERROR()           ferror( OP_FILE )

/*
 *  G_FLUSH()
 **************
 *      Flushes the graphics device (terminal) output buffer . There is no
 *      returned value .
*/
#define G_FLUSH()          ! fflush( OP_FILE )

/*
 *  G_PUTC(c)
 **************
 *      Writes the character `c' to the graphics device (terminal) .
 *      There is no returned result.
*/
#define G_PUTC(c)           fputc( c, OP_FILE )

/*
 *  G_PUTS(string)
 *******************
 *      Writes the character string `string' to the graphics device
 *      (terminal) . The there is no returned result .
*/
#define G_PUTS(string)      fputs( string, OP_FILE )

/*
 *  HYPOT(point)
 ****************
 *      Length of the vector "point" which is of type "coord_pair" .
*/
#define HYPOT(point)  \
        ( sqrt( (point)._x * (point)._x + (point)._y * (point)._y ) )

/*  LIMIT( x, min, max )
 *************************
 *      Limits the value of `x' to between `min' and `max'
*/
#define LIMIT(x,min,max)  (x) <= (min) ? (min) : ( (x) >= (max) ? (max) : (x) )

/*
 *  MAX( x, y )
 ****************
 *      The returned value is the maximum of `x' and `y' .
*/
#define MAX(x,y)        ( (x) > (y) ? (x) : (y) )

/*
 *  MIN( x, y )
 ****************
 *      The returned value is the minimum of `x' and `y' .
*/
#define MIN(x,y)        ( (x) > (y) ? (y) : (x) )

/*
 *  START_UP(caller,ret_type)
 *********************
 *      Calls `t_begin' with the default arguments `stdout' and `NULL'
 *  if `t_begin' has not previously been called .
*/
#define START_UP(caller,ret_type) \
    if ( ! dummy ) \
      { \
        if ( ! t_begin( stdout, mt_str ) ) \
            return (ret_type) make_error( ERR_CALL + ERR_PRINT, caller, \
                mt_str ); \
        (void) g_putmsg( NOTE, caller, no_begin ); \
      }
/*
 *  SWAPD( a, b )
 *****************
 *    swaps the two double precision floating point values `a' and `b'
*/
#define SWAPD(a,b)       { double temp = b; b = a; a = temp; }

/*
 *  TOLOWER( c )
 ****************
 *      Converts character `c' to lower case if it is upper case
*/
#define TOLOWER(c)      'A' <= (c) && (c) <= 'Z' ? (c) + 'a' - 'A' : (c)




/* *********************************************************************** */
/*                                                                         */
/*                              TYPE DECLARATIONS                          */
/*                                                                         */
/* *********************************************************************** */

/*
 * device mode - plotter is on or off
*/
enum dev_mode
  {
    ON_GRAPHICS,         /* plotter is switched on   */
    OFF_GRAPHICS         /* plotter is switched off  */
  };

/*
 *      The `data level' of the Visual 500 terminal controls how the
 *   terminal sets or reset pixel dots . The `data level' is set for
 *   use in alphagraphics, vector, point plot or incremental point plot
 *   mode . These `data levels' are,
*/
enum dev_dots     /* data level of terminal */
  {
    DOT_ON,             /* draw visible vectors, `or' them onto the screen  */
    DOT_OFF,            /* draw invisible vectors, switch dots off          */
    DOT_REV,            /* the compliment of the new data is plotted        */
    DOT_NEW             /* replace pixel contents with new data - not used  */
  };

typedef struct            /* TERMINAL STATUS                     */
  {                       /*     DATA TYPE                       */
    enum dev_mode mode;   /* terminal mode                       */
    enum dev_dots dots;   /* data level                          */
    int           colour; /* foreground colour                   */
    int           dash;   /* dashed line style                   */
    int           thick;  /* line and test width                 */
    coord_pair    cursor; /* alpha graphics text cursor position */
    coord_pair    pen;    /* vector graphics pen position        */
  }
    Plot_Status;

/*
 *  This defines the capabilities, selected options and general setup of
 *  the plotter
*/
typedef struct
  {
    short int dev_type;     /* SCREEN or HARDCOPY */
    short int is_colour;    /* TRUE or FALSE      */
    short int msg_type;     /* SILENT, INFORMATION, NOTE or ABORT */
    coord_pair ll_limit;    /* lower left hardclip limit  */
    coord_pair ur_limit;    /* upper right hardclip limit */
    coord_pair tty_siz;     /* size of teletype characters*/
    double pen_thick;       /* line width */
    coord_pair pg_size;     /* size of graphics page */
    short int alph_columns; /* number of character columns on the page */
    short int alph_lines;   /* number of character lines on the page   */
    double resolution;      /* resolution of the graphics device */
  }
    Setup_List;

typedef struct
  {
    Setup_List setup;       /* properties determined by options settings   */
    Text_Attribute stroke;  /* stroke text attributes                      */
    struct
      {                     /* Plotter status  */
        Plot_Status requir; /*     required    */
        Plot_Status actual; /*     actual      */
      }
        plotter;
    FILE *op_file;          /* file pointer for the graphics output device */
    clip_area VAR_PTR view_clip; /* pointer to viewport clipping area */
  }
    Gfx_Status;


/* ************************************************************************ */
/*                                                                          */
/*                     GLOBAL VARIABLE DECLARATIONS                         */
/*                                                                          */
/* ************************************************************************ */

extern StrokeFont
    asc_font,       /* font for ASCII chars   */
    sym_font;       /* font for special chars */
static Gfx_Status VAR_PTR dummy = (Gfx_Status VAR_PTR) 0;
                        /* complete description of dummy device */

/*
 *
 *      Common error messages and strings
*/
static char
  arg_str[72],        /* for printing argument list */
  VAR_PTR inactive  = "Inactive - hardcopy device",
  VAR_PTR err_write = "Cannot write to output file",
  VAR_PTR err_null  = "Arg is NULL pointer",
  VAR_PTR no_begin  =
      "t_begin not called. Auto-start with no options and output to stdout",
  VAR_PTR mt_str    = "";

/*
 *      Declare static and all external library functions whose returned
 *      value is used.
*/
static boolean
    g_putmsg(), set_dots(), set_colour(), set_style(), diff_point(),
    set_options();
boolean t_parse_options();
static boolean t_newpag(), t_gr_off();
static char VAR_PTR pr_dots(), VAR_PTR pr_mode(), VAR_PTR pr_txtset(),
    VAR_PTR pr_ch_def(), VAR_PTR pr_status();
static void clipped_char();
static int t_txtstr();

/* int fflush(), ferror();    declared in <stdio.h> */
int abs(), free(), fputc(), fputs(), fflush();
double fabs();
char *malloc();


/* ************************************************************************ */
/*                                                                          */
/*                              Utilities                                   */
/*                                                                          */
/* ************************************************************************ */

static boolean g_putmsg( mess_type, caller, message )
/*
 *    Put the message "message" on the output file . The name of the calling
 * function is "caller" . The message is only printed if the message type is
 * greater than the setup minimum . If the message type is UNKNOWN, the message
 * and calling function are not printed . If the message type is USE_FUNC, a
 * message of the form
 *   using "caller"("message")
 * where "message" is an argument list description .
*/
int mess_type;          /* message type : SILENT, USE_FUNC, INFORMATION, */
                        /*               NOTE, ABORT, UNKNOWN, CONTINUED */
char VAR_PTR caller,    /* name of calling function */
     VAR_PTR message;   /* message text */
  {
    static boolean hanging_msg = FALSE;     /* message is not finished yet */
    boolean have_written = FALSE,   /* have written to O/P file    */
            continued = FALSE;      /* this message is to be continued, */
                                    /* do not terminate with \n */
/*
 *      extract the continuation flag from the message type
*/
    if ( ( continued = CONTINUED & mess_type ) )
        mess_type &= ~ CONTINUED;
    if ( SETUP.msg_type <= mess_type )
/*
 *      write the message type and name of calling function
*/
      {
        if ( ! hanging_msg )
          {
/*          INFORMATION + NOTE */
            if ( ( have_written = INFORMATION==mess_type || NOTE==mess_type ) )
              {
                (void) G_PUTS(
                    INFORMATION == mess_type ? "information" : "note" );
                if ( caller )
                  {
                    (void) G_PUTS( " from " );
                    (void) G_PUTS( caller );
                  }
                (void) G_PUTS( " : " );
              }
            else if ( ( have_written = mess_type == USE_FUNC ) )
              {
                (void) G_PUTS( " using " );
                if ( caller )
                    (void) G_PUTS( caller );
                else
                    (void) G_PUTS( "????" );
                (void) G_PUTS( "( " );
               }
            /* else  SILENT + UNKNOWN */
          }
        else
            have_written = TRUE;
/*
 *      write message text
*/
        if ( message && ( have_written == TRUE ) )
            (void) G_PUTS( message );
        if ( mess_type == USE_FUNC )
            (void) G_PUTS( " )" );
        if ( have_written && ! continued )
            (void) G_PUTC( '\n' );
        if ( G_ERROR() )
            return make_error( ERR_ABORT + ERR_PRINT, caller, err_write );
      }
    hanging_msg = continued;
    return TRUE;
  }


static boolean set_dots( caller, dot )
/*
 *  sets the actual data level to `dot' .
*/
char VAR_PTR caller;    /* name of calling function */
enum dev_dots dot;
  {
    PLOTTER.actual.dots = dot;
    (void) g_putmsg( INFORMATION + CONTINUED, caller,
       (char VAR_PTR) "Setting data level = " );
    (void) g_putmsg( INFORMATION, caller,
       (char VAR_PTR) pr_dots( &PLOTTER.actual ) );
    return ! G_ERROR();
  }


static boolean set_colour( caller, pen )
/*
 *  set the actual colour to `ipen'
*/
char VAR_PTR caller;    /* name of calling function */
int pen;
  {
    char ascii_colour[2];   /* ASCII colour number */
    ascii_colour[0] = ( PLOTTER.actual.colour = pen ) + '0';
    ascii_colour[1] = NUL;
    (void) g_putmsg( INFORMATION | CONTINUED, caller,
       (char VAR_PTR) "Setting colour to " );
    (void) g_putmsg( INFORMATION, caller,
       (char VAR_PTR) ascii_colour );
    return ! G_ERROR();
  }


static boolean set_style( caller, style )
/*
 *  set the actual dash line pattern to `style'
*/
char VAR_PTR caller;    /* name of calling function */
int style;
  {
    char ascii_dash[2];   /* ASCII line style number */
    ascii_dash[0] = ( PLOTTER.actual.dash = style ) + '0';
    ascii_dash[1] = NUL;
    (void) g_putmsg( INFORMATION | CONTINUED, caller,
       (char VAR_PTR) "Setting dashed line style to " );
    (void) g_putmsg( INFORMATION, caller,
       (char VAR_PTR) ascii_dash );
    return ! G_ERROR();
  }


static char VAR_PTR pr_dots( status_ptr )
/*
 *  Convert data level (dots) to character string
*/
Plot_Status VAR_PTR status_ptr;
  {
    switch( status_ptr->dots )
      {
        case DOT_ON  : return (char VAR_PTR) "dots on";
        case DOT_OFF : return (char VAR_PTR) "dots off";
        case DOT_REV : return (char VAR_PTR) "complement dots";
        case DOT_NEW : return (char VAR_PTR) "replace dots";
      }
    return (char VAR_PTR) "";
  }


static char VAR_PTR pr_mode( status_ptr )
/*
 *  Convert graphics on/off status to character string
*/
Plot_Status VAR_PTR status_ptr;
  {
    switch( status_ptr->mode )
      {
        case ON_GRAPHICS  : return (char VAR_PTR) "graphics on";
        case OFF_GRAPHICS : return (char VAR_PTR) "graphics off";
      }
    return (char VAR_PTR) "";
  }


static char VAR_PTR pr_txtset( attrib )
/*
 * Interpret ``attrib'' as a text attribute and return a string which
 * identifies the attribute .
*/
int attrib;     /* text attribute */
  {
    switch ( attrib )
      {
        case T_CH_RESET :   return (char VAR_PTR) "T_CH_RESET";
        case T_CH_BOLD :    return (char VAR_PTR) "T_CH_BOLD";
        case T_CH_NOBOLD :  return (char VAR_PTR) "T_CH_NOBOLD";
        case T_ASCII :      return (char VAR_PTR) "T_ASCII";
        case T_SYMBOL :     return (char VAR_PTR) "T_SYMBOL";
        case T_CH_ALIGN :   return (char VAR_PTR) "T_CH_ALIGN";
        case T_CH_HEIGHT :  return (char VAR_PTR) "T_CH_HEIGHT";
        case T_CH_ASPECT :  return (char VAR_PTR) "T_CH_ASPECT";
        case T_CH_SLANT :   return (char VAR_PTR) "T_CH_SLANT";
        case T_CH_ANGLE :   return (char VAR_PTR) "T_CH_ANGLE";
        case T_CH_GAP :     return (char VAR_PTR) "T_CH_GAP";
        case -T_CH_ALIGN :  return (char VAR_PTR) "-T_CH_ALIGN";
        case -T_CH_HEIGHT : return (char VAR_PTR) "-T_CH_HEIGHT";
        case -T_CH_ASPECT : return (char VAR_PTR) "-T_CH_ASPECT";
        case -T_CH_SLANT :  return (char VAR_PTR) "-T_CH_SLANT";
        case -T_CH_ANGLE :  return (char VAR_PTR) "-T_CH_ANGLE";
        case -T_CH_GAP :    return (char VAR_PTR) "-T_CH_GAP";
      }
    return (char VAR_PTR) "unknown";
  }


static char VAR_PTR pr_ch_def( what )
/*
 * Interpret ``what'' as a required action for ``t_ch_def'' and return a
 * string which identifies the action .
*/
int what;     /* required action for ``t_ch_def'' */
  {
    switch ( what )
      {
        case T_FT_SIZE   :  return (char VAR_PTR) "T_FT_SIZE";
        case T_FT_EXIST  :  return (char VAR_PTR) "T_FT_EXIST";
        case T_FT_LENGTH :  return (char VAR_PTR) "T_FT_LENGTH";
        case T_FT_FIRST  :  return (char VAR_PTR) "T_FT_FIRST";
        case T_FT_LAST   :  return (char VAR_PTR) "T_FT_LAST";
        case T_FT_MIDDLE :  return (char VAR_PTR) "T_FT_MIDDLE";
        case T_FT_POINT  :  return (char VAR_PTR) "T_FT_POINT";
      }
    return (char VAR_PTR) "unknown";
  }


static char VAR_PTR pr_status( attrib )
/*
 * Interpret ``attrib'' as a status attribute  and return a string which
 * identifies the status attribute .
*/
int attrib;     /* text attribute */
  {
    switch ( attrib )
      {
        case T_ST_DEVICE :  return (char VAR_PTR) "T_ST_DEVICE";
        case T_ST_CATEG  :  return (char VAR_PTR) "T_ST_CATEG";
        case T_ST_SCREEN :  return (char VAR_PTR) "T_ST_SCREEN";
        case T_ST_MONO   :  return (char VAR_PTR) "T_ST_MONO";
        case T_ST_CLIP   :  return (char VAR_PTR) "T_ST_CLIP";
        case T_ST_MAX    :  return (char VAR_PTR) "T_ST_MAX";
        case T_ST_MIN    :  return (char VAR_PTR) "T_ST_MIN";
        case T_ST_CHAR   :  return (char VAR_PTR) "T_ST_CHAR";
        case T_ST_LINES  :  return (char VAR_PTR) "T_ST_LINES";
        case T_ST_PIXSIZ :  return (char VAR_PTR) "T_ST_PIXSIZ";
        case T_ST_A_CURS :  return (char VAR_PTR) "T_ST_A_CURS";
        case T_ST_R_CURS :  return (char VAR_PTR) "T_ST_R_CURS";
        case T_ST_A_PEN  :  return (char VAR_PTR) "T_ST_A_PEN";
        case T_ST_R_PEN  :  return (char VAR_PTR) "T_ST_R_PEN";
      }
    return (char VAR_PTR) "unknown";
  }


static void clipped_char( ll_point, caller, charact )
/*
 *  Check if parts of drawn character are clipped away or likely to be clipped
 *  away . This is only a crude test based on the page margins an the position
 * of the lower-left corner of the character, and so the results are only
 * approximate .
*/
coord_pair VAR_PTR ll_point;    /* location of lower left corner of character */
char VAR_PTR caller,            /* name of calling function */
     charact;                   /* character to be drawn */
  {
    char message[40];
    double margin = STROKE.height * MAX( 1.0, STROKE.aspect );
    if ( ll_point->_x < SETUP.ll_limit._x - margin ||
         ll_point->_x > SETUP.ur_limit._x + margin ||
         ll_point->_y < SETUP.ll_limit._y - margin ||
         ll_point->_y > SETUP.ur_limit._y + margin )
        (void) sprintf( message, "Character `%c' fully or partly clipped",
            charact );
    else if ( ll_point->_x < SETUP.ll_limit._x + margin ||
              ll_point->_x > SETUP.ur_limit._x - margin ||
              ll_point->_y < SETUP.ll_limit._y + margin ||
              ll_point->_y > SETUP.ur_limit._y - margin )
        (void) sprintf( message, "Character `%c' possibly (partly) clipped",
            charact );
    else
        (void) sprintf( message, "Character `%c' drawn", charact );
    (void) g_putmsg( NOTE, caller, (char VAR_PTR) message );
  }


static int t_txtstr( string, end_nul, caller )
/*
 * Prints a label of printable ASCII , starting at the current required cursor
 * position and using the current colour . Only characters within the viewport
 * are written to the screen . The returned value is the number of characters
 * written.
 * The character string `string' may be NUL terminated or, if only a fragment
 * of the string is to be printed, `end_nul' is a pointer to where the NUL
 * should be.
*/
char VAR_PTR string,    /* start address of character string to be printed  */
     VAR_PTR end_nul,   /* address of phantom end of string NUL */
     VAR_PTR caller;    /* name of calling function */
  {
    boolean first_cl = TRUE;
    int char_count = 0;     /* number of characters sent */
/*
 *      for a colour device,
 *      set the colour and data level.
*/
    if ( SETUP.is_colour )
      {
        if ( PLOTTER.requir.dots != PLOTTER.actual.dots )
            (void) set_dots( caller, PLOTTER.requir.dots );
        if ( PLOTTER.requir.colour != PLOTTER.actual.colour )
            (void) set_colour( caller, PLOTTER.requir.colour );
      }
    else
      {
/*
 *      for a monochrome device,
 *      set colour to black or white (dashed lines not allowed)
*/
        if ( ! PLOTTER.requir.colour && SETUP.dev_type == SCREEN )
          {
            if ( PLOTTER.actual.dots != DOT_OFF )
                (void) set_dots( caller, DOT_OFF );
          }
        else if ( PLOTTER.requir.dots != PLOTTER.actual.dots )
            (void) set_dots( caller, PLOTTER.requir.dots );
      }
    if ( G_ERROR() )
        return FALSE;
    for ( ; *string && string != end_nul; string++ )
      {
        if ( ' ' <= *string && *string <= '~' )
          {
/*
 ********     write the character if it is in the viewport
*/
            if ( PLOTTER.requir.cursor._x >= min_clip( VIEW_CLIP, _x ) &&
                 PLOTTER.requir.cursor._x <=
                     max_clip( VIEW_CLIP, _x ) + SETUP.tty_siz._x      &&
                 PLOTTER.requir.cursor._y >= min_clip( VIEW_CLIP, _y ) &&
                 PLOTTER.requir.cursor._y <=
                     max_clip( VIEW_CLIP, _y ) + SETUP.tty_siz._y )
              {
/*
 *          place the cursor in the correct place
*/
                if ( diff_point( (coord_pair VAR_PTR) &PLOTTER.requir.cursor,
                                (coord_pair VAR_PTR) &PLOTTER.actual.cursor ) )
                    PLOTTER.actual.cursor = PLOTTER.requir.cursor;
/*
 *          send text parameters to the plotter
*/
                char_count++;
                PLOTTER.actual.cursor._x += SETUP.tty_siz._x;
              }
            else if ( first_cl )
              {
                (void) g_putmsg( NOTE, caller,
                    (char VAR_PTR) "Character(s) clipped away" );
                first_cl = FALSE;
              }
            PLOTTER.requir.cursor._x += SETUP.tty_siz._x;
          }
        else
            (void) g_putmsg( NOTE, caller,
                (char VAR_PTR) "Non-printable character" );
      }
    PLOTTER.actual.pen = PLOTTER.actual.cursor;
    return char_count;
  }


static boolean diff_point( coord_1, coord_2 )
/*
 * Decides if the two points *``coord_1'' and *``coord_2'' are different;
 * i.e. either the abscissae or the ordinates are different by more than
 * the device resolution SETUP.resultion .
*/
coord_pair VAR_PTR coord_1,
           VAR_PTR coord_2;
  {
    return fabs( coord_1->_x - coord_2->_x ) > SETUP.resolution ||
           fabs( coord_1->_y - coord_2->_y ) > SETUP.resolution;
  }


/* ************************************************************************ */
/*                                                                          */
/*                    Terminal driving routines                             */
/*                                                                          */
/* ************************************************************************ */


static clip_area VAR_PTR t_begin( device, options )
/*
 *        Start graphics .
 * Decodes the required ``options'' string, initializes the graphics device,
 * initializes variables and sets up the viewport clip. The returned value is
 * a pointer to the clip_area structure of the viewport . If the function
 * fails, it returns a NULL pointer. Appropriate error messages are produced
 * in these cases .
*/
FILE * device;        /* file pointer for graphics device */
char VAR_PTR options; /* options string */
  {
    char VAR_PTR name = "t_begin";
    CATCH_NULL( device, name, clip_area VAR_PTR );
    (void) sprintf( arg_str, "(FILE *)x%.8lx,\"%.50s\"",
        (unsigned long int) device, options );

/*
 *    GRAB GRAPHICS STATUS DATA
 *      Find out if ``device'' has been registered as a device .
 *      If it has, check that it is indeed a dummy device and grab
 *      the pointer to it graphics status data .
*/
    if ( t_register( T_ISREGIST, device, NONEXISTENT_DEVICE, (char VAR_PTR) 0) )
      {
        if ( (int) t_register( T_REG_TYPE, device, 0, (char VAR_PTR) 0 )
                != DUMMY_GRAPHICS_DEV )
            return (clip_area VAR_PTR) make_error( ERR_ABORT + ERR_PRINT, name,
                (char VAR_PTR) "Wrong driver. Device is not dummy device" );
        dummy = (Gfx_Status VAR_PTR) t_register( T_REG_DATA,
            device, NONEXISTENT_DEVICE, (char VAR_PTR) 0 );
        (void) file_error( OP_FILE );   /* point to graphics device */
/*
 *    DECODE THE OPTIONS AND ASSIGN PAGE & SETUP STRUCTURES
*/
        if ( ! set_options( options, name ) )
            return (clip_area VAR_PTR) NULL;
        (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
        (void) g_putmsg( NOTE, name,
            (char VAR_PTR) "Device already registered" );
      }
    else
      {
/*
 *    REGISTRATION
 *      ``device'' has never been registered, so allocate memory for the
 *      graphics status data and register the device as a HP/GL plotter .
*/
        dummy = (Gfx_Status VAR_PTR)
            malloc( (unsigned int) sizeof(Gfx_Status) );
        if ( ! dummy )
            return (clip_area VAR_PTR) make_error( ERR_ABORT + ERR_PRINT, name,
                (char VAR_PTR) "Cannot allocate memory" );
        if ( ! t_register( T_REGISTER, device,
                 DUMMY_GRAPHICS_DEV, (char VAR_PTR) dummy ) )
          {
            (void) free( (char *) dummy );
            dummy = (Gfx_Status VAR_PTR) 0;
            return (clip_area VAR_PTR) make_error( ERR_CALL + ERR_PRINT,
                name, mt_str );
          }
/*
 *    DECODE THE OPTIONS AND ASSIGN PAGE & SETUP STRUCTURES
*/
        (void) file_error( OP_FILE = device );   /* point to graphics device */
        if ( ! set_options( options, name ) )
          {
            if ( ! t_register( T_DEREGIST, OP_FILE,
                    DUMMY_GRAPHICS_DEV, (char VAR_PTR) dummy ) )
                (void) make_error( ERR_CALL + ERR_PRINT, name, mt_str );
            (void) free( (char *) dummy );
            dummy = (Gfx_Status VAR_PTR) 0;
            return (clip_area VAR_PTR) NULL;
          }
        (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
        (void) g_putmsg( NOTE, name, (char VAR_PTR) "Registering new device" );
/*
 *           reset character attributes and plotter status
*/
        (void) reset_font( (Text_Attribute VAR_PTR) &STROKE,
            MAX( SETUP.pg_size._x, SETUP.pg_size._y ) / 20.0 );
/*
 *      SET THE PLOTTER STATUS
 *      and put pen in lower left corner.
*/
        PLOTTER.requir.colour = BLACK;
        PLOTTER.requir.dash   = SOLID_LINE;
        PLOTTER.requir.pen    = SETUP.ll_limit;
        PLOTTER.requir.cursor._x = SETUP.ll_limit._x;
        PLOTTER.requir.cursor._y = SETUP.ur_limit._y - SETUP.tty_siz._y + 1.0;
/*
 *          set the plotter clipping window to the hard clip limits
*/
        if ( ! ( VIEW_CLIP = def_2rclip( (coord_pair VAR_PTR) &SETUP.ll_limit,
                 (coord_pair VAR_PTR) &SETUP.ur_limit,
                 (coord_pair VAR_PTR) &SETUP.ll_limit ) ) )
            return (clip_area VAR_PTR) make_error( ERR_CALL + ERR_PRINT,
                name, mt_str );
      }
/*
 *      SET THE PLOTTER STATUS
 *      and put pen in lower left corner.
*/
    PLOTTER.actual.colour = BLACK;
    PLOTTER.actual.dash = SOLID_LINE;
    PLOTTER.actual.cursor = PLOTTER.actual.pen = PLOTTER.requir.pen =
        SETUP.ll_limit;
    if ( G_ERROR() )
        return (clip_area VAR_PTR) make_error( ERR_WARN + ERR_PRINT,
            name, err_write );
    return VIEW_CLIP;
  }


#define ST_OPT  (char VAR_PTR) &st_opt

static boolean set_options( options, caller )
/*
 * Decode the options to ``t_begin'' and assign appropriate values to
 * the setup structures .
*/
char VAR_PTR options,   /* string -- list of options */
     VAR_PTR caller;    /* name of calling function  */
  {
/*
 *  table of plotter options
*/
    static Setup_List st_opt;
    static Option_Elem option_list[] = {
        { "SCREEN",           2, OPT_SHORT,  ST_OPT.dev_type,   SCREEN   },
        { "HARDCOPY",         1, OPT_SHORT,  ST_OPT.dev_type,   HARDCOPY },
        { "COLOUR",           1, OPT_SHORT,  ST_OPT.is_colour,  TRUE     },
        { "MONOCHROME",       1, OPT_SHORT,  ST_OPT.is_colour,  FALSE    },
        { "SILENT",           2, OPT_SHORT,  ST_OPT.msg_type,   SILENT   },
        { "NO_MESSAGES",      3, OPT_SHORT,  ST_OPT.msg_type,   SILENT   },
    { "INFORMATION_MESSAGES", 1, OPT_SHORT,  ST_OPT.msg_type,   INFORMATION },
        { "NOTE_MESSAGES",    3, OPT_SHORT,  ST_OPT.msg_type,   NOTE        },
        { "X_MINIMUM",        4, OPT_DOUBLE, ST_OPT.ll_limit._x, OPT_NUM_ARG },
        { "Y_MINIMUM",        4, OPT_DOUBLE, ST_OPT.ll_limit._y, OPT_NUM_ARG },
        { "X_MAXIMUM",        4, OPT_DOUBLE, ST_OPT.ur_limit._x, OPT_NUM_ARG },
        { "Y_MAXIMUM",        4, OPT_DOUBLE, ST_OPT.ur_limit._y, OPT_NUM_ARG },
        { "X_TTY",            3, OPT_DOUBLE, ST_OPT.tty_siz._x, OPT_NUM_ARG },
        { "Y_TTY",            3, OPT_DOUBLE, ST_OPT.tty_siz._y, OPT_NUM_ARG },
        { "LINE_WIDTH",       1, OPT_DOUBLE, ST_OPT.pen_thick,  OPT_NUM_ARG },
        { "RESOLUTION",       1, OPT_DOUBLE, ST_OPT.resolution, OPT_NUM_ARG },
        { (char *) 0,         0, 0, (char VAR_PTR) 0,   0 } };
/*
 *          clear any previous plotter setup
*/
    st_opt.dev_type = st_opt.is_colour = st_opt.msg_type = NOT_SET;
    st_opt.ll_limit._x = st_opt.ll_limit._y = st_opt.ur_limit._x =
        st_opt.ur_limit._y = st_opt.tty_siz._x = st_opt.tty_siz._y =
        st_opt.pen_thick = st_opt.resolution = (double) NOT_SET;
/*
 *           decode the options list
*/
    if ( options )
        if ( ! t_parse_options( options, (Option_Elem VAR_PTR) option_list,
               caller ) )
            return FALSE;
    SETUP = st_opt;
/*
 *          fix any setup values not determined by the options list
*/
    if ( SETUP.dev_type == NOT_SET )
        SETUP.dev_type = SCREEN;
    if ( SETUP.is_colour == NOT_SET )
        SETUP.is_colour = TRUE;
    if ( SETUP.msg_type == NOT_SET )
        SETUP.msg_type = NOTE;

    if ( SETUP.ur_limit._x == (double) NOT_SET )
        SETUP.ur_limit._x = DEF_XUR_LIM;
    else
        SETUP.ur_limit._x = LIMIT( SETUP.ur_limit._x, -MAX_LIMIT, MAX_LIMIT );

    if ( SETUP.ur_limit._y == (double) NOT_SET )
        SETUP.ur_limit._y = DEF_XUR_LIM;
    else
        SETUP.ur_limit._y = LIMIT( SETUP.ur_limit._y, -MAX_LIMIT, MAX_LIMIT );

    if ( SETUP.ll_limit._x == (double) NOT_SET )
        SETUP.ll_limit._x = DEF_XLL_LIM;
    else
        SETUP.ll_limit._x = LIMIT( SETUP.ll_limit._x, -MAX_LIMIT,
            SETUP.ur_limit._x - 1.0 );

    if ( SETUP.ll_limit._y == (double) NOT_SET )
        SETUP.ll_limit._y = DEF_YLL_LIM;
    else
        SETUP.ll_limit._y = LIMIT( SETUP.ll_limit._y, -MAX_LIMIT,
            SETUP.ur_limit._y - 1.0 );

    SETUP.pg_size._x = SETUP.ur_limit._x - SETUP.ll_limit._x;
    if ( SETUP.tty_siz._x == (double) NOT_SET )
        SETUP.tty_siz._x =  DEF_TTY_WID;
    else
        SETUP.tty_siz._x = LIMIT( SETUP.tty_siz._x,
            SETUP.pg_size._x / 1000.0, SETUP.pg_size._x );

    SETUP.pg_size._y = SETUP.ur_limit._y - SETUP.ll_limit._y;
    if ( SETUP.tty_siz._y == (double) NOT_SET )
        SETUP.tty_siz._y =  DEF_TTY_HEI;
    else
        SETUP.tty_siz._y = LIMIT( SETUP.tty_siz._y,
            SETUP.pg_size._y / 1000.0, SETUP.pg_size._y );

    if ( SETUP.pen_thick == (double) NOT_SET )
        SETUP.pen_thick = PEN_THICKNESS;
    else
        SETUP.pen_thick = LIMIT( SETUP.pen_thick, 0.1, 10.0 );

    if ( SETUP.resolution == (double) NOT_SET )
        SETUP.resolution = DEF_RESOL;
    else
        SETUP.resolution = LIMIT( SETUP.resolution, 0.001, 10.0 );

/*
 *          complete the "plotter unit" SETUP info and
 *          "pixel units" plotting limits
*/
    SETUP.resolution /= 2.0;
    SETUP.alph_columns = ( SETUP.ur_limit._x - SETUP.ll_limit._x) /
        SETUP.tty_siz._x;
    SETUP.alph_lines   = ( SETUP.ur_limit._y - SETUP.ll_limit._y ) /
        SETUP.tty_siz._y;
    return TRUE;
  }
#undef ST_OPT


static boolean t_dump( file_name )
/*
 *      Opens a file with name `file_name' and dumps the contents of the
 * graphics device into it . If t_dump cannot open or write to file the file,
 * an error message is generated and the returned value is FALSE .
*/
char VAR_PTR file_name;     /* name of dump file */
  {
    FILE * dump_ptr;
    char VAR_PTR name = "t_dump";
    CATCH_NULL( file_name, name, boolean );
    START_UP( name, boolean );
    (void) g_putmsg( USE_FUNC, name, file_name );
    if ( ( dump_ptr = fopen( (char *) file_name, "w" ) ) )
      {
        (void) g_putmsg( INFORMATION, name, "Opened dump file" );
        (void) fputs( "This is a simulated dump.", dump_ptr );
        (void) fclose( dump_ptr );
        if ( ferror( dump_ptr ) )
            return make_error( ERR_ABORT + ERR_PRINT, name,
                (char VAR_PTR) "Cannot write to dump file" );
      }
    return make_error( ERR_ABORT + ERR_PRINT, name,
            (char VAR_PTR) "Cannot open dump file" );
  }


static boolean t_end()
/*
 *      End of graphics . Delete character fonts and frees the memory allocated
 *  to them, deregister the device .
*/
  {
    char VAR_PTR name = "t_end";
    START_UP( name, boolean );
    (void) g_putmsg( USE_FUNC, name, mt_str );
    if ( ! t_newpag() || ! t_gr_off() )
        return make_error( ERR_CALL + ERR_PRINT, name, mt_str );
/*
 *      Deregister the device
*/
    if ( ! t_register( T_DEREGIST, OP_FILE,
            DUMMY_GRAPHICS_DEV, (char VAR_PTR) dummy ) )
        (void) make_error( ERR_CALL + ERR_PRINT, name, mt_str );
    (void) g_putmsg( INFORMATION, name, (char VAR_PTR) "Deregistering" );
/*
 *      Delete character fonts and frees the memory allocated to them
*/
    if ( asc_font.font_got )
        (void) clear_font( (StrokeFont VAR_PTR) &asc_font );
    if ( sym_font.font_got )
        (void) clear_font( (StrokeFont VAR_PTR) &sym_font );
    (void) g_putmsg( INFORMATION, name,
        (char VAR_PTR) "Clearing character fonts" );
/*
 *      Free the memory used by the graphics status data .
*/
    (void) free( (char *) dummy );
    dummy = (Gfx_Status VAR_PTR) 0;
    return TRUE;
  }


static boolean t_vwport( clip_ptr )
/*
 *      Set the clipping limits for the viewport . The old limits are replaced .
 * The `clip_ptr' must point to  clip_area which has been defined by
 * `def_2rclip()' . The function fails if
 *   a) clip_ptr is a NULL pointer
 *   b) clip_ptr points to something which is not a structure for a two
 *      dimensional rectangular clipping region .
*/
clip_area VAR_PTR clip_ptr;
  {
    char VAR_PTR name = "t_vwport";
    CATCH_NULL( clip_ptr, name, boolean );
    START_UP( name, boolean );
    (void) sprintf( arg_str, "x%.8lx", (unsigned long int) clip_ptr );
    (void) g_putmsg( USE_FUNC, name, arg_str );
    if ( clip_ptr->id_struct == CLIP2R_ID )
      {
        if ( min_clip( clip_ptr, _x ) >= SETUP.ll_limit._x &&
             max_clip( clip_ptr, _x ) <= SETUP.ur_limit._x &&
             min_clip( clip_ptr, _y ) >= SETUP.ll_limit._y &&
             max_clip( clip_ptr, _y ) <= SETUP.ur_limit._y )
          {
            if ( SETUP.msg_type <= NOTE )
                (void) what_clip( OP_FILE, clip_ptr );
            VIEW_CLIP = clip_ptr;
            return TRUE;
          }
        return make_error( ERR_ABORT + ERR_PRINT, name,
            (char VAR_PTR) "Outside device limits" );
      }
    return make_error( ERR_ABORT + ERR_PRINT, name,
        (char VAR_PTR) "Invalid pointer" );
  }


static boolean t_colour( ipen )
/*
 *    Selects specified colour ( or pen ) .
 *
 * If the device is monochrome, colours are simulated with various dashed line
 * styles and the only true ``colours'' available are dark and bright . The
 * returned value is TRUE if  0 <= ipen <= 8 . If these limits are exceeded,
 * the value is wrapped around into the range 1 to 8 . The dashed line
 * equivalents of the colours are,
 *
 *    IPEN    SCREEN                  PLOTTER
 *          LINE STYLE                COLOUR
 *     0      black  (background)     no pen = invisible
 *     1      solid  (green)          black
 *     2      dotted                  brown
 *     3    short dash                red
 *     4    long dash                 orange
 *     5    dot dash                  yellow
 *     6    short-dash long-dash      green
 *     7    2-short-dashes long-dash  blue
 *     8    3-dots long-dash          violet
*/
int ipen;                /* pen number or colour */
  {
    char VAR_PTR name = "t_colour";
    START_UP( name, boolean );
    (void) sprintf( arg_str, "%d", ipen );
    (void) g_putmsg( USE_FUNC, name, arg_str );
    PLOTTER.requir.colour = ( abs( ipen ) - 1 ) % 8 + 1;
    if ( 0 <= ipen && ipen <= 8 )
        return TRUE;
    (void) g_putmsg( NOTE, name, (char VAR_PTR) "Wrap around" );
    return FALSE;
  }


static boolean t_dashed( style )
/*
 *      Set the dashed line style . The returned value is TRUE if
 * 0 <= style <= 8 . If these limits are exceeded, the value is wrapped
 * around into the range 1 to 8 .
 *
 *    STYLE   SCREEN
 *          LINE STYLE
 *     0      black  - plot vector endpoints only (invisible)
 *     1      solid  (green)
 *     2      dotted
 *     3    short dash
 *     4    long dash
 *     5    dot dash
 *     6    short-dash long-dash
 *     7    2-short-dashes long-dash
 *     8    3-dots long-dash
 *
*/
int style;        /* dashed line style */
  {
    char VAR_PTR name = "t_dashed";
    START_UP( name, boolean );
    (void) sprintf( arg_str, "%d", style );
    (void) g_putmsg( USE_FUNC, name, arg_str );
    PLOTTER.requir.dash = ( abs( style ) - 1 ) % 8 + 1;
    if ( 0 <= style && style <= 8 )
        return TRUE;
    (void) g_putmsg( NOTE, name, (char VAR_PTR) "Wrap around" );
    return FALSE;
  }


static boolean t_thick( width )
/*
 * Set the ``width'' of lines and text . The returned value is TRUE
 * if 0 <= thickness.
*/
int width;      /* line thickness */
  {
    char VAR_PTR name = "t_thick";
    START_UP( name, boolean );
    (void) sprintf( arg_str, "%d", width );
    (void) g_putmsg( USE_FUNC, name, arg_str );
    PLOTTER.requir.thick = width;
    return 0 <= width;
  }


/* ************************************************************************ */
/*                                                                          */
/*                          Turn graphics on or off                         */
/*                                                                          */
/* ************************************************************************ */


static boolean t_gr_off()
/*
 *      Turn graphics off . I.e. put terminal into alpha_numeric mode . This
 *  function fails if it cannot write to the screen .
*/
  {
    char VAR_PTR name = "t_gr_off";
    START_UP( name, boolean );
    PLOTTER.requir.mode = PLOTTER.actual.mode = OFF_GRAPHICS;
    return g_putmsg( USE_FUNC, name, mt_str );
  }


static boolean t_gr_on()
/*
 *      Turn graphics on . I.e. if not already in vector or alphagraphics
 *  mode, put terminal into vector mode . This function fails if it cannot
 *  write to the screen .
*/
  {
    char VAR_PTR name = "t_gr_on";
    START_UP( name, boolean );
    PLOTTER.requir.mode = PLOTTER.actual.mode = ON_GRAPHICS;
    return g_putmsg( USE_FUNC, name, mt_str );
  }


/* ************************************************************************ */
/*                                                                          */
/*           Change Data Level, Visible, Invisible, Revers, Replace         */
/*                                                                          */
/* ************************************************************************ */


static boolean t_visib()
/*
 *  Draw visible vectors . Sets screen into "dots on" mode . This function
 *  always succeeds .
*/
  {
    char VAR_PTR name = "t_visib";
    START_UP( name, boolean );
    (void) g_putmsg( USE_FUNC, name, mt_str );
    if ( SETUP.dev_type == HARDCOPY )
        (void) g_putmsg( NOTE, name, inactive );
    else
        PLOTTER.requir.dots = DOT_ON;
    return TRUE;
  }


static boolean t_invis()
/*
 *  Draw invisible vectors . Sets screen into "dots off" mode . This function
 *  always succeeds .
*/
  {
    char VAR_PTR name = "t_invis";
    START_UP( name, boolean );
    (void) g_putmsg( USE_FUNC, name, mt_str );
    if ( SETUP.dev_type == HARDCOPY )
        (void) g_putmsg( NOTE, name, inactive );
    else
        PLOTTER.requir.dots = DOT_OFF;
    return TRUE;
  }


static boolean t_revers()
/*
 *  Reverse the pixel dots when plotting vectors, plotting points or
 *  alphagraphic characters . Sets screen into "compliment" mode . This function
 *  always succeeds .
*/
  {
    char VAR_PTR name = "revers";
    START_UP( name, boolean );
    (void) g_putmsg( USE_FUNC, name, mt_str );
    if ( SETUP.dev_type == HARDCOPY )
        (void) g_putmsg( NOTE, name, inactive );
    else
        PLOTTER.requir.dots = DOT_REV;
    return TRUE;
  }


static boolean t_replac()
/*
 *  Replace pixel contents with the new data . Sets screen into "replace"
 *  mode . This function always succeeds .
*/
  {
    char VAR_PTR name = "t_replac";
    START_UP( name, boolean );
    (void) g_putmsg( USE_FUNC, name, mt_str );
    if ( SETUP.dev_type == HARDCOPY )
        (void) g_putmsg( NOTE, name, inactive );
    else
        PLOTTER.requir.dots = DOT_NEW;
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                         Miscellaneous stuff                              */
/*                                                                          */
/* ************************************************************************ */


static boolean t_newpag()
/*
 * Clear the graphics screen or eject/replace the paper. This function fails
 * if it cannot write to the screen .
*/
  {
    char VAR_PTR name = "t_newpag";
    START_UP( name, boolean );
    (void) g_putmsg( USE_FUNC, name, mt_str );
    return ! G_ERROR();
  }


static boolean t_newscr()
/*
 * Clear the alphanumeric screen . This function fails if it cannot
 *  write to the screen .
*/
  {
    char VAR_PTR name = "t_newscr";
    START_UP( name, boolean );
    (void) g_putmsg( USE_FUNC, name, mt_str );
    if ( SETUP.dev_type == HARDCOPY && OP_FILE != stdout )
        (void) g_putmsg( NOTE, name, inactive );
    return ! G_ERROR();
  }


static boolean t_bell()
/*
 *      Ring bell . This function fails if it cannot write to the screen .
*/
  {
    char VAR_PTR name = "t_bell";
    START_UP( name, boolean );
    (void) g_putmsg( USE_FUNC, name, mt_str );
    if ( SETUP.dev_type == HARDCOPY && OP_FILE != stdout )
        (void) g_putmsg( NOTE, name, inactive );
    return ! G_ERROR();
  }


static boolean t_wait()
/*
 *  Waits for the user to press the carriage return . This function is inactive
 *  if output is to disk, but no error message is generated .
 *  The returned value is TRUE unless it cannot write to the screen .
*/
  {
    char VAR_PTR name = "t_wait";
    START_UP( name, boolean );
    (void) g_putmsg( USE_FUNC, name, mt_str );
    if ( SETUP.dev_type == HARDCOPY && OP_FILE != stdout )
        (void) g_putmsg( NOTE, name, inactive );
    else
        while( getchar() != '\n' );
    return ! G_ERROR();
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Text Writing                                   */
/*                                                                          */
/* ************************************************************************ */


static boolean t_txtmov( point )
/*
 *    Move the text cursor to coordinate `point' . The returned result is
 *  TRUE only if the next printable character is within the screen limits .
*/
register coord_pair *point;
  {
    char VAR_PTR name = "t_txtmov";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    (void) sprintf( arg_str, "%.8lx->(%.6g,%.6g)",
        (unsigned long int) point, point->_x, point->_y );
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
    PLOTTER.requir.cursor = *point;
    if ( point->_x >= SETUP.ll_limit._x &&
         point->_x <= SETUP.ur_limit._x - SETUP.tty_siz._x &&
         point->_y >= SETUP.ll_limit._y  &&
         point->_y <= SETUP.ur_limit._y - SETUP.tty_siz._y )
        return TRUE;
    (void) g_putmsg( NOTE, name, (char VAR_PTR) "Character not on screen/page");
    return FALSE;
  }


static boolean t_txtcur( row, column )
/*
 *    Move the text cursor to text row `row' and text column `column' .
 *    The text row is numbered from the top of the screen . The returned
 *    result is TRUE only if the next printable character is within screen
 *    limits . The row and column positions are calculated for the default
 *    alphagraphics character size .
*/
int row, column;
  {
    char VAR_PTR name = "t_txtcur";
    START_UP( name, boolean );
    (void) sprintf( arg_str, "%d,%d", row, column );
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
    PLOTTER.requir.cursor._x = SETUP.tty_siz._x * (double) column;
    PLOTTER.requir.cursor._y = SETUP.ur_limit._y + 1.0 -
        SETUP.tty_siz._y * (double) ( row + 1 );
    if ( 0 <= column && column <= SETUP.alph_columns  &&
            0 <= row && row <= SETUP.alph_lines )
        return TRUE;
    (void) g_putmsg( NOTE, name, (char VAR_PTR) "Character not on screen/page");
    return FALSE;
  }


static boolean t_txtset( attrib, value )
/*
 *  Set the stroke character attribute `attrib' to `*value' .
 *  If `attrib' is negative, the current value of the attribute is returned
 *  to `*value' . This function fails if the `attrib' cannot be recognized or
 *  `value' is a null pointer . Appropriate error messages are generated .
*/
int attrib;
char *value;
  {
    char VAR_PTR name = "t_txtset";
    double *dvalue = (double *) value;   /* pointer to a double `value' */
    START_UP( name, boolean );

    switch( attrib )
      {
        case T_CH_RESET :
        case T_CH_BOLD :
        case T_CH_NOBOLD :
        case T_ASCII :
        case T_SYMBOL :
            (void) sprintf( arg_str, "%d=%s", attrib, pr_txtset( attrib ) );
            break;
        case T_CH_ALIGN :
            CATCH_NULL( value, name, boolean );
            (void) sprintf( arg_str, "%d=%s, x%.8lx->\"%.2s\"", attrib,
                pr_txtset( attrib), (unsigned long int) value, (char *) value );
            break;
        case T_CH_HEIGHT :
        case T_CH_ASPECT :
        case T_CH_SLANT :
        case T_CH_ANGLE :
        case T_CH_GAP :
            CATCH_NULL( value, name, boolean );
            (void) sprintf( arg_str, "%d=%s, x%.8lx->%.6g", attrib,
                pr_txtset( attrib ), (unsigned long int) dvalue, *dvalue );
            break;
        default :
            CATCH_NULL( value, name, boolean );
            (void) sprintf( arg_str, "%d=%s, x%.8lx", attrib,
                pr_txtset( attrib ), (unsigned long int) value );
            break;
      }
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
    switch( attrib )
      {
        case  T_CH_RESET :
            (void) reset_font( (Text_Attribute VAR_PTR) &STROKE,
                MAX( SETUP.pg_size._x, SETUP.pg_size._y ) / 20.0 );
            break;
        case  T_CH_HEIGHT :
            if ( STROKE.height != *dvalue )
              {
                STROKE.height = LIMIT( *dvalue, 5.0, SETUP.ur_limit._y );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_HEIGHT :
            *dvalue = STROKE.height;
            break;
        case  T_CH_ASPECT :
            if ( STROKE.aspect != *dvalue )
              {
                STROKE.aspect = LIMIT( *dvalue, 0.2, 5.0 );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_ASPECT :
            *dvalue = STROKE.aspect;
            break;
        case  T_CH_SLANT :
            if ( STROKE.slant != *dvalue )
              {
                STROKE.slant = LIMIT( *dvalue, -80.0, 80.0 ) * PI / 180.0;
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_SLANT :
            *dvalue = STROKE.slant * 180.0 / PI;
            break;
        case  T_CH_ANGLE :
            if ( STROKE.angle != *dvalue )
              {
                STROKE.angle = PI / 180.0 * ( *dvalue -
                    360.0 * (double) (int) ( *dvalue / 360.0 ) );
                STROKE.angle = STROKE.angle < -PI ? STROKE.angle + 2.0 * PI :
                    ( STROKE.angle > PI ? STROKE.angle-2.0*PI : STROKE.angle );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_ANGLE :
            *dvalue = STROKE.angle * 180.0 / PI;
            break;
        case  T_CH_GAP :
            STROKE.gap = LIMIT( *dvalue, 0.05, 5.0 );
            break;
        case -T_CH_GAP :
            *dvalue = STROKE.gap;
            break;
        case  T_CH_ALIGN :
            STROKE.align.vert  = TOLOWER( *value );
            value++;
            STROKE.align.horiz = TOLOWER( *value );
            if ( STROKE.align.horiz != 'l' && STROKE.align.horiz != 'c' &&
                    STROKE.align.horiz != 'r' )
                STROKE.align.horiz = 'l';
            if ( STROKE.align.vert != 'b' && STROKE.align.vert != 'c' &&
                    STROKE.align.vert != 't' )
                STROKE.align.vert = 'b';
            break;
        case -T_CH_ALIGN :
            *value++ = STROKE.align.vert;
            *value++ = STROKE.align.horiz;
            *value   = NUL;
            break;
        case T_ASCII :
            STROKE.font_ptr = &asc_font;
            break;
        case T_SYMBOL :
            STROKE.font_ptr = &sym_font;
            break;
        case T_CH_BOLD :
            STROKE.bold = TRUE;
            break;
        case T_CH_NOBOLD :
            STROKE.bold = FALSE;
            break;
        default :
            return make_error( ERR_ABORT + ERR_PRINT, name,
                (char VAR_PTR) "Unknown stroke text attribute" );
      }
    return TRUE;
  }


static boolean t_symdrw( point, symbol )
/*
 *      Draws a centred plotting symbol `symbol' at position `point' according
 *      to the selected stroke character attributes . This function fails if
 *      the font for the symbol is not defined or if it cannot write to the
 *      screen .
*/
coord_pair VAR_PTR point;
char symbol;
  {
    int font_index = symbol - 32;
    coord_pair offset, lower_left;
    char VAR_PTR name = "t_symdrw";
    StrokeChar VAR_PTR char_ptr;
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    (void) sprintf( arg_str, "x%.8lx->(%.6g,%.6g),'%c'",
        (unsigned long int) point, point->_x, point->_y, symbol );
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
    PLOTTER.requir.pen = *point;
/*
 *      for a colour device,
 *      set the colour and data level.
*/
    if ( SETUP.is_colour )
      {
        if ( SETUP.dev_type == SCREEN &&
             PLOTTER.requir.dots != PLOTTER.actual.dots )
            (void) set_dots( name, PLOTTER.requir.dots );
        if ( PLOTTER.requir.colour != PLOTTER.actual.colour )
            (void)  set_colour( name, PLOTTER.requir.colour );
      }
    else
      {
/*
 *      for a monochrome device,
 *      set colour to black or white (dashed lines not allowed)
*/
        if ( ! PLOTTER.requir.colour )
          {
            if ( PLOTTER.actual.dots != DOT_OFF )
                (void) set_dots( name, DOT_OFF );
          }
        else if ( SETUP.dev_type == SCREEN  &&
                  PLOTTER.requir.dots != PLOTTER.actual.dots )
            (void) set_dots( name, PLOTTER.requir.dots );
      }
/*
 *      set dashed line style to solid
*/
    if ( PLOTTER.actual.dash != SOLID_LINE )
        (void) set_style( name, SOLID_LINE );
    if ( G_ERROR() )
        return FALSE;
/*
 *           if necessary read character fonts
*/
    if ( ! STROKE.font_ptr->font_got )
      {
        (void) g_putmsg( INFORMATION | CONTINUED, name,
            (char VAR_PTR) "Reading character font from file " );
        (void) g_putmsg( INFORMATION, name, file_font( STROKE.font_ptr ) );
        if ( ! read_font( STROKE.font_ptr, file_font( STROKE.font_ptr ), name) )
            return FALSE;
      }
/*
 *      CALCULATE LOCATION OF LOWER LEFT CORNER OF THE SYMBOL
 *
 *      determine the character cell size vector
*/
    if ( STROKE.changed )
      {
        (void) map_font( (Text_Attribute VAR_PTR) &STROKE,
            SETUP.pen_thick );
        (void) g_putmsg( INFORMATION, name,
           (char VAR_PTR) "Calculated character transformation matrix" );
      }
    char_ptr = STROKE.font_ptr->font_dir + font_index;
    offset._x = STROKE.font_ptr->font_wid * 0.5;
    offset._y = char_ptr->font_midpt;
/*
 *    determine start position
*/
    lower_left._x = point->_x -
        STROKE.map[0][0] * offset._x - STROKE.map[0][1] * offset._y;
    lower_left._y = point->_y -
        STROKE.map[1][0] * offset._x - STROKE.map[1][1] * offset._y;

/*
 *      DRAW THE SYMBOL
*/
    if ( 0 <= font_index && font_index < STROKE.font_ptr->font_siz )
        if ( char_ptr->font_char == symbol && char_ptr->font_npts > 0 )
          {
            clipped_char( (coord_pair VAR_PTR) &lower_left, name, symbol );
            return move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
          }
    return make_error( ERR_ABORT + ERR_PRINT, name,
        (char VAR_PTR) "Unknown symbol" );
  }


static int t_txtdrw( string )
/*
 *      Draws the text contained in the character string `string' at the
 *   current cursor location according to the selected stroke character
 *   attributes . The returned value is the number of characters drawn . If
 *   an error occurs, the returned value is 0 . An error may be generated
 *   because
 *     a) a character in `string' is unavailable in the specified font
 *     b) the character is not printable
 *     c) could not update the pen position after drawing the characters
 *     d) it cannot write to the terminal
*/
char *string;
  {
    boolean ok = TRUE;          /* success flag                             */
    int font_index,             /* index of the character in font directory */
        n_chars,                /* length of `string'                       */
        char_count=0;           /* counts each character as it is drawn     */
    coord_pair cell_size,       /* size of the character cell               */
               offset,
               point;           /* location of char cell lower left corner  */
    char VAR_PTR name = "t_txtdrw";
    CATCH_NULL( string, name, boolean );
    START_UP( name, int );
    (void) sprintf( arg_str, "x%.8lx->\"%.50s\"", (unsigned long int) string,
        string );
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
/*
 *      for a colour device,
 *      set the colour and data level.
*/
    if ( SETUP.is_colour )
      {
        if ( SETUP.dev_type == SCREEN &&
             PLOTTER.requir.dots != PLOTTER.actual.dots )
            ok &= set_dots( name, PLOTTER.requir.dots );
        if ( PLOTTER.requir.colour != PLOTTER.actual.colour )
            ok &= set_colour( name, PLOTTER.requir.colour );
      }
    else
      {
/*
 *      for a monochrome device,
 *      set colour to black or white (dashed lines not allowed)
*/
        if ( ! PLOTTER.requir.colour )
          {
            if ( PLOTTER.actual.dots != DOT_OFF )
                ok &= set_dots( name, DOT_OFF );
          }
        else if ( SETUP.dev_type == SCREEN &&
                  PLOTTER.requir.dots != PLOTTER.actual.dots )
            ok &= set_dots( name, PLOTTER.requir.dots );
      }
/*
 *      set dashed line style to solid
*/
    if ( PLOTTER.actual.dash != SOLID_LINE )
        ok &= set_style( name, SOLID_LINE );
    if ( G_ERROR() )
        return FALSE;
/*
 *           if necessary read character fonts
*/
    if ( ! STROKE.font_ptr->font_got )
      {
        (void) g_putmsg( INFORMATION | CONTINUED, name,
            (char VAR_PTR) "Reading character font from file " );
        (void) g_putmsg( INFORMATION, name, file_font( STROKE.font_ptr ) );
        if ( ! read_font( STROKE.font_ptr, file_font( STROKE.font_ptr ), name) )
            return FALSE;
      }
    if ( STROKE.changed )
      {
        (void) map_font( (Text_Attribute VAR_PTR) &STROKE,
            SETUP.pen_thick );
        (void) g_putmsg( INFORMATION, name,
           (char VAR_PTR) "Calculated character transformation matrix" );
      }
/*
 *      CALCULATE LOCATION OF LOWER LEFT CORNER OF THE STRING
 *
 *      determine the character cell size vector
*/
    n_chars = t_strlen( string );
    cell_size._x = STROKE.font_ptr->font_wid *
        ( 1.0 + STROKE.gap / STROKE.aspect );
    cell_size._y = STROKE.font_ptr->font_hei;
    switch( STROKE.align.horiz )
      {
        case 'r' :  offset._x = cell_size._x * n_chars;
                    break;
        case 'c' :  offset._x = cell_size._x * n_chars / 2.0;
                    break;
        case 'l' :  offset._x = 0.0;
                    break;
      }
    switch( STROKE.align.vert )
      {
        case 't' :  offset._y = cell_size._y;
                    break;
        case 'c' :  offset._y = cell_size._y / 2.0;
                    break;
        case 'b' :  offset._y = 0.0;
                    break;
      }
/*
 *    determine start position
*/
    point._x = PLOTTER.requir.cursor._x - STROKE.map[0][0] * offset._x -
               STROKE.map[0][1] * offset._y;
    point._y = PLOTTER.requir.cursor._y - STROKE.map[1][0] * offset._x -
               STROKE.map[1][1] * offset._y;
    cell_size._y = STROKE.map[1][0] * cell_size._x;
    cell_size._x = STROKE.map[0][0] * cell_size._x;
/*
 *      DRAW EACH CHARACTER IN THE STRING
*/
    for( ; *string; string++, char_count++ )
      {
        font_index = *string - 32;
        if ( 0 <= font_index && font_index < STROKE.font_ptr->font_siz )
            if ( (STROKE.font_ptr->font_dir)[font_index].font_char ==
                     *string &&
                 (STROKE.font_ptr->font_dir)[font_index].font_npts > 0 )
              {
                clipped_char( (coord_pair VAR_PTR) &point, name, *string );
                point._x += cell_size._x;
                point._y += cell_size._y;
              }
            else
                (void) make_error( ERR_WARN + ERR_PRINT, name,
                    (char VAR_PTR) "character unavailable" );
         else
             (void) make_error( ERR_WARN + ERR_PRINT, name,
                 (char VAR_PTR) "not printable character" );
      }
    PLOTTER.requir.cursor = point;
    if ( ! ( ok &= move_clip( (coord_triple VAR_PTR) &PLOTTER.requir.pen,
                              VIEW_CLIP ) ) )
        (void) make_error( ERR_CALL + ERR_PRINT, name, mt_str );
    return char_count;
  }


static char VAR_PTR t_ch_def( what, character, index, point )
/*
 * This function allows the user to obtain the shape of any character defined
 * in the ASCII and SYMBOL fonts. With the exception of string alignment, the
 * transformations implied by t_txtset are applied to the returned coordinates.
 * This function is merely a call to the function ``define_font'' with an
 * additional argument defining which set of text attributes to use .
*/
int what;           /* specifies what to do (T_FT_* defined values) */
char character;     /* character to fetch from the font */
int index;          /* coordinate number   */
coord_pair
    VAR_PTR point;  /* returned coordinate */
  {
    char VAR_PTR name = "t_ch_def",
         VAR_PTR ret_value;
    START_UP( name, char VAR_PTR );
/*
 *          Function used message
*/
    (void) sprintf( arg_str, "%d=%s, %c=%xh, %d, x%lx",
        what, pr_ch_def( what ), character, character, index,
        (long unsigned int) point );
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
    if ( ! STROKE.font_ptr->font_got )
      {
        (void) g_putmsg( INFORMATION | CONTINUED, name,
            (char VAR_PTR) "Reading character font from file " );
        (void) g_putmsg( INFORMATION, name, file_font( STROKE.font_ptr ) );
      }
    if ( STROKE.changed )
        (void) g_putmsg( INFORMATION, name,
           (char VAR_PTR) "Calculated character transformation matrix" );
    ret_value = define_font( (Text_Attribute VAR_PTR) &STROKE,
        SETUP.pen_thick, what, character, index, point );
    switch( what )
      {
        case T_FT_SIZE  :
        case T_FT_POINT :
            (void) sprintf( arg_str, "returned coordinate = (%.6g,%.6g)",
                point->_x, point->_y );
            break;
        case T_FT_EXIST :
        case T_FT_FIRST :
        case T_FT_LAST  :
        case T_FT_MIDDLE :
            (void) sprintf( arg_str, "returned value = %s",
                ( *ret_value ? "TRUE" : "FALSE" ) );
            break;
        case T_FT_LENGTH :
            (void) sprintf( arg_str, "returned value = %d",
                *( (int VAR_PTR) ret_value ) );
      }
    (void) g_putmsg( INFORMATION, name, (char VAR_PTR) arg_str );
    return ret_value;
  }


static char VAR_PTR t_txtget( string, max_chars )
/*
 *    Reads a maximum of `max_chars' characters from `stdin' into `string'
 * until a newline character is read . The newline character is not stored in
 * `string' . `string' is terminated with NUL . The characters are echoed to
 * the plotter . The echoed characters are confined to the viewport . If the
 * cursor is initially outside the viewport, t_txtget() starts echoing
 * characters at the top left corner of the viewport. T_txtget will not echo
 * any more characters if it runs out of space at the bottom of the viewport .
 * If an error errors or no characters are read, a NULL pointer is returned,
 * otherwise `string' is returned .
 *
 * Programming notes.
 *  t_txtget only changes plotter.requir.cursor when the location of the start
 *  of the buffered string must be moved .
*/
char VAR_PTR string;
int max_chars;
  {
    char VAR_PTR ptr,
         VAR_PTR name = "t_txtget";
    START_UP( name, char VAR_PTR );
    (void) sprintf( arg_str, "(char *) x%.8lx, %d",
        (unsigned long int) string, max_chars );
    (void) g_putmsg( USE_FUNC, name, arg_str );
    CATCH_NULL( string, name, char VAR_PTR );
    for( ptr = string; ( *ptr = getchar() ) != '\n' && max_chars--; ptr++ );
    *ptr = NUL;
    if ( ptr != string )
      {
/*
 *              ensure cursor is in the viewport
 *              ( place it in top left corner if it is not )
*/
        coord_pair max_curs;
        max_curs._x = max_clip( VIEW_CLIP, _x ) - SETUP.tty_siz._x;
        max_curs._y = max_clip( VIEW_CLIP, _y ) - SETUP.tty_siz._y;
        if ( PLOTTER.requir.cursor._x < min_clip( VIEW_CLIP, _x ) ||
             PLOTTER.requir.cursor._x > max_curs._x            ||
             PLOTTER.requir.cursor._y < min_clip( VIEW_CLIP, _y ) ||
             PLOTTER.requir.cursor._y > max_curs._y )
          {
            PLOTTER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
            PLOTTER.requir.cursor._y = max_curs._y;
          }
/*
 *              place characters in a buffer until it runs out of the viewport
*/
        if ( PLOTTER.requir.cursor._x <= max_curs._x &&
             PLOTTER.requir.cursor._y >= min_clip( VIEW_CLIP, _y ) )
          {
            char buffer[BUFFER_LEN],
                 *buffer_ptr = buffer;
            int cursor_col = PLOTTER.requir.cursor._x;
            for( ptr = string; *ptr; ptr++ )
                if ( ' ' <= *ptr && *ptr <= '~' &&
                    PLOTTER.requir.cursor._y >= min_clip( VIEW_CLIP, _y ) &&
                    PLOTTER.requir.cursor._y <= max_clip( VIEW_CLIP, _y ) )
                  {
                    *buffer_ptr++ = *ptr;
/*
 *                      print the buffer if the right hand margin is reached
*/
                    if ( (cursor_col += (int) SETUP.tty_siz._x ) > max_curs._x)
                      {
                        *buffer_ptr = NUL;
                        (void) t_txtstr( (char VAR_PTR) buffer, buffer_ptr,
                            name );
                        PLOTTER.requir.cursor._x = cursor_col
                                                 = min_clip( VIEW_CLIP,_x );
                        PLOTTER.requir.cursor._y -= SETUP.tty_siz._y;
                        buffer_ptr = buffer;
                      }
/*
 *                      print the buffer if it is full
*/
                    else if ( buffer_ptr - buffer >= BUFFER_LEN )
                      {
                        *buffer_ptr = NUL;
                        (void) t_txtstr( (char VAR_PTR) buffer, buffer_ptr,
                            name );
                        buffer_ptr = buffer;
                      }
                  }
/*
 *                      print whatever remains in the buffer
*/
            if ( buffer_ptr != buffer )
              {
                *buffer_ptr = NUL;
                (void) t_txtstr( (char VAR_PTR) buffer, buffer_ptr, name );
              }
          }
        PLOTTER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
        PLOTTER.requir.cursor._y -= SETUP.tty_siz._y;
      }
    if ( *string )
        (void) g_putmsg( INFORMATION, name, string );
    else
        (void) g_putmsg( NOTE, name,
            (char VAR_PTR) "No characters received, returning NULL pointer" );
    return *string ? string : (char VAR_PTR) 0;
  }


static int t_txtput( string )
/*
 *    Print text string `string' on the plotter, starting at the place
 *  last specified by `t_txtmov' or 't_txtcur', using the colour specified by
 *  `t_colour' . Only those characters within the viewport are written to the
 *  screen. The returned value is the number of characters written to the
 *  screen, or zero if an error has occurred .
 *
 *  The character string `string' may contain printable characters and the
 *  control codes,
 *    HT : \t : \011 : move to the right by one character space
 *    BS : \b : \010 : move to the left by one character space
 *    CR : \r : \015 : move to the beginning of the current line
 *    LF : \n : \012 : move down one character space
 *    VT :    : \013 : move up one character space
 *    EM :    : \031 : move to top left hand side of the screen
 *  Any other characters are ignored .
 *
 *  This function only breaks the string into segments terminated by cursor
 *  control characters and positions the required_cursor appropriately before
 *  calling t_txtstr to put the characters on the screen.
 *
 * Programming notes,
 *  t_txtput only changes plotter.requir.cursor when the location of the start
 *  of the buffered string must be moved .
*/
char VAR_PTR string;        /* pointer to start of the character string */
  {
    char VAR_PTR ptr,
         VAR_PTR name = "t_txtput";
    int  char_count = 0;    /* number of characters printed             */
    CATCH_NULL( string, name, boolean );
    START_UP( name, int );
    (void) sprintf( arg_str, "x%.8lx->\"%.50s\"",
        (unsigned long int) string, (char *) string );
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
    for ( ptr = string; *ptr; ptr++ )
      {
        if ( *ptr == LF || *ptr == VT || *ptr == EM ||
             *ptr == CR || *ptr == BS || *ptr == HT )
          {
            char_count += t_txtstr( string, ptr, name ) + 1;
            switch( *ptr )
              {
                case LF : PLOTTER.requir.cursor._y -= SETUP.tty_siz._y;
                          break;
                case VT : PLOTTER.requir.cursor._y += SETUP.tty_siz._y;
                          break;
                case EM : PLOTTER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
                          PLOTTER.requir.cursor._y = max_clip( VIEW_CLIP, _y )
                              - SETUP.tty_siz._y + 1;
                          break;
                case CR : PLOTTER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
                          break;
                case BS : PLOTTER.requir.cursor._x -= SETUP.tty_siz._x;
                          break;
                case HT : PLOTTER.requir.cursor._x += SETUP.tty_siz._x;
                          break;
              }
            string = ptr + 1;
          }
      }
    if ( string != ptr )
        char_count += t_txtstr( string, ptr, name );
    return char_count;
  }


static boolean t_boxfil( lower_left, upper_right, style )
/*
 *      draw a rectangular box on the screen with lower left corner at
 * `*lower_left' and upper right corner at `*upper_right', and fill the box
 * with a hatching of type `style' . The hatching styles are :
 *
 *      0 = empty, box outline only      5 = horizontal lines
 *      1 = fill the box completely      6 = vertical lines
 *      2 = 50% `grey' fill              7 = diagonal cross hatch
 *      3 = NW-SE diagonal lines         8 = horizontal-vertical cross
 *      4 = NE-SW diagonal lines                 hatch
 * The box outline is clipped to the viewport limits .
*/
coord_pair *lower_left,                 /* lower left corner of the box     */
           *upper_right;                /* upper right corner of the box    */
int style;                              /* hatching style                   */
  {
    coord_pair ll_corner, ur_corner;
    char VAR_PTR name = "t_boxfil";
    CATCH_NULL( lower_left && upper_right, name, boolean );
    START_UP( name, boolean );
    (void) sprintf( arg_str, "x%.8lx->(%.6g,%.6g), x%.8lx->(%.6g,%.6g), %d",
        (unsigned long int) lower_left, lower_left->_x, lower_left->_y,
        (unsigned long int) upper_right, upper_right->_x, upper_right->_y,
        style );
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
/*
 *      clip the rectangle to the screen limits
*/
    ll_corner._x = LIMIT( lower_left->_x,
                     min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
    ll_corner._y = LIMIT( lower_left->_y,
                     min_clip( VIEW_CLIP, _y ), max_clip( VIEW_CLIP, _y ) );
    ur_corner._x = LIMIT( upper_right->_x,
                     min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
    ur_corner._y = LIMIT( upper_right->_y,
                     min_clip( VIEW_CLIP, _y ), max_clip( VIEW_CLIP, _y ) );
/*
 *      ensure correct ordering of coordinates
*/
    if ( ll_corner._x > ur_corner._x )
        SWAPD( ll_corner._x , ur_corner._x );
    if ( ll_corner._y > ur_corner._y )
        SWAPD( ll_corner._y , ur_corner._y );
/*
 *      draw the box
*/
    if ( ur_corner._x - ll_corner._x > 0.0 &&
         ur_corner._y - ll_corner._y > 0.0 )
      {
/*
 *      for a colour device,
 *      set the colour and data level.
*/
        if ( SETUP.is_colour )
          {
            if ( SETUP.dev_type == SCREEN &&
                 PLOTTER.requir.dots != PLOTTER.actual.dots )
                (void) set_dots( name, PLOTTER.requir.dots );
            if ( PLOTTER.requir.colour != PLOTTER.actual.colour )
                (void) set_colour( name, PLOTTER.requir.colour );
          }
        else
          {
/*
 *      for a monochrome device,
 *      set colour to black or white (dashed lines not allowed)
*/
            if ( ! PLOTTER.requir.colour )
              {
                if ( PLOTTER.actual.dots != DOT_OFF )
                    (void) set_dots( name, DOT_OFF );
              }
            else if ( SETUP.dev_type == SCREEN &&
                      PLOTTER.requir.dots != PLOTTER.actual.dots )
                (void) set_dots( name, PLOTTER.requir.dots );
          }
        if ( lower_left->_x != ll_corner._x || lower_left->_y != ll_corner._y ||
            upper_right->_x != ur_corner._x || upper_right->_y != ur_corner._y )
            (void) g_putmsg( NOTE, name, (char VAR_PTR)
                "Box corner changed by clipping or re-ordering" );
        return ! G_ERROR();
      }
    else
        (void) g_putmsg( NOTE, name,
            (char VAR_PTR) "Box not drawn -- clipped away" );
    return FALSE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Cross Hairs Routines                           */
/*                                                                          */
/* ************************************************************************ */

static boolean t_xhrget( point, char_pressed )
/*
 *      Get cross-hair location . If the terminal is not already in cross-hairs
 *  mode, then it is put into cross-hairs mode . The cross-hairs are positioned
 *  by the user, and the location of the cross-hairs is transmitted when an
 *  alpha-numeric key is pressed . The returned value of `t_xhrget' is TRUE if
 *  `point' is within the current viewport and `t_xhrget' can write to the
 *  graphics device ; otherwise the returned value is FALSE .
 *
 *    Arguments
*/
coord_pair VAR_PTR point;   /* pointer to the coordinates of the cross-hairs */
char VAR_PTR char_pressed;  /* character pressed to identify `point'         */
 {
    char VAR_PTR name = "t_xhrget";
    CATCH_NULL( point && char_pressed, name, boolean );
    START_UP( name, boolean );
    (void) sprintf( arg_str,"(coord_pair *) x%.8lx, (char *) x%.8lx",
        (unsigned long int) point, (unsigned long int) char_pressed );
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
    if ( SETUP.dev_type == HARDCOPY )
        return make_error( ERR_WARN + ERR_PRINT, name, inactive );
    (void) fputs(
        "t_xhrget : enter a character, abscissa & ordinate for cross-hairs ",
        stdout );
    (void) fscanf( stdin, "%c%lf%lf",
        char_pressed, &(point->_x), &(point->_y) );
    (void) sprintf( arg_str, "cross-hairs at (%.6g,%.6g) %c",
        point->_x, point->_y, *char_pressed );
    (void) g_putmsg( INFORMATION, name, arg_str );
    if ( min_clip( VIEW_CLIP, _x ) <= point->_x &&
         max_clip( VIEW_CLIP, _x ) >= point->_x &&
         min_clip( VIEW_CLIP, _y ) <= point->_y &&
         max_clip( VIEW_CLIP, _y ) >= point->_y )
        return TRUE;
    (void) g_putmsg( NOTE, name,
        (char VAR_PTR) "Cross-hairs not in viewport" );
    return FALSE;
 }


static boolean t_xhrput( point )
/*
 *      Put the  cross-hairs at `point' . If the terminal is not already in
 *  cross-hairs mode, then it is put into cross-hairs mode . The returned value
 *  of `t_xhrput' is TRUE if `point' is within the current viewport and
 *  `t_xhrput' can write to the graphics device ; otherwise the returned value
 *  is FALSE .
 *
 *    Arguments
*/
coord_pair VAR_PTR point;   /* pointer to the coordinates of the cross-hairs */
 {
    char VAR_PTR name = "t_xhrput";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    (void) sprintf( arg_str, "x%.8lx->(%.6g,%.6g)", (unsigned long int) point,
        point->_x, point->_y );
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
    PLOTTER.requir.cursor = *point;
    if ( SETUP.dev_type == HARDCOPY )
        return make_error( ERR_WARN + ERR_PRINT, name, inactive );
    if ( min_clip( VIEW_CLIP, _x ) <= point->_x &&
         max_clip( VIEW_CLIP, _x ) >= point->_x &&
         min_clip( VIEW_CLIP, _y ) <= point->_y &&
         max_clip( VIEW_CLIP, _y ) >= point->_y )
        return TRUE;
    (void) g_putmsg( NOTE, name,
        (char VAR_PTR) "Cross-hairs not in viewport" );
    return FALSE;
 }


/* ************************************************************************ */
/*                                                                          */
/*                        Print terminal status                             */
/*                                                                          */
/* ************************************************************************ */


static char VAR_PTR t_status( attribute, value )
/*
 *      Put status attribute `attribute' into the coordinate pair `*value' .
*/
int attribute;
char VAR_PTR value;
  {
    char VAR_PTR name = "t_status",
         VAR_PTR dev_name = "dummy graphics device";
    CATCH_NULL( value, name, char VAR_PTR );
    if ( attribute != T_ST_SCREEN && attribute != T_ST_CATEG )
        START_UP( name, char VAR_PTR );
    if ( dummy )
      {
        (void) sprintf( arg_str, "%d=%s, (char *) x%.8lx", attribute,
            pr_status(attribute), (unsigned long int) value );
        (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
      }
    switch( attribute )
      {
        case T_ST_DEVICE:
            (void) t_strcpy( value, dev_name );
            (void) t_strcpy( arg_str, dev_name );
            break;
        case T_ST_SCREEN:
            *( (boolean VAR_PTR) value ) = SETUP.dev_type == SCREEN;
            break;
        case T_ST_CATEG:
            *( (int VAR_PTR) value ) = T_ST_FILE;
            (void) t_strcpy( arg_str, "T_ST_FILE" );
            break;
        case T_ST_MONO:
            *( (boolean VAR_PTR) value ) = ! SETUP.is_colour;
            break;
        case T_ST_CLIP:
            *( (clip_area VAR_PTR VAR_PTR) value ) = VIEW_CLIP;
            arg_str[0] = '\000';
            if ( SETUP.msg_type <= NOTE )
                (void) what_clip( OP_FILE, VIEW_CLIP );
            break;
        case T_ST_MAX:
            *( (coord_pair VAR_PTR) value) = SETUP.ur_limit;
            break;
        case T_ST_MIN:
            *( (coord_pair VAR_PTR) value) = SETUP.ll_limit;
            break;
        case T_ST_CHAR:
            *( (coord_pair VAR_PTR) value) = SETUP.tty_siz;
            break;
        case T_ST_LINES:
            ( (coord_pair VAR_PTR) value)->_x = (double) SETUP.alph_columns;
            ( (coord_pair VAR_PTR) value)->_y = (double) SETUP.alph_lines;
            break;
        case T_ST_PIXSIZ:
            ( (coord_pair VAR_PTR) value)->_x =
                ( (coord_pair VAR_PTR) value)->_y = PIXEL_SIZE;
            break;
        case T_ST_A_CURS:
            ( (coord_pair VAR_PTR) value)->_x = PLOTTER.actual.cursor._x;
            ( (coord_pair VAR_PTR) value)->_y = PLOTTER.actual.cursor._y;
            break;
        case T_ST_R_CURS:
            ( (coord_pair VAR_PTR) value)->_x = PLOTTER.requir.cursor._x;
            ( (coord_pair VAR_PTR) value)->_y = PLOTTER.requir.cursor._y;
            break;
        case T_ST_A_PEN:
            ( (coord_pair VAR_PTR) value)->_x = PLOTTER.actual.pen._x;
            ( (coord_pair VAR_PTR) value)->_y = PLOTTER.actual.pen._y;
            break;
        case T_ST_R_PEN:
            ( (coord_pair VAR_PTR) value)->_x = PLOTTER.requir.pen._x;
            ( (coord_pair VAR_PTR) value)->_y = PLOTTER.requir.pen._y;
            break;
        default:
            return (char VAR_PTR) make_error( ERR_ABORT + ERR_PRINT, name,
                (char VAR_PTR) "unknown attribute" );
      }
    if ( dummy )
      {
        if ( attribute == T_ST_SCREEN || attribute == T_ST_MONO )
            (void) sprintf( arg_str, "%s",
                *((boolean VAR_PTR) value ) ? "TRUE" : "FALSE" );
        else if ( attribute != T_ST_DEVICE && attribute != T_ST_CATEG &&
                  attribute != T_ST_CLIP )
            (void) sprintf( arg_str, "(%.6g,%.6g)",
                ((coord_pair VAR_PTR) value)->_x,
                ((coord_pair VAR_PTR) value)->_y );
        (void) g_putmsg( INFORMATION, name, (char VAR_PTR) arg_str );
      }
    return value;
  }


static boolean t_pr_ter( out_file )
FILE *out_file;
  {
    char VAR_PTR name = "t_pr_ter";
    CATCH_NULL( out_file, name, boolean );
    START_UP( name, boolean );
    (void) sprintf( arg_str, "(FILE *) x%.8lx", (unsigned long int) out_file );
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
    (void) fprintf( out_file, " TERMINAL STATUS\n" );
    (void) fprintf( out_file,
                    "               ---requir---     ---actual---\n" );
    (void) fprintf( out_file, "data level  %15s  %15s\n",
                    pr_dots( &PLOTTER.requir ), pr_dots( &PLOTTER.actual ) );
    (void) fprintf( out_file, "mode        %15s  %15s\n",
                    pr_mode( &PLOTTER.requir ), pr_mode( &PLOTTER.actual ) );
    (void) fprintf( out_file, "colour      %15d  %15d\n",
                    PLOTTER.requir.colour, PLOTTER.actual.colour );
    (void) fprintf( out_file, "dash style  %15d  %15d\n",
                    PLOTTER.requir.dash, PLOTTER.actual.dash );
    (void) fprintf( out_file, "cursor      (%6.1f,%6.1f)  (%6.1f,%6.1f)\n",
                    PLOTTER.requir.cursor._x, PLOTTER.requir.cursor._y,
                    PLOTTER.actual.cursor._x, PLOTTER.actual.cursor._y );
    (void) fprintf( out_file, "pen         (%6.1f,%6.1f)  (%6.1f,%6.1f)\n",
                    PLOTTER.requir.pen._x, PLOTTER.requir.pen._y,
                    PLOTTER.actual.pen._x, PLOTTER.actual.pen._y );
    if ( ferror( out_file ) )
        return make_error( ERR_WARN + ERR_PRINT, name,
            (char VAR_PTR) "Cannot write terminal status to file" );
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Vector Drawing                                 */
/*                                                                          */
/* ************************************************************************ */


static boolean t_vecmov( point )
/*
 *      move the line drawing pen to `point'
*/
coord_pair VAR_PTR point;
  {
    char VAR_PTR name = "t_vecmov";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    (void) sprintf( arg_str, "x%.8lx->(%.6g,%.6g)",
        (unsigned long int) point, point->_x, point->_y );
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );
    PLOTTER.requir.pen = *point;
    return move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
  }


static boolean t_vecdrw( point )
/*
 *      Draw from the previous pen position to `point' according to the
 *  required pen colour, dashed line style and dots-level . For a monochrome
 *  dashed line style takes precedence over colours . Lines are clipped to
 *  the viewport boundaries .
*/
coord_pair *point;
  {
    boolean ok = TRUE;          /* success flag */
    char VAR_PTR name = "t_vecdrw";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    (void) sprintf( arg_str, "x%.8lx->(%.6g,%.6g)",
        (unsigned long int) point, point->_x, point->_y );
    (void) g_putmsg( USE_FUNC, name, (char VAR_PTR) arg_str );

/*
 * If required dashed line style is 0, plot the point if it is in the viewport .
*/
    if ( ! PLOTTER.requir.dash )
      {
        ok &= move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
        (void) g_putmsg( NOTE, name, "drawing end-points only" );
        if ( min_clip( VIEW_CLIP, _x ) <= point->_x &&
             max_clip( VIEW_CLIP, _x ) >= point->_x &&
             min_clip( VIEW_CLIP, _y ) <= point->_y &&
             max_clip( VIEW_CLIP, _y ) >= point->_y )
          {
/*
 *              COLOUR devices
 *      Set the data level and colour.
*/
            if ( SETUP.is_colour )
              {
                if ( SETUP.dev_type == SCREEN &&
                     PLOTTER.actual.dots != PLOTTER.requir.dots )
                    ok &= set_dots( name, PLOTTER.requir.dots );
                if ( PLOTTER.actual.colour != PLOTTER.requir.colour )
                    ok &= set_colour( name, PLOTTER.requir.colour );
              }
            else
/*
 *              MONOCHROME devices
 *      If required colour is black, put terminal data level into DOT_OFF,
 *      otherwise use the required data level .
*/
              {
                if ( ! PLOTTER.requir.colour )
                  {
                    if ( PLOTTER.actual.dots != DOT_OFF )
                        ok &= set_dots( name, DOT_OFF );
                  }
                else if ( SETUP.dev_type == SCREEN &&
                          PLOTTER.actual.dots != PLOTTER.requir.dots )
                    ok &= set_dots( name, PLOTTER.requir.dots );
              }
/*
 *          plot the point
*/
            if ( PLOTTER.actual.dash )
                ok &= set_style( name, 0 );
            PLOTTER.actual.cursor._x = PLOTTER.actual.pen._x = point->_x;
            PLOTTER.actual.cursor._y = PLOTTER.actual.pen._y = point->_y;
          }
        else
            (void) g_putmsg( NOTE, name, "point clipped away" );
      }
/*
 *          Need to draw a line . First clip the endpoints .
*/
    else if ( clip( (coord_triple VAR_PTR) point, VIEW_CLIP ) )
      {
/*
 *              COLOUR devices
 *      Set the data level, colour and dashed line style .
*/
        if ( SETUP.is_colour )
          {
            if (  SETUP.dev_type == SCREEN &&
                  PLOTTER.actual.dots != PLOTTER.requir.dots )
                ok &= set_dots( name, PLOTTER.requir.dots );
            if ( PLOTTER.actual.colour != PLOTTER.requir.colour )
                ok &= set_colour( name, PLOTTER.requir.colour );
            if ( PLOTTER.actual.dash != PLOTTER.requir.dash )
                ok &= set_style( name, PLOTTER.requir.dash );
          }
        else
/*
 *              MONOCHROME devices
 *      If required colour is black, put terminal data level into DOT_OFF,
 *      otherwise use the required data level .
*/
          {
            if ( ! PLOTTER.requir.colour )
              {
                if ( PLOTTER.actual.dots != DOT_OFF )
                    ok &= set_dots( name, DOT_OFF );
              }
            else if ( SETUP.dev_type == SCREEN &&
                      PLOTTER.actual.dots != PLOTTER.requir.dots )
                ok &= set_dots( name, PLOTTER.requir.dots );
/*
 *          If a continuous line is required, use the dashed line style as
 *          pseudo-colour, otherwise just set the required dashed line style
*/
            if ( PLOTTER.requir.dash == 1 && PLOTTER.requir.colour )
              {
                if ( PLOTTER.requir.colour != PLOTTER.actual.dash )
                    ok &= set_style( name, PLOTTER.requir.colour );
              }
            else if ( PLOTTER.actual.dash != PLOTTER.requir.dash )
                ok &= set_style( name, PLOTTER.requir.dash );
          }
        if ( diff_point( (coord_pair VAR_PTR) &PLOTTER.requir.pen,
                         (coord_pair VAR_PTR) FIRST_clip( VIEW_CLIP ) ) ||
             diff_point( point, (coord_pair VAR_PTR) LAST_clip( VIEW_CLIP ) ) )
            (void) g_putmsg( INFORMATION, name,
                (char VAR_PTR) "end(s) of line clipped" );
/*
 *              draw the line segment
*/
        PLOTTER.actual.cursor._x = PLOTTER.actual.pen._x =
            last_clip( VIEW_CLIP, _x );
        PLOTTER.actual.cursor._y = PLOTTER.actual.pen._y =
            last_clip( VIEW_CLIP, _y );
      }
    else
        (void) g_putmsg( NOTE, name, (char VAR_PTR) "line clipped away" );
    PLOTTER.requir.pen = *point;
    if ( ! ok )
        (void) make_error( ERR_WARN + ERR_PRINT, name, err_write );
    return ok;
  }

/* ************************************************************************ */
/* ************************************************************************ */
/*                                                                          */
/*                    Array of pointers to driver functions                 */
/*                                                                          */
/* With the exception of the function ``t_driver'' which is used to select  */
/* the required driver, this array is the ONLY point of access to the       */
/* functions in the driver .                                                */
/*                                                                          */
/* ************************************************************************ */
/* ************************************************************************ */

/*
 * static clip_area VAR_PTR t_begin();
 * static boolean
 *     t_end(),    t_vwport(), t_colour(), t_dashed(), t_thick(),  t_gr_off(),
 *     t_gr_on(),  t_pr_ter(), t_dump(),   t_visib(),  t_invis(),  t_revers(),
 *     t_replac(), t_newpag(), t_newscr(), t_bell(),   t_wait(),   t_txtcur(),
 *     t_txtmov(), t_txtset(), t_symdrw(), t_xhrget(), t_xhrput(), t_vecdrw(),
 *     t_boxfil(), t_vecmov();
 * static char
 *         VAR_PTR t_txtget(), VAR_PTR t_status(), VAR_PTR t_ch_def();
 * static int t_txtput(), t_txtdrw();
 */

T_dev_fns Dum_dev_ptrs = {
            t_begin,  t_end,    t_vwport, t_gr_on,  t_gr_off,
            t_colour, t_dashed, t_thick,  t_visib,  t_invis,
            t_revers, t_replac, t_newpag, t_newscr, t_bell,
            t_wait,   t_txtmov, t_txtcur, t_txtset, t_symdrw,
            t_txtdrw, t_txtget, t_txtput, t_status, t_pr_ter,
            t_vecmov, t_vecdrw, t_dump,   t_xhrput, t_xhrget,
            t_boxfil, t_ch_def };

