/*
    Copyright (C) 1997  Peter Vernon Lanspeary

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "system.h"
#include <stdio.h>

#ifndef POINTER_DEFINED
#include POINTER_HEADER
#endif

#include ERROR_HEADER
/* ************************************************************************ */
/*                                                                          */
/*                      ERROR CONDITION HANDLERS                            */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/*      These functions allow the user to record errors detected by other   */
/*  functions . Pointers to the error message strings are stored on a       */
/*  queue .The error messages can be printed or written to a file . The     */
/*  user can test whether or not errors have occurred, and can clear any    */
/*  recorded errors .                                                       */
/*                                                                          */
/*      Any recorded error has a `status' which indicates the severity of   */
/*  the error . Permissible values of `status' -- first argument of         */
/*  `make_error', `is_error' and `dump_error' -- are,                       */
/*                                                                          */
/*       ERR_CALL  is used only with `make_error' .                         */
/*       ERR_ALL   is used with `is_error' or `dump_error' .                */
/*       ERR_FATAL is the highest message status .                          */
/*       ERR_NOTE  is the lowest message status .                           */
/*       ERR_PRINT is a flag used only with `make_error'                    */
/*                                                                          */
/*  The following values can be added to the `status' to indicate a `math'  */
/*  error . The presence of one of these additional information to be       */
/*  queued .                                                                */
/*       ERR_DOMAIN     Argument(s) not in domain of function               */
/*       ERR_OVER       Result is greater than maximum allowed              */
/*       ERR_UNDER      Result is indistinguishable from 0.0                */
/*       ERR_SING       Argument produces singularity                       */
/*       ERR_TLOSS      Argument produces total loss of significant digits  */
/*       ERR_PLOSS      Arg. produces partial loss of significant digits    */
/*                                                                          */
/* External functions used : fprintf(), free(), malloc()                    */
/*                                                                          */
/* ************************************************************************ */

#define SEVERITY_MASK  ( ERR_NOTE | ERR_WARN | ERR_ABORT | ERR_FATAL )
#define MATH_MASK      ( ERR_DOMAIN | ERR_OVER | ERR_UNDER | \
                         ERR_SING | ERR_TLOSS | ERR_PLOSS )
#define MIN_MATH       ERR_DOMAIN
/*
 *                              TYPE DEFINITIONS
 *                     ********************************
*/
typedef struct error_st         /* error message queue structure */
  {
    int  type;                  /* category of error             */
    char VAR_PTR location;      /* name of routine where error occurred */
    char VAR_PTR message;       /* error message                 */
    struct error_st VAR_PTR next_err;  /* pointer to next error message */
  }
    VAR_PTR Error;

/*
 *                             GLOBAL VARIABLES
 *                     ********************************
*/
static Error
  first_error = 0,          /* pointer to first logged error message    */
  curr_error = 0;           /* pointer to last logged error message     */
FILE *error_file = (FILE *) 0; /* output file for writing error messages */

static char
    VAR_PTR err_math[] = { "", "Domain", "Overflow", "Underflow", "Singularity",
            "Total loss of significance", "Partial loss of significance", "", },
                            /* Printed text for math errors */
    VAR_PTR err_sev[] = { "", "Note", "Warning", "Abort", "Fatal" },
                            /* printed text for error status/severity   */
    VAR_PTR err_unknown = "unknown function";

char *malloc();
int free(), exit();

/*
 *                                FUNCTIONS
 *                    ********************************
*/

boolean make_error( status, where, text )
/*
 *    Records an error message of severity `status' for dumping at a
 *  later time . `where' and `text' are pointers to static character
 *  strings . `where' is the name of the function where the error has
 *  been detected . `text' is the text of the error message .
 *  If the ERR_PRINT bit in `status' is set, the error message is written
 *  to the error file ( set by `file_error()', default is `stderr' ) .
 *  If status contains a `math' error value, then `text' is interpreted as
 *  a pointer to `double' and the text of the error message reports the
 *  nature of the error and , if `text' is not a NULL pointer, the value of
 *  *( (double VAR_PTR) text) . If the error is not a `math' error and `text'
 *  in not a NULL pointer, `text' should be a static character string so
 *  that it does not disappear when the calling function returns .
*/
int status;
char VAR_PTR where, VAR_PTR text;
  {
    Error new_ptr;
    if ( ( status & SEVERITY_MASK ) == ERR_CALL && curr_error )
        text = curr_error->location;
/*
 *          allocate memory for the queue entry
*/
    if ( ( new_ptr = (Error) malloc( sizeof( struct error_st ) ) ) )
      {
/*
 *          fix pointer to current error
*/
        if ( curr_error )
            curr_error->next_err = new_ptr;
        else
            first_error = new_ptr;
/*
 *          fill in the queue entries
 *  in the case of a math error with non-NULL `text', the numeric value should
 *  be saved in allocated memory first .
*/
        curr_error = new_ptr;
        curr_error->type = status;
        curr_error->location = where ? where : err_unknown;
        if ( ( status & SEVERITY_MASK ) == ERR_CALL &&
             ( status & MATH_MASK ) && text )
          {
            curr_error->message = malloc( (unsigned) sizeof(double) );
            *((double VAR_PTR) curr_error->message) = *((double VAR_PTR) text);
          }
        else
            curr_error->message  = text;
        curr_error->next_err = (Error) 0;
/*
 *          write the error ( if required )
*/
        if ( status & ERR_PRINT )
          {
            if ( ! error_file )
                error_file = stderr;
            if ( ( status & SEVERITY_MASK ) == ERR_CALL )
                (void) fprintf( error_file,
                    "Error in value returned to %s from %s", where, text );
            else
              {
                (void) fprintf( error_file, "%s error in %s: ",
                    err_sev[ status & SEVERITY_MASK ], where );
                if ( status & MATH_MASK )
                  {
                    (void) fputs( (char *)
                        err_math[ ( status & MATH_MASK ) / MIN_MATH ],
                        error_file );
                    if ( text )
                        (void) fprintf( error_file, " from argument %f",
                            *( (double VAR_PTR) text ) );
                  }
                else if ( text )
                    (void) fputs( (char *) text, error_file );
              }
            (void) fputc( '\n', error_file );
          }
      }
    else
      {
        (void) fputs( "\nCannot allocate memory in make_error\n", stderr );
        exit( 1 );
      }
    return FALSE;
  }


boolean file_error( op_file )
/*
 *   Sets the output error file used by `make_error()' and `dump_error()' .
 * The returned result is TRUE unless `op_file' is a NULL pointer .
*/
FILE *op_file;      /* pointer to error file */
  {
   if ( ! op_file )
      {
        error_file = op_file;
        return TRUE;
      }
    return FALSE;
  }


boolean is_error( status )
/*
 *    Has an error of status `status' or higher been recorded ?
 * Returns TRUE if an error of the required status has been recorded,
 * and FALSE if no error has been logged .
 *
 *  Synopsis :    boolean errors_exist = is_error( status );
*/
int status;
  {
    Error ptr;
    for( ptr = first_error; ptr; ptr = ptr->next_err )
        if ( ( ptr->type & SEVERITY_MASK ) >= status )
            return TRUE;
    return FALSE;
  }


void clear_error()
/*
 *    Clear all recorded errors .
 *
 *  Synopsis :    clear_error();
*/
  {
    Error next_ptr, ptr;
    for( ptr = first_error; ptr; ptr = next_ptr )
      {
        next_ptr = ptr->next_err;
        if ( ( ptr->type & SEVERITY_MASK ) == ERR_CALL &&
             ( ptr->type & MATH_MASK ) && ptr->message )
            (void) free( (char VAR_PTR) ptr->message );
        (void) free( (char VAR_PTR) ptr );
      }
    curr_error = first_error = 0;
  }


int dump_error( status )
/*
 *    Dumps all logged errors with a status `status' or higher onto the
 * error file ( set by `file_error()', default is `stderr' ) .
 *    Returns the number of errors printed .
 *
 *  Synopsis :    int no_of_errors = dump_error( status );
*/
int status;
  {
    Error ptr;
    int err_count = 0;
    if ( ! error_file )
        error_file = stderr;
    for( ptr = first_error; ptr; ptr = ptr->next_err )
        if ( ( ptr->type & SEVERITY_MASK ) >= status )
          {
            if ( ( ptr->type & SEVERITY_MASK ) == ERR_CALL )
                (void) fprintf( error_file,
                    "Error in value returned to %s from %s",
                    ptr->location, ptr->message );
            else
              {
                (void) fprintf( error_file, "%s error in %s: ",
                    err_sev[ ptr->type & SEVERITY_MASK ], ptr->location );
                if ( ptr->type & MATH_MASK )
                  {
                    (void) fputs( (char *)
                        err_math[ ( ptr->type & MATH_MASK ) / MIN_MATH ],
                        error_file );
                    if ( ptr->message )
                        (void) fprintf( error_file, " from argument %f",
                            *( (double VAR_PTR) ptr->message ) );
                  }
                else if ( ptr->message )
                    (void) fputs( (char *) ptr->message, error_file );
              }
            (void) fputc( '\n', error_file );
            err_count++;
          }
    if ( err_count )
        (void) fprintf( error_file, "%d errors printed\n", err_count );
    return err_count;
  }


