/*
    Copyright (C) 1997  Peter Vernon Lanspeary

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* ************************************************************************ */
/*                                                                          */
/*              Basic Graphics Terminal Graphics Driver for the             */
/*            Hercules Graphics Card for IBM PCs and true clones .          */
/*                N.B. This is hardware dependent stuff                     */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/*   This is a graphics driver for the Hercules Graphics Card . While       */
/*   graphics is switched on, all terminal I/O should be through this       */
/*   driver . Separate pens are used for drawing lines and for text . Text  */
/*   can be written onto the graphics screen either with a stroke           */
/*   generating font or using the terminal's firmware pixel font .          */
/*      Most functions are of type `boolean' and return a TRUE or FALSE     */
/*   result to indicate success or failure . Error messages are placed on   */
/*   the error queue -- which can be examined at the discretion of the user.*/
/*      External functions used :                                           */
/*         boolean make_error();                                            */
/*         anything fprintf(), fputc(), fputs(), free();                    */
/*         int ferror(), fflush(), fscanf(), strlen();                      */
/*         double fabs(), sin(), cos();                                     */
/*         char *malloc(), *strcpy();                                       */
/*         clip_area *def_2rclip();s();                                     */
/*   Any programs which use this driver must #include DEVICE_G_HEADER .     */
/*   These functions must be linked with "clip.c" and "error.c". Functions  */
/*   in "clip.c" and "error.c" should be included in the object library     */
/*   "herc_g.lib". The documentation on "clip.c" and "error.c" should be    */
/*   read before continuing.                                                */
/*                                                                          */
/* ************************************************************************ */

#include "system.h"
#include <stdio.h>
#include DEVICE_G_HEADER
#include FONT_SOURCE_HEADER
#include ERROR_HEADER
#include "dev_udef.h"
/*
 *  non printable ASCII characters
*/
#define NUL  '\000'   /* Nul */
#define BEL  '\007'   /* Bell */
#define BS   '\010'   /* Backspace */
#define HT   '\011'   /* Horizontal Tabulation */
#define LF   '\012'   /* Line Feed */
#define VT   '\013'   /* Vertical Tabulation */
#define CR   '\015'   /* Carriage Return */
#define EM   '\031'   /* End of Medium */
#define SP   '\040'   /* Space */
#define DEL  '\177'   /* Delete */

#define PI   3.141592653589793238462643

#define HERC_X       719.0 /* viewport upper right coordinate - pixels */
#define HERC_Y       347.0 /*   for the Visual 500 terminal            */

#define TEXT_PAGE     0    /* Hercules Page 0 at 0B0000h for text      */
#define GRAPHICS_PAGE 1    /* Hercules Page 1 at 0B8000h for graphics  */
#define ALPH_HEIGHT   8.0  /* height in pixels of alphagraphics char   */
#define ALPH_WIDTH    8.0  /* width in pixels of alphagraphics char    */
#define ALPH_LINES    43   /* number lines on alphagraphics screen     */
#define ALPH_COLUMNS  90   /* number columns on alphagraphics screen   */
#define PEN_THICKNESS 1.0  /* pen or line thickness                    */
#define SOLID_LINE    -1   /* pixel mask for drawing solid lines       */
#define X_PIXEL_SIZE  0.3  /* horiz. and vertical size of a pixel (mm) */
#define Y_PIXEL_SIZE  0.5  /* horiz. and vertical size of a pixel (mm) */
#define CLIP2R_ID   0x3c22 /* CL2R two dimensional rectangular clipping*/

/*
 *      Dashed line styles
 *   bit pattern     Hex    dec itype  Description
 * 1111111111111111  FFFF     -1  0   solid
 * 0011001100110011  3333  13107  1   dotted
 * 0011111100111111  3F3F  16191  2   short dash
 * 0001111111111111  1FFF   8191  3   long dash
 * 0011001111111111  33FF  13311  4   dot, dash
 * 0011110011111111  3CFF  15615  5   short dash, long dash
 * 0011001100111111  333F  13119  6   2 short dashes, long dash
 * 0010010010011111  0000   9407  7   nothing
*/

/* ************************************************************************ */
/*                                                                          */
/*                             GRAPHICS MACROS                              */
/*                                                                          */
/* ************************************************************************ */
/*
 *  ABBREVIATIONS
 ******************
*/
#define  TER        herc->ter
#define  STROKE     herc->stroke
#define  VIEW_CLIP  herc->view_clip
/*
 *  CATCH_NULL(pointer,caller,ret_type)
 ****************************************
 *      If "pointer" has a NULL value, an "Arg is NULL pointer" error message
 * is generated and the function is aborted . Otherwise, nothing is done .
*/
#define CATCH_NULL(pointer,caller,ret_type) \
  if ( ! (pointer) )    \
      return (ret_type) make_error( ERR_ABORT, caller, err_null )
/*
 *  MAX( x, y )
 ****************
 *      The returned value is the maximum of `x' and `y' . */
#define MAX(x,y)        ( (x) > (y) ? (x) : (y) )
/*
 *  MIN( x, y )
 ****************
 *      The returned value is the minimum of `x' and `y' . */
#define MIN(x,y)        ( (x) > (y) ? (y) : (x) )
/*
 *  LIMIT( x, min, max )
 *************************
 *      Limits the value of `x' to between `min' and `max' */
#define LIMIT(x,min,max)  (x) <= (min) ? (min) : ( (x) >= (max) ? (max) : (x) )
/*
 *  TOLOWER( c )
 ****************
 *      Converts character `c' to lower case if it is upper case */
#define TOLOWER(c)      'A' <= c && c <= 'Z' ? c + 'a' - 'A' : c;
/*
 *  START_UP(caller,ret_type)
 ******************************
 *      Calls `t_begin' with the default arguments `stdout' and `NULL'
 *  if `t_begin' has not previously been called .*/
#define START_UP(caller,ret_type) \
    if ( ! herc ) \
        if ( ! t_begin( stdout, mt_str ) ) \
            return (ret_type) make_error( ERR_CALL, caller, mt_str )
/*
 *  SWAPD( a, b )
 *****************
 *      swaps the two double precision floating point values `a' and `b' */
#define SWAPD(a,b)      { double temp = b; b = a; a = temp; }


/* ************************************************************************ */
/*                                                                          */
/*                             TYPE DECLARATIONS                            */
/*                                                                          */
/* ************************************************************************ */

enum dev_mode     /* terminal mode */
  {
    ALPH_SCREEN,  /* alphanumeric screen  - normal, Herc Page 0 */
    GRAP_SCREEN,  /* graphics screen - Herc Page 1 */
  };

/*
 *      The `data level' controls how points, lines, text and boxes are put
 *   on the graphics screen . These `data levels' are,
*/
enum dev_dots     /* data level */
  {
    DOT_ON,       /* draw visible vectors, `or' them onto the screen */
    DOT_OFF,      /* draw invisible vectors, switch dots off         */
    DOT_REV,      /* the compliment of the new data is plotted       */
    DOT_NEW       /* replace pixel contents with new data - not used */
  };

typedef struct            /* TERMINAL STATUS   */
  {                       /*     DATA TYPE     */
    enum dev_mode mode;   /* terminal mode     */
    enum dev_dots dots;   /* data level        */
    int           colour; /* foreground colour */
    int           dash;   /* dashed line style */
    int           thick;  /* line and test width                 */
    coord_pair    cursor; /* alpha graphics text cursor position */
    coord_pair    pen;    /* vector graphics pen position */
  }
    ter_status;

/* Graphics status description -- this data structure contains everything
 * the driver knows about the graphics device .
*/
typedef struct
  {
    struct
      {
        ter_status requir;       /* required terminal status */
        ter_status actual;       /* actual terminal status   */
      }
        ter;
    Text_Attribute stroke;       /* stroke text attributes */
    clip_area VAR_PTR view_clip; /* pointer to viewport clipping area */
    FILE * handle;               /* device identifier */
  }
    Gfx_Status;

/* ************************************************************************ */
/*                                                                          */
/*                             GLOBAL VARIABLES                             */
/*                                                                          */
/* ************************************************************************ */

extern StrokeFont
    asc_font,   /* font for ASCII chars   */
    sym_font;   /* font for special chars */

static Gfx_Status
    VAR_PTR herc = (Gfx_Status VAR_PTR) 0; /* pointer to graphics status */

static int
    style_mask[] = { 0, SOLID_LINE, 13107, 16191, 8191, 13311, 15615, 13119,
                     9407 };

/*
 *      Assembler (Firmware) functions
*/
void Hg_clear(), Hg_dots(), Hg_on(), Hg_off(), Hg_style();
int  Hg_box(), Hg_draw(), Hg_dump(), Hg_getc(), Hg_getx(), Hg_move(),
     Hg_point(), Hg_putc(), Hg_putx();

static char VAR_PTR pr_dots(), VAR_PTR pr_mode();
static boolean t_stkchr(), next_cursor(), prev_cursor();
static void put_char();

char   *malloc();
int    abs(), fclose(), fprintf(), free();
double fabs();
FILE   *fopen();

/*
 *              Common error messages
*/
static char
    VAR_PTR err_firmware = "Illegal firmware parameter",
    VAR_PTR err_null   = "Arg is NULL pointer",
    VAR_PTR lab_txtget = "t_txtget",
    VAR_PTR mt_str     = "";

/* ************************************************************************ */
/*                                                                          */
/*                         TERMINAL DRIVING ROUTINES                        */
/*                                                                          */
/* ************************************************************************ */


static clip_area VAR_PTR t_begin( device, options )
/*
 *     Start graphics .
 * Puts the screen in vector mode, sets up the terminal, initializes
 * variables, reads the stroke-character fonts and sets up the viewport
 * clip. The returned value is a pointer to the clip_area structure of the
 * viewport . The function fails ( i.e. returns a NULL pointer ) if cannot
 *   a) not read one of the characters from the font files
 *   b) move the firmware pen
 *   c) define the viewport clip_area
 * Appropriate error messages are produced in these cases .
*/
FILE * device;         /* device handle */
char VAR_PTR options;  /* unused */
  {
    char VAR_PTR name = "t_begin";
    coord_pair origin, upper_right;     /* viewport corners */
/*
 *      initial terminal status -- read only
*/
    static struct
      {
        ter_status requir; /* required terminal status */
        ter_status actual; /* actual terminal status   */
      }
    init_ter = { { GRAP_SCREEN, DOT_ON, 1, 1, 0,
                 { 0.0, HERC_Y - ALPH_HEIGHT + 1 }, { 0.0, 0.0 } },
                 { GRAP_SCREEN, DOT_ON, -1, 1, 0,
                 { 0.0, 0.0 }, { 0.0, 0.0 } } };
    CATCH_NULL( device, name, clip_area VAR_PTR );
/*
 *      Find out if ``device'' has been registered as a device .
 *      If it has, check that it is indeed a Hercules card and grab the
 *      pointer to it graphics status data .
*/
    if ( t_register( T_ISREGIST, device, 0, (char VAR_PTR) 0 ) )
      {
        if ( (int) t_register( T_REG_TYPE, device, NONEXISTENT_DEVICE,
                (char VAR_PTR) 0 ) != VISUAL500_TERMINAL )
            return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "Wrong driver. Device is not Hercules" );
        herc = (Gfx_Status VAR_PTR) t_register( T_REG_DATA,
            device, NONEXISTENT_DEVICE, (char VAR_PTR) 0 );
      }
    else
      {
/*
 *      ``device'' has never been registered, so allocate memory for the
 *      graphics status data and register the device as a Hercules card .
*/
        herc = (Gfx_Status VAR_PTR) malloc( (unsigned int) sizeof(Gfx_Status) );
        if ( ! herc )
            return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "Cannot allocate memory" );
        if ( ! t_register( T_REGISTER, device,
                 HERCULES_GRAPHICS, (char VAR_PTR) herc ) )
          {
            (void) free( (char *) herc );
            herc = (Gfx_Status VAR_PTR) 0;
            return (clip_area VAR_PTR) make_error( ERR_CALL, name, mt_str );
          }
        herc->handle = device;
/*
 *           reset character attributes and terminal status
*/
        (void) reset_font( (Text_Attribute VAR_PTR) &STROKE,
            HERC_Y / 20.0 );
        TER.requir = init_ter.requir;
/*
 *          initialize viewport corners
*/
        origin._x = origin._y = 0.0;
        upper_right._x = HERC_X;
        upper_right._y = HERC_Y;
        if ( ! ( VIEW_CLIP = def_2rclip( (coord_pair VAR_PTR) &origin,
                (coord_pair VAR_PTR) &upper_right,
                (coord_pair VAR_PTR) &origin ) ) )
          {
            (void) free( (char *) herc );
            herc = (Gfx_Status VAR_PTR) 0;
            return (clip_area VAR_PTR) make_error( ERR_CALL, name, mt_str );
          }
      }
/*
 *          put graphics on, Dots-on, solid line style
*/
    Hg_style( SOLID_LINE );
    Hg_dots( DOT_ON );
    if ( ! Hg_move( 0, 0 ) )
        return (clip_area VAR_PTR) make_error( ERR_WARN, name, err_firmware );
    Hg_on();
    TER.actual = init_ter.actual;
    return VIEW_CLIP;
  }


static boolean t_end()
/*
 *     End of graphics .
 * Delete character fonts and frees the memory allocated to them, clear the
 * graphics screen and turn graphics off . This function always succeeds .
*/
  {
    char VAR_PTR name = "t_end";
/*
 *      Deregister the device and free the memory used by the graphics status
 *      data .
*/
    START_UP( name, boolean );
    if ( ! t_register( T_DEREGIST, herc->handle,
            HERCULES_GRAPHICS, (char VAR_PTR) herc ) )
        (void) make_error( ERR_WARN, name, mt_str );
    (void) free( (char *) herc );
/*
 *      Delete character fonts and frees the memory allocated to them
*/
    if ( asc_font.font_got )
        (void) clear_font( (StrokeFont VAR_PTR) &asc_font );
    if ( sym_font.font_got )
        (void) clear_font( (StrokeFont VAR_PTR) &sym_font );
/*
 *      Clear the graphics screen and turn graphics off .
*/
    Hg_clear( GRAPHICS_PAGE );
    Hg_off();
    return TRUE;
  }


static boolean t_vwport( clip_ptr )
/*
 *      Set the clipping limits for the viewport . The old limits are replaced .
 * The `clip_ptr' must point to  clip_area which has been defined by
 * `def_2rclip()' . The function fails if
 *   a) clip_ptr is a NULL pointer
 *   b) clip_ptr points to something which is not a structure for a two
 *      dimensional rectangular clipping region .
*/
clip_area VAR_PTR clip_ptr;
  {
    char VAR_PTR name = "t_vwport";
    START_UP( name, boolean );
    CATCH_NULL( clip_ptr, name, boolean );
    if ( clip_ptr->id_struct == CLIP2R_ID )
        if ( -0.5 < min_clip( clip_ptr, _x ) &&
                    max_clip( clip_ptr, _x ) <= HERC_X &&
             -0.5 < min_clip( clip_ptr, _y ) &&
                    max_clip( clip_ptr, _y ) <= HERC_Y )
          {
            VIEW_CLIP = clip_ptr;
            return TRUE;
          }
    return make_error(ERR_ABORT, name, (char VAR_PTR) "Invalid pointer" );
  }


static boolean t_colour( ipen )
/*
 *    Selects specified colour ( or pen ) .
 * Because the Hercules screen is monochrome, colours are simulated with
 * various dashed line styles . The only true ``colours'' available are dark
 * and bright . The returned value is TRUE if  0 <= ipen <= 8 . If these limits
 * are exceeded, the value is wrapped around into the range 1 to 8 . The dashed
 * line equivalents of the colours are,
 *
 * IPEN   SCREEN
 *      LINE STYLE
 *  0   black  (background)
 *  1   solid  (green)
 *  2   dotted
 *  3   short dash
 *  4   long dash
 *  5   dot dash
 *  6   short-dash long-dash
 *  7   2-short-dashes long-dash
 *  8   3-dots long-dash
*/
int ipen;  /* pen number or colour */
  {
    START_UP( (char VAR_PTR) "t_colour", boolean );
    TER.requir.colour = ( abs( ipen ) - 1 ) % 8 + 1;
    return 0 <= ipen && ipen <= 8;
  }


static boolean t_dashed( style )
/*
 *      Set the dashed line style . The returned value is TRUE if
 * 0 <= style <= 8 . If these limits are exceeded, the value is wrapped
 * around into the range 1 to 8 .
 *
 * STYLE   SCREEN
 *       LINE STYLE
 *  0   black  - plot vector endpoints only (invisible)
 *  1   solid  (green)
 *  2   dotted
 *  3   short dash
 *  4   long dash
 *  5   dot dash
 *  6   short-dash long-dash
 *  7   2-short-dashes long-dash
 *  8   3-dots long-dash
*/
int style;  /* dashed line style */
  {
    START_UP( (char VAR_PTR) "t_dashed", boolean );
    TER.requir.dash = ( abs( style ) - 1 ) % 8 + 1;
    return 0 <= style && style <= 8;
  }


static boolean t_thick( width )
/*
 * Set the ``width'' of lines and text . The returned value is TRUE
 * if 0 <= thickness.
*/
int width;      /* line thickness */
  {
    START_UP( (char VAR_PTR) "t_thick", boolean );
    TER.requir.thick = width;
    return 0 <= width;
  }


/* ************************************************************************ */
/*                                                                          */
/*                         Turn graphics on or off                          */
/*                                                                          */
/* ************************************************************************ */


static boolean t_gr_off()
/*
 *      Turn graphics off .
 * I.e. switch to the alpha_numeric mode . This function always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_gr_off", boolean );
    if ( TER.actual.mode != ALPH_SCREEN )
      {
        Hg_off();
        TER.actual.mode = ALPH_SCREEN;
      }
    TER.requir.mode = ALPH_SCREEN;
    return TRUE;
  }


static boolean t_gr_on()
/*
 *      Turn graphics on .
 * I.e. switch to the graphics screen . This function always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_gr_on", boolean );
    if ( TER.actual.mode != GRAP_SCREEN )
      {
        Hg_on();
        TER.actual.mode = GRAP_SCREEN;
      }
    TER.requir.mode = GRAP_SCREEN;
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*          Change Data Level, Visible, Invisible, Revers, Replace          */
/*                                                                          */
/* ************************************************************************ */

static boolean t_visib()
/*
 *  Draw visible vectors .
 *  Sets screen into "dots on" mode . This function always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_visib", boolean );
    TER.requir.dots = DOT_ON;
    return TRUE;
  }


static boolean t_invis()
/*
 *   Draw invisible vectors .
 *  Sets screen into "dots off" mode . This function always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_invis", boolean );
    TER.requir.dots = DOT_OFF;
    return TRUE;
  }


static boolean t_revers()
/*
 *  Reverse the pixel dots when plotting vectors, plotting points or
 * alphagraphic characters . Sets screen into "compliment" mode . This function
 * always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_revers", boolean );
    TER.requir.dots = DOT_REV;
    return TRUE;
  }


static boolean t_replac()
/*
 *  Replace pixel contents with the new data . Sets screen into "replace"
 *  mode . This function always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_replac", boolean );
    TER.requir.dots = DOT_NEW;
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Miscellaneous stuff                            */
/*                                                                          */
/* ************************************************************************ */


static boolean t_newpag()
/*
 *  Clear the graphics screen . This function always succeeds .
*/
  {
    Hg_clear( GRAPHICS_PAGE );
    return TRUE;
  }


static boolean t_newscr()
/*
 *  Clear the alphanumeric screen . This function always succeeds .
*/
  {
    Hg_clear( TEXT_PAGE );
    return TRUE;
  }


static boolean t_bell()
/*
 *  Ring bell . This function always succeeds .
*/
  {
    (void) fputc( BEL, stdout );
    return TRUE;
  }


static boolean t_wait()
/*
 * Waits for the to press the carriage return . The returned value is TRUE .
*/
  {
    while( Hg_getc() != '\r' );
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                              Text Writing                                */
/*                                                                          */
/* ************************************************************************ */


static boolean t_txtmov( point )
/*
 *    Move the text cursor to coordinate `point' . The returned result is
 *  TRUE only if the next printable character is within the screen limits .
*/
register coord_pair *point;
  {
    char VAR_PTR name = "t_txtmov";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    TER.requir.cursor = *point;
    return -0.5 < point->_x  &&  point->_x <= HERC_X - ALPH_WIDTH &&
           -0.5 < point->_y  &&  point->_y <= HERC_Y - ALPH_HEIGHT;
  }


static boolean t_txtcur( row, column )
/*
 * Move the text cursor to text row `row' and text column `column' .
 * The text row is numbered from the top of the screen . The returned
 * result is TRUE only if the next printable character is within screen
 * limits . The row and column positions are calculated for the default
 * alphagraphics character size .
*/
int row, column;
  {
    char VAR_PTR name = "t_txtcur";
    START_UP( name, boolean );
    TER.requir.cursor._x = ALPH_WIDTH * column;
    TER.requir.cursor._y = HERC_Y - ALPH_HEIGHT + 1 - ALPH_HEIGHT * row;
    return 0 <= column  &&  column <= ALPH_COLUMNS - 1  &&
           0 <= row     &&  row    <= ALPH_LINES - 1;
  }

static boolean t_txtset( attribute, value )
/*
 *  Set the stroke character attribute `attribute' to `*value' .
 *  If `attribute' is negative, the current value of the attribute is returned
 *  to `*value' . This function fails if the `attribute' cannot be recognized or
 *  `value' is a null pointer . Appropriate error messages are generated .
*/
int attribute;
char VAR_PTR value;
  {
    char VAR_PTR name = "t_txtset";
    double VAR_PTR dvalue;  /* pointer to a double `value' */
    dvalue = (double VAR_PTR) value;
    if ( attribute != T_CH_RESET  && attribute != T_CH_BOLD  &&
         attribute != T_CH_NOBOLD && attribute != T_ASCII    &&
         attribute != T_SYMBOL )
        CATCH_NULL( value, name, boolean );
    START_UP( name, boolean );
    switch( attribute )
      {
        case  T_CH_RESET  :
            (void) reset_font( (Text_Attribute VAR_PTR) &STROKE, HERC_Y / 20.0 );
            break;
        case  T_CH_HEIGHT :
            if ( STROKE.height != *dvalue )
              {
                STROKE.height = LIMIT( *dvalue, 5.0, HERC_Y );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_HEIGHT :
            *dvalue = STROKE.height;
            break;
        case  T_CH_ASPECT :
            if ( STROKE.aspect != *dvalue )
              {
                STROKE.aspect = LIMIT( *dvalue, 0.2, 5.0 );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_ASPECT :
            *dvalue = STROKE.aspect;
            break;
        case  T_CH_SLANT :
            if ( STROKE.slant != *dvalue )
              {
                STROKE.slant = LIMIT( *dvalue, -80.0, 80.0 ) * PI / 180.0;
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_SLANT :
            *dvalue = STROKE.slant * 180.0 / PI;
            break;
        case  T_CH_ANGLE  :
            if ( STROKE.angle != *dvalue )
              {
                STROKE.angle = PI / 180.0 * ( *dvalue -
                    360.0 * (double) (int) ( *dvalue / 360.0 ) );
                STROKE.angle = STROKE.angle < -PI ? STROKE.angle + 2.0 * PI :
                    ( STROKE.angle > PI ? STROKE.angle-2.0*PI : STROKE.angle );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_ANGLE :
            *dvalue = STROKE.angle * 180.0 / PI;
            break;
        case  T_CH_GAP :
            STROKE.gap = LIMIT( *dvalue, 0.05, 5.0 );
            break;
        case -T_CH_GAP :
            *dvalue = STROKE.gap;
            break;
        case  T_CH_ALIGN :
            STROKE.align.vert  = TOLOWER( *value );
            value++;
            STROKE.align.horiz = TOLOWER( *value );
            if ( STROKE.align.horiz != 'l' && STROKE.align.horiz != 'c' &&
                    STROKE.align.horiz != 'r' )
                STROKE.align.horiz = 'l';
            if ( STROKE.align.vert != 'b' && STROKE.align.vert != 'c' &&
                    STROKE.align.vert != 't' )
                STROKE.align.vert = 'b';
            break;
        case -T_CH_ALIGN :
            *value++ = STROKE.align.vert;
            *value++ = STROKE.align.horiz;
            *value   = NUL;
            break;
        case T_ASCII :
            STROKE.font_ptr = (StrokeFont VAR_PTR) &asc_font;
            break;
        case T_SYMBOL :
            STROKE.font_ptr = (StrokeFont VAR_PTR) &sym_font;
            break;
        case T_CH_BOLD :
            STROKE.bold = TRUE;
            break;
        case T_CH_NOBOLD :
            STROKE.bold = FALSE;
            break;
        default :
            return make_error( ERR_ABORT, name,
                (char VAR_PTR) "Unknown stroke text attribute" );
      }
    return TRUE;
  }


static char VAR_PTR t_ch_def( what, character, index, point )
/*
 * This function allows the user to obtain the shape of any character defined
 * in the ASCII and SYMBOL fonts. With the exception of string alignment, the
 * transformations implied by t_txtset are applied to the returned coordinates.
 * This function is merely a call to the function ``define_font'' with an
 * additional argument defining which set of text attributes to use .
*/
int what;           /* specifies what to do (T_FT_* defined values) */
char character;     /* character to fetch from the font */
int index;          /* coordinate number   */
coord_pair
    VAR_PTR point;  /* returned coordinate */
  {
    return define_font( (Text_Attribute VAR_PTR) &STROKE, PEN_THICKNESS,
        what, character, index, point );
  }


static boolean t_symdrw( point, symbol )
/*
 * Draws a centred plotting symbol `symbol' at position `point' according
 * of the selected stroke character attributes . This function fails if
 * the font for the symbol is not defined or if there is a firmware error .
*/
coord_pair VAR_PTR point;
char       symbol;
  {
    coord_pair offset, lower_left;
    char VAR_PTR name = "t_symdrw";
    int font_index = symbol - 32;   /* index of character in font directory */
    StrokeChar VAR_PTR char_ptr;
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    TER.requir.pen = *point;
/*
 *          SET UP THE TERMINAL
 * set colour to black or white (dashed lines not allowed)
*/
    if ( ! TER.requir.colour )
      {
        if ( TER.actual.dots != DOT_OFF )
            Hg_dots( TER.actual.dots = DOT_OFF );
      }
    else if ( TER.requir.dots != TER.actual.dots )
            Hg_dots( TER.actual.dots = TER.requir.dots );
/*
 *          set dashed line style to solid
*/
    if ( TER.actual.dash != 1 )
      {
        Hg_style( SOLID_LINE );
        TER.actual.dash = 1;
      }
/*
 *          if necessary, read character fonts
*/
    if ( ! STROKE.font_ptr->font_got )
        if ( ! read_font( STROKE.font_ptr, file_font( STROKE.font_ptr), name ) )
            return FALSE;
/*
 *           CALCULATE LOCATION OF LOWER LEFT CORNER OF THE SYMBOL
 *           determine the character cell size vector
*/
    if ( STROKE.changed )
        (void) map_font( (Text_Attribute VAR_PTR) &STROKE, PEN_THICKNESS );
    char_ptr = STROKE.font_ptr->font_dir + font_index;
    offset._x = STROKE.font_ptr->font_wid * 0.5;
    offset._y = char_ptr->font_midpt;
/*
 *           determine start position
*/
    lower_left._x = point->_x -
        STROKE.map[0][0] * offset._x - STROKE.map[0][1] * offset._y;
    lower_left._y = point->_y -
        STROKE.map[1][0] * offset._x - STROKE.map[1][1] * offset._y;
/*
 *           DRAW THE SYMBOL
*/
    if ( 0 <= font_index && font_index < STROKE.font_ptr->font_siz )
        if ( char_ptr->font_char == symbol && char_ptr->font_npts > 0 )
          {
            boolean ok = bold_font( (Text_Attribute VAR_PTR) &STROKE,
               (coord_pair VAR_PTR) &lower_left,
               (char VAR_PTR) &font_index, t_stkchr );
            return ok & move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
          }
    return make_error( ERR_ABORT, name, (char VAR_PTR) "Unknown symbol" );
  }


static int t_txtdrw( string )
/*
 *   Draws the text contained in the character string `string' at the
 * current cursor location according to the selected stroke character
 * attributes . The returned value is the number of characters drawn . If
 * an error occurs, the returned value is 0 . An error may be generated
 * because
 *   a) a character in `string' is unavailable in the specified font
 *   b) the character is not printable
 *   c) could not update the pen position after drawing the characters
 *   d) there is a firmware error
*/
char VAR_PTR string;
  {
    boolean ok = TRUE;    /* success flag */
    int font_index,       /* index of the character in font directory */
        n_chars,          /* length of `string' */
        char_count=0;     /* counts each character as it is drawn */
    char VAR_PTR name = "t_txtdrw";
    coord_pair cell_size, /* size of the character cell */
               offset,    /* offset of location from `point' */
               point;     /* location of char cell lower left corner  */
    CATCH_NULL( string, name, int );
    START_UP( name, int );
/*
 *          SET UP THE TERMINAL
 *
 *          set colour to black or white (dashed lines not allowed)
*/
    if ( ! TER.requir.colour )
      {
        if ( TER.actual.dots != DOT_OFF )
            Hg_dots( TER.actual.dots = DOT_OFF );
      }
    else if ( TER.requir.dots != TER.actual.dots )
            Hg_dots( TER.actual.dots = TER.requir.dots );
/*
 *          set dashed line style to solid
*/
    if ( TER.actual.dash != 1 )
      {
        Hg_style( SOLID_LINE );
        TER.actual.dash = 1;
      }
/*
 *          if necessary, read character fonts
*/
    if ( ! STROKE.font_ptr->font_got )
        if ( ! read_font( STROKE.font_ptr, file_font( STROKE.font_ptr), name ) )
            return FALSE;
    if ( STROKE.changed )
        (void) map_font( (Text_Attribute VAR_PTR) &STROKE, PEN_THICKNESS );
/*
 *          CALCULATE LOCATION OF LOWER LEFT CORNER OF THE STRING
 *          determine the character cell size vector
*/
    n_chars = t_strlen( string );
    cell_size._x = STROKE.font_ptr->font_wid *
        ( 1.0 + STROKE.gap / STROKE.aspect );
    cell_size._y = STROKE.font_ptr->font_hei;
    switch( STROKE.align.horiz )
      {
        case 'r' :  offset._x = cell_size._x * n_chars; break;
        case 'c' :  offset._x = cell_size._x * n_chars / 2.0; break;
        case 'l' :  offset._x = 0.0; break;
      }
    switch( STROKE.align.vert )
      {
        case 't' :  offset._y = cell_size._y; break;
        case 'c' :  offset._y = cell_size._y / 2.0; break;
        case 'b' :  offset._y = 0.0; break;
      }
/*
 *          determine start position
*/
    point._x = TER.requir.cursor._x - STROKE.map[0][0] * offset._x -
               STROKE.map[0][1] * offset._y;
    point._y = TER.requir.cursor._y - STROKE.map[1][0] * offset._x -
               STROKE.map[1][1] * offset._y;
    cell_size._y = STROKE.map[1][0] * cell_size._x;
    cell_size._x = STROKE.map[0][0] * cell_size._x;
/*
 *          DRAW EACH CHARACTER IN THE STRING
*/
    for( ; *string; string++, char_count++ )
      {
        font_index = *string - 32;
        if ( 0 <= font_index && font_index < STROKE.font_ptr->font_siz )
            if ( (STROKE.font_ptr->font_dir)[font_index].font_char == *string &&
                 (STROKE.font_ptr->font_dir)[font_index].font_npts > 0 )
              {
                ok &= bold_font( (Text_Attribute VAR_PTR) &STROKE,
                                 (coord_pair VAR_PTR) &point,
                                 (char VAR_PTR) &font_index, t_stkchr );
                point._x += cell_size._x;
                point._y += cell_size._y;
              }
            else
                (void) make_error( ERR_WARN, name,
                    (char VAR_PTR) "character unavailable" );
         else
             (void) make_error( ERR_WARN, name,
                 (char VAR_PTR) "not printable character" );
      }
    TER.requir.cursor = point;
    ok &= move_clip( (coord_triple VAR_PTR) &TER.requir.pen, VIEW_CLIP );
    if ( ! ok )
        (void) make_error( ERR_CALL, name, mt_str );
    return char_count;
  }


#define DEL_CURSOR  -1
#define PUT_CURSOR  0
#define IN_VIEW      ( TER.requir.cursor._x >= min_clip( VIEW_CLIP, _x ) && \
      TER.requir.cursor._x <= max_cursor._x && \
      TER.requir.cursor._y >= min_clip( VIEW_CLIP, _y ) && \
      TER.requir.cursor._y <= max_cursor._y )
static coord_pair max_cursor;

static char VAR_PTR t_txtget( string, max_chars )
/*
 *    Reads a maximum of `max_chars' characters from `stdin' into `string'
 * until a newline character is read . The newline character is not stored in
 * `string' . `string' is terminated with NUL . If using the alpha-numeric
 * screen, this function simply gets() . If using the graphics screen this
 * function is again like gets() except that the echoed character is confined
 * to the viewport . If the cursor is initially outside the viewport,
 * t_txtget() starts echoing characters at the top left corner of the viewport.
 * T_txtget will not echo any more characters if it runs out of space at the
 * bottom of the viewport .If an error errors or no characters are read,
 * a NULL pointer is returned, otherwise `string' is returned .
*/
char VAR_PTR string;
int max_chars;
  {
    char VAR_PTR ptr;
    CATCH_NULL( string, lab_txtget, char VAR_PTR );
    START_UP( lab_txtget, char VAR_PTR );
    if ( TER.actual.mode == GRAP_SCREEN )
      {
        boolean in_viewport;
        max_cursor._x = max_clip(VIEW_CLIP,_x) - ALPH_WIDTH + 1;
        max_cursor._y = max_clip(VIEW_CLIP,_y) - ALPH_HEIGHT + 1;
/*
 *              ensure cursor is in the viewport
*/
        if ( TER.requir.cursor._x < min_clip(VIEW_CLIP,_x) ||
             TER.requir.cursor._x > max_cursor._x ||
             TER.requir.cursor._y < min_clip(VIEW_CLIP,_y) ||
             TER.requir.cursor._y > max_cursor._y )
          {
            TER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
            TER.requir.cursor._y = max_cursor._y;
          }
        if ( in_viewport = IN_VIEW )
            put_char( PUT_CURSOR );
/*
 *              loop for getting characters
*/
        for( ptr = string; ( *ptr = Hg_getc() ) != '\r' && max_chars--; )
          {
/*              test for DEL key
*/
            if ( *ptr == BS )
              {
                if ( string != ptr )
                  {
                    if ( in_viewport )
                        put_char( DEL_CURSOR );
                    if ( in_viewport = prev_cursor() )
                        put_char( PUT_CURSOR );
                    ptr--;
                  }
              }
/*              test for printable character
*/
            else if ( ' ' <= *ptr && *ptr <= '~' )
              {
                if ( in_viewport )
                  {
                    put_char( DEL_CURSOR );
                    put_char( *ptr );
                  }
                if ( in_viewport = next_cursor() )
                    put_char( PUT_CURSOR );
                ptr++;
              }
            else
                ptr++;
          }
        if ( in_viewport )
            put_char( DEL_CURSOR );
        TER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
        TER.requir.cursor._y -= ALPH_HEIGHT;
      }
    else
/*
 *              alpha-numeric screen
*/
        for( ptr = string; ( *ptr = getchar() ) != '\n' && max_chars--; ptr++ );
    *ptr = NUL;
    return ptr != string ? string : (char VAR_PTR) 0;
  }


static int t_txtput( string )
/*
 *    Put a text string `string' on the GRAPHICS screen, starting at the place
 *  last specified by `t_txtmov' or 't_txtcur', using the colour specified by
 *  `t_colour' ( which can only be bright or dark ) and the data level
 *  specified by `t_visib', `t_invis', `t_revers' or `t_replac' . Only those
 *  characters within the viewport are written to the screen . The returned
 *  value is the number of characters written to the screen, or zero if an
 *  error has occurred .
 *    The character string `string' may contain printable characters and the
 *  control codes,
 *    HT : \t : \011 : move to the right by one character space
 *    BS : \b : \010 : move to the left by one character space
 *    CR : \r : \015 : move to the beginning of the current line
 *    LF : \n : \012 : move down one character space
 *    VT :    : \013 : move up one character space
 *    EM :    : \031 : move to top left hand side of the screen
 *  Any other characters are ignored .
 *   DOT_ON,  write foreground colour text onto the existing screen,
 *   DOT_OFF, write a dark background onto the existing screen,
 *   DOT_REV, change the background into foreground colour, write dark text,
 *   DOT_NEW, foreground colour text is written onto a dark background.
*/
char VAR_PTR string;
  {
    int char_count = 0; /* number of characters sent */
    boolean ok = TRUE;  /* success flag */
    boolean txt_space = FALSE;
    char VAR_PTR name = "t_txtput";
    CATCH_NULL( string, name, int );
    START_UP( name, int );
    TER.actual.dots = TER.requir.dots;
    if ( ! TER.requir.colour )
      {
        if ( TER.requir.dots == DOT_ON )
            TER.actual.dots = DOT_OFF;
        else if ( TER.requir.dots == DOT_REV )
          {
            TER.actual.dots = DOT_NEW;
            txt_space = TRUE;
          }
        else if ( TER.requir.dots == DOT_NEW )
            txt_space = TRUE;
      }
    Hg_dots( TER.actual.dots );
    max_cursor._x = max_clip(VIEW_CLIP,_x) - ALPH_WIDTH + 1;
    max_cursor._y = max_clip(VIEW_CLIP,_y) - ALPH_HEIGHT + 1;
/*
 *        start the real work here
*/
    for( ; *string && ok; string++ )
      {
        if ( ' ' <= *string && *string <= '~' )
          {
/*
 *******     put character on the screen if it is in the viewport
*/
            if ( min_clip( VIEW_CLIP, _x ) <= TER.requir.cursor._x &&
                 min_clip( VIEW_CLIP, _y ) <= TER.requir.cursor._y &&
                 max_cursor._x >= TER.requir.cursor._x &&
                 max_cursor._y >= TER.requir.cursor._y )
              {
/*
 *       place the curor at the correct position
*/
                if ( fabs( TER.requir.cursor._x - TER.actual.cursor._x ) > 0.5
                  || fabs( TER.requir.cursor._y - TER.actual.cursor._y ) > 0.5 )
                  {
                    if ( ok &= Hg_move( (int) ( TER.requir.cursor._x + 0.5 ),
                                        (int) ( TER.requir.cursor._y + 0.5 ) ) )
                        TER.actual.cursor = TER.requir.cursor;
                    else
                       (void) make_error( ERR_WARN, name, err_firmware );
                  }
/*
 *        write the character
*/
                Hg_putc( txt_space ? ' ' : *string );
                char_count++;
                TER.actual.cursor._x +=ALPH_WIDTH;
              }
            TER.requir.cursor._x += ALPH_WIDTH;
          }
/*
 *******      book keeping for control codes
*/
        else
          {
            switch( *string )
              {
                case LF: TER.requir.cursor._y -= ALPH_HEIGHT;
                         break;
                case VT: TER.requir.cursor._y += ALPH_HEIGHT;
                         break;
                case EM: TER.requir.cursor._y = max_cursor._y;
                         TER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
                         break;
                case CR: TER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
                         break;
                case BS: TER.requir.cursor._x -= ALPH_WIDTH;
                         break;
                case HT: TER.requir.cursor._x += ALPH_WIDTH;
               }
          }
      }
    return ok ? char_count : 0;
  }


/* ************************************************************************ */
/*                                                                          */
/*                         Cross Hairs Routines                             */
/*                                                                          */
/* ************************************************************************ */


static boolean t_xhrget( point, char_pressed )
/*
 *      Get cross-hair location . The cross-hairs are positioned by the user,
 *  and the location of the cross-hairs is transmitted when an ASCII character
 *  key is pressed . The returned value of `t_xhrget' is TRUE if `point' is
 *  within the current viewport ; otherwise the returned value is FALSE . If
 *  necessary, t_xhrget switches to the graphics screen .
*/
coord_pair VAR_PTR point;   /* pointer to the coordinates of the cross-hairs */
char VAR_PTR char_pressed;  /* character pressed to identify `point' */
 {
    int x, y;
    char VAR_PTR name = "t_xhrget";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    if ( TER.actual.mode != GRAP_SCREEN )
      {
        Hg_on();
        TER.actual.mode = GRAP_SCREEN;
      }
    *char_pressed = Hg_getx( (int VAR_PTR) &x, (int VAR_PTR) &y );
    point->_x = x;
    point->_y = y;
    return min_clip( VIEW_CLIP, _x ) <= point->_x &&
           max_clip( VIEW_CLIP, _x ) >= point->_x &&
           min_clip( VIEW_CLIP, _y ) <= point->_y &&
           max_clip( VIEW_CLIP, _y ) >= point->_y;
 }


static boolean t_xhrput( point )
/*
 *      Put the  cross-hairs at the coordinate `point' . The returned value
 *  of `t_xhrput' is TRUE if `point' is within the current viewport and
 *  `t_xhrput' can write to the graphics device ; otherwise the returned
 *  value is FALSE .
*/
coord_pair VAR_PTR point; /* pointer to the coordinates of the cross-hairs */
 {
    char VAR_PTR name = "t_xhrput";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    if ( -0.5 < point->_x  &&  point->_x <= HERC_X + 0.5 &&
         -0.5 < point->_y  &&  point->_y <= HERC_Y + 0.5 )
      {
/*
 *        send `point', load cross-hair & put into cross-hair mode
*/
        boolean ok; /* success flag */
        if ( TER.actual.mode != GRAP_SCREEN )
          {
            Hg_on();
            TER.actual.mode = GRAP_SCREEN;
          }
        ok = Hg_putx( (int) ( point->_x + 0.5 ), (int) ( point->_y + 0.5 ) );
    return ok && min_clip( VIEW_CLIP, _x ) <= point->_x &&
                 max_clip( VIEW_CLIP, _x ) >= point->_x &&
                 min_clip( VIEW_CLIP, _y ) <= point->_y &&
                 max_clip( VIEW_CLIP, _y ) >= point->_y;
      }
    return FALSE;
 }

/* ************************************************************************ */
/*                                                                          */
/*                        Print terminal status                             */
/*                                                                          */
/* ************************************************************************ */


static char VAR_PTR t_status( attribute, value )
/*    Put status attribute `attribute' into the coordinate pair `*value' .
*/
int attribute;
char VAR_PTR value;
  {
    char VAR_PTR name = "t_status";
    CATCH_NULL( value, name, char VAR_PTR );
    if ( attribute == T_ST_R_PEN  || attribute == T_ST_A_PEN  ||
         attribute == T_ST_R_CURS || attribute == T_ST_A_CURS ||
         attribute == T_ST_CLIP )
        START_UP( name, char VAR_PTR );
    switch( attribute )
      {
        case T_ST_DEVICE:
            (void) t_strcpy( value, "Hercules Graphics" );
            break;
        case T_ST_SCREEN:
            *( (boolean VAR_PTR) value ) = TRUE;
            break;
        case T_ST_CATEG:
            *( (int VAR_PTR) value ) = T_ST_HARD;
            break;
        case T_ST_MONO:
            *( (boolean VAR_PTR ) value ) = TRUE;
            break;
        case T_ST_CLIP:
            *( (clip_area VAR_PTR VAR_PTR) value ) = VIEW_CLIP;
            break;
        case T_ST_MAX:
            ((coord_pair VAR_PTR) value)->_x = HERC_X;
            ((coord_pair VAR_PTR) value)->_y = HERC_Y;
            break;
        case T_ST_MIN:
            ((coord_pair VAR_PTR) value)->_x =
                ((coord_pair VAR_PTR) value)->_y = 0.0;
            break;
        case T_ST_CHAR:
            ((coord_pair VAR_PTR) value)->_x = ALPH_WIDTH;
            ((coord_pair VAR_PTR) value)->_y = ALPH_HEIGHT;
            break;
        case T_ST_LINES:
            ((coord_pair VAR_PTR) value)->_x = ALPH_COLUMNS;
            ((coord_pair VAR_PTR) value)->_y = ALPH_LINES;
            break;
        case T_ST_PIXSIZ:
            ((coord_pair VAR_PTR) value)->_x = X_PIXEL_SIZE;
            ((coord_pair VAR_PTR) value)->_y = Y_PIXEL_SIZE;
            break;
        case T_ST_A_CURS:
            ((coord_pair VAR_PTR) value)->_x = TER.actual.cursor._x;
            ((coord_pair VAR_PTR) value)->_y = TER.actual.cursor._y;
            break;
        case T_ST_R_CURS:
            ((coord_pair VAR_PTR) value)->_x = TER.requir.cursor._x;
            ((coord_pair VAR_PTR) value)->_y = TER.requir.cursor._y;
            break;
        case T_ST_A_PEN:
            ((coord_pair VAR_PTR) value)->_x = TER.actual.pen._x;
            ((coord_pair VAR_PTR) value)->_y = TER.actual.pen._y;
            break;
        case T_ST_R_PEN:
            ((coord_pair VAR_PTR) value)->_x = TER.requir.pen._x;
            ((coord_pair VAR_PTR) value)->_y = TER.requir.pen._y;
            break;
        default:
            return (char VAR_PTR) make_error( ERR_ABORT, name,
               (char VAR_PTR) "unknown attribute" );
      }
    return value;
  }



static boolean t_pr_ter( out_file )
FILE * out_file;
  {
    char VAR_PTR name = "t_pr_ter";
    CATCH_NULL( out_file, name, boolean );
    START_UP( name, boolean );
    (void) fprintf( out_file, " TERMINAL STATUS\n" );
    (void) fprintf( out_file,
            "               ---requir---     ---actual---\n" );
    (void) fprintf( out_file, "data level  %15s  %15s\n",
            pr_dots( (ter_status VAR_PTR) &TER.requir ),
            pr_dots( (ter_status VAR_PTR) &TER.actual ) );
    (void) fprintf( out_file, "mode        %15s  %15s\n",
            pr_mode( (ter_status VAR_PTR) &TER.requir ),
            pr_mode( (ter_status VAR_PTR) &TER.actual ) );
    (void) fprintf( out_file, "colour      %15d  %15d\n",
            TER.requir.colour, TER.actual.colour );
    (void) fprintf( out_file, "dash style  %15d  %15d\n",
            TER.requir.dash, TER.actual.dash );
    (void) fprintf( out_file,
            "cursor      (%6.1f,%6.1f)  (%6.1f,%6.1f)\n",
            TER.requir.cursor._x, TER.requir.cursor._y,
            TER.actual.cursor._x, TER.actual.cursor._y );
    (void) fprintf( out_file,
            "pen         (%6.1f,%6.1f)  (%6.1f,%6.1f)\n",
            TER.requir.pen._x, TER.requir.pen._y,
            TER.actual.pen._x, TER.actual.pen._y );
    if ( ferror( out_file ) )
        return make_error( ERR_WARN, name,
            (char VAR_PTR) "Cannot write terminal status to file" );
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                             Vector Drawing                               */
/*                                                                          */
/* ************************************************************************ */


static boolean t_vecmov( point )
/*
 * Move the line drawing pen to `point'
*/
coord_pair VAR_PTR point;
  {
    char VAR_PTR name = "t_vecmov";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    TER.requir.pen = *point;
    return move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
  }


static boolean t_vecdrw( point )
/*
 *      Draw from the previous pen position to `point' according to the
 *  required pen colour, dashed line style and dots-level . Dashed line style
 *  takes precedence over colours . Lines are clipped to the viewport
 *  boundaries .
*/
coord_pair VAR_PTR point;
  {
    boolean ok = TRUE; /* success flag */
    char VAR_PTR name = "t_vecdrw";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    TER.requir.pen = *point;
/*
 *      If required dashed line style is 0, plot the point if it is in the
 *      viewport .
*/
    if ( ! TER.requir.dash )
      {
        ok &= move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
        if ( min_clip( VIEW_CLIP, _x ) <= point->_x &&
             max_clip( VIEW_CLIP, _x ) >= point->_x &&
             min_clip( VIEW_CLIP, _y ) <= point->_y &&
             max_clip( VIEW_CLIP, _y ) >= point->_y )
          {
/*
 *      If required colour is black, put terminal data level into DOT_OFF,
 *      otherwise use the required data level .
*/
            if ( ! TER.requir.colour )
              {
                if ( TER.actual.dots != DOT_OFF )
                    Hg_dots( TER.actual.dots = DOT_OFF );
              }
            else if ( TER.actual.dots != TER.requir.dots )
                Hg_dots( TER.actual.dots = TER.requir.dots );
/*
 *      plot the point
*/
            TER.actual.cursor._x = TER.actual.pen._x = point->_x;
            TER.actual.cursor._y = TER.actual.pen._y = point->_y;
            Hg_point( (int) ( point->_x + 0.5 ), (int) ( point->_y + 0.5 ) );
          }
      }
/*
 *      Need to draw a line . First clip the endpoints .
*/
    else if ( clip( (coord_triple VAR_PTR) point, VIEW_CLIP ) )
      {
/*
 *      If required colour is black, put terminal data level into DOT_OFF,
 *          otherwise use the required data level .
*/
        if ( ! TER.requir.colour )
          {
            if ( TER.actual.dots != DOT_OFF )
                Hg_dots( TER.actual.dots = DOT_OFF );
          }
        else if ( TER.actual.dots != TER.requir.dots )
            Hg_dots( TER.actual.dots = TER.requir.dots );
/*
 *      If a continuous line is required, use the dashed line style as
 *      pseudo-colour, otherwise just set the required dashed line style
*/
        if ( TER.requir.dash == 1 && TER.requir.colour )
          {
            if ( TER.requir.colour != TER.actual.dash )
              {
                Hg_style( style_mask[TER.requir.colour] );
                TER.actual.dash = TER.requir.colour;
              }
          }
        else if ( TER.actual.dash != TER.requir.dash )
          {
            Hg_style( style_mask[TER.requir.dash] );
            TER.actual.dash = TER.requir.dash;
          }
/*
 *      draw the vector
*/
        ok = Hg_move( (int) ( first_clip( VIEW_CLIP, _x ) + 0.5 ),
                      (int) ( first_clip( VIEW_CLIP, _y ) + 0.5 ) );
        TER.actual.cursor._x = TER.actual.pen._x = last_clip( VIEW_CLIP, _x );
        TER.actual.cursor._y = TER.actual.pen._y = last_clip( VIEW_CLIP, _y );
        ok &= Hg_draw( (int) ( last_clip( VIEW_CLIP, _x ) + 0.5 ),
                       (int) ( last_clip( VIEW_CLIP, _y ) + 0.5 ) );
      }
    if ( ! ok )
        (void) make_error( ERR_WARN, name, err_firmware );
    return ok;
  }


static boolean t_boxfil( lower_left, upper_right, style )
/*
 *      Draw a rectangular box on the screen with lower left corner at
 * `*lower_left' and upper right corner at `*upper_right', and fill the box
 * with a hatching of type `style' . The hatching styles are :
 *
 *  0 = empty, box outline only  5 = horizontal lines
 *  1 = fill the box completely  6 = vertical lines
 *  2 = 50% `grey' fill          7 = diagonal cross hatch
 *  3 = NW-SE diagonal lines     8 = horizontal-vertical cross
 *  4 = NE-SW diagonal lines         hatch
 * The box outline is clipped to the viewport limits .
*/
coord_pair VAR_PTR lower_left,
           VAR_PTR upper_right;
int style;
  {
    coord_pair ll_corner, ur_corner;
    char VAR_PTR name = "t_boxfil";
    CATCH_NULL( lower_left && upper_right, name, boolean );
    START_UP( name, boolean );
    ll_corner = *lower_left;
    ur_corner = *upper_right;
/*
 *      ensure correct ordering of coordinates
*/
    if ( ll_corner._x > ur_corner._x )
        SWAPD( ll_corner._x , ur_corner._x );
    if ( ll_corner._y > ur_corner._y )
        SWAPD( ll_corner._y , ur_corner._y );
/*
 *      clip the rectangle to the screen limits
*/
    ll_corner._x = LIMIT( ll_corner._x,
        min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
    ll_corner._y = LIMIT( ll_corner._y,
        min_clip( VIEW_CLIP, _y ), max_clip( VIEW_CLIP, _y ) );
    ur_corner._x = LIMIT( ur_corner._x,
        min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
    ur_corner._y = LIMIT( ur_corner._y,
        min_clip( VIEW_CLIP, _y ), max_clip( VIEW_CLIP, _y ) );
/*
 *      draw the box
*/
    if ( ur_corner._x > ll_corner._x && ur_corner._y > ll_corner._y )
      {
        TER.actual.dots = TER.requir.dots;
        if ( ! TER.requir.colour &&
                ( TER.requir.dots == DOT_ON || TER.requir.dots == DOT_NEW ) )
          {
            TER.actual.dots = DOT_OFF;
            if ( TER.requir.dots == DOT_NEW )
                style = 1;
          }
        Hg_dots( TER.actual.dots );
        if ( Hg_box( (int) ( ll_corner._x + 0.5 ), (int) ( ll_corner._y + 0.5),
                     (int) ( ur_corner._x + 0.5 ), (int) ( ur_corner._y + 0.5),
                     ( abs( style ) - 1 ) % 8 + 1 ) )
            return TRUE;
        (void) make_error( ERR_WARN, name, err_firmware );
      }
    return FALSE;
  }


static boolean t_dump( file_name )
/*
 *      Opens a printable binary file with name `file_name' and dumps the
 * contents of the graphics screen into it . If t_dump cannot open or write
 * to file the file, an error message is generated and the returned value is
 * FALSE . The file can be printed by the system command -
 *       COPY/B file_name LPT1:
*/
char VAR_PTR file_name;  /* name of dump file */
  {
    char print_bytes[8], /* buffer for an 8 x 8 block of pixels */
         VAR_PTR name = "t_dump";
    int row, column;     /* character row and column numbrs */
    boolean ok = FALSE;  /* success flag */
    FILE *dump_ptr;      /* pointer to the printable dump file */

    CATCH_NULL( file_name, name, boolean );
    if ( dump_ptr = fopen( (char *) file_name, "wb" ) )
      {
        unsigned int dashes = 0x8080;
        ok = 3 == fwrite( "\033A\010", 1, 3, dump_ptr );  /* line spacing    */
/*             ESC, 'A', BS */
        for( row = 0; ok && row <= ALPH_LINES; row++ )
          {
            int screen_width = HERC_X + 1;
            ok &= 3 == fwrite( "\033*\006", 1, 3, dump_ptr );  /* start line */
/*                 ESC, '*', ACK                               no. of pixels */
            ok &= 1 == fwrite( (char *) &screen_width, 2, 1, dump_ptr );
            for( column = 0; ok && column <= ALPH_COLUMNS - 1; column++ )
              {
                Hg_dump( row, column, print_bytes );
                ok &= 8 == fwrite( print_bytes, 1, 8, dump_ptr );
             }
            ok &= 1 == fwrite( "\n", 1, 1, dump_ptr );/*end of line-line feed*/
          }
        if ( ok )
            ok &= 8 == fwrite( "\033J\004\033*\004\000\004", 1, 8, dump_ptr );
/*             ESC, 'J', 4, ESC, '*', 4, 0, 4 */
        for( column = 0; ok && column < 512; column++ )
            ok &= 1 == fwrite( (char *) &dashes, 2, 1, dump_ptr );
        if ( ok )
            ok &= 6 == fwrite( "\012\033\002\007\007\032", 1, 6, dump_ptr );
/*             LF, ESC, 2, BEL, BEL, EOF */
        (void) fclose( dump_ptr );
        if ( ! ok )
            (void) make_error( ERR_ABORT, name,
                (char VAR_PTR) "Cannot write to file" );
      }
    else
        (void) make_error( ERR_ABORT, name,
            (char VAR_PTR) "Cannot open file" );
    return ok;
  }


/* ************************************************************************ */
/* ************************************************************************ */
/*                                                                          */
/*                    Array of pointers to driver functions                 */
/*                                                                          */
/* With the exception of the function ``t_driver'' which is used to select  */
/* the required driver, this array is the ONLY point of access to the       */
/* functions in the driver .                                                */
/*                                                                          */
/* ************************************************************************ */
/* ************************************************************************ */

T_dev_fns Herc_dev_ptrs = {
            t_begin,  t_end,    t_vwport, t_gr_on,  t_gr_off,
            t_colour, t_dashed, t_thick,  t_visib,  t_invis,
            t_revers, t_replac, t_newpag, t_newscr, t_bell,
            t_wait,   t_txtmov, t_txtcur, t_txtset, t_symdrw,
            t_txtdrw, t_txtget, t_txtput, t_status, t_pr_ter,
            t_vecmov, t_vecdrw, t_dump,   t_xhrput, t_xhrget,
            t_boxfil, t_ch_def };


/* *********************************************************************** */
/*                                                                         */
/*                     driver utilities for herc_g.c                       */
/*      these functions cannot be called directly by the user              */
/*                                                                         */
/* *********************************************************************** */

static boolean t_stkchr( index, point )
/*
 *      Draws character identified by `index', at location `point' .
 *      This is a single pass -- i.e. no boldface.
*/
char VAR_PTR index;
coord_pair VAR_PTR point;
  {
    int point_count;   /* counts the points plotted for each char */
    boolean ok = TRUE, /* success flag     */
            pen_up;    /* move with pen up */
    float VAR_PTR coord_ptr;   /* pointer to font coordinate values */
    coord_pair coord;  /* coordinate from the font - to plot */
    double temp;
    StrokeChar VAR_PTR char_ptr =
        STROKE.font_ptr->font_dir + *( (int VAR_PTR) index );
/*
 *      step through the coordinates which define the character
*/
    coord_ptr = char_ptr->font_coord;
    for( point_count = 0;
         ok && point_count < char_ptr->font_npts;
         point_count++ )
      {
/*
 *      calculate stroke coordinate position in pixel units
*/
        pen_up = ! point_count;
        coord._x = *coord_ptr++;
        coord._y = *coord_ptr++;
        if ( coord._x < 0.0 )
          {
            pen_up = TRUE;
            coord._x = -coord._x;
          }
        if ( coord._y < 0.0 )
          {
            pen_up = TRUE;
            coord._y = -coord._y;
          }
        coord._y -= char_ptr->font_shift;
        temp = STROKE.map[0][0] * coord._x + STROKE.map[0][1] * coord._y
                + point->_x;
        coord._y = STROKE.map[1][0] * coord._x + STROKE.map[1][1] * coord._y
                + point->_y;
        coord._x = temp;
/*
 *      plot the point
*/
        if ( pen_up )
            ok = move_clip( (coord_triple VAR_PTR) &coord, VIEW_CLIP );
        else if ( clip( (coord_triple VAR_PTR) &coord, VIEW_CLIP ) )
          {
/*
 *      draw vector on the screen
*/
            ok &= Hg_move( (int) ( first_clip( VIEW_CLIP, _x ) + 0.5 ),
                           (int) ( first_clip( VIEW_CLIP, _y ) + 0.5 ) );
            TER.actual.cursor._x = TER.actual.pen._x
                                 = last_clip( VIEW_CLIP, _x );
            TER.actual.cursor._y = TER.actual.pen._y
                                 = last_clip( VIEW_CLIP, _y );
            ok &= Hg_draw( (int) ( last_clip( VIEW_CLIP, _x ) + 0.5 ),
                           (int) ( last_clip( VIEW_CLIP, _y ) + 0.5 ) );
          }
      }
    if ( ! ok )
      (void) make_error( ERR_WARN, (char VAR_PTR) "t_stkchr", err_firmware );
    return ok;
  }


static void put_char( character )
/*
 *      Place a tty character on the screen at the required cursor location.
*/
char character;
  {
/*
 *      move to required cursor position
*/
    if ( Hg_move( (int) ( TER.requir.cursor._x + 0.5 ),
                  (int) ( TER.requir.cursor._y + 0.5 ) ) )
      {
/*
 *      write printable character
*/
        if ( ' ' <= character && character <= '~' )
          {
            TER.actual.dots = TER.requir.dots;
            if ( ! TER.requir.colour )
              {
                if ( TER.requir.dots == DOT_ON )
                    TER.actual.dots = DOT_OFF;
                else if ( TER.requir.dots == DOT_REV )
                  {
                    TER.actual.dots = DOT_NEW;
                    character = ' ';
                  }
                else if ( TER.requir.dots == DOT_NEW )
                    character = ' ';
              }
            Hg_dots( TER.actual.dots );
            if ( ! Hg_putc( character ) )
                make_error( ERR_WARN, lab_txtget, err_firmware );
          }
/*
 *      write or delete the cursor
*/
        else if ( character == PUT_CURSOR || character == DEL_CURSOR )
          {
            int ll_x = (int) ( TER.requir.cursor._x + 0.5 ),
                ll_y = (int) ( TER.requir.cursor._y + 0.5 );
            Hg_dots( TER.actual.dots = DOT_REV );
            Hg_box( ll_x, ll_y, ll_x + (int) ALPH_WIDTH - 1,
                                ll_y + (int) ALPH_HEIGHT - 1, 1 );
          }
        TER.actual.cursor = TER.requir.cursor;
        TER.actual.cursor._x += ALPH_WIDTH;
        TER.actual.pen = TER.actual.cursor;
      }
    else
        (void) make_error( ERR_WARN, lab_txtget, err_firmware );
  }


static boolean next_cursor()
/*
 *      advance the required location of the cursor
*/
  {
    if ( ( TER.requir.cursor._x += ALPH_WIDTH ) > max_cursor._x )
      {
        TER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
        TER.requir.cursor._y -= ALPH_HEIGHT;
      }
    return IN_VIEW;
  }


static boolean prev_cursor()
/*
 *      move the required location of the cursor back one step
*/
  {
    if ( ( TER.requir.cursor._x -= ALPH_WIDTH ) < min_clip( VIEW_CLIP, _x ) )
      {
        TER.requir.cursor._x =
            ( (int) ( ( max_clip( VIEW_CLIP, _x ) - min_clip( VIEW_CLIP, _x ) )
            / ALPH_WIDTH ) - 1 ) * ALPH_WIDTH + min_clip( VIEW_CLIP, _x );
        TER.requir.cursor._y += ALPH_HEIGHT;
      }
    return IN_VIEW;
  }


static char VAR_PTR pr_dots( status_ptr )
/*
 *      returns verbal interpretation of the data level
*/
ter_status * status_ptr;
  {
    switch( status_ptr->dots )
      {
        case DOT_ON  : return (char VAR_PTR) "dots on";
        case DOT_OFF : return (char VAR_PTR) "dots off";
        case DOT_REV : return (char VAR_PTR) "complement dots";
        case DOT_NEW : return (char VAR_PTR) "replace dots";
      }
    return mt_str;
  }


static char VAR_PTR pr_mode( status_ptr )
/*
 *      returns verbal interpretation of the screen mode
*/
ter_status *status_ptr;
  {
    switch( status_ptr->mode )
      {
        case ALPH_SCREEN : return (char VAR_PTR) "alphanumeric";
        case GRAP_SCREEN : return (char VAR_PTR) "graphics";
      }
    return mt_str;
  }

