/*
    Copyright (C) 1997  Peter Vernon Lanspeary

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* ************************************************************************ */
/*                                                                          */
/*                Basic Graphics Terminal Graphics Driver for               */
/*                     Hewlett/Packard Graphics Language                    */
/*                   N.B. This is hardware dependent stuff                  */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/* This is a graphics driver for the Hewlett/Packard Graphics plotters      */
/* which uses the Hewlett/Packard Graphics Language (HP/GL). Output is sent */
/* to a file (which may be your terminal). The plotter may be operated in   */
/* eavesdrop mode if it is equipped with this ability . In order to         */
/* maintain compatibility with eavesdrop mode, there should be no output to */
/* ``stdout'' while graphics is switched on. Separate virtual pens are used */
/* for drawing lines and for text . Text is written using the plotter's     */
/* default font whenever possible . Most functions are of type `boolean'    */
/* and return a TRUE or FALSE result to indicate success or failure . Error */
/* messages are placed on the error queue -- which can be examined at       */
/* the discretion of the user.                                              */
/*                                                                          */
/* Any programs which use this driver must #include DEVICE_G_HEADER .       */
/* These functions must be linked with "clip.c" and "error.c" . The docu-   */
/* mentation on "clip.c" and "error.c" should be read before continuing .   */
/*                                                                          */
/* This driver does require use of the Hewlett/Packard plotters' ability to */
/* talk (as well as listen) if the output device is ``stdout''. t_newscr(), */
/* t_bell() and t_wait() are inactive unless the output deviceis ``stdout''.*/
/* The functions t_visib(), t_invis(), t_revers(), t_replac(), t_xhrput()   */
/* and t_xhrget() are inactive .                                            */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/* The function `t_begin()' has two arguments . The first `device' argument */
/* is the FILE* pointer to the output device. If ``device == stdout'', the  */
/* driver assumes that the plotter is operating in eavesdrop mode and will  */
/* request output from the plotter ( i.e. two-way communication ) . This    */
/* enables the driver to determine the plotter type, paper type, and in     */
/* some cases, the pen type .                                               */
/* The second argument is a character string of selected options . The      */
/* available options are the words                                          */
/*   "METRIC" (default) for ISO standard paper ( A4, A3, A2, A1, A0 ),      */
/*   "US"               for ANSI standard paper ( A, B, C, D, E ),          */
/*   "A4" (default)     210 x 297 mm ( or 8.5 x 11 in. ) paper size         */
/*   "A3"               297 x 420 mm ( or 11 x 17 in. ) paper size          */
/*   "A2"               420 x 594 mm ( or 17 x 22 in. ) paper size          */
/*   "A1"               594 x 840 mm ( or 22 x 34 in. ) paper size          */
/*   "A0"               840 x 1188 mm ( or 34 x 44 in. ) paper size         */
/*   "WIDE" (default)   X axis is parallel to long side of the paper        */
/*   "TALL"             Y axis is parallel to long side of the paper        */
/*   "BUILT_IN_FONTS" (default)  use plotter's buit-in fonts for ASCII chars*/
/*   "PRIVATE_FONTS"    use the driver's fonts for ASCII characters         */
/*   "GP_IB"            produce output for GP-IB communications with plotter*/
/*   "FILE_OUTPUT"      force one-way communication -- as if to a file      */
/*   "FIBRE_PEN" (default)  fibre-tipped pens                               */
/*   "ROLLER_PEN"           roller-ball pens                                */
/*   "DRAUGHTING"           draughting pens                                 */
/*   "TRANSPARENCY"         fibre-tipped transparency pens                  */
/*   "NORMAL" (default) normal-sized plotting area | ignored if not a 7580B */
/*   "EXPAND"           expand plotting area       | or 7585B plotter       */
/*   "7470A" (default)  HP 7470 A  plotter                                  */
/*   "7475A"            HP 7475 A  plotter                                  */
/*   "7550A"            HP 7550 A  or  HP 7550B  plotter                    */
/*   "7580B"            HP 7580 B  plotter                                  */
/*   "7585B"            HP 7585 B  plotter                                  */
/*   "COLORPRO"         HP ColorPro plotter                                 */
/*   "TTY_SIZE" {number} height of the teletype character cell              */
/* Words may be abbreviated and in upper, lower or mixed case . For example */
/* "tall A2 us PRIV draught 7585" specifies tall plots on C (17" x 22" ANSI)*/
/* sized paper in a HP 7585B plotter using draughting pens and the driver's */
/* fonts for stroke-based ASCII characters . Some of the selected options   */
/* are ignored if the plotter is in eavesdrop mode .                        */
/*                                                                          */
/* How to load paper into the plotter                                       */
/*  +---------+-------+--------+          +---------+-------+--------+      */
/*  | plotter | paper | orien- |          | plotter | paper | orien- |      */
/*  |  type   | size  | tation |          |  type   | size  | tation |      */
/*  +---------+-------+--------+          +---------+-------+--------+      */
/*  | HP7470A | A4, A |  vert  |          | HP7550A | A4, A |  wide  |      */
/*  | HP7475A | A4, A |  horiz |          |         | A3, B |  vert  |      */
/*  |         | A3, B |  vert  |          | HP7550B | A4, A |  horiz |      */
/*  | HP7585B | A4, A |  horiz |          |         | A3, B |  vert  |      */
/*  |         | A3, B |  vert  |          | HP7580B | A4, A |  horiz |      */
/*  |         | A2, C |  horiz |          |         | A3, B |  vert  |      */
/*  |         | A1, D |  vert  |          |         | A2, C |  vert  |      */
/*  |         | A0, E |  vert  |          |         | A1, D |  vert  |      */
/*  | ColorPro| A4, A |  vert  |          |         |       |        |      */
/*  +---------+-------+--------|          +---------+-------+--------|      */
/*    vert  = long side is vertical                                         */
/*    horiz = long side is horizontal                                       */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/*      RS-232 / GP-IB   USAGE                                              */
/*    By default, this driver assumes that the communication medium to the  */
/* plotter is an RS-232 serial line using an Xon-Xoff handshake protocol .  */
/* The driver expects that when the plotter returns data to the computer,   */
/* the operating system echos the data back to the plotter, and so it is    */
/* not necessary to set your terminal line to "no echo" . With an RS-232    */
/* line, "ESC.(" and "ESC.)" are used to turn the plotter "on" and "off"    */
/* and other device control instructions are used to set Xon/Xoff handshake */
/* protocol. If output is not to the user's terminal, the "GP_IB" option may*/
/* be used to suppress these device control instructions ( "ESC.(" & "ESC.)"*/
/* so that the data can be sent via a GP-IB ( i.e. General purpose interface*/
/* bus -- formerly HP-IB ) . If your operating system supports GP-IB as a   */
/* "file" device, you can use the GP-IB without intermediate files .        */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/*      SYSTEMS OF UNITS                                                    */
/* The graphics device standard specifies that the graphics units available */
/* to the user are "pixels" i.e. equal in size to the basic resolution of   */
/* the plotter . The origin of the "pixel" coordinate system (0,0) is the   */
/* lower left corner of the paper or screen . The paper is aligned so that  */
/* the x-axis is horizontal and the y-axis is vertical .                    */
/*      At the lowest level, the pixel units are converted to "plotter"     */
/* units by adding an offset and if necessary, by rotating 90 degrees       */
/* anticlockwise .                                                          */
/*      Usually, "pixel" coordinates are expressed as of type "coord_pair", */
/* while "plotter" coordinates are expressed as of type "int_coord" .       */
/*                                                                          */
/* ************************************************************************ */

#include "system.h"             /* system dependent stuff */
#include <stdio.h>
#include DEVICE_G_HEADER
#include FONT_SOURCE_HEADER
#include "dev_udef.h"

/* ************************************************************************ */
/*                                                                          */
/*                                CONSTANTS                                 */
/*                                                                          */
/* ************************************************************************ */

/*
 *      Machine constants
*/
#define TRUE          1
#define FALSE         0
#define PI            3.141592653589793238462643
/*#define NULL        0       NULL pointer - already defined in stdio.h   */
/*#define EOF         (-1)    end of file  - already defined in stdio.h   */

/*
 *    non printable ASCII characters
*/
#define NUL  '\000'   /* Nul                   */
#define ETX  '\003'   /* End of transmission - */
                      /*      label terminator */
#define BEL  '\007'   /* Bell                  */
#define BS   '\010'   /* Backspace             */
#define HT   '\011'   /* Horizontal Tabulation */
#define LF   '\012'   /* Line Feed             */
#define VT   '\013'   /* Vertical Tabulation   */
#define CR   '\015'   /* Carriage Return       */
#define EM   '\031'   /* End of Medium         */

/*
 *          constants for drawing characters
*/
#define DEF_SPACE   1.5       /* size of cell size compared with character*/
#define CLIP2R_ID   0x3c22    /* CL2R two dimensional rectangular clipping*/
#define POINTS_ONLY   0       /* do not draw lines between points         */
#define SOLID_LINE    1       /* dashed line type                         */
#define NO_PEN        0       /* pen 0 - no pen is selected               */
#define BLACK         1       /* pen 1 - black pen                        */
#define MAX_LINE_LEN  80      /* max no. of chars on line sent to plotter */
#define BUFFER_LEN  MAX_LINE_LEN - 5 /* maximum echo buffer length        */
#define ID_STRLEN   5
#define TXT_JITTER    3       /* max. misalignment of text chars in pixels*/
#define BUILT_HEIGHT  1.0     /* height of built-in characters            */
#define BUILT_WIDTH   0.6     /* width of built-in characters             */
#define BUILT_GAP     0.3     /* gap between built-in characters          */
/*
 *          definition of available options
*/
#define MAX_WORD_LEN   15     /* max length of word in the options list   */
#define NOT_SET        -1     /* option not specified in options string   */
#define MET_PAPER       0     /* Metric type paper - ie. A4 A3 A2 A1 A0   */
#define US_PAPER        1     /* "US" or "English" type paper, ie. A B... */
#define A4_PAPER        0     /* Paper size A4 or A */
#define A3_PAPER        1     /* Paper size A3 or B */
#define A2_PAPER        2     /* Paper size A2 or C */
#define A1_PAPER        3     /* Paper size A1 or D */
#define A0_PAPER        4     /* Paper size A0 or E */
#define WIDE_PAPER      0     /* "Horizontal" is parallel to long side of */
#define TALL_PAPER      1     /* "Vertical" is parallel to long side of paper */
#define HP7470A         0     /* plotter is  HP 7470 A */
#define HP7475A         1     /* plotter is  HP 7475 A */
#define HP7550A         2     /* plotter is  HP 7550 A */
#define HP7550B         3     /* plotter is  HP 7550 B */
#define HP7580B         4     /* plotter is  HP 7580 B */
#define HP7585B         5     /* plotter is  HP 7580 B */
#define COLORPRO        6     /* plotter is  HP ColorPro */
/*                 The order of the next 4 pen constants is IMPORTANT */
#define FIBRE_PEN       0     /* fibre-tipped pens */
#define ROLLER_PEN      1     /* ball-point pens   */
#define DRAFT_PEN       2     /* draughting pens   */
#define TRANS_PEN       3     /* transparency pens */
#define FIBRE_THICK    12     /*                  | for fibre-tipped pens */
#define ROLLER_THICK   12     /* line thickness   | for ball-point pens   */
#define DRAFT_THICK    20     /* in plotter units | for draughting pens   */
#define TRANS_THICK    28     /*                  | for transparency pens */
#define FIBRE_SPEED    32     /*           | for fibre-tipped pens */
#define ROLLER_SPEED   60     /* pen speed | for ball-point pens   */
#define DRAFT_SPEED    15     /* in cm/sec | for draughting pens   */
#define TRANS_SPEED    10     /*           | for transparency pens */
#define DEF_TTY_SIZE  240.0   /* default teletype (built-in) text size */
#define MIN_TTY_SIZ    80.0   /* minimum teletype (built-in) text size */
#define MAX_TTY_SIZ   400.0   /* maximum teletype (built-in) text size */

/* ************************************************************************ */
/*                                                                          */
/*                          Graphics Macros                                 */
/*                                                                          */
/* ************************************************************************ */
/*
 *  ABBREVIATIONS
 ******************
*/
#define  PLOTTER    hpgl->plotter
#define  STROKE     hpgl->stroke
#define  VIEW_CLIP  hpgl->view_clip
#define  OP_FILE    hpgl->op_file
#define  SETUP      hpgl->setup
#define  PAGE       hpgl->page

/*
 *  CATCH_NULL(pointer,caller,ret_type)
 ****************************************
 *      If "pointer" has a NULL value, an "Arg is NULL pointer" error message
 * is generated and the function is aborted . Otherwise, nothing is done .
 *      "pointer"   is the pointer to be tested,
 *      "caller"    is a character string naming the function
 *      "ret_type"  is the type of the returned zero value which indicates an
 *                    error condition
*/
#define CATCH_NULL(pointer,caller,ret_type) \
  if ( ! (pointer) )    \
      return (ret_type) make_error( ERR_ABORT, caller, err_null )

/*
 *  HYPOT(point)
 ****************
 *      Length of the vector "point" which is of type "coord_pair" .
*/
#define HYPOT(point)  \
        ( sqrt( (point)._x * (point)._x + (point)._y * (point)._y ) )

/*
 *  G_PUTC(c)
 **************
 *      Writes the character `c' to the graphics device (terminal) .
 *      There is no returned result.
*/
#define G_PUTC(c)           fputc( c, OP_FILE )

/*
 *  G_PUTS(string)
 *******************
 *      Writes the character string `string' to the graphics device
 *      (terminal) . The there is no returned result .
*/
#define G_PUTS(string)      fputs( string, OP_FILE )

/*
 *  G_ERROR()
 **************
 *      Determines the error status of the graphics device (terminal) .
 *      The returned value is TRUE if an error has occurred, otherwise FALSE .
*/
#define G_ERROR()           ferror( OP_FILE )

/*
 *  G_FLUSH()
 **************
 *      Flushes the graphics device (terminal) output buffer . There is no
 *      returned value .
*/
#define G_FLUSH()          fflush( OP_FILE )

/*  LIMIT( x, min, max )
 *************************
 *      Limits the value of `x' to between `min' and `max'
*/
#define LIMIT(x,min,max)  (x) <= (min) ? (min) : ( (x) >= (max) ? (max) : (x) )

/*
 *  MAX( x, y )
 ****************
 *      The returned value is the maximum of `x' and `y' .
*/
#define MAX(x,y)        ( (x) > (y) ? (x) : (y) )

/*
 *  MIN( x, y )
 ****************
 *      The returned value is the minimum of `x' and `y' .
*/
#define MIN(x,y)        ( (x) > (y) ? (y) : (x) )

/*
 *  SWAPD( a, b )
 *****************
 *    swaps the two double precision floating point values `a' and `b'
*/
#define SWAPD(a,b)       { double temp = b; b = a; a = temp; }

/*
 *  SWAP_SI( a, b )
 *******************
 *    swaps the two short integers `a' and `b'
*/
#define SWAP_SI(a,b)       { short int temp = b; b = a; a = temp; }

/*
 *  TOLOWER( c )
 ****************
 *      Converts character `c' to lower case if it is upper case
*/
#define TOLOWER(c)      'A' <= (c) && (c) <= 'Z' ? (c) + 'a' - 'A' : (c)

/*
 *  START_UP()
 ***************
 *      Calls `t_begin' with the default arguments `stdout' and `NULL'
 *  if `t_begin' has not previously been called .
*/
#define START_UP(caller,ret_type) \
    if ( ! hpgl ) \
        if ( ! t_begin( stdout, file_opt ) ) \
            return (ret_type) make_error( ERR_CALL, caller, mt_str )

/* ************************************************************************ */
/*                                                                          */
/*                          Type declarations                               */
/*                                                                          */
/* ************************************************************************ */

/*
 * device mode - plotter is on or off
*/
enum dev_mode
  {
    ON_PLOTTER,         /* plotter is switched on   */
    OFF_PLOTTER,        /* plotter is switched off  */
  };

/*
 * coordinate pair of short integers
*/
typedef struct
  {
    short int _ix;       /* abscissa */
    short int _iy;       /* ordinate */
  }
    int_coord;

/*
 * plotter status -- purpose is to remember what the plotter is doing
*/
typedef struct
  {
    enum dev_mode mode;     /* terminal mode                       */
    int           colour;   /* foreground colour                   */
    int           dash;     /* dashed line style                   */
    int           thick;    /* line and test width                 */
    int           fill_type; /* fill type for boxes etc.           */
    coord_pair    cursor;   /* alpha graphics text cursor position */
    coord_pair    pen;      /* vector graphics pen position        */
    coord_pair    ch_size;  /* character width and height in pixels*/
    coord_pair    ch_step;  /* displacement between characters     */
    double        ch_angle; /* direction of writing - radians ccw  */
    double        ch_slant; /* shear factor for character slant    */
  }
    Plot_Status;

/*
 * element of the P1 & P2 list for each plotter
*/
typedef struct
  {
    short int Pp_type;      /* paper type - US or metric */
    short int Pp_size;      /* paper size - A4 to A0     */
    int_coord P1_ll;        /* P1 - plotter units        */
    int_coord P2_ur;        /* P2 - plotter units        */
    int_coord Pll_marg;     /* width of left & bottom margin - plotter units */
    int_coord Pur_marg;     /* width of right & top margin - plotter units */
  }
    P1P2_elem;

/*
 *  This defines the capabilities, selected options and general setup of
 *  the plotter
*/

typedef struct
  {
    short int paper_type; /* Metric or US paper type                    */
    short int paper_size; /* A4, A3, A2, A1, A0 paper size              */
    short int paper_dirn; /* wide or tall orientation                   */
    short int exact_text; /* draw text exactly stroke by stroke         */
                          /*  - otherwise use plotter's internal font   */
    short int rs_232;     /* using RS-232 serial line - otherwise GP-IB */
    short int op_todisk;  /* HP-GL output is to disk - otherwise stdout */
    short int pen_type;   /* pen type                                   */
    short int expand_80;  /* reduced margin size - HP7580 & 7585 only   */
    short int plotter_id; /* HP7470, 7475, 7550, 7580, 7585, or ColotPro*/
    double tty_size;      /* height of teletype character cell          */
  }
    Setup_List;

typedef struct
  {
    short int paper_rota; /* coordinates rotated by 90 degrees          */
    short int no_of_pens; /* maximum number of pens in the plotter      */
    short int pen_thick;  /* thickness of line drawn by the pen - pixels*/
    short int alph_columns; /* number of character columns on the page  */
    short int alph_lines; /* number of character lines on the page      */
    boolean set_pthick;   /* set pen thickness on 1st selection of a pen  */
/*      integer coordinates are in "plotter" units                      */
    int_coord plot_size;  /* width and height of available plotting area*/
    int_coord offset;     /* for transforming "pixel" to "plotter" units*/
    P1P2_elem VAR_PTR P1P2_ptr; /* pointer to entry in paper size table */
/*      double coordinates are in "pixel units"                         */
    double tty_width;     /* width of tty char cell; height is tty_size */
    coord_pair ll_limit;  /* lower left plotting limit  - set by P1     */
    coord_pair ur_limit;  /* upper right plotting limit - set by P2     */
  }
    Page_List;

typedef struct
  {
    Setup_List setup;       /* properties determined by options settings */
    Page_List  page;        /* description of page set by `t_begin()' */
    Text_Attribute stroke;  /* stroke text attributes                 */
    struct
      {                     /* required and actual plotter status    */
        Plot_Status requir;
        Plot_Status actual;
      }
        plotter;
    FILE *op_file;          /* file pointer for the graphics output device  */
    clip_area VAR_PTR view_clip;/* pointer to viewport clipping area        */
    boolean hanging_cmd;    /* driver is writing a PD command - not finished*/
    int point_cntr;         /* no. of coordinate point on a line of output  */
    int pen_used;           /* bit switch - flags pens which have been used */
  }
    Gfx_Status;

/* ************************************************************************ */
/*                                                                          */
/*                            Global variables                              */
/*                                                                          */
/* ************************************************************************ */

extern StrokeFont
    asc_font,       /* font for ASCII chars   */
    sym_font;       /* font for special chars */
/*
 *  Dummy built-in font (Courier) contained in the plotter -- the numbers
 *  are determined empirically from an Apple Laserwriter .
*/
static StrokeFont built_font =
  { (StrokeChar VAR_PTR) 0, /* directory of character descriptions */
    TRUE,           /* character font file has been read   */
    96,             /* number of characters */
    BUILT_HEIGHT,   /* height of characters */
    BUILT_WIDTH,    /* width of characters  */
    BUILT_GAP       /* gap between characters */ };

static Gfx_Status VAR_PTR hpgl;
                /* complete description of the Hewlett-Packard plotter */

/*
 *      Identification string and pixel size for each plotter
*/
static char
    VAR_PTR ID_string[] = { "7470A", "7475A", "7550A", "7550B",
        "7580B", "7585B", "7440A" };
static double
    resolution[] =        { 0.02495, 0.02487, 0.02500, 0.02500,
        0.02500, 0.02500, 0.02495 };

/*
 *
 *      Common error messages
*/
static char
  VAR_PTR err_write = "Cannot write to graphics device",
  VAR_PTR err_null  = "Arg is NULL pointer",
  VAR_PTR mt_str    = "",
  VAR_PTR file_opt  = "file",
  VAR_PTR inactive  = "inactive";

/*
 *      Declare static, standard and math library functions
*/
static P1P2_elem VAR_PTR find_P1P2(), VAR_PTR compare_P1P2();
static int tty_str(), t_txtstr(), t_stkchr();
static char VAR_PTR pr_mode();
static int_coord VAR_PTR plot_units();
static void txt_params(), hp_horiz_hatch(), hp_vert_hatch(), hp_fill_hatch(),
    hp_NWSE_hatch(),  hp_NESW_hatch(), hp_diag_hatch(), hp_terminate();
static boolean find_pen_type(), hp_move(), hp_draw(), hp_pen(), hp_style(),
    hp_box(), hp_on(), hp_off(), set_options();

boolean t_parse_options();
int t_stncmp();
int abs(), free(), fputc(), fputs(), fflush();
double fabs(), sin(), cos(), sqrt();
char *malloc(), *strcat();

/* ************************************************************************ */
/*                                                                          */
/*                    Plotter driving routines                              */
/*                                                                          */
/* ************************************************************************ */


static boolean t_dump()
/*
 *      Dummy for the nonexistent screen dump routine
*/
  {
    return make_error( ERR_ABORT, (char VAR_PTR) "t_dump",
         (char VAR_PTR) "You've gotta be joking !" );
  }


static clip_area VAR_PTR t_begin( device, options )
/*
 *        Start graphics .
 * Decodes the required ``options'' string, initializes the plotter, initializes
 * variables and sets up the viewport clip. The returned value is a pointer to
 * the clip_area structure of the viewport . If the function fails, it returns
 * a NULL pointer. Appropriate error messages are produced in these cases .
*/
FILE *device;    /* file pointer for graphics device - usually a disk file */
char VAR_PTR options;   /* options string */
  {
    int_coord start_point;
    char VAR_PTR name = "t_begin";

    CATCH_NULL( device && options, name, clip_area VAR_PTR );
/*
 *    GRAB GRAPHICS STATUS DATA
 *      Find out if ``device'' has been registered as a device .
 *      If it has, check that it is indeed a Hewlett Packard plotter and grab
 *      the pointer to it graphics status data .
*/
    if ( t_register( T_ISREGIST, device, NONEXISTENT_DEVICE, (char VAR_PTR) 0) )
      {
        if ( (int) t_register( T_REG_TYPE, device, 0, (char VAR_PTR) 0 )
                != HEWLETT_GL_PLOTTER )
            return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "Wrong driver. Device is not HP/GL plotter" );
        hpgl = (Gfx_Status VAR_PTR) t_register( T_REG_DATA,
            device, NONEXISTENT_DEVICE, (char VAR_PTR) 0 );
/*
 *    DECODE THE OPTIONS AND ASSIGN PAGE & SETUP STRUCTURES
*/
        if ( ! set_options( options, name, FALSE ) )
            return (clip_area VAR_PTR) NULL;
      }
    else
      {
/*
 *    REGISTRATION
 *      ``device'' has never been registered, so allocate memory for the
 *      graphics status data and register the device as a HP/GL plotter .
*/
        hpgl = (Gfx_Status VAR_PTR)
            malloc( (unsigned int) sizeof(Gfx_Status) );
        if ( ! hpgl )
            return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "Cannot allocate memory" );
        if ( ! t_register( T_REGISTER, device,
                 HEWLETT_GL_PLOTTER, (char VAR_PTR) hpgl ) )
          {
            (void) free( (char *) hpgl );
            hpgl = (Gfx_Status VAR_PTR) 0;
            return (clip_area VAR_PTR) make_error( ERR_CALL, name, mt_str );
          }
/*
 *    DECODE THE OPTIONS AND ASSIGN PAGE & SETUP STRUCTURES
*/
        OP_FILE = device;
        if ( ! set_options( options, name, TRUE ) )
          {
            if ( ! t_register( T_DEREGIST, OP_FILE,
                    HEWLETT_GL_PLOTTER, (char VAR_PTR) hpgl ) )
                (void) make_error( ERR_CALL, name, mt_str );
            (void) free( (char *) hpgl );
            hpgl = (Gfx_Status VAR_PTR) 0;
            return (clip_area VAR_PTR) NULL;
          }
/*
 *           reset character attributes and plotter status
*/
        (void) reset_font( (Text_Attribute VAR_PTR) &STROKE,
            (double) MAX( PAGE.plot_size._ix, PAGE.plot_size._iy ) / 20.0 );
        if ( SETUP.exact_text )
            STROKE.font_ptr = (StrokeFont VAR_PTR) &asc_font;
        else
            STROKE.font_ptr = (StrokeFont VAR_PTR) &built_font;
/*
 *      SET THE PLOTTER STATUS
 *      and put pen in lower left corner.
*/
        PLOTTER.requir.colour = NO_PEN;
        PLOTTER.requir.dash   = SOLID_LINE;
        PLOTTER.requir.pen    = PAGE.ll_limit;
        PLOTTER.requir.cursor._x = PAGE.ll_limit._x;
        PLOTTER.requir.cursor._y = PAGE.ur_limit._y - SETUP.tty_size + 1.0;
        PLOTTER.requir.ch_slant  = PLOTTER.requir.ch_angle = 0.0;
        PLOTTER.requir.ch_size._x = PLOTTER.requir.ch_size._y =
            PLOTTER.requir.ch_step._x = PLOTTER.requir.ch_step._y = NOT_SET;
/*
 *          set the plotter clipping window to the hard clip limits
*/
        if ( ! ( VIEW_CLIP = def_2rclip( (coord_pair VAR_PTR) &PAGE.ll_limit,
                 (coord_pair VAR_PTR) &PAGE.ur_limit,
                 (coord_pair VAR_PTR) &PAGE.ll_limit ) ) )
            return (clip_area VAR_PTR) make_error( ERR_CALL, name, mt_str );
      }
/*
 *      SET THE PLOTTER STATUS
 *      and put pen in lower left corner.
*/
    (void) hp_pen( NO_PEN );
    PLOTTER.actual.dash = SOLID_LINE;
    PLOTTER.actual.cursor = PLOTTER.actual.pen = PLOTTER.requir.pen =
        PAGE.ll_limit;
    PLOTTER.actual.ch_size._x = PLOTTER.actual.ch_size._y =
        PLOTTER.actual.ch_step._x = PLOTTER.actual.ch_step._y =
        PLOTTER.actual.ch_slant = PLOTTER.actual.ch_angle = 0;
    (void) hp_move( plot_units( (coord_pair VAR_PTR) &PAGE.ll_limit,
        (int_coord VAR_PTR) &start_point ) );
    if ( G_ERROR() )
        return (clip_area VAR_PTR) make_error( ERR_WARN, name, err_write );
    return VIEW_CLIP;
  }


#define ST_OPT  (char VAR_PTR) &st_opt

static boolean set_options( options, caller, first_call )
/*
 * Decode the options to ``t_begin'' and assign appropriate values to
 * the setup and page data structures . This may require talking to the
 * plotter .
*/
char VAR_PTR options,   /* string -- list of options */
     VAR_PTR caller;    /* name of calling function  */
boolean first_call;     /* TRUE = first call to t_begin with this output file */
  {
    short int VAR_PTR clear_ptr;
/*
 *     Paper size tables for each plotter
 * N.B. at least one list of scaling points is required for each plotter
*/
    static P1P2_elem
/*
 ***** HP7470A plotter ***** */
      P1P2_7470[] = {
        { MET_PAPER, A4_PAPER, {0,0}, {10900,7650}, {560,320}, {440,440} },
        { US_PAPER,  A4_PAPER, {0,0}, {10300,7650}, {560,320}, {336,676} } },
/*
 ***** HP7475A plotter ***** */
      P1P2_7475[] = {
        { MET_PAPER, A4_PAPER, {0,0}, {11040,7721},  {385,140}, {510,590} },
        { US_PAPER,  A4_PAPER, {0,0}, {10365,7962},  {400,145}, {460,590} },
        { MET_PAPER, A3_PAPER, {0,0}, {16158,11040}, {590,415}, {155,510} },
        { US_PAPER,  A3_PAPER, {0,0}, {16640,10365}, {600,380}, {120,460} } },
/*
 ***** HP7550 plotter ***** */
      P1P2_7550[] = {
        { MET_PAPER, A4_PAPER, {0,0}, {10870,7600},  {500,195}, {510,605} },
        { US_PAPER,  A4_PAPER, {0,0}, {10170,7840},  {500,240}, {500,550} },
        { MET_PAPER, A3_PAPER, {0,0}, {15970,10870}, {580,490}, {220,520} },
        { US_PAPER,  A3_PAPER, {0,0}, {16450,10170}, {550,500}, {220,500} } },
/*
 ***** normal HP7580 plotter ***** */
      P1P2_n7580[] = {
 { MET_PAPER, A4_PAPER, {-3000,-4830},  {3000,4830},  {600,600}, {1560,600} },
 { US_PAPER,  A4_PAPER, {-3100,-4850},  {3100,4850},  {600,600}, {1560,600} },
 { MET_PAPER, A3_PAPER, {-7200,-5200},  {7200,5200},  {1560,600},{600,600} },
 { US_PAPER,  A3_PAPER, {-7420,-4850},  {7420,4850},  {1560,600},{600,600} },
 { MET_PAPER, A2_PAPER, {-10680,-7680}, {10680,7680}, {1560,600},{600,600} },
 { US_PAPER,  A2_PAPER, {-9960,-7900},  {9960,7900},  {1560,600},{600,600} },
 { MET_PAPER, A1_PAPER, {-15600,-11100},{15600,11100},{1560,600},{600,600} },
 { US_PAPER,  A1_PAPER, {-16000,-10450},{16000,10450},{1560,600},{600,600} } },
/*
 ***** expand HP7580 plotter ***** */
      P1P2_e7580[] = {
 { MET_PAPER, A4_PAPER, {-3400,-5600},  {3400,5600},  {200,200}, {1160,200} },
 { US_PAPER,  A4_PAPER, {-3510,-5260},  {3510,5260},  {200,200}, {1160,200} },
 { MET_PAPER, A3_PAPER, {-7600,-5620},  {7600,5620},  {1160,200},{200,200} },
 { US_PAPER,  A3_PAPER, {-7820,-5250},  {7820,5250},  {1160,200},{200,200} },
 { MET_PAPER, A2_PAPER, {-11020,-8080}, {11020,8080}, {1160,200},{200,200} },
 { US_PAPER,  A2_PAPER, {-10360,-8300}, {10360,8300}, {1160,200},{200,200} },
 { MET_PAPER, A1_PAPER, {-16000,-11500},{16000,11500},{1160,200},{200,200} },
 { US_PAPER,  A1_PAPER, {-16450,-10800},{16450,10800},{1160,200},{200,200} } },
/*
 ***** normal HP7585 plotter ***** */
      P1P2_n7585[] = {
 { MET_PAPER, A4_PAPER, {-3000,-4830},  {3000,4830},  {600,600}, {1560,600} },
 { US_PAPER,  A4_PAPER, {-3100,-4850},  {3100,4850},  {600,600}, {1560,600} },
 { MET_PAPER, A3_PAPER, {-7200,-5200},  {7200,5200},  {1560,600},{600,600} },
 { US_PAPER,  A3_PAPER, {-7420,-4840},  {7420,4850},  {1560,600},{600,600} },
 { MET_PAPER, A2_PAPER, {-7200,-11100}, {7200,11100}, {600,600}, {1560,600} },
 { US_PAPER,  A2_PAPER, {-7420,-10450}, {7420,10450}, {600,600}, {1560,600} },
 { MET_PAPER, A1_PAPER, {-15600,-11100},{15600,11100},{1560,600},{1560,600} },
 { US_PAPER,  A1_PAPER, {-16000,-10450},{16000,10450},{1560,600},{1560,600} },
 { MET_PAPER, A0_PAPER, {-22500,-16000},{22500,16000},{1560,600},{1560,600} },
{ US_PAPER,  A0_PAPER, {-21100,-16500},{21100,16500},{1560,600},{1560,600} } },
/*
 ***** expand HP7585 plotter ***** */
      P1P2_e7585[] = {
 { MET_PAPER, A4_PAPER, {-3400,-5600},  {3400,5600},  {200,200}, {1160,200} },
 { US_PAPER,  A4_PAPER, {-3510,-5260},  {3510,5260},  {200,200}, {1160,200} },
 { MET_PAPER, A3_PAPER, {-7600,-5620},  {7600,5620},  {1160,200},{200,200} },
 { US_PAPER,  A3_PAPER, {-7820,-5250},  {7820,5250},  {1160,200},{200,200} },
 { MET_PAPER, A2_PAPER, {-7600,-11550}, {7600,11550}, {200,200}, {1160,200} },
 { US_PAPER,  A2_PAPER, {-7820,-10850}, {7820,10850}, {200,200}, {1160,200} },
 { MET_PAPER, A1_PAPER, {-16000,-11500},{16000,11500},{1160,200},{1160,200} },
 { US_PAPER,  A1_PAPER, {-16450,-10800},{16450,10800},{1160,200},{1160,200} },
 { MET_PAPER, A0_PAPER, {-22900,-16400},{22900,16400},{1160,200},{1160,200} },
{ US_PAPER,  A0_PAPER, {-21500,-16900},{21500,16900},{1160,200},{1160,200} } },
/*
 ***** HP ColorPro plotter ***** */
      P1P2_ColorPro[] = {
        { MET_PAPER, A4_PAPER, {0,0}, {10900,7650}, {560,375}, {450,395} },
        { US_PAPER,  A4_PAPER, {0,0}, {10300,7650}, {560,375}, {1050,395} } };
/*
 *  plotter options
*/
    static Setup_List st_opt;
    static Option_Elem option_list[] = {
        { "METRIC",         1, OPT_SHORT,  ST_OPT.paper_type, MET_PAPER },
        { "US",             1, OPT_SHORT,  ST_OPT.paper_type, US_PAPER },
        { "A4",             2, OPT_SHORT,  ST_OPT.paper_size, A4_PAPER },
        { "A3",             2, OPT_SHORT,  ST_OPT.paper_size, A3_PAPER },
        { "A2",             2, OPT_SHORT,  ST_OPT.paper_size, A2_PAPER },
        { "A1",             2, OPT_SHORT,  ST_OPT.paper_size, A1_PAPER },
        { "A0",             2, OPT_SHORT,  ST_OPT.paper_size, A0_PAPER },
        { "WIDE",           1, OPT_SHORT,  ST_OPT.paper_dirn, WIDE_PAPER },
        { "TALL",           2, OPT_SHORT,  ST_OPT.paper_dirn, TALL_PAPER },
        { "BUILT_IN_FONTS", 1, OPT_SHORT,  ST_OPT.exact_text, FALSE },
        { "PRIVATE_FONTS",  1, OPT_SHORT,  ST_OPT.exact_text, TRUE  },
        { "GP_IB",          1, OPT_SHORT,  ST_OPT.rs_232,     FALSE },
        { "FILE_OUTPUT",    3, OPT_SHORT,  ST_OPT.op_todisk,  TRUE  },
        { "FIBRE_PEN",      3, OPT_SHORT,  ST_OPT.pen_type,   FIBRE_PEN },
        { "ROLLER_PEN",     1, OPT_SHORT,  ST_OPT.pen_type,   ROLLER_PEN },
        { "DRAUGHTING",     1, OPT_SHORT,  ST_OPT.pen_type,   DRAFT_PEN },
        { "TRANSPARENCY",   2, OPT_SHORT,  ST_OPT.pen_type,   TRANS_PEN },
        { "NORMAL",         1, OPT_SHORT,  ST_OPT.expand_80,  FALSE },
        { "EXPAND",         1, OPT_SHORT,  ST_OPT.expand_80,  TRUE  },
        { "7470A",          4, OPT_SHORT,  ST_OPT.plotter_id, HP7470A },
        { "7475A",          4, OPT_SHORT,  ST_OPT.plotter_id, HP7475A },
        { "7550A",          5, OPT_SHORT,  ST_OPT.plotter_id, HP7550A },
        { "7550B",          5, OPT_SHORT,  ST_OPT.plotter_id, HP7550B },
        { "7580B",          4, OPT_SHORT,  ST_OPT.plotter_id, HP7580B },
        { "7585B",          4, OPT_SHORT,  ST_OPT.plotter_id, HP7585B },
        { "7440A",          3, OPT_SHORT,  ST_OPT.plotter_id, COLORPRO },
        { "COLORPRO",       1, OPT_SHORT,  ST_OPT.plotter_id, COLORPRO },
        { "TTY_SIZE",       2, OPT_DOUBLE, ST_OPT.tty_size,   OPT_NUM_ARG },
        { (char VAR_PTR) 0, 0, 0,          (char VAR_PTR) 0,  0 } };
/*
 *          table of line thicknesses and pen speeds for each type of pen
*/
    static struct
      {
        short int p_thick;  /* pen line thickness in plotter units */
        short int p_speed;  /* pen speed in cm/sec */
      }
    pen_list[] = { { FIBRE_THICK,  FIBRE_SPEED  },   /* fibre-tipped pen */
                   { ROLLER_THICK, ROLLER_SPEED },   /* ball-point pen   */
                   { DRAFT_THICK,  DRAFT_SPEED  },   /* draughting pen   */
                   { TRANS_THICK,  TRANS_SPEED  } }; /* transparency pen */
/*
 *          clear any previous plotter setup
*/
    for( clear_ptr =  (short int VAR_PTR) &st_opt.paper_type;
         clear_ptr <= (short int VAR_PTR) &st_opt.plotter_id; )
        *clear_ptr++ = NOT_SET;
    st_opt.tty_size  = NOT_SET;
/*
 *           decode the options list
*/
    if ( options )
        if ( ! t_parse_options( options, (Option_Elem VAR_PTR) option_list,
               caller ) )
            return FALSE;
    SETUP = st_opt;
/*
 *      ensure the default setup values if they are NOT_SET and
 *      remove any inconsistencies in the setup values ( if possible )
*/
    if ( SETUP.rs_232 == NOT_SET || ( ! SETUP.rs_232 && OP_FILE == stdout ) )
        SETUP.rs_232 = TRUE;
    if ( SETUP.op_todisk == NOT_SET )
        SETUP.op_todisk = OP_FILE != stdout;
/*      next test is redundant but is included in case the rules are changed */
    else if ( ! SETUP.op_todisk && OP_FILE != stdout )
        SETUP.op_todisk = TRUE;
    if ( SETUP.paper_type == NOT_SET )
        SETUP.paper_type = MET_PAPER;
    if ( SETUP.paper_dirn == NOT_SET )
        SETUP.paper_dirn = WIDE_PAPER;
    if ( SETUP.exact_text == NOT_SET )
        SETUP.exact_text = FALSE;
    if ( SETUP.tty_size == (double) NOT_SET )
        SETUP.tty_size = DEF_TTY_SIZE;
    else
        SETUP.tty_size = LIMIT( SETUP.tty_size, MIN_TTY_SIZ, MAX_TTY_SIZ );
/*
 *          NON-INTERACTIVE -- OUTPUT TO DISK FILE
 *************************************************
*/
    if ( SETUP.op_todisk )
      {
/*
 *          set any unspecified options to the default values
 *
 *  -- but don't set the pen speed; unless the options force a particular
 *     pen speed, the plotter is allowed to default the pen speed to the
 *     value determined by the carousel type .
*/
        if ( SETUP.plotter_id == NOT_SET )
            SETUP.plotter_id = HP7470A;
        if ( SETUP.paper_size == NOT_SET )
            SETUP.paper_size = A4_PAPER;
        if ( SETUP.expand_80 == NOT_SET )
            SETUP.expand_80 = FALSE;
/*
 *          set the P1 & P2 limits according to the plotter type
*/
        switch ( SETUP.plotter_id )
          {
            case HP7475A :
                PAGE.no_of_pens = 6;
                PAGE.P1P2_ptr = find_P1P2( (P1P2_elem VAR_PTR) P1P2_7475, 4 );
                break;
            case HP7550A :
            case HP7550B :
                PAGE.no_of_pens = 8;
                PAGE.P1P2_ptr = find_P1P2( (P1P2_elem VAR_PTR) P1P2_7550, 4 );
                break;
            case HP7580B :
                PAGE.no_of_pens = 8;
                if ( SETUP.expand_80 )
                    PAGE.P1P2_ptr =
                        find_P1P2( (P1P2_elem VAR_PTR) P1P2_e7580, 8 );
                else
                    PAGE.P1P2_ptr =
                        find_P1P2( (P1P2_elem VAR_PTR) P1P2_n7580, 8 );
                break;
            case HP7585B :
                PAGE.no_of_pens = 8;
                if ( SETUP.expand_80 )
                    PAGE.P1P2_ptr =
                        find_P1P2( (P1P2_elem VAR_PTR) P1P2_e7585, 10 );
                else
                    PAGE.P1P2_ptr =
                        find_P1P2( (P1P2_elem VAR_PTR) P1P2_n7585, 10 );
                break;
            case COLORPRO :
                PAGE.no_of_pens = 8;
                PAGE.P1P2_ptr =
                    find_P1P2( (P1P2_elem VAR_PTR) P1P2_ColorPro, 2 );
                break;
            case HP7470A :
                /* fall through */
            default :
                PAGE.no_of_pens = 2;
                PAGE.P1P2_ptr = find_P1P2( (P1P2_elem VAR_PTR) P1P2_7470, 2 );
                break;
          }
/*
 *          switch on and initialize plotter - don't use hp_on()
 *              ESC . ( -- Plotter on
 *              ESC . J -- Abort device control instructions
 *              ESC . I -- Set Handshake Mode 2, Xon/Xoff
 *                  81;;17: -- Xoff threshold, omit enquiry character,
 *                             CTRL-Q is the X-on trigger character
 *              ESC . N -- Set extended output and handshake mode
 *                  ;19:    -- omit intercharacter delay,
 *                             CTRL-S is the X-off trigger character
 *              N.B. output terminator is CR LF for GPIB .
*/
        if ( SETUP.rs_232 && first_call )
            (void) G_PUTS( "\033.(\033.J\033.I81;;17:\033.N;19:" );
      }
    else
/*
 *          INTERACTIVE -- OUTPUT DIRECT TO PLOTTER
 *          ***************************************
*/
      {
        char OI_string[8];
        static P1P2_elem OW_limit;
        SETUP.expand_80 = FALSE;
/*
 *          switch on and initialize plotter - don't use hp_on()
 *          set handshake to Xon/Xoff
*/
        if ( SETUP.rs_232 && first_call )
            (void) G_PUTS( "\033.(\033.J" );
        (void) G_PUTS( "IN;OI;" );
        (void) G_FLUSH();
        PLOTTER.actual.mode = ON_PLOTTER;
        (void) fgets( OI_string, 8, stdin );
/*
 *          set the P1 & P2 limits according to the plotter type
*/
        if ( ! t_stncmp( (char VAR_PTR) OI_string,
                (char VAR_PTR) ID_string[HP7470A], ID_STRLEN ) )
          {
            SETUP.plotter_id = HP7470A;
            PAGE.no_of_pens = 2;
            (void) G_PUTS( "IN;OW;" );
            (void) G_FLUSH();
            (void) fscanf( stdin, "%hd,%hd,%hd,%hd%*c",
                &OW_limit.P1_ll._ix, &OW_limit.P1_ll._iy,
                &OW_limit.P2_ur._ix, &OW_limit.P2_ur._iy );
            PAGE.P1P2_ptr = compare_P1P2( (P1P2_elem VAR_PTR) &OW_limit,
                (P1P2_elem VAR_PTR) P1P2_7470, 2, 0 );
          }
        else if ( ! t_stncmp( (char VAR_PTR) OI_string,
                (char VAR_PTR) ID_string[HP7475A], ID_STRLEN ) )
          {
            SETUP.plotter_id = HP7475A;
            PAGE.no_of_pens = 6;
            (void) G_PUTS( "IN;OH;" );
            (void) G_FLUSH();
            (void) fscanf( stdin, "%hd,%hd,%hd,%hd%*c",
                &OW_limit.P1_ll._ix, &OW_limit.P1_ll._iy,
                &OW_limit.P2_ur._ix, &OW_limit.P2_ur._iy );
            PAGE.P1P2_ptr = compare_P1P2( (P1P2_elem VAR_PTR) &OW_limit,
                (P1P2_elem VAR_PTR) P1P2_7475, 4, 0 );
          }
        else if ( ! t_stncmp( (char VAR_PTR) OI_string,
                (char VAR_PTR) ID_string[HP7550A], ID_STRLEN ) )
          {
            SETUP.plotter_id = HP7550A;
            PAGE.no_of_pens = 8;
            (void) G_PUTS( "IN;OH;" );
            (void) G_FLUSH();
            (void) fscanf( stdin, "%hd,%hd,%hd,%hd%*c",
                &OW_limit.P1_ll._ix, &OW_limit.P1_ll._iy,
                &OW_limit.P2_ur._ix, &OW_limit.P2_ur._iy );
            PAGE.P1P2_ptr = compare_P1P2( (P1P2_elem VAR_PTR) &OW_limit,
                (P1P2_elem VAR_PTR) P1P2_7550, 4, 0 );
            if ( ! find_pen_type( caller ) )
                return FALSE;
          }
        else if ( ! t_stncmp( (char VAR_PTR) OI_string,
                (char VAR_PTR) ID_string[HP7550B], ID_STRLEN ) )
          {
            SETUP.plotter_id = HP7550B;
            PAGE.no_of_pens = 8;
            (void) G_PUTS( "IN;OH;" );
            (void) G_FLUSH();
            (void) fscanf( stdin, "%hd,%hd,%hd,%hd%*c",
                &OW_limit.P1_ll._ix, &OW_limit.P1_ll._iy,
                &OW_limit.P2_ur._ix, &OW_limit.P2_ur._iy );
            PAGE.P1P2_ptr = compare_P1P2( (P1P2_elem VAR_PTR) &OW_limit,
                (P1P2_elem VAR_PTR) P1P2_7550, 4, 0 );
            if ( ! find_pen_type( caller ) )
                return FALSE;
          }
        else if ( ! t_stncmp( (char VAR_PTR) OI_string,
                (char VAR_PTR) ID_string[HP7580B], ID_STRLEN ) )
          {
            SETUP.plotter_id = HP7580B;
            PAGE.no_of_pens = 8;
            (void) G_PUTS( "IN;OH;" );
            (void) G_FLUSH();
            (void) fscanf( stdin, "%hd,%hd,%hd,%hd%*c",
                &OW_limit.P1_ll._ix, &OW_limit.P1_ll._iy,
                &OW_limit.P2_ur._ix, &OW_limit.P2_ur._iy );
            if ( ! ( PAGE.P1P2_ptr =
                    compare_P1P2( (P1P2_elem VAR_PTR) &OW_limit,
                                  (P1P2_elem VAR_PTR) P1P2_n7580, 8, 160 ) ) )
                if ( ( PAGE.P1P2_ptr =
                        compare_P1P2( (P1P2_elem VAR_PTR) &OW_limit,
                                  (P1P2_elem VAR_PTR) P1P2_e7580, 8, 160 ) ) )
                    SETUP.expand_80 = TRUE;
            if ( ! find_pen_type( caller ) )
                return FALSE;
          }
        else if ( ! t_stncmp( (char VAR_PTR) OI_string,
                (char VAR_PTR) ID_string[HP7585B], ID_STRLEN ) )
          {
            SETUP.plotter_id = HP7585B;
            PAGE.no_of_pens = 8;
            (void) G_PUTS( "IN;OH;" );
            (void) G_FLUSH();
            (void) fscanf( stdin, "%hd,%hd,%hd,%hd%*c",
                &OW_limit.P1_ll._ix, &OW_limit.P1_ll._iy,
                &OW_limit.P2_ur._ix, &OW_limit.P2_ur._iy );
            if ( ! ( PAGE.P1P2_ptr =
                    compare_P1P2( (P1P2_elem VAR_PTR) &OW_limit,
                                  (P1P2_elem VAR_PTR) P1P2_n7585, 10, 160 ) ) )
                if ( ( PAGE.P1P2_ptr =
                        compare_P1P2( (P1P2_elem VAR_PTR) &OW_limit,
                                  (P1P2_elem VAR_PTR) P1P2_e7585, 10, 160 ) ) )
                    SETUP.expand_80 = TRUE;
            if ( ! find_pen_type( caller ) )
                return FALSE;
          }
        else if ( ! t_stncmp( (char VAR_PTR) OI_string,
                (char VAR_PTR) ID_string[COLORPRO], ID_STRLEN ) )
          {
            SETUP.plotter_id = COLORPRO;
            PAGE.no_of_pens = 8;
            (void) G_PUTS( "IN;OW;" );
            (void) G_FLUSH();
            (void) fscanf( stdin, "%hd,%hd,%hd,%hd%*c",
                &OW_limit.P1_ll._ix, &OW_limit.P1_ll._iy,
                &OW_limit.P2_ur._ix, &OW_limit.P2_ur._iy );
            PAGE.P1P2_ptr = compare_P1P2( (P1P2_elem VAR_PTR) &OW_limit,
                (P1P2_elem VAR_PTR) P1P2_ColorPro, 2, 0 );
          }
        else
            return make_error( ERR_ABORT, caller,
                (char VAR_PTR) "Unknown plotter" );
        if ( PAGE.P1P2_ptr )
          {
            SETUP.paper_type = PAGE.P1P2_ptr->Pp_type;
            SETUP.paper_size = PAGE.P1P2_ptr->Pp_size;
/*
 *          set handshake to Xon/Xoff
*/
            if ( SETUP.rs_232 && first_call )
                (void) G_PUTS( "\033.I81;;17:\033.N;19:" );
          }
      }
/*
 *          (re-)initialize plotter and specify hard-clip limits
*/
    if ( PAGE.P1P2_ptr )
        (void) fprintf( OP_FILE, "IN;IP%d,%d,%d,%d;\n",
                 PAGE.P1P2_ptr->P1_ll._ix, PAGE.P1P2_ptr->P1_ll._iy,
                 PAGE.P1P2_ptr->P2_ur._ix, PAGE.P1P2_ptr->P2_ur._iy );
    else
         return make_error( ERR_ABORT, caller,
                (char VAR_PTR) "Illegal paper size" );
/*
 *      set the plotter clipping window
*/
    (void) fprintf( OP_FILE, "IW%d,%d,%d,%d;\n",
        PAGE.P1P2_ptr->P1_ll._ix, PAGE.P1P2_ptr->P1_ll._iy,
        PAGE.P1P2_ptr->P2_ur._ix, PAGE.P1P2_ptr->P2_ur._iy );
/*
 *      set pen speed and thickness ( for shading )
*/
    if ( SETUP.pen_type != NOT_SET )
      {
        (void) fprintf( OP_FILE, "VS%d;\n", pen_list[SETUP.pen_type].p_speed );
        PAGE.pen_thick = pen_list[SETUP.pen_type].p_thick;
/*          use PT instruction ? */
        PAGE.set_pthick = SETUP.plotter_id == HP7475A ||
            SETUP.plotter_id == HP7550A || SETUP.plotter_id == HP7550B;
      }
    else
      {
        SETUP.pen_type  = FIBRE_PEN;
        PAGE.pen_thick = FIBRE_THICK;
        PAGE.set_pthick = FALSE;             /* don't use PT instruction */
        if ( SETUP.plotter_id == HP7470A || SETUP.plotter_id == HP7475A )
            (void) fprintf( OP_FILE, "VS%d;\n", pen_list[FIBRE_PEN].p_speed );
      }
/*
 *          complete the "plotter unit" setup info and
 *          "pixel units" plotting limits
*/
    PAGE.plot_size._ix = PAGE.P1P2_ptr->P2_ur._ix - PAGE.P1P2_ptr->P1_ll._ix;
    PAGE.plot_size._iy = PAGE.P1P2_ptr->P2_ur._iy - PAGE.P1P2_ptr->P1_ll._iy;
    PAGE.tty_width = SETUP.tty_size * built_font.font_wid/ built_font.font_hei;
    if (  ( PAGE.paper_rota   =
             ( ( SETUP.paper_dirn == WIDE_PAPER ) ^
               ( PAGE.plot_size._ix > PAGE.plot_size._iy ) ) ) )
      {
        PAGE.ur_limit._x = ( PAGE.ll_limit._x = PAGE.P1P2_ptr->Pll_marg._iy )
                            + PAGE.plot_size._iy;
        PAGE.ur_limit._y = ( PAGE.ll_limit._y = PAGE.P1P2_ptr->Pur_marg._ix )
                            + PAGE.plot_size._ix;
        PAGE.offset._ix = (int) PAGE.ll_limit._y + PAGE.P1P2_ptr->P2_ur._ix;
        PAGE.offset._iy = (int) PAGE.ll_limit._x - PAGE.P1P2_ptr->P1_ll._iy;
        PAGE.alph_columns = PAGE.plot_size._iy / PAGE.tty_width;
        PAGE.alph_lines   = PAGE.plot_size._ix / SETUP.tty_size;
      }
    else
      {
        PAGE.ur_limit._x = ( PAGE.ll_limit._x = PAGE.P1P2_ptr->Pll_marg._ix )
                            + PAGE.plot_size._ix;
        PAGE.ur_limit._y = ( PAGE.ll_limit._y = PAGE.P1P2_ptr->Pll_marg._iy )
                            + PAGE.plot_size._iy;
        PAGE.offset._ix = (int) PAGE.ll_limit._x - PAGE.P1P2_ptr->P1_ll._ix;
        PAGE.offset._iy = (int) PAGE.ll_limit._y - PAGE.P1P2_ptr->P1_ll._iy;
        PAGE.alph_columns = PAGE.plot_size._ix / PAGE.tty_width;
        PAGE.alph_lines   = PAGE.plot_size._iy / SETUP.tty_size;
      }
    return TRUE;
  }
#undef ST_OPT


static boolean t_end()
/*
 *      End of graphics . Delete character fonts and frees the memory allocated
 *  to them, deregister the device, eject the paper .
*/
  {
    boolean ok = TRUE;
    char VAR_PTR name = "t_end";
    START_UP( name, boolean );
/*
 *      Put the pen back in its carousel, move to the lower left corner and
 *      eject the page .
*/
    if ( PLOTTER.actual.mode != ON_PLOTTER )
        ok &= hp_on();
    ok &= hp_pen( NO_PEN );
    ok &= hp_move( (int_coord VAR_PTR) &PAGE.P1P2_ptr->P1_ll );
    if ( SETUP.plotter_id == HP7550A || SETUP.plotter_id == HP7550B )
        (void) G_PUTS( "PG;\n" );
    if ( SETUP.rs_232 )
        (void) G_PUTS( "\033.)\n" );
    PLOTTER.actual.mode = PLOTTER.requir.mode = OFF_PLOTTER;
    if ( ! ok || G_ERROR() )
        return make_error( ERR_ABORT, name, err_write );

/*
 *      Deregister the device and free the memory used by the graphics status
 *      data .
*/
    if ( ! t_register( T_DEREGIST, OP_FILE,
            HEWLETT_GL_PLOTTER, (char VAR_PTR) hpgl ) )
        (void) make_error( ERR_CALL, name, mt_str );
    (void) free( (char *) hpgl );
    hpgl = (Gfx_Status VAR_PTR) 0;
/*
 *      Delete character fonts and frees the memory allocated to them
*/
    if ( asc_font.font_got )
        (void) clear_font( (StrokeFont VAR_PTR) &asc_font );
    if ( sym_font.font_got )
        (void) clear_font( (StrokeFont VAR_PTR) &sym_font );
    return TRUE;
  }


static boolean t_vwport( clip_ptr )
/*
 *      Set the clipping limits for the viewport . The old limits are replaced .
 * The `clip_ptr' must point to  clip_area which has been defined by
 * `def_2rclip()' . The function fails if
 *   a) clip_ptr is a NULL pointer
 *   b) clip_ptr points to something which is not a structure for a two
 *      dimensional rectangular clipping region .
 *   c) cannot write to the file or plotter
 *   d) the new clipping limits are beyond the limits of the plotter
 *
 * Argument
*/
clip_area VAR_PTR clip_ptr;
  {
    char VAR_PTR name = "t_vwport";
    START_UP( name, boolean );
    CATCH_NULL( clip_ptr, name, boolean );
    if ( clip_ptr->id_struct == CLIP2R_ID )
      {
        if ( min_clip( clip_ptr, _x ) >= PAGE.ll_limit._x &&
             max_clip( clip_ptr, _x ) <= PAGE.ur_limit._x &&
             min_clip( clip_ptr, _y ) >= PAGE.ll_limit._y &&
             max_clip( clip_ptr, _y ) <= PAGE.ur_limit._y )
          {
            int_coord ll_corner, ur_corner;
            if ( hpgl->hanging_cmd )
                hp_terminate();
            if ( PLOTTER.actual.mode != ON_PLOTTER )
                (void) hp_on();
            (void) plot_units( (coord_pair VAR_PTR) MIN_clip( clip_ptr ),
                (int_coord VAR_PTR) &ll_corner );
            (void) plot_units( (coord_pair VAR_PTR) MAX_clip( clip_ptr ),
                (int_coord VAR_PTR) &ur_corner );
            if ( ll_corner._ix > ur_corner._ix )
                SWAP_SI( ll_corner._ix, ur_corner._ix );
            (void) fprintf( OP_FILE, "IW%d,%d,%d,%d;\n", ll_corner._ix,
                ll_corner._iy, ur_corner._ix, ur_corner._iy );
            VIEW_CLIP = clip_ptr;
            if ( G_ERROR() )
                return make_error( ERR_ABORT, name, err_write );
            else
                return TRUE;
          }
        return make_error( ERR_ABORT, name,
            (char VAR_PTR) "Outside device limits" );
      }
    return make_error( ERR_ABORT, name,
        (char VAR_PTR) "Invalid pointer" );
  }


static boolean t_colour( ipen )
/*
 *    Selects specified colour ( or pen ) .
 *
 * Integers from 0 to 8 are valid values of `ipen' .
 * The returned value is TRUE if  0 <= ipen <= 8 . If these limits are
 * exceeded, the value is wrapped around into the range 1 to 8 .
 * If the plotter has fewer than 8 pens, the value of `ipen' is further
 * wrapped around the number of available pens .
 *
 *    IPEN  COLOUR     IPEN  COLOUR     IPEN  COLOUR
 *     0    no pen      1    black       2    red
 *     3    orange      4    yellow      5    green
 *     6    blue        7    violet      8    brown
 *
 *   Argument
*/
int ipen;                /* pen number or colour */
  {
    START_UP( (char VAR_PTR) "t_colour", boolean );
    ipen = ( abs( ipen ) - 1 ) % 8 + 1;
    PLOTTER.requir.colour = ( ipen - 1 ) % PAGE.no_of_pens + 1;
    return 0 <= ipen && ipen <= 8;
  }


static boolean t_dashed( style )
/*
 *      Set the dashed line style . The returned value is TRUE if
 * 0 <= style <= 8 . If these limits are exceeded, the value is wrapped
 * around into the range 1 to 8 .
 *      A value `style == 8' will not produce a `3-dots long-dash' line
 * style ( which is the the device driver standard ) --- very short dashes
 * are used instead.
 *
 *    STYLE   SCREEN
 *          LINE STYLE
 *     0    plot vector endpoints only
 *     1    solid
 *     2    dotted
 *     3    short dash
 *     4    long dash
 *     5    dot dash
 *     6    short-dash long-dash
 *     7    2-short-dashes long-dash
 *     8    very short dashes
 *
*/
int style;        /* dashed line style */
  {
    START_UP( (char VAR_PTR) "t_dashed", boolean );
    PLOTTER.requir.dash = ( abs( style ) - 1 ) % 8 + 1;
    return 0 <= style && style <= 8;
  }


static boolean t_thick( width )
/*
 *      Set the ``width'' of lines and text . The returned value is TRUE
 * if 0 <= thickness.
*/
int width;      /* line thickness */
  {
    START_UP( (char VAR_PTR) "t_thick", boolean );
    PLOTTER.requir.thick = width;
    return 0 <= width;
  }


/* ************************************************************************ */
/*                                                                          */
/*                          Turn graphics on or off                         */
/*                                                                          */
/* ************************************************************************ */


static boolean t_gr_off()
/*
 *      Turn the plotter off if the output is not to disk - so that if
 *  using eavesdrop mode the user can use the terminal. This function fails
 *  if it cannot write to the plotter .
*/
  {
    char VAR_PTR name = "t_gr_off";
    START_UP( name, boolean );
    PLOTTER.requir.mode = OFF_PLOTTER;
    if ( PLOTTER.actual.mode != OFF_PLOTTER )
        if ( ! hp_off() )
            return make_error( ERR_WARN, name, err_write );
    return TRUE;
  }


static boolean t_gr_on()
/*
 *      Turn the plotter on if the output device is ``stdout'' - so that a user
 *  who has used t_gr_off() with eavesdrop mode can resume plotting . This
 *  function fails if it cannot write to ``stdout''.
*/
  {
    char VAR_PTR name = "t_gr_on";
    START_UP( name, boolean );
    PLOTTER.requir.mode = ON_PLOTTER;
    if ( PLOTTER.actual.mode != ON_PLOTTER )
        if ( ! hp_on() )
            return make_error( ERR_WARN, name, err_write );
    return TRUE;
  }



/* ************************************************************************ */
/*                                                                          */
/*           Change Data Level, Visible, Invisible, Revers, Replace         */
/*                                                                          */
/* ************************************************************************ */

#define t_visib t_dummy     /* Draw visible vectors.   */
                            /* - the only thing which the plotter can do */
#define t_invis t_dummy     /* Draw invisible vectors. */
#define t_revers t_dummy    /* Reverse the pixel dots. */
#define t_replac t_dummy    /* Replace pixel contents with the new data. */

static boolean t_dummy()
/*
 *  This function does nothing .
*/
  {
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                         Miscellaneous stuff                              */
/*                                                                          */
/* ************************************************************************ */

static boolean t_newpag()
/*
 * Replace the paper in the plotter . If the plotter has ``autofeed'' facility
 * ( e.g. HP7550 ), this is automatic . For other plotters, if output is not
 * to disk, a message is placed on the user's terminal, and the user must
 * replace the paper manually . If output is to disk and the plotter is not
 * a HP7550, this function is inactive.
*/
  {
    char VAR_PTR name = "t_newpag";
    START_UP( name, boolean );
    if ( SETUP.plotter_id == HP7550A || SETUP.plotter_id == HP7550B )
      {
        if ( hpgl->hanging_cmd )
            hp_terminate();
        (void) G_PUTS( "PG;\n" );
      }
    else if ( ! SETUP.op_todisk )
      {
        if ( PLOTTER.actual.mode == ON_PLOTTER )
            (void) hp_off();
        (void) G_FLUSH();
        (void) fputs( "Replace paper in plotter. Press <RETURN>\n", stdout );
        while( getchar() != '\n' );
      }
    else
        return make_error( ERR_WARN, name, inactive );
    if ( G_ERROR() )
        return make_error( ERR_WARN, name, err_write );
    return TRUE;
  }


static boolean t_newscr()
/*
 * Clear the alphanumeric screen by the simple expedient of writing 50
 * newline characters to the screen. This function is inactive if output
 * is to disk .
*/
  {
    char VAR_PTR name = "t_newscr";
    START_UP( name, boolean );
    if ( ! SETUP.op_todisk )
      {
        int lf_count = 0;
        if ( PLOTTER.actual.mode == ON_PLOTTER )
            (void) hp_off();
        while( lf_count++ <= 50 )
            (void) putchar( '\n' );
        if ( G_ERROR() )
            return make_error( ERR_WARN, name, err_write );
        return TRUE;
      }
    return make_error( ERR_WARN, name, inactive );
  }


static boolean t_bell()
/*
 *      Ring bell . This function is inactive unless the output is not to
 * disk, and it fails if it cannot write to the plotter .
*/
  {
    char VAR_PTR name = "t_bell";
    START_UP( name, boolean );
    if ( ! SETUP.op_todisk )
      {
        if ( PLOTTER.actual.mode == ON_PLOTTER )
            (void) hp_off();
        (void) putchar( BEL );
        if ( G_ERROR() )
            return make_error( ERR_WARN, name, err_write );
        return TRUE;
      }
    return make_error( ERR_WARN, name, inactive );
  }


static boolean t_wait()
/*
 *  Waits for the user to press the carriage return . This function is inactive
 *  if output is to disk, but no error message is generated .
 *  The returned value is TRUE unless it cannot write to the screen .
*/
  {
    char VAR_PTR name = "t_wait";
    START_UP( name, boolean );
    if ( ! SETUP.op_todisk )
      {
        if ( PLOTTER.actual.mode == ON_PLOTTER )
            (void) hp_off();
        (void) G_FLUSH();
        if ( G_ERROR() )
            return make_error( ERR_WARN, name, err_write );
        while( getchar() != '\n' );
      }
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Text Writing                                   */
/*                                                                          */
/* ************************************************************************ */


static boolean t_txtmov( point )
/*
 *    Move the text cursor to coordinate `point' . The returned result is
 *  TRUE only if the next printable character is within the screen limits .
 *
 * Argument
*/
register coord_pair VAR_PTR point;
  {
    char VAR_PTR name = "t_txtmov";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    PLOTTER.requir.cursor = *point;
    return point->_x >  PAGE.ll_limit._x - 0.5 &&
           point->_x <= PAGE.ur_limit._x + 0.5 - PAGE.tty_width &&
           point->_y >  PAGE.ll_limit._y - 0.5 &&
           point->_y <= PAGE.ur_limit._y + 0.5 - SETUP.tty_size;
  }


static boolean t_txtcur( row, column )
/*
 *    Move the text cursor to text row `row' and text column `column' .
 *    The text row is numbered from the top of the screen . The returned
 *    result is TRUE only if the next printable character is within screen
 *    limits . The row and column positions are calculated for the default
 *    alphagraphics character size .
 *
 * Argument
*/
int row, column;
  {
    START_UP( (char VAR_PTR) "t_txtcur", boolean );
    PLOTTER.requir.cursor._x = PAGE.ll_limit._x + PAGE.tty_width * column;
    PLOTTER.requir.cursor._y = PAGE.ur_limit._y - SETUP.tty_size + 1.0 -
        SETUP.tty_size * row;
    return 0 <= column  &&  column <= PAGE.alph_columns  &&
           0 <= row     &&  row    <= PAGE.alph_lines;
  }


static boolean t_txtset( attribute, value )
/*
 *  Set the stroke character attribute `attribute' to `*value' .
 *  If `attribute' is negative, the current value of the attribute is returned
 *  to `*value' . This function fails if the `attribute' cannot be recognized or
 *  `value' is a null pointer . Appropriate error messages are generated .
*/
int attribute;
char VAR_PTR value;
  {
    double VAR_PTR dvalue;    /* pointer to a double `value'              */
    char VAR_PTR name = "t_txtset";
    dvalue = (double VAR_PTR) value;
    if ( attribute != T_CH_RESET  && attribute != T_CH_BOLD  &&
             attribute != T_CH_NOBOLD && attribute != T_ASCII    &&
             attribute != T_SYMBOL )
        CATCH_NULL( value, name, boolean );
    START_UP( name, boolean );
    switch( attribute )
      {
        case  T_CH_RESET :
            (void) reset_font( (Text_Attribute VAR_PTR) &STROKE,
                (double) MAX( PAGE.plot_size._ix, PAGE.plot_size._iy ) / 20.0 );
            if ( SETUP.exact_text )
                STROKE.font_ptr = (StrokeFont VAR_PTR) &asc_font;
            else
                STROKE.font_ptr = (StrokeFont VAR_PTR) &built_font;
            break;
        case  T_CH_HEIGHT :
            if ( STROKE.height != *dvalue )
              {
                STROKE.height = LIMIT( *dvalue, 50.0,
                    MIN( PAGE.plot_size._ix, PAGE.plot_size._iy ) );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_HEIGHT :
            *dvalue = STROKE.height;
            break;
        case  T_CH_ASPECT :
            if ( STROKE.aspect != *dvalue )
              {
                STROKE.aspect = LIMIT( *dvalue, 0.2, 5.0 );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_ASPECT :
            *dvalue = STROKE.aspect;
            break;
        case  T_CH_SLANT :
            if ( STROKE.slant != *dvalue )
              {
                STROKE.slant = LIMIT( *dvalue, -80.0, 80.0 ) * PI / 180.0;
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_SLANT :
            *dvalue = STROKE.slant * 180.0 / PI;
            break;
        case  T_CH_ANGLE :
            if ( STROKE.angle != *dvalue )
              {
                STROKE.angle = PI / 180.0 * ( *dvalue -
                    360.0 * (double) (int) ( *dvalue / 360.0 ) );
                STROKE.angle = STROKE.angle < -PI ? STROKE.angle + 2.0 * PI :
                    ( STROKE.angle > PI ? STROKE.angle-2.0*PI : STROKE.angle );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_ANGLE :
            *dvalue = STROKE.angle * 180.0 / PI;
            break;
        case T_CH_GAP :
            STROKE.gap = LIMIT( *dvalue, 0.05, 5.0 );
            break;
        case -T_CH_GAP :
            *dvalue = STROKE.gap;
            break;
        case  T_CH_ALIGN :
            STROKE.align.vert  = TOLOWER( *value );
            value++;
            STROKE.align.horiz = TOLOWER( *value );
            if ( STROKE.align.horiz != 'l' && STROKE.align.horiz != 'c' &&
                    STROKE.align.horiz != 'r' )
                STROKE.align.horiz = 'l';
            if ( STROKE.align.vert != 'b' && STROKE.align.vert != 'c' &&
                    STROKE.align.vert != 't' )
                STROKE.align.vert = 'b';
            break;
        case -T_CH_ALIGN :
            *value++ = STROKE.align.vert;
            *value++ = STROKE.align.horiz;
            *value   = NUL;
            break;
        case T_ASCII :
            if ( SETUP.exact_text )
              {
                if ( STROKE.font_ptr != (StrokeFont VAR_PTR) &asc_font )
                  {
                    STROKE.font_ptr = (StrokeFont VAR_PTR) &asc_font;
                    STROKE.changed = TRUE;
                  }
              }
            else
              {
                if ( STROKE.font_ptr != (StrokeFont VAR_PTR) &built_font )
                  {
                    STROKE.font_ptr = (StrokeFont VAR_PTR) &built_font;
                    STROKE.changed = TRUE;
                  }
              }
            break;
        case T_SYMBOL :
            if ( STROKE.font_ptr != (StrokeFont VAR_PTR) &sym_font )
              {
                STROKE.font_ptr = (StrokeFont VAR_PTR) &sym_font;
                STROKE.changed = TRUE;
              }
            break;
        case T_CH_BOLD :
            STROKE.bold = TRUE;
            break;
        case T_CH_NOBOLD :
            STROKE.bold = FALSE;
            break;
        default :
            return make_error( ERR_ABORT, name,
                (char VAR_PTR) "Unknown stroke text attribute" );
      }
    return TRUE;
  }


static char VAR_PTR t_ch_def( what, character, index, point )
/*
 * This function allows the user to obtain the shape of any character defined
 * in the ASCII and SYMBOL fonts. With the exception of string alignment, the
 * transformations implied by t_txtset are applied to the returned coordinates.
 * This function is merely a call to the function ``define_font'' with an
 * additional argument defining which set of text attributes to use .
*/
int what;           /* specifies what to do (T_FT_* defined values) */
char character;     /* character to fetch from the font */
int index;          /* coordinate number   */
coord_pair
    VAR_PTR point;  /* returned coordinate */
  {
    START_UP( (char VAR_PTR) "t_ch_def", char  VAR_PTR );
/*
 *      Must switch to stroke based ascii font; builtin font does
 *      not have the required information - it is only a dummy font
*/
    if ( STROKE.font_ptr == &built_font )
      {
        char VAR_PTR ret_value;     /* returned value */
        STROKE.font_ptr = &asc_font;
        ret_value = define_font( (Text_Attribute VAR_PTR) &STROKE,
            (double) PAGE.pen_thick, what, character, index, point );
        STROKE.font_ptr = &built_font;
        return ret_value;
      }
    return define_font( (Text_Attribute VAR_PTR) &STROKE,
        (double) PAGE.pen_thick, what, character, index, point );
  }


static boolean t_symdrw( point, symbol )
/*
 *      draws a centred plotting symbol `symbol' at position `point' according
 *      to the selected stroke character attributes . This function fails if
 *      the font for the symbol is not defined or if it cannot write to the
 *      plotter/file .
*/
coord_pair VAR_PTR point;
char symbol;
  {
    int font_index = symbol - 32;
    char VAR_PTR name = "t_symdrw";
    coord_pair ofset, lower_left;
    StrokeChar VAR_PTR char_ptr;
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
/*
 *              SET UP THE PLOTTER
*/
    if ( hpgl->hanging_cmd )
        hp_terminate();
    if ( PLOTTER.actual.mode != ON_PLOTTER )
        (void) hp_on();
/*
 *      set dashed line style to solid
*/
    if ( PLOTTER.actual.dash != 1 )
        (void) hp_style( SOLID_LINE );
    if ( PLOTTER.actual.colour != PLOTTER.requir.colour )
        (void) hp_pen( PLOTTER.requir.colour );
/*
 *           if necessary read character font
*/
    if ( ! STROKE.font_ptr->font_got )
        if ( ! read_font( STROKE.font_ptr, file_font( STROKE.font_ptr ), name) )
            return FALSE;
/*
 *      CALCULATE LOCATION OF LOWER LEFT CORNER OF THE SYMBOL
 *
 *      determine the character cell size vector
*/
    if ( STROKE.changed )
        (void) map_font( (Text_Attribute VAR_PTR) &STROKE,
            (double) PAGE.pen_thick );
    char_ptr = STROKE.font_ptr->font_dir + font_index;
    ofset._x = STROKE.font_ptr->font_wid * 0.5;
    ofset._y = char_ptr->font_midpt;
/*
 *    determine start position
*/
    lower_left._x = point->_x -
        STROKE.map[0][0] * ofset._x - STROKE.map[0][1] * ofset._y;
    lower_left._y = point->_y -
        STROKE.map[1][0] * ofset._x - STROKE.map[1][1] * ofset._y;
    if ( STROKE.font_ptr != &built_font )
/*
 *    draw the character in the string exactly as specified by the
 *       driver's character font
*/
      {
        if ( 0 <= font_index && font_index < STROKE.font_ptr->font_siz )
            if ( char_ptr->font_char == symbol && char_ptr->font_npts > 0 )
              {
                boolean ok = bold_font( (Text_Attribute VAR_PTR) &STROKE,
                    (coord_pair VAR_PTR) &lower_left,
                    (char VAR_PTR) &font_index, t_stkchr );
                PLOTTER.requir.pen = *point;
                return ok && move_clip( (coord_triple VAR_PTR) point,
                    VIEW_CLIP );
              }
      }
    else if ( ' ' <= symbol && symbol <= '~' )
/*
 *      use the plotter's character font -- ASCII font only
*/
      {
        boolean ok;
        char sym_str[2];
        double step = STROKE.height * STROKE.aspect *
                          ( 1.0 + STROKE.gap / STROKE.aspect );
        sym_str[0] = symbol;
        sym_str[1] = NUL;
        PLOTTER.requir.ch_size._y = STROKE.height;
        PLOTTER.requir.ch_size._x = STROKE.height * STROKE.aspect;
        PLOTTER.requir.ch_step._x = step * cos( STROKE.angle );
        PLOTTER.requir.ch_step._y = step * sin( STROKE.angle );
        PLOTTER.requir.ch_angle   = STROKE.angle;
        PLOTTER.requir.ch_slant   = STROKE.slant;
        ok = bold_font( (Text_Attribute VAR_PTR) &STROKE,
          (coord_pair VAR_PTR) &lower_left, (char VAR_PTR) sym_str, t_txtstr );
        PLOTTER.requir.pen = *point;
        return ok && move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
      }
    return make_error( ERR_ABORT, name, (char VAR_PTR) "Unknown symbol" );
  }


static int t_txtdrw( string )
/*
 *      Draws the text contained in the character string `string' at the
 *   current cursor location according to the selected stroke character
 *   attributes . The returned value is the number of characters drawn . If
 *   an error occurs, the returned value is 0 . An error may be generated
 *   because
 *     a) a character in `string' is unavailable in the specified font
 *     b) the character is not printable
 *     c) could not update the pen position after drawing the characters
 *     d) it cannot write to the plotter
*/
char VAR_PTR string;
  {
    boolean ok = TRUE;          /* success flag                             */
    int font_index,             /* index of the character in font directory */
        n_chars,                /* length of `string'                       */
        char_count=0;           /* counts each character as it is drawn     */
    coord_pair cell_size,       /* size of the character cell               */
               ofset,
               point;           /* location of char cell lower left corner  */
    char VAR_PTR name = "t_txtdrw";
    CATCH_NULL( string, name, boolean );
    START_UP( name, int );
/*
 *              SET UP THE PLOTTER
*/
    if ( hpgl->hanging_cmd )
        hp_terminate();
    if ( PLOTTER.actual.mode != ON_PLOTTER )
        (void) hp_on();
/*
 *      set dashed line style to solid and the required colour
*/
    if ( PLOTTER.actual.dash != 1 )
        (void) hp_style( SOLID_LINE );
    if ( PLOTTER.actual.colour != PLOTTER.requir.colour )
        ok &= hp_pen( PLOTTER.requir.colour );
/*
 *           if necessary read character fonts
*/
    if ( ! STROKE.font_ptr->font_got )
        if ( ! read_font( STROKE.font_ptr, file_font( STROKE.font_ptr ), name) )
            return 0;
    if ( STROKE.changed )
      (void) map_font( (Text_Attribute VAR_PTR) &STROKE,
          (double) PAGE.pen_thick );
/*
 *      CALCULATE LOCATION OF LOWER LEFT CORNER OF THE STRING
 *
 *      determine the character cell size vector
*/
    n_chars = t_strlen( string );
    cell_size._x = STROKE.font_ptr->font_wid *
        ( 1.0 + STROKE.gap / STROKE.aspect );
    cell_size._y = STROKE.font_ptr->font_hei;
    switch( STROKE.align.horiz )
      {
        case 'r' :  ofset._x = cell_size._x * n_chars;
                    break;
        case 'c' :  ofset._x = cell_size._x * n_chars / 2.0;
                    break;
        case 'l' :  ofset._x = 0.0;
                    break;
      }
    switch( STROKE.align.vert )
      {
        case 't' :  ofset._y = cell_size._y;
                    break;
        case 'c' :  ofset._y = cell_size._y / 2.0;
                    break;
        case 'b' :  ofset._y = 0.0;
                    break;
      }
/*
 *    determine start position
*/
    point._x = PLOTTER.requir.cursor._x - STROKE.map[0][0] * ofset._x -
               STROKE.map[0][1] * ofset._y;
    point._y = PLOTTER.requir.cursor._y - STROKE.map[1][0] * ofset._x -
               STROKE.map[1][1] * ofset._y;
    if ( STROKE.font_ptr != &built_font )
      {
        cell_size._y = STROKE.map[1][0] * cell_size._x;
        cell_size._x = STROKE.map[0][0] * cell_size._x;
/*
 *    draw each character in the string exactly as specified by the
 *       driver's character font
*/
        for( ; *string; string++, char_count++ )
          {
            font_index = *string - 32;
            if ( 0 <= font_index && font_index < STROKE.font_ptr->font_siz )
                if ( (STROKE.font_ptr->font_dir)[font_index].font_char ==
                         *string &&
                     (STROKE.font_ptr->font_dir)[font_index].font_npts > 0 )
                  {
                    ok &= bold_font( (Text_Attribute VAR_PTR) &STROKE,
                        (coord_pair VAR_PTR) &point,
                        (char VAR_PTR) &font_index, t_stkchr );
                    point._x += cell_size._x;
                    point._y += cell_size._y;
                  }
                else
                    (void) make_error( ERR_WARN, name,
                        (char VAR_PTR) "character unavailable" );
             else
                 (void) make_error( ERR_WARN, name,
                     (char VAR_PTR) "not printable character" );
          }
      }
    else
/*
 *      use the plotter's character font
*/
      {
        double step = STROKE.height * ( STROKE.aspect + STROKE.gap );
        PLOTTER.requir.ch_size._y = STROKE.height;
        PLOTTER.requir.ch_size._x = STROKE.height * STROKE.aspect;
        PLOTTER.requir.ch_step._x = step * cos( STROKE.angle );
        PLOTTER.requir.ch_step._y = step * sin( STROKE.angle );
        PLOTTER.requir.ch_angle   = STROKE.angle;
        PLOTTER.requir.ch_slant   = STROKE.slant;
        return bold_font( (Text_Attribute VAR_PTR) &STROKE,
            (coord_pair VAR_PTR) &point, string, t_txtstr );
      }
    PLOTTER.requir.cursor = point;
    if ( ! ( ok &= move_clip( (coord_triple VAR_PTR) &PLOTTER.requir.pen,
            VIEW_CLIP ) ) )
        (void) make_error( ERR_CALL, name, mt_str );
    return char_count;
  }


static char VAR_PTR t_txtget( string, max_chars )
/*
 *    Reads a maximum of `max_chars' characters from `stdin' into `string'
 * until a newline character is read . The newline character is not stored in
 * `string' . `string' is terminated with NUL . The characters are echoed to
 * the plotter . The echoed characters are confined to the viewport . If the
 * cursor is initially outside the viewport, t_txtget() starts echoing
 * characters at the top left corner of the viewport. T_txtget will not echo
 * any more characters if it runs out of space at the bottom of the viewport .
 * If an error errors or no characters are read, a NULL pointer is returned,
 * otherwise `string' is returned .
 *
 * Programming notes.
 *  t_txtget only changes plotter.requir.cursor when the location of the start
 *  of the buffered string must be moved .
*/
char VAR_PTR string;
int max_chars;
  {
    char VAR_PTR ptr;
    char VAR_PTR name = "t_txtget";
    CATCH_NULL( string, name, char VAR_PTR );
    START_UP( name, char VAR_PTR );
    if ( PLOTTER.actual.mode != OFF_PLOTTER )
        (void) hp_off();
    for( ptr = string; ( *ptr = getchar() ) != '\n' && max_chars--; ptr++ );
    *ptr = NUL;
    if ( ptr != string )
      {
/*
 *              ensure cursor is in the viewport
 *              ( place it in top left corner if it is not )
*/
        coord_pair max_curs;
        max_curs._x = max_clip( VIEW_CLIP, _x ) - PAGE.tty_width + 1;
        max_curs._y = max_clip( VIEW_CLIP, _y ) - SETUP.tty_size + 1;
        if ( PLOTTER.requir.cursor._x < min_clip( VIEW_CLIP, _x ) ||
             PLOTTER.requir.cursor._x > max_curs._x ||
             PLOTTER.requir.cursor._y < min_clip( VIEW_CLIP, _y ) ||
             PLOTTER.requir.cursor._y > max_curs._y )
          {
            PLOTTER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
            PLOTTER.requir.cursor._y = max_curs._y;
          }
/*
 *              place characters in a buffer until it runs out of the viewport
*/
        if ( PLOTTER.requir.cursor._x <= max_curs._x &&
             PLOTTER.requir.cursor._y >= min_clip( VIEW_CLIP, _x ) )
          {
            char buffer[BUFFER_LEN],
                 * buffer_ptr = buffer;
            int cursor_col = PLOTTER.requir.cursor._x;
            PLOTTER.requir.ch_size._x = PAGE.tty_width /
                ( 1.0 + built_font.font_gap / built_font.font_wid );
            PLOTTER.requir.ch_size._y = SETUP.tty_size /
                ( 1.0 + built_font.font_gap / built_font.font_wid );
            PLOTTER.requir.ch_step._x = PAGE.tty_width;
            PLOTTER.requir.ch_step._y = PLOTTER.requir.ch_angle =
                                        PLOTTER.requir.ch_slant = 0.0;
            for( ptr = string; *ptr; ptr++ )
                if ( ' ' <= *ptr && *ptr <= '~' &&
                    PLOTTER.requir.cursor._y >= min_clip( VIEW_CLIP, _y ) &&
                    PLOTTER.requir.cursor._y <= max_curs._y )
                  {
                    *buffer_ptr++ = *ptr;
/*
 *                      print the buffer if the right hand margin is reached
 *                      or if the buffer is full
*/
                    cursor_col += (int) PAGE.tty_width;
                    if ( cursor_col > max_curs._x )
                      {
                        *buffer_ptr = NUL;
                        (void) tty_str( (char VAR_PTR) buffer, buffer_ptr );
                        buffer_ptr = buffer;
                        PLOTTER.requir.cursor._x = cursor_col
                                                 = min_clip( VIEW_CLIP,_x );
                        PLOTTER.requir.cursor._y -= SETUP.tty_size;
                      }
                    else if ( buffer_ptr - buffer >= BUFFER_LEN )
                      {
                        *buffer_ptr = NUL;
                        (void) tty_str( (char VAR_PTR) buffer, buffer_ptr );
                        buffer_ptr = buffer;
                      }
                  }
/*
 *                      print whatever remains in the buffer
*/
            if ( buffer_ptr != buffer )
              {
                *buffer_ptr = NUL;
                (void) tty_str( (char VAR_PTR) buffer, buffer_ptr );
              }
          }
        PLOTTER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
        PLOTTER.requir.cursor._y -= SETUP.tty_size;
      }
    return *string ? string : (char VAR_PTR) 0;
  }


static int t_txtput( string )
/*
 *    Print text string `string' on the plotter, starting at the place
 *  last specified by `t_txtmov' or 't_txtcur', using the colour specified by
 *  `t_colour' . Only those characters within the viewport are written to the
 *  screen. The returned value is the number of characters written to the
 *  screen, or zero if an error has occurred .
 *
 *  The character string `string' may contain printable characters and the
 *  control codes,
 *    HT : \t : \011 : move to the right by one character space
 *    BS : \b : \010 : move to the left by one character space
 *    CR : \r : \015 : move to the beginning of the current line
 *    LF : \n : \012 : move down one character space
 *    VT :    : \013 : move up one character space
 *    EM :    : \031 : move to top left hand side of the screen
 *  Any other characters are ignored .
 *
 *  This function only breaks the string into segments terminated by cursor
 *  control characters and positions the required_cursor appropriately before
 *  calling t_txtstr to put the characters on the screen.
 *
 * Programming notes,
 *  t_txtput only changes plotter.requir.cursor when the location of the start
 *  of the buffered string must be moved .
*/
char VAR_PTR string;    /* pointer to start of the character string */
  {
    char VAR_PTR name = "t_txtput";
    char VAR_PTR ptr;
    int  char_count = 0;     /* number of characters printed */
    CATCH_NULL( string, name, boolean );
    START_UP( name, int );
    PLOTTER.requir.ch_size._x = PAGE.tty_width /
        ( 1.0 + built_font.font_gap / built_font.font_wid );
    PLOTTER.requir.ch_size._y = SETUP.tty_size /
        ( 1.0 + built_font.font_gap / built_font.font_wid );
    PLOTTER.requir.ch_step._x = PAGE.tty_width;
    PLOTTER.requir.ch_step._y = PLOTTER.requir.ch_angle =
        PLOTTER.requir.ch_slant = 0.0;

    for ( ptr = string; *ptr; ptr++ )
      {
        if ( *ptr == LF || *ptr == VT || *ptr == EM ||
             *ptr == CR || *ptr == BS || *ptr == HT )
          {
            char_count += tty_str( string, ptr ) + 1;
            switch( *ptr )
              {
                case LF : PLOTTER.requir.cursor._y += -SETUP.tty_size;
                          break;
                case VT : PLOTTER.requir.cursor._y += SETUP.tty_size;
                          break;
                case EM : PLOTTER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
                          PLOTTER.requir.cursor._y = max_clip( VIEW_CLIP, _y )
                              - SETUP.tty_size + 1;
                          break;
                case CR : PLOTTER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
                          break;
                case BS : PLOTTER.requir.cursor._x += - PAGE.tty_width;
                          break;
                case HT : PLOTTER.requir.cursor._x += PAGE.tty_width;
                          break;
              }
            string = ptr + 1;
          }
      }
    if ( string != ptr )
        char_count += tty_str( string, ptr );
    return char_count;
  }


static boolean t_boxfil( lower_left, upper_right, style )
/*
 *      draw a rectangular box on the screen with lower left corner at
 * `*lower_left' and upper right corner at `*upper_right', and fill the box
 * with a hatching of type `style' . The hatching styles are :
 *
 *      0 = empty, box outline only      5 = horizontal lines
 *      1 = fill the box completely      6 = vertical lines
 *      2 = 50% `grey' fill              7 = diagonal cross hatch
 *      3 = NW-SE diagonal lines         8 = horizontal-vertical cross
 *      4 = NE-SW diagonal lines                 hatch
 * The box outline is clipped to the viewport limits .
*/
coord_pair VAR_PTR lower_left,  /* lower left corner of the box  */
           VAR_PTR upper_right; /* upper right corner of the box */
int style;                      /* hatching style                */
  {
    char VAR_PTR name = "t_boxfil";
    coord_pair ll_corner, ur_corner;
    CATCH_NULL( lower_left && upper_right, name, boolean );
    START_UP( name, boolean );
/*
 *      clip the rectangle to the screen limits
*/
    ll_corner._x = LIMIT( lower_left->_x,
                     min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
    ll_corner._y = LIMIT( lower_left->_y,
                     min_clip( VIEW_CLIP, _y ), max_clip( VIEW_CLIP, _y ) );
    ur_corner._x = LIMIT( upper_right->_x,
                     min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
    ur_corner._y = LIMIT( upper_right->_y,
                     min_clip( VIEW_CLIP, _y ), max_clip( VIEW_CLIP, _y ) );
/*
 *      ensure correct ordering of coordinates
*/
    if ( ll_corner._x > ur_corner._x )
        SWAPD( ll_corner._x , ur_corner._x );
    if ( ll_corner._y > ur_corner._y )
        SWAPD( ll_corner._y , ur_corner._y );
/*
 *      draw the box
*/
    if ( fabs( ur_corner._x - ll_corner._x ) > 1.0 &&
         fabs( ur_corner._y - ll_corner._y ) > 1.0 )
      {
        boolean    ok = TRUE;
        int_coord  ll_p_units, ur_p_units;
        static int rotated_style[] = { 0, 1, 2, 4, 3, 6, 5, 7, 8 };
        if ( PLOTTER.actual.mode != ON_PLOTTER )
            ok &= hp_on();
        if ( PLOTTER.actual.colour != PLOTTER.requir.colour )
            ok &= hp_pen( PLOTTER.requir.colour );
        if (  PLOTTER.actual.dash != 1 )
            ok &= hp_style( 1 );
        style = ( abs( style ) - 1 ) % 8 + 1;
        ok &= hp_box(
            plot_units( (coord_pair VAR_PTR) &ll_corner,
                (int_coord VAR_PTR) &ll_p_units ),
            plot_units( (coord_pair VAR_PTR) &ur_corner,
                (int_coord VAR_PTR) &ur_p_units ),
            PAGE.paper_rota ? rotated_style[style] : style );
        if ( ! ok )
            (void) make_error( ERR_WARN, name, err_write );
        PLOTTER.actual.cursor = PLOTTER.actual.pen = ll_corner;
        return ok;
      }
    return FALSE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Cross Hairs Routines                           */
/*                                                                          */
/* ************************************************************************ */


static boolean t_xhrget( point, char_pressed )
/*
 * Is inactive. Does nothing .
*/
coord_pair VAR_PTR point;   /* pointer to the coordinates of the cross-hairs */
char VAR_PTR char_pressed;  /* character pressed to identify `point'         */
 {
    return make_error( ERR_WARN, (char VAR_PTR) "t_xhrget", inactive );
 }


static boolean t_xhrput( point )
/*
 * Is inactive. Does nothing .
*/
coord_pair VAR_PTR point;  /* pointer to the coordinates of the cross-hairs */
  {
    return make_error( ERR_WARN, (char VAR_PTR) "t_xhrput", inactive );
  }


/* ************************************************************************ */
/*                                                                          */
/*                        Print terminal status                             */
/*                                                                          */
/* ************************************************************************ */


static char VAR_PTR t_status( attribute, value )
/*
 *  Put status attribute 'attribute' into '*value' which may be a character
 *  string or a coordinate pair .
*/
int attribute;
char VAR_PTR value;
  {
    char VAR_PTR name = "t_status";
    CATCH_NULL( value, name, char VAR_PTR );
    if ( attribute != T_ST_SCREEN && attribute != T_ST_CATEG )
        START_UP( name, char VAR_PTR );
    switch( attribute )
      {
        case T_ST_DEVICE:
            (void) t_strcpy( value, ID_string[SETUP.plotter_id] );
            value[ID_STRLEN] = ' ';  /* replace '\n' with a space */
            (void) t_strcpy( value + ID_STRLEN + 1, "Hewlett/Packard plotter" );
            break;
        case T_ST_SCREEN:
            *( (boolean VAR_PTR) value ) = FALSE;
            break;
        case T_ST_CATEG:
            *( (int VAR_PTR) value ) = T_ST_FILE;
            break;
        case T_ST_MONO:
            *( (boolean VAR_PTR) value ) = FALSE;
            break;
        case T_ST_MAX:
            ( (coord_pair VAR_PTR) value)->_x = PAGE.ur_limit._x;
            ( (coord_pair VAR_PTR) value)->_y = PAGE.ur_limit._y;
            break;
        case T_ST_MIN:
            ( (coord_pair VAR_PTR) value)->_x = PAGE.ll_limit._x;
            ( (coord_pair VAR_PTR) value)->_y = PAGE.ll_limit._y;
            break;
        case T_ST_CHAR:
            ( (coord_pair VAR_PTR) value)->_y = SETUP.tty_size;
            ( (coord_pair VAR_PTR) value)->_x = PAGE.tty_width;
            break;
        case T_ST_LINES:
            ( (coord_pair VAR_PTR) value)->_x = PAGE.alph_columns;
            ( (coord_pair VAR_PTR) value)->_y = PAGE.alph_lines;
            break;
        case T_ST_PIXSIZ:
            ( (coord_pair VAR_PTR) value)->_x =
                ( (coord_pair VAR_PTR) value)->_y =
                resolution[SETUP.plotter_id];
            break;
        case T_ST_A_CURS:
            ( (coord_pair VAR_PTR) value)->_x = PLOTTER.actual.cursor._x;
            ( (coord_pair VAR_PTR) value)->_y = PLOTTER.actual.cursor._y;
            break;
        case T_ST_R_CURS:
            ( (coord_pair VAR_PTR) value)->_x = PLOTTER.requir.cursor._x;
            ( (coord_pair VAR_PTR) value)->_y = PLOTTER.requir.cursor._y;
            break;
        case T_ST_A_PEN:
            ( (coord_pair VAR_PTR) value)->_x = PLOTTER.actual.pen._x;
            ( (coord_pair VAR_PTR) value)->_y = PLOTTER.actual.pen._y;
            break;
        case T_ST_R_PEN:
            ( (coord_pair VAR_PTR) value)->_x = PLOTTER.requir.pen._x;
            ( (coord_pair VAR_PTR) value)->_y = PLOTTER.requir.pen._y;
            break;
        default:
            return (char VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "unknown attribute" );
      }
    return value;
  }


static boolean t_pr_ter( out_file )
FILE *out_file;
  {
    char VAR_PTR name = "t_pr_ter";
    CATCH_NULL( out_file, name, boolean );
    START_UP( name, boolean );
    (void) fprintf( out_file, " TERMINAL STATUS\n" );
    (void) fprintf( out_file,
                    "               ---requir---     ---actual---\n" );
    (void) fprintf( out_file, "mode        %15s  %15s\n",
                    pr_mode( (Plot_Status VAR_PTR) &PLOTTER.requir ),
                    pr_mode( (Plot_Status VAR_PTR) &PLOTTER.actual ) );
    (void) fprintf( out_file, "colour      %15d  %15d\n",
                    PLOTTER.requir.colour, PLOTTER.actual.colour );
    (void) fprintf( out_file, "dash style  %15d  %15d\n",
                    PLOTTER.requir.dash, PLOTTER.actual.dash );
    (void) fprintf( out_file, "fill type  %15d  %15d\n",
                    PLOTTER.requir.fill_type, PLOTTER.actual.fill_type );
    (void) fprintf( out_file, "cursor      (%6.0f,%6.0f)  (%6.0f,%6.0f)\n",
                    PLOTTER.requir.cursor._x, PLOTTER.requir.cursor._y,
                    PLOTTER.actual.cursor._x, PLOTTER.actual.cursor._y );
    (void) fprintf( out_file, "pen         (%6.0f,%6.0f)  (%6.0f,%6.0f)\n",
                    PLOTTER.requir.pen._x, PLOTTER.requir.pen._y,
                    PLOTTER.actual.pen._x, PLOTTER.actual.pen._y );
    (void) fprintf( out_file, "char. size  (%6.0f,%6.0f)  (%6.0f,%6.0f)\n",
                    PLOTTER.requir.ch_size._x, PLOTTER.requir.ch_size._y,
                    PLOTTER.actual.ch_size._x, PLOTTER.actual.ch_size._y );
    (void) fprintf( out_file, "char. angle %15f  %15f\n",
                    PLOTTER.requir.ch_angle, PLOTTER.actual.ch_angle );
    (void) fprintf( out_file, "char. slant %15f  %15f\n",
                    PLOTTER.requir.ch_slant, PLOTTER.actual.ch_slant );
    (void) fprintf( out_file, "char. step  (%6.0f,%6.0f)  (%6.0f,%6.0f)\n",
                    PLOTTER.requir.ch_step._x, PLOTTER.requir.ch_step._y,
                    PLOTTER.actual.ch_step._x, PLOTTER.actual.ch_step._y );
    if ( ferror( out_file ) )
        return make_error( ERR_WARN, name,
            (char VAR_PTR) "Cannot write terminal status to file" );
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Vector Drawing                                 */
/*                                                                          */
/* ************************************************************************ */


static boolean t_vecmov( point )
/*
 *      move the line drawing pen to `point'
*/
coord_pair VAR_PTR point;
  {
    char VAR_PTR name = "t_vecmov";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    PLOTTER.requir.pen = *point;
    return move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
  }


static boolean t_vecdrw( point )
/*
 *      Draw from the previous pen position to `point' according to the
 *  required pen colour, dashed line style and dots-level . Dashed line style
 *  takes precedence over colours . Lines are clipped to the viewport
 *  boundaries .
*/
coord_pair VAR_PTR point;
  {
    boolean ok = TRUE;     /* success flag */
    int_coord hp_coord;
    char VAR_PTR name = "t_vecdrw";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    PLOTTER.requir.pen = *point;
/*
 *          If required dashed line style is 0, plot vector endpoint onlyput terminal into POINT_PLOT
 *          mode, and plot the point if it is in the viewport .
*/
    if ( ! PLOTTER.requir.dash )
      {
        ok &= move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
        if ( PLOTTER.requir.colour &&
             min_clip( VIEW_CLIP, _x ) <= point->_x &&
             max_clip( VIEW_CLIP, _x ) >= point->_x &&
             min_clip( VIEW_CLIP, _y ) <= point->_y &&
             max_clip( VIEW_CLIP, _y ) >= point->_y )
          {
/*
 *          If necessary, switch the plotter on and select the correct dashed
 *          line style and pen colour .
*/
            if ( PLOTTER.actual.mode != ON_PLOTTER )
                ok &= hp_on();
            if ( PLOTTER.actual.colour != PLOTTER.requir.colour )
                ok &= hp_pen( PLOTTER.requir.colour );
            if ( PLOTTER.actual.dash != PLOTTER.requir.dash )
                ok &= hp_style( PLOTTER.requir.dash );
/*
 *          plot the point
*/
            if ( ok &= hp_draw( plot_units( point,
                    (int_coord VAR_PTR) &hp_coord ) ) )
                PLOTTER.actual.cursor = PLOTTER.actual.pen = *point;
          }
      }
/*
 *          Need to draw a line . First clip the endpoints .
*/
    else if ( clip( (coord_triple VAR_PTR) point, VIEW_CLIP ) )
      {
/*
 *          If necessary, switch the plotter on and select the correct dashed
 *          line style and pen colour .
*/
            if ( PLOTTER.actual.mode != ON_PLOTTER )
                ok &= hp_on();
            if ( PLOTTER.actual.colour != PLOTTER.requir.colour )
                ok &= hp_pen( PLOTTER.requir.colour );
            if ( PLOTTER.actual.dash != PLOTTER.requir.dash )
                ok &= hp_style( PLOTTER.requir.dash );
/*
 *          Now we are all set up, we can worry about drawing the vector
 *          First check if the pen is starting from the correct place .
*/
        if ( fabs( PLOTTER.actual.pen._x - first_clip(VIEW_CLIP,_x) ) > 0.5 ||
             fabs( PLOTTER.actual.pen._y - first_clip(VIEW_CLIP,_y) ) > 0.5 )
            ok &= hp_move( plot_units(
                (coord_pair VAR_PTR) FIRST_clip( VIEW_CLIP ),
                (int_coord VAR_PTR) &hp_coord ) );
/*
 *          send vector to the plotter
*/
        if ( ok &= hp_draw( plot_units(
              (coord_pair VAR_PTR) LAST_clip( VIEW_CLIP ),
              (int_coord VAR_PTR) &hp_coord ) ) )
          {
            PLOTTER.actual.cursor._x = PLOTTER.actual.pen._x =
                last_clip( VIEW_CLIP, _x );
            PLOTTER.actual.cursor._y = PLOTTER.actual.pen._y =
                last_clip( VIEW_CLIP, _y );
              }
      }
    if ( ! ok )
        (void) make_error( ERR_WARN, name, err_write );
    return ok;
  }


/* ************************************************************************ */
/* ************************************************************************ */
/*                                                                          */
/*                    Array of pointers to driver functions                 */
/*                                                                          */
/* With the exception of the function ``t_driver'' which is used to select  */
/* the required driver, this array is the ONLY point of access to the       */
/* functions in the driver .                                                */
/*                                                                          */
/* ************************************************************************ */
/* ************************************************************************ */

/*
 * static clip_area VAR_PTR t_begin();
 * static boolean
 *     t_end(),    t_vwport(), t_colour(), t_dashed(), t_thick(),  t_gr_off(),
 *     t_gr_on(),  t_pr_ter(), t_dump(),   t_visib(),  t_invis(),  t_revers(),
 *     t_replac(), t_newpag(), t_newscr(), t_bell(),   t_wait(),   t_txtcur(),
 *     t_txtmov(), t_txtset(), t_symdrw(), t_xhrget(), t_xhrput(), t_vecdrw(),
 *     t_boxfil(), t_vecmov();
 * static char
 *         VAR_PTR t_txtget(), VAR_PTR t_status(), VAR_PTR t_ch_def();
 * static int t_txtput(), t_txtdrw();
 */

T_dev_fns HPgl_dev_ptrs = {
            t_begin,  t_end,    t_vwport, t_gr_on,  t_gr_off,
            t_colour, t_dashed, t_thick,  t_visib,  t_invis,
            t_revers, t_replac, t_newpag, t_newscr, t_bell,
            t_wait,   t_txtmov, t_txtcur, t_txtset, t_symdrw,
            t_txtdrw, t_txtget, t_txtput, t_status, t_pr_ter,
            t_vecmov, t_vecdrw, t_dump,   t_xhrput, t_xhrget,
            t_boxfil, t_ch_def };

/* *********************************************************************** */
/*                                                                         */
/*                      Utilities for hpgl_g.c                             */
/*          These functions cannot be called directly by the user          */
/*                                                                         */
/* *********************************************************************** */


static P1P2_elem VAR_PTR find_P1P2( P1P2_table, table_size )
/*
 *      Find the entry in the paper size table which matches the setup paper
 *      type and size .
*/
P1P2_elem VAR_PTR P1P2_table;
int table_size;
  {
    int counter;
    for( counter = 0;
        counter++ < table_size && ( P1P2_table->Pp_type != SETUP.paper_type ||
                                    P1P2_table->Pp_size != SETUP.paper_size );
        P1P2_table++ );
    return counter <= table_size ? P1P2_table : (P1P2_elem VAR_PTR) 0;
  }


static P1P2_elem VAR_PTR compare_P1P2( OW_limit, P1P2_table, table_size,
    tolerance )
/*
 *      Find the entry in the paper size table which matches the paper size
 * information (ie. window coordinates) obtained from the plotter -- and then
 * set then paper margins .
*/
P1P2_elem VAR_PTR OW_limit, VAR_PTR P1P2_table;
int table_size, tolerance;
  {
    int counter;
    int_coord ll_dif, ur_dif;
    for( counter = 0;
         counter++ < table_size &&
       ( abs( ll_dif._ix = OW_limit->P1_ll._ix - P1P2_table->P1_ll._ix ) >
             tolerance ||
         abs( ll_dif._iy = OW_limit->P1_ll._iy - P1P2_table->P1_ll._iy ) >
             tolerance ||
         abs( ur_dif._ix = OW_limit->P2_ur._ix - P1P2_table->P2_ur._ix ) >
             tolerance ||
         abs( ur_dif._iy = OW_limit->P2_ur._iy - P1P2_table->P2_ur._iy ) >
             tolerance );
        P1P2_table++ );
    if ( counter <= table_size )
      {
        OW_limit->Pll_marg._ix = P1P2_table->Pll_marg._ix + ll_dif._ix;
        OW_limit->Pll_marg._iy = P1P2_table->Pll_marg._iy + ll_dif._iy;
        OW_limit->Pur_marg._ix = P1P2_table->Pur_marg._ix + ur_dif._ix;
        OW_limit->Pur_marg._iy = P1P2_table->Pur_marg._iy + ur_dif._iy;
        OW_limit->Pp_type = P1P2_table->Pp_type;
        OW_limit->Pp_size = P1P2_table->Pp_size;
        return OW_limit;
      }
    return (P1P2_elem VAR_PTR) 0;
  }


static boolean find_pen_type( caller )
/*
 *      Asks the plotter what kind of pens it has . The "OT;" has already
 *      been sent to the plotter by t_begin() .
*/
char VAR_PTR caller;        /* name of calling function */
  {
    short int pen_mask;
    (void) G_PUTS( "OT;" );
    (void) G_FLUSH();
    (void) fscanf( stdin, "%hd,%hd%*c", &SETUP.pen_type, &pen_mask );
    if ( SETUP.pen_type > 0 )
      {
        if ( pen_mask != ( 1 << PAGE.no_of_pens ) - 1 )
            (void) make_error( ERR_WARN, caller,
                (char VAR_PTR) "Some pens are missing" );
        return TRUE;
      }
    return make_error( ERR_ABORT, caller,
        (char VAR_PTR) "Missing or unknown carousel" );
  }


static int t_stkchr( index, end_nul, point )
/*
 *      Draws character identified by `index', at location `point' .
 *      This is a single pass -- i.e. no boldface.
*/
char VAR_PTR index,
     VAR_PTR end_nul;      /* phantom end of string pointer - required */
                           /* compatibility with bold_font & t_txtstr  */
coord_pair VAR_PTR point;
  {
    int point_count;        /* counts the points plotted for each char*/
    boolean ok = TRUE,      /* success flag                           */
            pen_up;         /* move with pen up                       */
    float VAR_PTR coord_ptr;/* pointer to font coordinate values      */
    coord_pair coord;       /* coordinate from the font - to plot     */
    int_coord int_point;    /* "plotter" units point set to plotter   */
    double temp;
    StrokeChar VAR_PTR char_ptr =
        STROKE.font_ptr->font_dir + *( (int VAR_PTR) index );
/*
 *      step through the coordinates which define the character
*/
    coord_ptr =  char_ptr->font_coord;
    for( point_count = 0;
         ok && point_count < char_ptr->font_npts;
         point_count++ )
      {
/*
 *      calculate stroke coordinate position in pixel units
*/
        pen_up = ! point_count;
        coord._x = *coord_ptr++;
        coord._y = *coord_ptr++;
        if ( coord._x < 0.0 )
          {
            pen_up = TRUE;
            coord._x = -coord._x;
          }
        if ( coord._y < 0.0 )
          {
            pen_up = TRUE;
            coord._y = -coord._y;
          }
        coord._y -= char_ptr->font_shift;
        temp = STROKE.map[0][0] * coord._x + STROKE.map[0][1] * coord._y
                + point->_x;
        coord._y = STROKE.map[1][0] * coord._x + STROKE.map[1][1] * coord._y
                + point->_y;
        coord._x = temp;
/*
 *      plot the point
*/
        if ( pen_up )
            ok = move_clip( (coord_triple VAR_PTR) &coord, VIEW_CLIP );
        else if ( clip( (coord_triple VAR_PTR) &coord, VIEW_CLIP ) )
          {
/*
 *      only `move' if the vector does not start from the end of the last one .
*/
            if ( fabs( PLOTTER.actual.pen._x - first_clip( VIEW_CLIP, _x ) )
                    > 0.5 || 0.5 <
                 fabs( PLOTTER.actual.pen._y - first_clip( VIEW_CLIP, _y ) ) )
                ok = hp_move( plot_units(
                    (coord_pair VAR_PTR) FIRST_clip( VIEW_CLIP ),
                    (int_coord VAR_PTR) &int_point ) );
            if ( ok &= hp_draw( plot_units(
                    (coord_pair VAR_PTR) LAST_clip( VIEW_CLIP ),
                    (int_coord VAR_PTR) &int_point ) ) )
              {
                PLOTTER.actual.cursor._x = PLOTTER.actual.pen._x =
                    last_clip( VIEW_CLIP, _x );
                PLOTTER.actual.cursor._y = PLOTTER.actual.pen._y =
                    last_clip( VIEW_CLIP, _y );
              }
          }
      }
    if ( ! ok )
      (void) make_error( ERR_WARN, (char VAR_PTR) "t_stkchr", err_write );
    return ok;
  }


static int tty_str( string, end_nul )
/*
 * This is simply calls ``t_txtstr()'' for the purpose of writing teletype
 * text ``string'' . The text location is the required cursor location
 * displaced upwards by 0.26 character cell heights .
*/
char VAR_PTR string,
     VAR_PTR end_nul;
  {
    int count; /* number of characters printed- returned value from t_txtstr */
    double displ = 0.26 * SETUP.tty_size;
    PLOTTER.requir.cursor._y += displ;
    count = t_txtstr( string, end_nul,
                      (coord_pair VAR_PTR) &PLOTTER.requir.cursor );
    PLOTTER.requir.cursor._y -= displ;
    return count;
  }


static int t_txtstr( string, end_nul, point )
/*
 * Prints a label of printable ASCII , starting at the current required cursor
 * postion and using the current colour . Only characters within the viewport
 * are written to the screen . The returned value is the number of characters
 * written.
 * The character string `string' may be NUL terminated or, if only a fragment
 * of the string is to be printed, `end_nul' is a pointer to where the NUL
 * should be.
 *
*/
char VAR_PTR string,    /* start address of character string to be printed */
     VAR_PTR end_nul;   /* address of phantom end of string NUL */
coord_pair VAR_PTR point;
  {
    boolean ok = TRUE,              /* success flag */
            label_cmd = FALSE,      /* label command has been sent */
            not_setup = TRUE;       /* text parameters have not been sent */
    int char_count = 0;             /* number of characters sent */
    double vw_marg = MAX( PLOTTER.actual.ch_size._x,
        PLOTTER.actual.ch_size._y ); /* excess viewport size */
    PLOTTER.requir.cursor = *point;
    for( ; *string && string != end_nul && ok; string++ )
      {
        if ( ' ' <= *string && *string <= '~' )
          {
/*
 ********     write the character if it is in the viewport
*/
            if ( PLOTTER.requir.cursor._x >= min_clip(VIEW_CLIP,_x)- vw_marg &&
                 PLOTTER.requir.cursor._x <= max_clip(VIEW_CLIP,_x)+ vw_marg &&
                 PLOTTER.requir.cursor._y >= min_clip(VIEW_CLIP,_y)- vw_marg &&
                 PLOTTER.requir.cursor._y <= max_clip(VIEW_CLIP,_y)+ vw_marg )
              {
/*
 *          place the cursor in the correct place
*/
                if ( fabs( PLOTTER.requir.cursor._x-PLOTTER.actual.cursor._x )
                         > TXT_JITTER || TXT_JITTER <
                   fabs( PLOTTER.requir.cursor._y-PLOTTER.actual.cursor._y ) )
                  {
                    int_coord int_point;
                    if ( PLOTTER.actual.mode != ON_PLOTTER )
                        ok &= hp_on();
                    if ( hpgl->hanging_cmd )
                        hp_terminate();
                    if ( label_cmd )
                      {
                        (void) G_PUTC( ETX );
                        (void) G_PUTC( '\n' );
                      }
                    label_cmd = FALSE;
                    if( ok &= hp_move( plot_units(
                          (coord_pair VAR_PTR) &PLOTTER.requir.cursor,
                          (int_coord VAR_PTR) &int_point ) ) )
                        PLOTTER.actual.cursor = PLOTTER.requir.cursor;
                  }
/*
 *          send text parameters to the plotter
*/
                if ( ! label_cmd )
                  {
                    if ( not_setup )
                      {
                        txt_params();
                        not_setup = FALSE;
                      }
                    (void) G_PUTS( "LB" );
                    label_cmd = TRUE;
                  }
                char_count++;
                (void) G_PUTC( *string );
                PLOTTER.actual.cursor._x += PLOTTER.actual.ch_step._x;
                PLOTTER.actual.cursor._y += PLOTTER.actual.ch_step._y;
              }
            PLOTTER.requir.cursor._x += PLOTTER.requir.ch_step._x;
            PLOTTER.requir.cursor._y += PLOTTER.requir.ch_step._y;
          }
      }
    if ( label_cmd )
      {
        (void) G_PUTC( ETX );
        (void) G_PUTC( '\n' );
      }
    PLOTTER.actual.pen = PLOTTER.actual.cursor;
    if ( ! ok )
        (void) make_error( ERR_WARN, (char VAR_PTR) "t_txtstr", err_write );
    return ok ? char_count : 0;
  }

#define HP_TINY       0.0001  /* smallest "sendable" decimal number       */

static void txt_params()
/*
 *  Ensure the plotter is set up to write labels in the correct manner .
 *  This requires setting pen colour, text slant, text direction, character
 *  spacing and character size .
*/
  {
    boolean step_changed = FALSE;
    if ( PLOTTER.actual.mode != ON_PLOTTER )
      (void) hp_on();
    if ( hpgl->hanging_cmd )
      hp_terminate();
/*
 *      colour
*/
    if ( PLOTTER.actual.colour != PLOTTER.requir.colour )
        (void) hp_pen( PLOTTER.requir.colour );
/*
 *      text slant
*/
    if ( fabs( PLOTTER.actual.ch_slant - PLOTTER.requir.ch_slant ) > HP_TINY )
      {
        if ( fabs( PLOTTER.requir.ch_slant ) > HP_TINY )
            (void) fprintf( OP_FILE, "SL%7.4f;\n", PLOTTER.requir.ch_slant );
        else
            (void) G_PUTS( "SL;\n" );
        PLOTTER.actual.ch_slant = PLOTTER.requir.ch_slant;
      }
/*
 *      text angle - direction
*/
    if ( fabs( PLOTTER.actual.ch_angle - PLOTTER.requir.ch_angle ) > HP_TINY )
      {
        double hp_angle  = PLOTTER.requir.ch_angle;
        if ( PAGE.paper_rota )
            hp_angle += PI / 2.0;
        if ( fabs( hp_angle ) > HP_TINY )
            (void) fprintf( OP_FILE, "DI%7.4f,%7.4f;\n",
               cos( hp_angle ), sin( hp_angle ) );
        else
            (void) G_PUTS( "DI;\n" );
        PLOTTER.actual.ch_angle = PLOTTER.requir.ch_angle;
        step_changed = TRUE;
      }
/*
 *      character size
*/
    if ( fabs( PLOTTER.actual.ch_size._x - PLOTTER.requir.ch_size._x )
           > HP_TINY || HP_TINY <
         fabs( PLOTTER.actual.ch_size._y - PLOTTER.requir.ch_size._y ) )
      {
        (void) fprintf( OP_FILE, "SI%5.3f,%5.3f;\n",
            ( PLOTTER.actual.ch_size._x = PLOTTER.requir.ch_size._x ) *
                resolution[SETUP.plotter_id] / 10.0,
            ( PLOTTER.actual.ch_size._y = PLOTTER.requir.ch_size._y ) *
                resolution[SETUP.plotter_id] / 10.0 );
        step_changed = TRUE;
      }
/*
 *      character spacing
*/

    if ( ( SETUP.plotter_id == HP7550A || SETUP.plotter_id == HP7550B ||
           SETUP.plotter_id == HP7580B || SETUP.plotter_id == HP7585B ) &&
         ( fabs( PLOTTER.actual.ch_step._x - PLOTTER.requir.ch_step._x )
                 > HP_TINY || step_changed || HP_TINY <
           fabs( PLOTTER.actual.ch_step._y - PLOTTER.requir.ch_step._y ) ) )
      {
        (void) fprintf( OP_FILE, "ES%7.4f;\n", HYPOT( PLOTTER.requir.ch_step )
            / PLOTTER.requir.ch_size._x / DEF_SPACE - 1.0 );
        PLOTTER.actual.ch_step = PLOTTER.requir.ch_step;
      }
    else if ( step_changed )
      {
        double step = PLOTTER.actual.ch_size._x * DEF_SPACE;
        PLOTTER.actual.ch_step._x = step * cos( PLOTTER.actual.ch_angle );
        PLOTTER.actual.ch_step._y = step * sin( PLOTTER.requir.ch_angle );
      }
  }
#undef  HP_TINY


static char VAR_PTR pr_mode( status_ptr )
/*
 *      Returns a verbal interpretation of plotter status (on or off) .
*/
Plot_Status VAR_PTR status_ptr;
  {
    switch( status_ptr->mode )
      {
        case OFF_PLOTTER  : return (char VAR_PTR) "plotter off";
        case ON_PLOTTER   : return (char VAR_PTR) "plotter on";
      }
    return mt_str;
  }


/* ************************************************************************ */
/*                                                                          */
/*                   Coordinate conversion routines                         */
/*                                                                          */
/* ************************************************************************ */


static int_coord VAR_PTR plot_units( pix_point, plot_point )
/*
 *      Convert coordinates from "pixel" units into "plotter" units .
 * If necessary, a coordinate rotation of 90 degrees is done .
*/
coord_pair VAR_PTR pix_point;
int_coord  VAR_PTR plot_point;
  {
    if ( PAGE.paper_rota )
      {
        plot_point->_ix = PAGE.offset._ix - (int) ( pix_point->_y + 0.5 );
        plot_point->_iy = (int) ( pix_point->_x + 0.5 ) - PAGE.offset._iy;
      }
    else
      {
        plot_point->_ix = (int) ( pix_point->_x + 0.5 ) - PAGE.offset._ix;
        plot_point->_iy = (int) ( pix_point->_y + 0.5 ) - PAGE.offset._iy;
      }
    return plot_point;
  }


/* ************************************************************************ */
/*                                                                          */
/*                        Basic output routines                             */
/*                                                                          */
/* ************************************************************************ */


static boolean hp_move( ipoint )
/*
 *  move with the pen up to ``ipoint'' ( in plotter units ) using a PU command
*/
int_coord VAR_PTR ipoint;
  {
    if ( hpgl->hanging_cmd )
        hp_terminate();
    (void) fprintf( OP_FILE, "PU%d,%d;\n", ipoint->_ix, ipoint->_iy );
    return ! G_ERROR();
  }


#define MAX_DIGITS    6       /* max no. of digits in an abscissa or      */
                              /* ordinate value                           */

static boolean hp_draw( ipoint )
/*
 *  move with the pen down to ``ipoint'' ( in plotter units ) using a PD
 *  command . Successive "PDx,y" commands are abbreviated by using the
 *  "PDx1,y1,x2,y2, ..." format and placing as many coordinate pairs as
 *  possible on a line .
*/
int_coord VAR_PTR ipoint;
  {
    if ( ! hpgl->hanging_cmd )
        (void) G_PUTS( "PD" );
    else if ( ( MAX_LINE_LEN - 4 ) / 2 / ( MAX_DIGITS + 1 ) < hpgl->point_cntr )
      {
        hpgl->point_cntr = 0;
        (void) G_PUTS( ";\nPD" );
      }
    else
        (void) G_PUTC( ',' );
    (void) fprintf( OP_FILE, "%d,%d", ipoint->_ix, ipoint->_iy );
    hpgl->point_cntr ++;
    hpgl->hanging_cmd = TRUE;
    return ! G_ERROR();
  }
#undef MAX_DIGITS


static boolean hp_pen( pen_number )
/*
 *      Select pen number `pen_number' . If `pen_number' exceeds the number
 * of available pens, it is wrapped around the number of available pens .
 * pen_number is assumed to be in the range 0 to 8 .
 * If use of the PT instruction is enabled, and the pen is being selected for
 * the first time, then a PT is also sent .
*/
int pen_number;
  {
    boolean ok;
    if ( hpgl->hanging_cmd )
        hp_terminate();
    if ( pen_number )
      {
        int mask = 0;
        (void) fprintf( OP_FILE, "SP%d;\n",
            pen_number = ( pen_number - 1 ) % PAGE.no_of_pens + 1 );
        if ( PAGE.set_pthick &&
             ! ( hpgl->pen_used & ( mask = 1 << pen_number ) ) )
            (void) fprintf( OP_FILE, "PT%3.1f;\n",
                (double) PAGE.pen_thick * resolution[SETUP.plotter_id] );
        hpgl->pen_used |= mask;
      }
    else
        (void) G_PUTS( "SP;\n" );
    if ( ( ok = ! G_ERROR() ) )
        PLOTTER.actual.colour = pen_number;
    return ok;
  }


static boolean hp_style( style )
/*
 *    set the dashed line style of the Hewlett Packard plotter
*/
int style;        /* dashed line style  */
  {
    boolean ok;
    if ( hpgl->hanging_cmd )
        hp_terminate();
    switch( style )
      {
        case 0 : /* plot vector endpoints only */
                 (void) G_PUTS( "LT0;\n" );
                 break;
        case 1 : /* solid */
                 (void) G_PUTS( "LT;\n" );
                 break;
        case 2 : /* dotted */
                 (void) G_PUTS( "LT1,0.5;\n" );
                 break;
        case 3 : /* short dash */
        case 4 : /* long dash  */
        case 5 : /* dot dash   */
        case 6 : /* short-dash long-dash */
        case 7 : /* 2-short-dashes long-dash */
                 (void) fprintf( OP_FILE, "LT%d,4;\n", style - 1 );
                 break;
        case 8 : /* very short dashes */
                 (void) G_PUTS( "LT2,2;\n" );
                 break;
      }
    if ( ( ok = ! G_ERROR() ) )
        PLOTTER.actual.dash = style;
    return ok;
  }


#define HATCH_SPACE 120     /* line spacing for horizontal hatching */

static boolean hp_box( lower_left, upper_right, style )
/*
 *  Draw a box with corner coordinates `*lower_left' and `*upper_right', and
 *  fill it with the hatching type `style' .
 *  This function reserves the right to swap coordinate values if they are not
 *  in the correct order . The coordinates are in plotter units .
*/
int_coord *lower_left,
          *upper_right;
int style;
  {
/*
 *          ensure that the corner coordinates are correctly ordered
*/
    if ( upper_right->_ix < lower_left->_ix )
        SWAP_SI( upper_right->_ix, lower_left->_ix );
    if ( upper_right->_iy < lower_left->_iy )
        SWAP_SI( upper_right->_iy, lower_left->_iy );
    if ( hpgl->hanging_cmd )
        hp_terminate();
/*
 *                  ********** SMART PLOTTERS **********
*/
    if ( SETUP.plotter_id == HP7475A ||
         SETUP.plotter_id == HP7550A ||
         SETUP.plotter_id == HP7550B )
      {
/*
 *          move to lower left corner
*/
        (void) hp_move( lower_left );
        if ( style )
          {
/*
 *          fill the box with the required pattern
*/
            if ( PLOTTER.actual.fill_type != style )
              {
                switch ( style )
                  {
                    case 1 : (void) G_PUTS( "FT;\n" ); break;
                    case 2 :
                        (void) fprintf( OP_FILE, "FT3,%d,0;\n", PAGE.pen_thick);
                        break;
                    case 3 :
                        (void) fprintf( OP_FILE, "FT3,%d,135;\n", HATCH_SPACE );
                        break;
                    case 4 :
                        (void) fprintf( OP_FILE, "FT3,%d,45;\n", HATCH_SPACE );
                        break;
                    case 5 :
                        (void) fprintf( OP_FILE, "FT3,%d,0;\n", HATCH_SPACE );
                        break;
                    case 6 :
                        (void) fprintf( OP_FILE, "FT3,%d,90;\n", HATCH_SPACE );
                        break;
                    case 7 :
                        (void) fprintf( OP_FILE, "FT4,%d,45;\n", HATCH_SPACE );
                        break;
                    case 8 :
                        (void) fprintf( OP_FILE, "FT4,%d,0;\n", HATCH_SPACE );
                        break;
                  }
                PLOTTER.actual.fill_type = style;
              }
            (void) fprintf( OP_FILE, "RA%d,%d;\n",
                            upper_right->_ix, upper_right->_iy );
          }
/*
 *          draw the empty box
*/
        (void) fprintf( OP_FILE, "EA%d,%d;\n",
                        upper_right->_ix, upper_right->_iy );
      }
/*
 *              *********** STUPID PLOTTERS *********
 *          in this case, it is necessary to draw the box and fill pattern
 *          line by painful line - this may generate a lot HPGL
*/
    else if ( SETUP.plotter_id == HP7470A ||
              SETUP.plotter_id == HP7580B ||
              SETUP.plotter_id == HP7585B ||
              SETUP.plotter_id == COLORPRO )
      {
/*
 *          fill the box with the required pattern
*/
        if ( style )
            switch ( style )
              {
                case 1 : hp_fill_hatch( lower_left, upper_right );
                         break;
                case 2 : hp_horiz_hatch(
                             (int) lower_left->_ix,  (int) lower_left->_iy,
                             (int) upper_right->_ix, (int) upper_right->_iy,
                             PAGE.pen_thick * 2 );
                         break;
                case 3 : hp_NWSE_hatch(
                             (int) lower_left->_ix,  (int) lower_left->_iy,
                             (int) upper_right->_ix, (int) upper_right->_iy,
                             PAGE.pen_thick + (int) ( 1.41 * HATCH_SPACE ) );
                         break;
                case 4 : hp_NESW_hatch(
                             (int) lower_left->_ix,  (int) lower_left->_iy,
                             (int) upper_right->_ix, (int) upper_right->_iy,
                             PAGE.pen_thick + (int) ( 1.41 * HATCH_SPACE ) );
                         break;
                case 5 : hp_horiz_hatch(
                             (int) lower_left->_ix,  (int) lower_left->_iy,
                             (int) upper_right->_ix, (int) upper_right->_iy,
                             PAGE.pen_thick + HATCH_SPACE );
                         break;
                case 6 : hp_vert_hatch(
                             (int) lower_left->_ix,  (int) lower_left->_iy,
                             (int) upper_right->_ix, (int) upper_right->_iy,
                             PAGE.pen_thick + HATCH_SPACE );
                         break;
                case 7 : hp_NWSE_hatch(
                             (int) lower_left->_ix,  (int) lower_left->_iy,
                             (int) upper_right->_ix, (int) upper_right->_iy,
                             PAGE.pen_thick + (int) ( 1.41 * HATCH_SPACE ) );
                         hp_NESW_hatch(
                             (int) lower_left->_ix,  (int) lower_left->_iy,
                             (int) upper_right->_ix, (int) upper_right->_iy,
                             PAGE.pen_thick + (int) ( 1.41 * HATCH_SPACE ) );
                         break;
                case 8 : hp_horiz_hatch(
                             (int) lower_left->_ix,  (int) lower_left->_iy,
                             (int) upper_right->_ix, (int) upper_right->_iy,
                             PAGE.pen_thick + HATCH_SPACE );
                         hp_vert_hatch(
                             (int) lower_left->_ix,  (int) lower_left->_iy,
                             (int) upper_right->_ix, (int) upper_right->_iy,
                             PAGE.pen_thick + HATCH_SPACE );
                         break;
              }
/*
 *          draw the empty box
*/
        (void) hp_move( lower_left );
        (void) fprintf( OP_FILE, "PD%d,%d,%d,%d,%d,%d,%d,%d;\n",
            lower_left->_ix,  upper_right->_iy,
            upper_right->_ix, upper_right->_iy,
            upper_right->_ix, lower_left->_iy,
            lower_left->_ix, lower_left->_iy );
      }
    return ! G_ERROR();
  }

#undef HATCH_SPACE

static void hp_horiz_hatch( ll_x, ll_y, ur_x, ur_y, spacing )
/*
 *  Fill the box with a hatching pattern of horizontal lines, starting from
 *  the bottom of the box and going to the top .
 *  The direction of drawing successive lines is reversed in order to
 *  eliminate unnecessary pen motion .
*/
int ll_x, ll_y,     /* lower left corner of the box  */
    ur_x, ur_y,     /* upper right corner of the box */
    spacing;        /* vertical space between hatching lines */
  {
    int swap_buffer;
    for( ll_y += spacing; ll_y < ur_y; ll_y += spacing )
      {
        (void) fprintf( OP_FILE, "PU%d,%d;PD%d,%d;\n",
                        ll_x, ll_y, ur_x, ll_y );
        swap_buffer = ur_x;
        ur_x = ll_x;
        ll_x = swap_buffer;
      }
  }


static void hp_vert_hatch( ll_x, ll_y, ur_x, ur_y, spacing )
/*
 *  Fill the box with a hatching pattern of vertical lines, starting from
 *  the left hand side of the box and going to the right hand side .
 *  The direction of drawing successive lines is reversed in order to
 *  eliminate unnecessary pen motion .
*/
int ll_x, ll_y,     /* lower left corner of the box  */
    ur_x, ur_y,     /* upper right corner of the box */
    spacing;        /* vertical space between hatching lines */
  {
    int swap_buffer;
    for( ll_x += spacing; ll_x < ur_x; ll_x += spacing )
      {
        (void) fprintf( OP_FILE, "PU%d,%d;PD%d,%d;\n",
                        ll_x, ll_y, ll_x, ur_y );
        swap_buffer = ur_y;
        ur_y = ll_y;
        ll_y = swap_buffer;
      }
  }


static void hp_fill_hatch( ll_corner, ur_corner )
/*
 *  Fill the box with solid colour,  starting from the bottom of the box
 *  and going to the top .
 *  The direction of drawing successive lines is reversed and the pen is not
 *  lifted in order to eliminate unnecessary pen motion . This routine
 *  generates rather a lot of HPGL .
*/
int_coord *ll_corner,   /* lower left corner of the box  */
          *ur_corner;   /* upper right corner of the box */
  {
    short int swap_buffer;
    int_coord left, right;
    left._ix  = ll_corner->_ix;
    right._ix = ur_corner->_ix;
    if ( ( right._iy = left._iy = ll_corner->_iy + PAGE.pen_thick ) <
         ( ur_corner->_iy - PAGE.pen_thick ) )
      {
        (void) hp_move( (int_coord VAR_PTR) &right );
        (void) hp_draw( (int_coord VAR_PTR) &left );
        for( right._iy = left._iy += PAGE.pen_thick;
             right._iy <= ur_corner->_iy - PAGE.pen_thick;
             right._iy = left._iy += PAGE.pen_thick )
          {
            (void) hp_draw( (int_coord VAR_PTR) &left );
            (void) hp_draw( (int_coord VAR_PTR) &right );
            swap_buffer = right._ix;
            right._ix = left._ix;
            left._ix = swap_buffer;
          }
      }
  }


static void hp_NWSE_hatch( ll_x, ll_y, ur_x, ur_y, spacing )
/*
 *  Fill the box with a hatching pattern of diagonal North-west to South-east
 *  lines, starting from the bottom right corner of the box .
 *  The direction of drawing successive lines is reversed in order to
 *  eliminate unnecessary pen motion .
*/
int ll_x, ll_y,     /* lower left corner of the box  */
    ur_x, ur_y,     /* upper right corner of the box */
    spacing;        /* vertical or horizontal spacing between lines */
  {
    int width  = ur_x - ll_x,   /* width of the box */
        height = ur_y - ll_y,   /* height of the box */
        counter = 0,            /* hatching line counter */
        step,
        var;
/*
 *          draw lines above the bottom right corner
*/
    for( var = ll_y; ( step = MIN( width,  ur_y - var ) ) > 0; var += spacing )
        hp_diag_hatch( ur_x, var, ur_x - step, var + step, counter++ );
/*
 *          draw lines to the left of the bottom right corner
*/
    for( var = ur_x - spacing; ( step = MIN( height, var - ll_x ) ) > 0;
            var -= spacing )
        hp_diag_hatch( var, ll_y, var - step, ll_y + step, counter++ );
  }


static void hp_NESW_hatch( ll_x, ll_y, ur_x, ur_y, spacing )
/*
 *  Fill the box with a hatching pattern of diagonal North-east to South-west
 *  lines, starting from the bottom left corner of the box .
 *  The direction of drawing successive lines is reversed in order to
 *  eliminate unnecessary pen motion .
*/
int ll_x, ll_y,     /* lower left corner of the box  */
    ur_x, ur_y,     /* upper right corner of the box */
    spacing;        /* vertical or horizontal spacing between lines */
  {
    int width  = ur_x - ll_x,   /* width of the box */
        height = ur_y - ll_y,   /* height of the box */
        counter = 0,            /* hatching line counter */
        step,
        var;
/*
 *          draw lines above the bottom left corner
*/
    for( var = ll_y; ( step = MIN( width,  ur_y - var ) ) > 0; var += spacing )
        hp_diag_hatch( ll_x, var, ll_x + step, var + step,
                       counter++ );
/*
 *          draw lines to the right of the bottom left corner
*/
    for( var = ll_x + spacing; ( step = MIN( height, ur_x - var ) ) > 0;
            var += spacing )
        hp_diag_hatch( var, ll_y, var + step, ll_y + step, counter++ );
  }


static void hp_diag_hatch( x1, y1, x2, y2, counter )
/*
 *  Draws a line between endpoints (x1,y1) and (x2,y2) .
 *  If `counter' is odd, the line is drawn from (x1,y1) to (x2,y2) .
 *  If `counter' is even, the line is drawn from (x2,y2) to (x1,y1) .
 *  This is used for drawing diagonal lines for hatching .
*/
int x1, y1,         /* first end point of the line */
    x2, y2,         /* second end point of the line */
    counter;        /* line counter */
  {
    if ( counter & 1 )
        (void) fprintf( OP_FILE, "PU%d,%d;PD%d,%d;\n", x1, y1, x2, y2 );
    else
        (void) fprintf( OP_FILE, "PU%d,%d;PD%d,%d;\n", x2, y2, x1, y1 );
  }


static boolean hp_on()
/*
 *  If HP/GL is sent to "stdout" ( i.e. the user's terminal ), the plotter is
 *  placed in the programmed on state by sending "ESC.(" . This is useful to
 *  control the behaviour of functions which do different things in different
 * plotter modes . The resulting plotter mode is OFF_PLOTTER .
*/
  {
    boolean ok = TRUE;
    if ( hpgl->hanging_cmd )
        hp_terminate();
    if ( OP_FILE == stdout && SETUP.rs_232 )
      {
        (void) G_PUTS( "\033.(" );
        ok = ! G_ERROR();
      }
    if ( ok )
      PLOTTER.actual.mode = ON_PLOTTER;
    return ok;
  }


static boolean hp_off()
/*
 *  If HP/GL is sent to "stdout" ( i.e. the user's terminal ), the plotter is
 *  placed in the programmed off state by sending "ESC.)" . This allows use of
 *  standard I/O with the terminal when the plotter is used in eavesdrop mode .
 *  The resulting plotter mode is OFF_PLOTTER .
*/
  {
    boolean ok = TRUE;
    if ( hpgl->hanging_cmd )
        hp_terminate();
    if ( OP_FILE == stdout && SETUP.rs_232 )
      {
        (void) G_PUTS( "\033.)" );
        (void) G_FLUSH();
        ok = ! G_ERROR();
      }
    if ( ok )
      PLOTTER.actual.mode = OFF_PLOTTER;
    return ok;
  }


static void hp_terminate()
/*
 *          write a HP-GL command terminator
*/
  {
    (void) G_PUTS( ";\n" );
    hpgl->hanging_cmd = FALSE;
    hpgl->point_cntr = 0;
  }



