/*
    Copyright (C) 1997  Peter Vernon Lanspeary

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* ************************************************************************ */
/*                                                                          */
/*                Basic Graphics Terminal Graphics Driver for               */
/*                   PostScript page description language                   */
/*                   N.B. This is hardware dependent stuff                  */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/*   This is a graphics driver for a PostScript interpreter which produces  */
/*   output on a graphics device. The default graphics device is an Apple   */
/*   Laserwriter. Output is sent to a file (which may be your terminal).    */
/*   Separate virtual pens are used for drawing lines and for text . Text   */
/*   is written using the plotter's default font whenever possible .        */
/*                                                                          */
/*      The PostScript output of this driver does not conform to file       */
/*   structuring conventions because downloaded procedures are scattered    */
/*   throughout the file, and so a page description may depend upon a       */
/*   previous page . This means that reordering pages or selecting page     */
/*   subsets of the PostScript file with a filter will produce nonsense .   */
/*                                                                          */
/*      Most functions are of type `boolean' and return a TRUE or FALSE     */
/*   result to indicate success or failure . Error messages are placed on   */
/*   the error queue -- which can be examined at the discretion of the user.*/
/*                                                                          */
/*   Any programs which use this driver must #include DEVICE_G_HEADER .     */
/*   These functions must be linked with "clip.c" and "error.c" .           */
/*   The documentation on "clip.c" and "error.c" should be read before      */
/*   continuing .                                                           */
/*                                                                          */
/*   The functions t_visib(), t_invis(), t_revers(), t_replac(), t_gr_off(),*/
/*   t_gr_on(), t_xhrput() and t_xhrget() are inactive. The functions       */
/*   t_newscr(), t_bell() and  t_wait() are inactive unless output is to    */
/*   the users terminal .                                                   */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/* The function `t_begin()' has two arguments . The first `device' argument */
/* is the FILE* pointer to the output device.                               */
/* The second argument is a character string of selected options . The      */
/* available options are the words                                          */
/*  ( the first three are page types for the Apple Laserwriter )            */
/*   "LETTER"           imageable region of 8.0 x 10.92 inches , centred on */
/*                      an 8.5 x 11 inch page, with 0.25 inch margins on    */
/*                      the right and left and 0.04 inch margins on top and */
/*                      bottom . This is the standard page type for all but */
/*                      legal-sized paper .                                 */
/*                          0 setpagetype                                   */
/*   "NOTE"             imageable region of 7.69 x 10.16 inches, centred on */
/*                      an 8.5 x 11 inch page, with 0.405 inch margins on   */
/*                      the right and left and 0.42 inch margins on top and */
/*                      bottom . This page type can be used for jobs        */
/*                      requiring unusually large amounts of VM; the smaller*/
/*                      frame buffer allows the VM to be about 100k bytes   */
/*                      larger than normal                                  */
/*                          1 setpagetype                                   */
/*   "LEGAL"            imageable region of 6.72 x 13.0 inches , centred on */
/*                      an 8.5 x 14 inch page, with 0.89 inch margins on    */
/*                      the right and left and 0.5 inch margins on top and  */
/*                      bottom . This is the standard page type for legal-  */
/*                      sized paper .                                       */
/*   "WIDE" (default)   X axis is parallel to long side of the paper        */
/*   "TALL"             Y axis is parallel to long side of the paper        */
/*   "BUILT_IN_FONTS"  (default) use plotter's buit-in fonts for ASCII chars*/
/*   "PRIVATE_FONTS"    use the driver's fonts for ASCII characters         */
/*"STANDARD_MONOCHROME" standard monochrome PostScript only                 */
/*   "STANDARD_COLOUR"  standard colour PostScript only                     */
/*  "APPLE_LASERWRITER" (degault) Apple Laserwriter (monochrome)            */
/*  "LINE_WIDTH number" line thickness in points (default=1)                */
/*  "TTY_SIZE number"   Size of teletype text (default=10)                  */
/*                                                                          */
/* Words may be abbreviated and in upper, lower or mixed case . For example */
/* "tall A2 us PRIV" specifies tall plots on C (17" x 22" ANSI) sized paper */
/* using the driver's fonts for stroke-based ASCII characters .             */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/* Technique for the PostScript device driver                               */
/*                                                                          */
/* This driver downloads and then uses PostScript procedures to achieve     */
/* many of its functions . The procedures are initialy stored in the file   */
/* defined as POSTSCRIPT_PROCS in the file "system.h" . When "t_begin()" is */
/* called, these procedures are read and stored in memory ready for use .   */
/* When a procedure is required for the first time, it is downloaded to the */
/* postscript file or device and it is called . If necessary, any procedures*/
/* which it calls are also downloaded . If a procedure is required again,   */
/* it is simply called . The POSTSCRIPT_PROCS  file can be in any legal     */
/* postscript format and may include comments . The guidelines for writing  */
/* PostScript procedures are                                                */
/*  1) proper documentation                                                 */
/*  2) do not use local variables - use only the stack for local variables  */
/*  3) use whatever intelligence exists in the interpreter                  */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/*      SYSTEMS OF UNITS                                                    */
/* The graphics device standard specifies that the graphics units available */
/* to the user are "pixels" i.e. equal in size to the basic resolution of   */
/* the plotter . The origin of the "pixel" coordinate system (0,0) is the   */
/* lower left corner of the paper or screen . The paper is aligned so that  */
/* the x-axis is horizontal and the y-axis is vertical .                    */
/*      The PostScript language does not define the basic device (pixel)    */
/* resolution . For example, the pixel size of the Apple Laserwriter is 300 */
/* dots per inch . The default units of the plotter are `points'. There are */
/* 72 `points' in one inch. Note that these are not true publisher's points.*/
/*                                                                          */
/* ************************************************************************ */

#include "system.h"             /* system dependent stuff */
#include <stdio.h>
#include <stdlib.h>
#include DEVICE_G_HEADER
#include FONT_SOURCE_HEADER
#include "dev_udef.h"

/* ************************************************************************ */
/*                                                                          */
/*                                CONSTANTS                                 */
/*                                                                          */
/* ************************************************************************ */

/*
 *      Machine constants
*/
#define TRUE          1
#define FALSE         0
#define PI            3.141592653589793238462643
/*#define NULL        0       NULL pointer - already defined in stdio.h   */
/*#define EOF         (-1)    end of file  - already defined in stdio.h   */

/*
 *      definitions for PostScript procedures
*/
#define NO_PS_PROCS     55      /* maximum number of PostScript pocedures */
#define NAME_LEN        40      /* maximum number of chars in procedure name */
#define PROC_LEN        1024    /* maximum number of chars in procedure text */
/*
 *    non printable ASCII characters
*/
#define NUL  '\000'   /* Nul                   */
#define BEL  '\007'   /* Bell                  */
#define VT   '\013'   /* Vertical Tabulation   */
#define EM   '\031'   /* End of Medium         */

/*
 *          constants for drawing teletype characters
*/

#define CLIP2R_ID   0x3c22    /* CL2R two dimensional rectangular clipping*/
#define SOLID_LINE    1       /* dashed line type                         */
#define BLACK         1       /* pen 1 - black pen                        */
#define POINT_SIZE (25.4/72)  /* size of a `pixel' unit in mm             */
#define PS_JITTER     0.1     /* assumed maximum device resolution / 2    */
#define BUFFER_LEN    80      /* maximum echo (to plotter) buffer length  */
#define LINE_WIDTH    1.0     /* DEFAULT line width - in points           */
/*
 *  proportions of built-in character font "character box" dimensions
 *  actually occupied by ink -- these have been measured (empirically)
*/
#define BUILT_HEIGHT_FACTOR 0.5669
#define BUILT_WIDTH_FACTOR  0.4308
#define BUILT_BOX_RATIO     0.6     /* width / height of character box */
#define BUILT_GAP_FACTOR   ( BUILT_BOX_RATIO - BUILT_WIDTH_FACTOR )
                                    /* size of gap */
/*
 *  FONT SPECIFIERS
 *      bit groups used to compose
 *  PLOTTER.requir.ch_font and PLOTTER.actual.ch_font .
 *      one term is selected from each group to define the font
 *
 *  font type group ; font is standard or bold
*/
#define STD_FONT       0
#define BOLD_FONT      1
/*
 *  selected character set group
*/
#define ASCII_FONT     0      /* ASCII characters */
#define SYMBOL_FONT    2      /* special symbols and Greek letters */
/*
 *  font source group
 *    private_font is defined by the driver and is downloaded when required
 *    builtin_font is 10 point Courier defined in the PostScript device
*/
#define PRIVATE_FONT   0
#define BUILTIN_FONT   4
#define LAST_FONT      ( BOLD_FONT | ASCII_FONT | BUILTIN_FONT )
#define FIRST_FONT     ( STD_FONT  | ASCII_FONT | PRIVATE_FONT )
/*
 *  bit masks used to isolate the value of a bit group
*/
#define BOLD_MASK      BOLD_FONT
#define CHARSET_MASK   SYMBOL_FONT
#define BUILTIN_MASK   BUILTIN_FONT
/*
 *          definition of AVAILABLE OPTIONS
*/
#define MAX_WORD_LEN   18     /* max length of word in the options list   */
#define NOT_SET        -1     /* option not specified in options string   */
#define MET_PAPER       0     /* Metric type paper - ie. A4 A3 A2 A1 A0   */
#define US_PAPER        1     /* "US" or "English" type paper, ie. A B... */
#define WIDE_PAPER      0     /* "Horizontal" is parallel to long side of */
#define TALL_PAPER      1     /* "Vertical" parallel to long side of paper*/
#define DEFAULT_PAPER   0     /* default image region in Laserwriter      */
#define LETTER_PAPER    1     /* 8.0 x 10.92 inches available on A4 paper */
#define NOTE_PAPER      2     /* 7.69 x 10.16 inches availbale on A4 paper*/
#define LEGAL_PAPER     3     /* 6.72 x 13.0 inches on legal-sized paper  */
#define VIRTUAL_PAPER   4     /* use all of the paper                     */
#define LASERWRITER     0     /* Apple Laserwriter                        */
#define MONOCHROME_STANDARD 1 /* standard PostScript for monochrome device*/
#define COLOUR_STANDARD 2     /* standard PostScript for colour device    */
#define DEF_LINE_WIDTH  1.0   /* default line thickness                   */
#define DEF_TTY_SIZE    10.0  /* default teletype (built-in) text size    */

/*
 *      Array indices for PostScript procedures
 *
 *              these correspond directly to driver functions */
#define FIRST_PROC      0
#define DRAW_PROC       (FIRST_PROC)      /* draw    */
#define MOVE_PROC       (DRAW_PROC+1)     /* move    */
#define POINT_PROC      (MOVE_PROC+1)     /* point   */
#define DASH_PROC       (POINT_PROC+1)    /* dashed  */
#define COLOUR_PROC     (DASH_PROC+1)     /* colour  */
#define GREY_PROC       (COLOUR_PROC+1)   /* grey_scale */
#define INIT_PROC       (GREY_PROC+1)     /* initialize */
#define VWPORT_PROC     (INIT_PROC+1)     /* viewport   */
#define NEWPAG_PROC     (VWPORT_PROC+1)   /* newpage */
#define FINISH_PROC     (NEWPAG_PROC+1)   /* finish  */
#define TTY_PROC        (FINISH_PROC+1)   /* LR_text */
#define TEXT_PROC       (TTY_PROC+1)       /* GR_text */
#define BOX_PROCS       (TEXT_PROC+1)       /* box_#   */
/*             other  defined PostScript variables */
#define STD_TXT_DEF     (BOX_PROCS+9)     /* std_builtin_txt  */
#define BOLD_TXT_DEF    (STD_TXT_DEF+1)   /* bold_builtin_txt */
#define TTY_SIZE_DEF    (BOLD_TXT_DEF+1)  /* TTY_size */
#define GR_FONT_DEF     (TTY_SIZE_DEF+1)  /* GR_font */
#define TXT_MAT_DEF     (GR_FONT_DEF+1)   /* GR_text_matrix   */
#define CHAR_GAP_DEF    (TXT_MAT_DEF+1)    /* GR_char_spacing  */
/*              for defining and selecting character fonts  */
#define DEF_CH_PROC     (CHAR_GAP_DEF+1)    /* define_char    */
#define T_ST_FONT_PROC  (DEF_CH_PROC+1)     /* txt_start_font */
#define T_EN_FONT_PROC  (T_ST_FONT_PROC+1)  /* txt_end_font   */
#define T_MK_BOLD_PROC  (T_EN_FONT_PROC+1)  /* txt_make_bold  */
#define S_ST_FONT_PROC  (T_MK_BOLD_PROC+1)  /* sym_start_font */
#define S_EN_FONT_PROC  (S_ST_FONT_PROC+1)  /* sym_end_font   */
#define S_MK_BOLD_PROC  (S_EN_FONT_PROC+1)  /* sym_make_bold  */
#define LAST_PROC       S_MK_BOLD_PROC
#define CALLABLE_PROCS  LAST_PROC+1

/* ************************************************************************ */
/*                                                                          */
/*                          Graphics Macros                                 */
/*                                                                          */
/* ************************************************************************ */

/*
 *  ABBREVIATIONS
 ******************
*/
#define  PLOTTER    ps_device->plotter
#define  STROKE     ps_device->stroke
#define  VIEW_CLIP  ps_device->view_clip
#define  OP_FILE    ps_device->op_file
#define  LOADED     ps_device->loaded
#define  SETUP      ps_device->setup
#define  PAGE       ps_device->page

/*
 *  CATCH_NULL(pointer,caller,ret_type)
 ****************************************
 *      If "pointer" has a NULL value, an "Arg is NULL pointer" error message
 * is generated and the function is aborted . Otherwise, nothing is done .
 *      "pointer"   is the pointer to be tested,
 *      "caller"    is a character string naming the function
 *      "ret_type"  is the type of the returned zero value which indicates an
 *                    error condition
*/
#define CATCH_NULL(pointer,caller,ret_type) \
  if ( ! (pointer) )    \
      return (ret_type) make_error( ERR_ABORT, caller, err_null )

/*
 *  G_PUTS(string)
 *******************
 *      Writes the character string `string' to the graphics device
 *      (terminal) . The there is no returned result .
*/
#define G_PUTS(string)      fputs( string, OP_FILE )

/*
 *  G_PUTC(c)
 **************
 *      Writes the character `c' to the graphics device (terminal) .
 *      There is no returned result.
*/
#define G_PUTC(c)           fputc( c, OP_FILE )

/*
 *  G_ERROR()
 **************
 *      Determines the error status of the graphics device (terminal) .
 *      The returned value is TRUE if an error has occurred, otherwise FALSE .
*/
#define G_ERROR()           ferror( OP_FILE )

/*
 *  G_FLUSH()
 **************
 *      Flushes the graphics device (terminal) output buffer . There is no
 *      returned value .
*/
#define G_FLUSH()          fflush( OP_FILE )

/*  LIMIT( x, min, max )
 *************************
 *      Limits the value of `x' to between `min' and `max'
*/
#define LIMIT(x,min,max)  (x) <= (min) ? (min) : ( (x) >= (max) ? (max) : (x) )

/*
 *  LOAD_PS(callable_index)
 ****************************
 * Downloads the user-callable PostScript procedure with index
 * ``callable_index'' into the list of callable procedures .
 * Note that the list of callable procedures is ordered whereas the procedure
 * directory is random (not ordered) . This macro provides a translation from
 * the ordered to the random index . The argument ``callable_index'' is nearly
 * always a constant, and so the indexing into ``proc_list'' (which is a static
 * variable) followed by the offset ``pr_index'' should compile to a single
 * constant .
*/
#define LOAD_PS(callable_index) \
    load_ps_proc( proc_list[callable_index]->pr_index )

/*
 *  MAX( x, y )
 ****************
 *      The returned value is the maximum of `x' and `y' .
*/
#define MAX(x,y)        ( (x) > (y) ? (x) : (y) )

/*
 *  MIN( x, y )
 ****************
 *      The returned value is the minimum of `x' and `y' .
*/
#define MIN(x,y)        ( (x) > (y) ? (y) : (x) )

/*
 *  TOLOWER( c )
 ****************
 *      Converts character `c' to lower case if it is upper case
*/
#define TOLOWER(c)      'A' <= (c) && (c) <= 'Z' ? (c) + 'a' - 'A' : (c)

/*
 *  START_UP()
 ***************
 *      Calls `t_begin' with the default arguments `stdout' and `NULL'
 *  if `t_begin' has not previously been called .
*/
#define START_UP(caller,ret_type) \
    if ( ! ps_device ) \
        if ( ! t_begin( stdout, mt_str ) ) \
            return (ret_type) make_error( ERR_CALL, caller, mt_str )

/*
 *  SWAPD( a, b )
 *****************
 *    swaps the two double precision floating point values `a' and `b'
*/
#define SWAPD(a,b)       { double temp = b; b = a; a = temp; }

/* ************************************************************************* */
/*                                                                           */
/*                           DATA TYPE DECLARATIONS                          */
/*                                                                           */
/* ************************************************************************* */


typedef struct proc_struc   /* STORAGE FOR POSTSCRIPT PROCEDURES */
  {
    char VAR_PTR pr_name;   /* name of the procedure */
    char VAR_PTR pr_string; /* text of the procedure */
    int          pr_index;  /* index of the procedure into the procedures */
                            /* directory and loaded-status arrays */
    struct pr_dl_struc      /* linked list of procedures called by this one */
      {
        struct proc_struc  VAR_PTR pr_depend; /* pointer to called procedure */
        struct pr_dl_struc VAR_PTR pr_next;  /* pointer to next list element */
      }
        VAR_PTR pr_list;    /* pointer to first list element */
  }
     Proc_entry, VAR_PTR Procedure;

typedef struct pr_dl_struc  /* pointer to linked list element  */
    Proc_dep_list;          /* of called PostScript procedures */

/*
 * plotter status -- purpose is to remember what the plotter is doing
*/
typedef struct
  {
    int           colour;   /* foreground colour                   */
    int           dash;     /* dashed line style                   */
    int           thick;    /* line and test width                 */
    int           ch_font;  /* specifies the character font        */
    coord_pair    cursor;   /* alpha graphics text cursor position */
    coord_pair    pen;      /* vector graphics pen position        */
    double        ch_gap;   /* extra space between characters      */
  }
    Plot_Status;

/*
 * element of the hardclip limits list for each printer
 *      units of coordinates are not rotated ( i.e. they are for a tall page )
*/
typedef struct
  {
    short int  Pp_type;     /* paper type - US or metric */
    short int  Pp_size;     /* paper size - LETTER, NOTE or LEGAL */
    coord_pair P_dimens;    /* dimensions of paper - (w,h) */
    coord_pair Pll_clip;    /* lower left hard clip limit  */
    coord_pair Pur_clip;    /* upper right hard clip limit */
  }
    Hclp_Elem;

/*
 *  This defines the capabilities, selected options and general setup of
 *  the plotter
*/
typedef struct
  {
    short int plotter_id;   /* APPLE_LASERWRITER, MONOCHROME_STANDARD,    */
                            /* COLOUR_STANDARD PostScript */
    short int paper_type;   /* Metric or US paper type    */
    short int paper_size;   /* LETTER, NOTE or LEGAL      */
    short int paper_dirn;   /* wide or tall orientation   */
    short int built_in;     /* use plotter's internal font*/
    double line_width;      /* line_thickness in points   */
    double tty_size;        /* size of teletype text      */
  }
    Setup_List;

typedef struct
  {
    short int alph_columns; /* number of character columns on the page */
    short int alph_lines;   /* number of character lines on the page   */
    short int is_monochr;   /* the device is monochrome - use grey scales */
    Hclp_Elem VAR_PTR hclp_ptr;  /* entry in paper size table - not rotated */
    coord_pair plot_size;   /* width, height of available area */
    coord_pair ll_limit;    /* lower left plotting limit  */
    coord_pair ur_limit;    /* upper right plotting limit */
    double alph_width;      /* width of tty character     */
    double alph_height;     /* height of tty character    */
  }
    Page_List;

typedef struct
  {
    Setup_List setup;       /* properties determined by options settings   */
    Page_List  page;        /* description of page size set by `t_begin()' */
    Text_Attribute stroke;  /* stroke text attributes                      */
    struct
      {                     /* Plotter status  */
        Plot_Status requir; /*     required    */
        Plot_Status actual; /*     actual      */
      }
        plotter;
    FILE *op_file;          /* file pointer for the graphics output device */
    clip_area VAR_PTR view_clip; /* pointer to viewport clipping area */
    struct                  /* sub-structure for keeping track of what has */
      {                     /* been downloaded */
        boolean asc_fnt;         /* ASCII character font  */
        boolean sym_fnt;         /* symbol character font */
        boolean matrix;          /* character transformation matrix  */
        char procs[NO_PS_PROCS]; /* listed PostScript procedures     */
            /* Each array element indicates the loaded-status of the */
            /* corresponding PostScript procedure in the directory array */
      }
        loaded;
    boolean Courier_used;   /* built in Courier font has been used         */
    boolean Cour_bold_used; /* built in Courier-Bold font has been used    */
    boolean empty_page;     /* nothing has been put on the page yet        */
  }
    Gfx_Status;

/* ************************************************************************ */
/*                                                                          */
/*                            Static variables                              */
/*                                                                          */
/* ************************************************************************ */

extern StrokeFont
    asc_font,       /* font for ASCII chars   */
    sym_font;       /* font for special chars */
/*
 *  Dummy built-in font (Courier) contained in the plotter -- the numbers
 *  are determined empirically from an Apple Laserwriter .
*/
static StrokeFont built_font =
  { (StrokeChar VAR_PTR) 0, /* directory of character descriptions */
    TRUE,                   /* character font file has been read   */
    96,                     /* number of characters */
    BUILT_HEIGHT_FACTOR * DEF_TTY_SIZE, /* height of characters */
    BUILT_WIDTH_FACTOR * DEF_TTY_SIZE,  /* width of characters  */
    BUILT_GAP_FACTOR * DEF_TTY_SIZE     /* gap between characters */ };

static Proc_entry VAR_PTR proc_dir[NO_PS_PROCS];
                        /* directory of PostScript procedures */
static Gfx_Status VAR_PTR ps_device = (Gfx_Status VAR_PTR) 0;
                        /* complete description of PostScript device */
/*
 *          List of PostScript procedures called directly by the driver
 *          Note that the Postscript procedures are global, but the names
 *          of those procedures are local to ``t_begin()'' and are accessed
 *          only through  proc_list[array_index]->pr_name .
*/
static Procedure proc_list[CALLABLE_PROCS];

/*
 *      Common error messages
*/
static char
  VAR_PTR err_write = "Cannot write to graphics device",
  VAR_PTR err_null  = "Arg is NULL pointer",
  VAR_PTR mt_str  = "", /* auto t_begin options */
  VAR_PTR lab_inactive = "inactive",
  VAR_PTR lab_bug = "Bug in ps_g -- illegal required font",
/*
 *   array of character font names
 *      N.B. order of this list is important
*/
  VAR_PTR font_names[] = {
    "std_private_txt",      /* 0 = STD_FONT  | ASCII_FONT | PRIVATE_FONT  */
    "bold_private_txt",     /* 1 = BOLD_FONT | ASCII_FONT | PRIVATE_FONT  */
    "std_private_sym",      /* 2 = STD_FONT  | SYMBOL_FONT | PRIVATE_FONT */
    "bold_private_sym",     /* 3 = BOLD_FONT | SYMBOL_FONT | PRIVATE_FONT */
    "std_builtin_txt",      /* 4 = STD_FONT  | ASCII_FONT | BUILTIN_FONT  */
    "bold_builtin_txt",     /* 5 = BOLD_FONT | ASCII_FONT | BUILTIN_FONT  */
    "no font" };

/*
 *      external library functions used by the driver
*/
boolean t_parse_options(), t_isspace();
int t_stncmp();
/* char *malloc(); */
/* int  abs(), free(); */
double fabs();
/*
 *      built-in utilities
*/
static int
  t_txtstr();
static boolean
  chain_ps_procs(), free_ps_procs(), load_ps_proc(), print_ps_proc(),
  read_ps_procs(),  ps_header(),     ps_trailer(),   ps_pmdraw(),
  ps_style(),       ps_font(),       ps_send_map(),  ps_text_gap(),
  set_colour(),     set_options(),   make_font(),
  t_newpag(),       t_vwport();
/* static void ps_text_map(); */
static Procedure
  find_ps_proc(),  make_ps_proc();
static Hclp_Elem
  VAR_PTR find_hclp();
static char
  VAR_PTR str_name();

/* ************************************************************************ */
/*                                                                          */
/*                       DEVICE DRIVER UTILITIES                            */
/*                                                                          */
/* ************************************************************************ */


static boolean make_font( font_ptr, font_proc_offset, caller )
/*
 *       Reads a character font specification from the text file `file_name',
 * into memory and constructs `standard' and `bold' PostScript fonts . The
 * fonts are formed by calling predefined  PostScript procedures and creating
 * one PostScript procedure to define the shape of each character .
*/
StrokeFont VAR_PTR font_ptr;    /* (pointer to) character font directory  */
int        font_proc_offset;    /* index into proc_list of the first      */
                                /* PostScript procedure for making the font */
                                /* T_ST_FONT_PROC or S_ST_FONT_PROC       */
char       VAR_PTR caller;      /* name of calling function */
  {
    char VAR_PTR command, /* "lineto" or "moveto" */
         VAR_PTR dict_id; /* first 3 chars used to specify dictionary */
    boolean moveto,       /* TRUE if a move (not  a draw) is required */
            start_char;   /* starting a new character                 */
    int point_count;      /* number of points in continuous line      */
    float VAR_PTR ptr;    /* pointer to coordinate abscissa */
    coord_pair point;     /* current coord point */
    StrokeChar
      VAR_PTR ch_def_ptr, /* pointer to stroke-font character definition */
      VAR_PTR last_ch_def;/* pointer to last definition in the font */
/*
 *      names of PostScript procedures used to generate private fonts are
 *  "define_char", "sym_start_font", "sym_end_font", "sym_make_bold"
 *                 "txt_start_font", "txt_end_font", "txt_make_bold"
 *
 *      download the required PostScript procedures
*/
    if ( LOAD_PS( font_proc_offset ) &&
         LOAD_PS( font_proc_offset + T_EN_FONT_PROC - T_ST_FONT_PROC ) &&
         LOAD_PS( font_proc_offset + T_MK_BOLD_PROC - T_ST_FONT_PROC ) &&
         LOAD_PS( DEF_CH_PROC ) )
      {
/*
 *          begin the PostScript font defintions
 *
 *          dict_id is actually "txt_start_font" or "sym_start_font" -- the
 *          first 3 characters are used later on to specify the dictionary
*/
        dict_id = proc_list[font_proc_offset]->pr_name;
        (void) G_PUTS( (char *) dict_id );
        (void) G_PUTC( '\n' );
/*
 *          create the PostScript character definitions
*/
        for( ch_def_ptr = font_ptr->font_dir,
             last_ch_def = font_ptr->font_dir + font_ptr->font_siz - 1;
             ch_def_ptr <= last_ch_def; ch_def_ptr++ )
            if ( ch_def_ptr &&
                 ch_def_ptr->font_char != NUL && ch_def_ptr->font_npts > 0 )
              {
                (void) G_PUTS( "{\n" );
                start_char = TRUE;
                for ( point_count = 0, ptr = ch_def_ptr->font_coord;
                      point_count < ch_def_ptr->font_npts;
                      point_count++,   ptr += 2 )
                  {
                    moveto = *ptr < 0.0 || *(ptr+1) < 0.0;
                    point._x = fabs( *ptr );
                    point._y = fabs( *(ptr+1) );
/*
 *          define a clause in the PostScript procedure
 *          The start of a line segment is signalled by negative x or y
*/
                    if ( moveto || start_char )
                      {
                        command = "moveto";
                        start_char = FALSE;
                      }
                    else
                        command = "lineto";
                    (void) fprintf( OP_FILE, "%5.2f %5.2f %s\n", point._x,
                        point._y - ch_def_ptr->font_shift, command );
                  }
                (void) fprintf( OP_FILE, "} %d std_%.3s_dict define_char\n",
                    ch_def_ptr->font_char, dict_id );
              }
/*
 *          terminate the PostScript font definitions
*/
        (void) G_PUTS( (char *) proc_list[font_proc_offset + T_EN_FONT_PROC -
            T_ST_FONT_PROC]->pr_name );
        (void) G_PUTC( '\n' );
        (void) G_PUTS( (char *) proc_list[font_proc_offset + T_MK_BOLD_PROC -
            T_ST_FONT_PROC]->pr_name );
        (void) G_PUTC( '\n' );
        if ( G_ERROR() )
            return make_error( ERR_ABORT, caller, err_write );
        return TRUE;
      }
    return 0;
  }


static boolean ps_font( font_id, caller )
/*
 *      Generates the PostScript definingthe required character font . This
 * may involve downloading character fonts to the PostScript device .
*/
int font_id;            /* specifies required font -- plotter.requir.ch_font */
char VAR_PTR caller;    /* name of calling function */
  {
    char VAR_PTR name = "ps_font";
/*
 *          check argument
*/
    if ( font_id < FIRST_FONT || font_id  > LAST_FONT )
        return make_error( ERR_ABORT, name,
            (char VAR_PTR) "Unknown font id" );
/*
 *        if necessary read the ASCII or symbol character font
*/
    if ( ! STROKE.font_ptr->font_got )
        if ( ! read_font( STROKE.font_ptr, file_font( STROKE.font_ptr ),
                   caller ) )
            return FALSE;
    if ( ( PLOTTER.requir.ch_font & CHARSET_MASK ) == ASCII_FONT )
      {
/*
 **************** ASCII CHARACTER FONT
 *
 *        if necessary download the ASCII character font to the PostScript file
*/
        if ( ( font_id & BUILTIN_MASK ) == PRIVATE_FONT )
          {
            if ( ! LOADED.asc_fnt )
                if ( make_font( (StrokeFont VAR_PTR) &asc_font,
                         T_ST_FONT_PROC, caller ) )
                   LOADED.asc_fnt = TRUE;
                else
                   return FALSE;
          }
/*
 *          record the use of a builtin ASCII character font
*/
        else if ( font_id == ( BUILTIN_FONT | ASCII_FONT | STD_FONT ) )
          {
            if ( ! LOAD_PS( STD_TXT_DEF ) )
                return FALSE;
            ps_device->Courier_used = TRUE;
          }
        else if ( font_id == ( BUILTIN_FONT | ASCII_FONT | BOLD_FONT ) )
          {
            if ( ! LOAD_PS( BOLD_TXT_DEF ) )
                return FALSE;
            ps_device->Cour_bold_used = TRUE;
          }
        else /* illegal font_id.plotter_status */
            return make_error( ERR_ABORT, name, lab_bug );
      }
    else if ( ( PLOTTER.requir.ch_font & CHARSET_MASK ) == SYMBOL_FONT )
/*
 **************** SYMBOL CHARACTER FONT
 *
 *        if necessary download the symbol character font to the PostScript file
 *        There is no built-in symbol font; so if used, it must be downloaded
*/
      {
        if ( ! LOADED.sym_fnt )
            if ( make_font( (StrokeFont VAR_PTR) &sym_font,
                         S_ST_FONT_PROC, caller ) )
               LOADED.sym_fnt = TRUE;
            else
               return FALSE;
      }
    else
        return make_error( ERR_ABORT, name, lab_bug );
/*
 *          dummy download the PostScript variable definition
 * so that the font selection is not overwritten by downloading other procedures
*/
    (LOADED.procs)[ proc_list[ GR_FONT_DEF ]->pr_index ] = TRUE;
/*
 *          choose the font
 *      /GR_font "font_name" def
*/
    (void) fprintf( OP_FILE, "/%s %s def\n",
        proc_list[GR_FONT_DEF]->pr_name, font_names[ font_id ] );
    if ( G_ERROR() )
        return make_error( ERR_ABORT, name, err_write );
    PLOTTER.actual.ch_font = font_id;
    return TRUE;
  }


static boolean ps_header()
/*
 *  Writes structuring header comments to the PostScript file .
 *  Note that the file is nonconforming because page descriptions are not
 *  independent. For this reason, pages are not individually labelled with
 *  ``%%Page: label ordinal'' . Pages are not counted because the driver
 *  does not open and close the output file .
*/
  {
    (void) G_PUTS( "%! Adobe PostScript -- " );
    (void) G_PUTS( "Does not conform to structuring conventions.\n" );
    (void) G_PUTS( "%%DocumentFonts: (atend)\n" );
    (void) G_PUTS( "%%Creator: ps_g, PostScript Graphics Driver\n" );
    (void) G_PUTS( "%%CreationDate: ??\n" );
    (void) G_PUTS( "%%EndComments\n" );
    (void) G_PUTS( "%%EndProlog\n" );
    return ! G_ERROR();
  }


static boolean ps_trailer()
/*
 *  Writes structuring trailer comments to the PostScript file .
*/
  {
    (void) G_PUTS( "%%Trailer\n" );
    (void) G_PUTS( "%%DocumentFonts:" );
    if ( ps_device->Courier_used )
        (void) G_PUTS( " Courier" );
    if ( ps_device->Cour_bold_used )
        (void) G_PUTS( " Courier-Bold" );
    (void) G_PUTC( '\n' );
    return ! G_ERROR();
  }


static boolean ps_pmdraw( point, proc_index )
/*
 *  PostScript point move or draw .
 *      This function is designed for using the PostScript procedures
 *  `P' (point), `M' (move) or `D' (draw) to the point ``point'' .
*/
coord_pair VAR_PTR point;  /* (x,y) coordinate */
int proc_index;            /* array index of required PostScript procedure */
  {
/*
 *          download the PostScript procedure
*/
    if ( ! LOAD_PS( proc_index ) )
        return FALSE;
/*
 *          call the PostScript procedure point, move or draw
 *          -- resolution is 0.1 points ( 1/720 inches = 1/32 mm )
*/
    (void) fprintf( OP_FILE, "%.1f %.1f %s\n",
        point->_x, point->_y, proc_list[proc_index]->pr_name );
    PLOTTER.actual.cursor = PLOTTER.actual.pen = *point;
    return ! G_ERROR();
  }


static boolean ps_style( style, proc_index )
/*
 *  PostScript style procedures .
 *      This function is designed for using the PostScript procedures `dashed',
 *  `colour' or `grey_scale' .
*/
int style,      /* dashed line style, colour or grey scale */
    proc_index; /* array index of required PostScript procedure */
  {
/*
 *          download the PostScript procedure
*/
    if ( ! LOAD_PS( proc_index ) )
        return FALSE;
/*
 *          call the PostScript procedure dashed, colour or grey_scale
*/
    (void) fprintf( OP_FILE, "%d %s\n", style,
        proc_list[proc_index]->pr_name );
    return ! G_ERROR();
  }


static boolean set_colour()
/*
 *  satisfies the required colour by setting either the greyscale or the
 *  colour if it is not already correct .
*/
  {
    boolean ok = TRUE;
    if ( PLOTTER.actual.colour != PLOTTER.requir.colour )
      {
        if ( PAGE.is_monochr )
            ok = ps_style( PLOTTER.requir.colour, GREY_PROC );
        else
            ok = ps_style( PLOTTER.requir.colour, COLOUR_PROC );
        if ( ok )
            PLOTTER.actual.colour = PLOTTER.requir.colour;
      }
    return ok;
  }


static int ps_send_map()
/*
 *      Download  the transformation matrix for drawing stroke characters
 * to the PostScript device as GR_text_matrix . Note that this is dummy
 * download -- so that the text mapping matrix is not overwritten by
 * downloading other procedures .
*/
  {
    (LOADED.procs)[ proc_list[ TXT_MAT_DEF ]->pr_index ] = TRUE;
/*
 *          (re)define the text transformation matrix /GR_text_matrix
*/
    (void) fprintf( OP_FILE, "/%s [ %.5g %.5g %.5g %.5g 0 0 ] def\n",
        proc_list[TXT_MAT_DEF]->pr_name, STROKE.map[0][0], STROKE.map[1][0],
        STROKE.map[0][1], STROKE.map[1][1] );
    LOADED.matrix = TRUE;
    return ! G_ERROR();
  }


static Hclp_Elem VAR_PTR find_hclp( hclp_table, table_size )
Hclp_Elem VAR_PTR hclp_table;
int       table_size;
  {
    int counter;
    for( counter = 0;
        counter++ < table_size && ( hclp_table->Pp_type != SETUP.paper_type ||
                                    hclp_table->Pp_size != SETUP.paper_size );
        hclp_table++ ) counter = ++counter - 1;
    return counter <= table_size ? hclp_table : (Hclp_Elem VAR_PTR) 0;
  }


static boolean ps_text_gap( extra_space )
/*
 *      Redefines the PostScript variable GR_char_spacing.
*/
double extra_space;
  {
/*
 *          dummy download the PostScript variable definition so that
 * the character spacing is not overwritten by downloading other procedures
*/
    (LOADED.procs)[ proc_list[ CHAR_GAP_DEF ]->pr_index ] = TRUE;
/*
 *          redefine char_gap
*/
    (void) fprintf( OP_FILE, "/%s %f def\n",
        proc_list[CHAR_GAP_DEF]->pr_name, extra_space );
    PLOTTER.actual.ch_gap = extra_space;
    return ! G_ERROR();
  }


/* ************************************************************************ */
/*                                                                          */
/*                       CALLABLE DRIVER FUNCTIONS                          */
/*                                                                          */
/* ************************************************************************ */


static boolean t_dump()
/*
 *      Dummy for the nonexistent screen dump routine
*/
  {
    return make_error( ERR_ABORT, (char VAR_PTR) "t_dump",
        (char VAR_PTR) "You've gotta be joking !" );
  }


static clip_area VAR_PTR t_begin( device, options )
/*
 *        Start graphics .
 * Decodes the required ``options'' string, initializes the plotter, initializes
 * variables and sets up the viewport clip. The returned value is a pointer to
 * the clip_area structure of the viewport . If the function fails, it returns
 * a NULL pointer. Appropriate error messages are produced in these cases .
*/
FILE *device;    /* file pointer for graphics device - usually a disk file */
char VAR_PTR options;   /* options string */
  {
    boolean ok,
            rotated;        /* coordinate system is rotated for WIDE page */
    FILE *ps_file;
    char VAR_PTR name = "t_begin",
         VAR_PTR ps_header_fname;
/*
 *    These are the names of the PostScript procedures -- IN THE CORRECT ORDER
*/
    static char VAR_PTR proc_names[] =
      { "D",        /* DRAW_PROC */
        "M",        /* MOVE_PROC */
        "P",        /* POINT_PROC */
        "dash",     /* DASH_PROC */
        "colour",   /* COLOUR_PROC */
        "grey",     /* GREY_PROC */
        "initialize",   /* INIT_PROC */
        "viewport", /* VWPORT_PROC */
        "newpage",  /* NEWPAG_PROC */
        "finish",   /* FINISH_PROC */
        "TTY",      /* TTY_PROC */
        "T",        /* TEXT_PROC */
        "B_0",      /* BOX_PROCS+0 */
        "B_1",      /* BOX_PROCS+1 */
        "B_2",      /* BOX_PROCS+2 */
        "B_3",      /* BOX_PROCS+3 */
        "B_4",      /* BOX_PROCS+4 */
        "B_5",      /* BOX_PROCS+5 */
        "B_6",      /* BOX_PROCS+6 */
        "B_7",      /* BOX_PROCS+7 */
        "B_8",      /* BOX_PROCS+8 */
        "std_builtin_txt",  /* STD_TXT_DEF */
        "bold_builtin_txt", /* BOLD_TXT_DEF */
        "TTY_size",         /* TTY_SIZE_DEF */
        "GR_font",          /* GR_FONT_DEF */
        "text_matrix",      /* TXT_MAT_DEF */
        "char_gap",         /* CHAR_GAP_DEF */
        "define_char",      /* DEF_CH_PROC */
        "txt_start_font",   /* T_ST_FONT_PROC */
        "txt_end_font",     /* T_EN_FONT_PROC */
        "txt_make_bold",    /* T_MK_BOLD_PROC */
        "sym_start_font",   /* S_ST_FONT_PROC */
        "sym_end_font",     /* S_EN_FONT_PROC */
        "sym_make_bold" };  /* S_MK_BOLD_PROC */
    CATCH_NULL( device, name, clip_area VAR_PTR );
/*
 *    GRAB GRAPHICS STATUS DATA
 *      Find out if ``device'' has been registered as a device .
 *      If it has, check that it is indeed a PostScript printer and grab the
 *      pointer to it graphics status data .
*/
    if ( t_register( T_ISREGIST, device, NONEXISTENT_DEVICE, (char VAR_PTR) 0))
      {
        if ( (int) t_register( T_REG_TYPE, device, 0, (char VAR_PTR) 0 )
                != POSTSCRIPT_PRINTER )
            return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "Wrong driver. Device is not PostScript" );
        ps_device = (Gfx_Status VAR_PTR) t_register( T_REG_DATA,
            device, NONEXISTENT_DEVICE, (char VAR_PTR) 0 );
/*
 *    DECODE THE OPTIONS AND ASSIGN PAGE & SETUP STRUCTURES
*/
        if ( ! set_options( options, (boolean VAR_PTR) &rotated, name ) )
            return (clip_area VAR_PTR) NULL;
      }
    else
      {
/*
 *    GET POSTSCRIPT PROCEDURES
 *          if no PostScript devices are registered
*/
        char VAR_PTR ch_ptr;
        if ( ! t_register( T_ISREGIST, (FILE *) 0, POSTSCRIPT_PRINTER,
             (char VAR_PTR) 0 ) )
          {
            if ( ( ps_header_fname = getenv( "G_PSPROC" ) ) )
              {
                if ( ( ps_file = fopen( ps_header_fname, "r" ) ) )
                  {
                    int proc_index;
/*
 *          read PostScript procedures and compile a directory
*/
                    ok = read_ps_procs( ps_file );
                    ok &= chain_ps_procs();
                    (void) fclose( ps_file );
                    if ( ! ok )
                        return (clip_area VAR_PTR) NULL;
/*
 *          locate all of the procedures called directly by the driver
*/

                    for ( proc_index = FIRST_PROC;
                        proc_index <= LAST_PROC && ( ok = (Procedure) 0 !=
                            ( proc_list[proc_index] = find_ps_proc(
                                (char VAR_PTR) proc_names[proc_index] ) ) );
                        proc_index++ );
                    if ( ! ok )
                        return (clip_area VAR_PTR) NULL;
		  }
                else
                    return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                     (char VAR_PTR) "Cannot open PostScript procedures file" );
              }
            else
                return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                   (char VAR_PTR) "No environment name G_PSPROC" );
          }
/*
 *    REGISTRATION
 *      ``device'' has never been registered, so allocate memory for the
 *      graphics status data and register the device as a PostScript printer .
*/
        ps_device = (Gfx_Status VAR_PTR)
            malloc( (unsigned int) sizeof(Gfx_Status) );
        if ( ! ps_device )
            return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "Cannot allocate memory" );
        if ( ! t_register( T_REGISTER, device,
                 POSTSCRIPT_PRINTER, (char VAR_PTR) ps_device ) )
          {
            (void) free( (char *) ps_device );
            ps_device = (Gfx_Status VAR_PTR) 0;
            return (clip_area VAR_PTR) make_error( ERR_CALL, name, mt_str );
          }
/*
 *      nothing has been downloaded yet
*/
        for ( ch_ptr = LOADED.procs;
              ch_ptr <= LOADED.procs + NO_PS_PROCS; *ch_ptr++ = FALSE );
        LOADED.asc_fnt = LOADED.sym_fnt = LOADED.matrix = FALSE;
/*
 *    DECODE THE OPTIONS AND ASSIGN PAGE & SETUP STRUCTURES
*/
        if ( ! set_options( options, (boolean VAR_PTR) &rotated, name ) )
          {
            if ( ! t_register( T_DEREGIST, OP_FILE,
                    POSTSCRIPT_PRINTER, (char VAR_PTR) ps_device ) )
                (void) make_error( ERR_WARN, name, mt_str );
            (void) free( (char *) ps_device );
            ps_device = (Gfx_Status VAR_PTR) 0;
            return (clip_area VAR_PTR) NULL;
          }
/*
 *    START THE POSTSCRIPT FILE
 *      Write structuring header comments to the PostScript file.
*/
        OP_FILE = device;
        ok &= ps_header();
/*
 *      reset character attributes and plotter status
*/
        (void) reset_font( (Text_Attribute VAR_PTR) &STROKE,
            MAX( PAGE.plot_size._x, PAGE.plot_size._y ) / 20.0 );
        if ( SETUP.built_in )
          {
            STROKE.font_ptr = (StrokeFont VAR_PTR) &built_font;
            PLOTTER.requir.ch_font = STD_FONT  | ASCII_FONT | BUILTIN_FONT;
          }
        else
          {
            STROKE.font_ptr = (StrokeFont VAR_PTR) &asc_font;
            PLOTTER.requir.ch_font = STD_FONT  | ASCII_FONT | PRIVATE_FONT;
          }
        PLOTTER.requir.dash      = SOLID_LINE;
        PLOTTER.requir.colour    = BLACK;
        PLOTTER.requir.thick     = 0;
        PLOTTER.requir.pen       = PAGE.ll_limit;
        PLOTTER.requir.cursor._x = PAGE.ll_limit._x;
        PLOTTER.requir.cursor._y = PAGE.ur_limit._y - PAGE.alph_height;
        PLOTTER.requir.ch_gap    = 0.0;
/*
 *          now fix the basic size of characters drawn by built_in font
*/
    built_font.font_hei = BUILT_HEIGHT_FACTOR * SETUP.tty_size;
    built_font.font_wid = BUILT_WIDTH_FACTOR * SETUP.tty_size;
    built_font.font_gap = BUILT_GAP_FACTOR * SETUP.tty_size;;
/*
 *      INITIALIZE OR RE-INITIALIZE THE POSTSCRIPT STATUS
 *      If required, set the imageable region
*/
    switch ( SETUP.paper_size )
      {
        case LETTER_PAPER : G_PUTS( "letter\n" ); break;
        case NOTE_PAPER :   G_PUTS( "note\n" );   break;
        case LEGAL_PAPER :  G_PUTS( "legal\n" );  break;
      }
/*
 *      dummy download TTY_size definition so that it cannot be overwritten
 *      by accidentally downloading it
*/
    LOADED.procs[ proc_list[ TTY_SIZE_DEF ]->pr_index ] = TRUE;
/*
 *      call the PostScript initialization procedure
 *      if necessary with rotated coordinate system
*/
    if ( ! LOAD_PS( INIT_PROC ) )
        return (clip_area VAR_PTR) NULL;
    (void) fprintf( OP_FILE, "%.1f %.1f %.2f %s\n", SETUP.tty_size,
        ( rotated ? PAGE.hclp_ptr->P_dimens._x : -1.0 ), SETUP.line_width,
        proc_list[INIT_PROC]->pr_name );
/*
 *      set the plotter clipping window to the hard clip limits
*/
        if ( ( VIEW_CLIP = def_2rclip( (coord_pair VAR_PTR) &PAGE.ll_limit,
                (coord_pair VAR_PTR) &PAGE.ur_limit,
                (coord_pair VAR_PTR) &PAGE.ll_limit ) ) )
            ok &= t_vwport( VIEW_CLIP );
        else
            return (clip_area VAR_PTR) make_error( ERR_CALL, name, mt_str );
      }
/*
 *      SET THE PLOTTER STATUS
 *      and put pen in lower left corner.
*/
    PLOTTER.actual.dash = SOLID_LINE;
    PLOTTER.actual.colour = BLACK;
    PLOTTER.actual.thick  = 0;
    PLOTTER.actual.ch_font = LAST_FONT + 1;
    PLOTTER.actual.ch_gap = 0.0;
    ok &= ps_pmdraw( (coord_pair VAR_PTR) &PAGE.ll_limit, MOVE_PROC );
    if ( ! ok && G_ERROR() )
        return (clip_area VAR_PTR) make_error( ERR_WARN, name, err_write );
    return VIEW_CLIP;
  }


static boolean set_options( options, rotated, caller )
/*
 *    Decode the options and assign the page & setup structures .
 *
*/
char VAR_PTR options;   /* options string */
boolean VAR_PTR rotated;/* is the coordinate system rotated ? */
char VAR_PTR caller;    /* name of calling function */
  {
    double half_width;
/*
 *     Paper size tables for each printer.
 * N.B. at least one list of scaling points is required for each printer
*/
    static Hclp_Elem
      hclp_apple[] = {       /* ***** Apple Laserwriter ***** */
/*                              size           lower left     upper right */
/*                            of paper           margin         margin    */
{ MET_PAPER,DEFAULT_PAPER, {595.28, 841.89}, {12.96, 9.36}, {577.20, 832.80} },
{ MET_PAPER, LETTER_PAPER, {595.28, 841.89}, {14.16, 7.92}, {597.60, 784.32} },
{ MET_PAPER, NOTE_PAPER,   {595.28, 841.89}, {26.40,30.48}, {567.60, 811.44} },
{ US_PAPER, DEFAULT_PAPER, {612.00, 792.00}, {18.00, 2.88}, {594.00, 789.12} },
{ US_PAPER,  LETTER_PAPER, {612.00, 792.00}, {18.00, 2.88}, {594.00, 789.12} },
{ US_PAPER,  NOTE_PAPER,   {612.00, 792.00}, {29.16,30.24}, {582.84, 761.76} },
{ US_PAPER,  LEGAL_PAPER,  {612.00,1008.00}, {64.08,36.00}, {547.92, 972.00} } },

      hclp_stand[] = {       /* ***** standard PostScript ***** */
{ MET_PAPER, VIRTUAL_PAPER, {595.28, 841.89}, { 0.0, 0.0}, {595.28, 841.89} },
{ US_PAPER,  VIRTUAL_PAPER, {612.00, 792.00}, { 0.0, 0.0}, {612.00, 792.00} } };
/*
 *  plotter options
 *        option string   min. abbr.  variable set    value to assign
 *                          length
 *  an assigned value of NOT_SET signals that the next word in the option
 *  string is the value to be assigned
*/
    static Setup_List st_opt;
    static Option_Elem option_list[] = {
{ "METRIC",              2, OPT_SHORT,  (char VAR_PTR) &st_opt.paper_type, MET_PAPER },
{ "US",                  1, OPT_SHORT,  (char VAR_PTR) &st_opt.paper_type, US_PAPER },
{ "LETTER_PAPER",        3, OPT_SHORT,  (char VAR_PTR) &st_opt.paper_size, LETTER_PAPER },
{ "NOTE_PAPER",          1, OPT_SHORT,  (char VAR_PTR) &st_opt.paper_size, NOTE_PAPER },
{ "LEGAL_PAPER",         3, OPT_SHORT,  (char VAR_PTR) &st_opt.paper_size, LEGAL_PAPER },
{ "WIDE",                1, OPT_SHORT,  (char VAR_PTR) &st_opt.paper_dirn, WIDE_PAPER },
{ "TALL",                2, OPT_SHORT,  (char VAR_PTR) &st_opt.paper_dirn, TALL_PAPER },
{ "BUILT_IN_FONTS",      1, OPT_SHORT,  (char VAR_PTR) &st_opt.built_in,   TRUE },
{ "PRIVATE_FONTS",       1, OPT_SHORT,  (char VAR_PTR) &st_opt.built_in,   FALSE },
{ "APPLE_LASERWRITER",   1, OPT_SHORT,  (char VAR_PTR) &st_opt.plotter_id, LASERWRITER },
{ "MONOCHROME_STANDARD", 2, OPT_SHORT,  (char VAR_PTR) &st_opt.plotter_id, MONOCHROME_STANDARD},
{ "COLOUR_STANDARD",     1, OPT_SHORT,  (char VAR_PTR) &st_opt.plotter_id, COLOUR_STANDARD },
{ "LINE_WIDTH",          2, OPT_DOUBLE, (char VAR_PTR) &st_opt.line_width, OPT_NUM_ARG },
{ "TTY_SIZE",            2, OPT_DOUBLE, (char VAR_PTR) &st_opt.tty_size,   OPT_NUM_ARG },
{ (char VAR_PTR) 0,      0, 0, (char VAR_PTR) 0,   0 } };
/*
 *          SET THE OPTIONS
 *
 *          clear any previous plotter setup
*/
    st_opt.plotter_id = st_opt.paper_type = st_opt.paper_size =
        st_opt.paper_dirn = st_opt.built_in = NOT_SET;
    st_opt.line_width = st_opt.tty_size = NOT_SET;
/*
 *           decode the options list
*/
    if ( options )
        if ( ! t_parse_options( options, (Option_Elem VAR_PTR) option_list,
                caller ) )
            return FALSE;
/*
 *      ensure the default setup values if they are NOT_SET and
 *      remove any inconsistencies in the setup values ( if possible )
*/
    if ( st_opt.plotter_id == NOT_SET )
        st_opt.plotter_id = LASERWRITER;
    if ( st_opt.paper_type == NOT_SET )
        st_opt.paper_type = MET_PAPER;
    if ( st_opt.paper_dirn == NOT_SET )
        st_opt.paper_dirn = WIDE_PAPER;
    if ( st_opt.built_in == NOT_SET )
        st_opt.built_in = TRUE;
    if ( st_opt.line_width == NOT_SET )
        st_opt.line_width = LINE_WIDTH;
    else
        st_opt.line_width = LIMIT( st_opt.line_width, 0.1, 10.0 );
    if ( st_opt.tty_size == NOT_SET )
        st_opt.tty_size = DEF_TTY_SIZE;
    else
        st_opt.tty_size = LIMIT( st_opt.tty_size, 5.0, 20.0 );
/*
 *      Now copy the static options selected to the Graphics Status area
*/
    SETUP = st_opt;
/*
 *      This driver always behaves as though writing to a disk file
 *      i.e. it is non-interactive .
 *
 *          set the hard clip limits according to the plotter type
*/
    switch ( SETUP.plotter_id )
      {
        case MONOCHROME_STANDARD :
            SETUP.paper_size = VIRTUAL_PAPER;
            PAGE.is_monochr = TRUE;
            PAGE.hclp_ptr = find_hclp( (Hclp_Elem VAR_PTR) hclp_stand, 2 );
            break;
        case COLOUR_STANDARD :
            SETUP.paper_size = VIRTUAL_PAPER;
            PAGE.is_monochr = TRUE;
            PAGE.hclp_ptr = find_hclp( (Hclp_Elem VAR_PTR) hclp_stand, 2 );
            break;
        case LASERWRITER :
        default :
            if ( SETUP.paper_size == NOT_SET )
                SETUP.paper_size = DEFAULT_PAPER;
            PAGE.is_monochr = TRUE;
            PAGE.hclp_ptr = find_hclp( (Hclp_Elem VAR_PTR) hclp_apple, 5 );
            break;
      }
    if ( ! PAGE.hclp_ptr )
        return make_error( ERR_ABORT, caller,
            (char VAR_PTR) "Illegal paper size" );
/*
 *          complete the "plotter unit" setup info and plotting limits
*/
    PAGE.plot_size._x =
            PAGE.hclp_ptr->Pur_clip._x - PAGE.hclp_ptr->Pll_clip._x;
    PAGE.plot_size._y =
            PAGE.hclp_ptr->Pur_clip._y - PAGE.hclp_ptr->Pll_clip._y;
    PAGE.alph_width = BUILT_BOX_RATIO * ( PAGE.alph_height = st_opt.tty_size );
    *rotated = ( SETUP.paper_dirn == WIDE_PAPER ) ^
            ( PAGE.plot_size._x > PAGE.plot_size._y );
    if ( *rotated )
      {
        PAGE.ll_limit._x =  PAGE.hclp_ptr->Pll_clip._y;
        PAGE.ll_limit._y =  PAGE.hclp_ptr->P_dimens._x -
                            PAGE.hclp_ptr->Pur_clip._x;
        PAGE.ur_limit._x =  PAGE.hclp_ptr->Pur_clip._y;
        PAGE.ur_limit._y =  PAGE.hclp_ptr->P_dimens._x -
                            PAGE.hclp_ptr->Pll_clip._x;
        PAGE.alph_columns = PAGE.plot_size._y / PAGE.alph_width;
        PAGE.alph_lines   = PAGE.plot_size._x / PAGE.alph_height;
      }
    else
      {
        PAGE.ll_limit = PAGE.hclp_ptr->Pll_clip;
        PAGE.ur_limit = PAGE.hclp_ptr->Pur_clip;
        PAGE.alph_columns = PAGE.plot_size._x / PAGE.alph_width;
        PAGE.alph_lines   = PAGE.plot_size._y / PAGE.alph_height;
      }
    half_width = SETUP.line_width / 2.0;
    PAGE.ll_limit._x += half_width;
    PAGE.ll_limit._y += half_width;
    PAGE.ur_limit._x -= half_width;
    PAGE.ur_limit._y -= half_width;
    return TRUE;
  }


static boolean t_end()
/*
 *      End of graphics . Delete character fonts and frees the memory allocated
 *  to them, dumps the bitmap image to the paper the graphics screen and
 *  writes the trailer comments .
*/
  {
    boolean ok;
    char VAR_PTR name = "t_end";
    START_UP( name, boolean );
/*
 *              dump the bitmap image to paper
*/
    ok = t_newpag();
/*
 *              download and execute PostScript cleanup procedure
*/
    if ( ! LOAD_PS( FINISH_PROC ) )
        return FALSE;
    (void) G_PUTS( (char *) proc_list[FINISH_PROC]->pr_name );
    (void) G_PUTC( '\n');
/*
 *              write trailer comments
*/
    if ( ! ps_trailer() || G_ERROR() )
        return make_error( ERR_ABORT, name, err_write );
/*
 *      Deregister the device and free the memory used by the graphics status
 *      data .
*/
    if ( ! t_register( T_DEREGIST, OP_FILE,
            POSTSCRIPT_PRINTER, (char VAR_PTR) ps_device ) )
        (void) make_error( ERR_WARN, name, mt_str );
    (void) free( (char *) ps_device );
    ps_device = (Gfx_Status VAR_PTR) 0;
/*
 *      Delete character fonts and frees the memory allocated to them
*/
    if ( asc_font.font_got )
        (void) clear_font( (StrokeFont VAR_PTR) &asc_font );
    if ( sym_font.font_got )
        (void) clear_font( (StrokeFont VAR_PTR) &sym_font );
/*
 *              If there no other PostScript devices are registered, free
 *              memory occupied by PostScript procs and force reread PostScript
 *              procs by clearing the first procedure pointer .
*/
    if ( ! t_register( T_ISREGIST, (FILE *) 0, POSTSCRIPT_PRINTER,
             (char VAR_PTR) 0 ) )
      {
        ok &= free_ps_procs();
        proc_dir[0] = (Proc_entry VAR_PTR) NULL;
      }
    return ok;
  }


static boolean t_vwport( clip_ptr )
/*
 *      Set the clipping limits for the viewport . The old limits are replaced .
 * The `clip_ptr' must point to  clip_area which has been defined by
 * `def_2rclip()' . The function fails if
 *   a) clip_ptr is a NULL pointer
 *   b) clip_ptr points to something which is not a structure for a two
 *      dimensional rectangular clipping region .
 *   c) cannot write to the file or plotter
 *   d) the new clipping limits are beyond the limits of the plotter
 *
 * Argument
*/
clip_area VAR_PTR clip_ptr;
  {
    char VAR_PTR name = "t_vwport";
    CATCH_NULL( clip_ptr, name, boolean );
    START_UP( name, boolean );
    if ( clip_ptr->id_struct == CLIP2R_ID )
      {
        if ( min_clip( clip_ptr, _x ) >= PAGE.ll_limit._x &&
             max_clip( clip_ptr, _x ) <= PAGE.ur_limit._x &&
             min_clip( clip_ptr, _y ) >= PAGE.ll_limit._y &&
             max_clip( clip_ptr, _y ) <= PAGE.ur_limit._y )
          {
/*
 *          download and call the PostScript procedure
*/
            double half_width = SETUP.line_width / 2.0;
            if ( ! LOAD_PS( VWPORT_PROC ) )
                return FALSE;
            (void) fprintf( OP_FILE, "%.1f %.1f %.1f %.1f %s\n",
                min_clip( clip_ptr, _x ) - half_width,
                min_clip( clip_ptr, _y ) - half_width,
                max_clip( clip_ptr, _x ) + half_width,
                max_clip( clip_ptr, _y ) + half_width,
                proc_list[VWPORT_PROC]->pr_name );
            VIEW_CLIP = clip_ptr;
            if ( G_ERROR() )
                return make_error( ERR_ABORT, name, err_write );
            else
                return TRUE;
          }
        return make_error( ERR_ABORT, name,
            (char VAR_PTR) "Outside device limits" );
      }
    return make_error( ERR_ABORT, name,
        (char VAR_PTR) "Invalid pointer" );
  }


static boolean t_colour( ipen )
/*
 *    Selects specified colour ( or pen ) .
 *
 * Integers from 0 to 8 are valid values of `ipen' .
 * Because the Visual 500 terminal is monochrome, colours are simulated with
 * various dashed line styles . The only true ``colours'' available are dark
 * The returned value is TRUE if  0 <= ipen <= 8 . If these limits are
 * exceeded, the value is wrapped around into the range 1 to 8 .
 *
 * If the plotter has fewer than 8 pens, the value of `ipen' is further
 * wrapped around the number of available pens .
 *
 *    IPEN  COLOUR     IPEN  COLOUR     IPEN  COLOUR
 *     0    no pen      1    black       2    red
 *     3    orange      4    yellow      5    green
 *     6    blue        7    violet      8    brown
 *
 *   Argument
*/
int ipen;                /* pen number or colour */
  {
    START_UP( (char VAR_PTR) "t_colour", boolean );
    PLOTTER.requir.colour = ( abs( ipen ) - 1 ) % 8 + 1;
    return 0 <= ipen && ipen <= 8;
  }


static boolean t_dashed( style )
/*
 *      Set the dashed line style . The returned value is TRUE if
 * 0 <= style <= 8 . If these limits are exceeded, the value is wrapped
 * around into the range 1 to 8 .
 *      A value `style == 8' will not produce a `3-dots long-dash' line
 * style ( which is the the device driver standard ) --- very short dashes
 * are used instead.
 *
 *    STYLE   SCREEN
 *          LINE STYLE
 *     0    plot vector endpoints only
 *     1    solid
 *     2    dotted
 *     3    short dash
 *     4    long dash
 *     5    dot dash
 *     6    short-dash long-dash
 *     7    2-short-dashes long-dash
 *     8    very short dashes
 *
*/
int style;        /* dashed line style */
  {
    START_UP( (char VAR_PTR) "t_dashed", boolean );
    PLOTTER.requir.dash = ( abs( style ) - 1 ) % 8 + 1;
    return 0 <= style && style <= 8;
  }


static boolean t_thick( width )
/*
 * Set the ``width'' of lines and text . The returned value is TRUE
 * if 0 <= thickness.
*/
int width;      /* line thickness */
  {
    START_UP( (char VAR_PTR) "t_thick", boolean );
    PLOTTER.requir.thick = width;
    return 0 <= width;
  }


/* ************************************************************************ */
/*                                                                          */
/*                          Turn graphics on or off                         */
/*                                                                          */
/* ************************************************************************ */


#define t_gr_off t_dummy
/*
 *      Turn the plotter off . The PostScript output is always a file, and so
 *  this function need not do anything . There is no provision for eavesdrop
 *  mode . This function is simply a call to ``t_dummy()'' which does nothing .
*/


#define t_gr_on t_dummy
/*
 *      Turn the plotter on . The PostScript output is always a file, and so
 *  this function need not do anything . There is no provision for eavesdrop
 *  mode . This function is simply a call to ``t_dummy()'' which does nothing .
*/


/* ************************************************************************ */
/*                                                                          */
/*           Change Data Level, Visible, Invisible, Revers, Replace         */
/*  Plotter (hardcopy) devices only draw in VISIBLE mode, and so calls to   */
/*  ``t_visib()'', ``t_invis()'', ``t_revers()'' & ``t_replac()'' are all   */
/*  calls to a dummy function which does nothing except return TRUE .       */
/*                                                                          */
/* ************************************************************************ */


#define t_visib  t_dummy     /* Draw visible vectors   */
#define t_invis  t_dummy     /* Draw invisible vectors */
#define t_revers t_dummy     /* Reverse the pixel dots */
#define t_replac t_dummy     /* Replace pixel contents with new data */

static boolean t_dummy()
/*
 *  This function does nothing -- properly !!
*/
  {
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                         Miscellaneous stuff                              */
/*                                                                          */
/* ************************************************************************ */


static boolean t_newpag()
/*
 * Prints the page .
*/
  {
    char VAR_PTR name = "t_new_pag";
    double half_width = SETUP.line_width / 2.0;
    START_UP( name, boolean );
    if ( ps_device->empty_page )
        return TRUE;
/*
 *          call the PostScript newpage procedure
*/
    if ( ! LOAD_PS( NEWPAG_PROC ) )
        return FALSE;
    (void) fprintf( OP_FILE, "%s\n", proc_list[NEWPAG_PROC]->pr_name );
    ps_device->empty_page = TRUE;
/*
 *          call the PostScript procedure viewport to restore the
 *          clipping path -- which is the device default after showpage
*/
    if ( ! LOAD_PS( VWPORT_PROC ) )
        return FALSE;
    (void) fprintf( OP_FILE, "%.1f %.1f %.1f %.1f %s\n",
        min_clip( VIEW_CLIP, _x ) - half_width,
        min_clip( VIEW_CLIP, _y ) - half_width,
        max_clip( VIEW_CLIP, _x ) + half_width,
        max_clip( VIEW_CLIP, _y ) + half_width,
        proc_list[VWPORT_PROC]->pr_name );
    if ( G_ERROR() )
        return make_error( ERR_ABORT, name, err_write );
    return TRUE;
  }


static boolean t_newscr()
/*
 * If output is to stdout, clear the alphanumeric screen by the simple
 * expedient of writing 50 newline characters to the screen ; otherwise do
 * nothing .
*/
  {
    char VAR_PTR name = "t_newscr";
    START_UP( name, boolean );
    if ( OP_FILE == stdout )
      {
        int lf_count = 0;
        while( lf_count++ <= 50 )
            (void) G_PUTC( '\n' );
        if ( G_ERROR() )
            return make_error( ERR_WARN, name, err_write );
        return TRUE;
      }
    return make_error( ERR_WARN, name, lab_inactive );
  }


static boolean t_bell()
/*
 *      If output is to stdout, ring bell ; otherwise do nothing .
*/
  {
    char VAR_PTR name = "t_bell";
    START_UP( name, boolean );
    if ( OP_FILE == stdout )
      {
        (void) G_PUTC( BEL );
        if ( G_ERROR() )
            return make_error( ERR_WARN, name, err_write );
        return TRUE;
      }
    return make_error( ERR_WARN, name, lab_inactive );
  }


static boolean t_wait()
/*
 *  If output is to stdout, waits for the user to press the carriage return ;
 *  otherwise do nothing ( but don't generate an error message ) .
 *  The returned value is TRUE unless it cannot write to stdout .
*/
  {
    char VAR_PTR name = "t_wait";
    START_UP( name, boolean );
    if ( OP_FILE == stdout )
      {
        (void) G_FLUSH();
        if ( G_ERROR() )
            return make_error( ERR_WARN, name, err_write );
        while( getchar() != '\n' );
      }
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                        Print terminal status                             */
/*                                                                          */
/* ************************************************************************ */


static char VAR_PTR t_status( attribute, value )
/*
 *  Put status attribute 'attribute' into '*value' which may be a character
 *  string or a coordinate pair .
*/
int attribute;
char VAR_PTR value;
  {
    char VAR_PTR name = "t_status";
/*
 *      Names of available PostScript devices
*/
    static char VAR_PTR ID_string[] = {   "Apple Laserwriter PostScript",
        "Monochrome standard PostScript", "Colour standard PostScript" };
    CATCH_NULL( value, name, char VAR_PTR );
    if ( attribute == T_ST_SCREEN )
        *( (boolean VAR_PTR) value ) = FALSE;
    else if ( attribute == T_ST_CATEG )
        *( (int VAR_PTR) value ) = T_ST_FILE;
    else
      {
        START_UP( name, char VAR_PTR );
        switch( attribute )
          {
            case T_ST_DEVICE:
                (void) t_strcpy( value, ID_string[SETUP.plotter_id] );
                break;
            case T_ST_MONO:
                *( (boolean VAR_PTR) value ) = PAGE.is_monochr;
                break;
            case T_ST_CLIP:
                *( (clip_area VAR_PTR VAR_PTR) value ) = VIEW_CLIP;
                break;
            case T_ST_MAX:
                ( (coord_pair VAR_PTR) value)->_x = PAGE.ur_limit._x;
                ( (coord_pair VAR_PTR) value)->_y = PAGE.ur_limit._y;
                break;
            case T_ST_MIN:
                ( (coord_pair VAR_PTR) value)->_x = PAGE.ll_limit._x;
                ( (coord_pair VAR_PTR) value)->_y = PAGE.ll_limit._y;
                break;
            case T_ST_CHAR:
                ( (coord_pair VAR_PTR) value)->_x = PAGE.alph_width;
                ( (coord_pair VAR_PTR) value)->_y = PAGE.alph_height;
                break;
            case T_ST_LINES:
                ( (coord_pair VAR_PTR) value)->_x = PAGE.alph_columns;
                ( (coord_pair VAR_PTR) value)->_y = PAGE.alph_lines;
                break;
            case T_ST_PIXSIZ:
                ( (coord_pair VAR_PTR) value)->_x =
                    ( (coord_pair VAR_PTR) value)->_y = POINT_SIZE;
                break;
            case T_ST_A_CURS:
                ( (coord_pair VAR_PTR) value)->_x = PLOTTER.actual.cursor._x;
                ( (coord_pair VAR_PTR) value)->_y = PLOTTER.actual.cursor._y;
                break;
            case T_ST_R_CURS:
                ( (coord_pair VAR_PTR) value)->_x = PLOTTER.requir.cursor._x;
                ( (coord_pair VAR_PTR) value)->_y = PLOTTER.requir.cursor._y;
                break;
            case T_ST_A_PEN:
                ( (coord_pair VAR_PTR) value)->_x = PLOTTER.actual.pen._x;
                ( (coord_pair VAR_PTR) value)->_y = PLOTTER.actual.pen._y;
                break;
            case T_ST_R_PEN:
                ( (coord_pair VAR_PTR) value)->_x = PLOTTER.requir.pen._x;
                ( (coord_pair VAR_PTR) value)->_y = PLOTTER.requir.pen._y;
                break;
            default:
                return (char VAR_PTR) make_error( ERR_ABORT, name,
                    (char VAR_PTR) "unknown attribute" );
          }
      }
    return value;
  }


static boolean t_pr_ter( out_file )
FILE *out_file;
  {
    char VAR_PTR name = "t_pr_ter";
    CATCH_NULL( out_file, name, boolean );
    START_UP( name, boolean );
    (void) fprintf( out_file, " TERMINAL STATUS\n" );
    (void) fprintf( out_file,
        "               ---requir---     ---actual---\n" );
    (void) fprintf( out_file, "colour      %15d  %15d\n",
        PLOTTER.requir.colour, PLOTTER.actual.colour );
    (void) fprintf( out_file, "dash style  %15d  %15d\n",
        PLOTTER.requir.dash, PLOTTER.actual.dash );
    (void) fprintf( out_file, "font        %s  %s\n",
        font_names[PLOTTER.requir.ch_font],
        font_names[PLOTTER.actual.ch_font] );
    (void) fprintf( out_file, "cursor      (%6.1f,%6.1f)  (%6.1f,%6.1f)\n",
        PLOTTER.requir.cursor._x, PLOTTER.requir.cursor._y,
        PLOTTER.actual.cursor._x, PLOTTER.actual.cursor._y );
    (void) fprintf( out_file, "pen         (%6.1f,%6.1f)  (%6.1f,%6.1f)\n",
        PLOTTER.requir.pen._x, PLOTTER.requir.pen._y,
        PLOTTER.actual.pen._x, PLOTTER.actual.pen._y );
    (void) fprintf( out_file, "char spacing %15.3f  %15.3f\n",
        PLOTTER.requir.ch_gap, PLOTTER.actual.ch_gap );
    if ( ferror( out_file ) )
        return make_error( ERR_WARN, name,
            (char VAR_PTR) "Cannot write terminal status to file" );
    return TRUE;
  }


static boolean t_boxfil( lower_left, upper_right, style )
/*
 *      draw a rectangular box on the screen with lower left corner at
 * `*lower_left' and upper right corner at `*upper_right', and fill the box
 * with a hatching of type `style' . The hatching styles are :
 *
 *      0 = empty, box outline only      5 = horizontal lines
 *      1 = fill the box completely      6 = vertical lines
 *      2 = 50% `grey' fill              7 = diagonal cross hatch
 *      3 = NW-SE diagonal lines         8 = horizontal-vertical cross
 *      4 = NE-SW diagonal lines                 hatch
 * The box outline is clipped to the viewport limits .
*/
coord_pair VAR_PTR lower_left,  /* lower left corner of the box     */
           VAR_PTR upper_right; /* upper right corner of the box    */
int style;                      /* hatching style                   */
  {
    coord_pair ll_corner, ur_corner;
    char VAR_PTR name = "t_boxfil";
    CATCH_NULL( lower_left && upper_right, name, boolean );
    START_UP( name, boolean );
/*
 *      clip the rectangle to the viewport limits
*/
    ll_corner._x = LIMIT( lower_left->_x,
        min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
    ll_corner._y = LIMIT( lower_left->_y,
        min_clip( VIEW_CLIP, _y ), max_clip( VIEW_CLIP, _y ) );
    ur_corner._x = LIMIT( upper_right->_x,
        min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
    ur_corner._y = LIMIT( upper_right->_y,
        min_clip( VIEW_CLIP, _y ), max_clip( VIEW_CLIP, _y ) );
/*
 *      ensure correct ordering of coordinates
*/
    if ( ll_corner._x > ur_corner._x )
        SWAPD( ll_corner._x , ur_corner._x );
    if ( ll_corner._y > ur_corner._y )
        SWAPD( ll_corner._y , ur_corner._y );
/*
 *      draw the box
*/
    if ( fabs( ur_corner._x - ll_corner._x ) > 0.0 &&
         fabs( ur_corner._y - ll_corner._y ) > 0.0 )
      {
        if ( ! set_colour() )
            return make_error( ERR_ABORT, name, err_write );
        style = ( abs( style ) - 1 ) % 8 + 1;
        if ( ! LOAD_PS( BOX_PROCS + style ) )
            return FALSE;
        (void) fprintf( OP_FILE, "%.1f %.1f %.1f %.1f %s\n",
            ll_corner._x, ll_corner._y, ur_corner._x, ur_corner._y,
            proc_list[BOX_PROCS+style]->pr_name );
        PLOTTER.actual.cursor = PLOTTER.actual.pen = ll_corner;
        if ( G_ERROR() )
            return make_error( ERR_ABORT, name, err_write );
        PLOTTER.actual.dash = SOLID_LINE;       /* done by the PostScript */
        ps_device->empty_page = FALSE;
        return TRUE;
      }
    return FALSE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Vector Drawing                                 */
/*                                                                          */
/* ************************************************************************ */


static boolean t_vecmov( point )
/*
 *      move the line drawing pen to `point'
*/
coord_pair VAR_PTR point;
  {
    char VAR_PTR name = "t_vecmov";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    PLOTTER.requir.pen = *point;
    return move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
  }


static boolean t_vecdrw( point )
/*
 *      Draw from the previous pen position to `point' according to the
 *  required colour, dashed line style and dots-level . Dashed line style
 *  takes precedence over colours . Lines are clipped to the viewport
 *  boundaries .
*/
coord_pair VAR_PTR point;
   {
    boolean ok = TRUE;       /* success flag */
    char VAR_PTR name = "t_vecdrw";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    PLOTTER.requir.pen = *point;
/*
 *          If required dashed line style is 0, plot the point if it
 *          is in the viewport .
*/
    if ( ! PLOTTER.requir.dash )
      {
        ok &= move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
        if ( PLOTTER.requir.colour &&
             min_clip( VIEW_CLIP, _x ) <= point->_x &&
             max_clip( VIEW_CLIP, _x ) >= point->_x &&
             min_clip( VIEW_CLIP, _y ) <= point->_y &&
             max_clip( VIEW_CLIP, _y ) >= point->_y )
          {
/*
 *          If necessary, select the correct dashed line style and pen colour .
*/
            ok = set_colour();
            if ( PLOTTER.actual.dash != SOLID_LINE )
                if ( ok &= ps_style( SOLID_LINE, DASH_PROC ) )
                    PLOTTER.actual.dash = SOLID_LINE;
/*
 *          plot the point
*/
            ok &= ps_pmdraw( point, POINT_PROC );
            ps_device->empty_page = FALSE;
          }
      }
/*
 *          Need to draw a line . First clip the endpoints .
*/
    else if ( clip( (coord_triple VAR_PTR) point, VIEW_CLIP ) )
      {
/*
 *          If necessary, select the correct dashed line style and pen colour .
*/
            ok = set_colour();
            if ( PLOTTER.actual.dash != PLOTTER.requir.dash )
                if ( ok &= ps_style( PLOTTER.requir.dash, DASH_PROC ) )
                    PLOTTER.actual.dash = PLOTTER.requir.dash;
/*
 *          Now we are all set up, we can worry about drawing the vector
 *          First check if the pen is starting from the correct place .
 *          -- if not, move to it.
*/
        if ( fabs( PLOTTER.actual.pen._x - first_clip(VIEW_CLIP,_x) ) > 0.01 ||
             fabs( PLOTTER.actual.pen._y - first_clip(VIEW_CLIP,_y) ) > 0.01 )
          ok &= ps_pmdraw( (coord_pair VAR_PTR) FIRST_clip( VIEW_CLIP ),
              MOVE_PROC );
/*
 *          draw to the clipped endpoint
*/
        ok &= ps_pmdraw( (coord_pair VAR_PTR) LAST_clip( VIEW_CLIP ),
            DRAW_PROC );
        ps_device->empty_page = FALSE;
      }
    if ( ! ok )
        (void) make_error( ERR_WARN, name, err_write );
    return ok;
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Cross Hairs Routines                           */
/*                                                                          */
/* ************************************************************************ */


static boolean t_xhrget( point, char_pressed )
/*
 * Is inactive. Does nothing .
*/
coord_pair VAR_PTR point;   /* pointer to the coordinates of the cross-hairs */
char VAR_PTR char_pressed;  /* character pressed to identify `point'         */
  {
    return make_error( ERR_WARN, (char VAR_PTR) "t_xhrget", lab_inactive );
  }


static boolean t_xhrput( point )
/*
 * Is inactive. Does nothing .
*/
coord_pair VAR_PTR point; /* pointer to the coordinates of the cross-hairs */
  {
    return make_error( ERR_WARN, (char VAR_PTR) "t_xhrput", lab_inactive );
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Text Writing                                   */
/*                                                                          */
/* ************************************************************************ */


static boolean t_txtmov( point )
/*
 *    Move the text cursor to coordinate `point' . The returned result is
 *  TRUE only if the next printable character is within the hardclip limits .
 *
 * Argument
*/
register coord_pair VAR_PTR point;
  {
    char VAR_PTR name = "t_txtmov";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    PLOTTER.requir.cursor = *point;
    return point->_x >  PAGE.ll_limit._x &&
           point->_x <= PAGE.ur_limit._x - PAGE.alph_width &&
           point->_y >  PAGE.ll_limit._y &&
           point->_y <= PAGE.ur_limit._y - PAGE.alph_height;
  }


static boolean t_txtcur( row, column )
/*
 *    Move the text cursor to text row `row' and text column `column' .
 *    The text row is numbered from the top of the screen . The returned
 *    result is TRUE only if the next printable character is within screen
 *    limits . The row and column positions are calculated for the default
 *    alphagraphics character size .
 *
 * Argument
*/
int row, column;
  {
    START_UP( (char VAR_PTR) "t_txtcur", boolean );
    PLOTTER.requir.cursor._x = PAGE.ll_limit._x + PAGE.alph_width * column;
    PLOTTER.requir.cursor._y = PAGE.ur_limit._y - PAGE.alph_height *
        ( row + 1 );
    return 0 <= column  &&  column <= PAGE.alph_columns  &&
           0 <= row     &&  row    <= PAGE.alph_lines;
  }


static boolean t_txtset( attribute, value )
/*
 *  Set the stroke character attribute `attribute' to `*value' .
 *  If `attribute' is negative, the current value of the attribute is returned
 *  to `*value' . This function fails if the `attribute' cannot be recognized or
 *  `value' is a null pointer . Appropriate error messages are generated .
*/
int attribute;
char VAR_PTR value;
  {
    char VAR_PTR name = "t_txtset";
    double VAR_PTR dvalue;    /* pointer to a double `value' */
    dvalue = (double VAR_PTR) value;
    if ( attribute != T_CH_RESET  && attribute != T_CH_BOLD  &&
             attribute != T_CH_NOBOLD && attribute != T_ASCII    &&
             attribute != T_SYMBOL )
        CATCH_NULL( value, name, boolean );
    START_UP( name, boolean );
    switch( attribute )
      {
        case  T_CH_RESET :
            (void) reset_font( (Text_Attribute VAR_PTR) &STROKE,
                MAX( PAGE.plot_size._x, PAGE.plot_size._y ) / 20.0 );
            if ( SETUP.built_in )
              {
                STROKE.font_ptr = (StrokeFont VAR_PTR) &built_font;
                PLOTTER.requir.ch_font = STD_FONT  | ASCII_FONT | BUILTIN_FONT;
              }
            else
              {
                STROKE.font_ptr = (StrokeFont VAR_PTR) &asc_font;
                PLOTTER.requir.ch_font = STD_FONT  | ASCII_FONT | PRIVATE_FONT;
              }
            break;
        case  T_CH_HEIGHT :
            if ( STROKE.height != *dvalue )
              {
                STROKE.height = LIMIT( *dvalue, 3.0,
                    MIN( PAGE.plot_size._x, PAGE.plot_size._y ) );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_HEIGHT :
            *dvalue = STROKE.height;
            break;
        case  T_CH_ASPECT :
            if ( STROKE.aspect != *dvalue )
              {
                STROKE.aspect = LIMIT( *dvalue, 0.2, 5.0 );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_ASPECT :
            *dvalue = STROKE.aspect;
            break;
        case  T_CH_SLANT :
            if ( STROKE.slant != *dvalue )
              {
                STROKE.slant = LIMIT( *dvalue, -80.0, 80.0 ) * PI / 180.0;
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_SLANT :
            *dvalue = STROKE.slant * 180.0 / PI;
            break;
        case  T_CH_ANGLE :
            if ( STROKE.angle != *dvalue )
              {
                STROKE.angle = PI / 180.0 * ( *dvalue -
                    360.0 * (double) (int) ( *dvalue / 360.0 ) );
                STROKE.angle = STROKE.angle < -PI ? STROKE.angle + 2.0 * PI :
                    ( STROKE.angle > PI ? STROKE.angle-2.0*PI : STROKE.angle );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_ANGLE :
            *dvalue = STROKE.angle * 180.0 / PI;
            break;
        case  T_CH_GAP :
            STROKE.gap = LIMIT( *dvalue, 0.05, 5.0 );
            break;
        case -T_CH_GAP :
            *dvalue = STROKE.gap;
            break;
        case  T_CH_ALIGN :
            STROKE.align.vert  = TOLOWER( *value );
            value++;
            STROKE.align.horiz = TOLOWER( *value );
            if ( STROKE.align.horiz != 'l' && STROKE.align.horiz != 'c' &&
                    STROKE.align.horiz != 'r' )
                STROKE.align.horiz = 'l';
            if ( STROKE.align.vert != 'b' && STROKE.align.vert != 'c' &&
                    STROKE.align.vert != 't' )
                STROKE.align.vert = 'b';
            break;
        case -T_CH_ALIGN :
            *value++ = STROKE.align.vert;
            *value++ = STROKE.align.horiz;
            *value   = NUL;
            break;
        case T_ASCII :
            if ( SETUP.built_in )
              {
                if ( STROKE.font_ptr != (StrokeFont VAR_PTR) &built_font )
                  {
                    STROKE.font_ptr = (StrokeFont VAR_PTR) &built_font;
                    PLOTTER.requir.ch_font = ( PLOTTER.requir.ch_font &
                        ~( CHARSET_MASK | BUILTIN_MASK ) ) | ASCII_FONT |
                        BUILTIN_FONT;
                    STROKE.changed = TRUE;
                  }
              }
            else
              {
                if ( STROKE.font_ptr != (StrokeFont VAR_PTR) &asc_font )
                  {
                    STROKE.font_ptr = (StrokeFont VAR_PTR) &asc_font;
                    PLOTTER.requir.ch_font = ( PLOTTER.requir.ch_font &
                        ~( CHARSET_MASK | BUILTIN_MASK ) ) | ASCII_FONT |
                        PRIVATE_FONT;
                    STROKE.changed = TRUE;
                  }
              }
            break;
        case T_SYMBOL :
            if ( STROKE.font_ptr != (StrokeFont VAR_PTR) &sym_font )
              {
                PLOTTER.requir.ch_font = ( PLOTTER.requir.ch_font &
                    ~( CHARSET_MASK | BUILTIN_MASK ) ) | SYMBOL_FONT |
                    PRIVATE_FONT;
                STROKE.font_ptr = (StrokeFont VAR_PTR) &sym_font;
                STROKE.changed = TRUE;
              }
            break;
        case T_CH_BOLD :
            PLOTTER.requir.ch_font = ( PLOTTER.requir.ch_font & ~BOLD_MASK )
                | BOLD_FONT;
            STROKE.bold = TRUE;
            break;
        case T_CH_NOBOLD :
            PLOTTER.requir.ch_font = ( PLOTTER.requir.ch_font & ~BOLD_MASK )
                | STD_FONT;
            STROKE.bold = FALSE;
            break;
        default :
            return make_error( ERR_ABORT, name,
                (char VAR_PTR) "Unknown STROKE text attribute" );
      }
    return TRUE;
  }


static char VAR_PTR t_ch_def( what, character, index, point )
/*
 * This function allows the user to obtain the shape of any character defined
 * in the ASCII and SYMBOL fonts. With the exception of string alignment, the
 * transformations implied by t_txtset are applied to the returned coordinates.
 * This function is merely a call to the function ``define_font'' with an
 * additional argument defining which set of text attributes to use .
*/
int what;           /* specifies what to do (T_FT_* defined values) */
char character;     /* character to fetch from the font */
int index;          /* coordinate number   */
coord_pair
    VAR_PTR point;  /* returned coordinate */
  {
    START_UP( (char VAR_PTR) "t_ch_def", char  VAR_PTR );
/*
 *      Must switch to stroke based ascii font; builtin font does
 *      not have the required information - it is only a dummy font
*/
    if ( STROKE.font_ptr == &built_font )
      {
        char VAR_PTR ret_value;     /* returned value */
        STROKE.font_ptr = &asc_font;
        ret_value = define_font( (Text_Attribute VAR_PTR) &STROKE,
            SETUP.line_width, what, character, index, point );
        STROKE.font_ptr = &built_font;
        return ret_value;
      }
    return define_font( (Text_Attribute VAR_PTR) &STROKE, SETUP.line_width,
        what, character, index, point );
  }


static boolean t_symdrw( point, symbol )
/*
 *      draws a centred plotting symbol `symbol' at position `point' according
 *      to the selected stroke character attributes . This function fails if
 *      the font for the symbol is not defined or if it cannot write to the
 *      plotter/file .
*/
coord_pair VAR_PTR point;
char symbol;
  {
    double vw_marg;
    coord_pair ofset, lower_left;
    char VAR_PTR name = "t_symdrw";
    CATCH_NULL( point, name, boolean );
    if ( symbol < ' ' || symbol > '~' )
        return make_error( ERR_ABORT, name, (char VAR_PTR) "Unknown symbol" );
    START_UP( name, boolean );
    PLOTTER.requir.pen = *point;
    vw_marg = STROKE.height * MAX( STROKE.aspect, 1.0 );
/*
 *          only draw the character if some part of it can fall
 *          within the viewport
*/
    if ( point->_x >= min_clip(VIEW_CLIP,_x) - vw_marg &&
         point->_x <= max_clip(VIEW_CLIP,_x) + vw_marg &&
         point->_y >= min_clip(VIEW_CLIP,_y) - vw_marg &&
         point->_y <= max_clip(VIEW_CLIP,_y) + vw_marg )
      {
/*
 *          set the correct colour
*/
        boolean ok = set_colour();
/*
 *           if necessary define the character fonts
*/
        if ( PLOTTER.requir.ch_font != PLOTTER.actual.ch_font )
            if ( ! ps_font( PLOTTER.requir.ch_font, name ) )
                return FALSE;
/*
 *      CALCULATE LOCATION OF LOWER LEFT CORNER OF THE SYMBOL
 *
*/
        if ( STROKE.changed )
          {
            (void) map_font( (Text_Attribute VAR_PTR) &STROKE,
                SETUP.line_width );
            LOADED.matrix = FALSE;
            ok = ps_send_map();
          }
        ofset._x = STROKE.font_ptr->font_wid * 0.5;
        ofset._y = STROKE.font_ptr->font_dir[ symbol - 32 ].font_midpt;
/*
 *      determine start position
*/
        lower_left._x = point->_x -
            STROKE.map[0][0] * ofset._x - STROKE.map[0][1] * ofset._y;
        lower_left._y = point->_y -
            STROKE.map[1][0] * ofset._x - STROKE.map[1][1] * ofset._y;
/*
 *      if necessary move to the correct position
*/
        if ( fabs( PLOTTER.actual.pen._x - lower_left._x ) > PS_JITTER ||
             fabs( PLOTTER.actual.pen._y - lower_left._y ) > PS_JITTER )
            ok &= ps_pmdraw( (coord_pair VAR_PTR) &lower_left, MOVE_PROC );
/*
 *      generate the call to "T" - download
*/
        if ( ! LOAD_PS( TEXT_PROC ) )
            return FALSE;
/*
 *      ( ) and \ are special characters in a string and must be escaped
*/
        if ( symbol == '(' || symbol == ')' || symbol == '\\' )
            (void) fprintf( OP_FILE, "(\\%c) %s\n", symbol,
                proc_list[TEXT_PROC]->pr_name );
        else
            (void) fprintf( OP_FILE, "(%c) %s\n", symbol,
                proc_list[TEXT_PROC]->pr_name );
/*
 *      don't forget that the pen is moved by the character cell width
 *      in the direction of drawing the text
*/
        ofset._x *= 2.0 * ( 1.0 + STROKE.gap / STROKE.aspect );
        PLOTTER.actual.pen._x += STROKE.map[0][0] * ofset._x;
        PLOTTER.actual.pen._y += STROKE.map[1][0] * ofset._x;
        ps_device->empty_page = FALSE;
        if ( G_ERROR() || ! ok )
            return make_error( ERR_ABORT, name, err_write );
      }
    return TRUE;
  }


static int t_txtdrw( string )
/*
 *      Draws the text contained in the character string `string' at the
 *   current cursor location according to the selected stroke character
 *   attributes . The returned value is the number of characters drawn . If
 *   an error occurs, the returned value is 0 . An error may be generated
 *   because
 *     a) a character in `string' is unavailable in the specified font
 *     b) the character is not printable
 *     c) could not update the pen position after drawing the characters
 *     d) it cannot write to the plotter
*/
char VAR_PTR string;
  {
    char VAR_PTR name = "t_txtdrw";
    boolean ok = TRUE,          /* success flag                             */
            open_paren = FALSE; /* initial ( of PostScript string is written*/
    int n_chars,                /* length of `string'                       */
        char_count=0;           /* counts each character as it is drawn     */
    coord_pair cell_size,       /* character cell size or step              */
               ofset;
    double vw_marg;             /* extra margin for the viewport            */
    CATCH_NULL( string, name, boolean );
    START_UP( name, int );
/*
 *           if necessary define the character fonts
*/
    if ( PLOTTER.requir.ch_font != PLOTTER.actual.ch_font )
        if ( ! ps_font( PLOTTER.requir.ch_font, name ) )
            return FALSE;
/*
 *      CALCULATE LOCATION OF LOWER LEFT CORNER OF THE STRING
 *
 *      determine the character cell size vector
*/
    n_chars = t_strlen( string );
    cell_size._x = STROKE.font_ptr->font_wid *
        ( 1.0 + STROKE.gap / STROKE.aspect );
    cell_size._y = STROKE.font_ptr->font_hei;
    switch( STROKE.align.horiz )
      {
        case 'r' :  ofset._x = cell_size._x * n_chars;
                    break;
        case 'c' :  ofset._x = cell_size._x * n_chars / 2.0;
                    break;
        case 'l' :  ofset._x = 0.0;
                    break;
      }
    switch( STROKE.align.vert )
      {
        case 't' :  ofset._y = cell_size._y;
                    break;
        case 'c' :  ofset._y = cell_size._y / 2.0;
                    break;
        case 'b' :  ofset._y = 0.0;
                    break;
      }
/*
 *    determine start position
*/
    if ( STROKE.changed )
      {
        (void) map_font( (Text_Attribute VAR_PTR) &STROKE,
            SETUP.line_width );
/*      don't need to actually supply the correct line width because */
/*          bold_font() is not called */
        LOADED.matrix = FALSE;
      }
    PLOTTER.requir.cursor._x -= STROKE.map[0][0] * ofset._x +
        STROKE.map[0][1] * ofset._y;
    PLOTTER.requir.cursor._y -= STROKE.map[1][0] * ofset._x +
        STROKE.map[1][1] * ofset._y;
    cell_size._y = STROKE.map[1][0] * cell_size._x;
    cell_size._x = STROKE.map[0][0] * cell_size._x;
    vw_marg = STROKE.height * MAX( STROKE.aspect, 1.0 );
/*
 *    draw each character in the string exactly as specified by the
 *       driver's character font
*/
    for( ; *string; string++, char_count++ )
      {
/*
 *          only draw the character if some part of it can fall
 *          within the viewport
*/
        if ( PLOTTER.requir.cursor._x >= min_clip( VIEW_CLIP, _x ) - vw_marg &&
             PLOTTER.requir.cursor._x <= max_clip( VIEW_CLIP, _x ) + vw_marg &&
             PLOTTER.requir.cursor._y >= min_clip( VIEW_CLIP, _y ) - vw_marg &&
             PLOTTER.requir.cursor._y <= max_clip( VIEW_CLIP, _y ) + vw_marg )
            if ( ' ' <= *string && *string <= '~' )
              {
                if ( ! open_paren )
                  {
/*
 *          set the correct colour
*/
                    ok &= set_colour();
/*
 *          if necessary download the text transformation matrix
*/
                    if ( ! LOADED.matrix )
                        ok &= ps_send_map();
/*
 *          for adjusting space between characters
*/
                    PLOTTER.requir.ch_gap =
                        STROKE.gap * STROKE.font_ptr->font_wid / STROKE.aspect -
                        STROKE.font_ptr->font_gap;
                    if ( PLOTTER.requir.ch_gap != PLOTTER.actual.ch_gap )
                        ok &= ps_text_gap( PLOTTER.requir.ch_gap );
/*
 *      if necessary move to the correct position
*/
                    if ( fabs( PLOTTER.actual.pen._x - PLOTTER.requir.cursor._x
                         ) > PS_JITTER || PS_JITTER < fabs(
                         PLOTTER.actual.pen._y - PLOTTER.requir.cursor._y ) )
                        ok &= ps_pmdraw( (coord_pair VAR_PTR)
                            &PLOTTER.requir.cursor, MOVE_PROC );
/*
 *      download the graphics text procedure if necessary
*/
                    if ( ! LOAD_PS( TEXT_PROC) )
                        return FALSE;
                    (void) G_PUTC( '(' );
                    open_paren = TRUE;
                  }
/*      ( ) and \ are special characters in a string and must be escaped */
                if ( *string == '(' || *string == ')' || *string == '\\' )
                    (void) G_PUTC( '\\' );
                (void) G_PUTC( *string );
              }
             else
                 (void) make_error( ERR_WARN, name,
                     (char VAR_PTR) "not printable character" );
        PLOTTER.requir.cursor._x += cell_size._x;
        PLOTTER.requir.cursor._y += cell_size._y;
      }
    if ( open_paren )
        (void) fprintf( OP_FILE, ") %s\n",
            (char *) proc_list[TEXT_PROC]->pr_name );
    PLOTTER.actual.pen = PLOTTER.actual.cursor = PLOTTER.requir.cursor;
    ps_device->empty_page &= ! char_count;
    return ok && G_ERROR()
        ? (int) make_error( ERR_ABORT, name, err_write )
        : char_count;
  }


static char VAR_PTR t_txtget( string, max_chars )
/*
 *    Reads a maximum of `max_chars' characters from `stdin' into `string'
 * until a newline character is read . The newline character is not stored in
 * `string' . `string' is terminated with NUL . The characters are echoed to
 * the plotter . The echoed characters are confined to the viewport . If the
 * cursor is initially outside the viewport, t_txtget() starts echoing
 * characters at the top left corner of the viewport. T_txtget will not echo
 * any more characters if it runs out of space at the bottom of the viewport .
 * If an error errors or no characters are read, a NULL pointer is returned,
 * otherwise `string' is returned .
 *
 * Programming notes.
 *  t_txtget only changes plotter.requir.cursor when the location of the start
 *  of the buffered string must be moved .
*/
char VAR_PTR string;
int max_chars;
  {
    char VAR_PTR ptr,
         VAR_PTR name = "t_txtget";
    CATCH_NULL( string, name, char VAR_PTR );
    START_UP( name, char VAR_PTR );
    for( ptr = string; ( *ptr = getchar() ) != '\n' && max_chars--; ptr++ );
    *ptr = NUL;
    if ( ptr != string )
      {
/*
 *              ensure cursor is in the viewport
 *              ( place it in top left corner if it is not )
*/
        coord_pair max_cursor;
        max_cursor._x = max_clip( VIEW_CLIP, _x ) - PAGE.alph_width;
        max_cursor._y = max_clip( VIEW_CLIP, _y ) - PAGE.alph_height;
        if ( PLOTTER.requir.cursor._x < min_clip( VIEW_CLIP, _x ) ||
             PLOTTER.requir.cursor._x > max_cursor._x ||
             PLOTTER.requir.cursor._y < min_clip( VIEW_CLIP, _y ) ||
             PLOTTER.requir.cursor._y > max_cursor._y )
          {
            PLOTTER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
            PLOTTER.requir.cursor._y = max_cursor._y;
          }
/*
 *              place characters in a buffer until it runs out of the viewport
*/
        if ( PLOTTER.requir.cursor._x <= max_cursor._x &&
             PLOTTER.requir.cursor._y >= min_clip( VIEW_CLIP, _y ) )
          {
            char buffer[BUFFER_LEN],
                 VAR_PTR buffer_ptr = buffer;
            int cursor_col = PLOTTER.requir.cursor._x;
            for( ptr = string; *ptr; ptr++ )
                if ( ' ' <= *ptr && *ptr <= '~' &&
                    PLOTTER.requir.cursor._y >= min_clip( VIEW_CLIP, _y ) &&
                    PLOTTER.requir.cursor._y <= max_cursor._y )
                  {
                    *buffer_ptr++ = *ptr;
/*
 *                      print the buffer if the right hand margin is reached
*/
                    cursor_col += (int) PAGE.alph_width;
                    if ( cursor_col > max_cursor._x )
                      {
                        *buffer_ptr = NUL;
                        (void) t_txtstr( (char VAR_PTR) buffer, buffer_ptr,
                            (coord_pair VAR_PTR) &PLOTTER.requir.cursor,
                            name );
                        PLOTTER.requir.cursor._x = cursor_col
                                                 = min_clip( VIEW_CLIP,_x );
                        PLOTTER.requir.cursor._y -= PAGE.alph_height;
                        buffer_ptr = buffer;
                      }
/*
 *                      print the buffer if it is full
*/
                    else if ( buffer_ptr - buffer >= BUFFER_LEN )
                      {
                        *buffer_ptr = NUL;
                        (void) t_txtstr( (char VAR_PTR) buffer, buffer_ptr,
                            (coord_pair VAR_PTR) &PLOTTER.requir.cursor,
                            name );
                        buffer_ptr = buffer;
                      }
                  }
/*
 *                      print whatever remains in the buffer
*/
            if ( buffer_ptr != buffer )
              {
                *buffer_ptr = NUL;
                (void) t_txtstr( (char VAR_PTR) buffer, buffer_ptr,
                    (coord_pair VAR_PTR) &PLOTTER.requir.cursor,
                    name );
              }
          }
        PLOTTER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
        PLOTTER.requir.cursor._y -= PAGE.alph_height;
      }
    return *string ? string : (char VAR_PTR) 0;
  }


static int t_txtput( string )
/*
 *    Print text string `string' on the plotter, starting at the place
 *  last specified by `t_txtmov' or 't_txtcur', using the colour specified by
 *  `t_colour' . Only those characters within the viewport are written to the
 *  screen. The returned value is the number of characters written to the
 *  screen, or zero if an error has occurred .
 *
 *  The character string `string' may contain printable characters and the
 *  control codes,
 *    HT : \t : \011 : move to the right by one character space
 *    BS : \b : \010 : move to the left by one character space
 *    CR : \r : \015 : move to the beginning of the current line
 *    LF : \n : \012 : move down one character space
 *    VT :    : \013 : move up one character space
 *    EM :    : \031 : move to top left hand side of the screen
 *  Any other characters are ignored .
 *
 *  This function only breaks the string into segments terminated by cursor
 *  control characters and positions the required_cursor appropriately before
 *  calling t_txtstr to put the characters on the screen.
 *
 * Programming notes,
 *  t_txtput only changes PLOTTER.requir.cursor when the location of the start
 *  of the buffered string must be moved .
*/
char VAR_PTR string;            /* pointer to start of the character string */
  {
    char VAR_PTR ptr,
         VAR_PTR name = "t_txtput";
    int     char_count = 0;     /* number of characters printed             */
    CATCH_NULL( string, name, boolean );
    START_UP( name, int );
    ps_device->Courier_used = TRUE;
    for ( ptr = string; *ptr; ptr++ )
      {
        if ( *ptr == '\n' || *ptr == VT   || *ptr == EM   ||
             *ptr == '\r' || *ptr == '\b' || *ptr == '\t' )
          {
            char_count += t_txtstr( string, ptr,
                (coord_pair VAR_PTR) &PLOTTER.requir.cursor, name ) + 1;
            switch( *ptr )
              {
                case '\n': PLOTTER.requir.cursor._y += -PAGE.alph_height;
                           break;
                case VT  : PLOTTER.requir.cursor._y += PAGE.alph_height;
                           break;
                case EM  : PLOTTER.requir.cursor._x= min_clip( VIEW_CLIP, _x );
                           PLOTTER.requir.cursor._y = max_clip( VIEW_CLIP, _y )
                               - PAGE.alph_height;
                           break;
                case '\r': PLOTTER.requir.cursor._x= min_clip( VIEW_CLIP, _x );
                           break;
                case '\b': PLOTTER.requir.cursor._x += -PAGE.alph_width;
                           break;
                case '\t': PLOTTER.requir.cursor._x += PAGE.alph_width;
                           break;
              }
            string = ptr + 1;
          }
      }
    if ( string != ptr )
        char_count += t_txtstr( string, ptr,
            (coord_pair VAR_PTR) &PLOTTER.requir.cursor, name );
    return char_count;
  }


static int t_txtstr( string, end_nul, point, called_by )
/*
 * Prints a label of printable ASCII , starting at the current required cursor
 * postion and using the current colour . Only characters within the viewport
 * are written to the screen . The returned value is the number of characters
 * written.
 * The character string `string' may be NUL terminated or, if only a fragment
 * of the string is to be printed, `end_nul' is a pointer to where the NUL
 * should be.
*/
char       VAR_PTR string,      /* string to be printed */
           VAR_PTR end_nul;     /* address of phantom end of string NUL */
coord_pair VAR_PTR point;       /* starting location of the string */
char       VAR_PTR called_by;   /* name of calling function */
  {
    boolean ok = TRUE,          /* success flag */
            open_paren = FALSE; /* opening ( is sent to file */
    int char_count = 0;         /* number of characters sent */
    PLOTTER.requir.cursor = *point;
    for( ; *string && string != end_nul && ok; string++ )
      {
        if ( ' ' <= *string && *string <= '~' )
          {
/*
 ********     write the character if it is in the viewport
*/
            if ( PLOTTER.requir.cursor._x >= min_clip( VIEW_CLIP, _x ) &&
                 PLOTTER.requir.cursor._x <= max_clip( VIEW_CLIP, _x ) &&
                 PLOTTER.requir.cursor._y >= min_clip( VIEW_CLIP, _y ) &&
                 PLOTTER.requir.cursor._y <= max_clip( VIEW_CLIP, _y ) )
              {
                if ( ! open_paren )
                  {
                    ok &= set_colour();
                    ok &= LOAD_PS( TTY_PROC );
/*
 *          Place the cursor in the correct place and write the opening
 *          ( of the PostScript string.
*/
                    if ( fabs( PLOTTER.requir.cursor._x -
                               PLOTTER.actual.cursor._x ) > PS_JITTER ||
                         fabs( PLOTTER.requir.cursor._y -
                               PLOTTER.actual.cursor._y ) > PS_JITTER )
                        ok &= ps_pmdraw( (coord_pair VAR_PTR)
                            &PLOTTER.requir.cursor, MOVE_PROC );
                    (void) G_PUTC( '(' );
                    open_paren = TRUE;
                  }
/*
 *          write the character to the PostScript string
*/
                char_count++;
/*      ( ) and \ are special characters in a string and must be escaped */
                if ( *string == '(' || *string == ')' || *string == '\\' )
                    (void) G_PUTC( '\\' );
                (void) G_PUTC( *string );
                PLOTTER.actual.cursor._x += PAGE.alph_width;
              }
            PLOTTER.requir.cursor._x += PAGE.alph_width;
          }
      }
/*
 *          close the PostScript string and call the LR_text procedure
*/
    if ( ok && open_paren )
        (void) fprintf( OP_FILE, ") %s\n",
            (char *) proc_list[TTY_PROC]->pr_name );
    PLOTTER.actual.pen = PLOTTER.actual.cursor;
    ps_device->empty_page &= ! char_count;
    if ( ! ok || G_ERROR() )
        return (int) make_error( ERR_WARN, called_by, err_write );
    if ( char_count > 0 )
        ps_device->Courier_used |= TRUE;
    return char_count;
  }


/* ************************************************************************ */
/* ************************************************************************ */
/*                                                                          */
/*                    Array of pointers to driver functions                 */
/*                                                                          */
/* With the exception of the function ``t_driver'' which is used to select  */
/* the required driver, this array is the ONLY point of access to the       */
/* functions in the driver .                                                */
/*                                                                          */
/* ************************************************************************ */
/* ************************************************************************ */

/*
 * static clip_area VAR_PTR t_begin();
 * static boolean
 *     t_end(),    t_vwport(), t_colour(), t_dashed(), t_thick(),  t_gr_off(),
 *     t_gr_on(),  t_pr_ter(), t_dump(),   t_visib(),  t_invis(),  t_revers(),
 *     t_replac(), t_newpag(), t_newscr(), t_bell(),   t_wait(),   t_txtcur(),
 *     t_txtmov(), t_txtset(), t_symdrw(), t_xhrget(), t_xhrput(), t_vecdrw(),
 *     t_boxfil(), t_vecmov();
 * static char
 *         VAR_PTR t_txtget(), VAR_PTR t_status(), VAR_PTR t_ch_def();
 * static int t_txtput(), t_txtdrw();
 */

T_dev_fns Post_dev_ptrs = {
            t_begin,  t_end,    t_vwport, t_gr_on,  t_gr_off,
            t_colour, t_dashed, t_thick,  t_visib,  t_invis,
            t_revers, t_replac, t_newpag, t_newscr, t_bell,
            t_wait,   t_txtmov, t_txtcur, t_txtset, t_symdrw,
            t_txtdrw, t_txtget, t_txtput, t_status, t_pr_ter,
            t_vecmov, t_vecdrw, t_dump,   t_xhrput, t_xhrget,
            t_boxfil, t_ch_def };


/* ************************************************************************* */
/*                                                                           */
/*                      PostScript procedure handlers                        */
/*                                                                           */
/*      These functions read the source file of PostScript procedures,       */
/* interpret its contents, construct a directory of Postscript procedures,   */
/* destroy the directory, find a named procedure in the directory, download  */
/* a procedure to the graphics device or file, and print the contents of a   */
/* directory entry to a file .                                               */
/*                                                                           */
/* ************************************************************************* */

static Procedure
   make_ps_proc( name_str, name_len, proc_str, proc_len, dir_index )
/*
 * Allocate space for a PostScript procedure with name "name_str" and procedure
 * text "proc_str" . "dir_ptr" is a pointer to an entry in the directory of
 * Post_Script procedures . If the required memory cannot be allocated, a NULL
 * pointer is returned .
*/
char VAR_PTR name_str,
     VAR_PTR proc_str;  /* name and procedure text strings */
int name_len, proc_len, /* length of name and procedure text strings */
    dir_index;          /* index into procedure directory of the procedure */
  {
    char VAR_PTR name = "make_ps_proc";
    Procedure ps_proc;
    CATCH_NULL( name_str && proc_str, name, Procedure );
    if ( dir_index < 0 || dir_index >= NO_PS_PROCS )
        return (Procedure) make_error( ERR_ABORT, name,
            (char VAR_PTR) "Illegal procedure index" );
/*
 *      allocate memory for the Procedure structure
*/
    if ( ( ps_proc = (Proc_entry VAR_PTR)
         malloc( (unsigned) sizeof( Proc_entry ) ) ) )
      {
/*
 *      allocate memory for procedure name string and copy the name into it
*/
        if ( ( ps_proc->pr_name = (char VAR_PTR)
                 malloc( (unsigned) sizeof(char) * ( name_len + 1 ) ) ) )
          {

            (void) t_strcpy( ps_proc->pr_name, name_str );
/*
 *      allocate memory for procedure text string and copy the text into it
*/
            if ( ( ps_proc->pr_string =
                     malloc( (unsigned) sizeof(char) * ( proc_len + 1 ) ) ) )
              {
                (void) t_strcpy( (char *) ps_proc->pr_string,
                    (char *) proc_str );
/*
 *      reset cleared and loaded flags, dependency list is empty
*/
                ps_proc->pr_index = dir_index;
                ps_proc->pr_list  = (Proc_dep_list VAR_PTR) NULL;
                return ps_proc;
              }
/*
 *      clean up and return with error if space cannot be allocated
*/
            (void) free( ps_proc->pr_name );
          }
        (void) free( (char *) ps_proc );
        ps_proc = (Procedure) 0;
      }
    return (Procedure) make_error( ERR_ABORT, name,
        (char VAR_PTR) "Cannot allocate memory" );
  }


static boolean print_ps_proc( out_file, ps_proc )
/*
 *     Print the contents of the PostScript procedure ``ps_proc'' to file
 * ``out_file''. The returned value is FALSE if cannot write to the file.
*/
FILE *out_file;
Procedure ps_proc;
  {
    char VAR_PTR name = "print_ps_proc";
    Proc_dep_list VAR_PTR list; /*  to next called-procedure list element */
    CATCH_NULL( out_file && ps_proc, name, boolean );
    (void) fprintf( out_file, "\n***** PostScript procedure *****\n" );
/*
 *      print name of procedure
*/
    if ( ps_proc->pr_name )
        (void) fprintf( out_file, "   ** name       : %s\n",
                        ps_proc->pr_name );
    else
        (void) make_error( ERR_WARN, name,
            (char VAR_PTR) "Procedure has no name" );
/*
 *      print directory index
*/
    (void) fprintf( out_file, "   ** directory index : %d\n",
        ps_proc->pr_index );
/*
 *      print text of the procedure
*/
    if ( ps_proc->pr_string )
        (void) fprintf( out_file, "   ** procedure  : %s\n",
                        ps_proc->pr_string );
    else
        (void) make_error( ERR_WARN, name,
            (char VAR_PTR) "Procedure has no text" );
/*
 *      print the list of called procedures
*/
    (void) fprintf( out_file, "   ** calls      :" );
    for( list = ps_proc->pr_list; list; list = list->pr_next )
        if ( list->pr_depend )
            (void) fprintf( out_file, " %s", list->pr_depend->pr_name );
    (void) fprintf( out_file, "\n" );

    if ( ferror( out_file ) )
        return make_error( ERR_WARN, name,
            (char VAR_PTR) "Cannot write to file" );
    return TRUE;
  }


static boolean read_ps_procs( ps_file )
FILE *ps_file;
/*
 *  Function for reading a PostScript file, stripping the comments away
 *  and placing the procedures and variable definitions which are embedded
 *  in this file into a directory of PostScript ``Procedure''s . The
 *  ``Procedure''s can then be downloaded in the correct order to a PostScript
 *  file .
*/
  {
    boolean ok = TRUE,
            first_char = TRUE,      /* TRUE if at the start of a line   */
            found_bsl  = FALSE,     /* TRUE if a backslash is found     */
            comment_1  = FALSE,     /* TRUE if a comment starts from column 1 */
            in_comment = FALSE,     /* TRUE if in a comment             */
            in_name    = FALSE,     /* TRUE if reading a procedure name */
            in_proc    = FALSE,     /* TRUE if reading procedure text   */
            in_string  = FALSE,     /* TRUE if reading a text string    */
            in_white   = FALSE;     /* TRUE if previous char was white space */
    int c,                          /* character read from PostScript file   */
        open_braces = 0,            /* number of unpaired `{'s */
        found_def   = 0,     /* = 3 when procedure terminator "def" is found */
        dir_index   = 0;            /* initial index into procedure directory*/
    char name_str[NAME_LEN],        /* name of procedure */
         VAR_PTR name_ptr = name_str, /* pointer to 1st blank in name_str*/
         proc_str[PROC_LEN],        /* procedure text */
         VAR_PTR proc_str_ptr = proc_str, /* pointer to 1st blank in proc_str*/
         VAR_PTR name = "read_ps_procs";

    CATCH_NULL( ps_file, name, boolean );
/*
 *          read the file until we get to the end
*/
    while ( EOF != ( c = getc( ps_file ) ) && ok )
      {
/*
 *          if we are in the middle of a comment, ignore the character, but
 *          watch for the newline character which marks the end of the comment.
*/
        if ( in_comment )
          {
            if ( c == '\n' )
              {
                in_comment = FALSE;
                first_char = TRUE;
/*
 *          if the comment did not begin at the start of the line, and we are
 *          reading a procedure name or procedure body, copy the newline char
 *          to the procedure string .
*/
                if ( ! comment_1 )
                    if ( in_name || in_proc )
                        *proc_str_ptr++ = c;
              }
          }
/*
 *          If the char is a '%', we are at the beginning of a comment, and
 *          any name we are currently defining must be terminated.
 *          Terminating a name does not require generation of a white-space
 *          separator because the end-of-comment will automatically do this.
*/
        else if ( c == '%' && ! in_string )
          {
            comment_1 = first_char;
            in_comment = TRUE;
            if ( in_name )
              {
               in_name = FALSE;
               in_proc = TRUE;
              }
          }
/*
 *          this ``else if'' simply compresses out multiple white-space
 *          characters
*/
        else if ( ! ( t_isspace(c) && in_white ) || in_string )
          {
            first_char = c == '\n';
            in_white = t_isspace(c);
/*
 *          N.B. no account is taken of character string () syntax
 *          if we are in a procedure name or procedure body, add the character
 *          to the procedure string
*/
            if ( in_proc || in_name )
                *proc_str_ptr++ = c;
/*
 *          if we are in a procedure name, look for the terminating white-space
 *          character . If a name is complete, make a new Procedure, and copy
 *          the name into the Procedure . Any other character (except '%' which
 *          has already been filtered out) is copied to the name string .
 *       The character has already been added to the temporary procedure string.
*/
            if ( in_name )
              {
                if ( in_white )
                  {
                    in_name = FALSE;
                    in_proc = TRUE;
                  }
                else
                    *name_ptr++ = c;
              }
/*
 *          if we are not in a procedure, a '/' character indicates the start
 *          of the name of a new procedure .
 *       The character has not been added to the temporary procedure string.
 *       Char the name string but don't include the character '/' in the name
*/
            else if ( c == '/' && ! in_proc )
              {
                in_name = TRUE;
/*  clear the name string and reset the name pointer */
                for ( name_ptr = name_str; *name_ptr; *name_ptr++ = '\000' );
                name_ptr = name_str;
/*  clear the procedure string and reset the procedure string pointer */
                for ( proc_str_ptr = proc_str; *proc_str_ptr;
                      *proc_str_ptr++ = '\000' );
                proc_str_ptr = proc_str;
                *proc_str_ptr++ = c;
              }
/*
 *          if we are in a string, watch out for the end of the string
*/
            else if ( in_string )
              {
                in_string = c != ')' || ! found_bsl;
                found_bsl = c == '\\';
              }
/*
 *          we are now in the body of a procedure or variable definition
 *          The method is to search for '{', '}' and 'def' . If 'def' is
 *          found without any unpaired '{'s remaining, the end of the procedure
 *          has been found . The procedure can then be pasted into the
 *          procedure directory .
*/
            else if ( in_proc )
              {
                if ( c == '{' )
                    open_braces++;
                else if ( c == '}' )
                    if ( open_braces <= 0 )
                        return make_error( ERR_ABORT, name,
                            (char VAR_PTR) "Syntax error : too many }s" );
                    else
                        open_braces--;
                else if ( 'd' + found_def == c )
                  {
                    found_def++;
                    if ( found_def == 3 && ! open_braces )
                      {
/*
 *          We have now found the end of a procedure . Then put the procedure
 *          into the directory.
*/
                        in_proc = FALSE;
                        found_def = 0;
/*          properly terminate the name and procedure strings -- Experience */
/*          has shown that this piece of defensive code is necessary        */
                        *name_ptr = *proc_str_ptr = NUL;
                        if ( ( proc_dir[ dir_index ] =
                                 make_ps_proc( (char VAR_PTR) name_str,
                                     (int) ( name_ptr - name_str ),
                                     (char VAR_PTR) proc_str,
                                     (int) ( proc_str_ptr - proc_str ),
                                     dir_index ) ) )
                            dir_index++;
                        else
                            return FALSE;
                      }
                  }
                else
                    found_def = 0;
              }
/*
 *          if we get to this point, we have found some PostScript which
 *          is not in a procedure or variable definition, and so we can
 *          ignore it
*/
          }
      }
    return TRUE;
  }


static Procedure find_ps_proc( proc_name )
/*
 *   Search in the procedure directory for the PostScript procedure
 *   ``proc_name'' . If the procedure is found, it is returned . If the
 *   procedure cannot be found, an error is reported and a NULL result
 *   is returned .
*/
char VAR_PTR proc_name;     /* name of required PostScript procedure */
  {
    Procedure VAR_PTR proc_ptr;  /* pointer to directory entry */
    char VAR_PTR name = "find_ps_proc";
    CATCH_NULL( proc_name, name, Procedure );
/*
 *          compare the name string in directory entries with the required
 *          procedure name and return the procedure if a match is found
*/
    for ( proc_ptr = proc_dir; proc_ptr < proc_dir + NO_PS_PROCS; proc_ptr++ )
        if ( *proc_ptr )
            if ( (*proc_ptr)->pr_name )
              {
                if ( ! t_stncmp( proc_name, (*proc_ptr)->pr_name, 80 ) )
                    return *proc_ptr;
              }
            else
                return (Procedure) make_error( ERR_ABORT, name,
                    (char VAR_PTR) "Procedure has no name" );
    return (Procedure) make_error( ERR_ABORT, name,
        (char VAR_PTR) "Cannot find procedure" );
  }


static boolean load_ps_proc( proc_index )
/*
 * This is a recursive procedure which downloads the procedure and
 * every other procedure which it calls to the graphics output file .
*/
int proc_index;     /* directory index of the PostScript procedure */
  {
    boolean ok = TRUE;
    Proc_dep_list VAR_PTR list; /* to next called-procedure list element */
    char VAR_PTR name = "load_ps_proc";
    Procedure ps_proc;
    if ( proc_index < 0 || proc_index >= NO_PS_PROCS )
        return make_error( ERR_ABORT, name,
            (char VAR_PTR) "Illegal procedure index" );
/*
 *          if the procedure has already been downloaded, don't do it again
 *          if the procedure has no text, PANIC .
*/
    if ( LOADED.procs[ proc_index ] )
        return TRUE;
    ps_proc = proc_dir[ proc_index ];
    if ( ! ps_proc->pr_string )
        return (boolean) make_error( ERR_ABORT, name,
            (char VAR_PTR) "Procedure has no text" );
/*
 *          first, load the procedures which `ps_proc' calls
*/
    for( list = ps_proc->pr_list; list && ok; list = list->pr_next )
        if ( list->pr_depend )
            ok = load_ps_proc( list->pr_depend->pr_index );
/*
 *          load the text of the procedure
*/
    (void) G_PUTS( (char *) ( ps_proc->pr_string ) );
    (void) G_PUTC( '\n' );
    if ( G_ERROR() )
        return make_error( ERR_ABORT, name, err_write );
    return LOADED.procs[ proc_index ] = TRUE;
  }


static boolean free_ps_procs( )
/*
 * Free all of the memory occupied by the PostScript procedures in the directory
 * `proc_dir' .
*/
  {
    boolean ok = TRUE;
    Procedure *proc_ptr;    /* pointer to directory entry */
    Proc_dep_list VAR_PTR list,  /* pointer to called-procedure list element */
             VAR_PTR next;  /* pointer to next called-procedure list element */
    char VAR_PTR name = "free_ps_procs";

    for ( proc_ptr = proc_dir; proc_ptr < proc_dir + NO_PS_PROCS; proc_ptr++ )
        if ( *proc_ptr )
          {
/*
 *          De-allocate the memory occupied by the text of the procedure
*/
            if ( (*proc_ptr)->pr_string )
                (void) free( (char *) (*proc_ptr)->pr_string );
            else
                (void) make_error( ERR_WARN, name,
                    (char VAR_PTR) "Procedure has no text" );
/*
 *          clear the list of called procedures
*/
            for( list = (*proc_ptr)->pr_list; list; list = next )
              {
                next = list->pr_next;
                (void) free( (char *) list );
              }
/*
 *          clear the procedure name
*/
            if ( (*proc_ptr)->pr_name )
                (void) free( (*proc_ptr)->pr_name );
            else
                return make_error( ERR_WARN, name,
                    (char VAR_PTR) "Procedure has no name" );
/*
 *          finally, release the procedure
*/
            (void) free( (char *) (*proc_ptr) );
            *proc_ptr = (Procedure) 0;
          }
    return ok;
  }


static boolean chain_ps_procs( )
/*
 *  Examines the text of all procedures in the directory for names of other
 *  procedures . Any such names found are put into a linked list of ``called
 *  procedures'' .
*/
  {
    char VAR_PTR name = "chain_ps_procs";
    Proc_dep_list VAR_PTR VAR_PTR list_ptr;  /* pointer called-procedure list */
    Procedure VAR_PTR name_ptr, /* pointer to directory entry -name */
              VAR_PTR proc_ptr; /* pointer to directory entry -procedure text */

/*
 *          this loop steps through the text of each procedure
*/
    for ( proc_ptr = proc_dir; proc_ptr < proc_dir + NO_PS_PROCS; proc_ptr++ )
        if ( *proc_ptr )
            if ( (*proc_ptr)->pr_name && (*proc_ptr)->pr_string )
              {
                list_ptr = &( (*proc_ptr)->pr_list );
/*
 *          this loop fetches the name of each procedure in the directory
 *          and attempts to find it in the procedure text
*/
                for ( name_ptr = proc_dir; name_ptr < proc_dir + NO_PS_PROCS;
                      name_ptr++ )
                    if ( *name_ptr && *name_ptr != *proc_ptr )
                        if ( (*name_ptr)->pr_name )
                          {
                            if ( str_name( (*name_ptr)->pr_name,
                                          (*proc_ptr)->pr_string ) )
                              {
                                if ( ( *list_ptr =(Proc_dep_list VAR_PTR )
                              malloc ( (unsigned) sizeof( Proc_dep_list ) ) ) )
                                  {
                                    (*list_ptr)->pr_next = (Proc_dep_list *) 0;
                                    (*list_ptr)->pr_depend = *name_ptr;
                                    list_ptr = (Proc_dep_list VAR_PTR VAR_PTR)
                                               &( (*list_ptr)->pr_next );
                                  }
                                else
                                    return make_error( ERR_ABORT, name,
                                      (char VAR_PTR) "Cannot allocate memory" );
                              }
                          }
                        else
                            return make_error( ERR_ABORT, name,
                                (char VAR_PTR) "Procedure has no name" );
              }
            else
                return make_error( ERR_ABORT, name,
                    (char VAR_PTR) "Procedure has no name or no text" );
    return TRUE;
  }

/* ************************************************************************* */
/*                                                                           */
/*                    END OF PostScript PROCEDURE HANDLERS                   */
/*                                                                           */
/* ************************************************************************* */

/* ************************************************************************* */
/*                                                                           */
/*                              GENERAL UTILITIES                            */
/*                                                                           */
/* ************************************************************************* */

static char VAR_PTR str_name( name, string )
/*
 *      The function `str_name' returns a pointer to the first character of
 * the substring in `string' which is the same as `name' . If `name' is not
 * contained in string `string' , a NULL pointer is returned .
*/
  char VAR_PTR name, VAR_PTR string;
  {
    boolean match;           /* a match is achieved */
    char VAR_PTR nam_ptr,
         VAR_PTR str_ptr;    /* moving pointers to `name' & `string' */
    int diflen = t_strlen( string ) - t_strlen( name );
                             /* difference between lengths of strings */
/*
 * Method : `string' is used as a pointer to the first character of the
 *         substring in `string' which matches `name'
*/
    while( diflen-- >= 0 )
      {
        match = TRUE;
        for ( str_ptr = string, nam_ptr = name;
              *nam_ptr && ( match &= *nam_ptr++ == *str_ptr ); str_ptr++ );
        if ( match && ( t_isspace( *str_ptr ) || ! *str_ptr ) )
            return string;
        else
            while ( *string && ! t_isspace( *string++ ) )
                diflen--;
      }
    return (char VAR_PTR) 0;
  }

/* ************************************************************************* */
/*                                                                           */
/*                          END OF GENERAL UTILITIES                         */
/*                                                                           */
/* ************************************************************************* */
