/*
    Copyright (C) 1997  Peter Vernon Lanspeary

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* ************************************************************************ */
/*                                                                          */
/*      Basic Graphics Terminal Graphics Driver for the Sun workstation     */
/*                 N.B. This is hardware dependent stuff                    */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/*   This  is a  graphics  driver for  the Sun  workstation.  This driver   */
/*   opens a Suntool window for graphics.  Separate pens  are used for      */
/*   drawing lines and  for text . Text can  be written into  the graphics  */
/*   window  either with a  stroke generating font or using the default     */
/*   pixel font .                                                           */
/*                                                                          */
/*      Most functions are of type `boolean' and return a TRUE or FALSE     */
/*   result to indicate success or failure . Error messages are placed on   */
/*   the error queue -- which can be examined at the discretion of the user.*/
/*                                                                          */
/*   Any programs which use this driver must #include DEVICE_G_HEADER .     */
/*   These functions must be linked with "clip.c" and "error.c"             */
/*   The documentation on "clip.c" and "error.c" should be read before      */
/*   continuing .                                                           */
/*                                                                          */
/* BUGS                                                                     */
/*  1) If the application is busy doing something other than graphics, the  */
/*     mouse cursor will not track across the graphics window .             */
/*  2) It is possible to scroll off the canvas . The image can be brought   */
/*     back by doing a Resize without changing the size of the window .     */
/*  3) Sometimes, if you do a lot of scrolling, the graphics cursor becomes */
/*     stuck as a scrolling cursor and the crosshairs disappear .           */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/* The second argument of ``t_begin()'' is a character string of selected   */
/* options . The available options are the words                            */
/*   "LEFT {integer}"   distance of the left edge of the graphics window    */
/*                      from the left edge of the screen (default = 0)      */
/*   "TOP {integer}"    distance of the top edge of the graphics window     */
/*                      from the top edge of the screen (default = 0)       */
/*   "WIDTH {integer}"  width of the graphics window (default = maximum)    */
/*   "HEIGHT {integer}" height of the graphics window (default = maximum)   */
/*   "OPEN"             begin with an open graphics window (default)        */
/*   "CLOSED"           begin with an graphics window closed to an icon     */
/* Words may be abbreviated and in upper, lower or mixed case .             */
/*                                                                          */
/* ************************************************************************ */

#include "system.h"             /* system dependent stuff */
#include <stdio.h>
#include DEVICE_G_HEADER
#include FONT_SOURCE_HEADER
#include "dev_udef.h"
#include <suntool/sunview.h>
#include <suntool/canvas.h>
#include <suntool/panel.h>
#include <suntool/scrollbar.h>
#include <sunwindow/rect.h>

/*
 *      Machine constants
*/
#define PI            3.141592653589793238462643

/*
 *    non printable ASCII characters
 *  delete unused characters when finished
*/
#define NUL  '\000'   /* Nul                        */
#define BEL  '\007'   /* Bell                       */
#define BS   '\010'   /* Backspace                  */
#define CR   '\015'   /* Carriage Return            */
#define EM   '\031'   /* End of Medium              */
#define HT   '\011'   /* Horizontal Tabulation      */
#define LF   '\012'   /* Line Feed                  */
#define VT   '\013'   /* Vertical Tabulation        */
#define ESC  '\033'   /* Escape                     */
#define DEL  '\177'   /* Delete                     */

#define MIN_FRAME_WIDTH  64     /* minimum allowed frame width              */
#define MIN_FRAME_HEIGHT 64     /* minimum allowed frame height             */
#define ALPH_HEIGHT (SETUP.t_sysfont->pf_defaultsize).y /* character height */
#define ALPH_WIDTH  (SETUP.t_sysfont->pf_defaultsize).x /* character width  */
#define PIX_SIZE    0.3048                           /* size of a pixel (mm)*/
#define PEN_THICKNESS 1.0       /* pen or line thickness                    */
#define CMS_SIZE      16        /* size of colour map segment               */
#define SOLID_LINE    1         /* line style for solid line                */
#define CLIP2R_ID 0x3c22        /* two dimensional rectangular clipping     */
#define NOT_SET       -1        /* indicates that an option is not set      */

/*
 *      Sun_g determines the width and height of the workstation screen when
 *  the graphics is fired up with `t_begin'. As far as I know there are two
 *  sizes of Sun workstation screen . The definitions are only given here to
 *  to give the reader some idea of the resolution of the screen .
 *
 * medium resolution
 *   #define FULL_SCR_X  1152
 *   #define FULL_SCR_Y   900
 *
 * high resolution
 *   #define FULL_SCR_X  1600
 *   #define FULL_SCR_Y  1280
 *
 *      FULL_SCR_X and FULL_SCR_Y are the width and height of the graphics base
 *  frame (window). The useful size of the drawing canvas ( which fits inside
 *  the borders of the base frame ) is slightly less than this, and is stored
 *  in `sun_max._x' and `sun_max._y' .
*/
/*
 *      The drawing canvas event procedure is required when waiting for
 *      input from the user . Only `t_xhrget', `t_txtget' and `t_wait'
 *      require a response from the user. Before releasing control to the
 *      notifier, these functions set the value of `waiting_func' so
 *      that the `canvas_event_proc'edure can determine what action to take.
 *      The allowable values for `waiting_func' are
*/
#define T_NONE    1000
#define T_XHRGET  1001
#define T_TXTGET  1002
#define T_WAIT    1003
#define T_END     1004

/* ************************************************************************ */
/*                                                                          */
/*                          Graphics Macros                                 */
/*                                                                          */
/* ************************************************************************ */
/*
 *  ABBREVIATIONS
 ******************
*/
#define  TER        sun3->ter
#define  STROKE     sun3->stroke
#define  VIEW_CLIP  sun3->view_clip
#define  SETUP      sun3->setup
#define  GFX_WIN    sun3->gfx_win

/*
 *  CATCH_NULL(pointer,caller,ret_type)
 ****************************************
 *      If "pointer" has a NULL value, an "Arg is NULL pointer" error message
 * is generated and the function is aborted . Otherwise, nothing is done .
 *      "pointer"   is the pointer to be tested,
 *      "caller"    is a character string naming the function
 *      "ret_type"  is the type of the returned zero value which indicates an
 *                    error condition
*/
#define CATCH_NULL(pointer,caller,ret_type) \
  if ( ! (pointer) )    \
      return (ret_type) make_error( ERR_ABORT, caller, mt_str )
/*
 *  LIMIT( x, min, max )
 *************************
 *      Limits the value of `x' to between `min' and `max'
*/
#define LIMIT(x,min,max)  (x) <= (min) ? (min) : ( (x) >= (max) ? (max) : (x) )

/*
 *  MAX( x, y )
 ****************
 *      The returned value is the maximum of `x' and `y' .
*/
#define MAX(x,y)        ( (x) > (y) ? (x) : (y) )

/*
 *  MIN( x, y )
 ****************
 *      The returned value is the minimum of `x' and `y' .
*/
#define MIN(x,y)        ( (x) > (y) ? (y) : (x) )

/*
 *  START_UP(name,ret_type)
 ****************************
 *      Calls `t_begin' with the default arguments `stdout' and `mt_str'
 *  (empty string) if `t_begin' has not previously been called .
*/
#define START_UP(name,ret_type) \
    if ( ! sun3 ) \
        if ( ! t_begin( stdout, mt_str ) ) \
            return (ret_type) make_error( ERR_CALL, name, mt_str )

/*
 *  SWAPD( a, b )
 *****************
 *    swaps the two double precision floating point values `a' and `b'
*/
#define SWAPD(a,b)       { double temp = b; b = a; a = temp; }

/*
 *  TOLOWER( c )
 *      Converts character `c' to lower case if it is upper case
*/
#define TOLOWER(c)      ( 'A' <= (c) && (c) <= 'Z' ? (c) + 'a' - 'A' : (c) )

/*
 *  SUN_ROP( dot )
 *      Sets the `raster_op' to the correct value for `dot'
*/
#define SUN_ROP(dot)    ( rop_list[ (int) ( dot ) ] )

/*
 *  SUN_COLOUR( Colour )
 *      Sets the colour part of the `raster op' .
*/
#define SUN_COLOUR(Colour)   ( SETUP.colour_screen & \
                               PIX_COLOR( colour_list[ ( Colour ) ] ) )


/* ************************************************************************ */
/*                                                                          */
/*                          Type declarations                               */
/*                                                                          */
/* ************************************************************************ */

enum dev_mode     /* terminal mode */
  {
    GRAPH_CLOSED, /* graphics window is closed to an icon - all t_txtget    */
                  /*   & t_txtput I/O is through this parent window         */
    GRAPH_OPEN,   /* graphics window is open - all t_txtget                 */
                  /*   & t_txtput I/O is through this graphics window       */
    GRAPH_QUIT    /* graphics base frames and hierarchies do not exist      */
  };

/*
 *      The `raster operation' of the workstation screen controls how it sets
 *   or reset pixel dots . The `raster operation' is set for putting text and
 *   drawing vector & points . The `raster operations' used by sun_g form only
 *   a small subset of those available ; they are,
*/
enum dev_dots     /* data level of terminal */
  {
    DOT_ON,             /* draw visible vectors, `or' them onto the screen  */
    DOT_OFF,            /* draw invisible vectors, switch dots off          */
    DOT_REV,            /* the compliment of the new data is plotted        */
    DOT_NEW,            /* replace pixel contents with new data - not used  */
    DOT_NOP             /* nothing                                          */
  };

typedef struct            /* TERMINAL STATUS                     */
  {                       /*     DATA TYPE                       */
    enum dev_mode mode;   /* terminal mode                       */
    enum dev_dots dots;   /* data level                          */
    int           colour; /* foreground colour                   */
    int           dash;   /* dashed line style                   */
    int           thick;  /* line and test width                 */
    coord_pair    cursor; /* alpha graphics text cursor position */
    coord_pair    pen;    /* vector graphics pen position        */
  }
    ter_status;

typedef struct      /*    these properties are written only by     */
  {                 /* t_begin() when creating the graphics window */
    coord_pair sun_max;     /* maximum coordinate value                     */
    double alph_columns;    /* number columns on alphagraphics screen       */
    double alph_home;       /* height from char origin to top of cell       */
    double alph_lines;      /* number lines on alphagraphics screen         */
    int colour_screen;      /* -1 for colour screen, 0 for black and white  */
    int full_scr_x;         /* width of the workstation screen in pixels    */
    int full_scr_y;         /* height of the workstation screen in pixels   */
/*
 *              Opaque handles to Sunview devices
*/
    Frame   gfx_frame,      /* main graphics frame (window)            */
            quit_frame;     /* base frame for "quit disabled" message  */
    Panel   quit_panel;     /* panel for "quit disabled" message       */
    Panel_item quit_msg;    /* panel_item defining "quit disabled" msg */
    Canvas  t_canvas;       /* graphics canvas (subwindow)             */
    Pixwin  *t_pixwin;      /* pixel memory map                        */
    Pixrect *col_pattern;   /* temporary - for filling coloured boxes  */
    Pixfont *t_sysfont;     /* for system pixel font                   */
    Icon    t_icon;         /* icon for the graphics base frame        */
    Scrollbar left_scrbar,  /* left scrollbar - for vertical scrolling */
              top_scrbar;   /* top scrollbar - for horizontal scrolling*/
  }
    setup_list;

typedef struct  /* This data structure contains changing   */
  {             /* parameters which indicate window status */
    boolean frame_event;  /* gfx_frame event has been dispatched by notifier*/
    boolean show_xhairs;    /* cross hairs are drawn on the screen          */
    int xhair_x;            /* abscissa of received crosshair position      */
    int xhair_y;            /* ordinate of received crosshair position      */
    int waiting_func;       /* name of function waiting for input           */
    int t_return_value;     /* value returned from t_canvas event procedure */
    int was_arrow_key;      /* arrow escape sequence detector               */
  }
    Window_Status;

typedef struct                  /* Complete graphics status desciption */
  {
    ter_status ter;
    Text_Attribute stroke;      /* stroke text attributes              */
    clip_area VAR_PTR view_clip;/* pointer to viewport clipping area   */
    setup_list setup;           /* setup parameters -- not initialized */
                                /* all values are obtaind from SunView */
    Window_Status gfx_win;      /* window status -- not initialized    */
                                /* all values are obtaind from SunView */
    FILE *handle;               /* handle for the graphics window      */
  }
    Gfx_Status;

/* ************************************************************************ */
/*                                                                          */
/*                          Global variables                                */
/*                                                                          */
/* ************************************************************************ */

extern StrokeFont
    asc_font,                   /* font for ASCII chars   */
    sym_font;                   /* font for special chars */

static Gfx_Status VAR_PTR sun3 = (Gfx_Status VAR_PTR) 0;

/*
 *
 *              data for defining colour map segment, list of raster ops,
 *                  dashed line patterns, t_canvas' cursor and gfx_frame's icon
 *
 *             black    orange     green     violet    -----black-----
 *                   red     yellow     blue      brown              white
 *  index        0    1    2    3    4    5    6    7  complementary  15
 *  colour no.   0    2    3    4    5    6    7    8                 1
*/
static unsigned char
   t_red[]   = { 0, 255, 255, 255,   0,   0, 160, 200, 0,0,0,0,0,0,0, 255 },
   t_green[] = { 0,   0, 128, 255, 255,   0,   0, 100, 0,0,0,0,0,0,0, 255 },
   t_blue[]  = { 0,   0,   0,   0,   0, 255, 224,   0, 0,0,0,0,0,0,0, 255 };

static short int
   colour_list[] = { 0, 15, 1, 2, 3, 4, 5, 6, 7 },
/*
 *      ops for drawing lines
*/
   rop_list[] = { PIX_SRC | PIX_DST,            /* DOT_ON,  paint */
                  PIX_NOT( PIX_SRC ) & PIX_DST, /* DOT_OFF, erase */
                  PIX_SRC ^ PIX_DST,            /* DOT_REV, inverting paint */
                  PIX_SRC,                      /* DOT_NEW, write */
                  PIX_DST };                    /* no-op */
/*
 *
 *      Common error messages
*/
static char
    VAR_PTR err_write  = "Graphics base frame does not exist",
    VAR_PTR err_null   = "Arg is NULL pointer",
    VAR_PTR err_fixup  = "device status conflict fixed",
    VAR_PTR mt_str   = "";

/* ************************************************************************* */
/*                                                                           */
/*          Sunview Interface and Pixrect functions and macros used .        */
/*                                                                           */
/* ************************************************************************* */
/*
 * #define
 *      *canvas_pixwin( Canvas canvas )
 * Event
 *      *canvas_window_event( Canvas canvas, Event *event )
 * Cursor
 *      cursor_create( <attribute-list> attributes )
 * caddr_t
 *      cursor_get( Cursor cursor, Cursor_attribute attribute )
 * void cursor_set( Cursor cursor, <attribute-list> attributes )
 *
 * #define event_id( Event *event )
 * #define event_is_button( Event *event )
 * #define event_is_up( Event *event )
 * #define event_set_x( Event *event, int x )
 * #define event_set_y( Event *event, int y )
 * #define event_x( Event *event )
 * #define event_y( Event *event )
 *
 * Icon icon_create( <attribute-list> attributes )
 *
 * #define mpr_static( Pixrect name, int w, int h, int depth, short int *image )
 * Notify_error
 *      notify_dispatch()
 * Notify_error
 *      notify_interpose_destroy_func( Notify_client client,
 *                                     Notify_func destroy_func )
 * Notify_error
 *      notify_interpose_event_func( Notify_client, Notify_func event_func,
 *                                   Notify_event_type type )
 * Notify_value
 *      notify_next_destroy_func( Notify_client client, Event *event,
 *                                Notify_arg arg, Notify_event_type type )
 * Notify_error
 *      notify_veto_destroy( Notify_client client )
 *
 * Panel_item
 *      panel_create_item( Panel panel, <item-type> item_type,
 *                         <attribute-list> attributes )
 *
 * int  pr_dump( Pixrect *input_pr, FILE *output, colormap *colormap,
 *               int type, int copy_flag )
 * int  pr_rop( Pixrect *pr, int dx, int dy, int dw, int dh, int op,
 *              Pixrect *pr, int sx, int sy )
 *
 * int    pw_batch_off( Pixwin *pw )
 * int    pw_batch_on( Pixwin *pw )
 * int    pw_char( Pixwin *pw, int x, int y, int op, Pixfont *font, char c )
 * int    pw_pfsysclose()
 * Pixfont
 *      * pw_pfsysopen()
 * int    pw_putcolormap( Pixwin *pw, int index, int count, unsigned char red[],
 *                      unsigned char green[], unsigned char blue[] )
 * int    pw_replrop( Pixwin *pw, int dx, int dy, int dw, int dh, int op,
 *                    Pixrect *pr, int sx, int sy )
 * int    pw_rop( Pixwin *pw, int dx, int dy, int dw, int dh, int op,
 *              Pixrect *sp, int sx, int sy )
 * (void) pw_setcmsname( Pixwin *pw, char cmsname[NAMESIZE] )
 * int    pw_stencil( Pixwin *dpw, int dx, int dy, int dw, int dh, int op,
 *                  Pixrect *stpr, int stx, int sty,
 *                  Pixrect *spr, int sx, int sy )
 * int    pw_writebackground( Pixwin *pw, int dx, int dy, int dw, int dh, int op )
 * int    pw_vector( Pixwin *pw, int x0, int y0, int x1, int y1,
 *                 int op, int value )
 *
 * Scrollbar
 *      scrollbar_create( <attribute-list> attributes )
 * caddr_t
 *      scrollbar_get( Scrollbar scrollbar, Scrollbar_attribute attribute )
 * void scrollbar_scroll_to( Scrollbar scrollbar, long int new_view_start )
 *
 * Window
 *      window_create( Window owner, <window-type> type,
 *                     <attribute-list> attributes )
 * int  window_destroy( Window win )
 * #define  window_fit( Window win )
 * caddr_t
 *      window_get( Window win, Window_attribute attribute )
 * caddr_t
 *      window_loop( Frame subframe )
 * void window_return( caddr_t value )
 * int  window_set( Window win, <attribute-list> attributes )
*/
/* ************************************************************************* */
/*                                                                           */
/*             Stroke-based character font handler functions used .          */
/*                                                                           */
/* ************************************************************************* */
/*
 * boolean clear_font(), map_font(), regist_font(), reset_font(),
 *         t_parse_options();
 * int     bold_font(), t_strlen(), read_font();
 * char    VAR_PTR define_font(), VAR_PTR file_font(), VAR_PTR t_strcpy();
*/
/* ************************************************************************* */
/*                                                                           */
/*                          Error handler function used .                    */
/*                                                                           */
/* ************************************************************************* */
/*
 * boolean make_error();
*/
/* ************************************************************************* */
/*                                                                           */
/*                      Clipping function and macros used .                  */
/*                                                                           */
/* ************************************************************************* */
/*
 * #define first_clip()
 * #define last_clip()
 * #define max_clip()
 * #define min_clip()
 * clip_area * def2r_clip();
 * boolean clip(), move_clip();
*/

/* ************************************************************************* */
/*                                                                           */
/*                               Internal functions .                        */
/*                                                                           */
/* ************************************************************************* */

static boolean
    create_quit_msg(), next_cursor(), open_frame(), prev_cursor(), t_stkchr();
static char
    VAR_PTR pr_dots(), VAR_PTR pr_mode(), sun_getc();
static int t_txtstr();
static Notify_value
    catch_resize(), catch_destroy();
static void
    canvas_event_proc(), catch_noquit(), flush_frame_events(), put_char(),
    remove_crosshairs(), sun_box(), sun_line(), t_scroll_adjust();
extern Notify_error notify_dispatch();

/* ************************************************************************* */
/*                                                                           */
/*                    Declare K & R standard functions                       */
/*                                                                           */
/* ************************************************************************* */

int abs(), fclose(), fputc(), fflush();
char *malloc();
FILE *fopen();
/* don't declare fprintf() - it is declared differently on different systems */
/* don't declare ferror()  - it is a macro */
/* don't declare getchar() - it is a macro to fgetc() */


/* ************************************************************************ */
/*                                                                          */
/*                    Terminal driving routines                             */
/*                                                                          */
/* ************************************************************************ */


static clip_area VAR_PTR t_begin( device, options )
/*
 *        Start graphics .
 * If ``device'' has been registered as a device, grab the pointer to the
 * graphics status data .
 * If ``device'' has never been registered, allocate memory for the graphics
 * status data, register the device as a Sun 3 Workstation, create the
 * graphics base frame, scrollbars, cursor, canvas and  quit disable message .
 *      In both cases, the options string is parsed and used .
*/
FILE *device;           /* unused pointer to output file */
char VAR_PTR options;   /* unused pointer to options string */
  {
    coord_pair origin;                /* viewport origin */
    Cursor gfx_cursor;
    char VAR_PTR name = "t_begin";
    static Rect VAR_PTR fullscr_rect; /* pointer to full screen size */
/*
 *             initial required terminal status
*/
    static ter_status
        init_ter = { GRAPH_QUIT, DOT_ON, 1, 1, 0, { 0.0, 0.0 }, { 0.0, 0.0 } };
/*
 *              These are the properties which can be set by the ``options''
 *              argument. None of these variables are visible to other functions
*/
    static int
      gfx_left   = NOT_SET,     /* location of left edge of graphics window */
      gfx_top    = NOT_SET,     /* location of top edge of graphics window */
      gfx_width  = NOT_SET,     /* width of graphics window */
      gfx_height = NOT_SET,     /* height of graphics window */
      gfx_closed = NOT_SET;     /* start with closed graphics window */
/*
 *              table of legal options
*/
    static Option_Elem option_list[] = {
      { "LEFT",   1, OPT_INT, (char VAR_PTR) &gfx_left,   OPT_NUM_ARG },
      { "TOP",    1, OPT_INT, (char VAR_PTR) &gfx_top,    OPT_NUM_ARG },
      { "WIDTH",  1, OPT_INT, (char VAR_PTR) &gfx_width,  OPT_NUM_ARG },
      { "HEIGHT", 1, OPT_INT, (char VAR_PTR) &gfx_height, OPT_NUM_ARG },
      { "OPEN",   1, OPT_INT, (char VAR_PTR) &gfx_closed, FALSE },
      { "CLOSED", 1, OPT_INT, (char VAR_PTR) &gfx_closed, TRUE },
      { (char VAR_PTR) 0, 0, 0, (char VAR_PTR) 0, 0.0 } };
/*
 *              define the static data arrays and Pixrects for the mouse cursor
 *              ( which is 16 x 16 cross) and the icon for the graphics window
 *              ( which is a 64 x 64 bitmap image) .
*/
    static short int
      cursor_pixrectdata[] = {
        0x8002,0x4004,0x2008,0x1010,0x0820,0x0440,0x0280,0x0100,
        0x0280,0x0440,0x0820,0x1010,0x2008,0x4004,0x8002,0x0000 },
      t_icon_data[] = {
        0x0400,0x0008,0x0001,0x0000,0x0200,0x0008,0x0001,0x0000,
        0x0100,0x0004,0x0001,0x0000,0x0080,0x0004,0x0002,0x0000,
        0x0040,0x0004,0x0002,0x0000,0x0020,0x0002,0x0002,0x0000,
        0x0010,0x0002,0x0002,0x0002,0x0008,0x0002,0x0004,0x0004,
        0x0004,0x0002,0x0004,0x0008,0x0002,0x0001,0x0004,0x0010,
        0x0001,0x0000,0x0004,0x0010,0x0000,0x8000,0x0000,0x0020,
        0x0000,0x4000,0x0000,0x0040,0x0000,0x2000,0x0000,0x0080,
        0x0000,0x1000,0x0000,0x0100,0x0000,0x0800,0x7FF0,0x0200,
        0x0000,0x0401,0xC01C,0x0000,0x0000,0x0006,0x0003,0x0000,
        0x0000,0x0008,0x0000,0x8000,0x0000,0x0010,0x0000,0x4000,
        0x0000,0x0060,0x0000,0x2000,0x0000,0x0040,0x0000,0x1000,
        0xE000,0x0080,0x0000,0x0803,0x1E00,0x0100,0x0000,0x040C,
        0x01C0,0x0200,0x0000,0x0200,0x0038,0x0200,0x0000,0x0200,
        0x0007,0x0400,0x0000,0x0300,0x0000,0x0400,0x0000,0x0100,
        0x0000,0x0C00,0x0000,0x0100,0x0000,0x0800,0x0000,0x0080,
        0x0000,0x0800,0x0000,0x0080,0x0000,0x0800,0x0000,0x0080,
        0x0000,0x0800,0x0000,0x0080,0x0000,0x0800,0x0000,0x0080,
        0x0000,0x0800,0x0000,0x0080,0x0000,0x0800,0x0000,0x0080,
        0x0000,0x0800,0x0000,0x0080,0x0002,0x0C00,0x0000,0x0080,
        0x007C,0x0400,0x0000,0x0100,0x1F80,0x0400,0x0000,0x0100,
        0xE000,0x0400,0x0000,0x0100,0x0000,0x0200,0x0000,0x0200,
        0x0000,0x0200,0x0000,0x0200,0xFFFF,0xFFFF,0xFFFF,0xFFFE,
        0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
        0x0000,0x0000,0x0000,0x0000,0x0000,0x07FF,0xFFFF,0x8000,
        0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
        0x0000,0x0000,0x0000,0x01A0,0x03FF,0xFFFF,0xFFF8,0x0260,
        0x0000,0x0000,0x0000,0x0220,0x0000,0x0000,0x0000,0x0220,
        0x0000,0x0000,0x0000,0x0220,0xFFFF,0xFFFF,0xFE00,0x0260,
        0x0000,0x0000,0x0000,0x01A0,0x0000,0x0000,0x0003,0xF020,
        0x0000,0x0000,0x0000,0x0220,0xFFFF,0xFFF0,0x0000,0x01C0,
        0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
        0x0000,0x0000,0x0000,0x0000,0xFFFF,0x0000,0x0000,0x0000 };
    mpr_static_static( cursor_pixrect, 16, 16, 1, cursor_pixrectdata );
    mpr_static_static( t_icon_pixrect, 64, 64, 1, t_icon_data );

    CATCH_NULL( device, name, clip_area VAR_PTR );
/*
 *      Find out if ``device'' has been registered as a device .
 *      If it has, check that it is indeed a Sun 3 workstation and grab the
 *      pointer to it graphics status data .
*/
    if ( t_register( T_ISREGIST, device, 0, (char VAR_PTR) 0 ) )
      {
        if ( (int) t_register( T_REG_TYPE, device, 0, (char VAR_PTR) 0 )
                 != SUN_3_WORKSTATION )
            return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
              (char VAR_PTR) "Wrong driver. Device is not a Sun3 workstation" );
        sun3 = (Gfx_Status VAR_PTR) t_register( T_REG_DATA, device, 0,
            (char VAR_PTR) 0 );
      }
    else
      {
/*
 *      ``device'' has never been registered, so allocate memory for the
 *      graphics status data and register the device as a Sun 3 Workstation .
*/
        sun3 = (Gfx_Status VAR_PTR) malloc( (unsigned int) sizeof(Gfx_Status) );
        if ( ! sun3 )
            return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "Cannot allocate memory" );
        if ( ! t_register( T_REGISTER, device,
                 SUN_3_WORKSTATION, (char VAR_PTR) sun3 ) )
          {
            (void) free( (char *) sun3 );
            sun3 = (Gfx_Status VAR_PTR) 0;
            return (clip_area VAR_PTR) make_error( ERR_CALL, name, mt_str );
          }
        sun3->handle = device;  /* point to the graphics device */
/*
 *              Open a graphics base frame of the default size with
 *              label, icon but do not make it visible on the screen
*/
        if ( ( SETUP.t_icon = icon_create( ICON_IMAGE, &t_icon_pixrect, 0 ) )
              && ( SETUP.gfx_frame = window_create( (Window) 0, FRAME,
                   WIN_SHOW, FALSE, FRAME_LABEL, "sun_g",
                   FRAME_ICON, SETUP.t_icon, 0  ) ) )
/*
 *              Temporarily set the size of the graphics base frame to the full
 *              size of the screen . This is done so that the graphics canvas
 *              can be the same size -- less the thickness of the surrounding
 *              borders . The size and location of the base frame may be
 *              adjusted later .
*/
          {
            if ( fullscr_rect = (Rect *) window_get( SETUP.gfx_frame,
                     WIN_SCREEN_RECT ) )
              {
                (void) window_set( SETUP.gfx_frame,
                    WIN_WIDTH, SETUP.full_scr_x = fullscr_rect->r_width,
                    WIN_HEIGHT, SETUP.full_scr_y = fullscr_rect->r_height, 0 );
/*
 *              Create scrollbars with default parameters
*/
                if ( ( SETUP.top_scrbar = scrollbar_create(
                           SCROLL_PAGE_BUTTONS, FALSE, 0 ) ) &&
                     ( SETUP.left_scrbar = scrollbar_create(
                           SCROLL_PAGE_BUTTONS, FALSE, 0 ) ) &&
/*
 *              Create the cursor . The cursor is a simple cross and the
 *              crosshairs are the full size of the graphics canvas . Both the
 *              cursor and  the crosshairs are ``exclusive-or''ed with the
 *              screen data in order to maximize visibility .
*/
                     ( gfx_cursor = cursor_create( CURSOR_SHOW_CURSOR, TRUE,
                          CURSOR_IMAGE, &cursor_pixrect,
                          CURSOR_XHOT, 7,      /* hot-spot is in the   */
                          CURSOR_YHOT, 7,      /* middle of the cursor */
                          CURSOR_OP, PIX_SRC ^ PIX_DST,  /* exclusive or */
                          CURSOR_SHOW_CROSSHAIRS, FALSE,
                          CURSOR_CROSSHAIR_OP, PIX_SRC ^ PIX_DST, 0 ) ) &&
/*
 *              Create the graphics canvas to just fit inside the base frame .
 *              Attach the cursor and ensure that the size of canvas does not
 *              change when the size of the base frame is changed . Do not
 *              create a retained pixrect until the colourmap segment has been
 *              defined . Force notifier to ignore all events except scroll
 *              request and mouse buttons .
*/
                     ( SETUP.t_canvas = window_create( SETUP.gfx_frame, CANVAS,
                          WIN_CURSOR, gfx_cursor,
                          CANVAS_AUTO_SHRINK, FALSE, /* ensure fixed */
                          CANVAS_AUTO_EXPAND, FALSE, /* size canvas  */
                          CANVAS_AUTO_CLEAR,  FALSE, /*no clear before repaint*/
                          CANVAS_RETAINED,    FALSE,
                          WIN_EVENT_PROC, canvas_event_proc,
                          WIN_CONSUME_KBD_EVENT, WIN_NO_EVENTS,
                          WIN_IGNORE_PICK_EVENT, WIN_NO_EVENTS,
                          WIN_CONSUME_PICK_EVENTS, SCROLL_REQUEST,
                              WIN_MOUSE_BUTTONS, 0, 0 ) ) &&
/*
 *              Create quit disable message for blocking a mouse-generated
 *              quit (base frame) event .
*/
                     create_quit_msg() &&
/*
 *              Grab the handle of the virtual pixel map .
*/
                     ( SETUP.t_pixwin = canvas_pixwin( SETUP.t_canvas ) ) &&
/*
 *              Open the pixel font -- just use the default system font .
*/
                     ( SETUP.t_sysfont = pw_pfsysopen() ) )
                  {
/*
 *              define the colour map if a colour_screen
*/
                    if ( SETUP.t_pixwin->pw_pixrect->pr_depth > 1 )
                      {
                        (void) pw_setcmsname( SETUP.t_pixwin, "t_cms" );
                        (void) pw_putcolormap( SETUP.t_pixwin, 0, CMS_SIZE,
                                        t_red, t_green, t_blue );
                        SETUP.colour_screen = 0xffffffff;
                      }
                    (void) window_set( SETUP.t_canvas,
                        CANVAS_RETAINED, TRUE, 0 );
/*
 *              Determine the size of the graphics canvas .
*/
                    SETUP.sun_max._x = (double) ( (int)
                        window_get( SETUP.t_canvas, CANVAS_WIDTH ) - 1 );
                    SETUP.sun_max._y = (double) ( (int)
                        window_get( SETUP.t_canvas, CANVAS_HEIGHT ) -1 );
 /*
  *             Determine the number of tty text lines and columns  on the
  *             screen, and the distance from the character origin to top of
  *             the cell .
*/
                    SETUP.alph_lines =
                        ( (int) SETUP.sun_max._y + 1 ) / ALPH_HEIGHT;
                    SETUP.alph_columns =
                        ( (int) SETUP.sun_max._x + 1 ) / ALPH_WIDTH;
                    SETUP.alph_home = (double)
                        ( SETUP.t_sysfont->pf_char[ (int) 'h' ].pc_home.y );
/*
 *              Install the interposed frame RESIZE event handler and
 *              install the interposed quit handler .
*/
                    (void) notify_interpose_event_func( SETUP.gfx_frame,
                        catch_resize, NOTIFY_SAFE );
                    (void) notify_interpose_destroy_func( SETUP.gfx_frame,
                        catch_destroy );
                  }
                else
                  {
                    (void)  make_error( ERR_ABORT, name, (char VAR_PTR)
                  "Cannot create scrollbar, cursor, canvas or quit message;" );
                    return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                        (char VAR_PTR) "or cannot get pixwin or system font" );
                  }
              }
            else
                return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                    (char VAR_PTR) "Cannot find screen size" );
          }
        else
            return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "Cannot create base frame or icon" );
        origin._x = origin._y = 0.0;
        if ( ! ( VIEW_CLIP = def_2rclip( (coord_pair VAR_PTR) &origin,
                                         (coord_pair VAR_PTR) &(SETUP.sun_max),
                                         (coord_pair VAR_PTR) &origin ) ) )
            (void) make_error( ERR_CALL, name, mt_str );
/*
 *              Set the graphics pen and text cursor postion .
 *              Set default stroke text parameters .
*/
        (void) reset_font( (Text_Attribute VAR_PTR) &STROKE,
             MIN( fullscr_rect->r_width, fullscr_rect->r_height ) / 20.0 );
        TER = init_ter;
        TER.cursor._y = SETUP.sun_max._y + SETUP.alph_home;
      }
/*
 *              WHAT FOLLOWS IS DONE EVERY TIME t_begin() IS CALLED
 *
 *                  OPTIONS
 *              First clear the parameters which can be changed by the options .
 *              The decode the options list set all the parameter still
 *              cleared to the default values and finally ensure that the
 *              values are valid .
*/
    gfx_left = gfx_top = gfx_width = gfx_height = gfx_closed = NOT_SET;
    if ( options )
        if ( ! t_parse_options( options, (Option_Elem VAR_PTR) option_list,
                 name ) )
             return (clip_area VAR_PTR) NULL;
/*
 *              All values NOT_SET are given default values .
*/
    if ( gfx_left == NOT_SET )
        gfx_left = SETUP.full_scr_x / 10;
    if ( gfx_top == NOT_SET )
        gfx_top = SETUP.full_scr_y / 10;
    if ( gfx_width == NOT_SET )
        gfx_width = 9 * SETUP.full_scr_x / 10 - gfx_left;
    if ( gfx_height == NOT_SET )
        gfx_height = 9 * SETUP.full_scr_y / 10 - gfx_top;
    if ( gfx_closed == NOT_SET )
        gfx_closed = FALSE;
/*
 *              Ensure valid window location and size .
*/
    gfx_left   = LIMIT( gfx_left, 0, SETUP.full_scr_x - MIN_FRAME_WIDTH );
    gfx_top    = LIMIT( gfx_top, 0, SETUP.full_scr_y - MIN_FRAME_HEIGHT );
    gfx_width  = LIMIT( gfx_width,
         MIN_FRAME_WIDTH, SETUP.full_scr_x - gfx_left );
    gfx_height = LIMIT( gfx_height,
         MIN_FRAME_HEIGHT, SETUP.full_scr_y - gfx_top );
/*
 *              Set the location and size of the base frame . The default
 *              location of the top-left corner is the top left corner of the
 *              screen ( i.e. (0,0) ) . The window is displayed and is opened
 *              ( unless the CLOSED option is selected ) .
*/
    (void) window_set( SETUP.gfx_frame,
        WIN_X,        gfx_left,
        WIN_Y,        gfx_top,
        WIN_WIDTH,    gfx_width,
        WIN_HEIGHT,   gfx_height,
        FRAME_CLOSED, gfx_closed,
        WIN_SHOW,     TRUE, 0 );
    TER.mode = GRAPH_OPEN;
    flush_frame_events();
    return VIEW_CLIP;
  }


static boolean t_end()
/*
 *      End of graphics . Delete character fonts and frees the memory allocated
 *  to them, destroy the graphics base frames . This function always succeeds .
*/
  {
    char VAR_PTR name = "t_end";
/*
 *      Deregister the device and free the memory used by the graphics status
 *      data .
*/
    START_UP( name, boolean );
    if ( ! t_register( T_DEREGIST, sun3->handle,
            SUN_3_WORKSTATION, (char VAR_PTR)  sun3 ) )
        (void) make_error( ERR_WARN, name, mt_str );
    (void) free( (char *) sun3 );
/*
 *              Remove the character fonts .
*/
    if ( asc_font.font_got )
        (void) clear_font( (StrokeFont VAR_PTR) &asc_font );
    if ( sym_font.font_got )
        (void) clear_font( (StrokeFont VAR_PTR) &sym_font );
    if ( TER.mode != GRAPH_QUIT )
      {
/*
 *              Destroy all base frames - including the `quit disabled'
 *              message frame without asking for confirmation .
*/
        GFX_WIN.waiting_func = T_END;
        (void) window_set( SETUP.gfx_frame, FRAME_NO_CONFIRM, TRUE, 0 );
        (void) window_destroy( SETUP.gfx_frame );
        (void) window_set( SETUP.quit_frame, FRAME_NO_CONFIRM, TRUE, 0 );
        (void) window_destroy( SETUP.quit_frame );
/*
 *              close system font
*/
        (void) pw_pfsysclose();
        TER.mode = GRAPH_QUIT;
      }
    return TRUE;
  }


static boolean t_vwport( clip_ptr )
/*
 *      Set the clipping limits for the viewport . The old limits are replaced .
 * The `clip_ptr' must point to  clip_area which has been defined by
 * `def_2rclip()' . The function fails if
 *   a) clip_ptr is a NULL pointer
 *   b) clip_ptr points to something which is not a structure for a two
 *      dimensional rectangular clipping region .
 *
 * Argument
*/
clip_area VAR_PTR clip_ptr;
  {
    char VAR_PTR name = "t_vwport";
    START_UP( name, boolean );
    CATCH_NULL( clip_ptr, name, boolean );
    if ( clip_ptr->id_struct == CLIP2R_ID )
        if ( min_clip( clip_ptr, _x ) >  -0.5 &&
             max_clip( clip_ptr, _x ) <= SETUP.sun_max._x &&
             min_clip( clip_ptr, _y ) >  -0.5 &&
             max_clip( clip_ptr, _y ) <= SETUP.sun_max._y )
          {
            VIEW_CLIP = clip_ptr;
            flush_frame_events();
            return TRUE;
          }
    return make_error(ERR_ABORT, name, (char VAR_PTR) "Invalid pointer" );
  }


static boolean t_colour( ipen )
/*
 *    Selects specified colour ( or pen ) .
 *
 * On the monochrome workstation, colours are simulated with various dashed
 * line styles . The only true ``colours'' available are dark and bright . The
 * returned value is TRUE if  0 <= ipen <= 8 . If these limits are exceeded,
 * the value is wrapped around into the range 1 to 8 . The dashed line
 * equivalents of the colours are,
 *
 *    IPEN  MONOCHROME SCREEN         COLOUR SCREEN
 *          LINE STYLE                  COLOUR
 *     0    white (background)        black
 *     1    black (foreground)        white
 *     2    dotted                    red
 *     3    short dash                orange
 *     4    long dash                 yellow
 *     5    dot dash                  green
 *     6    short-dash long-dash      blue
 *     7    2-short-dashes long-dash  violet
 *     8    3-dots long-dash          brown
 *
 *   Argument
*/
int ipen;                /* pen number or colour */
  {
    START_UP( (char VAR_PTR) "t_colour", boolean );
    TER.colour = ( abs( ipen ) - 1 ) % 8 + 1;
    flush_frame_events();
    return 0 <= ipen && ipen <= 8;
  }


static boolean t_dashed( style )
/*
 *      Set the dashed line style . The returned value is TRUE if
 * 0 <= style <= 8 . If these limits are exceeded, the value is wrapped
 * around into the range 1 to 8 .
 *
 *    STYLE   SCREEN
 *          LINE STYLE
 *     0      white  - plot vector endpoints only (invisible)
 *     1      solid  (black)
 *     2      dotted
 *     3    short dash
 *     4    long dash
 *     5    dot dash
 *     6    short-dash long-dash
 *     7    2-short-dashes long-dash
 *     8    3-dots long-dash
 *
*/
int style;        /* dashed line style */
  {
    START_UP( (char VAR_PTR) "t_dashed", boolean );
    TER.dash = ( abs( style ) - 1 ) % 8 + 1;
    flush_frame_events();
    return 0 <= style && style <= 8;
  }



static boolean t_thick( width )
/*
 * Set the ``width'' of lines and text . The returned value is TRUE
 * if 0 <= thickness.
*/
int width;      /* line thickness */
  {
    START_UP( (char VAR_PTR) "t_thick", boolean );
    TER.thick = width;
    flush_frame_events();
    return 0 <= width;
  }


/* ************************************************************************ */
/*                                                                          */
/*           Change Data Level, Visible, Invisible, Revers, Replace         */
/*                                                                          */
/* ************************************************************************ */


static boolean t_visib()
/*
 *  Draw visible vectors . Sets screen into "dots on" mode . This function
 *  always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_visib", boolean );
    TER.dots = DOT_ON;
    return TRUE;
  }


static boolean t_invis()
/*
 *  Draw invisible vectors . Sets screen into "dots off" mode . This function
 *  always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_invis", boolean );
    TER.dots = DOT_OFF;
    return TRUE;
  }


static boolean t_revers()
/*
 *  Reverse the pixel dots when plotting vectors, plotting points or
 *  alphagraphic characters . Sets screen into "compliment" mode . This function
 *  always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_revers", boolean );
    TER.dots = DOT_REV;
    return TRUE;
  }


static boolean t_replac()
/*
 *  Replace pixel contents with the new data . Sets screen into "replace"
 *  mode . This function always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_replac", boolean );
    TER.dots = DOT_NEW;
    return TRUE;
  }



/* ************************************************************************ */
/*                                                                          */
/*                          Turn graphics on or off                         */
/*                                                                          */
/* ************************************************************************ */


static boolean t_gr_off()
/*
 *      Turn graphics off, i.e. close the graphics base frame to an icon. The
 * returned result is FALSE if the graphics base frame does not exist.
*/
  {
    char VAR_PTR name = "t_gr_off";
    START_UP( name, boolean );
    if ( TER.mode != GRAPH_QUIT )
      {
/*
 *          Close the window only if it is open;
 *          check that the terminal status is correct .
*/
        if ( ! ( boolean) window_get( SETUP.gfx_frame, FRAME_CLOSED ) )
          {
            if ( TER.mode != GRAPH_OPEN )
                (void) make_error( ERR_WARN, name, err_fixup );
            (void) window_set( SETUP.gfx_frame, FRAME_CLOSED, TRUE, 0 );
          }
        flush_frame_events();
        TER.mode = GRAPH_CLOSED;
        return TRUE;
      }
    return make_error( ERR_ABORT, name, err_write );
  }


static boolean t_gr_on()
/*
 *      Turn graphics on, i.e. open (display) the graphics base frame. The
 * returned result is FALSE if the graphics base frame does not exist.
*/
  {
    char VAR_PTR name = "t_gr_on";
    START_UP( name, boolean );
    flush_frame_events();
    return open_frame( name );
  }


static boolean open_frame( caller )
/*
 * Turn graphics on, i.e. open (display) the graphics base frame. The
 * returned result is FALSE if the graphics base frame does not exist.
 * This function is for internal use by other driver functions only .
*/
char VAR_PTR caller;    /* name of calling function */
  {
    if ( TER.mode != GRAPH_QUIT )
      {
        if ( ( boolean) window_get( SETUP.gfx_frame, FRAME_CLOSED ) )
          {
            if ( TER.mode != GRAPH_CLOSED )
                (void) make_error( ERR_WARN, caller, err_fixup );
            (void) window_set( SETUP.gfx_frame, FRAME_CLOSED, FALSE, 0 );
          }
        TER.mode = GRAPH_OPEN;
        return TRUE;
      }
    return make_error( ERR_ABORT, caller, err_write );
  }


/* ************************************************************************ */
/*                                                                          */
/*                         Miscellaneous stuff                              */
/*                                                                          */
/* ************************************************************************ */


static boolean t_newpag()
/*
 * Clear the graphics screen . This function fails if the graphics base frame
 * does not exist.
*/
  {
    char VAR_PTR name = "t_newpag";
    START_UP( name, boolean );
    if ( TER.mode != GRAPH_QUIT )
      {
        remove_crosshairs();
/*
 *              Clear the graphics canvas .
*/
        (void) pw_writebackground( SETUP.t_pixwin, 0, 0,
            1 + (int) SETUP.sun_max._x, 1 + (int) SETUP.sun_max._y,
            PIX_SRC );  /* write background colour only */
        flush_frame_events();
        return TRUE;
      }
    return make_error( ERR_WARN, name, err_write );
  }


static boolean t_newscr()
/*
 * Clear the window from which the program is run . This function fails if it
 * cannot write to the parent window.
 * As a temporary measure, we do it by sending 50 '\n's to the screen. The
 * proper way is to send the appropriate terminal escape sequence .
*/
  {
    int counter = 0;
    boolean ok = TRUE;
    START_UP( (char VAR_PTR) "t_newscr", boolean );
    while( counter++ < (int) SETUP.alph_lines && ok )
        ok &= '\n' == putchar( '\n' );
    flush_frame_events();
    return ok;
  }


static boolean t_bell()
/*
 *      Ring bell . This function fails if it cannot write to the parent
 *  window.
*/
  {
    START_UP( (char VAR_PTR) "t_bell", boolean );
    flush_frame_events();
    return fputc( BEL, stdout ) == BEL && fflush( stdout );
  }


static boolean t_wait()
/*
 *  Waits for the to press the carriage return . If the graphics frame is open,
 *  the cursor must be inside the it . If the graphics frame is closed, the
 *  cursor must be in the window from which the program is running . The user
 *  can close the graphics window while `t_wait()" is executing, but cannot
 *  open it . The returned value is always TRUE .
*/
  {
    START_UP( (char VAR_PTR) "t_wait", boolean );
    if ( TER.mode == GRAPH_OPEN )
      {
        remove_crosshairs();
/*
 *              Allow canvas to receive ASCII keyboard events .
*/
        (void) window_set( SETUP.t_canvas, WIN_CONSUME_KBD_EVENT,
                    WIN_ASCII_EVENTS, 0 );
/*
 *              Call the notifier until the `wait' request has been satisfied
 *              by the canvas event procedure .
*/
        for( GFX_WIN.waiting_func = T_WAIT;
                GFX_WIN.waiting_func == T_WAIT && TER.mode == GRAPH_OPEN; )
            (void) notify_dispatch();
/*
 *              Canvas ignores ASCII keyboard events .
*/
        (void) window_set( SETUP.t_canvas, WIN_IGNORE_KBD_EVENT,
            WIN_NO_EVENTS, 0 );
        if ( TER.mode == GRAPH_OPEN )
            return TRUE;
      }
/*
 *              Wait for CR from the parent window if the graphics frame window
*/
    while( getchar() != '\n' );
    flush_frame_events();
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                   Writing Stroke based character                         */
/*                                                                          */
/* ************************************************************************ */


static boolean t_txtmov( point )
/*
 *    Move the text cursor to coordinate `point' . The returned result is
 *  TRUE only if the next printable character is within the screen limits .
 *
 * Argument
*/
register coord_pair VAR_PTR point;
  {
    char VAR_PTR name = "t_txtmov";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    TER.cursor = *point;
    flush_frame_events();
    return
      -0.5 < point->_x &&
             point->_x <= SETUP.sun_max._x - ALPH_WIDTH + 0.5 &&
      -0.5 < point->_y &&
             point->_y <= SETUP.sun_max._y + SETUP.alph_home + 0.5;
  }


static boolean t_txtcur( row, column )
/*
 *    Move the text cursor to text row `row' and text column `column' .
 *    The text row is numbered from the top of the screen . The returned
 *    result is TRUE only if the next printable character is within screen
 *    limits . The row and column positions are calculated for the default
 *    alphagraphics character size .
 *
 * Argument
*/
int row, column;
  {
    START_UP( (char VAR_PTR) "t_txtcur", boolean );
    TER.cursor._x = ALPH_WIDTH * column;
    TER.cursor._y = SETUP.sun_max._y + SETUP.alph_home - ALPH_HEIGHT * row;
    flush_frame_events();
    return 0 <= column  &&  column <= SETUP.alph_columns  &&
           0 <= row     &&  row    <= SETUP.alph_lines;
  }


static boolean t_txtset( attribute, value )
/*
 *  Set the stroke character attribute `attribute' to `*value' .
 *  If `attribute' is negative, the current value of the attribute is returned
 *  to `*value' . This function fails if the `attribute' cannot be recognized or
 *  `value' is a null pointer . Appropriate error messages are generated .
*/
int attribute;
char VAR_PTR value;
  {
    char VAR_PTR name = "t_txtset";
    double *dvalue;                 /* pointer to a double `value' */
    dvalue = (double *) value;
    if ( attribute != T_CH_RESET  && attribute != T_CH_BOLD  &&
             attribute != T_CH_NOBOLD && attribute != T_ASCII    &&
             attribute != T_SYMBOL )
        CATCH_NULL( value, name, boolean );
    START_UP( name, boolean );
    flush_frame_events();
    switch( attribute )
      {
        case  T_CH_RESET :
            (void) reset_font( (Text_Attribute VAR_PTR) &STROKE,
                MAX( SETUP.sun_max._x, SETUP.sun_max._y ) / 20.0 );
            break;
        case  T_CH_HEIGHT :
            if ( STROKE.height != *dvalue )
              {
                STROKE.height = LIMIT( *dvalue, 5.0,
                    MIN( SETUP.sun_max._x, SETUP.sun_max._y ) );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_HEIGHT :
            *dvalue = STROKE.height;
            break;
        case  T_CH_ASPECT :
            if ( STROKE.aspect != *dvalue )
              {
                STROKE.aspect = LIMIT( *dvalue, 0.2, 5.0 );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_ASPECT :
            *dvalue = STROKE.aspect;
            break;
        case  T_CH_SLANT :
            if ( STROKE.slant != *dvalue )
              {
                STROKE.slant = LIMIT( *dvalue, -80.0, 80.0 ) * PI / 180.0;
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_SLANT :
            *dvalue = STROKE.slant * 180.0 / PI;
            break;
        case  T_CH_ANGLE :
            if ( STROKE.angle != *dvalue )
              {
                STROKE.angle = PI / 180.0 * ( *dvalue -
                    360.0 * (double) (int) ( *dvalue / 360.0 ) );
                STROKE.angle = STROKE.angle < -PI ? STROKE.angle + 2.0 * PI :
                    ( STROKE.angle > PI ? STROKE.angle-2.0*PI : STROKE.angle );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_ANGLE :
            *dvalue = STROKE.angle * 180.0 / PI;
            break;
        case  T_CH_GAP :
            STROKE.gap = LIMIT( *dvalue, 0.05, 5.0 );
            break;
        case -T_CH_GAP :
            *dvalue = STROKE.gap;
            break;
        case  T_CH_ALIGN :
            STROKE.align.vert  = TOLOWER( *value );
            value++;
            STROKE.align.horiz = TOLOWER( *value );
            if ( STROKE.align.horiz != 'l' && STROKE.align.horiz != 'c' &&
                    STROKE.align.horiz != 'r' )
                STROKE.align.horiz = 'l';
            if ( STROKE.align.vert != 'b' && STROKE.align.vert != 'c' &&
                    STROKE.align.vert != 't' )
                STROKE.align.vert = 'b';
            break;
        case -T_CH_ALIGN :
            *value++ = STROKE.align.vert;
            *value++ = STROKE.align.horiz;
            *value   = NUL;
            break;
        case T_ASCII :
            STROKE.font_ptr = (StrokeFont VAR_PTR) &asc_font;
            break;
        case T_SYMBOL :
            STROKE.font_ptr = (StrokeFont VAR_PTR) &sym_font;
            break;
        case T_CH_BOLD :
            STROKE.bold = TRUE;
            break;
        case T_CH_NOBOLD :
            STROKE.bold = FALSE;
            break;
        default :
            return make_error( ERR_ABORT, name,
                (char VAR_PTR) "Unknown stroke text attribute" );
      }
    return TRUE;
  }


static boolean t_symdrw( point, symbol )
/*
 *      draws a centred plotting symbol `symbol' at position `point' according
 *      to the selected stroke character attributes . This function fails if
 *      the font for the symbol is not defined or if it cannot write to the
 *      screen .
*/
coord_pair VAR_PTR point;
char       symbol;
  {
    char VAR_PTR name = "t_symdrw";
    START_UP( name, boolean );
    if ( TER.mode != GRAPH_QUIT )
      {
        int font_index = symbol - 32;
        coord_pair offset, lower_left;
        StrokeChar VAR_PTR char_ptr;

        CATCH_NULL( point, name, boolean );
        TER.pen = *point;
        remove_crosshairs();
/*
 *           if necessary, read character fonts
*/
        if ( ! STROKE.font_ptr->font_got )
            if ( ! read_font( STROKE.font_ptr, file_font( STROKE.font_ptr ),
                              name ) )
                return FALSE;
/*
 *      CALCULATE LOCATION OF LOWER LEFT CORNER OF THE SYMBOL
 *
 *      determine the character cell size vector
*/
        if ( STROKE.changed )
            (void) map_font( (Text_Attribute VAR_PTR) &STROKE, PEN_THICKNESS );
        char_ptr = STROKE.font_ptr->font_dir + font_index;
        offset._x = STROKE.font_ptr->font_wid * 0.5;
        offset._y = char_ptr->font_midpt;
/*
 *    determine start position
*/
        lower_left._x = point->_x -
            STROKE.map[0][0] * offset._x - STROKE.map[0][1] * offset._y;
        lower_left._y = point->_y -
            STROKE.map[1][0] * offset._x - STROKE.map[1][1] * offset._y;
/*
 *      DRAW THE SYMBOL
*/
        if ( 0 <= font_index && font_index < STROKE.font_ptr->font_siz )
          {
            if ( char_ptr->font_char == symbol && char_ptr->font_npts > 0 )
              {
                boolean ok;
                pw_batch_on( SETUP.t_pixwin );
                ok = bold_font( (Text_Attribute VAR_PTR) &STROKE,
                            (coord_pair VAR_PTR) &lower_left,
                            (char VAR_PTR) &font_index, t_stkchr );
                pw_batch_off( SETUP.t_pixwin );
/*              don't flush_frame_events() -- it is done in t_stkdrw */
                return ok & move_clip( (coord_triple VAR_PTR ) point,
                    VIEW_CLIP );
              }
          }
      }
    return make_error( ERR_ABORT, name, (char VAR_PTR) "Unknown symbol" );
  }


static int t_txtdrw( string )
/*
 *      Draws the text contained in the character string `string' at the
 *   current cursor location according to the selected stroke character
 *   attributes . The returned value is the number of characters drawn . If
 *   an error occurs, the returned value is 0 . An error may be generated
 *   because
 *     a) a character in `string' is unavailable in the specified font
 *     b) the character is not printable
 *     c) could not update the pen position after drawing the characters
 *     d) it cannot write to the terminal
*/
char VAR_PTR string;    /* character string to draw */
  {
    int char_count=0;   /* counts each character as it is drawn */
    char VAR_PTR name = "t_txtdrw";
    START_UP( name, int );
    if ( TER.mode != GRAPH_QUIT )
      {
        boolean ok = TRUE;      /* success flag                             */
        int font_index,         /* index of the character in font directory */
            n_chars;            /* length of `string'                       */
        coord_pair cell_size,   /* size of the character cell               */
                   offset,
                   point;       /* location of char cell lower left corner  */
        CATCH_NULL( string, name, int );
/*
 *           if necessary, read character fonts
*/
        if ( ! STROKE.font_ptr->font_got )
            if ( ! read_font( STROKE.font_ptr, file_font( STROKE.font_ptr ),
                              name ) )
                return 0;
/*
 *      CALCULATE LOCATION OF LOWER LEFT CORNER OF THE STRING
 *
 *      determine the character cell size vector
*/
        remove_crosshairs();
        if ( STROKE.changed )
            (void) map_font( (Text_Attribute VAR_PTR) &STROKE, PEN_THICKNESS );
        n_chars = t_strlen( string );
        cell_size._x = STROKE.font_ptr->font_wid *
            ( 1.0 + STROKE.gap / STROKE.aspect );
        cell_size._y = STROKE.font_ptr->font_hei;
        switch( STROKE.align.horiz )
          {
            case 'r' :  offset._x = cell_size._x * n_chars;
                        break;
            case 'c' :  offset._x = cell_size._x * n_chars / 2.0;
                        break;
            case 'l' :  offset._x = 0.0;
                        break;
          }
        switch( STROKE.align.vert )
          {
            case 't' :  offset._y = cell_size._y;
                        break;
            case 'c' :  offset._y = cell_size._y / 2.0;
                        break;
            case 'b' :  offset._y = 0.0;
                        break;
          }
/*
 *    determine start position
*/
        point._x = TER.cursor._x - STROKE.map[0][0] * offset._x -
                   STROKE.map[0][1] * offset._y;
        point._y = TER.cursor._y - STROKE.map[1][0] * offset._x -
                   STROKE.map[1][1] * offset._y;
        cell_size._y = STROKE.map[1][0] * cell_size._x;
        cell_size._x = STROKE.map[0][0] * cell_size._x;
/*
 *      DRAW EACH CHARACTER IN THE STRING
*/
        for( ; *string; string++, char_count++ )
          {
            font_index = *string - 32;
            if ( 0 <= font_index && font_index < STROKE.font_ptr->font_siz )
                if ( (STROKE.font_ptr->font_dir)[font_index].font_char ==
                         *string &&
                     (STROKE.font_ptr->font_dir)[font_index].font_npts > 0 )
                  {
                    pw_batch_on( SETUP.t_pixwin );
                    ok &= bold_font( (Text_Attribute VAR_PTR) &STROKE,
                                     (coord_pair VAR_PTR) &point,
                                     (char VAR_PTR) &font_index, t_stkchr );
                    pw_batch_off( SETUP.t_pixwin );
/*                  don't flush_frame_events() -- it is done in t_stkdrw */
                    point._x += cell_size._x;
                    point._y += cell_size._y;
                  }
                else
                    (void) make_error( ERR_WARN, name,
                        (char VAR_PTR) "character unavailable" );
             else
                 (void) make_error( ERR_WARN, name,
                     (char VAR_PTR) "not printable character" );
          }
        TER.cursor = point;
        if ( !( ok &= move_clip((coord_triple VAR_PTR) &TER.pen, VIEW_CLIP ) ) )
            (void) make_error( ERR_CALL, name, err_null );
      }
    return char_count;
  }


static boolean t_stkchr( index, point )
/*
 *      Draws character identified by `index', at location `point' .
 *      This is a single pass -- i.e. no boldface .
*/
char VAR_PTR index;       /* pointer to index into the font of the character */
coord_pair VAR_PTR point; /* location of character */
  {
    enum dev_dots txt_dots;     /* indicates the raster op to be used */
    int point_count;       /* counts the points plotted for each char */
    boolean pen_up;             /* move with pen up                   */
    float VAR_PTR coord_ptr;    /* pointer to font coordinate values  */
    coord_pair cord;            /* coordinate from the font - to plot */
    double temp;
    StrokeChar VAR_PTR char_ptr =
        STROKE.font_ptr->font_dir + *( (int VAR_PTR) index );
/*
 *      set colour (dashed lines not allowed)
*/
    if ( ! TER.colour )
        txt_dots = DOT_OFF;
    else if ( SETUP.colour_screen && TER.colour > 1 &&
              TER.dots == DOT_ON )
        txt_dots = DOT_NEW;
    else
        txt_dots = TER.dots;
/*
 *      step through the coordinates which define the character
*/
    coord_ptr =  char_ptr->font_coord;
    for( point_count = 0; point_count < char_ptr->font_npts; point_count++ )
      {
/*
 *      calculate stroke coordinate position in pixel units
*/
        pen_up = ! point_count;
        cord._x = *coord_ptr++;
        cord._y = *coord_ptr++;
        if ( cord._x < 0.0 )
          {
            pen_up = TRUE;
            cord._x = -cord._x;
          }
        if ( cord._y < 0.0 )
          {
            pen_up = TRUE;
            cord._y = -cord._y;
          }
        cord._y -= char_ptr->font_shift;
        temp = STROKE.map[0][0] * cord._x + STROKE.map[0][1] * cord._y
                + point->_x;
        cord._y = STROKE.map[1][0] * cord._x + STROKE.map[1][1] * cord._y
                + point->_y;
        cord._x = temp;
/*
 *      plot the point
*/
        if ( pen_up )
            (void) move_clip( (coord_triple VAR_PTR) &cord, VIEW_CLIP );
        else if ( clip( (coord_triple VAR_PTR ) &cord, VIEW_CLIP ) )
            (void) pw_vector( SETUP.t_pixwin,
                (int) ( first_clip( VIEW_CLIP, _x ) + 0.5 ),
                (int) ( SETUP.sun_max._y - first_clip( VIEW_CLIP, _y) + 0.5 ),
                (int) ( last_clip( VIEW_CLIP, _x ) + 0.5 ),
                (int) ( SETUP.sun_max._y - last_clip( VIEW_CLIP, _y ) + 0.5 ),
                SUN_ROP( txt_dots ) |
                    ( TER.dots == DOT_OFF || TER.dots == DOT_REV ?
                        0 : SUN_COLOUR( TER.colour ) ),
                15 );
      }
    flush_frame_events();
    return TRUE;
  }


static char VAR_PTR t_ch_def( what, character, index, point )
/*
 * This function allows the user to obtain the shape of any character defined
 * in the ASCII and SYMBOL fonts. With the exception of string alignment, the
 * transformations implied by t_txtset are applied to the returned coordinates.
 * This function is merely a call to the function ``define_font'' with an
 * additional argument defining which set of text attributes to use .
*/
int what;           /* specifies what to do (T_FT_* defined values) */
char character;     /* character to fetch from the font */
int index;          /* coordinate number   */
coord_pair
    VAR_PTR point;  /* returned coordinate */
  {
    START_UP( (char VAR_PTR) "t_ch_def", char VAR_PTR );
    return define_font( (Text_Attribute VAR_PTR) &STROKE, PEN_THICKNESS,
        what, character, index, point );
  }


/* ************************************************************************ */
/*                                                                          */
/*                     Reading and Writing tty character                    */
/*                                                                          */
/* ************************************************************************ */

#define DEL_CURSOR  '\001'
#define PUT_CURSOR  '\000'
#define IN_VIEW     ( TER.cursor._x >= min_clip( VIEW_CLIP, _x ) && \
      TER.cursor._x <= max_cursor._x && \
      TER.cursor._y >= min_clip( VIEW_CLIP, _y ) && \
      TER.cursor._y <= max_cursor._y )
static coord_pair max_cursor;

static char VAR_PTR t_txtget( string, max_chars )
/*
 *    Reads a maximum of `max_chars' characters from `stdin' into `string'
 * until a newline character is read . The newline character is not stored in
 * `string' . `string' is terminated with NUL . If using the alpha-numeric
 * screen, this function simply gets() . If using the graphics screen this
 * function is again like gets() except that the echoed character is confined
 * to the viewport . If the cursor is initially outside the viewport,
 * t_txtget() starts echoing characters at the top left corner of the viewport.
 * T_txtget will not echo any more characters if it runs out of space at the
 * bottom of the viewport .If an error occurs or no characters are read, a NULL
 * pointer is returned, otherwise `string' is returned .
*/
char VAR_PTR string;    /* result string */
int max_chars;          /* maximum number of characters in result string */
  {
    char VAR_PTR ptr;   /* pointer to current character in the result string */
    char VAR_PTR name = "t_txtget";
    START_UP( name, char VAR_PTR );
    if ( TER.mode == GRAPH_OPEN )
      {
/*
 *              Get characters from the graphics window if it is open .
 *              *******************************************************
*/
        boolean in_viewport;
        CATCH_NULL( string, name, char VAR_PTR );
/*
 *              Ensure cursor is in the viewport if possible .
*/
        remove_crosshairs();
        max_cursor._x = max_clip(VIEW_CLIP,_x) - ALPH_WIDTH + 1;
        max_cursor._y = max_clip(VIEW_CLIP,_y) + SETUP.alph_home;
        if ( ! IN_VIEW )
          {
            TER.cursor._x = min_clip( VIEW_CLIP, _x );
            TER.cursor._y = max_cursor._y;
          }
/*
 *              Make a cursor on the screen where the first character
 *              will be placed .
*/
        if ( in_viewport = IN_VIEW )
            put_char( PUT_CURSOR );
/*
 *              Loop for getting characters .
*/
        for( ptr = string; ( *ptr = sun_getc() ) != '\r' && max_chars--; )
          {
/*
 *              Test for DEL key . Then move the cursor back one position .
*/
            if ( *ptr == DEL )
              {
                if ( string < ptr )
                  {
                    pw_batch_on( SETUP.t_pixwin );
/*
 *              Don't do anything if *ptr is the first character in the string .
 *              Delete the cursor, move it back one space and rewrite it .
*/
                    if ( in_viewport )
                        put_char( DEL_CURSOR );
                    if ( in_viewport = prev_cursor() )
                        put_char( PUT_CURSOR );
                    pw_batch_off( SETUP.t_pixwin );
                    ptr--;
                  }
              }
/*
 *              Test for printable character
*/
            else if ( ' ' <= *ptr && *ptr <= '~' )
              {
                pw_batch_on( SETUP.t_pixwin );
/*
 *              Only write the character if it is within the viewport .
 *              Delete the cursor, write the character and rewrite the cursor .
*/
                if ( in_viewport )
                  {
                    put_char( DEL_CURSOR );
                    put_char( *ptr );
                  }
                if ( in_viewport = next_cursor() )
                    put_char( PUT_CURSOR );
                pw_batch_off( SETUP.t_pixwin );
                ptr++;
              }
            else
                ptr++;
          }
        if ( in_viewport )
            put_char( DEL_CURSOR );
        TER.cursor._x = min_clip( VIEW_CLIP, _x );
        TER.cursor._y -= ALPH_HEIGHT;
      }
    else /* ( TER.mode == GRAPH_CLOSED ) */
/*
 *              If the graphics window is closed, simply use the parent window .
*/
        for( ptr = string; ( *ptr = getchar() ) != '\n' && max_chars--; ptr++ );
    *ptr = NUL;
    return string;
  }


static char sun_getc()
/*
 * Gets a printable ASCII character from the keyboard .
*/
  {
/*
 *              Enable keyboard events .
*/
    (void) window_set( SETUP.t_canvas, WIN_CONSUME_KBD_EVENTS, WIN_ASCII_EVENTS,
        WIN_RIGHT_KEYS, WIN_LEFT_KEYS, WIN_TOP_KEYS, 0, 0 );
/*
 *              Call the notifier until the ``t_txtget()'' character request
 *              has been satisfied by the canvas event procedure .
*/
    for( GFX_WIN.waiting_func = T_TXTGET; GFX_WIN.waiting_func == T_TXTGET; )
        (void) notify_dispatch();
/*
 *              Disable keyboard events .
*/
    (void) window_set( SETUP.t_canvas, WIN_IGNORE_KBD_EVENT, WIN_NO_EVENTS, 0 );
    flush_frame_events();
    return (char) GFX_WIN.t_return_value;
  }


static void put_char( character )
/*
 *      Place a printable ASCII character or cursor on the graphics screen.
*/
char character;
  {
/*
 *      Raster operations for writing characters to black and white screens .
*/
   static short int bw_char_op[2][4] =
     { { PIX_NOT( PIX_SRC ) & PIX_DST,  /* colour=0  DOT_ON  */
         PIX_NOT( PIX_SRC ) & PIX_DST,  /* colour=0  DOT_OFF */
         PIX_CLR,                       /* colour=0  DOT_REV */
         PIX_CLR },                     /* colour=0  DOT_NEW */
       { PIX_SRC | PIX_DST,             /* colour>0  DOT_ON  */
         PIX_NOT( PIX_SRC ) & PIX_DST,  /* colour>0  DOT_OFF */
         PIX_NOT( PIX_SRC ),            /* colour>0  DOT_REV */
         PIX_SRC } };                   /* colour>0  DOT_NEW */
    if ( ' ' <= character && character <= '~' )
      {
/*
 ********              colour screen
 *              Colour screens are difficult . A different technique for each
 *              ``TER.dots'' .
*/
        if ( SETUP.colour_screen )
          {
            if ( TER.dots == DOT_ON )
              {
/*
 *              Paint the required colour through the character shape stencil .
*/
                (void) pw_stencil( SETUP.t_pixwin,
                    (int) ( TER.cursor._x + 0.5 ),
                    (int) ( SETUP.sun_max._y - TER.cursor._y + 0.5 +
                            SETUP.alph_home ),
                    ALPH_WIDTH, ALPH_HEIGHT,
                    PIX_SRC | SUN_COLOUR( TER.colour ),
                    (SETUP.t_sysfont->pf_char)[character].pc_pr, 0, 0,
                    (Pixrect *)0, 0, 0 );
              }
            else if ( TER.dots == DOT_NEW  )
              {
/*
 *              Paints the character in the foreground colour and also paints
 *              the background .
*/
                (void) pw_char( SETUP.t_pixwin,
                    (int) ( TER.cursor._x + 0.5 ),
                    (int) ( SETUP.sun_max._y - TER.cursor._y + 0.5 ),
                    PIX_SRC | SUN_COLOUR( TER.colour ),
                    SETUP.t_sysfont,
                    TER.colour ? character : ' ' );
              }
            else
              {
                if ( TER.dots == DOT_REV )
/*
 *              For reversed text, first paint a box in the foreground colour .
*/
                    (void) pw_rop( SETUP.t_pixwin,
                        (int) ( TER.cursor._x + 0.5 ),
                        (int) ( SETUP.sun_max._y - TER.cursor._y + 0.5 +
                                SETUP.alph_home ),
                        ALPH_WIDTH, ALPH_HEIGHT,
                        PIX_SRC | SUN_COLOUR( TER.colour ),
                        (Pixrect *)0, 0, 0 );
/*
 *              For reversed and ``DOTS_OFF'' text write with the background
 *              colour .
*/
                (void) pw_char( SETUP.t_pixwin,
                    (int) ( TER.cursor._x + 0.5 ),
                    (int) ( SETUP.sun_max._y - TER.cursor._y + 0.5 ),
                    PIX_DST & PIX_NOT( PIX_SRC ),
                    SETUP.t_sysfont, character );
              }
          }
        else
/*
 ********              black and white screen
*/
          {
                (void) pw_char( SETUP.t_pixwin,
                    (int) ( TER.cursor._x + 0.5 ),
                    (int) ( SETUP.sun_max._y - TER.cursor._y + 0.5 ),
                    bw_char_op[ TER.colour > 0 ][ (int) TER.dots ],
                SETUP.t_sysfont, character );
          }
      }
/*
 *          Paint or remove the cursor . The action is the same in both cases,
 *          but they are named differently for legibility of the code .
*/
    else if ( character == PUT_CURSOR || character == DEL_CURSOR )
        (void) pw_char( SETUP.t_pixwin,
            (int) ( TER.cursor._x + 0.5 ),
            (int) ( SETUP.sun_max._y - TER.cursor._y + 0.5 ),
            PIX_NOT( PIX_DST ), SETUP.t_sysfont, ' ' );
  }


static boolean next_cursor()
/*
 *      Advance the position of the text cursor.
 * This includes moving to the beginning of the next line if the right boundary
 * of the viewport is encountered .
*/
  {
    if ( ( TER.cursor._x += ALPH_WIDTH ) > max_cursor._x )
      {
        TER.cursor._x = min_clip( VIEW_CLIP, _x );
        TER.cursor._y -= ALPH_HEIGHT;
      }
    return IN_VIEW;
  }


static boolean prev_cursor()
/*
 *      Place the text cursor at its previous position.
 * This includes moving to the end of the previous line if the left boundary of
 * the viewport is encountered .
*/
  {
    if ( ( TER.cursor._x -= ALPH_WIDTH ) < min_clip( VIEW_CLIP, _x ) )
      {
        TER.cursor._x =
            ( (int) ( ( max_clip( VIEW_CLIP, _x ) - min_clip( VIEW_CLIP, _x ) )
            / ALPH_WIDTH ) - 1 ) * ALPH_WIDTH + min_clip( VIEW_CLIP, _x );
        TER.cursor._y += ALPH_HEIGHT;
      }
    return IN_VIEW;
  }
#undef IN_VIEW
#undef PUT_CURSOR
#undef DEL_CURSOR


static int t_txtput( string )
/*
 *    Put a text string `string' on the GRAPHICS screen, starting at the place
 *  last specified by `t_txtmov' or 't_txtcur', using the colour specified by
 *  `t_colour' as foreground and the data level specified by `t_visib',
 *  `t_invis', `t_revers' or `t_replac' . Only those characters within the
 *  viewport are written to the screen . The returned value is the number of
 *  characters written to the screen, or zero if an error has occurred .
 *
 *  The character string `string' may contain printable characters and the
 *  control codes,
 *    HT : \t : \011 : move to the right by one character space
 *    BS : \b : \010 : move to the left by one character space
 *    CR : \r : \015 : move to the beginning of the current line
 *    LF : \n : \012 : move down one character space
 *    VT :    : \013 : move up one character space
 *    EM :    : \031 : move to top left hand side of the screen
 *  Any other characters are ignored .
 *
 *    DOT_ON,  write foreground colour text onto the existing screen,
 *    DOT_OFF, write a dark background onto the existing screen,
 *    DOT_REV, change the background into foreground colour, write dark text,
 *    DOT_NEW, foreground colour text is written onto a dark background.
 *
 *  This function only breaks the string into segments terminated by cursor
 *  control characters and postions the required_cursor appropriately before
 *  calling t_txtstr to put the characters on the screen.
*/
char VAR_PTR string;            /* pointer to start of the character string */
  {
    int char_count = 0;     /* number of characters printed             */
    char VAR_PTR name = "t_txtput";
    START_UP( name, int );
    if ( TER.mode != GRAPH_QUIT )
      {
        char    *ptr,
                ctrl_char;     /* temporary storage for cursor control char*/
        CATCH_NULL( string, name, int );
        remove_crosshairs();
        max_cursor._x = max_clip(VIEW_CLIP,_x) - ALPH_WIDTH + 1;
        max_cursor._y = max_clip(VIEW_CLIP,_y) + SETUP.alph_home;
        for ( ptr = string; *ptr; ptr++ )
          {
            if ( *ptr == LF || *ptr == VT || *ptr == EM ||
                 *ptr == CR || *ptr == BS || *ptr == HT )
              {
                ctrl_char = *ptr;
                *ptr = NUL;
                char_count += t_txtstr( string ) + 1;
                switch( *ptr = ctrl_char )
                  {
                    case LF : TER.cursor._y += -ALPH_HEIGHT;
                              break;
                    case VT : TER.cursor._y += ALPH_HEIGHT;
                              break;
                    case EM : TER.cursor._x = min_clip( VIEW_CLIP, _x );
                              TER.cursor._y = max_cursor._y;
                              break;
                    case CR : TER.cursor._x = min_clip( VIEW_CLIP, _x );
                              break;
                    case BS : TER.cursor._x += -ALPH_WIDTH;
                              break;
                    case HT : TER.cursor._x += ALPH_WIDTH;
                              break;
                  }
                string = ptr + 1;
              }
          }
        if ( string != ptr )
            char_count += t_txtstr( string );
        flush_frame_events();
      }
    else
        (void) make_error( ERR_ABORT, name, err_write );
    return char_count;
  }


static int t_txtstr( string )
/*
 * Puts a string of printable ASCII characters onto the screen, starting at the
 * current required cursor postion and using the current colour and data level.
 * Only characters within the viewport are written to the screen . The returned
 * value is the number of characters written.
 *
*/
char VAR_PTR string;
  {
    int char_count = 0;             /* number of characters sent             */
    pw_batch_on( SETUP.t_pixwin );
    for( ; *string; string++ )
      {
        if ( ' ' <= *string && *string <= '~' )
          {
/*
 *        put character on the screen if it is in the viewport
*/
            if ( min_clip( VIEW_CLIP, _x ) <= TER.cursor._x &&
                 max_cursor._x >= TER.cursor._x &&
                 min_clip( VIEW_CLIP, _y ) <= TER.cursor._y &&
                 max_cursor._y >= TER.cursor._y )
              {
                put_char( *string );
                char_count++;
              }
            TER.cursor._x += ALPH_WIDTH;
          }
      }
    pw_batch_off( SETUP.t_pixwin );
    return char_count;
  }


/* ************************************************************************ */
/*                                                                          */
/*                              Drawing boxes                               */
/*                                                                          */
/* ************************************************************************ */


static boolean t_boxfil( lower_left, upper_right, style )
/*
 *      draw a rectangular box on the screen with lower left corner at
 * `*lower_left' and upper right corner at `*upper_right', and fill the box
 * with a hatching of type `style' . The hatching styles are :
 *
 *      0 = empty, box outline only      5 = horizontal lines
 *      1 = fill the box completely      6 = vertical lines
 *      2 = 50% `grey' fill              7 = diagonal cross hatch
 *      3 = NW-SE diagonal lines         8 = horizontal-vertical cross
 *      4 = NE-SW diagonal lines                 hatch
 * The box outline is clipped to the viewport limits .
*/
coord_pair VAR_PTR lower_left,  /* lower left corner of the box     */
           VAR_PTR upper_right; /* upper right corner of the box    */
int style;                      /* hatching style                   */
  {
    char VAR_PTR name = "t_boxfil";
    START_UP( name, boolean );
    if ( TER.mode != GRAPH_QUIT )
      {
        coord_pair ll_corner, ur_corner;
        CATCH_NULL( lower_left && upper_right, name, boolean );
        ll_corner = *lower_left;
        ur_corner = *upper_right;
/*
 *      ensure correct ordering of coordinates
*/
        if ( ll_corner._x > ur_corner._x )
            SWAPD( ll_corner._x , ur_corner._x );
        if ( ll_corner._y > ur_corner._y )
            SWAPD( ll_corner._y , ur_corner._y );
/*
 *      clip the rectangle to the screen limits
*/
        ll_corner._x = LIMIT( ll_corner._x,
            min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
        ll_corner._y = LIMIT( ll_corner._y,
            min_clip( VIEW_CLIP, _y ), max_clip( VIEW_CLIP, _y ) );
        ur_corner._x = LIMIT( ur_corner._x,
            min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
        ur_corner._y = LIMIT( ur_corner._y,
            min_clip( VIEW_CLIP, _y ), max_clip( VIEW_CLIP, _y ) );
/*
 *      draw the box
*/
        if ( ur_corner._x - ll_corner._x > 0.0 &&
             ur_corner._y - ll_corner._y > 0.0 )
          {
            remove_crosshairs();
            sun_box( (int)( ll_corner._x + 0.5 ),
                     (int)( SETUP.sun_max._y - ur_corner._y + 0.5 ),
                     (int)( ur_corner._x + 0.5 ),
                     (int)( SETUP.sun_max._y - ll_corner._y + 0.5 ),
                     ( abs( style ) - 1 ) % 8 + 1 );
            flush_frame_events();
            return TRUE;
          }
        return FALSE;
      }
    return make_error( ERR_ABORT, name, err_write );
  }


#define BOX_PR_HEIGHT   8
#define BOX_PR_WIDTH    16
#define COLOUR_PR_DEPTH 8

static void sun_box( x_left, y_top, x_right, y_bottom, style )
/*
 *      draw a box on the canvas - with appropriate fill pattern,
 *  N.B. There is no checking of the arguments.
*/
int x_left, y_top,              /* top left pixwin coordinate           */
    x_right, y_bottom,          /* bottom right pixwin coordinate       */
    style;                      /* style of fill pattern 0 to 8         */
  {
/*
 *              first calculate the size of the box
*/
    boolean using_colour;
    int rop;
    int width = x_right - x_left - 1;
    int height = y_bottom - y_top - 1;
/*
 *              Arrays of Pixrect data defining the hatching patterns .
*/
    static short int
      fill_prdata[] = { 0xffff, 0xffff, 0xffff, 0xffff,
                        0xffff, 0xffff, 0xffff, 0xffff },
      grey_prdata[] = { 0xaaaa, 0x5555, 0xaaaa, 0x5555,
                        0xaaaa, 0x5555, 0xaaaa, 0x5555 },
      NWSE_prdata[] = { 0x8080, 0x4040, 0x2020, 0x1010,
                        0x0808, 0x0404, 0x0202, 0x0101 },
      NESW_prdata[] = { 0x0101, 0x0202, 0x0404, 0x0808,
                        0x1010, 0x2020, 0x4040, 0x8080 },
      vert_prdata[] = { 0x1111, 0x1111, 0x1111, 0x1111,
                        0x1111, 0x1111, 0x1111, 0x1111 },
      horz_prdata[] = { 0x0000, 0x0000, 0x0000, 0xffff,
                        0x0000, 0x0000, 0x0000, 0xffff },
      cros_prdata[] = { 0x8282, 0x4444, 0x2828, 0x1010,
                        0x2828, 0x4444, 0x8282, 0x0101 },
      mesh_prdata[] = { 0x1111, 0x1111, 0x1111, 0xffff,
                        0x1111, 0x1111, 0x1111, 0xffff };
/*
 *              Define the Pixrects containing the hatching patterns and
 *              the array of pointers to these Pixrects .
*/
  mpr_static_static( fill_prect, BOX_PR_WIDTH, BOX_PR_HEIGHT, 1, fill_prdata );
  mpr_static_static( grey_prect, BOX_PR_WIDTH, BOX_PR_HEIGHT, 1, grey_prdata );
  mpr_static_static( NWSE_prect, BOX_PR_WIDTH, BOX_PR_HEIGHT, 1, NWSE_prdata );
  mpr_static_static( NESW_prect, BOX_PR_WIDTH, BOX_PR_HEIGHT, 1, NESW_prdata );
  mpr_static_static( vert_prect, BOX_PR_WIDTH, BOX_PR_HEIGHT, 1, vert_prdata );
  mpr_static_static( horz_prect, BOX_PR_WIDTH, BOX_PR_HEIGHT, 1, horz_prdata );
  mpr_static_static( cros_prect, BOX_PR_WIDTH, BOX_PR_HEIGHT, 1, cros_prdata );
  mpr_static_static( mesh_prect, BOX_PR_WIDTH, BOX_PR_HEIGHT, 1, mesh_prdata );
    static Pixrect *(box_pixrect_list[]) =
        { &fill_prect, &grey_prect, &NWSE_prect, &NESW_prect,
          &vert_prect, &horz_prect, &cros_prect, &mesh_prect };
/*
 *              This is intermediate Pixrect for DOT_NEW filling colour boxes .
*/
    static short int colour_data[8];
    mpr_static_static( colour_pattern, BOX_PR_WIDTH, BOX_PR_HEIGHT,
                COLOUR_PR_DEPTH, colour_data );
/*
 *              Define the raster operation
*/
    if ( ! TER.colour && TER.dots == DOT_ON )
        rop = SUN_ROP( DOT_OFF );
    else if ( ! TER.colour && TER.dots == DOT_NEW )
      {
        rop = SUN_ROP( DOT_OFF );
        style = 1;
      }
    else
        rop = SUN_ROP(TER.dots );
    if ( ( using_colour = SETUP.colour_screen && TER.colour > 1 ) &&
            ( TER.dots == DOT_ON || TER.dots == DOT_NEW ) )
        rop = SUN_ROP( DOT_NEW ) | SUN_COLOUR( TER.colour );
/*
 *              Draw the box outline .
*/
    pw_batch_on( SETUP.t_pixwin );
    if ( x_left <= x_right )
        (void) pw_vector( SETUP.t_pixwin, x_left, y_bottom, x_right, y_bottom,
            rop, 15 );
    if ( height >= 0 )
        (void) pw_vector( SETUP.t_pixwin, x_left, y_top, x_right, y_top,
            rop, 15 );
    if ( height > 0 )
      {
        (void) pw_vector( SETUP.t_pixwin, x_left, --y_bottom, x_left, ++y_top,
            rop, 15 );
        (void) pw_vector( SETUP.t_pixwin, x_right, y_bottom, x_right, y_top,
            rop, 15 );
      }
/*
 *              Fill the box with the required pattern . A different technique
 *              is used for colour and monochrome screens . ``fill_Pr_ptr'' is
 *              a pointer to the required fill pattern Pixrect .
*/
    if ( 0 < style && style <= 8 && width > 0 && height > 0 )
      {
        Pixrect * fill_Pr_ptr = box_pixrect_list[style - 1];
/*
 *              DOT_ON on a colour screen is a special case in which the fill
 *              pattern Pixrect is used as a stencil through which the required
 *              foreground colour is painted onto the screen - in sets of 16x8
 *              8 pixels . Note that this is very slow .
*/
        if ( using_colour && TER.dots == DOT_ON )
          {
            int cell_x, cell_y, cell_height;
            for( cell_y = y_top; cell_y <= y_bottom;
                    cell_y += BOX_PR_HEIGHT )
              {
                cell_height = MIN( BOX_PR_HEIGHT, y_bottom - cell_y + 1 );
                for( cell_x = x_left + 1; cell_x <= x_right - 1;
                        cell_x += BOX_PR_WIDTH )
                    (void) pw_stencil( SETUP.t_pixwin, cell_x, cell_y,
                        MIN( BOX_PR_WIDTH, x_right - cell_x ), cell_height,
                        rop, fill_Pr_ptr, 0, 0, (Pixrect *) 0, 0, 0 );
              }
          }
/*
 *              For DOT_NEW, first make a coloured copy of the fill pattern
 *              Pixrect; then replicate it onto the screen .
*/
        else if ( using_colour && TER.dots == DOT_NEW )
          {
            (void) pr_rop( &colour_pattern, 0, 0, BOX_PR_WIDTH, BOX_PR_HEIGHT,
                rop, fill_Pr_ptr, 0, 0 );
            (void) pw_replrop( SETUP.t_pixwin, ++x_left, y_top, width, height,
                rop & ~ PIX_COLOR( -1 ), &colour_pattern, 0, 0 );
          }
/*
 *              For monochrome screen and DOT_REV, DOT_OFF colour screen .
*/
        else
            (void) pw_replrop( SETUP.t_pixwin, ++x_left, y_top, width, height,
                rop, box_pixrect_list[ --style ], 0, 0 );
      }
    pw_batch_off( SETUP.t_pixwin );
  }

#undef BOX_PR_HEIGHT
#undef BOX_PR_WIDTH
#undef COLOUR_PR_DEPTH


/* ************************************************************************ */
/*                                                                          */
/*                           Cross Hairs Routines                           */
/*                                                                          */
/* ************************************************************************ */


static boolean t_xhrget( point, char_pressed )
/*
 *      Get cross-hair location . The cross-hairs are positioned by the user,
 *  and the location of the cross-hairs is transmitted when an ASCII character
 *  key or a mouse button is pressed . The returned value of `t_xhrget' is
 *  TRUE if `point' is within the current viewport ; otherwise the returned
 *  value is FALSE . If necessary, t_xhrget opens the graphics screen .
 *
 *    Arguments
*/
coord_pair VAR_PTR point;   /* pointer to the coordinates of the cross-hairs */
char VAR_PTR char_pressed;  /* character pressed to identify `point'         */
  {
    Cursor gfx_cursor;
    char VAR_PTR name = "t_xhrget";
    START_UP( name, boolean );
    if ( TER.mode != GRAPH_QUIT )
      {
        CATCH_NULL( point, name, boolean );
/*
 *              Ensure that the graphics frame is open .
*/
        (void) open_frame( name );
/*
 *              Draw crosshairs -- if not already showing them ( use SunView to
 *              find out .
*/
        gfx_cursor = window_get( SETUP.t_canvas, WIN_CURSOR );
        if ( ! cursor_get( gfx_cursor, CURSOR_SHOW_CROSSHAIRS ) )
          {
            (void) cursor_set( gfx_cursor, CURSOR_SHOW_CROSSHAIRS, TRUE,
                                     CURSOR_SHOW_CURSOR,     FALSE, 0 );
            (void) window_set( SETUP.t_canvas, WIN_CURSOR, gfx_cursor, 0 );
          }
/*
 *              Get crosshairs position .
 *              Enable keyboard events, then call the notifier until the
 *              ``crosshair_get'' request has been satisfied by the canvas
 *              event procedure .
*/
        (void) window_set( SETUP.t_canvas, WIN_CONSUME_KBD_EVENTS,
            WIN_ASCII_EVENTS, WIN_RIGHT_KEYS, WIN_LEFT_KEYS, WIN_TOP_KEYS, 0,
            0 );
        for( GFX_WIN.waiting_func= T_XHRGET; GFX_WIN.waiting_func == T_XHRGET; )
            (void) notify_dispatch();
/*
 *              get the returned values from the global graphic window status
*/
        point->_x = GFX_WIN.xhair_x;
        point->_y = SETUP.sun_max._y - GFX_WIN.xhair_y;
        *char_pressed = (char) GFX_WIN.t_return_value;
        remove_crosshairs();
        return min_clip( VIEW_CLIP, _x ) <= point->_x &&
               max_clip( VIEW_CLIP, _x ) >= point->_x &&
               min_clip( VIEW_CLIP, _y ) <= point->_y &&
               max_clip( VIEW_CLIP, _y ) >= point->_y;
      }
    else
        return make_error( ERR_ABORT, name, err_write );
  }


static boolean t_xhrput( point )
/*
 *      Put the  cross-hairs at `point' . If the terminal is not already in
 *  cross-hairs mode, then it is put into cross-hairs mode . The returned value
 *  of `t_xhrput' is TRUE if `point' is within the current viewport and
 *  `t_xhrput' can write to the graphics device ; otherwise the returned value
 *  is FALSE . If necessary, t_xhrput opens the graphics screen .
 *
 *    Arguments
*/
coord_pair VAR_PTR point;   /* pointer to the coordinates of the cross-hairs */
  {
    Event put_x;
    Cursor gfx_cursor;
    char VAR_PTR name = "t_xhrput";
    START_UP( name, boolean );
    if ( TER.mode != GRAPH_QUIT )
      {
        CATCH_NULL( point, name, boolean );
        if ( -0.5 < point->_x  &&  point->_x <= SETUP.sun_max._x + 0.5 &&
             -0.5 < point->_y  &&  point->_y <= SETUP.sun_max._y + 0.5 )
          {
/*
 *              ensure that the graphics frame is open
*/
            (void) open_frame( name );
/*
 *              Draw crosshairs if necessary .
*/
            gfx_cursor = window_get( SETUP.t_canvas, WIN_CURSOR );
            if ( ! cursor_get( gfx_cursor, CURSOR_SHOW_CROSSHAIRS ) )
              {
                (void) cursor_set( gfx_cursor, CURSOR_SHOW_CROSSHAIRS, TRUE,
                    CURSOR_SHOW_CURSOR, FALSE, 0 );
                (void) window_set( SETUP.t_canvas, WIN_CURSOR,
                    gfx_cursor, 0 );
              }
/*
 *            Translate coordinate of the shifted event from the logical
 *            (scrollable) canvas to the canvas subwindow space and the
 *            place the crosshairs in the required position .
*/
            event_set_x( &put_x, (int) ( point->_x + 0.5 ) );
            event_set_y( &put_x, (int) ( SETUP.sun_max._y - point->_y + 0.5 ) );
            (void) canvas_window_event( SETUP.t_canvas, &put_x );
            (void) window_set( SETUP.t_canvas,
                WIN_MOUSE_XY, event_x( &put_x ), event_y( &put_x ), 0 );
            flush_frame_events();
            return min_clip( VIEW_CLIP, _x ) <= point->_x &&
                   max_clip( VIEW_CLIP, _x ) >= point->_x &&
                   min_clip( VIEW_CLIP, _y ) <= point->_y &&
                   max_clip( VIEW_CLIP, _y ) >= point->_y;
          }
        else
            return FALSE;
      }
    return make_error( ERR_ABORT, name, err_write );
  }


static void remove_crosshairs()
/*
 *      remove crosshairs
*/
  {
    Cursor gfx_cursor = window_get( SETUP.t_canvas, WIN_CURSOR );
    if ( cursor_get( gfx_cursor, CURSOR_SHOW_VERT_HAIR ) )
      {
        (void) cursor_set( gfx_cursor, CURSOR_SHOW_CROSSHAIRS, FALSE,
                                   CURSOR_SHOW_CURSOR, TRUE, 0 );
        (void) window_set( SETUP.t_canvas, WIN_CURSOR, gfx_cursor, 0 );
      }
  }


/* ************************************************************************ */
/*                                                                          */
/*                        Print terminal status                             */
/*                                                                          */
/* ************************************************************************ */


static char VAR_PTR t_status( attribute, value )
/*
 *  Put status attribute 'attribute' into '*value' which may be a character
 *  string or a coordinate pair .
*/
int attribute;
char VAR_PTR value;
  {
    char VAR_PTR name = "t_status";
    CATCH_NULL( value, name, char VAR_PTR );
    if ( attribute != T_ST_SCREEN && attribute != T_ST_CATEG )
        START_UP( name, char VAR_PTR );
    switch( attribute )
      {
        case T_ST_DEVICE:
            if ( SETUP.colour_screen )
                (void) t_strcpy( value,
                    (char VAR_PTR) "colour sun workstation" );
            else
                (void) t_strcpy( value,
                    (char VAR_PTR) "monochrome sun workstation");
            break;
        case T_ST_MONO:
            *( (boolean VAR_PTR) value ) = ! SETUP.colour_screen;
            break;
        case T_ST_SCREEN :
            *( (boolean VAR_PTR) value ) = TRUE;
            break;
        case T_ST_CATEG :
            *( (int VAR_PTR) value ) = T_ST_WIND;
            break;
        case T_ST_CLIP:
            *( (clip_area VAR_PTR VAR_PTR) value ) = VIEW_CLIP;
            break;
        case T_ST_MAX:
            ( (coord_pair VAR_PTR) value)->_x = SETUP.sun_max._x;
            ( (coord_pair VAR_PTR) value)->_y = SETUP.sun_max._y;
            break;
        case T_ST_MIN:
            ( (coord_pair VAR_PTR) value)->_x = 0.0;
            ( (coord_pair VAR_PTR) value)->_y = 0.0;
            break;
        case T_ST_CHAR:
            ( (coord_pair VAR_PTR) value)->_x = ALPH_WIDTH;
            ( (coord_pair VAR_PTR) value)->_y = ALPH_HEIGHT;
            break;
        case T_ST_LINES:
            ( (coord_pair VAR_PTR) value)->_x = SETUP.alph_columns;
            ( (coord_pair VAR_PTR) value)->_y = SETUP.alph_lines;
            break;
        case T_ST_PIXSIZ:
            ( (coord_pair VAR_PTR) value)->_x = PIX_SIZE;
            ( (coord_pair VAR_PTR) value)->_y = PIX_SIZE;
            break;
        case T_ST_A_CURS: /* fall through */
        case T_ST_R_CURS:
            ( (coord_pair VAR_PTR) value)->_x = TER.cursor._x;
            ( (coord_pair VAR_PTR) value)->_y = TER.cursor._y;
            break;
        case T_ST_A_PEN:  /* fall through */
        case T_ST_R_PEN:
            ( (coord_pair VAR_PTR) value)->_x = TER.pen._x;
            ( (coord_pair VAR_PTR) value)->_y = TER.pen._y;
            break;
        default:
            return (char VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "unknown attribute" );
      }
    return value;
  }


static boolean t_pr_ter( out_file )
/*
 *  Prints the terminal status on the file ``out_file'' .
*/
FILE VAR_PTR out_file;
  {
    char VAR_PTR name = "t_pr_ter";
    CATCH_NULL( out_file, name, boolean );
    START_UP( name, boolean );
    (void) fprintf( out_file, " TERMINAL STATUS\n" );
    (void) fprintf( out_file, "data level  %15s\n", pr_dots( &TER ) );
    (void) fprintf( out_file, "mode        %15s\n", pr_mode( &TER ) );
    (void) fprintf( out_file, "colour      %15d\n", TER.colour );
    (void) fprintf( out_file, "dash style  %15d\n", TER.dash );
    (void) fprintf( out_file, "cursor      (%6.1f,%6.1f)\n",
               TER.cursor._x, TER.cursor._y );
    (void) fprintf( out_file, "pen         (%6.1f,%6.1f)\n",
               TER.pen._x, TER.pen._y );
    if ( ferror( out_file ) )
        return make_error( ERR_WARN, name,
            (char VAR_PTR) "Cannot write terminal status to file" );
    return TRUE;
  }


static char VAR_PTR pr_dots( status_ptr )
ter_status VAR_PTR status_ptr;
  {
    switch( status_ptr->dots )
      {
        case DOT_ON  : return "dots on";
        case DOT_OFF : return "dots off";
        case DOT_REV : return "complement dots";
        case DOT_NEW : return "replace dots";
      }
    return mt_str;
  }


static char VAR_PTR pr_mode( status_ptr )
ter_status VAR_PTR status_ptr;
  {
    switch( status_ptr->mode )
      {
        case GRAPH_QUIT   : return "quit";
        case GRAPH_CLOSED : return "closed";
        case GRAPH_OPEN   : return "open";
      }
    return mt_str;
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Vector Drawing                                 */
/*                                                                          */
/* ************************************************************************ */

static boolean t_vecmov( point )
/*
 *      move the line drawing pen to `point'
*/
coord_pair VAR_PTR point;
  {
    char VAR_PTR name = "t_vecmov";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    TER.pen = *point;
    flush_frame_events();
    return move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
  }


static boolean t_vecdrw( point )
/*
 *      Draw from the previous pen position to `point' according to the
 *  required pen colour, dashed line style and dots-level . Dashed line style
 *  takes precedence over colours . Lines are clipped to the viewport
 *  boundaries .
*/
coord_pair VAR_PTR point;
  {
    char VAR_PTR name = "t_vecdrw";
    START_UP( name, boolean );
    if ( TER.mode != GRAPH_QUIT )
      {
        enum dev_dots vec_dots;
        int vec_style;
        CATCH_NULL( point, name, boolean );
        TER.pen = *point;
        remove_crosshairs();
/*
 *          If required dashed line style is 0, plot the point if it is
 *          in the viewport .
*/
        if ( ! TER.dash )
          {
            (void) move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
            if ( min_clip( VIEW_CLIP, _x ) <= point->_x &&
                 max_clip( VIEW_CLIP, _x ) >= point->_x &&
                 min_clip( VIEW_CLIP, _y ) <= point->_y &&
                 max_clip( VIEW_CLIP, _y ) >= point->_y )
              {
/*
 *   If required colour is black, put terminal data level into DOT_OFF,
 *   otherwise use the required data level . Then plot the point .
*/
                if ( ! TER.colour )
                    vec_dots = DOT_OFF;
                else
                    vec_dots = TER.dots;
                (void) pw_vector( SETUP.t_pixwin,
                    (int) ( point->_x + 0.5 ), (int) ( point->_y + 0.5 ),
                    (int) ( point->_x + 0.5 ), (int) ( point->_y + 0.5 ),
                    SUN_ROP( vec_dots ) | SUN_COLOUR( TER.colour ),
                    15 );
              }
          }
/*
 *          Need to draw a line . First clip the endpoints .
*/
        else if ( clip( (coord_triple VAR_PTR) point, VIEW_CLIP ) )
          {
/*
 *          If required colour is black, put terminal data level into DOT_OFF,
 *          otherwise use the required data level .
*/
            if ( ! TER.colour )
                vec_dots = DOT_OFF;
            else if ( SETUP.colour_screen && TER.colour > 1 &&
                      TER.dots == DOT_ON )
                vec_dots = DOT_NEW;
            else
                vec_dots = TER.dots;
/*
 *          If a continuous line is required on a black & white screen,
 *          use the dashed line style as pseudo-colour, otherwise just
 *          set the required dashed line style
*/
            if ( ! SETUP.colour_screen  && TER.dash == SOLID_LINE
                   && TER.colour )
                vec_style = TER.colour;
            else if ( TER.dash )
                vec_style = TER.dash;
/*
 *          Draw the vector -- solid line .
*/
            if ( vec_style == 1 )
                (void) pw_vector( SETUP.t_pixwin,
                    (int) ( first_clip( VIEW_CLIP, _x ) + 0.5 ),
                    (int) ( SETUP.sun_max._y - first_clip( VIEW_CLIP, _y ) +
                            0.5 ),
                    (int) ( last_clip( VIEW_CLIP, _x ) + 0.5 ),
                    (int) ( SETUP.sun_max._y - last_clip( VIEW_CLIP, _y ) +
                            0.5 ),
                    SUN_ROP( vec_dots ) |
                        ( TER.dots==DOT_OFF || TER.dots==DOT_REV ?
                            0 : SUN_COLOUR( TER.colour ) ),
                    15 );
            else
/*
 *          Draw the vector -- dashed line .
*/
                sun_line( (int) ( first_clip( VIEW_CLIP, _x ) + 0.5 ),
                    (int) ( SETUP.sun_max._y - first_clip( VIEW_CLIP, _y ) +
                            0.5 ),
                    (int) ( last_clip( VIEW_CLIP, _x ) + 0.5 ),
                    (int) ( SETUP.sun_max._y - last_clip( VIEW_CLIP, _y ) +
                            0.5 ),
                    SUN_ROP( vec_dots ) |
                        ( TER.dots==DOT_OFF || TER.dots==DOT_REV ?
                            0 : SUN_COLOUR( TER.colour ) ),
                    vec_style );
          }
        flush_frame_events();
        return TRUE;
      }
    return make_error( ERR_WARN, name, err_write );
  }


/*
 *          The dashed line ``Texture'' status records the pattern description,
 *          where in the sequence ``sun_line()'' should start drawing again, and
 *          whether the pen is up or down . This data is remembered between
 *          calls to ``sun_line()'' .
*/
typedef struct          /* information required for drawing dashed lines */
  {
    short int VAR_PTR pattern; /* pointer to array defining line style   */
    short int offset;   /* length of segment still to be drawn           */
    short int index;    /* index to pattern array                        */
    short int pen_down; /* TRUE if pen is down, FALSE if pen is raised   */
    short int cur_style;/* current line style                            */
  }
    Texture;


static void sun_line( x1, y1, x2, y2, op, style )
/*
 * Draw a dashed line from (x1,y1) to (x2,y2) with the pattern specified by
 * 'style' . This is and algorithm concocted to avoid the complications of
 * calling ``pw_line()'' (I couldn't make it work anyway) .
 *
 * The line is drawn in a sequence of segments . Each segment is drawn from
 * (a_x,a_y) to (b_x,b_y) using either x or y as the independent variable .
 * The variable selected as independent is the one with the largest step
 * i.e. ( y2-y1 or x2-x1 ) . The algorithm used to calculate the segment
 * endpoint does not accumulate round-off error and uses integer arithmetic
 * only, and always converges to the endpoint (x2,y2) .
 *
 * (x1,y1) = starting point of the dashed line
 *         = (*indep_1,*depend_1) for lines < 45 degrees from horizontal
 *         = (*depend_1,*indep_1) for lines < 45 degrees from horizontal
 * (x2,y2) = starting point of the dashed line
 *         = (*indep_2,*depend_2) for lines > 45 degrees from horizontal
 *         = (*depend_2,*indep_2) for lines > 45 degrees from horizontal
 * (a_x,a_y) = starting point of the dashed line
 *           = (*indep_a,*depend_a) for lines < 45 degrees from horizontal
 *           = (*depend_a,*indep_a) for lines < 45 degrees from horizontal
 * (b_x,b_y) = starting point of the dashed line
 *           = (*indep_b,*depend_b) for lines > 45 degrees from horizontal
 *           = (*depend_b,*indep_b) for lines > 45 degrees from horizontal
 *
*/
int x1, y1,     /* initial point in SunView's screen coordinates */
    x2, y2,     /* final point in SunView's screen coordinates   */
    op,         /* raster operation used                         */
    style;      /* dashed line style -- coded as for ``t_dashed()'' */
  {
    int VAR_PTR indep_1, VAR_PTR depend_1,  /* starting point of dashed line  */
        VAR_PTR indep_2,                /* finishing point of the dashed line */
        a_x, a_y,                           /* starting point of the segment  */
        VAR_PTR indep_a,                    /* starting point of the segment  */
        b_x, b_y,                           /* finishing point of the segment */
        VAR_PTR indep_b, VAR_PTR depend_b,  /* finishing point of the segment */
        indep_step, depend_step,            /* step size in x & y directions  */
        temp_swap;                          /* temporary value for swapping   */
    boolean add;                            /* TRUE=add, FALSE=subtract       */
/*
 *          Dashed line pattern descriptions (zero termination)
 *          ( pen_down, pen_up, pen_down, .... 0 )
 *          N.B. The first pattern (solid line) should never be used .
*/
    static short int
      solid_line[]  = { 0 },               /*--------------------------------*/
      dotted_line[] = { 3, 5, 0 },         /*---     ---     ---     ---     */
      short_dash[]  = { 11, 5, 0 },        /*-----------     -----------     */
      long_dash[]   = { 27, 5, 0 },        /*---------------------------     */
      dot_dash[]    = { 2, 5, 20, 5, 0 },  /*--     --------------------     */
      short_long[]  = { 6, 5, 16, 5, 0 },  /*------     ----------------     */
      short2_dash[] = { 3,5,3,5,11,5,0 },  /*---     ---     -----------     */
      dddot_dash[]  = { 2,5,2,5,2,5,8,5,0},/*--    --    --    ----------    */
/*
 *          Array of pointers to dashed line pattern descriptions .
*/
      VAR_PTR (dash_list[])= { solid_line, dotted_line, short_dash,  long_dash,
                               dot_dash, short_long,  short2_dash, dddot_dash };
    static Texture texture = { (short int VAR_PTR) 0, 0, 0, TRUE, NOT_SET };
/*
 *          Only change the remembered dash style parameters if the line type
 *          has been changed .
*/
    if ( style != texture.cur_style )
      {
        texture.pattern  = dash_list[ style - 1 ];
        texture.index    = 0;
        texture.pen_down = TRUE;
        texture.offset   = texture.pattern[0]; /* length of 1st segment */
        texture.cur_style = style;
      }
/*
 *          Initialize dependent and independent variables and pointers .
 *
 *          Line is inclined at less than 45 degrees from the horizontal .
*/
    if ( abs( indep_step = x2 - x1 ) > abs( depend_step = y2 - y1 ) )
      {
        indep_a  = (int VAR_PTR) &a_x;
        indep_b  = (int VAR_PTR) &b_x;
        depend_b = (int VAR_PTR) &b_y;
        indep_1  = (int VAR_PTR) &x1;
        depend_1 = (int VAR_PTR) &y1;
        indep_2  = (int VAR_PTR) &x2;
      }
    else
/*
 *          line is inclined at more than 45 degrees from the horizontal
*/
      {
        indep_a  = (int VAR_PTR) &a_y;
        indep_b  = (int VAR_PTR) &b_y;
        depend_b = (int VAR_PTR) &b_x;
        indep_1  = (int VAR_PTR) &y1;
        depend_1 = (int VAR_PTR) &x1;
        indep_2  = (int VAR_PTR) &y2;
        temp_swap = indep_step;
        indep_step = depend_step;
        depend_step = temp_swap;
      }
/*
 *          Are we adding or subtracting to get the finishing point of the
 *          independent variable ? Initialize the segment values .
*/
    add = *indep_2 >= *indep_1;
    b_x = a_x = x1;
    b_y = a_y = y1;
    texture.offset++;
/*
 *          Use batching to speed multiple operations
 *          Keep going around the loop until distance from the end of the line
 *          is less than the length of line segment remaining to be drawn .
*/
    pw_batch_on( SETUP.t_pixwin );
    while ( abs( *indep_2 - *indep_a ) > texture.offset )
      {
/*
 *          Calculate the finishing point of the segment .
*/
        if ( add )
            *indep_b = *indep_a + texture.offset;
        else
            *indep_b = *indep_a - texture.offset;
        *depend_b = ( ( *indep_b - *indep_1 ) * depend_step ) / indep_step +
                    *depend_1;
/*
 *          Only draw if the pen is down -- except for PIX_SRC raster operation
 *          where the background is painted when the pen is up . PIX_NOT(0) is
 *          used to extract the raster operation from ``op'' .
*/
        if ( texture.pen_down )
            (void) pw_vector( SETUP.t_pixwin, a_x, a_y, b_x, b_y, op, 15 );
        else if ( ( op & PIX_NOT(0) ) == PIX_SRC )
            (void) pw_vector( SETUP.t_pixwin, a_x, a_y, b_x, b_y, PIX_SRC, 0 );
/*
 *          Transfer the segment finish point to the segment start point in
 *          preparation for drawing the next segment .
*/
        a_x = b_x;
        a_y = b_y;
/*
 *          Increment to the next segment in the pattern description . If we get
 *          to the end of the pattern description, wrap around to the start .
*/
        texture.index++;
        if ( texture.pattern[ texture.index ] <= 0 )
            texture.index = 0;
/*
 *          The pen is raised or lowered after drawing each segment . At the
 *          end of one segment, all of the next segment remains to be drawn .
*/
        texture.pen_down = ! texture.pen_down;
        texture.offset = texture.pattern[ texture.index ];
      }
/*
 *          Draw the last segment to arrive at the finish point (x2,y2) .
 *          Then adjust the final offset .
*/
   if ( *indep_2 != *indep_b )
      {
        if ( texture.pen_down )
            (void) pw_vector( SETUP.t_pixwin, b_x, b_y, x2, y2, op, 15 );
        else if ( ( op & PIX_NOT(0) ) == PIX_SRC )
            (void) pw_vector( SETUP.t_pixwin, a_x, a_y, b_x, b_y, PIX_SRC, 0 );
        texture.offset -= abs( *indep_2 - *indep_b );
      }
   pw_batch_off( SETUP.t_pixwin );
 }


static boolean t_dump( file_name )
/*
 *  Dump the graphics image into a file with name ``file_name'' .
 * This function dumps the canvas backing Pixrect onto a standard ( Sun's
 * RT_STANDARD ) raster file which can then be passed through various filters
 * (if desired) . The dumped image is not the visible screen or merely part
 * of the visible image; it is the entire graphics image . It is your
 * responsibility to convert the file into a format suitable for particular
 * purpose .
*/
char VAR_PTR file_name;     /* name of dump raster file */
  {
    FILE *dump_ptr;         /* pointer to the printable dump file */
    char VAR_PTR name = "t_dump";
    boolean ok = FALSE;
    CATCH_NULL( file_name, name, boolean );
    START_UP( name, boolean );
    if ( dump_ptr = fopen( file_name, "w" ) )
      {
        colormap_t t_col_map;
        Rect t_pw_rect;
        t_pw_rect.r_left = t_pw_rect.r_top = 0;
        t_pw_rect.r_width = (int ) SETUP.sun_max._x + 1;
        t_pw_rect.r_height = ( int ) SETUP.sun_max._y + 1;
        if ( SETUP.colour_screen )
          {
            t_col_map.type   = RMT_EQUAL_RGB;
            t_col_map.length = 16;
            t_col_map.map[0] = t_red;
            t_col_map.map[1] = t_green;
            t_col_map.map[2] = t_blue;
          }
        else
          {
            t_col_map.type   = RMT_NONE;
            t_col_map.length = 0;
            t_col_map.map[2] = t_col_map.map[1] = t_col_map.map[2] =
                (unsigned char *) 0;
          }
/*
 *          Don't ``pw_lock()'' --- ``pw_unlock()'' around ``pr_dump()'' because
 *          ``pr_dump()'' uses system I/O to files which may jam and cause a
 *          block which requires the process to be ``kill''ed by a remote login
 *          ( ``rlogin'' ) or a  reboot .
 *          All the work is actually done by ``pr_dump()'' .
*/
        ok = ! pr_dump( SETUP.t_pixwin->pw_prretained, dump_ptr, &t_col_map,
                   RT_STANDARD, FALSE);
        (void) fclose( dump_ptr );
        return ok;
      }
    return make_error( ERR_ABORT, name, (char VAR_PTR) "Cannot open file" );
  }


/* ************************************************************************ */
/* ************************************************************************ */
/*                                                                          */
/*                    Array of pointers to driver functions                 */
/*                                                                          */
/* With the exception of the function ``t_driver'' which is used to select  */
/* the required driver, this array is the ONLY point of access to the       */
/* functions in the driver .                                                */
/*                                                                          */
/* ************************************************************************ */
/* ************************************************************************ */

/*
 * static clip_area VAR_PTR t_begin();
 * static boolean
 *     t_end(),    t_vwport(), t_colour(), t_dashed(), t_thick(),  t_gr_off(),
 *     t_gr_on(),  t_pr_ter(), t_dump(),   t_visib(),  t_invis(),  t_revers(),
 *     t_replac(), t_newpag(), t_newscr(), t_bell(),   t_wait(),   t_txtcur(),
 *     t_txtmov(), t_txtset(), t_symdrw(), t_xhrget(), t_xhrput(), t_vecdrw(),
 *     t_boxfil(), t_vecmov();
 * static char
 *         VAR_PTR t_txtget(), VAR_PTR t_status(), VAR_PTR t_ch_def();
 * static int t_txtput(), t_txtdrw();
 */

T_dev_fns Sun3_dev_ptrs = {
            t_begin,  t_end,    t_vwport, t_gr_on,  t_gr_off,
            t_colour, t_dashed, t_thick,  t_visib,  t_invis,
            t_revers, t_replac, t_newpag, t_newscr, t_bell,
            t_wait,   t_txtmov, t_txtcur, t_txtset, t_symdrw,
            t_txtdrw, t_txtget, t_txtput, t_status, t_pr_ter,
            t_vecmov, t_vecdrw, t_dump,   t_xhrput, t_xhrget,
            t_boxfil, t_ch_def };


/* *********************************************************************** */
/*                                                                         */
/*              Window and event handling functions                        */
/*                                                                         */
/* *********************************************************************** */


static void flush_frame_events()
/*
 *      Flush pending frame events .
 * I.e. call the notifier until there are no more pending events .
 * Frame events are created when the window is opened, closed, moved, resized,
 * repainted, hidden by another window or exposed .
 *      The flag variable ``frame_event'' is raised by the interposed frame
 * event procedure and is lowered by this function before calling the
 * dispatcher . This function calls event dispatcher until the frame event
 * handler is not called; i.e. all frame events are cleared .
*/
  {
    if ( TER.mode != GRAPH_QUIT )
      {
        do
          {
            GFX_WIN.frame_event = FALSE;
            (void) notify_dispatch();
          }
        while ( GFX_WIN.frame_event );
      }
  }


static Notify_value catch_resize( this_frame, event, arg, type )
/*
 *      Interposed event procedure for the graphics base frame `gfx_frame'.
 *  First, the notifier is called to deal with  the base frame event.
 *  If the bar frame is open and the event is `resize',
 *    1) If necessary the canvas window is adjusted so that it is not scrolled
 *       off the right or bottom edges of logical canvas.
 *    2) If the base frame does not fill the workstation screen, scrollbars
 *       are displayed at the top and left edges of the frame window. No
 *       scrollbars are displayed if the base frame occupies the entire screen.
*/
Frame this_frame;       /* handle of the graphics frame */
Event *event;           /* pointer to frame event       */
Notify_arg arg;         /* argument for notify_next_event_func */
Notify_event_type type; /* argument for notify_next_event_func */
  {
/*
 *      first deal with the frame event
*/
    Notify_value value = notify_next_event_func( this_frame,
        (Notify_event) event, arg, type );
    GFX_WIN.frame_event = TRUE;
    if ( event_id( event ) == WIN_RESIZE )
      {
/*
 *          This allows the user to the open the graphics frame if the
 *          crosshairs are in use .
*/
        if ( window_get( this_frame, FRAME_CLOSED ) )
          {
            if ( GFX_WIN.waiting_func == T_XHRGET ||
                 GFX_WIN.waiting_func == T_TXTGET )
              {
                (void) window_set( SETUP.gfx_frame, FRAME_CLOSED, FALSE, 0 );
                TER.mode = GRAPH_OPEN;
              }
            else
                TER.mode = GRAPH_CLOSED;
          }
        else
          {
            int frame_width, frame_height;
            TER.mode = GRAPH_OPEN;
/*
 *              This looks after the scrollbars .
 *              get size of base frame window
*/
            frame_width = (int) window_get( SETUP.gfx_frame, WIN_WIDTH );
            frame_height = (int) window_get( SETUP.gfx_frame, WIN_HEIGHT );
            if ( frame_width < SETUP.full_scr_x ||
                 frame_height < SETUP.full_scr_y )
              {
/*
 *              force display of scrollbars if the graphics frame is smaller
 *              than the maximum size (whole screen) .
*/
                if ( ! window_get( SETUP.t_canvas, WIN_HORIZONTAL_SCROLLBAR ) )
                  {
                    (void) window_set( SETUP.t_canvas,
                        WIN_HORIZONTAL_SCROLLBAR, SETUP.top_scrbar,
                        WIN_VERTICAL_SCROLLBAR,   SETUP.left_scrbar, 0 );
                  }
              }
/*
 *              Remove scrollbars if the graphics frame occupies the whole
 *              workstation screen, but first scroll to the top-left corner
 *              of the canvas and explicitly call the dispatcher .
*/
            else if ( window_get( SETUP.t_canvas, WIN_HORIZONTAL_SCROLLBAR ) )
              {
                scrollbar_scroll_to( SETUP.left_scrbar, (long int) 0 );
                scrollbar_scroll_to( SETUP.top_scrbar, (long int) 0 );
                (void) window_set( SETUP.t_canvas,
                    WIN_VERTICAL_SCROLLBAR, SETUP.left_scrbar,
                    WIN_HORIZONTAL_SCROLLBAR, SETUP.top_scrbar, 0 );
                (void) notify_dispatch();
                (void) notify_dispatch();
                (void) window_set( SETUP.t_canvas,
                    WIN_HORIZONTAL_SCROLLBAR, 0,
                    WIN_VERTICAL_SCROLLBAR,   0, 0 );
              }
          }
      }
    return value;
  }


static void t_scroll_adjust()
/*
 *  Adjust vertical and horizontal scrolling; i.e. prevent scrolling onto a
 *  nonexistent part of the canvas . This function will not do anything unless
 *  the scrollbars are actually attached to the canvas .
 *    The offset is guarranteed > 0 by SunView, so we don't need to check that
 *  one .
*/
  {
    int start, length;
    if ( window_get( SETUP.t_canvas, WIN_HORIZONTAL_SCROLLBAR ) )
/*
 *      Adjust vertical scrolling .
*/
      {
        start = (int) scrollbar_get( SETUP.left_scrbar, SCROLL_VIEW_START );
        length = (int) scrollbar_get( SETUP.left_scrbar, SCROLL_VIEW_LENGTH);
        if ( ( start + length ) > (int) SETUP.sun_max._y )
          {
            scrollbar_scroll_to( SETUP.left_scrbar,
                (long int) ( (int) SETUP.sun_max._y - length ) );
            (void) window_set( SETUP.t_canvas,
                 WIN_VERTICAL_SCROLLBAR, SETUP.left_scrbar, 0 );
          }
/*
 *      Adjust horizontal scrolling .
*/
        start = (int) scrollbar_get( SETUP.top_scrbar, SCROLL_VIEW_START );
        length = (int) scrollbar_get( SETUP.top_scrbar, SCROLL_VIEW_LENGTH );
        if ( ( start + length ) > (int) SETUP.sun_max._x )
          {
            scrollbar_scroll_to( SETUP.top_scrbar,
                (long int) ( (int) SETUP.sun_max._x - length ) );
            (void) window_set( SETUP.t_canvas,
                 WIN_HORIZONTAL_SCROLLBAR, SETUP.top_scrbar, 0 );
          }
      }
  }


static Notify_value catch_destroy( this_frame, status )
/*
 *      Interposed quit (destroy graphics base frame) event handling procedure
 *  for the graphics base frame `gfx_frame'.
 *  This event handler is called twice when quitting the frame,
 *    1) the first time is to confirm the quit (status == DESTROY_CHECKING).
 *       If the quit event is generated by the user, it is vetoed, and the
 *       "quit disabled" message is displayed until the user clicks a mouse
 *       button in the message window. No quit confirmation is required if the
 *       quit event is generated by `t_end()' .
 *    2) the second time is to destroy the base frame ( but only if the event
 *       comes from `t_end()' ).
*/
Frame this_frame;       /* handle of the frame to be destroyed */
Destroy_status status;  /* confirming ? or actually destroying ? */
  {
/*
 *              If confirming the quit and the event is not from ``t_end'' .
*/
    if ( status == DESTROY_CHECKING && GFX_WIN.waiting_func != T_END )
      {
/*
 *              Veto the quit, and display the "quit disabled" message .
 *              Further I/O is blocked until the user removes the message .
*/
        (void) notify_veto_destroy( this_frame );
        (void) window_loop( SETUP.quit_frame );
        return NOTIFY_DONE;
      }
    else
      {
/*
 *              Is a destroy event (rather than a confirm) or is from ``t_end''.
 *              Reset the "waiting_func" and call the destroy event procedure .
*/
        if ( GFX_WIN.waiting_func == T_END )
            GFX_WIN.waiting_func = T_NONE;
        (void) window_set( SETUP.gfx_frame, FRAME_NO_CONFIRM, TRUE, 0 );
        return notify_next_destroy_func( this_frame, status );
      }
  }


static boolean create_quit_msg()
/*
 *      Defines the "quit disabled" message which is displayed if the user
 * attempts to quit the graphics base frame. The message can only be removed
 * by placing the cursor in the message window and clicking a mouse button.
 * The click is caught by the panel event procedure `catch_noquit()' .
 * The returned value is TRUE only if Sunview can create all of the required
 * objects.
*/
  {
/*
 *              Define message frame, panel and panel_item .
 *              If any of these fail, abort .
*/
    if ( SETUP.quit_frame = window_create( (Window) 0, FRAME,
            FRAME_SHOW_LABEL, FALSE,
            WIN_SHOW,         FALSE,
            WIN_CONSUME_PICK_EVENT, WIN_MOUSE_BUTTONS,
            WIN_IGNORE_KBD_EVENT, WIN_NO_EVENTS,
            0 ) )
        if ( SETUP.quit_panel = window_create( SETUP.quit_frame, PANEL, 0 ) )
            if ( SETUP.quit_msg   = panel_create_item( SETUP.quit_panel,
                    PANEL_MESSAGE,
                    PANEL_LABEL_STRING,
                        "QUIT DISABLED. Click a mouse button in this window",
                    PANEL_EVENT_PROC, catch_noquit,
                    0 ) )
              {
                int quit_frame_x, quit_frame_y;
/*
 *              Fit the panel and frame sizes to the size of the message .
*/
                window_fit( SETUP.quit_panel );
                window_fit( SETUP.quit_frame );
/*
 *              place the message in the middle of the screen
*/
                quit_frame_x = ( SETUP.full_scr_x -
                    (int) window_get( SETUP.quit_frame, WIN_WIDTH ) ) / 2;
                quit_frame_y = ( SETUP.full_scr_y -
                    (int) window_get( SETUP.quit_frame, WIN_HEIGHT ) ) / 2;
                (void) window_set( SETUP.quit_frame,
                    WIN_X, MAX( quit_frame_x, 0 ),
                    WIN_Y, MAX( quit_frame_y, 0 ), 0 );
              }
    return SETUP.quit_frame && SETUP.quit_panel && SETUP.quit_msg;
  }


static void catch_noquit( item, event )
/*
 *    This is the panel event procedure for intercepting the mouse button
 * click required to remove the "quit disabled" message . This is done by
 * returning from `window_loop' when a mouse button is pressed .
*/
Panel_item item;    /* the ``quit disabled'' message panel */
Event *event;       /* pointer to the event to be dealt with */
  {
    if ( event_is_button( event ) )
      window_return( (caddr_t) TRUE );
  }


static void canvas_event_proc( this_canvas, event_ptr, arg )
/*
 *    This is the t_canvas event procedure for handling events generated
 * when `t_xhrget()', `t_txtget()' and `t_wait()' are waiting for input
 * from the user . The value returned to the "waiting_function" is put
 * into ``t_return_value''. When the driver function request is satisfied,
 * the ``waiting_func'' value is cleared so that the driver function can
 * exit its notifier loop and so that this procedure does not reprocess
 * the request . Note that scroll requests never get as far as this procedure.
*/
Window this_canvas;     /* handle of the canvas */
Event *event_ptr;       /* pointer to event structure */
caddr_t arg;            /* argument to event handling procedure */
  {
/*
 *            (x,y) step of crosshairs position for each arrow key .
 *            When an arrow key is pressed, the ANSI X3.64-7 bit ASCII
 *            code is sent as an event sequence to the graphics canvas.
 *            The codes are :  arrow_up     ESC  [  A
 *                             arrow_down   ESC  [  B
 *                             arrow_right  ESC  [  C
 *                             arrow_left   ESC  [  D
 *
 *                          A  B  C   D
*/
    static short int
        arrow_x_step[] = {  0, 0, 1, -1 },
        arrow_y_step[] = { -1, 1, 0,  0 };
    boolean mouse;                          /* event is a mouse button click  */
    int key_press = event_id( event_ptr );  /* event identifier,usually ASCII */
/*
 *            Return immediately if event is a SCROLL_REQUEST .
 *            Only respond to key or button press -- ignore release .
*/
    if ( key_press == SCROLL_REQUEST || event_is_up( event_ptr) )
        return;
/*
 *            Sort out which driver function is waiting to be served .
 *            ********************************************************
*/
    switch( GFX_WIN.waiting_func )
      {
/*
 *            t_xhrget() is waiting for a printable ASCII character
 *            - or mouse button click which is interpreted as a ' '
 *            - or an arrow key to move the crosshair by one pixel.
*/
        case T_XHRGET:
/*
 *            First character of arrow escape sequence .
*/
            if ( key_press == ESC )
                GFX_WIN.was_arrow_key = 1;
/*
 *            Second character of arrow escape sequence .
*/
            else if ( GFX_WIN.was_arrow_key == 1 && key_press == '[' )
                GFX_WIN.was_arrow_key = 2;
/*
 *            Third character of arrow escape sequence - move crosshairs .
*/
            else if ( GFX_WIN.was_arrow_key == 2 )
              {
                if ( 'A' <= key_press && key_press <= 'D' )
                  {
/*
 *            Define an event with the required location .
*/
                    Event put_x;
                    event_set_x( &put_x,
                       event_x( event_ptr ) + arrow_x_step[ key_press - 'A' ] );
                    event_set_y( &put_x,
                       event_y( event_ptr ) + arrow_y_step[ key_press - 'A' ] );
/*
 *            Translate coordinate of the shifted event from the logical
 *            (scrollable) canvas to the canvas subwindow space and the
 *            place the crosshairs .
*/
                    (void) canvas_window_event( SETUP.t_canvas, &put_x );
                    (void) window_set( SETUP.t_canvas,
                        WIN_MOUSE_XY, event_x( &put_x ), event_y( &put_x ), 0 );
                    GFX_WIN.was_arrow_key = 0;
                  }
              }
/*
 *            Printable ASCII character, mouse button or return or linefeed .
*/
            else if ( ' ' <= key_press && key_press <= '~' || key_press == '\n'
                || key_press == '\r' || ( mouse = event_is_button( event_ptr )))
              {
                (void) window_set( SETUP.t_canvas,
                            WIN_IGNORE_KBD_EVENT, WIN_NO_EVENTS, 0 );
                GFX_WIN.xhair_x = event_x( event_ptr );
                GFX_WIN.xhair_y = event_y( event_ptr );
/*
 *            This tells the notifier loop in ``t_xhrget()'' that a crosshairs
 *            location has been obtained .
*/
                if ( mouse == TRUE )
                    GFX_WIN.t_return_value = (int) ' ';
                else
                    GFX_WIN.t_return_value = key_press;
                GFX_WIN.waiting_func = T_NONE;
                GFX_WIN.was_arrow_key = 0;
              }
            else
                GFX_WIN.was_arrow_key = 0;
            break;
/*
 *            t_txtget() is waiting for a printable ASCII character
 *            take care not to respond to ANSI escape sequences
 *            using the same technique as for T_XHRGET .
*/
        case T_TXTGET:
            if ( key_press == ESC )
                GFX_WIN.was_arrow_key = 1;
            else if ( GFX_WIN.was_arrow_key == 1 && key_press == '[' )
                GFX_WIN.was_arrow_key = 2;
            else if ( GFX_WIN.was_arrow_key == 2 )
                GFX_WIN.was_arrow_key = 0;
            else if ( ' ' <= key_press && key_press <= '~' ||
                 key_press == '\r' || key_press == DEL )
              {
/*
 *            This tells the notifier loop in ``t_txtget()'' that an ASCII
 *            kaey has been pressed .
*/
                GFX_WIN.waiting_func = T_NONE;
                GFX_WIN.t_return_value = key_press;
                GFX_WIN.was_arrow_key = 0;
              }
            else
                GFX_WIN.was_arrow_key = 0;
            break;
/*
 *            t_wait()  is waiting for a CR
*/
        case T_WAIT  :
            if ( key_press == '\r' )
              {
/*
 *            This tells the notifier loop in ``t_wait()'' that the return
 *            key has been obtained .
*/
                GFX_WIN.waiting_func = T_NONE;
                GFX_WIN.t_return_value = key_press;
              }
            break;
/*
 *            Extra (spurious) event, nothing is waiting for input .
*/
        case T_NONE  :
        default      :
            break;
      }
    t_scroll_adjust();
  }


