/*
    Copyright (C) 1997  Peter Vernon Lanspeary

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* ************************************************************************ */
/*                                                                          */
/*    Basic Graphics Terminal Graphics Drivers for the Visual 500 and 550   */
/*                N.B. This is hardware dependent stuff                     */
/*                                                                          */
/* ************************************************************************ */
/*                                                                          */
/*   These are graphics drivers for the Visual 500 and 550 Terminals. There */
/*   is also an xterm (Tektronics 4014 emulation mode) driver, but many     */
/*   things in it do not work because the Tek 4014 does not have the neces- */
/*   sary graphics escape codes.  The xterm functions which do not work are */
/*   t_invis(), t_revers(), t_replac(), t_boxfil(), t_xhrput().  These are  */
/*   are deliberately and silently disabled.  Only five line styles are     */
/*   available in t_dashed(), rather than eight.                            */
/*   While graphics is switched on, all terminal I/O should be through this */
/*   driver. Separate pens are used for drawing lines and for text. Text    */
/*   can be written onto the graphics screen either with a stroke generating*/
/*   font or using the terminal's firmware pixel font.                      */
/*      Most functions are of type `boolean' and return a TRUE or FALSE     */
/*   result to indicate success or failure . Error messages are placed on   */
/*   the error queue -- which can be examined at the discretion of the user.*/
/*                                                                          */
/*   Graphics escape codes for the Tektronics 4014 graphics terminal are    */
/*   described in The Tektronics 4014 service manual, Part No. #070-1648-00 */
/*   August 1974.                                                           */
/*   Any programs which use this driver must #include DEVICE_G_HEADER .     */
/*   These functions must be linked with "clip.c" and "error.c"             */
/*   The documentation on "clip.c" and "error.c" should be read before      */
/*   continuing .                                                           */
/*                                                                          */
/*   External functions used :                                              */
/*       anything fprintf(), fputc(), fputs();                              */
/*       boolean make_error();                                              */
/*       int abs(), t_strlen();                                             */
/*       double fabs();                                                     */
/*       clip_area VAR_PTR def_2rclip();                                    */
/*       char VARPTR t_strcpy();                                            */
/*   The macro getchar() is also used .                                     */
/*                                                                          */
/*   N.B. There is no screen dump routine `t_dump' because you can dump the */
/*   contents of the graphics screen directly onto a printer via the        */
/*   terminal's auxilary port .                                             */
/*                                                                          */
/* ************************************************************************ */


#include "system.h"             /* system dependent stuff */
#include <stdio.h>
#include DEVICE_G_HEADER
#include FONT_SOURCE_HEADER
#include "dev_udef.h"

/*
 *      Machine constants
*/
/*#define NULL        0       NULL pointer - already defined in stdio.h   */
/*#define EOF         (-1)    end of file  - already defined in stdio.h   */

/*
 *    non printable ASCII characters
 *  delete unused characters when finished
*/
#define NUL  '\000'   /* Nul                        */
#define BEL  '\007'   /* Bell                       */
#define BS   '\010'   /* Backspace                  */
#define HT   '\011'   /* Horizontal Tabulation      */
#define LF   '\012'   /* Line Feed                  */
#define VT   '\013'   /* Vertical Tabulation        */
#define FF   '\014'   /* Form Feed                  */
#define CR   '\015'   /* Carriage Return            */
#define CAN  '\030'   /* Cancel                     */
#define EM   '\031'   /* End of Medium              */
#define ESC  '\033'   /* Escape                     */
#define SUB  '\032'   /* Substitute                 */
#define FS   '\034'   /* File Separator             */
#define GS   '\035'   /* Group Separator            */
#define US   '\037'   /* Unit Separator             */

#define PI            3.141592653589793238462643

#define   T4010_X   1023.0      /* viewport upper right coordinate - pixels */
#define   T4010_Y   779.0       /*  for the Tektronix 4010 terminal         */
#define   V500_X    767.0       /* viewport upper right coordinate - pixels */
#define   V500_Y    584.0       /*   for the Visual 500 terminal            */

#define ALPH_HEIGHT   17.0      /* height in pixels of alphagraphics char   */
#define ALPH_WIDTH    10.0      /* width in pixels of alphagraphics char    */
#define ALPH_LINES    34        /* number lines on alphagraphics screen     */
#define ALPH_COLUMNS  76        /* number columns on alphagraphics screen   */
#define PEN_THICKNESS 1.0       /* pen or line thickness                    */
#define PIXEL_SIZE    0.3       /* horiz. and vertical size of a pixel (mm) */
#define SOLID_LINE    1
#define NOT_USED      -1

#define   CLIP2R_ID 0x3c22      /* CL2R two dimensional rectangular clipping*/


/* ************************************************************************ */
/*                                                                          */
/*                          Graphics Macros                                 */
/*                                                                          */
/* ************************************************************************ */

/*
 *  ABBREVIATIONS
 ******************
*/
#define  TER        v500->ter
#define  STROKE     v500->stroke
#define  VIEW_CLIP  v500->view_clip


/*
 *  CATCH_NULL(pointer,caller,ret_type)
 ****************************************
 *      If "pointer" has a NULL value, an "Arg is NULL pointer" error message
 * is generated and the function is aborted . Otherwise, nothing is done .
 *      "pointer"   is the pointer to be tested,
 *      "caller"    is a character string naming the function
 *      "ret_type"  is the type of the returned zero value which indicates an
 *                    error condition
*/
#define CATCH_NULL(pointer,caller,ret_type) \
  if ( ! (pointer) )    \
      return (ret_type) make_error( ERR_ABORT, caller, err_null )

/*
 *  G_PUTC(c)
 **************
 *      Writes the character `c' to the graphics device (terminal) .
 *      There is no returned result.
*/
#define G_PUTC(c)           fputc( c, v500->t_op_dev )

/*
 *  G_PUTS(string)
 *******************
 *      Writes the character string `string' to the graphics device
 *      (terminal) . The there is no returned result .
*/
#define G_PUTS(string)      fputs( string, v500->t_op_dev )

/*
 *  G_ERROR()
 **************
 *      Determines the error status of the graphics device (terminal) .
 *      The returned value is TRUE if an error has occurred, otherwise FALSE .
*/
#define G_ERROR()           ferror( v500->t_op_dev )

/*
 *  G_FLUSH()
 **************
 *      Flushes the graphics device (terminal) output buffer . The returned
 *  value is TRUE if successful.
*/
#define G_FLUSH()          ! fflush( v500->t_op_dev )

/*
 *  LIMIT( x, min, max )
 *************************
 *      Limits the value of `x' to between `min' and `max'
*/
#define LIMIT(x,min,max)  (x) <= (min) ? (min) : ( (x) >= (max) ? (max) : (x) )

/*
 *  MAX( x, y )
 ****************
 *      The returned value is the maximum of `x' and `y' .
*/
#define MAX(x,y)        ( (x) > (y) ? (x) : (y) )

/*
 *  MIN( x, y )
 ****************
 *      The returned value is the minimum of `x' and `y' .
*/
#define MIN(x,y)        ( (x) > (y) ? (y) : (x) )

/*
 *  START_UP(name,ret_type)
 ****************************
 *      Calls `t_begin' with the default arguments `stdout' and `mt_str'
 *  (empty string) if `t_begin' has not previously been called .
*/
#define START_UP(name,ret_type) \
    if ( ! v500 ) \
        if ( ! (* device_ptr->t_BEGIN)( stdout, mt_str ) ) \
            return (ret_type) make_error( ERR_CALL, name, mt_str )

/*
 *  SWAPD( a, b )
 *****************
 *    swaps the two double precision floating point values `a' and `b'
*/
#define SWAPD(a,b)       { double temp = b; b = a; a = temp; }

/*
 *  TOLOWER( c )
 *      Converts character `c' to lower case if it is upper case
*/
#define TOLOWER(c)      'A' <= (c) && (c) <= 'Z' ? (c) + 'a' - 'A' : (c);


/* ************************************************************************ */
/*                                                                          */
/*                          Type declarations                               */
/*                                                                          */
/* ************************************************************************ */

enum dev_mode     /* terminal mode */
  {
    ALPH_NUMERIC,       /* alphanumeric mode  - normal, non graphics screen */
    CROSS_HAIR,         /* cross hair mode - display cross hairs            */
    ALPH_GRAPHIC,       /* alphagraphics mode - text on graphics screen     */
    VECTOR_MODE,        /* vector mode     - plot vector between end points */
    POINT_PLOT,         /* point plot mode - dashed line style 0 - no lines */
  };

/*
 *      The `data level' of the Visual 500 terminal controls how the
 *   terminal sets or reset pixel dots . The `data level' is set for
 *   use in alphagraphics, vector, point plot or incremental point plot
 *   mode . These `data levels' are,
*/
enum dev_dots     /* data level of terminal */
  {
    DOT_ON,             /* draw visible vectors, `or' them onto the screen  */
    DOT_OFF,            /* draw invisible vectors, switch dots off          */
    DOT_REV,            /* the compliment of the new data is plotted        */
    DOT_NEW             /* replace pixel contents with new data - not used  */
  };

typedef struct            /* TERMINAL STATUS                     */
  {                       /*     DATA TYPE                       */
    enum dev_mode mode;   /* terminal mode                       */
    enum dev_dots dots;   /* data level                          */
    int           colour; /* foreground colour                   */
    int           dash;   /* dashed line style                   */
    int           thick;  /* line and test width                 */
    coord_pair    cursor; /* alpha graphics text cursor position */
    coord_pair    pen;    /* vector graphics pen position        */
  }
    Ter_Status;

/*
 *      Graphics status description
 * -- this data structure contains everything the driver knows about the
 *    Visual 500 terminal
*/
typedef struct
  {
    int type;               /* device type -- one of VISUAL500_TERMINAL */
                            /* VISUAL_550_TERMINAL, XTERM_VT100TEK4014  */
    coord_pair scr_size;    /* size of the screen in pixels */
    struct
      {
        Ter_Status requir;  /* required terminal status */
        Ter_Status actual;  /* actual terminal status   */
      }
        ter;
    Text_Attribute stroke;  /* stroke text attributes   */
    FILE *t_op_dev;         /* file pointer for the graphics output device  */
    boolean last_draw;      /* last output was a `draw vector' */
    clip_area VAR_PTR view_clip;    /* pointer to viewport clipping area */
  }
    Gfx_Status;


/* ************************************************************************ */
/*                                                                          */
/*                          Global variables                                */
/*                                                                          */
/* ************************************************************************ */


extern StrokeFont
    asc_font,   /* font for ASCII chars   */
    sym_font;   /* font for special chars */

static Gfx_Status
    VAR_PTR v500 = (Gfx_Status VAR_PTR) 0;      /* pointer to graphics status */

/*
 *      Common error messages
*/
static char
  VAR_PTR err_write  = "Cannot write to graphics device",
  VAR_PTR err_null   = "Arg is NULL pointer",
  VAR_PTR mt_str     = "";
/*
 *      Declare static and all external library functions whose returned
 *      value is used.
*/
static char VAR_PTR pr_dots(), VAR_PTR pr_mode();
static int t_txtstr();
static boolean t4010_vec(), t_stkchr(),
               v500_dots(), v500_box(), v500_style();

/* int fflush();    declared in <stdio.h> */
double fabs();
int    abs(), free();
char  *malloc();

/* ************************************************************************ */
/*                                                                          */
/*                    Terminal driving routines                             */
/*                                                                          */
/* ************************************************************************ */


extern T_dev_fns VAR_PTR device_ptr;

static boolean t_dump( file_name )
/*
 *      Dummy for the nonexistent screen dump routine
*/
char VAR_PTR file_name;         /* name of file containing the screen dump */
  {
    return make_error( ERR_ABORT, (char VAR_PTR) "t_dump",
        (char VAR_PTR) "function unavailable" );
  }


static clip_area VAR_PTR t_begin( device_type, out_file, options )
/*
 *        Start graphics .
 *    Puts terminal in vector mode, sets up the terminal, initializes
 *    variables, reads the stroke-character fonts and sets up the viewport
 *    clip. The returned value is a pointer to the clip_area structure of the
 *    viewport . The function fails ( i.e. returns a NULL pointer ) if cannot
 *      a) define the viewport clip_area
 *      b) write to the terminal .
 *    Appropriate error messages are produced in these cases .
 *
 *  Argument
*/
int  device_type;     /* device type -- VISUAL500_TERMINAL, */
                      /*     VISUAL550_TERMINAL or XTERM_VT100TEK4014 */
FILE * out_file;      /* file pointer for graphics device - usually = stdout */
char VAR_PTR options; /* options string - not used for Visual 500 driver */
  {
    char VAR_PTR name  = "t_begin";
    coord_pair origin;          /* viewport lower left corner */
    static struct               /* initial terminal status -- read only */
      {
        Ter_Status requir;      /* required terminal status */
        Ter_Status actual;      /* actual terminal status */
      }
        init_ter = { { VECTOR_MODE, DOT_ON, 1, SOLID_LINE, 1,
                       { 0.0, V500_Y - ALPH_HEIGHT + 1}, { 0.0, 0.0 } },
                     { VECTOR_MODE, DOT_ON, NOT_USED, SOLID_LINE, NOT_USED,
                       { 0.0, 0.0 }, { 0.0, 0.0 } } };
    CATCH_NULL( out_file, name, clip_area VAR_PTR );
    if ( device_type != VISUAL500_TERMINAL &&
         device_type != VISUAL550_TERMINAL &&
         device_type != XTERM_VT100TEK4014 )
        return (clip_area VAR_PTR) make_error( ERR_ABORT, name, err_null );
/*
 *      Find out if ``out_file'' has been registered as a device .
 *      If it has, check that it is indeed a Visual 500 and grab the
 *      pointer to it graphics status data .
*/
    if ( t_register( T_ISREGIST, out_file, 0, (char VAR_PTR) 0 ) )
      {
        if ( (int) t_register( T_REG_TYPE, out_file, NONEXISTENT_DEVICE,
                (char VAR_PTR) 0 ) != device_type )
            return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "Wrong driver / device type" );
        v500 = (Gfx_Status VAR_PTR) t_register( T_REG_DATA,
            out_file, NONEXISTENT_DEVICE, (char VAR_PTR) 0 );
      }
    else
      {
/*
 *      ``out_file'' has never been registered, so allocate memory for the
 *      graphics status data and register the device as a Visual 500 .
*/
        v500 = (Gfx_Status VAR_PTR) malloc( (unsigned int) sizeof(Gfx_Status) );
        if ( ! v500 )
            return (clip_area VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "Cannot allocate memory" );
        if ( ! t_register( T_REGISTER, out_file,
                           device_type, (char VAR_PTR) v500 ) )
          {
            (void) free( (char *) v500 );
            v500 = (Gfx_Status VAR_PTR) 0;
            return (clip_area VAR_PTR) make_error( ERR_CALL, name, mt_str );
          }
        v500->t_op_dev = out_file;  /* point to the graphics device */

        v500->type = device_type;
        if ( device_type == XTERM_VT100TEK4014 )
          {
            v500->scr_size._x = T4010_X;
            v500->scr_size._y = T4010_Y;
          }
        else
          {
            v500->scr_size._x = V500_X;
            v500->scr_size._y = V500_Y;
          }
        init_ter.requir.cursor._y = v500->scr_size._y - ALPH_HEIGHT + 1;
/*
 *           reset character attributes and terminal status
*/
        (void) reset_font( (Text_Attribute VAR_PTR) &STROKE,
            v500->scr_size._y / 20.0 );
        TER.requir = init_ter.requir;
/*
 *      initialize the viewport corners
*/
        origin._x = origin._y = 0.0;
        if ( ! ( VIEW_CLIP = def_2rclip( (coord_pair VAR_PTR) &origin,
                (coord_pair VAR_PTR) &(v500->scr_size),
                (coord_pair VAR_PTR) &origin ) ) )
          {
            (void) free( (char *) v500 );
            v500 = (Gfx_Status VAR_PTR) 0;
            return (clip_area VAR_PTR) make_error( ERR_CALL, name, mt_str );
          }
      }
/*
 *      set terminal to
 *   Vector mode,      Dots-on,   Auto scaling = 3:4,
 *   SP code = non destructive,   Cell size = 10x17,
 *
*/
    if ( v500->type == XTERM_VT100TEK4014 )
        (void) G_PUTS( "\033[?38h\035\033/0d\033/1h\033/2h\033/3l\035\035" );
    else if (  v500->type == VISUAL500_TERMINAL ||
               v500->type == VISUAL550_TERMINAL )
        (void) G_PUTS( "\035\033/0d\033/1h\033/2h\033/3l\035\035" );
    if ( ! t4010_vec( 0, 0 ) )
        return (clip_area VAR_PTR) make_error( ERR_WARN, name, err_write );
    v500->last_draw = TRUE;
    TER.actual = init_ter.actual;
    return VIEW_CLIP;
  }


static clip_area VAR_PTR v500_begin( v500_ter, options )
FILE * v500_ter;
char VAR_PTR options;
  {
    return t_begin( VISUAL500_TERMINAL, v500_ter, options );
  }


static clip_area VAR_PTR v550_begin( v550_ter, options )
FILE * v550_ter;
char VAR_PTR options;
  {
    return t_begin( VISUAL550_TERMINAL, v550_ter, options );
  }


static clip_area VAR_PTR xterm_begin( xterm_ter, options )
FILE * xterm_ter;
char VAR_PTR options;
  {
    return t_begin( XTERM_VT100TEK4014, xterm_ter, options );
  }


static boolean t_vwport( clip_ptr )
/*
 *      Set the clipping limits for the viewport . The old limits are replaced .
 * The `clip_ptr' must point to  clip_area which has been defined by
 * `def_2rclip()' . The function fails if
 *   a) clip_ptr is a NULL pointer
 *   b) clip_ptr points to something which is not a structure for a two
 *      dimensional rectangular clipping region .
 *
 * Argument
*/
clip_area VAR_PTR clip_ptr;
  {
    char VAR_PTR name = "t_vwport";
    CATCH_NULL( clip_ptr, name, boolean );
    START_UP( name, boolean );
    if ( clip_ptr->id_struct == CLIP2R_ID )
      {
        if ( -0.5 < min_clip( clip_ptr, _x ) &&
                    max_clip( clip_ptr, _x ) <= v500->scr_size._x &&
             -0.5 < min_clip( clip_ptr, _y ) &&
                    max_clip( clip_ptr, _y ) <= v500->scr_size._y )
          {
            VIEW_CLIP = clip_ptr;
            return TRUE;
          }
        return make_error( ERR_ABORT, name,
            (char VAR_PTR) "Outside device limits" );
      }
   return make_error( ERR_ABORT, name, (char VAR_PTR) "Invalid pointer" );
  }


static boolean t_colour( ipen )
/*
 *    Selects specified colour ( or pen ) .
 *
 * Because the Visual 500 terminal is monochrome, colours are simulated with
 * various dashed line styles . The only true ``colours'' available are dark
 * and bright . The returned value is TRUE if  0 <= ipen <= 8 . If these limits
 * are exceeded, the value is wrapped around into the range 1 to 8 . The dashed
 * line equivalents of the colours are,
 *
 *    IPEN    SCREEN                  PLOTTER
 *          LINE STYLE                COLOUR
 *     0      black  (background)     no pen = invisible
 *     1      solid  (green)          black
 *     2      dotted                  brown
 *     3    short dash                red
 *     4    long dash                 orange
 *     5    dot dash                  yellow
 *     6    short-dash long-dash      green
 *     7    2-short-dashes long-dash  blue
 *     8    3-dots long-dash          violet
 *
 *   Argument
*/
int ipen;                /* pen number or colour */
  {
    START_UP( (char VAR_PTR) "t_colour", boolean );
    TER.requir.colour = ( abs( ipen ) - 1 ) % 8 + 1;
    return 0 <= ipen && ipen <= 8;
  }


static boolean t_dashed( style )
/*
 *      Set the dashed line style . The returned value is TRUE if
 * 0 <= style <= 8 . If these limits are exceeded, the value is wrapped
 * around into the range 1 to 8 .
 *
 *    STYLE   SCREEN
 *          LINE STYLE
 *     0      black  - plot vector endpoints only (invisible)
 *     1      solid  (green)
 *     2      dotted
 *     3    short dash
 *     4    long dash
 *     5    dot dash
 *     6    short-dash long-dash
 *     7    2-short-dashes long-dash
 *     8    3-dots long-dash
 *
*/
int style;        /* dashed line style */
  {
    START_UP( (char VAR_PTR) "t_dashed", boolean );
    TER.requir.dash = ( abs( style ) - 1 ) % 8 + 1;
    return 0 <= style && style <= 8;
  }


static boolean t_thick( width )
/*
 * Set the ``width'' of lines and text . The returned value is TRUE
 * if 0 <= thickness.
*/
int width;      /* line thickness */
  {
    START_UP( (char VAR_PTR) "t_thick", boolean );
    TER.requir.thick = width;
    return 0 <= width;
  }


/* ************************************************************************ */
/*                                                                          */
/*                          Turn graphics on or off                         */
/*                                                                          */
/* ************************************************************************ */


static boolean t_gr_off()
/*
 *      Turn graphics off . I.e. put terminal into alpha_numeric mode . This
 *  function fails if it cannot write to the screen .
*/
  {
    char VAR_PTR name =  "t_gr_off";
    int i;
    START_UP( name, boolean );
    if ( TER.actual.mode != ALPH_NUMERIC )
      {
        if ( v500->type == XTERM_VT100TEK4014 )
          {
           (void) G_PUTC( US );
           for( i=0; i<64; i++ )
              (void) G_PUTS( "                " );
           (void) G_PUTS( "\033\003" );
          }
        else
            (void) G_PUTC( CAN );
      }
    TER.requir.mode = ALPH_NUMERIC;
    (void) G_FLUSH();
    if ( G_ERROR() )
        return make_error( ERR_ABORT, name, err_write );
    TER.actual.mode = ALPH_NUMERIC;
    v500->last_draw = FALSE;
    return TRUE;
  }


static boolean t_gr_on()
/*
 *      Turn graphics on . I.e. if not already in vector or alphagraphics
 *  mode, put terminal into vector mode . This function fails if it cannot
 *  write to the screen .
*/
  {
    char VAR_PTR name = "t_gr_on()";
    START_UP( name, boolean );
    if ( TER.actual.mode == ALPH_NUMERIC && v500->type == XTERM_VT100TEK4014 )
        (void) G_PUTS( "\033[?38h" );
    if ( TER.actual.mode != VECTOR_MODE &&
         TER.actual.mode != ALPH_GRAPHIC )
      {
        (void) G_PUTC( GS );
        TER.requir.mode = VECTOR_MODE;
      }
    if ( G_ERROR() )
        return make_error( ERR_WARN, name, err_write );
    v500->last_draw = FALSE;
    TER.actual.mode = VECTOR_MODE;
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*           Change Data Level, Visible, Invisible, Revers, Replace         */
/*                                                                          */
/* ************************************************************************ */



static boolean t_visib()
/*
 *  Draw visible vectors . Sets screen into "dots on" mode . This function
 *  always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_visib", boolean );
    TER.requir.dots = DOT_ON;
    return TRUE;
  }


static boolean t_invis()
/*
 *  Draw invisible vectors . Sets screen into "dots off" mode . This function
 *  always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_invis", boolean );
    TER.requir.dots = DOT_OFF;
    return TRUE;
  }


static boolean t_revers()
/*
 *  Reverse the pixel dots when plotting vectors, plotting points or
 *  alphagraphic characters . Sets screen into "compliment" mode . This function
 *  always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_revers", boolean );
    TER.requir.dots = DOT_REV;
    return TRUE;
  }


static boolean t_replac()
/*
 *  Replace pixel contents with the new data . Sets screen into "replace"
 *  mode . This function always succeeds .
*/
  {
    START_UP( (char VAR_PTR) "t_replac", boolean );
    TER.requir.dots = DOT_NEW;
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                         Miscellaneous stuff                              */
/*                                                                          */
/* ************************************************************************ */

static boolean t_newpag()
/*
 * Clear the graphics screen . This function fails if it cannot write to the
 * screen .
*/
  {
    char VAR_PTR name = "t_newpag";
    START_UP( name, boolean );
    if ( TER.actual.mode == ALPH_NUMERIC && v500->type == XTERM_VT100TEK4014 )
        (void) G_PUTS( "\033[?38h" );
    if ( TER.actual.mode == ALPH_NUMERIC || TER.actual.mode == CROSS_HAIR )
      {
        (void) G_PUTC( GS );
        TER.actual.mode = VECTOR_MODE;
      }
    (void) G_PUTS( "\033\014" );
    (void) G_FLUSH();
    if ( G_ERROR() )
        return make_error( ERR_WARN, name, err_write );
    TER.actual.mode = ALPH_GRAPHIC;
    TER.actual.cursor._x = 0.0;
    TER.actual.cursor._y = v500->scr_size._y - ALPH_HEIGHT;
    TER.actual.dots = DOT_ON;
    v500->last_draw = FALSE;
    return TRUE;
  }


static boolean vis500_newscr()
/*
 * Clear the alphanumeric screen . This function fails if it cannot
 *  write to the screen .
*/
  {
    char VAR_PTR name = "t_newscr";
    START_UP( name, boolean );
    if ( TER.actual.mode != ALPH_NUMERIC )
        (void) G_PUTC( CAN );
    (void) G_PUTS( "\033w" );
    if ( G_ERROR() )
        return make_error( ERR_WARN, name, err_write );
    TER.actual.mode = ALPH_NUMERIC;
    v500->last_draw = FALSE;
    return TRUE;
  }


static boolean vis550_newscr()
/*
 * Clear the alphanumeric screen . This function fails if it cannot
 *  write to the screen .
*/
  {
    char VAR_PTR name = "t_newscr";
    START_UP( name, boolean );
    if ( TER.actual.mode != ALPH_NUMERIC )
        (void) G_PUTC( CAN );
    (void) G_PUTS( "\033[2J\033[H" );
    if ( G_ERROR() )
        return make_error( ERR_WARN, name, err_write );
    TER.actual.mode = ALPH_NUMERIC;
    v500->last_draw = FALSE;
    return TRUE;
  }


static boolean xterm_newscr()
/*
 * Clear the alphanumeric screen . This function fails if it cannot
 *  write to the screen .
*/
  {
    char VAR_PTR name = "t_newscr";
    START_UP( name, boolean );
    if ( TER.actual.mode != ALPH_NUMERIC )
        (void) G_PUTS( "\033\003" );
    (void) G_PUTC( FF );
    (void) G_FLUSH();
    if ( G_ERROR() )
        return make_error( ERR_WARN, name, err_write );
    TER.actual.mode = ALPH_NUMERIC;
    v500->last_draw = FALSE;
    return TRUE;
  }


static boolean t_bell()
/*
 *      Ring bell . This function fails if it cannot write to the screen .
*/
  {
    char VAR_PTR name = "t_bell";
    START_UP( name, boolean );
    v500->last_draw = FALSE;
    (void) G_PUTC( BEL );
    if ( G_ERROR() )
        return make_error( ERR_WARN, name, err_write );
    return TRUE;
  }


static boolean t_wait()
/*
 *  Waits for the user to press the carriage return .
 *  The returned value is TRUE unless it cannot write to the screen .
*/
  {
    char VAR_PTR name = "t_wait";
    START_UP( name, boolean );
    (void) G_FLUSH();
    if ( G_ERROR() )
        return make_error( ERR_WARN, name, err_write );
    while( getchar() != '\n' );
    if ( TER.actual.mode == VECTOR_MODE || TER.actual.mode == POINT_PLOT )
        TER.actual.mode = ALPH_GRAPHIC;
    return TRUE;
  }


static boolean t_end()
/*
 *      End of graphics .
*/
  {
    boolean ok;
    char VAR_PTR name = "t_end";
/*
 *      Deregister the device and free the memory used by the graphics status
 *      data .
*/
    START_UP( name, boolean );
    if ( ! t_register( T_DEREGIST, v500->t_op_dev,
            v500->type, (char VAR_PTR) v500 ) )
        (void) make_error( ERR_WARN, name, mt_str );
    (void) free( (char *) v500 );
/*
 *      Delete character fonts and frees the memory allocated to them
*/
    if ( asc_font.font_got )
        (void) clear_font( (StrokeFont VAR_PTR) &asc_font );
    if ( sym_font.font_got )
        (void) clear_font( (StrokeFont VAR_PTR) &sym_font );
/*
 *      Clear the graphics screen and turn graphics off .
*/
    ok = t_newpag();
    return t_gr_off() && ok;
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Text Writing                                   */
/*                                                                          */
/* ************************************************************************ */


static boolean t_txtmov( point )
/*
 *    Move the text cursor to coordinate `point' . The returned result is
 *  TRUE only if the next printable character is within the screen limits .
 *
 * Argument
*/
register coord_pair VAR_PTR point;
  {
    char VAR_PTR name = "t_txtmov";
    START_UP( name, boolean );
    CATCH_NULL( point, name, boolean );
    TER.requir.cursor = *point;
    return -0.5 < point->_x  &&  point->_x <= v500->scr_size._x - ALPH_WIDTH &&
           -0.5 < point->_y  &&  point->_y <= v500->scr_size._y - ALPH_HEIGHT;
  }


static boolean t_txtcur( row, column )
/*
 *    Move the text cursor to text row `row' and text column `column' .
 *    The text row is numbered from the top of the screen . The returned
 *    result is TRUE only if the next printable character is within screen
 *    limits . The row and column positions are calculated for the default
 *    alphagraphics character size .
 *
 * Argument
*/
int row, column;
  {
    START_UP( (char VAR_PTR) "t_txtcur", boolean );
    TER.requir.cursor._x = ALPH_WIDTH * column;
    TER.requir.cursor._y = v500->scr_size._y - ALPH_HEIGHT + 1.0 -
                           ALPH_HEIGHT * row;
    return 0 <= column  &&  column <= ALPH_COLUMNS  &&
           0 <= row     &&  row    <= ALPH_LINES;
  }


static boolean t_txtset( attribute, value )
/*
 *  Set the stroke character attribute `attribute' to `*value' .
 *  If `attribute' is negative, the current value of the attribute is returned
 *  to `*value' . This function fails if the `attribute' cannot be recognized or
 *  `value' is a null pointer . Appropriate error messages are generated .
*/
int attribute;
char VAR_PTR value;
  {
    char VAR_PTR name = "t_txtset";
    double VAR_PTR dvalue;       /* pointer to a double `value' */
    dvalue = (double VAR_PTR) value;
    if ( attribute != T_CH_RESET  && attribute != T_CH_BOLD  &&
             attribute != T_CH_NOBOLD && attribute != T_ASCII    &&
             attribute != T_SYMBOL )
        CATCH_NULL( value, name, boolean );
    START_UP( name, boolean );
    switch( attribute )
      {
        case  T_CH_RESET :
            (void) reset_font( (Text_Attribute VAR_PTR) &STROKE,
                       v500->scr_size._y / 20.0 );
            break;
        case  T_CH_HEIGHT :
            if ( STROKE.height != *dvalue )
              {
                STROKE.height = LIMIT( *dvalue, 5.0, v500->scr_size._y );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_HEIGHT :
            *dvalue = STROKE.height;
            break;
        case  T_CH_ASPECT :
            if ( STROKE.aspect != *dvalue )
              {
                STROKE.aspect = LIMIT( *dvalue, 0.2, 5.0 );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_ASPECT :
            *dvalue = STROKE.aspect;
            break;
        case  T_CH_SLANT :
            if ( STROKE.slant != *dvalue )
              {
                STROKE.slant = LIMIT( *dvalue, -80.0, 80.0 ) * PI / 180.0;
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_SLANT :
            *dvalue = STROKE.slant * 180.0 / PI;
            break;
        case  T_CH_ANGLE :
            if ( STROKE.angle != *dvalue )
              {
                STROKE.angle = PI / 180.0 * ( *dvalue -
                    360.0 * (double) (int) ( *dvalue / 360.0 ) );
                STROKE.angle = STROKE.angle < -PI ? STROKE.angle + 2.0 * PI :
                    ( STROKE.angle > PI ? STROKE.angle-2.0*PI : STROKE.angle );
                STROKE.changed = TRUE;
              }
            break;
        case -T_CH_ANGLE :
            *dvalue = STROKE.angle * 180.0 / PI;
            break;
        case  T_CH_GAP :
            STROKE.gap = LIMIT( *dvalue, 0.05, 5.0 );
            break;
        case -T_CH_GAP :
            *dvalue = STROKE.gap;
            break;
        case  T_CH_ALIGN :
            STROKE.align.vert  = TOLOWER( *value );
            value++;
            STROKE.align.horiz = TOLOWER( *value );
            if ( STROKE.align.horiz != 'l' && STROKE.align.horiz != 'c' &&
                    STROKE.align.horiz != 'r' )
                STROKE.align.horiz = 'l';
            if ( STROKE.align.vert != 'b' && STROKE.align.vert != 'c' &&
                    STROKE.align.vert != 't' )
                STROKE.align.vert = 'b';
            break;
        case -T_CH_ALIGN :
            *value++ = STROKE.align.vert;
            *value++ = STROKE.align.horiz;
            *value   = NUL;
            break;
        case T_ASCII :
            STROKE.font_ptr = (StrokeFont VAR_PTR) &asc_font;
            break;
        case T_SYMBOL :
            STROKE.font_ptr = (StrokeFont VAR_PTR) &sym_font;
            break;
        case T_CH_BOLD :
            STROKE.bold = TRUE;
            break;
        case T_CH_NOBOLD :
            STROKE.bold = FALSE;
            break;
        default :
            return make_error( ERR_ABORT, name,
                (char VAR_PTR) "Unknown stroke text attribute" );
      }
    return TRUE;
  }


static boolean t_symdrw( point, symbol )
/*
 *      Draws a centred plotting symbol `symbol' at position `point' according
 *      to the selected stroke character attributes . This function fails if
 *      the font for the symbol is not defined or if it cannot write to the
 *      screen .
*/
coord_pair VAR_PTR point;
char       symbol;
  {
    char VAR_PTR name = "t_symdrw";
    boolean ok = TRUE;
    int font_index = symbol - 32;
    coord_pair offset, lower_left;
    StrokeChar VAR_PTR char_ptr;
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    TER.requir.pen = *point;
/*
 *              SET UP THE TERMINAL
 *
 *      put terminal into vector mode
*/
    if ( TER.actual.mode == ALPH_NUMERIC && v500->type == XTERM_VT100TEK4014 )
        (void) G_PUTS( "\033[?38h" );
    if ( TER.actual.mode != VECTOR_MODE )
      {
        (void) G_PUTC( GS );
        if ( ok &= ! G_ERROR() )
            TER.actual.mode = VECTOR_MODE;
        v500->last_draw = FALSE;
      }
/*
 *      set colour to black or white (dashed lines not allowed)
*/
    if ( ! TER.requir.colour )
      {
        if ( TER.actual.dots != DOT_OFF )
            ok &= v500_dots( DOT_OFF );
      }
    else if ( TER.requir.dots != TER.actual.dots )
        ok &= v500_dots( TER.requir.dots );
/*
 *      set dashed line style to solid
*/
    if ( TER.actual.dash != SOLID_LINE )
        ok &= v500_style( SOLID_LINE );
/*
 *           if necessary read characTER fonts
*/
    if ( ! STROKE.font_ptr->font_got )
        if ( ! read_font( STROKE.font_ptr, file_font( STROKE.font_ptr ), name) )
            return FALSE;
/*
 *      CALCULATE LOCATION OF LOWER LEFT CORNER OF THE SYMBOL
 *
 *      determine the character cell size vector
*/
    if ( STROKE.changed )
        (void) map_font( (Text_Attribute VAR_PTR) &STROKE, PEN_THICKNESS );
    char_ptr = STROKE.font_ptr->font_dir + font_index;
    offset._x = STROKE.font_ptr->font_wid * 0.5;
    offset._y = char_ptr->font_midpt;
/*
 *    determine start position
*/
    lower_left._x = point->_x -
        STROKE.map[0][0] * offset._x - STROKE.map[0][1] * offset._y;
    lower_left._y = point->_y -
        STROKE.map[1][0] * offset._x - STROKE.map[1][1] * offset._y;

/*
 *      DRAW THE SYMBOL
*/
    if ( 0 <= font_index && font_index < STROKE.font_ptr->font_siz )
      {
        if ( char_ptr->font_char == symbol && char_ptr->font_npts > 0 )
          {
            boolean ok;
            ok = bold_font( (Text_Attribute VAR_PTR) &STROKE,
                            (coord_pair VAR_PTR) &lower_left,
                            (char VAR_PTR) &font_index, t_stkchr );
            v500->last_draw = FALSE;
            return ok & move_clip( (coord_triple VAR_PTR) point,
                VIEW_CLIP );
          }
      }
    return make_error( ERR_ABORT, name, (char VAR_PTR) "Unknown symbol" );
  }


static int t_txtdrw( string )
/*
 *      Draws the text contained in the character string `string' at the
 *   current cursor location according to the selected stroke character
 *   attributes . The returned value is the number of characters drawn . If
 *   an error occurs, the returned value is 0 . An error may be generated
 *   because
 *     a) a character in `string' is unavailable in the specified font
 *     b) the character is not printable
 *     c) could not update the pen position after drawing the characters
 *     d) it cannot write to the terminal
*/
char VAR_PTR string;
  {
    char VAR_PTR name = "t_txtdrw";
    boolean ok = TRUE;          /* success flag                             */
    int font_index,             /* index of the character in font directory */
        n_chars,                /* length of `string'                       */
        char_count=0;           /* counts each character as it is drawn     */
    coord_pair cell_size,       /* size of the character cell               */
               offset,
               point;           /* location of char cell lower left corner  */
    CATCH_NULL( string, name, boolean );
    START_UP( name, boolean );
/*
 *      SET UP THE TERMINAL
 *
 *      put terminal into vector mode
*/
    if ( TER.actual.mode == ALPH_NUMERIC && v500->type == XTERM_VT100TEK4014 )
        (void) G_PUTS( "\033[?38h" );
    if ( TER.actual.mode != VECTOR_MODE )
      {
        (void) G_PUTC( GS );
        if ( ok &= ! G_ERROR() )
            TER.actual.mode = VECTOR_MODE;
        v500->last_draw = FALSE;
      }
/*
 *           if necessary read character fonts
*/
    if ( ! STROKE.font_ptr->font_got )
        if ( ! read_font( STROKE.font_ptr, file_font( STROKE.font_ptr ), name) )
            return 0;
/*
 *      set colour to black or white (dashed lines not allowed)
*/
    if ( ! TER.requir.colour )
      {
        if ( TER.actual.dots != DOT_OFF )
            ok &= v500_dots( DOT_OFF );
      }
    else if ( TER.requir.dots != TER.actual.dots )
        ok &= v500_dots( TER.requir.dots );
/*
 *      set dashed line style to solid
*/
    if ( TER.actual.dash != SOLID_LINE )
        ok &= v500_style( SOLID_LINE );

    if ( STROKE.changed )
      (void) map_font( (Text_Attribute VAR_PTR) &STROKE, PEN_THICKNESS );
/*
 *      CALCULATE LOCATION OF LOWER LEFT CORNER OF THE STRING
 *
 *      determine the character cell size vector
*/
    n_chars = t_strlen( string );
    cell_size._x = STROKE.font_ptr->font_wid *
        ( 1.0 + STROKE.gap / STROKE.aspect );
    cell_size._y = STROKE.font_ptr->font_hei;
    switch( STROKE.align.horiz )
      {
        case 'r' :  offset._x = cell_size._x * n_chars;
                    break;
        case 'c' :  offset._x = cell_size._x * n_chars / 2.0;
                    break;
        case 'l' :  offset._x = 0.0;
                    break;
      }
    switch( STROKE.align.vert )
      {
        case 't' :  offset._y = cell_size._y;
                    break;
        case 'c' :  offset._y = cell_size._y / 2.0;
                    break;
        case 'b' :  offset._y = 0.0;
                    break;
      }
/*
 *    determine start position
*/
    point._x = TER.requir.cursor._x - STROKE.map[0][0] * offset._x -
               STROKE.map[0][1] * offset._y;
    point._y = TER.requir.cursor._y - STROKE.map[1][0] * offset._x -
               STROKE.map[1][1] * offset._y;
    cell_size._y = STROKE.map[1][0] * cell_size._x;
    cell_size._x = STROKE.map[0][0] * cell_size._x;
/*
 *      DRAW EACH CHARACTER IN THE STRING
*/
    for( ; *string; string++, char_count++ )
      {
        font_index = *string - 32;
        if ( 0 <= font_index && font_index < STROKE.font_ptr->font_siz )
            if ( (STROKE.font_ptr->font_dir)[font_index].font_char == *string &&
                 (STROKE.font_ptr->font_dir)[font_index].font_npts > 0 )
              {
                ok &= bold_font( (Text_Attribute VAR_PTR) &STROKE,
                                 (coord_pair VAR_PTR) &point,
                                 (char VAR_PTR) &font_index, t_stkchr );
                point._x += cell_size._x;
                point._y += cell_size._y;
              }
            else
                (void) make_error( ERR_WARN, name,
                    (char VAR_PTR) "character unavailable" );
         else
             (void) make_error( ERR_WARN, name,
                 (char VAR_PTR) "not printable character" );
      }
    TER.requir.cursor = point;
    if ( ! ( ok &= move_clip( (coord_triple VAR_PTR) &TER.requir.pen,
            VIEW_CLIP ) ) )
        (void) make_error( ERR_CALL, name, mt_str );
    return char_count;
  }


static char VAR_PTR t_ch_def( what, character, index, point )
/*
 * This function allows the user to obtain the shape of any character defined
 * in the ASCII and SYMBOL fonts. With the exception of string alignment, the
 * transformations implied by t_txtset are applied to the returned coordinates.
 * This function is merely a call to the function ``define_font'' with an
 * additional argument defining which set of text attributes to use .
*/
int what;           /* specifies what to do (T_FT_* defined values) */
char character;     /* character to fetch from the font */
int index;          /* coordinate number   */
coord_pair
    VAR_PTR point;  /* returned coordinate */
  {
    START_UP( (char VAR_PTR) "t_ch_def", char VAR_PTR );
    return define_font( (Text_Attribute VAR_PTR) &STROKE, PEN_THICKNESS,
        what, character, index, point );
  }


static char VAR_PTR t_txtget( string, max_chars )
/*
 *    Reads a maximum of `max_chars' characters from `stdin' into `string'
 * until a newline character is read . The newline character is not stored in
 * `string' . `string' is terminated with NUL . If not in alpha-numeric or
 * alpha-graphic mode, the terminal is put into alpha-graphic mode . The
 * characters are echoed to the screen as they are typed . If using the
 * alpha-numeric screen, this function simply gets() . If using the graphics
 * screen this function is again like gets() except that the echoed character
 * is confined to the viewport . If the cursor is initially outside the
 * viewport, t_txtget() starts echoing characters at the top left corner of
 * the viewport. If an error errors or no characters are read, a NULL pointer
 * is returned, otherwise `string' is returned .
*/
char VAR_PTR string;
int max_chars;
  {
    char VAR_PTR name = "t_txtget";
    boolean ok = TRUE;
    char VAR_PTR ptr;
    CATCH_NULL( string, name, char VAR_PTR );
    START_UP( name, char VAR_PTR );
    if ( TER.actual.mode != ALPH_NUMERIC )
      {
/*
 *              ensure cursor is in the viewport
*/
        if ( TER.requir.cursor._x < min_clip(VIEW_CLIP,_x) ||
             TER.requir.cursor._x > max_clip(VIEW_CLIP,_x) - ALPH_WIDTH + 1 ||
             TER.requir.cursor._y < min_clip(VIEW_CLIP,_y) ||
             TER.requir.cursor._y > max_clip(VIEW_CLIP,_y) - ALPH_HEIGHT + 1 )
          {
            TER.requir.cursor._x = min_clip(VIEW_CLIP,_x);
            TER.requir.cursor._y = max_clip(VIEW_CLIP,_y) - ALPH_HEIGHT + 1;
          }
/*
 *              if necessary, correctly position the cursor
*/
        if ( fabs( TER.requir.cursor._x - TER.actual.cursor._x ) > 0.5 ||
             fabs( TER.requir.cursor._y - TER.actual.cursor._y ) > 0.5 )
          {
            if ( TER.actual.mode != VECTOR_MODE )
              {
                (void) G_PUTC( GS );
                if ( ok &= ! G_ERROR() )
                    TER.actual.mode = VECTOR_MODE;
              }
            (void) G_PUTC( GS );
            ok &= ! G_ERROR();
            if ( ok &= t4010_vec( (int) ( TER.requir.cursor._x + 0.5 ),
                                  (int) ( TER.requir.cursor._y + 0.5 ) ) )
                TER.actual.cursor = TER.requir.cursor;
          }
      }
/*
 *              ensure that terminal is in alphanumeric or alphagraphic mode
*/
    if ( TER.actual.mode != ALPH_GRAPHIC && TER.actual.mode != ALPH_NUMERIC )
      {
        (void) G_PUTC( US );
        if ( ok &= ! G_ERROR() )
            TER.actual.mode = ALPH_GRAPHIC;
      }
    (void) G_FLUSH();
    if ( ! ( ok &= ! G_ERROR() ) )
        (void) make_error( ERR_WARN, name, err_write );
/*
 *              alpha-graphic screen
*/
    if ( TER.actual.mode == ALPH_GRAPHIC )
      {
/*
 *      N.B. Behaviour is unpredictable if characters run off the bottom of
 *           the screen or a control character (non-printable) is typed .
*/
        boolean lost_cursor = FALSE;
        if ( TER.actual.dots != TER.requir.dots )
            ok &= v500_dots( TER.requir.dots );
        for( ptr = string; ( *ptr = getchar() ) != '\n' && max_chars-- && ok;
            ptr++ )
          {
            lost_cursor |= *ptr < ' ' || *ptr > '~';
            if ( ( TER.requir.cursor._x += ALPH_WIDTH ) >
                   v500->scr_size._x - ALPH_WIDTH + 1 )
              {
                TER.actual.cursor._y = TER.requir.cursor._x = 0.0;
                lost_cursor |= ( TER.requir.cursor._y -= ALPH_HEIGHT ) < 0.0;
              }
          }
        TER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
        if ( lost_cursor )
            TER.requir.cursor._y = max_clip( VIEW_CLIP, _y ) - ALPH_HEIGHT + 1;
        else
            TER.requir.cursor._y -= ALPH_HEIGHT;
        (void) G_PUTC( GS );
        if ( ok &= ! G_ERROR() )
            TER.actual.mode = VECTOR_MODE;
        (void) G_PUTC( GS );
        ok &= ! G_ERROR();
        if ( ok &= t4010_vec( (int) ( TER.requir.cursor._x + 0.5 ),
                              (int) ( TER.requir.cursor._y + 0.5 ) ) )
            TER.actual.pen = TER.actual.cursor = TER.requir.cursor;
        else
            (void) make_error( ERR_WARN, name, err_write );
      }
    else
/*
 *              alpha-numeric screen
*/
      for( ptr = string; ( *ptr = getchar() ) != '\n' && max_chars--; ptr++ );
/*
 *              terminate string & return
*/
    *ptr = NUL;
    return ok && ptr != string ? string : (char VAR_PTR) 0;
  }


static int t_txtput( string )
/*
 *    Put a text string `string' on the GRAPHICS screen, starting at the place
 *  last specified by `t_txtmov' or 't_txtcur', using the colour specified by
 *  `t_colour' ( which can only be bright or dark ) as foreground and the data
 *  level specified by `t_visib', `t_invis', `t_revers' or `t_replac' . Only
 *  those characters within the viewport are written to the screen. The
 *  returned value is the number of characters written to the screen, or zero
 *  if an error has occurred .
 *
 *  The character string `string' may contain printable characters and the
 *  control codes,
 *    HT : \t : \011 : move to the right by one character space
 *    BS : \b : \010 : move to the left by one character space
 *    CR : \r : \015 : move to the beginning of the current line
 *    LF : \n : \012 : move down one character space
 *    VT :    : \013 : move up one character space
 *    EM :    : \031 : move to top left hand side of the screen
 *  Any other characters are ignored .
 *
 *    DOT_ON,  write foreground colour text onto the existing screen,
 *    DOT_OFF, write a dark background onto the existing screen,
 *    DOT_REV, change the background into foreground colour, write dark text,
 *    DOT_NEW, foreground colour text is written onto a dark background.
 *
 *  This function only breaks the string into segments terminated by cursor
 *  control characters and postions the required_cursor appropriately before
 *  calling t_txtstr to put the characters on the screen.
*/
char VAR_PTR string;         /* pointer to start of the character string */
  {
    char VAR_PTR name = "t_txtput";
    char VAR_PTR ptr;
    int  char_count = 0;     /* number of characters printed             */
    CATCH_NULL( string, name, boolean );
    START_UP( name, boolean );
    for ( ptr = string; *ptr; ptr++ )
      {
        if ( *ptr == LF || *ptr == VT || *ptr == EM ||
             *ptr == CR || *ptr == BS || *ptr == HT )
          {
            char_count += t_txtstr( string, ptr ) + 1;
            switch( *ptr )
              {
                case LF :   TER.requir.cursor._y += -ALPH_HEIGHT;
                            break;
                case VT :   TER.requir.cursor._y += ALPH_HEIGHT;
                            break;
                case EM :   TER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
                            TER.requir.cursor._y = max_clip( VIEW_CLIP, _y ) -
                                  ALPH_HEIGHT + 1;
                            break;
                case CR :   TER.requir.cursor._x = min_clip( VIEW_CLIP, _x );
                            break;
                case BS :   TER.requir.cursor._x += -ALPH_WIDTH;
                            break;
                case HT :   TER.requir.cursor._x += ALPH_WIDTH;
                            break;
              }
            string = ptr + 1;
          }
      }
    if ( string != ptr )
        char_count += t_txtstr( string, ptr );
    return char_count;
  }


static boolean t_boxfil( lower_left, upper_right, style )
/*
 *      draw a rectangular box on the screen with lower left corner at
 * `*lower_left' and upper right corner at `*upper_right', and fill the box
 * with a hatching of type `style' . The hatching styles are :
 *
 *      0 = empty, box outline only      5 = horizontal lines
 *      1 = fill the box completely      6 = vertical lines
 *      2 = 50% `grey' fill              7 = diagonal cross hatch
 *      3 = NW-SE diagonal lines         8 = horizontal-vertical cross
 *      4 = NE-SW diagonal lines                 hatch
 * The box outline is clipped to the viewport limits .
*/
coord_pair VAR_PTR lower_left,  /* lower left corner of the box     */
           VAR_PTR upper_right; /* upper right corner of the box    */
int style;                      /* hatching style                   */
  {
    char VAR_PTR name = "t_boxfil";
    coord_pair ll_corner, ur_corner;
    CATCH_NULL( lower_left && upper_right, name, boolean );
    START_UP( name, boolean );
    ll_corner = *lower_left;
    ur_corner = *upper_right;
/*
 *      ensure correct ordering of coordinates
*/
    if ( ll_corner._x > ur_corner._x )
        SWAPD( ll_corner._x , ur_corner._x );
    if ( ll_corner._y > ur_corner._y )
        SWAPD( ll_corner._y , ur_corner._y );
/*
 *      clip the rectangle to the screen limits
*/
    ll_corner._x = LIMIT( ll_corner._x,
                     min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
    ll_corner._y = LIMIT( ll_corner._y,
                     min_clip( VIEW_CLIP, _y ), max_clip( VIEW_CLIP, _y ) );
    ur_corner._x = LIMIT( ur_corner._x,
                     min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
    ur_corner._y = LIMIT( ur_corner._y,
                     min_clip( VIEW_CLIP, _y ), max_clip( VIEW_CLIP, _y ) );
/*
 *      draw the box
*/
    if ( ur_corner._x - ll_corner._x > 0.0 &&
         ur_corner._y - ll_corner._y > 0.0 )
      {
        boolean ok = TRUE;
        if ( TER.actual.mode == ALPH_NUMERIC &&
               v500->type == XTERM_VT100TEK4014 )
            (void) G_PUTS( "\033[?38h" );
        if ( TER.actual.mode != VECTOR_MODE && TER.actual.mode != ALPH_GRAPHIC)
          {
            (void) G_PUTC( GS );
            if ( ( ok = G_PUTC( GS ) ) )
                TER.actual.mode = VECTOR_MODE;
          }
        if ( ( TER.requir.dots == DOT_ON || TER.requir.dots == DOT_NEW ) &&
                ! TER.requir.colour )
          {
            if ( TER.actual.dots != DOT_OFF )
                ok &= v500_dots( DOT_OFF );
            if ( TER.requir.dots == DOT_NEW )
                style = 1;
          }
        else if ( TER.actual.dots != TER.requir.dots )
            ok &= v500_dots( TER.requir.dots );
        ur_corner._x -= ll_corner._x - 1;
        ur_corner._y -= ll_corner._y - 1;
        if ( TER.actual.dots == DOT_REV )
          {
            ok &= v500_box( (coord_pair VAR_PTR) &ll_corner,
                            (coord_pair VAR_PTR) &ur_corner, 0 );
            ll_corner._x += 1.0;
            ll_corner._y += 1.0;
            ur_corner._x -= 2.0;
            ur_corner._y -= 2.0;
            ok &= v500_box( (coord_pair VAR_PTR) &ll_corner,
                            (coord_pair VAR_PTR) &ur_corner, 0 );
          }
        ok &= v500_box( (coord_pair VAR_PTR) &ll_corner,
            (coord_pair VAR_PTR) &ur_corner, ( abs( style ) - 1 ) % 8 + 1 );
        if ( ! ok )
            (void) make_error( ERR_WARN, name, err_write );
        return ok;
      }
    return FALSE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Cross Hairs Routines                           */
/*                                                                          */
/* ************************************************************************ */


static boolean t_xhrget( point, char_pressed )
/*
 *      Get cross-hair location . If the terminal is not already in cross-hairs
 *  mode, then it is put into cross-hairs mode . The cross-hairs are positioned
 *  by the user, and the location of the cross-hairs is transmitted when an
 *  alpha-numeric key is pressed . The returned value of `t_xhrget' is TRUE if
 *  `point' is within the current viewport and `t_xhrget' can write to the
 *  graphics device ; otherwise the returned value is FALSE .
 *
 *    Arguments
*/
coord_pair VAR_PTR point;   /* pointer to the coordinates of the cross-hairs */
char VAR_PTR char_pressed;  /* character pressed to identify `point'         */
 {
    char VAR_PTR name = "t_xhrget";
    int high_x, high_y;
    boolean ok = TRUE;
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    if ( v500->t_op_dev != stdout )
        return make_error( ERR_ABORT, name,
            (char VAR_PTR) "Output is not to stdout" );
/*
 *      put the terminal in cross-hairs mode . i.e. switch on cross-hairs
 *      and wait for the user to press a key
*/
    if ( TER.actual.mode != CROSS_HAIR )
      {
        if ( TER.actual.mode == ALPH_NUMERIC &&
               v500->type == XTERM_VT100TEK4014 )
            (void) G_PUTS( "\033[?38h" );
        if ( TER.actual.mode != VECTOR_MODE )
          {
            (void) G_PUTC( GS );
            if ( ( ok = ! G_ERROR() ) )
               TER.actual.mode = VECTOR_MODE;
          }
        (void) G_PUTS( "\033\032" );
        if ( ok &= ! G_ERROR() )
            TER.actual.mode = CROSS_HAIR;
        (void) G_FLUSH();
        if ( ! ( ok &= ! G_ERROR() ) )
            (void) make_error( ERR_WARN, name, err_write );
      }
/*
 *      read the character key pressed . If it is a '\n', put the terminal
 *      back into vector mode immediately ( the echoed '\n' puts it into
 *      alphagraphics mode ) so that rubbish does not appear on the graphics
 *      screen .
*/
    *char_pressed = getchar();
    if ( *char_pressed == '\n' )
         (void) G_PUTC( GS );
/*
 *      read and decode the location of the crosshairs
*/
    high_x = getchar();
    point->_x = ( ( 31 & high_x ) << 5 ) + ( 31 & (int) getchar() );
    high_y = getchar();
    point->_y = ( ( 31 & high_y ) << 5 ) + ( 31 & (int) getchar() );
    TER.actual.mode = ALPH_GRAPHIC;
    v500->last_draw = FALSE;
/*
 *      flush the trailer codes
*/
    (void) getchar();
    return ok &&
           min_clip( VIEW_CLIP, _x ) <= point->_x &&
           max_clip( VIEW_CLIP, _x ) >= point->_x &&
           min_clip( VIEW_CLIP, _y ) <= point->_y &&
           max_clip( VIEW_CLIP, _y ) >= point->_y;
 }


static boolean t_xhrput( point )
/*
 *      Put the  cross-hairs at `point' . If the terminal is not already in
 *  cross-hairs mode, then it is put into cross-hairs mode . The returned value
 *  of `t_xhrput' is TRUE if `point' is within the current viewport and
 *  `t_xhrput' can write to the graphics device ; otherwise the returned value
 *  is FALSE .
 *
 *    Arguments
*/
coord_pair VAR_PTR point;   /* pointer to the coordinates of the cross-hairs */
 {
    char VAR_PTR name = "t_xhrput";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    if ( -0.5 < point->_x  &&  point->_x <= V500_X + 0.5 &&
         -0.5 < point->_y  &&  point->_y <= V500_Y + 0.5 )
      {
/*
 *        either put into vector mode, or prepare to draw dark vector
*/
        boolean ok;                     /* success flag                     */
        if ( TER.actual.mode == ALPH_NUMERIC &&
               v500->type == XTERM_VT100TEK4014 )
            (void) G_PUTS( "\033[?38h" );
        (void) G_PUTC( GS );
        ok = ! G_ERROR();
        TER.actual.mode = VECTOR_MODE;
/*
 *        send `point', load cross-hair & put into cross-hair mode
*/
        ok &= t4010_vec( (int) ( point->_x + 0.5 ), (int) ( point->_y + 0.5 ) );
        (void) G_PUTS( "\033/f\033\032" );
        if ( ! ( ok &= ! G_ERROR() ) )
            (void) make_error( ERR_WARN, name, err_write );
        return ok &&
           min_clip( VIEW_CLIP, _x ) <= point->_x &&
           max_clip( VIEW_CLIP, _x ) >= point->_x &&
           min_clip( VIEW_CLIP, _y ) <= point->_y &&
           max_clip( VIEW_CLIP, _y ) >= point->_y;
      }
    return FALSE;
 }


/* ************************************************************************ */
/*                                                                          */
/*                        Print terminal status                             */
/*                                                                          */
/* ************************************************************************ */


static char VAR_PTR t_status( attribute, value )
/*
 *     Put status attribute `attribute' into the coordinate pair `*value' .
*/
int attribute;
char VAR_PTR value;
  {
    char VAR_PTR name = "t_status";
    CATCH_NULL( value, name, char VAR_PTR );
    if ( attribute == T_ST_CLIP   || attribute == T_ST_A_CURS ||
         attribute == T_ST_R_CURS || attribute == T_ST_A_PEN ||
         attribute == T_ST_R_PEN )
        START_UP( name, char VAR_PTR );
    switch( attribute )
      {
        case T_ST_DEVICE:
            if ( v500->type == VISUAL500_TERMINAL )
                (void) t_strcpy( value, (char VAR_PTR) "Visual 500 terminal" );
            else if ( v500->type == VISUAL550_TERMINAL )
                (void) t_strcpy( value, (char VAR_PTR) "Visual 550 terminal" );
            else if ( v500->type == XTERM_VT100TEK4014 )
                (void) t_strcpy( value, (char VAR_PTR) "xterm- Tek4014 mode" );
            else
                return (char VAR_PTR) make_error( ERR_ABORT, name,
                    (char VAR_PTR) "BUG: unknown device type" );
            break;
        case T_ST_SCREEN:
            *( (boolean VAR_PTR) value ) = TRUE;
            break;
        case T_ST_CATEG:
            *( (int VAR_PTR) value ) = T_ST_FILE;
            break;
        case T_ST_MONO:
            *( (boolean VAR_PTR ) value ) = TRUE;
            break;
        case T_ST_CLIP:
            *( (clip_area VAR_PTR VAR_PTR) value ) = VIEW_CLIP;
            break;
        case T_ST_MAX:
            ( (coord_pair VAR_PTR) value)->_x = v500->scr_size._x;
            ( (coord_pair VAR_PTR) value)->_y = v500->scr_size._y;
            break;
        case T_ST_MIN:
            ( (coord_pair VAR_PTR) value)->_x = 0.0;
            ( (coord_pair VAR_PTR) value)->_y = 0.0;
            break;
        case T_ST_CHAR:
            ( (coord_pair VAR_PTR) value)->_x = ALPH_WIDTH;
            ( (coord_pair VAR_PTR) value)->_y = ALPH_HEIGHT;
            break;
        case T_ST_LINES:
            ( (coord_pair VAR_PTR) value)->_x = ALPH_COLUMNS;
            ( (coord_pair VAR_PTR) value)->_y = ALPH_LINES;
            break;
        case T_ST_PIXSIZ:
            ( (coord_pair VAR_PTR) value)->_x =
                 ( (coord_pair VAR_PTR) value)->_y = PIXEL_SIZE;
            break;
        case T_ST_A_CURS:
            ( (coord_pair VAR_PTR) value)->_x = TER.actual.cursor._x;
            ( (coord_pair VAR_PTR) value)->_y = TER.actual.cursor._y;
            break;
        case T_ST_R_CURS:
            ( (coord_pair VAR_PTR) value)->_x = TER.requir.cursor._x;
            ( (coord_pair VAR_PTR) value)->_y = TER.requir.cursor._y;
            break;
        case T_ST_A_PEN:
            ( (coord_pair VAR_PTR) value)->_x = TER.actual.pen._x;
            ( (coord_pair VAR_PTR) value)->_y = TER.actual.pen._y;
            break;
        case T_ST_R_PEN:
            ( (coord_pair VAR_PTR) value)->_x = TER.requir.pen._x;
            ( (coord_pair VAR_PTR) value)->_y = TER.requir.pen._y;
            break;
        default:
            return (char VAR_PTR) make_error( ERR_ABORT, name,
                (char VAR_PTR) "unknown attribute" );
      }
    return value;
  }


static boolean t_pr_ter( out_file )
FILE *out_file;
  {
    char VAR_PTR name = "t_pr_ter";
    CATCH_NULL( out_file, name, boolean );
    START_UP( name, boolean );
    (void) fprintf( out_file, " TERMINAL STATUS\n" );
    (void) fprintf( out_file,
                    "               ---requir---     ---actual---\n" );
    (void) fprintf( out_file, "data level  %15s  %15s\n",
                    pr_dots( (Ter_Status VAR_PTR) &TER.requir ),
                    pr_dots( (Ter_Status VAR_PTR) &TER.actual ) );
    (void) fprintf( out_file, "mode        %15s  %15s\n",
                    pr_mode( (Ter_Status VAR_PTR) &TER.requir ),
                    pr_mode( (Ter_Status VAR_PTR) &TER.actual ) );
    (void) fprintf( out_file, "colour      %15d  %15d\n",
                    TER.requir.colour, TER.actual.colour );
    (void) fprintf( out_file, "dash style  %15d  %15d\n",
                    TER.requir.dash, TER.actual.dash );
    (void) fprintf( out_file, "cursor      (%6.1f,%6.1f)  (%6.1f,%6.1f)\n",
                    TER.requir.cursor._x, TER.requir.cursor._y,
                    TER.actual.cursor._x, TER.actual.cursor._y );
    (void) fprintf( out_file, "pen         (%6.1f,%6.1f)  (%6.1f,%6.1f)\n",
                    TER.requir.pen._x, TER.requir.pen._y,
                    TER.actual.pen._x, TER.actual.pen._y );
    if ( ferror( out_file ) )
        return make_error( ERR_WARN, name,
            (char VAR_PTR) "Cannot write terminal status to file" );
    return TRUE;
  }


/* ************************************************************************ */
/*                                                                          */
/*                           Vector Drawing                                 */
/*                                                                          */
/* ************************************************************************ */

static boolean t_vecmov( point )
/*
 *      move the line drawing pen to `point'
*/
coord_pair VAR_PTR point;
  {
    char VAR_PTR name = "t_vecmov";
    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    v500->last_draw = FALSE;
    TER.requir.pen = *point;
    return move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
  }


static boolean t_vecdrw( point )
/*
 *      Draw from the previous pen position to `point' according to the
 *  required pen colour, dashed line style and dots-level . Dashed line style
 *  takes precedence over colours . Lines are clipped to the viewport
 *  boundaries .
*/
coord_pair VAR_PTR point;
  {
    char VAR_PTR name = "t_vecdrw";
    boolean ok = TRUE;      /* success flag */

    CATCH_NULL( point, name, boolean );
    START_UP( name, boolean );
    TER.requir.pen = *point;
/*
 *          If required dashed line style is 0, put terminal into POINT_PLOT
 *          mode, and plot the point if it is in the viewport .
*/
    if ( ! TER.requir.dash )
      {
        ok &= move_clip( (coord_triple VAR_PTR) point, VIEW_CLIP );
        if ( min_clip( VIEW_CLIP, _x ) <= point->_x &&
             max_clip( VIEW_CLIP, _x ) >= point->_x &&
             min_clip( VIEW_CLIP, _y ) <= point->_y &&
             max_clip( VIEW_CLIP, _y ) >= point->_y )
          {
            if ( TER.actual.mode == ALPH_NUMERIC &&
                   v500->type == XTERM_VT100TEK4014 )
                (void) G_PUTS( "\033[?38h" );
            if ( TER.actual.mode != POINT_PLOT )
              {
                (void) G_PUTC( FS );
                if ( ok &= ! G_ERROR() )
                    TER.actual.mode = POINT_PLOT;
              }
/*
 *          If required colour is black, put terminal data level into DOT_OFF,
 *          otherwise use the required data level .
*/
            if ( ! TER.requir.colour )
              {
                if ( TER.actual.dots != DOT_OFF )
                    ok &= v500_dots( DOT_OFF );
              }
            else if ( TER.actual.dots != TER.requir.dots )
                ok &= v500_dots( TER.requir.dots );
/*
 *          plot the point
*/
            TER.actual.cursor._x = TER.actual.pen._x = point->_x;
            TER.actual.cursor._y = TER.actual.pen._y = point->_y;
            ok &= t4010_vec( (int) ( point->_x + 0.5 ),
                             (int) ( point->_y + 0.5 ) );
            v500->last_draw = TRUE;
          }
      }
/*
 *          Need to draw a line . First clip the endpoints, and put the
 *          terminal into VECTOR_MODE .
*/
    else if ( clip( (coord_triple *) point, VIEW_CLIP ) )
      {
        if ( TER.actual.mode == ALPH_NUMERIC &&
               v500->type == XTERM_VT100TEK4014 )
            (void) G_PUTS( "\033[?38h" );
        if ( TER.actual.mode != VECTOR_MODE )
          {
            (void) G_PUTC( GS );
            if ( ok &= ! G_ERROR() )
                TER.actual.mode = VECTOR_MODE;
            v500->last_draw = FALSE;
          }
/*
 *          If required colour is black, put terminal data level into DOT_OFF,
 *          otherwise use the required data level .
*/
        if ( ! TER.requir.colour )
          {
            if ( TER.actual.dots != DOT_OFF )
              ok &= v500_dots( DOT_OFF );
          }
        else if ( TER.actual.dots != TER.requir.dots )
            ok &= v500_dots( TER.requir.dots );
/*
 *          If a continuous line is required, use the dashed line style as
 *          pseudo-colour, otherwise just set the required dashed line style
*/
        if ( TER.requir.dash == 1 && TER.requir.colour )
          {
            if ( TER.requir.colour != TER.actual.dash )
              {
                if ( ok &= v500_style( TER.requir.colour ) )
                    TER.actual.dash = TER.requir.colour;
                v500->last_draw = FALSE;
              }
          }
        else if ( TER.actual.dash != TER.requir.dash )
          {
            if ( ok &= v500_style( TER.requir.dash ) )
                TER.actual.dash = TER.requir.dash;
            v500->last_draw = FALSE;
          }
/*
 *          Now we are all set up, we can worry about drawing the vector
 *            First check if the vector starts from the end of the last one .
*/
        v500->last_draw &= fabs( TER.actual.pen._x - first_clip(VIEW_CLIP,_x) ) < 0.5
                  && fabs( TER.actual.pen._y - first_clip(VIEW_CLIP,_y) ) < 0.5;
/*
 *          send vector to the Tektronix 4010 terminal
*/
        if ( ! v500->last_draw )
          {
            (void) G_PUTC( GS );
            ok &= ! G_ERROR();
            ok &= t4010_vec( (int) ( first_clip( VIEW_CLIP, _x ) + 0.5 ),
                             (int) ( first_clip( VIEW_CLIP, _y ) + 0.5 ) );
          }
        TER.actual.cursor._x = TER.actual.pen._x = last_clip( VIEW_CLIP, _x );
        TER.actual.cursor._y = TER.actual.pen._y = last_clip( VIEW_CLIP, _y );
        ok &= t4010_vec( (int) ( last_clip( VIEW_CLIP, _x ) + 0.5 ),
                         (int) ( last_clip( VIEW_CLIP, _y ) + 0.5 ) );
        v500->last_draw = TRUE;
      }
    if ( ! ok )
        (void) make_error( ERR_WARN, name, err_write );
    return ok;
  }


/* ************************************************************************ */
/* ************************************************************************ */
/*                                                                          */
/*                    Array of pointers to driver functions                 */
/*                                                                          */
/* With the exception of the function ``t_driver'' which is used to select  */
/* the required driver, this array is the ONLY point of access to the       */
/* functions in the driver .                                                */
/*                                                                          */
/* ************************************************************************ */
/* ************************************************************************ */

/*
 * static clip_area VAR_PTR t_begin();
 * static boolean
 *     t_end(),    t_vwport(), t_colour(), t_dashed(), t_thick(),  t_gr_off(),
 *     t_gr_on(),  t_pr_ter(), t_dump(),   t_visib(),  t_invis(),  t_revers(),
 *     t_replac(), t_newpag(), t_newscr(), t_bell(),   t_wait(),   t_txtcur(),
 *     t_txtmov(), t_txtset(), t_symdrw(), t_xhrget(), t_xhrput(), t_vecdrw(),
 *     t_boxfil(), t_vecmov();
 * static char
 *         VAR_PTR t_txtget(), VAR_PTR t_status(), VAR_PTR t_ch_def();
 * static int t_txtput(), t_txtdrw();
 */

static boolean t_dummy()
  {
    return TRUE;
  }


T_dev_fns V500_dev_ptrs = {
            v500_begin,  t_end,    t_vwport, t_gr_on,  t_gr_off,
            t_colour, t_dashed, t_thick,  t_visib,  t_invis,
            t_revers, t_replac, t_newpag, vis500_newscr, t_bell,
            t_wait,   t_txtmov, t_txtcur, t_txtset, t_symdrw,
            t_txtdrw, t_txtget, t_txtput, t_status, t_pr_ter,
            t_vecmov, t_vecdrw, t_dump,   t_xhrput, t_xhrget,
            t_boxfil, t_ch_def };

T_dev_fns V550_dev_ptrs = {
            v550_begin,  t_end,    t_vwport, t_gr_on,  t_gr_off,
            t_colour, t_dashed, t_thick,  t_visib,  t_invis,
            t_revers, t_replac, t_newpag, vis550_newscr, t_bell,
            t_wait,   t_txtmov, t_txtcur, t_txtset, t_symdrw,
            t_txtdrw, t_txtget, t_txtput, t_status, t_pr_ter,
            t_vecmov, t_vecdrw, t_dump,   t_xhrput, t_xhrget,
            t_boxfil, t_ch_def };

T_dev_fns Xterm_dev_ptrs = {
            xterm_begin,  t_end,    t_vwport, t_gr_on,  t_gr_off,
            t_colour, t_dashed, t_thick,  t_dummy,  t_dummy,
            t_dummy, t_dummy, t_newpag, xterm_newscr, t_bell,
            t_wait,   t_txtmov, t_txtcur, t_txtset, t_symdrw,
            t_txtdrw, t_txtget, t_txtput, t_status, t_pr_ter,
            t_vecmov, t_vecdrw, t_dump,   t_dummy, t_xhrget,
            t_dummy, t_ch_def };


/* ************************************************************************ */
/*                                                                          */
/*                      Utilities for v500_g.c                              */
/*      These functions cannot be called directly by the user               */
/*                                                                          */
/* ************************************************************************ */


static boolean t_stkchr( index, end_nul, point )
/*
 *      Draws character identified by `index', at location `point' .
 *      This is a single pass -- i.e. no boldface.
*/
char VAR_PTR index,
     VAR_PTR end_nul;      /* phantom end of string pointer - required */
                           /* compatibility with bold_font & t_txtstr  */
coord_pair VAR_PTR point;
  {
    int point_count;            /* counts the points plotted for each char  */
    boolean ok = TRUE,          /* success flag                             */
            pen_up;             /* move with pen up                         */
    float VAR_PTR coord_ptr;    /* pointer to font coordinate values        */
    coord_pair coord;           /* coordinate from the font - to plot       */
    double temp;
    StrokeChar VAR_PTR char_ptr =
        STROKE.font_ptr->font_dir + *( (int VAR_PTR) index );
/*
 *      step through the coordinates which define the character
*/
    coord_ptr =  char_ptr->font_coord;
    for( point_count = 0;
         ok && point_count < char_ptr->font_npts;
         point_count++ )
      {
/*
 *      calculate stroke coordinate position in pixel units
*/
        pen_up = ! point_count;
        coord._x = *coord_ptr++;
        coord._y = *coord_ptr++;
        if ( coord._x < 0.0 )
          {
            pen_up = TRUE;
            coord._x = -coord._x;
          }
        if ( coord._y < 0.0 )
          {
            pen_up = TRUE;
            coord._y = -coord._y;
          }
        coord._y -= char_ptr->font_shift;
        temp = STROKE.map[0][0] * coord._x + STROKE.map[0][1] * coord._y
                + point->_x;
        coord._y = STROKE.map[1][0] * coord._x + STROKE.map[1][1] * coord._y
                + point->_y;
        coord._x = temp;
/*
 *      plot the point
*/
        if ( pen_up )
          {
            v500->last_draw = FALSE;
            ok = move_clip( (coord_triple VAR_PTR) &coord, VIEW_CLIP );
          }
        else if ( clip( (coord_triple VAR_PTR) &coord, VIEW_CLIP ) )
          {
/*
 *      First check if the vector starts from the end of the last one .
*/
            v500->last_draw &= fabs( TER.actual.pen._x -
                               first_clip( VIEW_CLIP, _x ) ) < 0.5
                      && fabs( TER.actual.pen._y -
                               first_clip( VIEW_CLIP, _y ) ) < 0.5;
/*
 *      send vector to the Tektronix 4010 terminal
*/
            if ( ! v500->last_draw )
              {
                (void) G_PUTC( GS );
                ok &= ! G_ERROR();
                ok &= t4010_vec( (int) ( first_clip(VIEW_CLIP,_x) + 0.5 ),
                                 (int) ( first_clip(VIEW_CLIP,_y) + 0.5 ) );
              }
            TER.actual.cursor._x = TER.actual.pen._x
                                 = last_clip( VIEW_CLIP, _x );
            TER.actual.cursor._y = TER.actual.pen._y
                                 = last_clip( VIEW_CLIP, _y );
            ok &= t4010_vec( (int) ( last_clip( VIEW_CLIP, _x ) + 0.5 ),
                             (int) ( last_clip( VIEW_CLIP, _y ) + 0.5 ) );
            v500->last_draw = TRUE;
          }
      }
    return ok;
  }


static int t_txtstr( string, end_nul )
/*
 * Puts a string of printable ASCII characters onto the screen, starting at the
 * current required cursor postion and using the current colour and data level.
 * Only characters within the viewport are written to the screen . The returned
 * value is the number of characters written.
 * The character string `string' may be NUL terminated or, if only a fragment
 * of the string is to be printed, `end_nul' is a pointer to where the NUL
 * should be.
 *
*/
char VAR_PTR string,    /* start address of character string to be printed */
     VAR_PTR end_nul;   /* address of phantom end of string NUL */
  {
    coord_pair start_box;              /* lower left & size of text box      */
/*
 *        enable/disable text and boxes and set data level switchs
*/
    boolean ok = TRUE,                 /* success flag                       */
            txt_space = FALSE,         /* write spaces rather than text      */
            destroy_space = FALSE;     /* write spaces destructively         */
    int char_count = 0;                /* number of characters sent          */
    enum dev_dots txt_dots = TER.requir.dots;

    if ( ( destroy_space = TER.requir.dots == DOT_REV ||
                           TER.requir.dots == DOT_NEW ) )
      {
        txt_dots = DOT_NEW;
        txt_space = ! TER.requir.colour;
/*
 *        write spaces destructively
*/
        (void) G_PUTS( "\035\033/2l" );
        if ( ( ok = ! G_ERROR() ) )
            TER.actual.mode = VECTOR_MODE;
/*
 *        initialize box location and size
*/
        start_box._x = LIMIT( TER.requir.cursor._x,
                       min_clip( VIEW_CLIP, _x ), max_clip( VIEW_CLIP, _x ) );
        start_box._y = TER.requir.cursor._y;
      }
    else if ( TER.requir.dots == DOT_ON && ! TER.requir.colour )
        txt_dots = DOT_OFF;
/*
 *        start the real work here
*/
    for( ; *string && string != end_nul && ok; string++ )
      {
        if ( ' ' <= *string && *string <= '~' )
          {
/*
 ********     put character on the screen if it is in the viewport
*/
            if ( min_clip(VIEW_CLIP,_x)              <= TER.requir.cursor._x &&
                 max_clip(VIEW_CLIP,_x)-ALPH_WIDTH+1 >= TER.requir.cursor._x &&
                 min_clip(VIEW_CLIP,_y)              <= TER.requir.cursor._y &&
                 max_clip(VIEW_CLIP,_y)-ALPH_HEIGHT+1 >= TER.requir.cursor._y )
              {
/*
 *                  write character
*/
                if ( TER.actual.mode == ALPH_NUMERIC &&
                       v500->type == XTERM_VT100TEK4014 )
                    (void) G_PUTS( "\033[?38h" );
                if ( fabs( TER.requir.cursor._x-TER.actual.cursor._x ) > 0.5 ||
                     fabs( TER.requir.cursor._y-TER.actual.cursor._y ) > 0.5 )
                  {
                    if ( TER.actual.mode != VECTOR_MODE )
                      {
                        (void) G_PUTC( GS );
                        if ( ok &= ! G_ERROR() )
                            TER.actual.mode = VECTOR_MODE;
                      }
                    (void) G_PUTC( GS );
                    ok &= ! G_ERROR();
                    if( ok &= t4010_vec( (int) ( TER.requir.cursor._x + 0.5 ),
                            (int) ( TER.requir.cursor._y + 0.5 ) ) )
                        TER.actual.cursor = TER.requir.cursor;
                  }
/*
 *        put into ALPHA_GRAPHICS mode if necessary and then write the character
*/
                if ( TER.actual.mode != ALPH_GRAPHIC )
                    if ( ok &= G_PUTC( US ) )
                        TER.actual.mode = ALPH_GRAPHIC;
                if ( TER.actual.dots != txt_dots )
                    ok &= v500_dots( txt_dots );
                (void) G_PUTC( txt_space ? ' ' : *string );
                if ( ok &= ! G_ERROR() )
                  {
                    char_count++;
                    TER.actual.cursor._x += ALPH_WIDTH;
                  }
              }
            TER.requir.cursor._x += ALPH_WIDTH;
          }
      }
    TER.actual.pen = TER.actual.cursor;
/*
 *                  draw box
*/
    if ( TER.requir.dots == DOT_REV && TER.requir.colour &&
         start_box._y                   >= min_clip( VIEW_CLIP, _y ) &&
         start_box._y + ALPH_HEIGHT - 1 <= max_clip( VIEW_CLIP, _y ) )
      {
        coord_pair size_box;                    /* size of text box     */
        size_box._x = LIMIT( TER.requir.cursor._x, min_clip( VIEW_CLIP, _x ),
                     max_clip( VIEW_CLIP, _x ) ) - start_box._x;
        size_box._y = ALPH_HEIGHT;
        if ( size_box._x >= ALPH_WIDTH )
          {
            if ( TER.actual.mode != ALPH_GRAPHIC &&
                 TER.actual.mode != VECTOR_MODE )
              {
                (void) G_PUTC( GS );
                if ( ( ok = ! G_ERROR() ) )
                    TER.actual.mode = VECTOR_MODE;
              }
            if ( TER.actual.dots != DOT_REV )
                ok &= v500_dots( DOT_REV );
            ok &= v500_box( (coord_pair VAR_PTR) &start_box,
                            (coord_pair VAR_PTR) &size_box, 1 );
/*              kludge due to bug in Visual 500 firmware */
            ok &= v500_box( (coord_pair VAR_PTR) &start_box,
                            (coord_pair VAR_PTR) &size_box, 0 );
            TER.actual.pen = TER.actual.cursor = start_box;
           }
      }
    if ( destroy_space )
      {
        (void) G_PUTS( "\035\033/2h" );
        if ( ok &= ! G_ERROR() )
            TER.actual.mode = VECTOR_MODE;
      }
    if ( ! ok )
        (void) make_error( ERR_WARN, (char VAR_PTR) "t_txtput", err_write );
    v500->last_draw = FALSE;
    return ok ? char_count : 0;
  }


static char VAR_PTR pr_dots( status_ptr )
/*
 *      Returns a verbal interpretation of the data level
*/
Ter_Status VAR_PTR status_ptr;
  {
    switch( status_ptr->dots )
      {
        case DOT_ON  : return "dots on";
        case DOT_OFF : return "dots off";
        case DOT_REV : return "complement dots";
        case DOT_NEW : return "replace dots";
      }
    return "";
  }


static char VAR_PTR pr_mode( status_ptr )
/*
 *      Returns a verbal interpretation of the terminal mode
*/
Ter_Status VAR_PTR status_ptr;
  {
    switch( status_ptr->mode )
      {
        case ALPH_NUMERIC : return "alphanumeric";
        case CROSS_HAIR   : return "cross hair";
        case ALPH_GRAPHIC : return "alphgraphic";
        case VECTOR_MODE  : return "vector";
        case POINT_PLOT   : return "point plot";
      }
    return "";
  }


/* ************************************************************************ */
/*                                                                          */
/*                        Basic output routines                             */
/*                                                                          */
/* ************************************************************************ */


#define   HI_SHIFT    5
#define   NUM_MASK    0x1f
#define   HI_Y_TAG    0x20
#define   LO_Y_TAG    0x60
#define   HI_X_TAG    0x20
#define   LO_X_TAG    0x40

static boolean t4010_vec( ix, iy )
/*
 *        Sends a vector to the Tektronics 4010 terminal .
*/
int ix, iy;
  {
    (void) G_PUTC( ( ( iy >> HI_SHIFT ) & NUM_MASK ) | HI_Y_TAG );
    (void) G_PUTC( ( iy & NUM_MASK ) | LO_Y_TAG );
    (void) G_PUTC( ( ( ix >> HI_SHIFT ) & NUM_MASK ) | HI_X_TAG );
    (void) G_PUTC( ( ix & NUM_MASK ) | LO_X_TAG );
    return ! G_ERROR();
  }

#undef   HI_SHIFT
#undef   NUM_MASK
#undef   HI_Y_TAG
#undef   LO_Y_TAG
#undef   HI_X_TAG
#undef   LO_X_TAG


static boolean v500_style( style )
/*
 *    set the dashed line style of the Visual 500 terminal
 *    The returned returned result is FALSE if style<1 or style>8,
 *      or characters cannot be sent to graphics device .
*/
int style;        /* dashed line style  */
  {
    static char v500_dash[] = "`acdbxxx";
    if ( 0 < style && style <= 8 )
      {
	if ( v500->type == XTERM_VT100TEK4014 )
          {
            (void) G_PUTC( ESC );
            (void) G_PUTC( v500_dash[ ( style - 1 ) % 5 ] );
          }
        else /* VISUAL500_TERMINAL, VISUAL550_TERMINAL */
          {
            switch( style )
              {
                case 6 : (void) G_PUTS( "\033/15615a" );
                         break;
                case 7 : (void) G_PUTS( "\033/13119a" );
                         break;
                case 8 : (void) G_PUTS( "\033/9407a" );
              }
            (void) G_PUTC( ESC );
            (void) G_PUTC( v500_dash[ style - 1 ] );
          }
        v500->last_draw = FALSE;
        TER.actual.dash = style;
        return ! G_ERROR();
      }
    return FALSE;
  }


static boolean v500_dots( set_dots )
/*
 *      set the data level of the Visual 500 terminal to `set_dots'
*/
enum dev_dots set_dots;
  {
    boolean ok;
    switch( set_dots )
     {
       case DOT_ON : (void) G_PUTS( "\033/0d" );
                     break;
       case DOT_OFF: (void) G_PUTS( "\033/1d" );
                     break;
       case DOT_REV: (void) G_PUTS( "\033/2d" );
                     break;
       case DOT_NEW: (void) G_PUTS( "\033/3d" );
     }
   if ( ( ok = ! G_ERROR() ) )
       TER.actual.dots = set_dots;
   v500->last_draw = FALSE;
   return ok;
  }


static boolean v500_box( lower_left, size, style )
coord_pair *lower_left,
           *size;
int style;
  {
    static char box_style[] = "@ABCDEFG";
    if ( 0 < style )
      {
        (void) G_PUTC( ESC );
        (void) G_PUTC( box_style[ style - 1 ] );
      }
    (void) fprintf( v500->t_op_dev, "\033/%d;%d;%d;%d",
             (int) ( lower_left->_x + 0.5 ), (int) ( lower_left->_y + 0.5 ),
             (int) size->_x,                 (int) size->_y );
    if ( style )
        (void) G_PUTC( 'y' );
    else
        (void) G_PUTC( 'x' );
    TER.actual.cursor = *lower_left;
    v500->last_draw = FALSE;
    return ! G_ERROR();
  }
