/*
    Copyright (C) 1997  Peter Vernon Lanspeary

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *                      device_g.h
*/
#ifndef DEVICE_G_DEFINED

#ifndef POINTER_DEFINED
#include POINTER_HEADER
#endif

#ifndef STD_TYPE_DEFINED
#include STD_TYPE_HEADER
#endif

#ifndef ERROR_DEFINED
#include ERROR_HEADER
#endif

#ifndef CLIP_DEFINED
#include CLIP_HEADER
#endif



/* ************************************************************************ */
/*                                                                          */
/*               Driving more than one graphics device .                    */
/*                                                                          */
/* The technique is to set up a table of function pointers for each driver .*/
/* This table is defined as an external struct variable in each driver      */
/* source file . A driver is selected by calling the function t_device with */
/* an integer argument specifying the device . Use one of the symbols       */
/* defined below . The default device is the dummy graphics module . The    */
/* driver functions are defined as macros which point to the appropriate    */
/* entry address in the table for the device .                              */
/*                                                                          */
/*                                                                          */
/*               Device selection arguments for t_device                    */
/*                                                                          */
/* ************************************************************************ */

#define NONEXISTENT_DEVICE  0   /* don't use this one, you clown             */
#define VISUAL500_TERMINAL  1   /* Visual 500 terminal                       */
#define HERCULES_GRAPHICS   2   /* IBM PC monochrome (Hercules) graphics     */
#define SUN_3_WORKSTATION   3   /* Sun 3 workstation - using Sunview windows */
#define HEWLETT_GL_PLOTTER  4   /* Hewlett/Packard Graphics Language plotter */
#define POSTSCRIPT_PRINTER  5   /* (Standard) PostScript printer             */
#define DUMMY_GRAPHICS_DEV  6   /* Dummy graphics device - for debugging     */
#define ENHANCED_GRAPHICS   7   /* IBM PC Enhanced graphics adaptor          */
#define X11_WINDOWS         8   /* X Windows Version 11                      */
#define DOT_MATRIX_PRINTER  9
#define VISUAL550_TERMINAL 10   /* Visual 550 terminal                       */
#define XTERM_VT100TEK4014 11   /* xterm tektronix 4014 emulator             */

/*
 *      This can replace the ``options'' argument of ``t_device()'' to prevent
 *      it from starting graphics (by calling ``t_begin()'') .
*/
#define T_SELECT_ONLY   ( (char VAR_PTR) -1L )
#define T_NOT_BEGIN     T_SELECT_ONLY

/* ************************************************************************* */
/*                                                                           */
/*                  Available actions for ``t_register()''                   */
/*                                                                           */
/* ************************************************************************* */

#define T_ISREGIST   0  /* is the device registered ?             */
#define T_REG_FILE  -1  /* return a file pointer or device handle */
#define T_REG_TYPE  -2  /* return the device type                 */
#define T_REG_COUNT -3  /* total number of registered devices     */
#define T_REG_MAXD  16  /* maximum number of simultaneous devices */

/* ************************************************************************* */
/*                                                                           */
/*              Stroke character attributes - used with t_txtset             */
/*                                                                           */
/* ************************************************************************* */

#define  T_CH_RESET   0     /* reset to default */
#define  T_CH_HEIGHT  1     /* height */
#define  T_CH_ASPECT  2     /* aspect ratio = height / width */
#define  T_CH_SLANT   3     /* clockwise slant (or slope) in degrees */
#define  T_CH_ANGLE   4     /* direction of writing in deg. ccw from horiz. */
#define  T_CH_GAP     5     /* gap between characters in character heights  */
#define  T_CH_ALIGN   6     /* specifies string origin */
#define  T_ASCII      7     /* ASCII character font */
#define  T_SYMBOL     8     /* symbol character font */
#define  T_CH_BOLD    9     /* boldface characters */
#define  T_CH_NOBOLD  10    /* standard (not boldface) characters */



/* ************************************************************************* */
/*                                                                           */
/*         Character font definition attributes - used with t_ch_def         */
/*                                                                           */
/* ************************************************************************* */

#define T_FT_SIZE   0       /* width & height of characters */
#define T_FT_EXIST  1       /* definition exists for specified character */
#define T_FT_LENGTH 2       /* number of coordinate pairs in char definition */
#define T_FT_FIRST  3       /* point is the first in a line segment */
#define T_FT_LAST   4       /* point is the last in a line segment */
#define T_FT_MIDDLE 5       /* point is in the middle of a line segment */
#define T_FT_POINT  6       /* coordinate of a point */



/* ************************************************************************* */
/*                                                                           */
/*                  Status attributes - used with t_status                   */
/*                                                                           */
/* ************************************************************************* */

#define  T_ST_DEVICE  1     /* the device name */
#define  T_ST_SCREEN  2     /* is a "screen" device */
#define  T_ST_CATEG   3     /* category of driver output */
#define  T_ST_MONO    4     /* is a monochrome device */
#define  T_ST_CLIP    5     /* pointer to the viewport clip_area */
#define  T_ST_MAX     6     /* upper right hardclip limit */
#define  T_ST_MIN     7     /* lower left hardclip limit */
#define  T_ST_CHAR    8     /* size of a teletype character */
#define  T_ST_LINES   9     /* number of columns and rows of teletype text */
#define  T_ST_PIXSIZ 10     /* width and height of a pixel in mm */
#define  T_ST_A_CURS 11     /* actual cursor location */
#define  T_ST_R_CURS 12     /* required cursor location */
#define  T_ST_A_PEN  13     /* actual pen location */
#define  T_ST_R_PEN  14     /* required pen location */
/* returned values for T_ST_CATEG */
#define  T_ST_FILE   32     /* output is to a file (or stdout) */
#define  T_ST_HARD   64     /* driver talks to dedicated hardware */
#define  T_ST_WIND  128     /* graphics in windowing environment */



/* ************************************************************************* */
/*                                                                           */
/*                           Driver function macros                          */
/*                                                                           */
/* ************************************************************************* */

#define t_begin(out_file,options)   (* device_ptr->t_BEGIN)(out_file,options)
#define t_end()                     (* device_ptr->t_END)()
#define t_vwport(clip_ptr)          (* device_ptr->t_VWPORT)(clip_ptr)
#define t_gr_on()                   (* device_ptr->t_GR_ON)()
#define t_gr_off()                  (* device_ptr->t_GR_OFF)()
#define t_colour(pen)               (* device_ptr->t_COLOUR)( pen )
#define t_dashed(style)             (* device_ptr->t_DASHED)( style )
#define t_thick(width)              (* device_ptr->t_THICK)( width )
#define t_visib()                   (* device_ptr->t_VISIB)()
#define t_invis()                   (* device_ptr->t_INVIS)()
#define t_revers()                  (* device_ptr->t_REVERS)()
#define t_replac()                  (* device_ptr->t_REPLAC)()
#define t_newpag()                  (* device_ptr->t_NEWPAG)()
#define t_newscr()                  (* device_ptr->t_NEWSCR)()
#define t_bell()                    (* device_ptr->t_BELL)()
#define t_wait()                    (* device_ptr->t_WAIT)()
#define t_txtmov(point)             (* device_ptr->t_TXTMOV)( point )
#define t_txtcur(row,column)        (* device_ptr->t_TXTCUR)( row, column )
#define t_txtset(attribute,value)   (* device_ptr->t_TXTSET)( attribute, value )
#define t_symdrw(point,symbol)      (* device_ptr->t_SYMDRW)( point, symbol )
#define t_txtdrw(string)            (* device_ptr->t_TXTDRW)( string )
#define t_txtget(string,max_chars)  (* device_ptr->t_TXTGET)( string, max_chars)
#define t_txtput(string)            (* device_ptr->t_TXTPUT)( string )
#define t_status(attribute,value)   (* device_ptr->t_STATUS)( attribute, value )
#define t_pr_ter(out_file)          (* device_ptr->t_PR_TER)( out_file )
#define t_vecmov(point)             (* device_ptr->t_VECMOV)( point )
#define t_vecdrw(point)             (* device_ptr->t_VECDRW)( point )
#define t_dump(file_name)           (* device_ptr->t_DUMP)( file_name )
#define t_xhrput(point)             (* device_ptr->t_XHRPUT)( point )
#define t_xhrget(point,char_pressed)            \
            (* device_ptr->t_XHRGET)( point, char_pressed )
#define t_boxfil(lower_left,upper_right,style)  \
            (* device_ptr->t_BOXFIL)( lower_left, upper_right, style )
#define t_ch_def(what,character,index,point)    \
            (* device_ptr->t_CH_DEF)( what, character, index, point )



/* ************************************************************************* */
/*                                                                           */
/*                           Driver function vector                          */
/*                                                                           */
/* ************************************************************************* */

typedef struct
  {
    clip_area VAR_PTR (VAR_PTR t_BEGIN)();  /* start graphics - open file     */
    boolean (VAR_PTR t_END)();              /* finish graphics - close file   */
    boolean (VAR_PTR t_VWPORT)();           /* define clipping viewport       */
    boolean (VAR_PTR t_GR_ON)();            /* show graphics screen           */
    boolean (VAR_PTR t_GR_OFF)();           /* hide graphics screen           */
    boolean (VAR_PTR t_COLOUR)();           /* select colour                  */
    boolean (VAR_PTR t_DASHED)();           /* select dashed line style       */
    boolean (VAR_PTR t_THICK)();            /* select line thickness          */
    boolean (VAR_PTR t_VISIB)();            /* write foreground colour only   */
    boolean (VAR_PTR t_INVIS)();            /* write background colour only   */
    boolean (VAR_PTR t_REVERS)();           /* reverse fore & background      */
    boolean (VAR_PTR t_REPLAC)();           /* write foreground  & background */
    boolean (VAR_PTR t_NEWPAG)();           /* clear graphics image           */
    boolean (VAR_PTR t_NEWSCR)();           /* clear teletype page            */
    boolean (VAR_PTR t_BELL)();             /* ring bell                      */
    boolean (VAR_PTR t_WAIT)();             /* wait for user to press RETURN  */
    boolean (VAR_PTR t_TXTMOV)();           /* move text cursor to point      */
    boolean (VAR_PTR t_TXTCUR)();           /* move text cursor to row,column */
    boolean (VAR_PTR t_TXTSET)();           /* set text attributes            */
    boolean (VAR_PTR t_SYMDRW)();           /* draw graphics symbol           */
    int     (VAR_PTR t_TXTDRW)();           /* draw text                      */
    char VAR_PTR (VAR_PTR t_TXTGET)();      /* get ASCII string from user     */
    boolean (VAR_PTR t_TXTPUT)();           /* write teletype text            */
    char VAR_PTR (VAR_PTR t_STATUS)();      /* return device status           */
    boolean (VAR_PTR t_PR_TER)();           /* print device status            */
    boolean (VAR_PTR t_VECMOV)();           /* move (with pen up)             */
    boolean (VAR_PTR t_VECDRW)();           /* draw (with pen down)           */
    boolean (VAR_PTR t_DUMP)();             /* dump graphics image to file    */
    int     (VAR_PTR t_XHRPUT)();           /* place cross hairs (pick)       */
    boolean (VAR_PTR t_XHRGET)();           /* get crosshairs (pick) location */
    boolean (VAR_PTR t_BOXFIL)();           /* draw a box                     */
    char VAR_PTR (VAR_PTR t_CH_DEF)();      /* return character shape info    */
  }
    T_dev_fns;

/*
 * All the function pointers are defined here . Some drivers are restricted
 * to particular hardware and are included by defining one (or more) of the
 * symbols IBM_PC, SUN3_WORKSTATION, X_WINDOWS .
*/
extern T_dev_fns
#ifdef IBM_PC
    Herc_dev_ptrs,  /* Hercules graphics for PC                 */
/*  EGA_dev_ptrs, *//* Enhanced Graphics Adapter for PC         */
#endif
#ifdef SUN3_WORKSTATION
    Sun3_dev_ptrs,  /* Sun 3 workstation                        */
#endif
#ifdef X_WINDOWS
/*  X11_dev_ptrs, *//* X windows terminal / workstation         */
#endif
    V500_dev_ptrs,  /* Visual 500 terminal                      */
    V550_dev_ptrs,  /* Visual 550 terminal                      */
    Xterm_dev_ptrs, /* xterm tektronix 4014 emulator            */
    HPgl_dev_ptrs,  /* Hewlett Packard graphics language plotter*/
    Post_dev_ptrs,  /* Standard Postscript hardcopy device      */
/*  DMxP_dev_ptrs,*//* Dot matrix printer                       */
    Dum_dev_ptrs;   /* Dummy graphics for debugging             */
/*
 *  ``device_ptr'' is used in all calls to driver function, and so can be
 *  declared as a ``static'' variable only if all calls to driver functions
 *  are within a single source file .
*/
#ifndef T_DEVICE_SOURCE_CODE
extern T_dev_fns
    VAR_PTR device_ptr;
#endif

FILE *t_device();

#define DEVICE_G_DEFINED
#endif

