c
c    Copyright (C) 1997  Peter Vernon Lanspeary
c
c    This program is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 2 of the License, or
c    (at your option) any later version.
c
c    This program is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program; if not, write to the Free Software
c    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
c


      PROGRAM THWAITES
c
c       Calculate the momentum thickness of a boundary layer formed on
c  a curved surface from a data file containing four columns of numbers .
c  The first column is the air speed , and the remaining columns are the
c  (x,y,z) coordinate of that air speed . An output file with the
c  same name as the input file is generated . It has nine columns . The
c  contents of the output file are,
c    1) column 1 : air speed     5) column 5 : momentum thickness
c    2) column 2 : x             6) column 6 : velocity gradient at the wall
c    3) column 3 : y             7) column 7 : H = shape factor
c    4) column 4 : z             8) column 8 : m = Thwaites form parameter
c                                9) column 9 : l = Thwaites friction parameter
c
c       A line in which the first non-blank character is a ';' is a comment
c  is copied to the output file and then ignored . Each profile is terminated
c  by a line which is not a comment line and contains only the word 'end' .
c
c  The mean flow is in the Z direction
c
c Constants
c  Hkbdlu = 5                  # keyboard logical unit no.
c  Hvdulu = 6                  # vdu logical unit no.
c  Hinlu  = 11                 # input file logical unit no.
c  Houtlu = 12                 # output file logical unit no.
c  Hnpts  = 250                # maximum number of data items in a sequence
c  Hnu    = 1.5e-5             # kinematic viscosity of air
c Variables
c  real L,                     # Thwaites friction factor
c       Theta,                 # momentum thickness
c       theta1,                # initial momentum thickness
c       M,                     # Thwaites form factor
c       U1(Hnpts),             # free stream speed
c       x(Hnpts), y(Hnpts), z(Hnpts),  # position
c       s(Hnpts),              # distance along the surface from leading edge
c       area(Hnpts),           # result of integrating U1**5
c       integrand(Hnpts),      # U1**5
c       qthwaites,             #          : thwaites parameters
c       qresp                  # function : real response
c
      implicit none
      integer *2 Hkbdlu, Hvdulu, Hinlu, Houtlu, Hnpts
      real Hnu
      parameter( Hkbdlu=5, Hvdulu=6, Hinlu=11, Houtlu=12, Hnpts=250,
     1  Hnu=1.50e-5 )
c
      character Cfile*64, Cline*80
      integer *4 i, iline, npts, jljstr
      real L, m, theta, r(Hnpts,7), U1(Hnpts), x(Hnpts), y(Hnpts),
     1  z(Hnpts), s(Hnpts), area(Hnpts), integrand(Hnpts), theta1,
     2  qThwaites, qresp
c
      equivalence ( r(1,1), U1(1) ), ( r(1,2), x(1) ), ( r(1,3), y(1) ),
     1  ( r(1,4), z(1) ), ( r(1,5), s(1) ), ( r(1,6), area(1) ),
     2  ( r(1,7), integrand(1) )
c
c       get initial momentum thickness
c
      write( Hvdulu, 15 )
15    format( ' Initial momentum thickness (mm) ? ', $ )
      theta1 = qresp( Hkbdlu, 0.0, 100.0, 4 ) / 1000.0
c
c       Open files
c
      write( Hvdulu, 1 )
1     format( ' Input file name ? ', $ )
      read( Hkbdlu, 2 ) Cfile
2     format( a )
      open( UNIT=Hinlu, FILE=Cfile, STATUS='old' )
      open( UNIT=Houtlu, FILE=Cfile, STATUS='new',
     1  CARRIAGECONTROL='list' )
      iline = 0
c     start loop                                                           1b
c
c     read data
13      npts = 1
c
c       start loop                                                         2b
3         iline = iline + 1
          read( Hinlu, 4, END=10 ) Cline
4         format( a )
          if ( jljstr( Cline ) .gt. 0 ) then                               3b
            if ( Cline(1:1) .eq. ';' ) then                                4b
              write( Houtlu, 2 ) Cline
            else if ( index( Cline, 'end' ) .gt. 0 ) then                  4m
              goto 5 ! end of series
            else                                                           4m
              read( Cline, *, ERR=8 ) U1(npts), x(npts), y(npts),
     1          z(npts)
c              write( Hvdulu, 6 ) U1(npts), x(npts), y(npts), z(npts)
              npts = npts + 1
              end if                                                       4e
            end if                                                         3e
          goto 3
c         end of loop                                                      2e
5       npts = npts - 1
c
c       sort the data - into order of ascending Z
c
        call sortsh( r, npts, 1, Hnpts, 4, 7 )
c
c       calculate distance from the 'leading edge'  and Thwaites integrand
c         leading edge is at ( x(1), y(1), 0.0 )
c
        area(1) = U1(1) **6 * theta1 **2 / Hnu / 0.45
        s(1) = 0.0
        integrand(1) = U1(1) ** 5
        do i = 2, npts                                                     2b
          s(i) = s(i-1) +  ( sqrt( ( x(i) - x(i-1) ) **2 +
     1      ( y(i) - y(i-1) ) **2 + ( z(i)- z(i-1) ) **2 ) )
          integrand(i) = U1(i) ** 5
          end do                                                           2e
c
c       do the integration
c
        call strapz( s, integrand, area, 1, npts )
c
c       calculate momentum thickness, Thwaites form factor,
c
c       write first line to file
c
        write( Hvdulu, 12 )
12      format( ' free stream  distance from  momentum    --Thwaites ',
     1    'parameters--' / '  air speed   leading edge   thickness  ',
     2    'press. grad.  skin friction' )
        write( Hvdulu, 14 ) U1(1), s(1), theta1, 0.0,
     1    qthwaites( 0.0, 'L' )
14      format( 1x, f13.6, 2g13.6, 2f13.4 )
        write( Houtlu, 6 ) U1(1), x(1), y(1), z(1), s(1), theta1,
     1      qThwaites( 0.0, 'H' ), 0.0, qThwaites( 0.0, 'L' )
6         format( 1x, 10g13.5 )
        do i = 2, npts                                                     2b
c
c       calculate momentum thickness, Thwaites form factor,
c
c        WRITE(6,6) AREA(I), U1(I)**6
          theta = sqrt( 0.45 * Hnu * area(i) / U1(i) **6 )
          m = - theta **2 / Hnu *
     1             ( U1(i) - U1(i-1) ) / ( s(i) - s(i-1) )
          L = qthwaites( m, 'L' )
c
c       write results to file
c
          write( Hvdulu, 14 ) U1(i), s(i), theta, M, L
          write( Houtlu, 6 ) U1(i), x(i), y(i), z(i), s(i), theta,
     1      L * U1(i) / Theta, qthwaites( m, 'H' ), M, L
          end do                                                           2e
        write( Houtlu, 7 )
7       format( ' end' )
        goto 13
c       end of loop                                                        2e
c
c       can only get here by error in or end of file
c
8       write( Hvdulu, 9 ) iline, Cline
9       format( ' ABORT : Cannot read file at line ', i8 / 1x, a )
        goto 3
c
c
10    write( Hvdulu, 11 )
11    format( ' end of data file' )
      stop
      end
