c
c    Copyright (C) 1997  Peter Vernon Lanspeary
c
c    This program is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 2 of the License, or
c    (at your option) any later version.
c
c    This program is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program; if not, write to the Free Software
c    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
c


      PROGRAM WT_BUOY
c
c Calculate the buoyant swirl in the win tunnel contraction.
c The second order differential equation for flow rotation angle, which
c was derived from the forced vortex theoretical model, is solved
c numerically for an axisymmetric settling chamber and contraction which
c has the same profile as the wind tunnel.
c
c Usage: wt_buoy [-profile <spec_file>] [-temperature <Tr>] [-velocity <Uout>]
c                [-begin <Xstart>] [-end <Xfinish>] [-step <Xstep>]
c                [-degrees] [-radians] [-angle <init_angle>] [-friction <Cf>]
c
c <spec_file> = name of file specifying the contraction profile:
c               consists of 2 columns of numbers,
c               1st column is distance from start of contraction (m),
c               2nd column is contraction radius (m)
c <Tr>        = inlet temperature nonuniformity divided by the ambient
c               absolute temperature
c <Uout>      = air speed at contraction outlet (m/s)
c <Xstart>    = calculation starts at this distance from the start of
c               contraction
c <Xfinish>   = calculation finishes at this distance from the start of
c               contraction
c <Xstep>     = integration step size
c <init_angle>= initial flow rotation angle
c <Cf>        = boundary layer skin friction coefficient
c
      include 'stdio.def'
      logical parse_args, read_profile
      call set_default_params()
      if ( parse_args() ) then
          call clip_params()
          if ( read_profile() ) then
              call solve_and_print()
              end if
          end if
      stop
      end



      SUBROUTINE SET_DEFAULT_PARAMS()
      include 'stdio.def'
      include 'wt_buoy.def'
      include 'wt_buoy.cmn'
c
      profile_file =  'contract.dat'
      Temp_ratio   =  1.0 / 300.0
      outlet_speed =  7.5
      x_start      = -0.760
      x_finish     =  1.650
      x_step       = 0.050
      degrees      = .TRUE.
      init_angle   = 60.0 * PI / 180.0
      frict_coeff  = 0.0
      return
      end


      SUBROUTINE CLIP_PARAMS()
      include 'stdio.def'
      real       MIN_TEMP,         MAX_TEMP,
     1           MIN_SPEED,        MAX_SPEED,
     2           MIN_X,            MAX_X,
     3           MIN_STEP,         MAX_STEP,
     4           MIN_CF,           MAX_CF
      parameter( MIN_TEMP =1.0e-6, MAX_TEMP =0.1,
     1           MIN_SPEED=1.0,    MAX_SPEED=10.0,
     2           MIN_X    =-3.0,   MAX_X    =3.0,
     3           MIN_STEP =0.005,  MAX_STEP =0.5,
     4           MIN_CF   =0.0,    MAX_CF   =0.01 )
      include 'wt_buoy.def'
      real x_temp
      include 'wt_buoy.cmn'
c
      if ( Temp_ratio .lt. MIN_TEMP ) then
          Temp_ratio = MIN_TEMP
          write( STD_ERR, 1 ) 'temperature', Temp_ratio
1         format( ' WT_BUOY : WARNING ', a, ' clipped at ', g12.5 )
      else if ( Temp_ratio .gt. MAX_TEMP ) then
          Temp_ratio = MAX_TEMP
          write( STD_ERR, 1 ) 'temperature', Temp_ratio
          end if
      if ( outlet_speed .lt. MIN_SPEED ) then
          outlet_speed = MIN_SPEED
          write( STD_ERR, 1 ) 'velocity', outlet_speed
      else if ( outlet_speed .gt. MAX_SPEED ) then
          outlet_speed = MAX_SPEED
          write( STD_ERR, 1 ) 'velocity', outlet_speed
          end if
      if ( x_start .lt. MIN_X ) then
          x_start = MIN_X
          write( STD_ERR, 1 ) 'begin', x_start
      else if ( x_start .gt. MAX_X ) then
          x_start = MAX_X
          write( STD_ERR, 1 ) 'begin', x_start
          end if
      if ( x_finish .lt. MIN_X ) then
          x_finish = MIN_X
          write( STD_ERR, 1 ) 'end', x_finish
      else if ( x_finish .gt. MAX_X ) then
          x_finish = MAX_X
          write( STD_ERR, 1 ) 'end', x_finish
          end if
      if ( x_finish .lt. x_start ) then
          x_temp = x_finish
          x_finish = x_start
          x_start  = x_temp
          end if
      if ( init_angle .lt. -PI ) then
          init_angle = -PI
          write( STD_ERR, 1 ) 'angle', init_angle
      else if ( init_angle .gt. PI ) then
          init_angle = PI
          write( STD_ERR, 1 ) 'angle', init_angle
          end if
      if ( frict_coeff .lt. MIN_CF ) then
          frict_coeff = MIN_CF
          write( STD_ERR, 1 ) 'friction', frict_coeff
      else if ( frict_coeff .gt. MAX_CF ) then
          frict_coeff = MAX_CF
          write( STD_ERR, 1 ) 'friction', frict_coeff
          end if
      return
      end



      LOGICAL FUNCTION PARSE_ARGS()
c
c    Parse the program arguments and open the necessary files . The
c program arguments, which are all optional, are
c
c Usage :
c  wt_buoy [-profile <spec_file>] [-temperature <Tr>] [-velocity <Uout>]
c          [-begin <Xstart>] [-end <Xfinish>] [-step <Xstep>] [-degrees]
c          [-radians] [-angle <init_angle>] [-friction <Cf>]
c
c  <spec_file> = name of file specifying the contraction profile: consists of
c                2 columns of numbers, 1st column is distance from start of
c                contraction (m), 2nd column is contraction radius (m)
c  <Tr>        = inlet temperature nonuniformity divided by the ambient
c                absolute temperature
c  <Uout>      = air speed at contraction outlet (m/s)
c  <Xstart>    = calculation starts at this distance from start of contraction
c  <Xfinish>   = calculation finishes at this distance from start of contraction
c  <Xstep>     = integration step size
c  <init_angle>= initial flow rotation angle
c  <Cf>        = boundary layer skin friction coefficient
c
c integer
c  arg_ptr        # pointer to next program argument
c  n_args         # number of program arguments
c  iargc          # function : number of program arguments
c  str_length     #          : length of string
c logical
c  is_abbr        # function : is abbreviation
c character
c  arg_str*64     # argument string
c  value*64       # value of argument
c subroutine
c  getarg
c
      include 'stdio.def'
      include 'wt_buoy.def'
c
      integer arg_ptr, n_args, iargc, str_length
      logical is_abbr
      character arg_str*64, value*64
      include 'wt_buoy.cmn'
c
c
c         Examine each argument in turn
c
      parse_args = .TRUE.
      arg_ptr = 1
      n_args = iargc()
1     if ( arg_ptr .lt. n_args ) then
          call getarg( arg_ptr, arg_str )
          arg_ptr = arg_ptr + 1
          if ( is_abbr( arg_str, '-profile', 2 ) ) then
              call getarg( arg_ptr, profile_file )
          else if ( is_abbr( arg_str, '-temperature', 2 ) ) then
              call getarg( arg_ptr, value )
              read( value, *, ERR=2) Temp_ratio
          else if ( is_abbr( arg_str, '-velocity', 2 ) ) then
              call getarg( arg_ptr, value )
              read( value, *, ERR=2 ) outlet_speed
          else if ( is_abbr( arg_str, '-begin', 2 ) ) then
              call getarg( arg_ptr, value )
              read( value, *, ERR=2 ) x_start
          else if ( is_abbr( arg_str, '-end', 2 ) ) then
              call getarg( arg_ptr, value )
              read( value, *, ERR=2 ) x_finish
          else if ( is_abbr( arg_str, '-step', 2 ) ) then
              call getarg( arg_ptr, value )
              read( value, *, ERR=2 ) x_step
          else if ( is_abbr( arg_str, '-degrees', 2 ) ) then
              degrees = .TRUE.
              arg_ptr = arg_ptr - 1
          else if ( is_abbr( arg_str, '-radians', 2 ) ) then
              arg_ptr = arg_ptr - 1
              degrees = .FALSE.
          else if ( is_abbr( arg_str, '-angle', 2 ) ) then
              call getarg( arg_ptr, value )
              read( value, *, ERR=2 ) init_angle
              if ( degrees ) init_angle = init_angle * PI / 180.0
          else if ( is_abbr( arg_str, '-friction', 2 ) ) then
              call getarg( arg_ptr, value )
              read( value, *, ERR=2 ) frict_coeff
          else
              write( STD_ERR, 4 ) arg_str(:str_length(arg_str))
4             format( ' WT_BUOY : Illegal argument ', a /
     1 ' Usage : wt_buoy [-profile <spec_file>] [-temperature <Tr>] ' /
     2 17x, '[-velocity <Uout>] [-begin <Xstart>] [-end <Xfinish>]' /
     3 17x, '[-step <Xstep>] [-degrees] [-radians] ',
     4      '[-angle <init_angle>] [-friction <Cf>]' )
              parse_args = .FALSE.
              return
              end if
          arg_ptr = arg_ptr + 1
          goto 1
c
2         write( STD_ERR, 3 ) value(:str_length(value)),
     1                        arg_str(:str_length(arg_str))
3         format( ' WT_BUOY : ERROR : Cannot decode value ', a,
     1            ' for argument ', a )
          goto 1
          end if
c         end of loop
c
      return
      end




      LOGICAL FUNCTION IS_ABBR( ABBREV, TEST, MIN_LEN )
c
c       EQual ABBReviation
c     Tests if the character string CSTR is a legal abbreviation of the
c character string CTEST .
c
c Arguments
c  character abbrev*(*), # operand character string               INPUT
c            test*(*)    # test string                            INPUT
c  integer   min_len     # minimum number of characters allowed   INPUT
c                        # in the abbreviation
c Variables
c  integer i,          # indexing counter
c          finish      # first letter in a word
c          str_length  # function: length of string
c
      include   'stdio.def'
      integer   i, finish, min_len, str_length
      character abbrev*(*), test*(*)
c
      finish = max( str_length( abbrev ), 1 )
      if ( finish .gt. str_length( test ) ) then
          is_abbr = .FALSE.
      else
          is_abbr = .TRUE.
          i = 1
1         if ( i .le. finish .and. is_abbr ) then
              is_abbr = is_abbr .and. abbrev(i:i) .eq. test(i:i)
              i = i + 1
              goto 1
              end if
        is_abbr = is_abbr .and. i .gt. min_len
        end if
      return
      end




      LOGICAL FUNCTION READ_PROFILE()
c
c Read the contraction profile from file 'profile_file'. Returned value
c is the number of coordinates read.
c
c logical   is_comment  # function : is arg a comment ?
c integer   str_length  # function : length of string
c character line        # line from file
c
      include 'stdio.def'
      include 'wt_buoy.def'
      logical is_comment
      integer str_length
      character line*80
      include 'wt_buoy.cmn'
c
      open( UNIT=PRF_LU, FILE=profile_file, STATUS='old', ERR=6 )
      len_profile = 1
c     start of loop
1         read( PRF_LU, 2, END=5 ) line
2         format( a )
          if ( is_comment( line ) ) goto 1
          read( line, *, ERR=5 )
     1        profile_x(len_profile), profile_r(len_profile)
          profile_x(len_profile) = profile_x(len_profile) / 1000.0
          profile_r(len_profile) = profile_r(len_profile) / 1000.0
          len_profile = len_profile + 1
          goto 1
c         end of loop
3     write( STD_ERR, 4 ) len_profile, line(:str_length(line))
4     format( ' WT_BUOY : ERROR : Cannot decode data line ', i5 /
     1        1x, a )
5     len_profile = len_profile - 1
      close( UNIT=PRF_LU )
      read_profile = len_profile .gt. 1
      return
6     write( STD_ERR, 7 ) profile_file(:str_length(profile_file))
7     format( ' WT_BUOY : FATAL : Cannot open file ', a )
      read_profile = .FALSE.
      return
      end




      INTEGER FUNCTION STR_LENGTH( STRING )
c
c    Find the location of the last non-blank character in the character
c string STRING . If the whole string is blank, the returned value is 0 .
c
      include 'stdio.def'
      character string*(*)
c
      str_length = len( string )
1     if ( str_length .ge. 1 ) then
          if ( string(str_length:str_length) .eq. ' ' ) then
              str_length = str_length - 1
          else
              return
              end if
          goto 1
          end if
      return
      end



      LOGICAL FUNCTION IS_COMMENT( STRING )
c
c A string is a comment if it is blank or the first non-blank is a ';'
c
      include 'stdio.def'
      integer i
      character string*(*)
c
      i = 1
      is_comment = .TRUE.
1     if ( i .le. len(string) ) then
          if ( string(i:i) .ne. ' ' ) then
              if ( string(i:i) .ne. ';' ) is_comment = .FALSE.
              return
              end if
          i = i + 1
          goto 1
          end if
      return
      end



      SUBROUTINE SOLVE_AND_PRINT()
c
c Solve the differential rotation equation an print the results
c
c real
c  location           # streamwise location, x
c  radius             # radius of duct at x
c  Richardson         # average Richardson number in the duct
c  rotation           # flow rotation angle
c  rotn_ang           # flow rotation angle in radians only
c  rotn_ang_deriv     # first derivative of flow rotation angle
c  swirl              # peripheral swirl angle of the flow
c  velocity           # air speed at x
c  calculate_radius   # function : calculate radius
c  qrknde             #          : Runge-Kutta-Nystrom 2nd order ODE solver
c  rotation_equation  #          : ODE to be solved
c
      include 'stdio.def'
      include 'wt_buoy.def'
      real location, radius, Richardson, rotation, rotn_ang,
     1     rotn_ang_deriv, swirl, velocity,
     2     calculate_radius, qrknde, rotation_equation
      character angle*7
      include 'wt_buoy.cmn'
      external rotation_equation
c
      Flow_rate_on_Pi = outlet_speed * calculate_radius( INFINITY ) **2
      if ( degrees ) then
          rotation = 180.0 / PI * init_angle
          angle = 'degrees'
      else
          rotation = init_angle
          angle = 'radians'
          end if
      location   = x_start
      radius     = calculate_radius( location )
      velocity   = Flow_rate_on_Pi / radius **2
      rotn_ang   = init_angle
      rotn_ang_deriv = 0.0

      Richardson = 0.5 / velocity **2
      write( STD_OUT, 1 ) Temp_ratio, x_finish - x_start, GRAVITY,
     1                    velocity, frict_coeff, angle, angle
1     format( ' ;' /
     1    ' ; Buoyant swirl in an axisymmetric contraction' /
     2    ' ;   Temperature nonuniformity ratio = ', g12.5 /
     3    ' ;   Length of duct                  = ', f7.3, ' m' /
     4    ' ;   Acceleration due to gravity     = ', f7.5, ' m/s/s' /
     5    ' ;   Inlet flow speed                = ', f8.5, ' m/s' /
     6    ' ;   Boundary layer friction coeff.  = ', f8.6 /
     7    ' ;  x    radius  speed  rotation      swirl' /
     8    ' ; (m)    (m)    (m/s)  (', a, ')   (', a, ')' )
      write( STD_OUT, 2 ) location, radius, velocity, rotation, 0.0
2     format( 1x, f7.3, f7.4, f8.4, 2g12.5e1 )
c
c     start_loop
3     if ( location .lt. x_finish ) then
          rotation = qrknde( location, rotn_ang, rotn_ang_deriv,
     1                       x_step, rotation_equation )
          swirl    = atan( rotn_ang_deriv * radius )
          radius   = calculate_radius( location )
          velocity = Flow_rate_on_Pi / radius **2
          Richardson =  Richardson + 1.0 / velocity **2
          if ( degrees ) then
              rotation = rotation * 180.0 / PI
              swirl    = swirl    * 180.0 / PI
              end if
          write( STD_OUT, 2 ) location, radius, velocity, rotation,
     1        swirl
          goto 3
          end if
c         end of loop
      Richardson = Temp_ratio * GRAVITY *
     1             ( Richardson - 0.5 / velocity **2 ) * x_step
      write( STD_OUT, 4 ) Richardson
4     format( ' ; ' / ' ; Average Richardson number = ', g12.5 )
      stop
      end



      REAL FUNCTION ROTATION_EQUATION( LOCATION, ROTATION_ANGLE,
     1                                 ROTATION_ANGLE_DERIV )
c
c Evaluates the second derivative of the flow rotation angle from the
c forced vortex formula
c                                           2
c Psi'' = ( Temp_ratio g cos(Psi) )/ ( 4 R U )  and  U = Flow_rate/(pi*R**2)
c
c Arguments
c  real location             # x
c       rotation_angle       # flow rotation angle
c       rotation_angle_deriv # first derivative of flow rotation angle
c
c local variables
c  real
c   radius            # duct radius
c   velocity          # air speed
c   calculate_radius  # function
c
      include 'stdio.def'
      include 'wt_buoy.def'
      real    location, rotation_angle, rotation_angle_deriv,
     1        radius, velocity, calculate_radius
      include 'wt_buoy.cmn'
c
      radius   = calculate_radius( location )
      velocity = Flow_rate_on_Pi / radius **2
      rotation_equation = Temp_ratio * GRAVITY * cos( rotation_angle ) /
     1                    4.0 / radius / velocity **2
     2                  - frict_coeff / radius * rotation_angle_deriv
      return
      end



      REAL FUNCTION CALCULATE_RADIUS( LOCATION )
c
c Calculate radius of duct at specified location by linear interpolation
c of the profile data
c
      include 'stdio.def'
      include 'wt_buoy.def'
      integer i, jdhunt
      real    location
      include 'wt_buoy.cmn'
c
      if ( location .le. profile_x(1) ) then
          calculate_radius = profile_r(1)
      else if ( location .ge. profile_x(len_profile) ) then
          calculate_radius = profile_r(len_profile)
      else
          i = jdhunt( profile_x, 1, len_profile, USE_PREVIOUS, location)
          calculate_radius =
     1        ( profile_r(i+1) - profile_r(i) ) /
     2        ( profile_x(i+1) - profile_x(i) ) *
     3        ( location - profile_x(i) ) + profile_r(i)
          end if
      return
      end



      REAL FUNCTION QRKNDE( X, Y, Z, H, F )
c
c       Runge- Kutta- Nystrom  solution of Differential Equation
c     Function for Runge- Kutta- Nystrom ( numerical method ) solution of a
c second order ordinary differential equation of the form :-
c       y" = d( dy/dx )/dx = f( x, y, y' ) ;   y' = dy/dx
c
c                       input                   output
c
c real x,             # independent variable  # increased by h
c      y,             # initial value of y    # result- value of y
c      z,             # initial value of y'   # final value of y'
c      h,             # increment size
c      QRKNDE                                 # result
c local variables
c real a,b,c,d,e,g    # intermediate values
c
      implicit undefined (a-z)
      real x, y, z, h, h2, a, b, c, d, e, g, f
c
      h2 = h / 2.0
      e = x + h2
      a = h2 * f( x, y, z )
      g = y + h2 * ( z + a / 2.0 )
      b = h2 * f( e, g, z + a )
      c = h2 * f( e, g, z + b )
      x = x + h
      d = h2 * f( x , y + h * ( z + c ), z + 2.0 * c )
      y = y + h * ( z + ( a + b + c ) / 3.0 )
      z = z + ( a + 2.0 * b + 2.0 * c + d ) / 3.0
      QRKNDE = y
      return
      end



      INTEGER FUNCTION JDHUNT( X_DATA, INCREM, N_POINTS, INIT_GUESS,
     1                         X0 )
c
c       inteval Doubling HUNT
c Function for locating the last element between the 1st and Nth elements in
c the array X of N real numbers , which in the case of an ascending array, is
c less than the value X0 . In the case of a descending array, it finds the
c last element which is greater than or equal to X0 . The returned value of
c JDHUNT is the index of this element in the array X . The search begins with
c an interval-doubling hunt until the target value X0 is straddled, and then
c the search interval is reduced by bisection search . INITIAL_GUESS is an
c initial guess for the returned result supplied by the calling program . If
c INITIAL_GUESS=-999, and JDHUNT has been used before, the previous returned
c result is used as the starting value . If INITIAL_GUESS is not within the
c bounds of the data array (e.g. 0), the search is by bisection only .
c    N.B. X must be an array of values arranged in STRICTLY ASCENDING or
c  STRICTLY DESCENDING order .
c
c     Arguments
c real x_data(*)        # data array
c integer increm,       # data array indexing increment
c         n_points      # number of data points used
c         init_guess    # starting point for search
c real x0               # search target value
c
c     Variables
c logical desc          # .true. if data array X is descending
c         not_pass1     # second or later call to this function
c integer nlast         # index of last point used
c         left, right,  # left & right hand indices of the search interval
c         centre        # moving search index
c         saved_result  # result of the previous call
c         step          # hunt interval size
c
      include 'stdio.def'
      real x_data(*), x0
      logical desc, not_pass1
      integer increm, n_points, init_guess,
     1        nlast, left, right, centre, saved_result, step
c
      save not_pass1, saved_result
      data not_pass1 / .FALSE. /
      data saved_result / 0 /
c
      if( n_points .le. 0 .or. increm .le. 0 ) return
      nlast = 1 + ( n_points - 1 ) * increm
      desc = x_data(1) .gt. x_data(nlast)
c
c             deal with the boundary conditions
c ( the method of left=0, right=nlast+1 from Numerical Recipes does not
c   work for nincr>1 . )
c
      if ( x0 .lt. x_data(1) .neqv. desc ) then                          1b
          left = 1
      else if ( x0 .gt. x_data(nlast) .neqv. desc ) then                 1m
          left = nlast
      else if ( x0 .eq. x_data(1) ) then                                 1m
          left = 1
      else if ( x0 .eq. x_data(nlast) ) then                             1m
          left = nlast - increm
      else                                                               1m
c
c             decide if a hunt is required
c
          if ( not_pass1 .and. init_guess .eq. -999 .or.
     1            init_guess .ge. 1 .and. init_guess .le. nlast ) then   2b
              if ( init_guess .eq. -999 ) then                           3b
                  left = saved_result
              else                                                       3m
                  left = init_guess
                  end if                                                 3e
              left = max( 1, min( nlast,
     1            ( ( left - 1 ) /  increm ) * increm + 1 ) )
c
c             doubling hunt procedure
c
              step = increm
              if ( x0 .ge. x_data(left) .eqv. desc ) then                3b
c                 start loop                                             4b
1                     right = left
                      left = right - step
                      step = 2 * step
                      if ( left .lt. 1 ) then                            5b
                          left = 1
                      else if ( x0 .lt. x_data(left) .neqv. desc ) then  5m
                          goto 1
                          end if                                         5e
c                     end loop                                           4e
              else                                                       3m
                  right = left
c                 start of loop                                          4b
2                     left = right
                      right = left + step
                      step = 2 * step
                      if ( right .gt. nlast ) then                       5b
                          right = nlast
                      else if ( x0 .ge. x_data(right) .neqv. desc ) then 5m
                          goto 2
                          end if                                         5e
c                     end of loop                                        4e
                  end if                                                 3e
          else                                                           2m
              left = 1
              right = nlast
              end if                                                     2e
c
c             bisection procedure
c ( the simple averaging of left & right does not find a valid array index
c   for centre )
c
3         if ( right - left .gt. increm ) then                           2b
              centre = left + ( right - left ) / 2 / increm * increm
              if ( x0 .gt. x_data(centre) .neqv. desc ) then             3b
                  left = centre
              else                                                       3m
                  right = centre
                  end if                                                 3e
              goto 3
              end if                                                     2e
          end if                                                         1e
      not_pass1 = .TRUE.
      saved_result = left
      JDHUNT = left
      return
      end


