c    Copyright (C) 1997  Peter Vernon Lanspeary
c
c    This documentation is free; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 2 of the License, or
c    (at your option) any later version.
c
c    This documentation is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this documentation; if not, write to the Free Software
c    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
c


c
c    DIG_HIST
c
c Processes a data stream obtained from digitized graph .
c compile :  f77 -o dig_hist -O dig_hist.f ~/lib/sun4/f77lib.a
c
c     The input data is assumed to take the form of two columns of numbers
c with comment lines acting as record separators. Each record is processed
c separately. Separating comments are copied to the output file.
c First the abscissae (first column) of the input are sorted into ascending
c order. A moving average (uniform weight) filter is then applied to the data
c to provide smoothing. A new data stream is then interpolated from this
c sorted and filtered data and written to the output .
c
c Synopsis :
c   dig_hist [-width   <linear_pass_band>]
c            [-ratio   <log_pass_band>]
c            [-start   <start_abscissa>]
c            [-finish  <finish_abscissa>]
c            [-length  <OP_record_length>]
c            [-min_err <minimum peak interpolation error>]
c            [-max_err <maximum peak interpolation error>]
c            [-min_inc <minimum interpolation abscissa increment>]
c            [-max_inc <maximum interpolation abscissa increment>]
c            [-inc_adj <interpolation abscissa increment adjustment factor]
c            [-adaptiv]  # adaptive step in interpolated data
c            [-unif]     # uniform step in interpolated data
c            [-nograph]
c            [-graph]
c            [-help    <help_file>]
c            [-data    <input_file>]
c            [-output  <output_file>]
c            [-param   <command_file>]
c
c   linear_pass_band  width of smoothing pass band, expressed as percentage
c                         of abscissa range, min=0, max=100, default=3
c   log_pass_band     width of smoothing pass band, expressed as
c                         right_edge_of_band = left_edge_of_band * log_pass_band
c                         min=1, max=100, default=1.2
c   start_abscissa    first abscissa value of output data record
c                      min & default = minimum input, max = max input
c   finish_abscissa   last abscissa value of output data record
c                      min = minimum input, max & default = max input
c   OP_record_length  length of output data record
c                      min = 1, max = MAX_LEN (default = DEF_OP_LEN )
c                      min = 1, max = MAX_POLY, default = 3
c   adapt_param_file  file containing adaptive interpolation parameters
c                     -- see below
c   help_file         help is written to this file (default = stderr)
c   input_file        default = stdin
c   output_file       default = stdout
c   command           parameters are read from this file
c
c Adaptive interpolation parameters
c **********************************
c  The adaptive interpolation parameters are :
c         description            name      min  max default       units
c minimum interpolation error  min_error   0.01   1   0.1    % * ordinate_range
c maximum interpolation error  max_error   0.1   10   1.0    % * ordinate_range
c minimum abscissa increment   min_increm  0.01   1   0.1    * OP_increm
c maximum abscissa increment   max_increm  1.0  100  10.0    * OP_increm
c increment adjustment factor  increm_adj  1.02   4   1.414
c
c  **maximum interpolation error is the maximum difference between filtered
c  ordinates and interpolation straight line segments, expressed as percentage
c  of the ordinate range.
c
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc


      PROGRAM DIG_HIST
c
c integer constants
c   DATA_ABS = 1      data array column no. for abscissa values
c   DATA_ORD = 2                                ordinate
c   FILT_ABS = 3                                filtered abscissa
c   FILT_ORD = 4                                filtered ordinate
c   INTP_ABS = 5                                interpolated abscissa
c   INTP_ORD = 6                                interpolated ordinate
c   MAX_LEN  = 16384   maximum input and output record length
c
c logical variables
c  adapt_interp       # adaptive step in interpolated data
c  do_plot            # plot data & results
c  log_p_band         # log pass band description (else linear)
c  parse_args         # parses program arguments and opens files
c  read_record        # reads a record from the input file
c
c integer variables
c  IP_len             # length of input data record
c  OP_len             # length of output data record
c  i_out
c
c real variables
c  param_list(PARAM_LEN) # operating parameters
c  b_width            # width of smoothing pass band,
c  record(MAX_LEN,6)  # storage for the input, filtered and interpolated data
c
      include 'stdio.def'
      include 'dig_hist.def'
      integer     MAX_LEN,
     1            DATA_ABS,     DATA_ORD,     FILT_ORD,
     2            FILT_ABS,     INTP_ABS,     INTP_ORD
      parameter ( MAX_LEN=16384,
     1            DATA_ABS=1,   DATA_ORD=2,   FILT_ABS=3,
     1            FILT_ORD=4,   INTP_ABS=5,   INTP_ORD=6 )
c
      logical adapt_interp, do_plot, log_p_band, parse_args, read_record
      integer IP_len, OP_len, i_out
      real    param_list(PARAM_LEN), b_width, record(MAX_LEN,6)
c
c         set default values
c
c
c         Get input file name & ouput file name from arg list and open the
c         files . Also get averaging length and other args .
c
      if ( parse_args( param_list ) ) then                                1b
c
c         check the values from PARSE_ARGS
c
          do_plot = int( param_list(GRAPHICS) + 0.5 ) .eq. YES
          if ( param_list(INPUT_LU) .le. 0.0 .or.
     1         param_list(OUTPUT_LU) .lt. 0.0 .or.
     2         param_list(HELP_LU) .lt. 0.0 ) then                        2b             2b
              write( STD_ERR, 1 )
1             format( ' BUG : Illegal logical unit from PARSE_ARGS' )
              stop
              end if                                                      2e
c
c         Fix bandwidth parameter;
c         use default filter pass-band width if necessary.
c
          if ( param_list(WIDTH) .ne. 0.0 ) then                          2b
              b_width = min( max( 0.0, param_list(WIDTH) ), 100.0 )
              log_p_band = .FALSE.
          else if ( param_list(RATIO) .ne. 0.0 ) then                     2m
              b_width = min( max( 1.0001, param_list(RATIO) ),
     1                       100.0 )
              log_p_band = .TRUE.
          else                                                            2m
              b_width = 0.0
              log_p_band = .FALSE.
              end if                                                      2e
c
c
c         INITIAL PROCESSING
c***************************
c         Read one data record
c
c         Start of main loop
2         if ( read_record( int( param_list(INPUT_LU) + 0.5 ),
     1                      int( param_list(OUTPUT_LU) + 0.5 ),
     2                      record(1,DATA_ABS), record(1,DATA_ORD),
     3                      IP_len, MAX_LEN ) ) then                      2b
c
c         Sort abscissae into ascending order
c
c                  SORTSH( X, START_ROW, END_ROW, ROWS, SORT_COL, COLS )
              call sortsh( record, IP_len, 1, MAX_LEN, DATA_ABS,
     1                     INTP_ORD )
c
c         Compress out any duplicated abscissa values
c
c                  COMP_DUP( ARRAY,  IN_ROWS, OUT_ROWS, COMP_COL,
c     1                     ROWS, COLUMNS, TOLERANCE )
              call comp_dup( record, IP_len,  IP_len,   DATA_ABS,
     1                       MAX_LEN, INTP_ORD, 0.2e-3 )
c
c         Plot sorted data (as dotted line)
c
              if ( do_plot ) then                                         3b
                  call ilgopen( ' ' )
                  call newscr()
                  call ilgraph( 1, 1, 27, 27, 1 )
                  if ( log_p_band ) call ilgslaw( 'log', 'lin' )
                  call ilgline( 0, 2, 1 )
                  call ilgdata( record(1,DATA_ABS), record(1,DATA_ORD),
     1                          1, IP_LEN, 0 )
                  end if                                                  3e
c
c         APPLY MOVING AVERAGE FILTER
c************************************
c
c         apply moving average filter (also generates new abscissa values)
c                  SMVOOTH( X, Y,
c                    XY_INCREM, N_POINTS, PASS_BAND, LOG?,
c                    X_OUT, OUT, OUT_INCR )
c
              call smvooth( record(1,DATA_ABS), record(1,DATA_ORD),
     1                      1, IP_len, b_width, log_p_band,
     2                      record(1,FILT_ABS), record(1,FILT_ORD), 1 )

c
c         Compress out any duplicated abscissa values
c      SUBROUTINE COMP_DUP( ARRAY, IN_ROWS, OUT_ROWS, COMP_COL,
c     1                     ROWS, COLUMNS, TOLERANCE )
c
              call comp_dup( record, IP_len, IP_len, FILT_ABS,
     1                       MAX_LEN, INTP_ORD, 0.2e-3 )
c
c         plot filtered data (as dashed line)
c
              if ( do_plot ) then                                         3b
                  call ilgline( 0, 1, 1 )
                  call ilgdata( record(1,FILT_ABS), record(1,FILT_ORD),
     1                          1, IP_LEN, 0 )
                  end if                                                  3e
c
c         INTERPOLATE FROM THE FILTERED DATA
c*******************************************
c         clip start and finish of output abscissa to avoid extrapolation
c
              adapt_interp = int( param_list(ADAPTIVE) + 0.5) .eq. YES
              OP_len = int( param_list(OP_LENGTH) + 0.5 )
              call interpolate(
     1            record(1,FILT_ABS), record(1,FILT_ORD), IP_len,
     2            record(1,INTP_ABS), record(1,INTP_ORD),
     3            OP_len, MAX_LEN,
     4            max( param_list(OP_START), record(1,FILT_ABS) ),
     5            min( param_list(OP_FINISH), record(IP_len,FILT_ABS) ),
     6            log_p_band, adapt_interp, param_list(ERROR_MIN) )
c
c         plot interpolation results (as solid line)
c
              if ( do_plot ) then                                         3b
                  call ilgline( 5, 0, 1 )
                  call ilgdata( record(1,INTP_ABS), record(1,INTP_ORD),
     1                   1, OP_len, 0 )
                  call ilgwait()
                  call ilgclos()
                  end if                                                  3e
c
c         write results to file
c
              do 4 i_out = 1, OP_len
                  write( int( param_list(OUTPUT_LU) + 0.5 ), 3 )
     1                record(i_out,INTP_ABS), record(i_out,INTP_ORD)
3                 format( 1x, 2g13.5 )
4                 continue
c
              goto 2
              end if                                                      2e
c             end of main loop
      else                                                                1m
          call GIVE_HELP( int( param_list(HELP_LU) + 0.5 ), MAX_LEN )
          end if                                                          1e
      stop
      end



      SUBROUTINE INTERPOLATE( IP_X, IP_Y, IP_LEN,
     1     OP_X, OP_Y, OP_LEN, OP_MAX,
     2     OP_START, OP_FINISH, LOG_P_BAND, ADAPT_INTERP, ADAPT_PARAMS )
c
c         Interpolate new data from specified abscissa starting point
c         end point and increment.
c
c arguments
c  real    IP_x(*),           # input data abscissae
c          IP_y(*)            # input data ordinates
c  integer IP_len             # number of input data coordinates
c  real    OP_x(*),           # output data abscissae
c          OP_y(*)            # output data ordinates
c  integer OP_len             # number of output data coordinates
c          OP_max             # array size for output
c  real    OP_start,          # output data start abscissa
c          OP_finish          # output data final abscissa
c  logical log_p_band         # log pass-band description (else linear)
c          adapt_interp       # adaptive linear interpolation (else uniform)
c  real    adapt_params(*)    # adaptive interpolation parameters
c
c integer constants
c  BISECTION_ONLY =  0  signals to JDHUNT: use bisection search only
c  USE_PREVIOUS = -999        : use previous result as starting point for search
c  ERROR_MIN  = 1
c  ERROR_MAX  = 2
c  INCREM_MIN = 3
c  INCREM_MAX = 4
c  ADJ_INCREM = 5
c
c local variables
c  integer
c   err_too_big       # number of successive iterations in which interpolation
c                     #     error was too big
c   err_too_small     # number of successive iterations in which interpolation
c                     #     error was too small
c   i_left, i_right   # input data indices at extremes of interpolation interval
c   i_out             # output data index
c   JDHUNT            # function : search by interval doubling hunt & bisection
c  real
c                     # interpolation parameters : maximum / minimum permitted
c   max_error         #    interpolation error
c   max_increm        #    output data abscissa increment
c   min_error         #    interpolation error
c   min_increm        #    output data abscissa increment
c   increm_adj        #    factor by which increment can be adjusted
c   finish            # OP_finish + rounding error tolerance
c   max_IP_y          # maximum input data ordinate
c   min_IP_y          # minimum input data ordinate
c   increm            # current output data abscissa increment
c   OP_increm         # initial output dat abscissa increment
c   peak_error        # maximum difference between input data and line
c                     #     segment from the interpolation
c   FIND_ERROR        # function : calculate "peak_error"
c   QLININT           #          : linear interpolation
c
c
      include 'stdio.def'
      integer BISECTION_ONLY, USE_PREVIOUS,
     1        ERROR_MIN,  ERROR_MAX, INCREM_MIN, INCREM_MAX, ADJ_INCREM
      parameter( BISECTION_ONLY = 0,  USE_PREVIOUS = -999,
     1           ERROR_MIN      = 1,  ERROR_MAX    = 2,
     2           INCREM_MIN     = 3,  INCREM_MAX   = 4,
     3           ADJ_INCREM     = 5 )
c
      logical log_p_band, adapt_interp
      integer IP_len, OP_len, OP_max, err_too_big, err_too_small,
     1        i_left, i_right, i_out, JDHUNT
      real    IP_x(*), IP_y(*), OP_x(*), OP_y(*), OP_start, OP_finish,
     1        adapt_params(*), max_error, max_increm, min_error,
     2        min_increm, increm_adj, finish, max_IP_y, min_IP_y,
     3        increm, OP_increm, peak_error, FIND_ERROR
c
c         SET UP INTERPOLATION PARAMETERS
c         *******************************
c                   min    max   default
c      min_error  = 0.01    1     0.1     % * ordinate_range
c      max_error  = 0.1    10     1       % * ordinate_range
c      max_increm = 1     100     10      * OP_increm
c      min_increm = 0.01    1     0.1     * OP_increm
c      increm_adj = 1.02    4     1.414
c
      if ( .not. log_p_band ) then                                        1b
          OP_increm = ( OP_finish - OP_start ) /
     1        real( max( OP_len - 1, 1 ) )
          finish = OP_finish + OP_increm / 1000.0
      else                                                                1m
          OP_increm = ( OP_finish / OP_start ) **
     1        ( 1.0 / real( max( OP_len - 1, 1 ) ) )
          finish = OP_finish * 1.0001
          end if                                                          1e
      if ( adapt_interp ) then                                            1b
          max_IP_y = IP_y(1)
          min_IP_y = IP_y(1)
          do 2 i_left = 1, IP_len                                         2b
              max_IP_y = max( max_IP_y, IP_y(i_left) )
              min_IP_y = min( min_IP_y, IP_y(i_left) )
2             continue                                                    2e
          max_IP_y = ( max_IP_y - min_IP_y ) / 100.0
          if ( 0.1 .le. adapt_params(ERROR_MIN) .and.
     1                  adapt_params(ERROR_MIN) .le. 1.0 ) then           2b
              min_error  = adapt_params(ERROR_MIN)  * max_IP_y
          else                                                            2m
              min_error  = 0.1 * max_IP_y
              end if                                                      2e
          if ( 0.01 .le. adapt_params(ERROR_MAX) .and.
     1                   adapt_params(ERROR_MAX) .le. 10.0 ) then         2b
              max_error  = adapt_params(ERROR_MAX) * max_IP_y
          else                                                            2m
              max_error  = 1.0 * max_IP_y
              end if                                                      2e
          if ( 1.0 .le. adapt_params(INCREM_MAX) .and.
     1                  adapt_params(INCREM_MAX) .le. 100.0 ) then        2b
              if ( log_p_band ) then
                  max_increm = OP_increm ** adapt_params(INCREM_MAX)
              else
                  max_increm = adapt_params(INCREM_MAX) * OP_increm
                  end if
          else                                                            2m
              if ( log_p_band ) then
                  max_increm = OP_increm ** 10.0
              else
                  max_increm = 10.0 * OP_increm
                  end if
              end if                                                      2e
          if ( 0.01 .le. adapt_params(INCREM_MIN) .and.
     1                   adapt_params(INCREM_MIN) .le. 1.0 ) then         2b
              if ( log_p_band ) then
                  min_increm = OP_increm ** adapt_params(INCREM_MIN)
              else
                  min_increm = adapt_params(INCREM_MIN) * OP_increm
                  end if
          else                                                            2m
              if ( log_p_band ) then
                  min_increm = OP_increm ** 0.1
              else
                  min_increm = 0.1 * OP_increm
                  end if
              end if                                                      2e
          if ( 1.02 .le. adapt_params(ADJ_INCREM) .and.
     1                   adapt_params(ADJ_INCREM) .le. 4.0 ) then         2b
              increm_adj = adapt_params(ADJ_INCREM)
          else                                                            2m
              increm_adj = 1.414
              end if                                                      2e
C      WRITE(6,9996) MIN_ERROR, MAX_ERROR, MIN_INCREM, MAX_INCREM,
C     1              INCREM_ADJ
C9996  FORMAT( 'MIN_ERROR MAX_ERROR MIN_INCREM MAX_INCREM INCREM_ADJ' /
C     1         5G13.5 )
          end if                                                          1e
c
c         START INTERPOLATION PROCESS
c         ***************************
c
      OP_x(1) = OP_start
      i_left  = JDHUNT( IP_x, 1, IP_len, BISECTION_ONLY, OP_x(1) )
      i_left  = min( max( 1, i_left ), IP_len - 1 )
      OP_y(1) = ( IP_y(i_left+1) - IP_y(i_left) ) /
     1          ( IP_x(i_left+1) - IP_x(i_left) ) *
     2          ( OP_x(1) - IP_x(i_left) ) + IP_y(i_left)
      if ( OP_len .gt. 1 ) then                                           1b
          err_too_big = 0
          err_too_small = 0
          i_out  = 2
C      WRITE(6,9998) OP_start, OP_finish, OP_increm, finish
C9998  FORMAT( ' OP_start/finish/increm finish = ', 4G12.5 )
          increm = OP_increm
1         if ( i_out .le. OP_max ) then                                   2b
              if ( .not. log_p_band ) then                                3b
                  OP_x(i_out) = OP_x(i_out-1) + increm
                  if ( OP_finish - OP_x(i_out-1) .gt. min_increm ) then   4b
                      if ( OP_x(i_out) .gt. OP_finish )
     1                    OP_x(i_out) = OP_finish
                      end if                                              4e
              else                                                        3m
                  OP_x(i_out) = OP_x(i_out-1) * increm
                  if ( OP_finish / OP_x(i_out-1) .gt. min_increm ) then   4b
                      if ( OP_x(i_out) .gt. OP_finish )
     1                    OP_x(i_out) = OP_finish
                      end if                                              4e
                  end if                                                  3e
              if ( OP_x(i_out) .lt. finish ) then                         3b
C      WRITE(6,9997) I_OUT,OP_X(I_OUT),FINISH
C9997  FORMAT( ' OP_X(', I5, ') = ', G13.5, '< FINISH =', G13.5 )
                  i_right = JDHUNT( IP_x, 1, IP_len, USE_PREVIOUS,
     1                              OP_x(i_out) )
                  i_right  = min( max( i_left, i_right ), IP_len - 1 )
                  OP_y(i_out) = ( IP_y(i_right+1) - IP_y(i_right) ) /
     1                          ( IP_x(i_right+1) - IP_x(i_right) ) *
     2                ( OP_x(i_out) - IP_x(i_right) ) + IP_y(i_right)
C              WRITE(6,9999) I_LEFT,  IP_X(I_LEFT),  IP_Y(I_LEFT),
C     1            I_RIGHT,   IP_X(I_RIGHT),   IP_Y(I_RIGHT),
C     2            I_RIGHT+1, IP_X(I_RIGHT+1), IP_Y(I_RIGHT+1),
C     3            I_OUT,     OP_X(I_OUT),     OP_Y(I_OUT)
C9999          FORMAT( ' IP_X-Y(', I5, ') = ', 2G13.5 /
C     1                ' IP_X-Y(', I5, ') = ', 2G13.5 /
C     1                ' IP_X-Y(', I5, ') = ', 2G13.5 /
C     2                ' OP_X-Y(', I5, ') = ', 2G13.5 / )
c
c         ADAPTIVE INTERPOLATION
c         **********************
c             first, find the peak interpolation
c
                  if ( adapt_interp .and. i_left .le. i_right ) then      4b
                      peak_error = find_error( IP_x, IP_y, i_left,
     1                    i_right, OP_x(i_out-1), OP_y(i_out-1) )
c
c         If the peak error is larger than the permitted minimum,
c           if it was previously too small, make the increment adjustment smaller
c                                too big, make the increment adjustment bigger
c           if increment is at its minimum, accept result & move on (i_out++),
c             otherwise decrease the increment
c           count successive occurences
c
C      WRITE(6,9994) MIN_ERROR, PEAK_ERROR, MAX_ERROR,
C     1              MIN_INCREM, INCREM, MAX_INCREM
C9994  FORMAT( 1X, G12.5, ' < PEAK_ERROR = ', G12.5, ' < ', G12.5 /
C     1        1X, G12.5, ' < INCREM = ', G12.5, ' < ', G12.5 )
                      if ( peak_error .gt. max_error ) then               5b
                          if ( err_too_small .gt. 0 ) then                6b
                              increm_adj = increm_adj / 2.0 + 0.5
C      WRITE(6,9995) 'too big & oscillating; shrink increm_adj ',
C     1    INCREM_ADJ
C9995  FORMAT( 1X, A, G12.5 )
                          else if ( err_too_big .gt. 1 ) then             6m
                              increm_adj = increm_adj * 2.0 - 1.0
C      WRITE(6,9995) 'too big again; expand increm_adj ', INCREM_ADJ
                              end if                                      6e
                          if ( increm .le. min_increm ) then              6b
C      WRITE(6,9995) 'increm at min, accept move on', increm
c                             accept result
                              i_out = i_out + 1
                              i_left = min( i_right + 1, IP_len )
                          else                                            6m
c                             decrease increment
                              if ( .not. log_p_band ) then                7b
                                  increm = min( increm / increm_adj,
     1                                          max_increm )
                              else                                        7m
                                  increm = min( increm **
     1                                ( 1.0 / increm_adj ), max_increm )
                                  end if                                  7e
C      WRITE(6,9995) 'too big; decrease INCREM ', INCREM
                              end if                                      6e
                          err_too_big   = err_too_big + 1
                          err_too_small = 0
c
c         If the peak error is smaller than the desired minimum,
c           if it was previously too big, make the increment adjustment smaller
c                                too small, make the increment adjustment bigger
c           if increment is at its maximum, accept result & move on (i_out++),
c             otherwise increase the increment
c           count successive occurences
c
                      else if ( peak_error .lt. min_error ) then          5m
                          if ( err_too_big .gt. 0 ) then                  6b
                              increm_adj = increm_adj / 2.0 + 0.5
C      WRITE(6,9995) 'too small & oscillating; shrink increm_adj ',
C     1    INCREM_ADJ
                          else if ( err_too_small .gt. 1 ) then           6m
                              increm_adj = increm_adj * 2.0 - 1.0
C      WRITE(6,9995) 'too small again; expand increm_adj ', INCREM_ADJ
                              end if                                      6e
                          if ( increm .ge. max_increm ) then              6b
c                             accept result
                              i_out = i_out + 1
                              i_left = min( i_right + 1, IP_len )
C      WRITE(6,9995) 'increm at max, accept move on', increm
                          else                                            6m
c                             increase increment
                              if ( .not. log_p_band ) then                7b
                                  increm = min( increm * increm_adj,
     1                                          max_increm )
                              else                                        7m
                                  increm = min( increm ** increm_adj,
     1                                          max_increm )
                                  end if                                  7e
C      WRITE(6,9995) 'too small; increase INCREM ', INCREM
                              end if                                      6e
                          err_too_big   = 0
                          err_too_small = err_too_small + 1
c
c         interpolation error is acceptable
c
                      else                                                5m
                          i_out = i_out + 1
                          i_left = min( i_right + 1, IP_len )
                          end if                                          5e
c
c         UNIFORM STEP INTERPOLATION
c         **************************
c
                  else                                                    4m
                      i_out = i_out + 1
                      i_left = min( i_right + 1, IP_len )
                      end if                                              4e
                  goto 1
                  end if                                                  3e
              end if                                                      2e
c             end of loop
          OP_len = i_out - 1
          end if                                                          1e
      return
      end



      REAL FUNCTION FIND_ERROR( X, Y, LEFT, RIGHT, REF_X, REF_Y )
c
c Calculates the maximum difference between the (X,Y) data coordinates and
c the straight line segment defined by the two reference points
c (REF_X(1),REF_Y(1)) and (REF_X(2),REF_Y(2)).
c
c Arguments
c  real    x(*), y(*)   # data coordinates
c  integer left, right  # array indices defining data interval to be checked
c  real    ref_x(2), ref_y(2) # 2 points defining straight line segment
c
c local variables
c  integer
c   i                   # data array index
c  real
c   slope, offset       # slope and ordinate offset of reference line segment
c   min_x, max_x        # minimum & maximum reference abscissa values
c
      include 'stdio.def'
      integer i, left, right
      real    min_x, max_x, offset, slope, ref_x(2), ref_y(2),
     1        x(*), y(*)
c
      find_error = 0.0
      slope  = ( ref_y(2) - ref_y(1) ) / ( ref_x(2) - ref_x(1) )
      offset = ( ( ref_y(2) + ref_y(1) ) -
     1           slope * ( ref_x(2) + ref_x(1) ) ) / 2.0
      max_x  = max( ref_x(1), ref_x(2) )
      min_x  = min( ref_x(1), ref_x(2) )
c
C      WRITE(6,9999) LEFT, RIGHT, SLOPE, OFFSET, MIN_X, MAX_X
C9999  FORMAT( ' FIND_ERROR: LEFT, RIGHT = ', 2I5 /
C     1        ' SLOPE, OFFSET, MIN_X, MAX_X ', 4G13.6 )
      if ( left .le. right ) then                                          1b
          do 1 i = left, right                                             2b
              if ( min_x .le. x(i) .and. x(i) .le. max_x ) then            3b
crms                  find_error = find_error +
crms     1                ( abs( x(i) * slope + offset - y(i) ) ) **2
                  find_error = max( find_error,
     1                abs( x(i) * slope + offset - y(i) ) )
C      WRITE(6,9998) I, X(I), Y(I), FIND_ERROR
C9998  FORMAT( ' X/Y(', I4, ')= ', 2g13.5, ' FIND_ERROR = ', G12.5 )
                  end if                                                   3e
1             continue                                                     2e
crms          find_error = sqrt( find_error ) / ( right - left + 1 )
          end if                                                           1e
      return
      end



      LOGICAL FUNCTION PARSE_ARGS( PARAM_LIST )
c
c    Parse the program arguments and open the necessary files . The
c program arguments, which are all optional, are
c
c Synopsis :
c   dig_hist [-width   <linear_pass_band>]
c            [-ratio   <log_pass_band>]
c            [-start   <start_abscissa>]
c            [-finish  <finish_abscissa>]
c            [-length  <OP_record_length>]
c            [-min_err <minimum peak interpolation error>]
c            [-max_err <maximum peak interpolation error>]
c            [-min_inc <minimum interpolation abscissa increment>]
c            [-max_inc <maximum interpolation abscissa increment>]
c            [-inc_adj <interpolation abscissa increment adjustment factor]
c            [-adaptiv]  # adaptive step in interpolated data
c            [-unif]     # uniform step in interpolated data
c            [-nograph]
c            [-graph]
c            [-help    <help_file>]
c            [-data    <input_file>]
c            [-output  <output_file>]
c            [-param   <command_file>]
c
c arg_str arg_type var_type var_index var_value
c -width   prefix  numeric  WIDTH      <linear_pass_band>
c -ratio   prefix  numeric  RATIO      <log_pass_band>
c -start   prefix  numeric  OP_START   <start_abscissa>
c -finish  prefix  numeric  OP_FINISH  <finish_abscissa>
c -length  prefix  numeric  OP_LENGTH  <OP_record_length>
c -min_err prefix  numeric  ERROR_MIN  <minimum peak interpolation error>
c -max_err prefix  numeric  ERROR_MAX  <maximum peak interpolation error>
c -min_inc prefix  numeric  INCREM_MIN <minimum interpolation abscissa increment>
c -max_inc prefix  numeric  INCREM_MAX <maximum interpolation abscissa increment>
c -inc_adj prefix  numeric  ADJ_INCREM <interpolation increm adjustment factor>
c -adaptiv switch  numeric  ADAPTIVE   YES
c -unif    switch  numeric  ADAPTIVE   NO
c -graph   switch  numeric  GRAPHICS   YES
c -nograph switch  numeric  GRAPHICS   NO
c -help    prefix file_name HELP_LU    <help_file_logical_unit_number>
c -data    prefix file_name INPUT_LU   <input_file_logical_unit_number>
c -output  prefix file_name OUTPUT_LU  <output_file_logical_unit_number>
c -param   prefix file_name PARAM_LU   <parameters_file_logical_unit_number>
c
c
c Arguments
c  real param_list(PARAM_LEN)   # list of returned parameters
c  index name        description                           min    max   default       units
c    1 = WIDTH       width of linear smoothing pass band    0     100     3
c                    expressed as percentage of abscissa range,
c    2 = RATIO       width of log smoothing pass band       1     100     1.2
c                    right_edge_of_band = left_edge_of_band * log_pass_band
c    3 = OP_START    first abscissa value of output      Min_I/P Max_I/P Min_I/P
c    4 = OP_FINISH   last abscissa value of output       Min_I/P Max_I/P Max_I/P
c    5 = OP_LENGTH   length of output data record           1     4096    100
c    6 = ERROR_MIN   minimum interpolation error            0.01     1    0.1
c    7 = ERROR_MAX   maximum interpolation error            0.1     10    1.0
c    8 = INCREM_MIN  minimum abscissa increment             0.01     1    0.1
c    9 = INCREM_MAX  maximum abscissa increment             1.0    100   10.0
c   10 = ADJ_INCREM  increment adjustment factor            1.02     4    1.414
c   11 = ADAPTIVE     YES=adaptive, NO=uniform interpolation
c   12 = GRAPHICS     YES or NO
c   13 = HELP_LU      help is written to this file (default = stderr)
c   14 = INPUT_LU     default = stdin
c   15 = OUTPUT_LU    default = stdout
c   16 = PARAM_LU     parameters file; not used externally
c
c    parameters 6 -- 10 are adaptive interpolation parameters
c  minimum/maximum interpolation error is the minimum/maximum difference
c  between filtered ordinates and interpolation straight line segments,
c  expressed as percentage of the ordinate range.
c  minimum & maximum abscissa increment are multiples of the 'uniform' increment
c
c
c integer
c  abbr_len(LIST_LEN) # minimum length of abbreviation for each arg
c  arg_code(LIST_LEN) # specifies arg type PREFIX_NUM, PREFIX_FIL, YES, NO
c  arg_index(LIST_LEN)# specifies which parameter each arg controls
c  arg_ptr            # points to next arg to be fetched
c  file_lun           # logical unit number of file to be opened
c  i                  # array index
c  par_lu             # logical unit number for parameters file
c  n_args             # total number of arguments
c  iargc              # function : argument count
c  new_lun            #          : find new logical unit number
c  str_length         #          : length of a string
c logical
c  get_param          # function : get next param from args or param file
c  is_abbr            #          : arg1 is abbreviation on arg2 ?
c real
c  param_defaults(PARAM_LEN) # default values for all parameters
c character
c  arg_names(LIST_LEN)*8     # list of valid arguments
c  arg_str*16         # program argument
c
c
      include 'stdio.def'
      include 'dig_hist.def'
      integer    LIST_LEN,
     1           PREFIX_NUM,    PREFIX_NFL,    PREFIX_OFL
      parameter( LIST_LEN=18,
     1           PREFIX_NUM=-1, PREFIX_NFL=-2, PREFIX_OFL=-3 )
c
      integer abbr_len(LIST_LEN), arg_code(LIST_LEN),
     1        arg_index(LIST_LEN), arg_ptr, file_lun, i, n_args,
     2        iargc, new_lun, par_lu, str_length
      logical get_param, is_abbr
      real    param_list(PARAM_LEN), param_defaults(PARAM_LEN)
      character arg_names(LIST_LEN)*8, arg_str*16
c
c
      data arg_names / '-width  ', '-ratio  ', '-start  ', '-finish ',
     1     '-length ', '-min_err', '-max_err', '-min_inc', '-max_inc',
     2     '-inc_adj', '-adaptiv', '-uniform', '-graph  ', '-nograph',
     3     '-help   ', '-data   ', '-output ', '-param  ' /
      data abbr_len / 2, 2, 2, 2, 2, 6, 6, 6, 6,
     1                2, 2, 2, 2, 2, 2, 2, 2, 2 /
      data arg_index / WIDTH,      RATIO,      OP_START,   OP_FINISH,
     1     OP_LENGTH,  ERROR_MIN,  ERROR_MAX,  INCREM_MIN, INCREM_MAX,
     2     ADJ_INCREM, ADAPTIVE,   ADAPTIVE,   GRAPHICS,   GRAPHICS,
     3     HELP_LU,    INPUT_LU,   OUTPUT_LU,  PARAM_LU /
      data arg_code /  PREFIX_NUM, PREFIX_NUM, PREFIX_NUM, PREFIX_NUM,
     1     PREFIX_NUM, PREFIX_NUM, PREFIX_NUM, PREFIX_NUM, PREFIX_NUM,
     2     PREFIX_NUM, YES,        NO,         YES,        NO,
     3     PREFIX_NFL, PREFIX_OFL, PREFIX_NFL, PREFIX_OFL /
c
c         WIDTH,        RATIO,         OP_START,     OP_FINISH,
c         OP_LENGTH,    ERROR_MIN,     ERROR_MAX,    INCREM_MIN,
c         INCREM_MAX,   ADJ_INCREM,    ADAPTIVE,     GRAPHICS,
c         HELP_LU,      INPUT_LU,      OUTPUT_LU,    PARAM_LU
      data param_defaults /
     1    3.0, 1.2,  -1.0e37, 1.0e37,
     2  100.0, 0.1,   1.0,    0.1,
     3   10.0, 1.414, 0.0,    0.0,
     4    0.0, 5.0,   6.0,    0.0 /
c
c         set the default values
c
      do 1 i = 1, PARAM_LEN
          param_list(i) = param_defaults(i)
1         continue
c
c         Examine each argument in turn
c
c     start of loop
      par_lu  = 0
      arg_ptr = 1
      n_args = iargc()
2     if ( arg_ptr .le. n_args .or. par_lu .gt. 0 ) then                   1b
c         get argument
          if ( get_param( par_lu, arg_ptr, arg_str, n_args ) ) then        2b
3             do 4 i = 1, LIST_LEN                                         3b
c                 if argument is an acceptable abbreviation
                  if ( is_abbr( arg_str, arg_names(i), abbr_len(i) ) )
     1                    then                                             4b
                      if ( arg_code(i) .eq. PREFIX_NUM ) then              5b
c                     argument prefixes a numeric value
                          if ( get_param( par_lu, arg_ptr, arg_str,
     1                                    n_args ) ) then                  6b
                              read( arg_str, *, ERR=5 )
     1                            param_list(arg_index(i))
                              end if                                       6e
                      else if ( arg_code(i) .eq. PREFIX_NFL .or.
     1                          arg_code(i) .eq. PREFIX_OFL ) then         5m
c                         argument prefixes a file name
                          if ( arg_code(i) .eq. PREFIX_NFL ) then          6b
                              param_list(arg_index(i)) = STD_OUT
                          else if ( arg_code(i) .eq. PREFIX_OFL ) then     6m
                              param_list(arg_index(i)) = STD_IN
                              end if                                       6e
                          if ( get_param( par_lu, arg_ptr, arg_str,
     1                                    n_args ) ) then                  6b
                              if ( arg_str(1:1) .eq. '-' )
     1                            goto 7
                              file_lun = new_lun()
                              open( UNIT=file_lun,    FILE=arg_str,
     1                            STATUS='unknown', ACCESS='sequential',
     2                            FORM='formatted', ERR=9 )
                              if ( arg_index(i) .eq. PARAM_LU )
     1                            par_lu = file_lun
                              param_list(arg_index(i)) = file_lun
                              end if                                       6e
                      else                                                 5m
c                         argument is a switch with pre-set value
                          param_list(arg_index(i)) = arg_code(i)
                          end if                                           5e
                      goto 2
                      end if                                               4e
4                 continue                                                 3e
                  end if                                                   2e
              goto 2
c
c     ERROR TRAPS
c
c         cannot decode numeric value of argument
5             write( STD_ERR, 6 ) arg_str(1:str_length(arg_str)),
     1                            arg_names(i)
6             format( ' Cannot decode number ', a, ' for arg ', a )
              goto 2
c
c         file name expected after argument
7             if ( arg_index(i) .eq. HELP_LU ) then                        2b
                  param_list(arg_index(i)) = STD_OUT
              else                                                         2m
                  write( STD_ERR, 8 ) arg_names(i),
     1                                arg_str(1:str_length(arg_str))
8                 format( ' Arg ', a, ' expects file name, not ', a )
                  end if                                                   2e
              goto 3
c
c         cannot open file
9             write( STD_ERR, 10 ) arg_names(i)(2:),
     1                             arg_str(1:str_length(arg_str))
10            format( ' Cannot open ', a, 'file ', a )
              goto 2
          end if                                                           1e
      parse_args = param_list(HELP_LU) .eq. real( STD_ERR )
      return
      end



      LOGICAL FUNCTION IS_ABBR( ABBREV, TEST, MIN_LEN )
c
c       EQual ABBReviation
c     Tests if the character string CSTR is a legal abbreviation of the
c character string CTEST .
c
c Arguments
c  character abbrev*(*), # operand character string               INPUT
c            test*(*)    # test string                            INPUT
c  integer   min_len     # minimum number of characters allowed   INPUT
c                        # in the abbreviation
c Variables
c  integer i,          # indexing counter
c          finish      # first letter in a word
c          str_length  # function: length of string
c
      include   'stdio.def'
      integer   i, finish, min_len, str_length
      character abbrev*(*), test*(*)
c
      finish = max( str_length( abbrev ), 1 )
      if ( finish .gt. str_length( test ) ) then
          is_abbr = .FALSE.
      else
          is_abbr = .TRUE.
          i = 1
1         if ( i .le. finish .and. is_abbr ) then
              is_abbr = is_abbr .and. abbrev(i:i) .eq. test(i:i)
              i = i + 1
              goto 1
              end if
        is_abbr = is_abbr .and. i .gt. min_len
        end if
      return
      end



      LOGICAL FUNCTION GET_PARAM( PAR_LU, ARG_PTR, ARG_STR, N_ARGS )
c
c Get parameter from program argument (if PAR_LU=0) or parameter file
c (if PAR_LU<>0).
c
c Arguments
c  integer   par_lu       # logical unit of parameter file
c            arg_ptr      # program argument number
c  character arg_str*(*)  # parameter
c  integer   n_args       # number of program arguments
c
c local variables
c  logical
c   get_token     # function : get token from opened file
c
      include   'stdio.def'
      integer   par_lu, arg_ptr, n_args
      logical   get_token
      character arg_str*(*)
c
      get_param = .FALSE.
      if ( par_lu .gt. 0 ) then
          if ( get_token( par_lu, arg_str ) ) then
              get_param = .TRUE.
          else
              close( UNIT=par_lu )
              par_lu = 0
              end if
          end if
      if ( par_lu .eq. 0 .and.
     1     0 .le. arg_ptr .and. arg_ptr .le. n_args ) then
          call getarg( arg_ptr, arg_str )
          arg_ptr = arg_ptr + 1
          get_param = .TRUE.
          end if
      return
      end



      LOGICAL FUNCTION GET_TOKEN( LUN, TOKEN )
c
c Extract the next token from an opened file with logical unit number
c LUN, place it in TOKEN and return a value of TRUE. The maximum line
c length is 128 characters. If no tokens are found in the file the
c returned value is FALSE.
c
c Arguments
c  integer   lun          # INPUT  # logica unit of opened file
c  character token*(*)    # OUTPUT # returned token
c
c local variables
c  integer
c   len_saved         # non blamk length of the saved string
c   start, finish     # start & finish indices in the saved string of the token
c   str_length        # function : non-blank length of string
c  character
c   saved*128         # saved string
c
      include 'stdio.def'
c
      integer lun, len_saved, start, finish, str_length
      character saved*128, token*(*)
c
      save saved, len_saved, start
      data saved / ' ' /
      data start, len_saved / 1, 0 /
c
c   start of main loop
1     if ( len_saved .le. 0 .or. start .gt. len_saved ) then
          read( lun, 2, END=5, ERR=5 ) saved
2         format( a )
          if ( saved(1:1) .eq. ';' ) goto 1
          len_saved = str_length( saved )
          start = 1
          end if
c
3     if ( start .le. len_saved ) then
          if ( saved(start:start) .eq. ' ' ) then
              start = start + 1
              goto 3
              end if
          finish = start
4         if ( finish .lt. len_saved ) then
              if ( saved(finish+1:finish+1) .ne. ' ' ) then
                  finish = finish + 1
                  goto 4
                  end if
              end if
          token = saved(start:finish)
          get_token = .TRUE.
          start = finish + 1
          return
          end if
      goto 1
c   end of main loop
5     get_token = .FALSE.
      return
      end



      LOGICAL FUNCTION READ_RECORD( INPUT_LU, OUTPUT_LU, X, Y, NPTS,
     1    MAX_PTS )
c
c  Reads a single record from the data file with logical unit number
c  INPUT_LU . Initial lines beginning with ';' are copied to the file
c  with logical unit number OUTLU . The following lines are interpreted
c  as (X,Y) coordinate pairs until the next comment line is encountered .
c  This new comment line is stored in a buffer until READ_RECORD is used
c  again, and it is then copied to the output file .
c
c Arguments
c  integer input_lu     # logical unit number of input file   INPUT
c          output_lu    # logical unit number of output file  INPUT
c  real    x(*),        # abscissa values                    OUTPUT
c          y(*)         # ordinate values                    OUTPUT
c  integer npts,        # number of input data coordinates   OUTPUT
c          max_pts      # maximum allowed size of data record INPUT
c
c logical variables
c   buffer_empty        # comment storage buffer is empty
c integer variables
c   columns             # number of items in a non-comment line
c   line_len            # length of a comment line
c   jljstn              # function : left justify & count columns
c   str_length          #          : length of a string
c character variables
c   Cbuffer *80         # comment buffer
c   Cline *80           # line read from input file
c
      include 'stdio.def'
      integer input_lu, output_lu, npts, max_pts
      real x(*), y(*)
c
      logical buffer_empty
      integer columns, line_len, jljstn, str_length
      character Cbuffer*81, Cline*81
      data buffer_empty / .TRUE. /
c
c             if the buffer is not empty flush it to output file
c
      npts = 0
      if ( .not. buffer_empty ) then
          write( output_lu, 1, ERR=10 ) Cbuffer(1:str_length(Cbuffer))
1         format( 1x, a )
          buffer_empty = .TRUE.
          end if
c
c             copy initial comments to output file and skip blank lines
c
c     start of loop
2         read( input_lu, 3, ERR=8, END=7 ) Cline
3         format( a )
          line_len = jljstn( Cline, columns )
          if ( line_len .le. 0 ) then
              goto 2
          else if ( Cline(1:1) .eq. ';' ) then
              write( output_lu, 1, ERR=10 ) Cline(1:line_len)
              goto 2
              end if
c         end of loop
c
c             read the data until the next comment is found
c
c     start of loop
4         if ( jljstn( Cline, columns ) .le. 0 ) then
              read( input_lu, 3, ERR=8, END=7 ) Cline
              goto 4
              end if
          if ( Cline(1:1) .ne. ';' ) then
              if ( columns .le. 2 ) then
                  read( Cline, *, ERR=5 ) x(npts+1), y(npts+1)
                  npts = npts + 1
                  if ( npts .gt. max_pts ) goto 7
                  end if
              read( input_lu, 3, ERR=8, END=7 ) Cline
              goto 4
c
5             write( STD_ERR, 6 )
6             format( ' WARNING: Cannot decode input data; skipped' )
              read( input_lu, 3, ERR=8, END=7 ) Cline
              goto 4
              end if
c         end of loop
c
c             fall through to here if a comment line is found
c
      Cbuffer = Cline
      buffer_empty = .FALSE.
7     read_record = npts .gt. 0
      return
c
c             trap : error reading input file
c
8     write( STD_ERR, 9 )
9     format( ' ABORT: Cannot read input file' )
      read_record = .FALSE.
      return
c
c             trap : error writing to output file
c
10    write( STD_ERR, 11 )
11    format( ' ABORT: Cannot write to output file' )
      read_record = .FALSE.
      return
      end


      SUBROUTINE COMP_DUP( ARRAY, IN_ROWS, OUT_ROWS, COMP_COL,
     1                     ROWS, COLUMNS, TOLERANCE )
c
c    COMPress out DUPlicate rows
c Compresses out the rows of the ARRAY, which contain a value in the
c column COMP_COL duplicating the previous row .
c
c Arguments
c  real    array(rows,columns) # IN / OUT # data array
c  integer in_rows             # IN # number of rows to be checked
c          out_rows           # OUT # number of rows remaining after compression
c          comp_col            # IN # column to be compressed
c          rows, columns       # IN # size of array
c  real tolerance              # IN # minimum difference for 2 values to be
c                                   # considered different
c                                   # tolerance = ABS( x - y) /
c                                   #             ( ABS( x ) + ABS( y ) )
c integer variables
c  col_index      # column index used for copying rows
c  dest_index     # row index of the destination row
c  ref_index      # row index of the reference row
c  src_index      # row index of the source row
c real variables
c  tol            # absolute value of tolerance
c
      include 'stdio.def'
c
      integer columns, col_index, comp_col, dest_index, in_rows,
     1        out_rows, ref_index, rows, src_index
      real array(rows,columns), tol, tolerance
c
c         check the arguments
c
      tol = abs( tolerance ) * 0.5
      if ( rows    .lt. 2 .or. columns .lt. 1    .or.
     1     in_rows .lt. 2 .or. in_rows .gt. rows )
     2    return
c
c         The source row is compared with the reference row . If there
c         is a significant difference, the source row is copied to the
c         destination row which immediately follows the reference row .
c
      ref_index  = 1
      dest_index = ref_index + 1
      do 2 src_index = 2, in_rows
          if ( abs( array(src_index,comp_col) -
     1              array(ref_index,comp_col) ) /
     2         ( abs( array(src_index,comp_col) ) +
     3           abs( array(ref_index,comp_col) ) ) .gt. tol ) then
c
c         Don't copy if the source row is the same as the destination row
c
              if ( src_index .ne. dest_index ) then
                  do 1 col_index = 1, columns
                      array(dest_index,col_index) =
     1                    array(src_index,col_index)
1                     continue
                  end if
              ref_index = dest_index
              dest_index = dest_index + 1
              end if
2          continue
      out_rows = ref_index
      return
      end


      SUBROUTINE GIVE_HELP( HELP_LU, MAX_LEN, DEF_OP_LEN )
c
c  Write help message to the file with logical unit number HELP_LU .
c
c  help_lu   = logical unit number of output file
c  max_len    = maximum input and output record lengths
c  def_op_len = default length of output record
c
      include 'stdio.def'
      integer help_lu, max_len, def_op_len
      write( help_lu, 1, ERR=5 )
1     format( '    DIG_HIST' //
     1 ' Smooth a data stream and interpolate a new stream.' //
     2 '   The input data is assumed to take the form of two columns' /
     3 ' of numbers with comment lines acting as record separators .' /
     4 ' Each record is processed separately. Separating comments are' /
     5 ' copied to the output file. Comments begin with a semicolon .'//
     6 ' First, the abscissae (first column) of the input are sorted' /
     7 ' into ascending order. A moving average (uniform weight)' /
     8 ' filter is then applied to the data to provide smoothing. A' /
     9 ' new data stream is then interpolated from this sorted and' /
     a ' filtered data and written to the output .' / )

      write( help_lu, 2, ERR=5 )
2     format( '   Synopsis :' /
     1 ' dig_hist [-width  <linear pass band: % of abscissa range>]' /
     2 10x, '[-ratio  <log pass band: abscissa range ratio>]' /
     3 10x, '[-start  <start abscissa of output>]' /
     4 10x, '[-finish <finish abscissa of output>]' /
     5 10x, '[-length <number of output record values>]' /
     6 10x, '[-min_err <minimum peak interpolation error>]' /
     8 10x, '[-max_err <maximum peak interpolation error>]' /
     9 10x, '[-min_inc <minimum interpolation abscissa increment>]' )

      write( help_lu, 3, ERR=5 )
3     format(
     1 10x, '[-max_inc <maximum interpolation abscissa increment>]' /
     2 10x, '[-inc_adj <abscissa increment adjustment factor]' /
     3 10x, '[-adaptiv]  # adaptive step in interpolated data' /
     4 10x, '[-unif]     # uniform step in interpolated data' /
     5 10x, '[-nograph]' / 10x, '[-graph]' /
     6 10x, '[-help    <help_file>]' / 10x, '[-data    <input_file>]' /
     7 10x, '[-output  <output_file>]' /
     8 10x, '[-param   <command_file>]' )
      return
c
5     write( STD_ERR, 6 )
6     format( 'ABORT : Cannot write to help file' )
      return
      end



      SUBROUTINE SMVOOTH( X, Y, XY_INCREM, N_POINTS, PASS_BAND,
     1                    LOG_SCALE, X_OUT, OUT, OUT_INCREM )
c MoVing smOOTHing
c
c Moving average smoothing filter for coordinate pairs with nonuniformly
c spaced abscissa values.
c
c   A pass band window of total width PASS_BAND is passed along the data
c coordinates (X,Y) . The data samples within the window are averaged and
c placed in the array OUT . The pass band window is centred on each output
c datum . Near the ends of the data, the width of the passband window is
c reduced so that it does exceed the abscissa limits .
c
c Arguments
c  real    x(*)       # abscissa values of unfiltered input data stream
c          y(*)       # ordinate values of unfiltered input data stream
c  integer xy_increm  # indexing increment for the input data stream
c          n_points   # number of data points
c  real    pass_band  # width of smoothing pass band, expressed as percentage
c                     #    of abscissa range, min=0, max=100, default=3
c  logical log_scale  # log description of band pass window
c  real    x_out(*)   # abscissa values for filtered output data stream
c          out(*)     # filtered output data stream ordinates
c  integer out_increm # indexing increment for the output data stream
c
c Local Variables
c  integer
c   i, j,             # input abscissa & ordinate array indices
c   i_out,            # output array index
c   left, right,      # indices of each end of pass band window
c   xy_last,          # index of last available datum
c   jdhunt            # function # hunt for value in array
c  real
c   half_width        # half width of smoothing pass band window
c  double precision
c   sum               # moving agregate
c
      include 'stdio.def'
      integer i, j, i_out, left, n_points, out_increm, right,
     1        xy_increm, xy_last, jdhunt
      logical log_scale
      real    half_width, pass_band, x(*), y(*), x_out(*), out(*)
      double precision sum
c
c         check arguments
c
      if ( xy_increm .le. 0 ) then                                        1b
          write( STD_ERR, 1 ) 'XY_INCREM', xy_increm
1         format( ' SMVOOTH : ABORT : Illegal ', a, ' = ', i12 )
      else if ( n_points .le. 0 ) then                                    1m
          write( STD_ERR, 1 ) 'N_POINTS', n_points
      else if ( out_increm .le. 0 ) then                                  1m
          write( STD_ERR, 1 ) 'OUT_INCREM', out_increm
      else if ( ( pass_band .le. 1.0 .or. pass_band .gt. 100.0 )
     1          .and. log_scale .or. .not. log_scale .and.
     2        ( pass_band .le. 0.0 .or. pass_band .gt. 100.0 ) ) then     1m
          write( STD_ERR, 2 ) 'PASS_BAND', pass_band
2         format( ' SMVOOTH : ABORT : Illegal ', a, ' = ', g15.7 )
      else                                                                1m
c
          xy_last = ( n_points - 1 ) * xy_increm + 1
          left  = -999
          right = -999
          i_out = 1
c
c         Pass band width is zero, just copy input to output
c
          if ( pass_band .eq. 0.0 ) then                                  2b
              i_out = 1
              do 7 i = 1, xy_last, xy_increm                              3b
                  x_out(i_out) = x(i)
                  out(i_out)   = y(i)
                  i_out = i_out + out_increm
7                 continue                                                3e
c
c           LOG PASS BAND WINDOW
c
          else if ( log_scale ) then                                      2m
              half_width = sqrt( pass_band )
              do 4 i = 1, xy_last, xy_increm                              3b
c
c           find location of the passband window
c
                  if ( x(i) / half_width .lt. x(1) ) then                 4b
                      left = 1
                      right = jdhunt( x, xy_increm, n_points, right,
     1                                x(i) **2 / x(1) )
                  else if ( x(i) * half_width .gt. x(xy_last) ) then      4m
                      left  = jdhunt( x, xy_increm, n_points, left,
     1                                x(i) **2 / x(xy_last) )
                      right = xy_last
                  else                                                    4m
                      left  = jdhunt( x, xy_increm, n_points, left,
     1                                x(i) / half_width )
                      right = jdhunt( x, xy_increm, n_points, right,
     1                                x(i) * half_width )
                      end if                                              4e
c                  call ilgline( 14, 0, 1 )
c                  call ilgdata( x(left), y(left), 1, 1, 0 )
c                  call ilgdata( x(right), y(right), 1, 1, 0 )
c
c           average within the passband window
c
                  if ( left + xy_increm .gt. right ) then                 4b
                      x_out(i_out) = x(i)
                      out(i_out)   = y(i)
                  else                                                    4m
                      sum = 0.0d0
                      do 3 j = left + xy_increm, right, xy_increm         5b
                          sum = sum + ( y( j ) + y( j - xy_increm ) ) *
     1                                log( x( j ) / x( j - xy_increm ) )
3                         continue                                        5e
                      x_out(i_out) = sqrt( x(left) * x(right) )
                      out(i_out) = 0.5 * sum / log( x(right) / x(left) )
                      end if                                              4e
c                  write( STD_OUT, 10 ) ' log', left, right,
c     1              ( right - left ) / xy_increm + 1, i_out, out(i_out)
c10                format( a, i5, '-', i4, ' n = ', i4, ' out(', i4,
c     1                ')=', g12.5 )
c                  call ilginvs()
c                  call ilgdata( x(left), y(left), 1, 1, 0 )
c                  call ilginvs()
c                  call ilgdata( x(right), y(right), 1, 1, 0 )
c                  call ilgline( 5, 0, 1 )
c                  call ilgdata( x_out(i_out), out(i_out), 1, 1, 0 )
                  i_out = i_out + out_increm
4                 continue                                                3m
c
c           LINEAR PASS BAND WINDOW
c           find location of the passband window
c
          else                                                            2m
              half_width = ( x(xy_last) - x(1) ) * pass_band * 0.005
              do 6 i = 1, xy_last, xy_increm                              3b
                  if ( x(i) - half_width .lt. x(1) ) then                 4b
                      left = 1
                      right = jdhunt( x, xy_increm, n_points, right,
     1                                2.0 * x(i) - x(1) )
                  else if ( x(i) + half_width .gt. x(xy_last) ) then      4m
                      left  = jdhunt( x, xy_increm, n_points, left,
     1                                2.0 * x(i) - x(xy_last) )
                      right = xy_last
                  else                                                    4m
                      left  = jdhunt( x, xy_increm, n_points, left,
     1                                x(i) - half_width )
                      right = jdhunt( x, xy_increm, n_points, right,
     1                                x(i) + half_width )
                      end if                                              4e
c                  call ilgline( 14, 0, 1 )
c                  call ilgdata( x(left), y(left), 1, 1, 0 )
c                  call ilgdata( x(right), y(right), 1, 1, 0 )
c
c           average within the passband window
c
                  if ( left + xy_increm .gt. right ) then                 4b
                      x_out(i_out) = x(i)
                      out(i_out) = y(i)
                  else                                                    4m
                      sum = 0.0d0
                      do 5 j = left + xy_increm, right, xy_increm         5b
                          sum = sum + ( y( j ) + y( j - xy_increm ) ) *
     1                                ( x( j ) - x( j - xy_increm ) )
5                         continue                                        5e
                      x_out(i_out) = ( x(left) + x(right) ) / 2.0
                      out(i_out) = 0.5 * sum / ( x(right) - x(left) )
                      end if                                              4e
c                  write( STD_OUT, 10 ) ' lin', left, right,
c     1              ( right - left ) / xy_increm + 1, i_out, out(i_out)
c                  write( STD_OUT, 11) i, i_out, x(i), x_out(i_out)
c11                format( 1x, i4, '=', i4, g12.5, '=', g12.5 )
c                  call ilginvs()
c                  call ilgdata( x(left), y(left), 1, 1, 0 )
c                  call ilginvs()
c                  call ilgdata( x(right), y(right), 1, 1, 0 )
c                  call ilgline( 5, 0, 1 )
c                  call ilgdata( x_out(i_out), out(i_out), 1, 1, 0 )
                  i_out = i_out + out_increm
6                 continue                                                3e
              end if                                                      2e
          end if                                                          1e
      return
      end




