c
c    Copyright (C) 1997  Peter Vernon Lanspeary
c
c    This program is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 2 of the License, or
c    (at your option) any later version.
c
c    This program is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program; if not, write to the Free Software
c    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
c


      PROGRAM SMOOTH_PROFILE_DATA
c
c  Generate and plot smooth spline curves from the input data
c
c  usage :
c    SMOOTH <n_knots> <order> [-e [<bias>] ] <input_file> [<output_file>]
c
c The source data is read from file <input_file> which consists of at least 2
c columns of numbers . The first column contains the abscissa values and every
c other column is a set of ordinate values . The input data is smoothed by
c fitting B-spline curves to each set of ordinates over the whole (abscissa)
c domain. The smoothed curves are written to the file <output_file> using the
c same number of columns as in <input_file> . <n_knots> is the number of
c spline knots. <order> is the order of the spline polynmials . The optional
c argument "-e" specifies logarithm transformation of abscissa values before
c fitting the spline . If present, <bias> specifies a bias added to the
c abscissae before the transformation .
c
c IMSL Version 10 is required
c
c constants
c   Hmax_knots=50   # maximum number of knots
c   Hmin_order=1    # minimum order of B-spline
c   Hmax_order=5    # maximum order of B-spline
c   Hmax_data=10000 # available data space
c   Hmax_rows=1000  # maximum number of rows in data array
c   Hn_smooth=200   # number of coordinates in a smooth curve
c variables
c  logical
c   log_dist        # abscissae are distributed logarithmically
c  integer
c   column          # data column number
c   ip_lu           # input file logical unit number
c   op_lu           # output file logical unit number
c   n_args          # number of command line arguments
c   n_knots         # number of knots
c   n_rows          # number of data rows
c   n_columns       # number of data columns
c   order           # order of B-spline
c   row             # data row number
c   row_index       # array index pointing to a row
c   first_file      # argument number of the first file argument
c   ARGtoINT        # function : convert command line argument to integer
c   iargc           # function : number of command line arguments
c   open_file       # function : open file
c   new_lun         # function : get new logical unit number
c   str_length      #          : non-blank length of string
c  real
c   bias                # bias for the logarithmic abscissa tranformation
c   array(Hmax_data)    # data array
c   weights(Hmax_rows)  # array of weights for use in calculation of spline
c   log_x(Hmax_rows)    # natural log of abscissae from data array
c   knots(Hmax_knots)   # abscissa of spline knots
c   smooth(Hmax_data)   # array for smooth curve coordinates
c   ordinate(Hmax_rows) # list of ordinates for calculation of spline
c   bsplines(Hmax_knots-Hmin_order) # B-spline coefficients
c   knot_step,          # increment of knot abscissae
c   log_min, log_max    # min & max of LOG_X values
c   x_step              # geometric abscissa increment for smooth curves
c   BSVAL               # IMSL function : B-spline evaluation
c
c  character
c   Carg*2          # possible 3rd argument
c   Cip_file*64     # name of input file
c   Cop_file*64     # name of output file
c   Ctitle*32       # graph title
c
      include 'stdio.def'
      integer Hmax_knots, Hmin_order, Hmax_order, Hmax_data,
     1    Hmax_rows, Hn_smooth
      parameter( Hmax_knots=50, Hmin_order=1, Hmax_order=5,
     1    Hmax_data=10000, Hmax_rows=1000, Hn_smooth=200 )
c
      logical log_dist
      integer column, ip_lu, n_args, n_columns, n_knots, n_rows, op_lu,
     1    order, row, row_index, first_file,
     2    ARGtoINT, iargc, new_lun, open_file, str_length
      real bias, array(Hmax_data), weights(Hmax_rows), log_x(Hmax_rows),
     1    knots(Hmax_knots), smooth(Hmax_data), ordinate(Hmax_rows),
     2    bsplines(Hmax_knots-Hmin_order),
     3    knot_step, log_min, log_max, x_step, BSVAL
      character Carg*64, Cip_file*64, Cop_file*64, Ctitle*32
c
      data log_dist / .FALSE. /
      data n_knots, order / 10, 3 /
      data bias / 0.0 /
      data Carg, Cip_file, Cop_file / ' ', ' ', ' ' /
c
c           INTERPRET INPUT
c           complain about command line arguments if there are fewer than 2
c
      n_args = iargc()
      if ( n_args .lt. 3 ) then
          write( STD_ERR, 1 )
1         format( ' usage' / ' SMOOTH <n_knots> <order> [-e [<bias>] ]',
     1            ' <input_file> [<output_file>]' )
          end if
c
c           determine number of spline knots and order of Bspline
c
      order = ARGtoINT( 2, 'Order of Bspline', Hmin_order, Hmax_order )
      n_knots = ARGtoINT( 1, 'Number of knots', order + 1, Hmax_knots )
c
c           determine if exponentially distributed abscissae are required
c
      if ( n_args .ge. 3 ) then
          call getarg( 3, Carg )
          if ( Carg(1:1) .eq. '-' ) then
              first_file = 4
              log_dist = Carg .eq. '-e'
              if ( n_args .ge. 4 ) then
                  call getarg( 4, Carg )
                  bias = 0.0
                  read( Carg, *, ERR=15 ) bias
                  first_file = 5
15                continue
                  end if
          else
              first_file = 3
              log_dist = .FALSE.
              end if
          end if
c
c           open input file -- get name from command line if possible
c
      if ( n_args .lt. first_file ) then
          Cip_file = '?'
      else
          call getarg( first_file, Cip_file )
          end if
      ip_lu = open_file( new_lun(), Cip_file, 'old', 0 )
      if ( ip_lu .le. 0 ) then
          write( STD_ERR, 2 ) Cip_file(1:str_length(Cip_file))
2         format( ' SMOOTH : Cannot open', a )
          stop
          end if
c
c           open output file -- get name from command line if possible
c
      if ( n_args .lt. first_file ) then
          Cop_file = ' '
      else
          call getarg( first_file + 1, Cop_file )
          end if
      op_lu = open_file( new_lun(), Cop_file, 'unknown', 0 )
      if ( op_lu .le. 0 ) then
          write( STD_ERR, 2 ) Cop_file(1:str_length(Cop_file))
          stop
          end if
c
c           read the data file
c
      call read_data( ip_lu, array, Hmax_data, n_rows, n_columns )
c
c           INITIALIZE ARRAY DATA
c           calculate limits of abscissa, log-abscissa and weights
c
      row_index = 1 + n_columns
      if ( log_dist ) then
          log_x(1) = log( array(1) + bias )
      WRITE(6,9999) 'LOG 1'
9999  FORMAT( 1X, A )
      else
          log_x(1) =  array(1)
      WRITE(6,9999) 'LIN 1'
          end if
      weights(1) = 1.0
      log_min = log_x(1)
      log_max = log_min
      do 3 row = 2, n_rows
          if ( log_dist ) then
              log_x(row) = log( array( row_index ) + bias )
      WRITE(6,9999) 'LOG 2'
          else
              log_x(row) = array( row_index )
      WRITE(6,9999) 'LIN 2'
              end if
          weights(row) = 1.0
          log_min = min( log_min, log_x(row) )
          log_max = max( log_max, log_x(row) )
          row_index = row_index + n_columns
3         continue
c
c           calculate knot positions
c
      knot_step = ( log_max - log_min ) / ( n_knots - 2 * order - 1 )
      knots( 1 ) = log_min - order * knot_step
      do 4 row = 2, n_knots
          knots( row ) = knots( row - 1 ) + knot_step
4         continue
c
c           calculate smooth curve abscissae
c
      if ( log_dist ) then
      WRITE(6,9999) 'LOG 3'
          smooth(1) = exp( log_min ) - bias
          x_step = ( ( exp(log_max) - bias ) / smooth(1) ) **
     1             ( 1.0 / real( Hn_smooth - 1 ) )
          do 5 row_index = 1 + n_columns,
     1                     1 + n_columns * ( Hn_smooth - 1 ), n_columns
              smooth(row_index) = smooth( row_index - n_columns ) *
     1            x_step
5             continue
      else
      WRITE(6,9999) 'LIN 3'
          smooth(1) = log_min
          x_step = ( log_max - log_min ) / real( Hn_smooth - 1 )
          do 6 row_index = 1 + n_columns,
     1                     1 + n_columns * ( Hn_smooth - 1 ), n_columns
              smooth(row_index) = smooth( row_index - n_columns ) +
     1            x_step
6             continue
          end if
c
c           CALCULATE ALL SETS OF SPLINES
c           once around this loop for each column of data
c
      do 11 column = 2, n_columns
c
c           copy from array column into ordinate vector
c
          row_index = column
          do 7 row = 1, n_rows
              ordinate(row) = array( row_index )
              row_index = row_index + n_columns
7             continue
c
c           calculate set of b-splines
c
          call BSLSQ( n_rows, log_x, ordinate, weights,
     1                order, knots, n_knots - order, bsplines )
c
c           calculate the smooth-curve data
c
          if ( log_dist ) then
      WRITE(6,9999) 'LOG 4'
              do 8 row_index = 1, 1 + n_columns * ( Hn_smooth - 1 ),
     1                 n_columns
                  smooth( row_index + column - 1 ) =
     1               BSVAL( log( smooth( row_index ) + bias ),
     2               order, knots, n_knots - order, bsplines )
8                 continue
          else
      WRITE(6,9999) 'LIN 4'
              do 9 row_index = 1, 1 + n_columns * ( Hn_smooth - 1 ),
     1                 n_columns
                  smooth( row_index + column - 1 ) =
     1                BSVAL( smooth( row_index ), order, knots,
     2                       n_knots - order, bsplines )
9                 continue
              end if
c
c           plot data and smooth curve
c
          if ( log_dist ) then
      WRITE(6,9999) 'LOG 5'
              call ilgslaw( 'log', 'lin' )
          else
      WRITE(6,9999) 'LIN 5'
              call ilgslaw( 'lin', 'lin' )
              end if
          call ilgline( 1, 0, 1 )
          call ilgdata( array(1), array(column), n_columns, n_rows, 1 )
          call ilgline( 0, 1, 1 )
          call ilgdata( smooth(1), smooth(column), n_columns,
     1        Hn_smooth, 0 )
          write( Ctitle, 10 ) column
10        format( '****** Column number ', i2, ' ******' )
          call ilgtitl( Ctitle )
11        continue
c
c           WRITE THE SMOOTH CURVE DATA
c
      call ilgwait()
      call ilgclos()
      row_index = 0
      if ( op_lu .ne. STD_OUT ) then
          write( STD_OUT, 12 ) Cop_file(1:str_length(Cop_file))
12        format( ' Writing file ', a )
          end if
      do 14 row = 1, Hn_smooth
          write( op_lu, 13 ) ( smooth( row_index + column ),
     1        column = 1, n_columns )
13        format( 1x, 8f9.4 : / ( 10x, 7f9.4 ) )
          row_index = row_index + n_columns
14        continue
      stop
      end



      INTEGER FUNCTION ARGtoINT( ARG_NO, PROMPT, MIN_VALUE, MAX_VALUE )
c
c Decode the ARG_NOth command line argument as an integer between the limits
c MIN_VALUE and MAX_VALUE . If the total number of arguments is less than
c ARG_NO, the decoded argument exceeds the limits, or the argument cannot be
c decoded, the value is read from the keyboard using JRESP .
c
c Arguments
c  integer arg_no         # command line argument specifier
c  character prompt*(*)   # user prompt string
c  integer min_value      # minimum allowed value
c          max_value      # maximum allowed value
c
c variables
c  integer iargc          # function : number of arguments
c          Jresp          # function : read integer
c  character Carg_str*64  # argument string
c
      include 'stdio.def'
c
      integer arg_no, max_value, min_value, iargc, Jresp
      character Carg_str*64, prompt*(*)
c
      if ( arg_no .gt. iargc() ) then
          write( STD_ERR, 1 ) prompt
1         format( 1x, a, ' ? ', $ )
          ARGtoINT = Jresp( STD_IN, min_value, max_value, 4 )
      else
          call getarg( arg_no, Carg_str )
          read( Carg_str, *, ERR=3 ) ARGtoINT
          if ( min_value .le. ARGtoINT .and. ARGtoINT .le. max_value )
     1        return
          write( STD_ERR, 2 ) ARGtoINT, min_value, max_value
2         format( ' Argument ', i6, ' exceeds limits ', i6, ' and ',
     1        i6 )
3         write( STD_ERR, 1 ) prompt
          ARGtoINT = Jresp( STD_IN, min_value, max_value, 4 )
          end if
      return
      end
