       PROGRAM PROXIMITY_LAW
c
c       Program to produce a wall proximity law which can be used by
c SBHATIA to correct hot-wire air speed readings for conduction of heat
c to the smooth wall . This wall proximity effect is only significant
c for  y_plus < 6 .
c
c Experimental data can be read from files which satisfy the following
c formatting rules
c    1) Any line beginning with a semicolon ';' is a comment . Comments
c       are ignored .
c    2) All non-comment lines contain 2 numbers . Anything after the
c       2 numbers is ignored .
c    3) The first number on a non-comment line is a  `y_plus' value .
c    4) The second number on a non-comment line is a non-dimensionalized
c       `u_plus' measurement or a velocity error `Delta_U_plus' estimate .
c
c Formatting rules for the experimental data output file
c    1) The first line contains a ; followed by the LABEL string
c    2) Comments can then be appended by the user . Each comment line
c       begins with a ';' . ( The last comment line entered by the user
c       contains only a '.' and is not included in the output file ) .
c    3) The data is written as (y_plus,u_data)  coordinate pairs, one on
c       each line .
c
c
c Local variables
c  logical
c   got_velocity, # experimental velocity data is defined
c   got_error,    # experimental error data is defined
c   got_law,      # wall proximity correction law is defined
c   got_friction  # a value for friction velocity is available
c   got_visc      # a value for kinematic viscosity is available
c   vel_to_err,   # function : converts velocity to error
c   equery,       # function : read logical value
c   read_data,    # function : read experimental data
c   make_corr_law # function : calculate correction law
c   vel_to_Nusselt# function : convert velocity error to Nusselt error
c  integer
c   idatlu,       # logical unit no. of data file
c   method,       # method of data correction
c                 # determined by the form of error data
c                 # = UNDEF_METHOD, VELOC_METHOD, HEAT_METHOD, MOD_H_METHOD
c   n_error,      # number of error data points
c   n_speed,      # number of velocity data points
c   data_type,    # VELOC_DATA, ERROR_DATA
c   item,         # item selected from menu
c   Jhw1_limit    # function : control calibration clipping limits
c   Jresp         # function : read integer
c   new_lun       # function : get new logical unit number
c  real
c   p_slope             # inferred slope of sublayer mean velocity profile
c   kinem_visc          # kinematic viscosity
c   friction_vel        # friction_velocity
c   y_speed(MAX_DATA),  # probe locations for velocity profile data
c   speed(MAX_DATA),    # speeds measured for velocity profile data
c   y_error(MAX_DATA),  # probe locations for experimental velocity error
c   error(MAX_DATA),
c   corr_law(MAX_TERMS) # coefficients for the correction law
c   Qresp               # function : read real
c  character
c   Cdata_type(2)*8     # 'velocity', 'error'
c
c
      include 'stdio.def'
      include 'prox_law.def'
      character Cno_data*20, Cno_law*26, Cno_visc*27, Cno_fric*25
      parameter( Cno_data = 'Do not have any data',
     1           Cno_law  = 'Do not have correction law',
     2           Cno_visc = 'Unknown kinematic viscosity',
     3           Cno_fric = 'Unknown friction velocity' )
c
      logical got_velocity, got_error, got_law, got_Nusselt,
     1        got_friction, got_visc, vel_to_err, equery, read_data,
     2        make_corr_law, read_corr_law, vel_to_Nusselt
      integer idatlu, method, n_error, n_speed, data_type,
     1        item, JHW1_limit, Jresp, new_lun
      real y_speed(MAX_DATA), speed(MAX_DATA), y_error(MAX_DATA),
     1     error(MAX_DATA), corr_law(MAX_TERMS), p_slope, kinem_visc,
     2     friction_vel, Qresp
      character Cdata_type(2)*8, Cmethod(0:3)*18
      data Cdata_type / 'velocity', 'error   ' /
      data Cmethod    / 'undefined         ', 'velocity error    ',
     1                  'simple heat flow  ', 'modified heat flow' /
c
      got_velocity = .false.
      got_error    = .false.
      got_law      = .false.
      data_type    = VELOC_DATA
      method       = UNDEF_METHOD
      idatlu       = new_lun()
c
c
c
c             start of main loop
c             ##################
c
c     start loop                                                          1b
c
c             select what to do
c             *****************
c
1         call newscr()
          write( STD_OUT, 2 )  Cdata_type(data_type), Cmethod(method),
     1        kinem_visc, friction_vel
2         format( ' select' /
     z        '     0) Exit' /
     1        '     1) Data type', 22x, a /
     2        '     2) Correction method', 11x, a /
     3        '     3) Kinematic viscosity', 10x, g12.5 /
     4        '     4) Friction velocity', 12x, g12.5 /
     5        '     5) Open hot-wire calibration file' /
     6        ' Experimental data' /
     7        '     6) Read' /
     8        '     7) Edit/plot' /
     9        '     8) Write' )
          write( STD_OUT, 3 )
3         format( ' Convert' /
     1        '     9) Velocity to velocity error' /
     2        '    10) Velocity error to Nusselt error' /
     3        ' Correction law' /
     4        '    11) Make' /
     5        '    12) Plot' /
     6        '    13) Write' /
     7        '    14) Read' /
     8        '    15) Test' /
     9        '    16) Clear ', $ )
          item = Jresp( STD_IN, 0, 16, MAX_TRIES  )
          if ( item .eq. 0 ) then                                         2b
              if ( Equery( STD_IN, ' Are you sure? ' ) ) then             3b
                  stop
                  end if                                                  3e
c
c             set the data type
c             *****************
c
          else if ( item .eq. 1 ) then                                    2m
              data_type = mod( data_type, 2 ) + 1
c
c             set the correction method
c             ************************
c
          else if ( item .eq. 2 ) then                                    2m
              method = mod( method, 3 ) + 1
c
c             set the kinematic viscosity
c             ***************************
c
          else  if ( item .eq. 3 ) then                                   2m
              write( STD_OUT, 5 )
5             format( ' Kinematic viscosity ? ', $ )
              kinem_visc = Qresp( STD_IN, 1.0e-5, 2.0e-5, MAX_TRIES )
              got_visc = .TRUE.
c
c             set the friction velocity
c             *************************
c
          else if ( item .eq. 4 ) then                                    2m
              write( STD_OUT, 6 )
6             format( ' Friction velocity ? ', $ )
              friction_vel = Qresp( STD_IN, 0.01, 2.0, MAX_TRIES )
              got_friction = .TRUE.
c
c             Open hot-wire calibration file
c             ******************************
c                 ( internal routine prompt for file name and record no )
c
          else if ( item .eq. 5 ) then                                    2m
              call ehw1_open( 0, ' ', 0 )
              call jhw1_method( 0 )
              item = JHW1_LIMIT( -6, 0.15 )
              item = JHW1_LIMIT( -7, 2.0 )
c
c             read experimental data
c             **********************
c
          else if ( item .eq. 6 ) then                                    2m
              if ( data_type .eq. VELOC_DATA ) then                       3b
                  got_velocity = read_data( idatlu, y_speed, speed,
     1                                      n_speed, MAX_DATA )
              else if ( data_type .eq. ERROR_DATA ) then                  3m
                  got_error = read_data( idatlu, y_error, error,
     1                                   n_error, MAX_DATA )
                  got_Nusselt = Equery( STD_IN,
     1                'Is this a Nusselt error (otherwise velocity) ?' )
                  end if                                                  3e
              if ( .not. got_velocity )
     1            call print_err( 'Could not read experimental data' )

c
c             edit/plot experimental data & correction law
c             ********************************************
c
          else if ( item .eq. 7 ) then                                    2m
              if ( data_type .eq. VELOC_DATA ) then                       3b
                  if ( got_velocity ) then                                4b
                      call ilgraph( 1, 1, 11, 11, 1 )
                      call sfudge( y_speed, speed, 1, n_speed,
     1                             n_speed, MAX_DATA )
                      call ilgclos
                  else                                                    4m
                      call print_err( Cno_data )
                      end if                                              4e
              else if ( data_type .eq. ERROR_DATA ) then                  3m
                  if ( got_error ) then                                   4b
                      call ilgraph( 1, 1, 11, 11, 1 )
                      call sfudge( y_error, error, 1, n_error,
     1                             n_error, MAX_DATA )
                      if ( got_law ) then                                 5b
                          if ( Equery( STD_IN,
     1                            'Plot correction law also ?' ) ) then   6b
                              call plot_corr_law( idatlu, corr_law,
     1                            .false., 1.0 )
                              end if                                      6e
                          end if                                          5e
                      call ilgclos()
                  else                                                    4m
                      call print_err( Cno_data )
                      end if                                              4e
                  end if                                                  3e
c
c             write experimental data
c             ***********************
c
          else if ( item .eq. 8 ) then                                    2m
              if ( data_type .eq. VELOC_DATA ) then                       3b
                  if ( got_velocity ) then                                4b
                      call write_data( idatlu, y_speed, speed, n_speed,
     1                    1, 1, 'mean velocity profile' )
                  else                                                    4m
                      call print_err( Cno_data )
                      end if                                              4e
              else if ( data_type .eq. ERROR_DATA ) then                  3m
                  if ( got_error ) then                                   4b
                      call write_data( idatlu, y_error, error, n_error,
     1                    1, 1, 'non_dimensional velocity error' )
                  else                                                    4m
                      call print_err( Cno_data )
                      end if                                              4e
                  end if                                                  3e
c
c             Convert velocity to error
c             *************************
c
          else if ( item .eq. 9 ) then                                    2m
              if ( .not. got_velocity ) then                              3b
                  call print_err( Cno_data )
              else                                                        3m
                  if ( got_error ) then                                   4b
                      if ( Equery( STD_IN, 'Will overwrite existing' //
     1                        ' error data. OK ?' ) ) then                5b
                          got_error = vel_to_err( y_speed, speed,
     1                      n_speed, y_error, error, n_error, p_slope )
                          end if                                          5e
                  else                                                    4m
                      got_error = vel_to_err( y_speed, speed, n_speed,
     1                        y_error, error, n_error, p_slope )
                      end if                                              4e
                  data_type = ERROR_DATA
                  got_Nusselt = .FALSE.
                  end if                                                  3e
c
c             Convert velocity error to Nusselt error
c             ***************************************
c
          else if ( item .eq. 10 ) then                                   2m
              if ( .not. got_error ) then                                 3b
                  call print_err( Cno_data )
              else if ( got_Nusselt ) then                                3m
                  call print_err( 'Already Nusselt error. ' //
     1                            'Regenerate velocity error.' )
              else if ( p_slope .le. 0 ) then                             3m
                  call print_err( 'Profile slope not available.' )
              else if ( .not. got_visc ) then                             3m
                  call print_err( Cno_visc )
              else if ( .not. got_friction ) then                         3m
                  call print_err( Cno_fric )
              else                                                        3m
                  got_Nusselt = vel_to_Nusselt( y_error, error, n_error,
     1                p_slope, kinem_visc, friction_vel, method )
                  end if                                                  3e
c
c             Make correction law
c             *******************
c
          else if ( item .eq. 11 ) then                                   2m
              if ( got_error ) then                                       3b
                  if ( got_law ) then                                     4b
                      if ( Equery( STD_IN,
     1                    'Overwrite existing correction law ?' ) ) then  5b
                          got_law = make_corr_law(
     1                        y_error, error, n_error,
     2                        corr_law, MAX_TERMS, got_law,
     3                        method, p_slope )
                          end if                                          5e
                  else                                                    4m
                      got_law = make_corr_law(
     1                    y_error, error, n_error,
     2                    corr_law, MAX_TERMS, got_law,
     3                    method, p_slope )
                      end if                                              4e
              else                                                        3m
                  call print_err( Cno_data )
                  end if                                                  3e
c
c             Plot the correction law
c             ***********************
c
          else if ( item .eq. 12 ) then                                   2m
              if ( got_law ) then                                         3b
                  call plot_corr_law( idatlu, corr_law, .true., 1.0 )
              else                                                        3m
                   call print_err( Cno_law )
                  end if                                                  3e
c
c             Write the correction law
c             ************************
c
          else if ( item .eq. 13 ) then                                   2m
              if ( got_law ) then                                         3b
                  call write_corr_law( corr_law )
              else                                                        3m
                  call print_err( Cno_law )
                  end if                                                  3e
c
c             Read the correction law
c             ***********************
c
          else if ( item .eq. 14 ) then                                   2m
              if ( got_law ) then                                         3b
                  if ( Equery( STD_IN,
     1                    'Overwrite existing correction law ?' ) ) then  4b
                      got_law = read_corr_law( corr_law, MAX_TERMS,
     1                    p_slope, method )
                      end if                                              4e
              else                                                        3m
                  got_law = read_corr_law( corr_law, MAX_TERMS,
     1                p_slope, method )
                  end if                                                  3e
              if ( .not. got_law )
     1            call print_err( Cno_law )
c
c             Test correction law
c             *******************
c
          else if ( item .eq. 15 ) then                                   2m
              if ( .not. got_law ) then                                   3b
                  call print_err( Cno_law )
              else if ( .not. got_visc ) then                             3m
                  call print_err( Cno_visc )
              else if ( .not. got_friction ) then                         3m
                  call print_err( Cno_fric )
              else                                                        3m
                  call test_corr_law( idatlu, corr_law, kinem_visc,
     1                friction_vel )
                  end if                                                  3e
          else if ( item .eq. 16 ) then                                   2m
              got_law = .FALSE.
              end if                                                      2e
          goto 1
c         end of loop                                                     1e
c
c               end of main loop
c               ################
      end


      SUBROUTINE PRINT_ERR( MESSAGE )
      include 'stdio.def'
      include 'prox_law.def'
      integer   str_length
      character message*(*)
c
      write( STD_ERR, 1 ) message(1:str_length( message ) )
1     format( 1x, a / ' Press <RETURN> to continue' )
      read( STD_IN, 2 )
2     format()
      return
      end



      LOGICAL FUNCTION READ_DATA( IDATLU, Y_PLUS, U_DATA, N_POINTS,
     1                            MAX_POINTS )
c
c       Read experimental data from file .
c Formatting rules
c    1) Any line beginning with a semicolon ';' is a comment . Comments
c       are ignored .
c    2) All non-comment lines contain 2 numbers . Anything after the
c       2 numbers is ignored .
c    3) The first number on a non-comment line is a  `y_plus' value .
c    4) The second number on a non-comment line is a non-dimensionalized
c       `u_plus' measurement or a velocity error `Delta_U_plus' estimate .
c
c Arguments
c  integer idatlu       # logical unit no. for the file
c  real y_plus(*),
c       u_data(*)
c  integer n_points     # number of data points read from file
c          max_points   # maximum number of data points
c
c Local Variables
c  integer i,           # points to first non-blank char in a line
c          fil_lu       # logical unit number of input file
c          length       # lenth of 'line'
c          line_no      # number of lines read from file
c          n_words      # number of words (tokens) on the line
c          jljstn       # function : left justify, count length & tokens
c          open_file    # function : open file
c  character line *80   # line read from the data file
c
      include 'stdio.def'
c
      integer fil_lu, line_no, n_points, n_words, max_points, length,
     1        idatlu, jljstn, open_file
      real y_plus(*), u_data(*)
      character line*81

c
c             open data file (and count length of file_name)
c
      read_data = .FALSE.
      fil_lu = open_file( idatlu, '?', 'old', 0 )
      if ( fil_lu .eq. 0 ) return
c
c             loop for reading data file
c             **************************
c
      line_no = 0
      n_points = 1
c
c             read a line, check for blank or comment
c
c     start of loop
1         read( fil_lu, 2, END=6, ERR=8 ) line
2         format( a )
          line_no = line_no + 1
          length = jljstn( line, n_words )
          if ( line(1:1) .eq. ';' .or. length .le. 0 .or.
     1          n_words .le. 0) then
              goto 1
c
c             skip if line does not have enough tokens
c
          else if ( n_words .lt. 2 ) then
              write( STD_ERR, 3 )
3             format( ' READ_DATA : line rejected - not enough data' )
c
c             don't run off the end of the data arrays, 'normal' exit
c
          else if ( n_points .gt. max_points ) then
              write( STD_ERR, 4 ) max_points
4             format( ' READ_DATA : reached end of data array. Read ',
     1                  i5, ' points .' )
              goto 6
          else
c
c             extract y_plus & u_data and then get the next line
c
              if ( n_words .gt. 2 ) then
                  write( STD_ERR, 5 )
5                 format( ' READ_DATA : surplus data ignored' )
                  end if
              read( line, *, ERR=10 ) y_plus(n_points), u_data(n_points)
              n_points = n_points + 1
              end if
          goto 1
c         end of loop
c
c             normal exit
c             ***********
c
6     write( STD_ERR, 7 ) n_points
7     format( ' READ_DATA : reached EOF . Read ', i5, ' points .' )
      n_points = n_points - 1
      read_data = .TRUE.
      close( UNIT=fil_lu )
      return
c
c             ABORT traps
c
8     write( STD_ERR, 9 ) line_no
9     format( ' READ_DATA : Cannot read file, line ', i6 )
      close( UNIT=fil_lu )
      return
c
10    write( STD_ERR, 11 ) line_no
11    format( ' READ_DATA : Cannot decode line ', i5 )
      close( UNIT=fil_lu )
      return
      end



      SUBROUTINE WRITE_DATA( IDATLU, Y_PLUS, U_DATA, N_POINTS, COLUMNS,
     1                       WIDTH, LABEL )
c
c       Writes experimental data to a file .
c Formatting rules for the output file
c    1) The first line contains a ; followed by the LABEL string
c    2) Comments can then be appended by the user . Each comment line
c       begins with a ';' . ( The last comment line entered by the user
c       contains only a '.' and is not included in the output file ) .
c    3) The data is written on each line as
c          y_plus(i), u_data(i,1) ... u_data(i,columns)
c
c Arguments
c  integer idatlu           # logical unit no. of data file
c  real y_plus(*),          # first column of data
c       u_data(width,*)     # remaining columns of data
c  integer n_points         # number of data points read from file
c          columns          # Number of columns of data
c          width            # width (ie no. of columns) of u_data
c  character label*(*)      # text string label for the data file
c
c Local Variables
c  integer i, j             # array, row & column indices
c          fil_lu           # logical unit number of opened file
c          open_file        # function : opens file
c  character line *80       # line read from the data file
c
      include 'stdio.def'
c
      integer i, j, idatlu, fil_lu, n_points, columns, width, open_file
      real y_plus(*), u_data(width,*)
      character label*(*), line*81
c
c             check columns argument
c
      if ( columns .gt. width .or. columns .le. 0 ) then
          write( STD_ERR, 1 ) columns
1         format( ' WRITE_DATA : ABORT : Illegal no. of data columns ',
     1            i12 )
          return
          end if
c
c             open data file (and count length of file_name)
c
      fil_lu = open_file( idatlu, '?', 'new', 0 )
      if ( fil_lu .eq. 0 ) return
c
c             write label and comments
c
      write( fil_lu, 2, ERR=11 ) label( : min( len(label), 128 ) )
2     format( ' ; ', a )
      if ( fil_lu .ne. STD_OUT ) then
          write( STD_OUT, 3 )
3         format( ' Enter comment (end with .)' )
c         start of loop
4             read( STD_IN, 5, ERR=4, END=13 ) line
5             format( a )
              if ( line .eq. '.' ) goto 6
              write( fil_lu, 2, ERR=11 ) line
              goto 4
          end if
c
c             loop for writing data file
c
6     if ( width .gt. 1 ) then
          do 8 i = 1, n_points
              write( fil_lu, 7, ERR=11 ) y_plus(i),
     1            ( u_data(j,i), j = 1, width )
7             format( 1x, f7.3, 8f9.5 : / ( 8x, 8f9.5 ) )
8             continue
      else
          do 10 i = 1, n_points
              write( fil_lu, 9, ERR=11 ) y_plus(i), u_data(1,i)
9             format( 1x, 2g14.6 )
10            continue
          end if
      if ( fil_lu .eq. STD_OUT ) then
          call ilgwait()
      else
          close( UNIT=fil_lu )
          end if
      return
c
c             ABORT traps
c
11    call print_err( ' WRITE_DATA : Cannot write to file ' )
      close( UNIT=fil_lu )
      return
c
13    stop 'ABORT'
      end



      LOGICAL FUNCTION VEL_TO_ERR( Y_SPEED, SPEED, N_SPEED,
     1                             Y_ERROR, ERROR, N_ERROR, SLOPE )
c
c     Converts the velocity profile information to a velocity error by
c     by subtracting an ideal ( i.e. linear ) mean velocity profile .
c     The slope of the ideal profile is determined by a least squares
c     fit with some of the data .
c
c Arguments
c  real y_speed(*),       # IN  # distance from wall for velocity data
c       speed(*),         # IN  # velocity data
c  integer n_speed        # IN  # no. of speed data points
c  real y_error(*),       # OUT # distance from wall for error data = y_speed
c       error(*),         # OUT # error data
c  integer n_error        # OUT # no. of error data points = n_speed
c  real slope             # slope of line of best fit - returned out only
c
c Local variables
c  logical first_try      # .TRUE. only for the first time around the loop
c          ilgread        # function : read cross-hairs position
c  integer i              # array index
c          count          # number of fitted data points
c  real min_fit, max_fit, # abscissa limits for fitted data
c       junk1, junk2,     # unused ordinate limits for fitted data
c       variance,         # variance of fitted data
c       covariance,       # covariance of fitted data
c       ideal_x(2), ideal_y(2)  # endpoints of line of best fit
c
      include 'stdio.def'
c
      logical first_try, ilgread
      integer n_speed, n_error, i, count
      real y_speed(*), speed(*), y_error(*), error(*),
     1     min_fit, max_fit, junk1, junk2, variance, covariance, slope,
     2     ideal_x(2), ideal_y(2)
      character cjunk*1
c
c                 plot the velocity data
c
      call ilgopen( ' ' )
      call ilgraph( 1, 1, 11, 11, 1 )
      call ilgline( 1, 0, 1 )
      call ilgdata( y_speed, speed, 1, n_speed, 1 )
      first_try = .TRUE.
c     start of loop
c
c                 Get the upper and lower limits of data to be fitted
c                 to the ideal (linear) profile .
c
1         call newscr()
          if ( .not. first_try ) then
              write( STD_OUT, 2 ) slope
2             format( ' Slope of ideal profile is ', f12.7 )
              end if
          write( STD_OUT, 3 )
3         format( ' Locate the upper and lower fitting limits .' /
     1            ' ( Enter to identical points to exit . )' )
4         if ( .not. ilgread( min_fit, junk1, cjunk ) ) goto 4
5         if ( .not. ilgread( max_fit, junk2, cjunk ) ) goto 5
c
c                 Exit from the loop if the slope of the ideal profile has
c                 been calculated and the two limit points are the same .
c
          if ( .not. first_try .and. abs( junk2 - junk1 ) .lt. 1.0e-7
     1         .and. abs( max_fit - min_fit ) .lt. 1.0e-7 )
     2        goto 8

c
c                 ensure min < max
c
          if ( max_fit .le. min_fit ) then
              junk1   = max_fit
              max_fit = min_fit
              min_fit = junk1
              end if
c
c                 fit an ideal profile to the data
c
          variance = 0.0
          covariance = 0.0
          count = 0
          do 6 i = 1, n_speed
              if ( y_speed(i) .ge. min_fit .and.
     1             y_speed(i) .le. max_fit ) then
                  covariance = covariance + y_speed(i) * speed(i)
                  variance = variance + y_speed(i) **2
                  count = count + 1
                  end if
6             continue
c
c                 draw the ideal profile
c                 ( but rub out any previous attempt )
c
          if ( count .gt. 0 .and. variance .gt. 0.0 ) then
              slope = covariance / variance
              call ilgline( 0, 1, 1 )
              if ( .not. first_try ) then
                  call ilginvs()
                  call ilgdata( ideal_x, ideal_y, 1, 2, 0 )
                  end if
              ideal_x(1) = min_fit
              ideal_y(1) = min_fit * slope
              ideal_x(2) = max_fit
              ideal_y(2) = max_fit * slope
              call ilgdata( ideal_x, ideal_y, 1, 2, 0 )
              first_try = .FALSE.
          else
              call print_err( 'No data between limits : ' //
     1                        'cannot find ideal profile' )
              vel_to_err = .FALSE.
              end if
          goto 1
c         end of loop
c
c                 shift distances from wall from speed to error stream
c                 and calculate the velocity error
c
8     if ( .not. first_try ) then
          do 9 i = 1, n_speed
              y_error(i) = y_speed(i)
              error(i) = speed(i) - slope * y_speed(i)
9             continue
          vel_to_err = .TRUE.
          n_error = n_speed
      else
          call print_err( ' No conversion performed' )
          vel_to_err = .FALSE.
          end if
      call ilgclos()
      return
      end



      LOGICAL FUNCTION VEL_TO_NUSSELT( Y_ERROR, ERROR, N_ERROR,
     1    P_SLOPE, KINEM_VISC, FRICTION_VEL, METHOD )
c
c    Converts velocity error to Nusselt number error of a form suitable for
c calculating a heat loss correction law .
c
c Arguments
c  real    y_error(*),  # INPUT  # distance from wall in wall units
c          error(*)     # IN/OUT # velocity error in wall units
c                       # OUTPUT #(dimensioned) Nusselt number error
c  integer n_error      # INPUT  # number of error data points
c  real    p_slope      # INPUT  # inferred slope of mean velocity profile
c          kinem_visc   # INPUT  # kinematic viscosity
c          friction_vel # INPUT  # friction_velocity
c  integer method       # INPUT  # correction method  HEAT_METHOD or MOD_H_METHOD
c
c Local variables
c  integer
c    data_lu          # logical unit no. of output file
c    i                # array index
c    item             # item selected from menu
c    poly_index       # pointer to polynomial coefficient
c    Jresp            # function : read integer
c    open_file        # function : open file
c  real
c    corr_fun(MAX_DATA)#correction function for adjusted convected heat flow
c    E2_inferred(MAX_DATA) # voltage squared for inferred convected heat flow
c    E2_meas(MAX_DATA)# voltage squared for total heat flow
c    expon
c    Nu_inferred(MAX_DATA) # correction function for inferred convected heat flow
c    U_inferred(MAX_DATA)  # mean velocity inferred from profile slope and y_plus
c    U_meas(MAX_DATA) # measured mean velocity
c    x_max, x_min     # abscissa limits
c    y_max, y_min     # ordinate limits
c    QHW1_UtoE2       # function : voltage squared from velocity
c    Qpoly            # function : evaluae polynomial
c    qresp            # function : read real
c  subroutine
c    newscr           # clear screen
c    limits           # find max & min in an array
c    ilgopen, ilgraph, ilgxlab, ilgylab, ilgnote,
c    ilgimit, ilgdata, ilgwait, ilgclos
c
c
      include 'stdio.def'
      include 'prox_law.def'
      include 'wallprox.def'
      character  abort_id*26, legend(3)*21
      parameter( abort_id = ' VEL_TO_NUSSELT : ABORT : ' )
c
      integer n_error, method, data_lu, i, item,
     1        poly_index, Jresp, open_file
      real y_error(*), error(*), p_slope, kinem_visc,
     1    friction_vel, E2_inferred(MAX_DATA), E2_meas(MAX_DATA),
     2    Nu_inferred(MAX_DATA), U_inferred(MAX_DATA), U_meas(MAX_DATA),
     3    x_max, x_min, y_max, y_min, corr_fun(MAX_DATA), expon, y,
     4    QHW1_UtoE2, Qpoly, qresp
c     subroutine newscr, limits
c    1    ilgopen, ilgraph, ilgxlab, ilgylab, ilgnote,
c    2    ilgimit, ilgdata, ilgwait, ilgclos
c
      save legend
      data legend / '|1r|20f|s1 = measured',
     1              '|2r|20f|s2 = inferred',
     2              '|3r|20f|s3 = corr fun' /
c
c             check arguments
c
      vel_to_Nusselt = .FALSE.
      if ( n_error .le. 0 ) then                                          1b
          write( STD_ERR, 1 ) abort_id, n_error
1         format( a, 'Number of data points. 0<= ', i12 )
      else if ( p_slope .lt. 0.5 .or. p_slope .gt. 1.5 ) then             1m
          write( STD_ERR, 2 ) abort_id, p_slope
2         format( a, 'Illegal inferred profile slope ', g12.5 )
      else if ( method .ne. HEAT_METHOD .and.
     1          method .ne. MOD_H_METHOD ) then                           1m
          write( STD_ERR, 4 ) abort_id
4         format( a, ' Correction method must be a heat flow method' )
      else                                                                1m
c
c         start of loop                                                   2b
5             continue
c
c             convert from velocity to Nusselt number error
c
              do 6 i = 1, n_error                                         3b
                  y = kinem_visc / friction_vel * y_error(i)
c
c                   inferred actual mean velocity in sublayer
                  U_inferred(i)  = p_slope * friction_vel * y_error(i)
                  E2_inferred(i) = Qhw1_utoe2( U_inferred(i) )
c
c                   mean velocity measured by the probe
                  U_meas(i)  = U_inferred(i) + friction_vel * error(i)
                  E2_meas(i) = Qhw1_utoe2( U_meas(i) )
                  if ( method .eq. MOD_H_METHOD ) then                    4b
                      Nu_inferred(i) = E2_meas(i) - E2_inferred(i)
                      corr_fun(i) = Nu_inferred(i) *
     1                    y ** ( 1.0 -  HW_CALIB_EXPON_N ) *
     2                    friction_vel ** ( 1.0 - 2.0*HW_CALIB_EXPON_N )
                  else                                                    4m
                      Nu_inferred(i) = y *
     1                    ( E2_meas(i) - E2_inferred(i) )
                      end if                                              4e
6                 continue                                                3e
c
c           Menu
              call newscr()
              write( STD_OUT, 8 )
8             format( ' Select' / ' -1) Abort' / '  0) Return (ok)' /
     7            '  5) Plot velocities' /
     8            '  6) Plot CTA voltages squared' /
     9            '  7) Plot correction functions' /
     b            '  9) Write data to file    ', $ )
              item = Jresp( STD_IN, -1, 9, MAX_TRIES )
c
c               Return to main menu
c
              if ( item .eq. -1 ) then                                    3b
                  vel_to_Nusselt = .FALSE.
                  return
              else if ( item .eq. 0 ) then                                3m
                  do 9 i = 1, n_error                                     4b
                      if ( method .eq. MOD_H_METHOD ) then                5b
                          error(i) = corr_fun(i)
                      else                                                5m
                          error(i) = Nu_inferred(i)
                          end if                                          5e
9                     continue                                            4e
                  vel_to_Nusselt = .TRUE.
                  return
c
c               Plot velocities
c
              else if ( item .eq. 5 ) then                                3m
                  x_min = INFINITY
                  x_max = - INFINITY
                  call limits( x_min, x_max, y_error, 1, n_error )
                  y_min = INFINITY
                  y_max = - INFINITY
                  call limits( y_min, y_max, U_meas, 1, n_error )
                  call limits( y_min, y_max, U_inferred, 1, n_error )
                  call ilgopen( ' ' )
                  call ilgraph( 1, 1, 11, 11, 1 )
                  call ilgxlab( 'y|t+' )
                  call ilgylab( 'U|t+' )
                  call ilgimit( x_min, y_min, x_max, y_max )
                  call ilgline( 1, 0, 1 )
                  call ilgnote( legend(1) )
                  call ilgdata( y_error, U_meas, 1, n_error, 0 )
                  call ilgline( 2, 0, 1 )
                  call ilgnote( legend(2) )
                  call ilgdata( y_error, U_inferred, 1, n_error, 0 )
                  call ilgwait()
                  call ilgclos()
c
c               Plot CTA voltages squared
c
              else if ( item .eq. 6 ) then                                3m
                  x_min = INFINITY
                  x_max = - INFINITY
                  call limits( x_min, x_max, y_error, 1, n_error )
                  y_min = INFINITY
                  y_max = - INFINITY
                  call limits( y_min, y_max, E2_meas, 1, n_error )
                  call limits( y_min, y_max, E2_inferred, 1, n_error )
                  call ilgopen( ' ' )
                  call ilgraph( 1, 1, 11, 11, 1 )
                  call ilgxlab( 'y|t+' )
                  call ilgylab( 'E|t2 (Volts|t2)' )
                  call ilgimit( x_min, y_min, x_max, y_max )
                  call ilgline( 1, 0, 1 )
                  call ilgnote( legend(1) )
                  call ilgdata( y_error, E2_meas, 1, n_error, 0 )
                  call ilgline( 2, 0, 1 )
                  call ilgnote( legend(2) )
                  call ilgdata( y_error, E2_inferred, 1, n_error, 0 )
                  call ilgwait()
                  call ilgclos()
c
c               Plot Correction functions
c
              else if ( item .eq. 7 ) then                                3m
                  x_min = INFINITY
                  x_max = - INFINITY
                  call limits( x_min, x_max, y_error, 1, n_error )
                  y_min = INFINITY
                  y_max = - INFINITY
                  call limits( y_min, y_max, Nu_inferred, 1, n_error )
                  call ilgopen( ' ' )
                  call ilgraph( 1, 1, 11, 11, 1 )
                  call ilgxlab( 'y|t+' )
                  call ilgylab( 'Correction function f(y|t+)' )
                  call ilgimit( x_min, y_min, x_max, y_max )
                  call ilgline( 2, 0, 1 )
                  call ilgnote( legend(2) )
                  call ilgdata( y_error, Nu_inferred, 1, n_error, 0 )
                  if ( method .eq. MOD_H_METHOD ) then                    4b
                      y_min = INFINITY
                      y_max = - INFINITY
                      call limits( y_min, y_max, corr_fun, 1, n_error )
                      call ilgraph( 1, 1, 0, 0, 0 )
                      call ilgimit( x_min, y_min, x_max, y_max )
                      call ilgline( 3, 0, 1 )
                      call ilgnote( legend(3) )
                      call ilgdata( y_error, corr_fun, 1, n_error, 0 )
                      end if                                              4e
                  call ilgwait()
                  call ilgclos()
c
c               Write data to file
c
              else if ( item .eq. 9 ) then                                3m
                  data_lu = open_file( 99, '?', 'new', 0 )
                  if ( data_lu .gt. 0 ) then                              4b
                      write( data_lu, 11 )
11                    format( '       Y^+   U_m   E^2_M   U_inf ',
     1                    'E^2_inf U_con E^2_con  Err_inf     Err_con',
     2                    '     Nu_y (adj)' )
                      do 13 i = 1, n_error                                5b
                          write( data_lu, 12 ) i, y_error(i), U_meas(i),
     1                        E2_meas(i), U_inferred(i), E2_inferred(i),
     2                        Nu_inferred(i), corr_fun(i)
12                        format( 1x, i3, 1x, f6.3, 1x, f6.4, 1x, f6.3,
     1                        1x, f6.4, 1x, f6.3, 1x, g12.5, 1x, g12.5 )
                          if ( data_lu .eq. STD_OUT ) then                6b
                              if ( mod( i, 20 ) .eq. 0 ) then             7b
                                  write( data_lu, 11 )
                                  call ilgwait()
                              else if ( i .eq. n_error ) then             7m
                                  call ilgwait()
                                  end if                                  7e
                              end if                                      6e
13                        continue                                        5e
                      if ( data_lu .ne. STD_OUT )                                           5m
     1                    close( UNIT = data_lu )
                      end if                                              4e
                  end if                                                  3e
              goto 5
c             end of loop                                                 2e
          end if                                                          1e
      return
      end



      SUBROUTINE LIMITS( ARR_MIN, ARR_MAX, ARRAY, INCREM, N_ARRAY )
      include 'stdio.def'
      integer increm, n_array, i
      real arr_min, arr_max, array(*)
c
      do 1 i = 1, ( n_array - 1 ) * increm + 1, increm
          arr_min = min( arr_min, array(i) )
          arr_max = max( arr_max, array(i) )
1         continue
      return
      end



      SUBROUTINE PLOT_CORR_LAW( IDATLU, CORR_LAW, EXTRAS, FRICTION_VEL )
c
c    Calculates a set of data points for the correction law and plots
c these points . After the initial plot, the curve can plotted again
c with different scaling . The data points can also be written to data
c file
c
c Arguments
c  integer idatlu     # logical unit no. of data file
c  real    corr_law(*)# array describing the correction law - described above
c  logical extras     # .TRUE. if extended plotting is allowed
c  real  friction_vel # mean friction velocity
c
c Local Variables
c  logical changed          # abscissa limits changed - must recalculate points
c  integer i,               # array index
c    item,                  # item selected from menu
c    x_scale, y_scale,      # 1=lin or 2=log scaling of axes
c    n_steps,               # number of points to plot
c    new_scale,             # 1=rescale in ilgdata
c    Jresp                  # function : read integer
c  real
c    y_data(MAX_POINTS),   # plotted data - distance from wall
c    corr_data(MAX_POINTS),# plotted data - velocity error
c    absc_min, absc_max, absc_step  # abscissa min, max & step size
c    x_min, y_min,          # lower left corner of graph
c    x_max, y_max,          # upper right corner of graph
c    Qresp,                 # function : read real number
c    Qpoly,                 # function : evaluate polynomial
c    prox_law               # function : evaluate wall proximity correction
c character Cscale(2)*3     # 'lin', 'log'
c
      include 'stdio.def'
      include 'wallprox.def'
      include 'prox_law.def'
c
      logical extras, changed
      integer idatlu, i, item, x_scale, y_scale, n_steps, new_scale,
     1        Jresp
      real y_data(MAX_POINTS), corr_data(MAX_POINTS), corr_law(*),
     1     absc_min, absc_max, absc_step, x_min, y_min, x_max, y_max,
     2     factor, friction_vel, Qresp, prox_law
      character Cscale(2)*3
      data Cscale / 'lin', 'log' /
c
c               default values
c
      n_steps = MAX_POINTS / 2
      absc_min = corr_law(Hdom_min)
      absc_max = corr_law(Hdom_max)
      absc_step = ( absc_max - absc_min ) / ( n_steps - 1 )
      factor = 1.0
      if ( extras ) then                                                  1b
          changed = .TRUE.
          new_scale = 1
          x_scale = 1
          y_scale = 1
c
c               select data limits etc
c
c         start loop                                                      2b
1             call newscr()
              write( STD_OUT, 2 ) absc_min, absc_max, absc_step,
     1            n_steps, Cscale( 3 - x_scale ), Cscale( 3 - y_scale )
2             format( ' Select' / '   0) Exit' /
     1            '   1) Abscissa minimum value ', g12.5 /
     2            '   2) Abscissa maximum value ', g12.5 /
     3            '   3) Abscissa step size     ', g12.5 /
     4            '   4) Number of values       ', i4 /
     5            '   5) Change window limits' /
     6            '   6) Change abscissa scaling law to ', a /
     7            '   7) Change ordinate scaling law to ', a /
     8            '   8) Open capture file' /
     9            '   9) Plot' /
     +            '  10) Write data points to file      ', $ )
              item = Jresp( STD_IN, 0, 10, MAX_TRIES )
c
c                     Exit
c
              if ( item .eq. 0 ) then                                     3b
                  call ilgslaw( 'lin', 'lin' )
                  call ilgclos
                  return
c
c                     Abscissa maximum value
c
              else if ( item .eq. 1 ) then                                3m
                  write( STD_OUT, 3 )
3                 format( ' What value ? ', $ )
                  absc_min = Qresp( STD_IN, corr_law(Hdom_min) * 0.5,
     1                corr_law(Hdom_max) * 1.5, 4 )
                  n_steps = ( absc_max - absc_min ) / absc_step + 1
                  absc_step = ( absc_max - absc_min ) / ( n_steps - 1 )
                  changed = .TRUE.
c
c                     Abscissa maximum value
c
              else if ( item .eq. 2 ) then                                3m
                  write( STD_OUT, 3 )
                  absc_max = Qresp( STD_IN, corr_law(Hdom_min) * 0.5,
     1                corr_law(Hdom_max) * 1.5, 4 )
                  n_steps = ( absc_max - absc_min ) / absc_step + 1
                  absc_step = ( absc_max - absc_min ) / ( n_steps - 1 )
                  changed = .TRUE.
c
c                     Abscissa step size
c
              else if ( item .eq. 3 ) then                                3m
                  write( STD_OUT, 3 )
                  absc_step = Qresp( STD_IN,
     1                ( absc_max - absc_min ) / ( MAX_POINTS - 1 ),
     2                ( absc_max - absc_min ), 4 )
                  n_steps = ( absc_max - absc_min ) / absc_step + 1
                  changed = .TRUE.
c
c                     Number of values
c
              else if ( item .eq. 4 ) then                                3m
                  write( STD_OUT, 3 )
                  n_steps = Jresp( STD_IN, 2, MAX_POINTS, 4 )
                  absc_step = ( absc_max - absc_min ) / ( n_steps - 1 )
                  changed = .TRUE.
c
c                     Change window limits
c
              else if ( item .eq. 5 ) then                                3m
                  write( STD_OUT, 4 )
4                 format( ' Lower left corner (coordinate) ? ', $ )
                  read ( STD_IN, *, ERR=9, END=11 ) x_min, y_min
                  write( STD_OUT, 5 )
5                 format( ' Upper right corner (coordinate) ? ', $ )
                  read ( STD_IN, *, ERR=9, END=11 ) x_max, y_max
                  call ilgwild( 'UU', x_min, y_min, x_max, y_max )
                  call ilgimit( x_min, y_min, x_max, y_max )
                  new_scale = 0
c
c                     Change abscissa scaling law
c
              else if ( item .eq. 6 ) then                                3m
                  x_scale = 3 - x_scale
                  call ilgslaw( Cscale(x_scale), Cscale(y_scale) )
                  new_scale = 1
c
c                     Change ordinate scaling law
c
              else if ( item .eq. 7 ) then                                3m
                  y_scale = 3 - y_scale
                  call ilgslaw( Cscale(x_scale), Cscale(y_scale) )
                  new_scale = 1
c
c                     Open capture file
c
              else if ( item .eq. 8 ) then                                3m
                  call ilgopen( '?' )
c
c                     Evaluate data points
c
              else if ( item .eq. 9 .or. item .eq. 10 ) then              3m
                  if ( changed ) then                                     4b
                      do 6 i = 1, n_steps                                 5b
                          y_data(i) = ( i - 1 ) * absc_step + absc_min
                          corr_data(i) = prox_law(y_data(i), corr_law) *
     1                        factor
6                         continue                                        5e
                      end if                                              4e
c
c                     plot the data
c
                  if ( item .eq. 9 ) then                                 4b
                      if ( new_scale .gt. 0 )
     1                    call ilgraph( 1, 1, 11, 11, 1 )
                      call ilgline( 0, 1, 1 )
                      call ilgdata( y_data, corr_data, 1, n_steps,
     1                              new_scale )
c
c                     write data to file
c
                  else if ( item .eq. 10 ) then                           4m
                      call write_data( idatlu, y_data, corr_data,
     1                    n_steps, 1, 1, 'correction law data' )
                  else
                      call print_err( 'BUG : Unknown item' )
                      end if                                              4e
              else
                  call print_err( 'BUG : Unknown item' )
                  end if                                                  3e
              goto 1
c             end of loop                                                 2e
      else                                                                1m
c
c               calculate the data points
c
          do 8 i = 1, n_steps                                             2b
              y_data(i) = ( i - 1 ) * absc_step + absc_min
              corr_data(i) = prox_law( y_data(i), corr_law ) * factor
8             continue                                                    2e
c
c               plot the data
c
          call ilgline( 0, 1, 1 )
          call ilgdata( y_data, corr_data, 1, n_steps, 0 )
          call ilgwait()
          end if                                                          1e
      return
c
c                 trap for input conversion error
c
9     call print_err( ' Did not understand that' )
      goto 1
c
11    stop 'ABORT'
      end



      LOGICAL FUNCTION MAKE_CORR_LAW( Y_ERROR, ERROR, N_ERROR,
     1                     CORR_LAW, LAW_LEN, GOT_LAW,
     2                     METHOD, P_SLOPE )
c
c    Make a correction law to fit the velocity error data . This is the
c only routine in this program which needs to know the details of the
c correction law
c
c    The correction law is defined by the array CORR_LAW . In CORR_LAW, we have
c  Htype   =1   # type of correction law
c  Horder  =2   # order of polynomial
c  Hv_fact =3   # independent vertical scaling factor
c  Hy_fact =4   # independent horizontal scaling factor
c  Hexpon  =5   # independent exponent
c  Hp_slope=6   # inferred profile slope
c  Hdom_min=7   # domain minimum
c  Hdom_max=8   # domain maximum
c  Hpol_err=9   # mean square error of polynomial fit
c  Hpoly   =10  # start of polynomial coefficients
c where `order' is the order of the polynomial
c
c Arguments
c  real    y_error(*) # experimental data - distance from wall
c          error(*)   # experimental data - measured error
c  integer n_error    # number of experimental data points
c  real    corr_law(*)# array describing the correction law - described above
c  integer law_len    # max allowed number of values in corr_law
c  logical got_law    # a correction law is already defined
c  integer method     # correction method  HEAT_METHOD or MOD_H_METHOD
c  real    p_slope    # inferred slope of sublayer mean velocity profile
c
c Local Variables
c  integer i          # array index
c          item       # item selected from menu
c          type, ttype# type of correction law
c          order      # order of fitted polynomial
c          n_points   # number of points to be fitted
c          Jresp      # function : read integer
c          str_length # function : location of last non-blank in a string
c  real ordinate(MAX_POINTS)    # ordinate for polynomial fitting
c       weight(MAX_POINTS)      # weights for polynomial fitting
c       y_factor                # y factor for correction law
c       vel_factor              # velocity factor for correction law
c       exponent                # exponent for correction law
c       max_domain              # maximum abscissa value for fitting
c       min_domain              # minimum abscissa value for fitting
c       del_x                   # abscissa step
c       px_min, px_max          # domain of fitted polynomial
c       x_min, y_min            # lower left corner of fitting data
c       x_max, y_max            # upper right left corner of fitting data
c       y_err                   # yerror(i) * y_factor
c       poly_x(POLY_PTS), poly_y(POLY_PTS), # arrays for fitting polynmials
c       Qresp                   # function : read real
c       Qpoly                   # function : evaluate polynomial
c  double precision work(2,MAX_TERMS)  # work space required by SPYORTH
c  character capture_file*64    # capture file name
c
      include 'stdio.def'
      include 'prox_law.def'
c
      logical changed, got_law
      integer i, item, law_len, method, n_error, type, order,
     1        n_points, ttype, Jresp
      real corr_law(*), error(*), y_error(*), abscissa(MAX_POINTS),
     1    ordinate(MAX_POINTS), weight(MAX_POINTS), x_min, y_min,
     2    x_max, y_max, temp, vel_factor, y_factor, exponent,
     3    min_domain, max_domain, del_x, px_min, px_max,
     4    p_slope, poly_x(POLY_PTS), poly_y(POLY_PTS), Qresp, Qpoly
      double precision work(2,MAX_TERMS)
      character capture_file*64
      include 'wallprox.def'
      data capture_file / ' ' /
c
      changed = .TRUE.
      if ( .not. got_law ) then
          if ( method .eq. VELOC_METHOD ) then
              type     = 4
              order    = 2
              exponent = 1.0
          else if ( method .eq. HEAT_METHOD ) then
              type     = 5
              order    = 3
              exponent = 1.0
          else if ( method .eq. MOD_H_METHOD ) then
              type     = 7
              order    = 3
              exponent = 0.0
          else
              call print_err( 'ABORT : Illegal correction method.' )
              make_corr_law = .FALSE.
              return
              end if
          vel_factor = 1.0
          y_factor   = 1.0
          min_domain = y_error(1)
          max_domain = y_error(1)
          do 1 i = 1, n_error
              min_domain = min( min_domain, y_error(i) )
              max_domain = max( max_domain, y_error(i) )
              weight(i) = 1.0
1             continue
          make_corr_law = .FALSE.
      else
          type  = int( corr_law(Htype) + 0.5 )
          order = int( corr_law(Horder) + 0.5 )
          vel_factor = corr_law(Hv_fact)
          y_factor   = corr_law(Hy_fact)
          exponent   = corr_law(Hexpon)
          min_domain = corr_law(Hdom_min)
          max_domain = corr_law(Hdom_max)
          make_corr_law = .TRUE.
          end if
c
c     start of loop
3         call newscr()
          write( STD_OUT, 4 ) type, order, vel_factor, y_factor,
     1         exponent, p_slope, min_domain, max_domain, Capture_file
4         format( ' What to do ?' / '   0) Exit'   /
     1            '   1) Set correction law type ', i4 /
     2            '   2) Set order of polynomial ', i4 /
     3            '   3) Velocity scale factor   ', g12.5 /
     4            '   4) Y (horiz) scale factor  ', g12.5 /
     5            '   5) Set y_plus exponent     ', g12.5 /
     6            '   6) Inferred profile slope  ', g12.5 /
     7            '   7) Minimum fitted y_plus   ', g12.5 /
     8            '   8) Maximum fitted y_plus   ', g12.5 /
     9            '   9) Capture file name       ', a /
     a            '  10) Calculate Correction law' /
     1            '  11) Edit/plot fitting data' /
     2            '  12) Set default correction law ', $ )
          item = Jresp( STD_IN, 0, 12, MAX_TRIES )
c
          if ( item .eq. 0 ) then
              call ilgclos()
              return
c
c               set default correction law
c               **************************
c
          else if ( item .eq. 12 ) then
              if ( method .eq. VELOC_METHOD ) then
                  type     = 4
                  order    = 2
                  exponent = 1.0
              else if ( method .eq. HEAT_METHOD ) then
                  type     = 5
                  order    = 3
                  exponent = 1.0
              else if ( method .eq. MOD_H_METHOD ) then
                  type     = 7
                  order    = 3
                  exponent = 1.0
              else
                  call print_err( 'ABORT : Illegal correction method.' )
                  make_corr_law = .FALSE.
                  return
                  end if
              vel_factor = 1.0
              y_factor   = 1.0
              min_domain = y_error(1)
              max_domain = y_error(1)
              do 5 i = 1, n_error
                  min_domain = min( min_domain, y_error(i) )
                  max_domain = max( max_domain, y_error(i) )
5                 continue
              changed = .TRUE.
c
c               set correction law type
c               ***********************
c
          else if ( item .eq. 1 ) then
              write( STD_OUT, 6 )
6             format( ' What value ? ', $ )
              ttype = Jresp( STD_IN, 1, 7, MAX_TRIES )
              if ( ttype .ne. 4 .and. ttype .ne. 5 .and.
     1             ttype .ne. 7 ) then
                  call print_err( 'Must select type 4, 5 or 7' )
              else
                  type = ttype
                  end if
              changed = .TRUE.
c
c               set order of the polynomial
c               ***************************
c
          else if ( item .eq. 2 ) then
              write( STD_OUT, 6 )
              order = Jresp( STD_IN, 1, law_len - Hpoly, MAX_TRIES )
              changed = .TRUE.
c
c               set the value of the velocity scale factor
c               ******************************************
c
          else if ( item .eq. 3 ) then
              write( STD_OUT, 6 )
              vel_factor = Qresp( STD_IN, 1.0e-5, 1.0e5, MAX_TRIES )
              corr_law(Hv_fact) = vel_factor
              changed = .TRUE.
c
c               set the value of the y scale factor
c               ***********************************
c
          else if ( item .eq. 4 ) then
              write( STD_OUT, 6 )
              y_factor = Qresp( STD_IN, 1.0e-5, 1.0e5, MAX_TRIES )
              corr_law(Hy_fact) = y_factor
              changed = .TRUE.
c
c               set the value of the exponent
c               *****************************
c
          else if ( item .eq. 5 ) then
              write( STD_OUT, 6 )
              exponent = Qresp( STD_IN, -20.0, 20.0, MAX_TRIES )
              changed = .TRUE.
c
c               Inferred profile slope
c               **********************
c
          else if ( item .eq. 6 ) then
              write( STD_OUT, 6 )
              p_slope = Qresp( STD_IN, 0.3, 3.0, MAX_TRIES )
c
c               Minimum fitted y_plus
c               *********************
c
          else if ( item .eq. 7 ) then
              write( STD_OUT, 6 )
              min_domain = Qresp( STD_IN, 0.0, 15.0, MAX_TRIES )
              corr_law(Hdom_min) = min_domain
              changed = .TRUE.
c
c               Maximum fitted y_plus
c               *********************
c
          else if ( item .eq. 8 ) then
              write( STD_OUT, 6 )
              max_domain = Qresp( STD_IN, 0.0, 15.0, MAX_TRIES )
              corr_law(Hdom_max) = max_domain
              changed = .TRUE.
c
c               Set capture file name
c               *********************
c
          else if ( item .eq. 9 ) then
              write( STD_OUT, 6 )
              call Cresp( STD_IN, capture_file, MAX_TRIES )
c
c               calculate the fitting coordinates
c               *********************************
c
          else if ( item .eq. 10 .or. item .eq. 11 ) then
              if ( changed ) then
                  n_points = 1
                  x_min = 1.0e37
                  y_min = 1.0e37
                  x_max = -1.0e37
                  y_max = -1.0e37
                  do 8 i = 1, n_error
                      if ( min_domain .le. y_error(i) .and.
     1                     max_domain .ge. y_error(i) ) then
CC**********************************************************************
CC**
CC**                 do not use correction law types 1, 2, 3 & 6
CC** c
CC** c               TYPE 1 - depends on formula for the law
CC** c
CC**                           if ( type .eq. 1 ) then
CC**                               ordinate(n_points) = error(i) /
CC**      1                            vel_factor / y_error(i) ** exponent
CC**                               abscissa(n_points) = y_error(i) / y_factor
CC** c
CC** c               TYPE 2 - depends on formula for the law
CC** c
CC**                           else if ( type .eq. 2 ) then
CC**                               ordinate(n_points) = error(i) /
CC**      1                            vel_factor / y_error(i) ** exponent
CC**                               abscissa(n_points) = y_factor / y_error(i)
CC** c
CC** c               TYPE 3 - depends on formula for the law
CC** c
CC**                           else if ( type .eq. 3 ) then
CC**                               ordinate(n_points) = error(i) / vel_factor
CC**                               abscissa(n_points) =
CC**      1                            ( y_error(i) / y_factor ) ** exponent
CC** c
CC** c               TYPE 6 - depends on formula for the law
CC** c
CC**                           else if ( type .eq. 5 ) then
CC**                               temp = error(i) / vel_factor /
CC**      1                            ( y_error(i) / y_factor ) ** exponent
CC**                               if ( temp .gt. 0.0 ) then
CC**                                   ordinate(n_points) = - log( temp )
CC**                                   abscissa(n_points) =
CC**      1                                y_error(i) / y_factor
CC**                               else
CC**                                   write( STD_OUT, 5 ) temp
CC**                                   n_points = n_points - 1
CC**                                   end if
CC**
CC**********************************************************************
c
c               TYPE 4 - depends on formula for the law
c
                          if ( type .eq. 4 ) then
                              temp = error(i) * y_error(i) /
     1                            exponent / vel_factor
                              if ( temp .gt. 0.0 ) then
                                  ordinate(n_points) = - log( temp )
                                  abscissa(n_points) =
     1                                ( y_error(i) / y_factor ) ** 2
                              else
                                  write( STD_OUT, 7 ) temp
7     format( ' Datum skipped - Cannot evaluate log(',g12.5,')' )
                                  n_points = n_points - 1
                                  end if
c
c               TYPE 5 - depends on formula for the law
c
                          else if ( type .eq. 5 ) then
                              temp = error(i) / exponent / vel_factor
                              if ( temp .gt. 0.0 ) then
                                  ordinate(n_points) = - log( temp )
                                  abscissa(n_points) =
     1                                y_error(i) / y_factor
                              else
                                  write( STD_ERR, 7 ) temp
                                  n_points = n_points - 1
                                  end if
c
c               TYPE 7 - depends on formula for the law
c
                          else if ( type .eq. 7 ) then
                              temp = error(i) / vel_factor
                              if ( temp .gt. 0.0 ) then
                                  ordinate(n_points) = - log( temp )
                                  abscissa(n_points) =
     1                                y_error(i) / y_factor
                              else
                                  write( STD_ERR, 7 ) temp
                                  n_points = n_points - 1
                                  end if
                          else
                              goto 12
                              end if
                          if ( n_points .gt. 0 ) then
                              x_min = min( x_min, abscissa(n_points) )
                              y_min = min( y_min, ordinate(n_points) )
                              x_max = max( x_max, abscissa(n_points) )
                              y_max = max( y_max, ordinate(n_points) )
                              end if
                          n_points = n_points + 1
                          end if
8                     continue
                  n_points = n_points - 1
                  changed = .false.
                  end if
c
c               calculate the correction law
c               ****************************
c
              if ( item .eq. 10 .and. n_points .gt. 0 ) then
                  corr_law(Htype)    = type
                  corr_law(Horder)   = order
                  corr_law(Hv_fact)  = vel_factor
                  corr_law(Hy_fact)  = y_factor
                  corr_law(Hexpon)   = exponent
                  corr_law(Hp_slope) = p_slope
                  corr_law(Hdom_min) = min_domain
                  corr_law(Hdom_max) = max_domain
                  call spyorth( abscissa, ordinate, weight, n_points,
     1                corr_law(Hpoly), order, corr_law(Hpol_err), work )
                  make_corr_law = .TRUE.
                  call prox_write( corr_law, ' ' )
                  call ilgwait()
c
c               edit/plot the fitting data
c               **************************
c
              else if ( item .eq. 11 .and. n_points .gt. 0 ) then
                  call ilgopen( capture_file )
                  call ilgraph( 1, 1, 11, 11, 1 )
                  if ( make_corr_law .or. got_law ) then
c
c               determine the domain of the fitted polynomial
c               this depends on the form of the fitting laws
c
                      if ( type .eq. 1 .or. type .eq. 5 .or.
     1                        type .eq. 7 ) then
                          px_min = min_domain
                          px_max = max_domain
                      else if ( type .eq. 2 ) then
                          px_min = 1.0 / max_domain
                          px_max = 1.0 / min_domain
                      else if ( type .eq. 3 ) then
                          if ( exponent .gt. 0.0 ) then
                              px_min = min_domain ** exponent
                              px_max = max_domain ** exponent
                          else
                              px_min = max_domain ** exponent
                              px_max = min_domain ** exponent
                              end if
                      else if ( type .eq. 4 ) then
                          px_min = min_domain **2
                          px_max = max_domain **2
                      else
                          goto 13
                          end if
c
c               generate the data points & then plot the polynomial
c
                      del_x = ( px_max - px_min ) / ( POLY_PTS - 1 )
                      do 11 i = 1, POLY_PTS
                          poly_x(i) = px_min + del_x * ( i - 1 )
                          poly_y(i) = qpoly( poly_x(i), corr_law(Hpoly),
     1                        int( corr_law(Horder) + 0.5 ) )
11                        continue
                      call ilgimit( x_min, y_min, x_max, y_max )
                      call ilgline( 0, 1, 1 )
                      call ilgdata( poly_x, poly_y, 1, POLY_PTS, 0 )
                      end if
                  call sfudge( abscissa, ordinate, 1, n_points,
     1                n_points, MAX_POINTS )
                  call ilgline( 0, 1, 1 )
                  call ilgclos()
              else if ( n_points .le. 0 ) then
                  call print_err( ' No data points inside domain' )
                  end if
              end if
          goto 3
c
c               unknown correction law
c
12        call print_err( 'Unknown correction law type' )
          goto 3
c
13    stop ' ABORT'
      end



      SUBROUTINE WRITE_CORR_LAW( CORR_LAW )
c
c    Write the correction law to a file .
c
c    The correction law is defined by the array CORR_LAW . In CORR_LAW, we have
c
c Arguments
c  real corr_law(*)   # array describing the correction law - described above
c
c Local Variables
c  character file_name*80
c
      include 'stdio.def'
      include 'prox_law.def'
c
      real corr_law(*)
      character file_name*80
c
c               open data file
c
      write( STD_OUT, 1 )
1     format( ' Name of correction law file ( blank = screen ) ? ', $ )
      call Cresp( STD_IN, file_name, MAX_TRIES )
c
c       write the correction law to file
c
      call prox_write( corr_law, file_name )
      if ( file_name .eq. ' ' )
     1    call ilgwait()
      return
      end



      LOGICAL FUNCTION READ_CORR_LAW( CORR_LAW, LAW_LEN,
     1                                P_SLOPE, METHOD )
c
c    Read a correction law to a file . The correction law is defined by
c the array CORR_LAW and MAX_TERMS . MAX_TERMS is number of items in CORR_LAW .
c
c Arguments
c  real corr_law(*)   # array describing the correction law - described above
c  integer law_len    # number of values in corr_law
c  real    p_slope    # inferred slope of sublayer mean velocity profile
c  integer method     # correction method (defined in prox_law.def)
c
c Local Variables
c  logical prox_read  # function : reads wall proximity correction law
c  integer str_length # function : location of last non-blank in a string
c  character file_name*80
c
      include 'stdio.def'
      include 'prox_law.def'
      include 'wallprox.def'
c
      logical prox_read
      integer law_len, method, str_length
      real corr_law(*), p_slope
      character file_name*82
c
c               open data file
c
1     write( STD_OUT, 2 )
2     format( ' Name of correction law file ? ', $ )
      call Cresp( STD_IN, file_name, MAX_TRIES )
      if ( str_length( file_name ) .le. 0 ) goto 1
c
c                get the correction law
c
      read_corr_law = prox_read( corr_law, law_len, file_name )
      p_slope = corr_law(Hp_slope)
      if ( int( corr_law(Htype) + 0.5 ) .eq. 4 ) then
          method = VELOC_METHOD
      else if ( int( corr_law(Htype) + 0.5 ) .eq. 5 ) then
          method = HEAT_METHOD
      else if ( int( corr_law(Htype) + 0.5 ) .eq. 7 ) then
          method = MOD_H_METHOD
      else
          method = UNDEF_METHOD
          end if
      return
4     stop 'Abort'
      end



      BLOCK DATA CORR_LAW_BLOCK
c
c         initialize stream data
c
      include 'stdio.def'
      include 'prox_law.def'
      include 'prox_law.cmn'
c
      data got_stream / .FALSE. /
      data n_streams, n_rows, stream, method / 0, 0, 1, 1 /
      end



      SUBROUTINE TEST_CORR_LAW( IDATLU, CORR_LAW, KINEM_VISC,
     1    FRICTION_VEL )
c
c   Test the wall proximity correction law defined by CORR_LAW .
c The test data and the correction law must be read from file .
c
c Argument
c   integer idatlu    # logical unit no. of data file
c   real corr_law(*)  # wall proximity correction law
c        kinem_visc   # INPUT  # kinematic viscosity
c        friction_vel # INPUT  # friction_velocity
c
c Data file format :-
c   The data file contains a number of velocity time series records,
c each of which is stored as one column of data . The first column of
c data is time (in seconds) . A ';' marks the beginning of a comment
c field in the file . The comment field continues until the end of line.
c The first non-comment line in the file is read as a sequence of probe
c locations (in mm from the wall) . The expected number of velocity
c records is the same as the number of probe locations . Any line which
c contains insufficient data is rejected . Any surplus data on a line is
c ignored .
c
c Local variables
c  logical Equery,        # function :
c          read_stream    # function : reads data stream
c  integer src, dest      # source and destination indices for copying
c                         # a data stream prior to correcting
c          count_clip     # display count of calibration clipping
c          i              # array index
c          item           # item selected from menu
c          istr           # sweeping stream number
c          unused         # unused return value
c          jresp          # function : read integer
c          JHW1_LIMIT     # function : control hot-wire calibration clipping
c  integer *2 AD_signal(MAX_ROWS)  # A/D converter readings
c          JHW1_E2toAD    # function : convert Volts_squared to A/D readings
c  real temp_vel(MAX_ROWS)         # temporary storage for corrected velocity
c       QHW1_UtoE2        # function : Convert speed to volts_squared
c       QHW1_E2toU        # function : Convert volts_squared to speed
c  character Cmethod(2)*8 # correction method options
c            Cyes*3       # function : NO, YES => ASCII
c
      include 'stdio.def'
      integer    MAX_FILT_LEN
      parameter( MAX_FILT_LEN=128 )
      include 'wallprox.def'
      include 'prox_law.def'
c
      logical   EFIR_read, Equery, read_stream
      integer   idatlu, src, dest, count_clip, i, item, istr, unused,
     1          filt_len, n_filtered,
     2          JFIR_filt, Jresp, JHW1_LIMIT
      double precision filt_coeffs(MAX_FILT_LEN)
      real      corr_law(*), kinem_visc, friction_vel,
     1          temp_vel(MAX_ROWS), QHW1_UtoE2, QHW1_E2toU
      integer *2 AD_signal(MAX_ROWS), JHW1_E2toAD
      character Cmethod(3)*8, Cyes*3, filt_file(MAX_COLS)*32
      external  QHW1_E2toU, QHW1_UtoE2
c
      include 'prox_law.cmn'
      save count_clip
      data count_clip / YES /
      data Cmethod / 'velocity', 'simple  ', 'modified' /
c
c                 write correction law to temporary file
c                 and clear all remembered file names and correction laws
c
      if ( int( corr_law(Htype) + 0.5 ) .eq. 4 ) then
          method = VELOC_METHOD
      else if ( int( corr_law(Htype) + 0.5 ) .eq. 5 ) then
          method = HEAT_METHOD
      else if ( int( corr_law(Htype) + 0.5 ) .eq. 7 ) then
          method = MOD_H_METHOD
          end if
      call prox_write( corr_law, 'prox_law.scr' )
      call sbhatia( velocity, 1, 0, 0.0, 0.0, ' ' )
      call prox_nusselt( AD_signal, temp_vel, 1, 0, 0.0, 0.0,
     1                   QHW1_E2toU, QHW1_UtoE2, ' ' )
c
c     start of loop                                                     1b
1         call newscr()
          write( STD_OUT, 2 ) Cmethod(method)
2         format( ' Select' / '   0) Exit' /
     1        '   1) read time series data' /
     2        '   2) single point correction' /
     3        '   3) correction method', 8x, a )
          if ( got_stream ) then                                        2b
              write( STD_OUT, 3 ) stream, stream, stream
3             format( '   4) select data stream         ', i2 /
     1                '   5) apply correction to stream ', i2 /
     2                '   6) correct all data' /
     3                '   7) plot time history of stream ', i2 /
     4                '   8) write time series data' /
     5                '   9) calculate means etc' /
     6                '  10) error statistics' /
     7                '  11) Switch debugging status' /
     8                '  12) Calibration clipping control' /
     9                '  13) Specify filter files' /
     a                '  14) filter all data' /
     b                '  15) clear corrected data ',$ )
              end if                                                    2e
          if ( got_stream ) then                                        2b
              item = Jresp( STD_IN, 0, 15, MAX_TRIES )
          else                                                          2m
              item = Jresp( STD_IN, 0, 3, MAX_TRIES )
              end if                                                    2e
c
c                   return to main program
c                   **********************
c
          if ( item .eq. 0 ) then                                       2b
              call ilgclos()
              return
c
c                  read data stream
c                   ****************
c
          else if ( item .eq. 1 ) then                                  2m
              if ( got_stream ) then                                    3b
                  if ( Equery( STD_IN, 'Write over existing data ?' ) ) 4b
     1                    then
                      got_stream = read_stream( idatlu )
                      end if                                            4e
              else                                                      3m
                  got_stream = read_stream( idatlu )
                  end if                                                3e
              if ( got_stream ) then                                    3b
                  do 4 i = 1, n_streams                                 4b
                      corrected(i) = .FALSE.
4                     continue                                          4e
              else                                                      3m
                  n_streams = 0
                  end if                                                3e
c
c                   single point correction
c                   ***********************
c
          else if ( item .eq. 2 ) then                                  2m
              call single_point( idatlu, method, kinem_visc,
     1                           count_clip )
c
c                   change correction method
c                   ************************
c
          else if ( item .eq. 3 ) then                                  2m
              method = mod( method - 1, 3 ) + 1
c
c                   select data stream
c                   ******************
c
          else if ( item .eq. 4 .and. n_streams .gt. 0 ) then           2m
              write( STD_OUT, 5 )
5             format( ' What value ? ', $ )
              stream = Jresp( STD_IN, 1, min( n_streams, MAX_COLS ), 4)
c
c                   apply correction -- selected stream only
c                   ****************************************
c
          else if ( item .eq. 5 .and. n_streams .gt. 0 ) then           2m
              if ( stream .gt. n_streams ) then                         3b
                  write( STD_ERR, 6 ) n_streams
6                 format( ' Only ', i2, ' data streams are available' )
                  call ilgwait()
              else if ( corrected(stream) ) then                        3m
                  write( STD_ERR, 7 ) stream
7                 format( ' Stream ', i2, ' is already corrected' )
                  call ilgwait()
              else                                                      3m
c
c                   BHATIA'S (VELOCITY) CORRECTION METHOD
c                   first, copy the data into the result space
c
                  if ( method .eq. VELOC_METHOD ) then                  4b
                      src = stream
                      dest = stream + n_streams
                      do 8 i = 1, n_data(stream)                        5b
                          velocity(dest) = velocity(src)
                          dest = dest + 2 * n_streams
                          src  = src  + 2 * n_streams
8                         continue                                      5e
c
c                   then do the correction in the result space
c
                      call sbhatia( velocity(stream + n_streams),
     1                    n_streams * 2, n_data(stream),
     2                    y(stream) / 1000.0, kinem_visc,
     3                    'prox_law.scr' )
                      corrected(stream) = .TRUE.
c
c                   NUSSELT CORRECTION METHOD
c                   first, convert the velocity data into voltage squared
c                   and put it into an integer array for use by PROX_NUSSELT
c
                  else if ( method .eq. HEAT_METHOD .or.
     1                      method .eq. MOD_H_METHOD ) then             4m
                      src = stream
                      do 9 i = 1, n_data(stream)                        5b
                          AD_signal(i) =
     1                        JHW1_E2toAD( QHW1_UtoE2( velocity( src )))
                          src = src + 2 * n_streams
9                         continue                                      5e
c
c                   do the correction into a temporary work array
c
                      call prox_nusselt( AD_signal, temp_vel,
     1                    1, n_data(stream), y(stream) / 1000.0,
     2                    kinem_visc, QHW1_E2toU, QHW1_UtoE2,
     3                    'prox_law.scr' )
c
c                   finally, copy the temporary work array to the result space
c
                      dest = stream + n_streams
                      do 10 i = 1, n_data(stream)                       5b
                          velocity(dest) = temp_vel(i)
                          dest = dest + 2 * n_streams
10                        continue                                      5e
c
c                   display calibration clip count
c
                      if ( count_clip .eq. YES ) then                   5b
                          unused = jhw1_limit( STD_OUT, 0.0 )
                          unused = jhw1_limit( -1, 0.0 )
                          call ilgwait()
                          end if                                        5e
c
                      corrected(stream) = .TRUE.
                  else
                      call print_err( 'BUG: Unknown correction method' )
                      end if                                            4e
                  end if                                                3e
c
c                   apply correction  to all streams
c                   ********************************
c
          else if ( item .eq. 6 .and. n_streams .gt. 0 ) then           2m
              do 16 istr = 1, n_streams                                 3b
                  if ( .not. corrected(istr) ) then                     4b
c
c                   BHATIA'S (VELOCITY) CORRECTION METHOD
c                   first, copy the data into the result space
c
                      if ( method .eq. VELOC_METHOD ) then              5b
                          src = istr
                          dest = istr + n_streams
                          do 12 i = 1, n_data(istr)                     6b
                              velocity(dest) = velocity(src)
                              dest = dest + 2 * n_streams
                              src  = src  + 2 * n_streams
12                            continue                                  6e
c
c                   then do the correction in the result space
c
                          call sbhatia( velocity(istr + n_streams),
     1                        n_streams * 2, n_data(istr),
     2                        y(istr) / 1000.0, kinem_visc,
     3                        'prox_law.scr' )
                          corrected(istr) = .TRUE.
                          write( STD_OUT, 13 ) istr
13                        format( ' Done stream ', i2 )
c
c                   NUSSELT CORRECTION METHOD
c                   first, convert the velocity data into voltage squared
c                   and put it into an integer array for use by PROX_NUSSELT
c
                      else if ( method .eq. HEAT_METHOD .or.
     1                          method .eq. MOD_H_METHOD ) then         5m
                          src = istr
                          do 14 i = 1, n_data(istr)                     6b
                              AD_signal(i) = JHW1_E2toAD( QHW1_UtoE2(
     1                            velocity( src ) ) )
                              src = src + 2 * n_streams
14                            continue                                  6e
c
c                   do the correction into a temporary work array
c
                          call prox_nusselt( AD_signal, temp_vel,
     1                        1, n_data(istr), y(istr) / 1000.0,
     2                        kinem_visc, QHW1_E2toU, QHW1_UtoE2,
     3                        'prox_law.scr' )
c
c                   finally, copy the temporary work array to the result space
c
                          dest = istr + n_streams
                          do 15 i = 1, n_data(istr)                     6b
                              velocity(dest) = temp_vel(i)
                              dest = dest + 2 * n_streams
15                            continue                                  6e
                          corrected(istr) = .TRUE.
                          write( STD_OUT, 13 ) istr
c
c                   display calibration clip count
c
                          if ( count_clip .eq. YES ) then               5b
                              unused = jhw1_limit( STD_OUT, 0.0 )
                              unused = jhw1_limit( -1, 0.0 )
                              end if                                    5e
                      else                                              5m
                          call print_err(
     1                        'BUG: Unknown correction method' )
                          end if                                        5e
                      end if                                            4e
16                continue                                              3e
              if ( count_clip .eq. YES ) call ilgwait()
c
c                   plot time history
c                   *****************
c
          else if ( item .eq. 7 .and. n_streams .gt. 0 ) then           2m
              if ( stream .gt. n_streams ) then                         3b
                  write( STD_ERR, 6 ) n_streams
                  call ilgwait()
              else                                                      3m
                  call plot_stream()
                  end if                                                3e
c
c                   write time history to file
c                   **************************
c
          else if ( item .eq. 8 .and. n_streams .gt. 0 ) then           2m
              call write_stream( idatlu )
c
c                   calculate means etc
c                   *******************
c
          else if ( item .eq. 9 .and. n_streams .gt. 0 ) then           2m
              if ( stream .gt. n_streams ) then                         3b
                  write( STD_ERR, 6 ) n_streams
                  call ilgwait()
              else                                                      3m
                  call mean_stream( idatlu, kinem_visc, friction_vel )
                  end if                                                3e
c
c                   calculate statistics of the error streams
c                   *****************************************
c
          else if ( item .eq. 10 .and. n_streams .gt. 0 ) then          2m
              if ( stream .gt. n_streams ) then                         3b
                  write( STD_ERR, 6 ) n_streams
                  call ilgwait()
              else                                                      3m
                  call error_stats( idatlu, kinem_visc, friction_vel,
     1                corr_law, temp_vel )
                  end if                                                3e
c
c                   switch debugging status
c                   ***********************
c
          else if ( item .eq. 11 ) then                                 2m
              call prox_nusselt( AD_signal, temp_vel, 1, -1,
     1                y(stream) / 1000.0, kinem_visc,
     2               QHW1_E2toU, QHW1_UtoE2, ' ' )
c
c                   Calibration clipping control
c                   ****************************
c
          else if ( item .eq. 12 ) then                                 2m
c             repeat                                                    3b
20                call newscr()
                  write( STD_OUT, 21 ) Cyes( count_clip )
21                format( '   2) Count calibration clipping : ', a )
                  unused = JHW1_LIMIT( -3, 0.0 )
                  if ( unused .eq. -2 )
     1                count_clip = mod( count_clip + 1, 2 )
                  if ( unused .ne. 0 ) goto 20
c                 until ( Exit from menu )                              3e
c
c                   Specify filter files
c                   ********************
c
          else if ( item .eq. 13 ) then                                 2m
              call get_filter_files( filt_file, n_streams )
c
c                   filter all data
c                   ***************
c
          else if ( item .eq. 14 ) then                                 2m
              do 23 istr = 1, n_streams                                 3b
                  if ( filt_file(istr) .ne. ' ' ) then                  4b
                      if ( efir_read( filt_file(istr), MAX_FILT_LEN,
     1                        filt_coeffs, filt_len ) ) then            5b
                          n_filtered = Jfir_filt( velocity(istr),
     1                        velocity(istr), 2 * n_streams,
     2                        n_data(istr), filt_coeffs, filt_len, 8 )
                          if ( n_filtered .gt. 0 ) then
                              n_data(istr) = n_filtered
                              end if
                          write( STD_OUT, 22 ) istr, n_filtered
22                        format( ' Filtered stream ', i2, ' ; ', i6,
     1                        ' data samples' )
                          end if                                        5e
                      end if                                            4e
23                continue                                              3e
          else if ( item .eq. 15 ) then                                 2m
              do 24 istr = 1, n_streams                                 3b
                  corrected(istr) = .FALSE.
24                continue                                              3e
              end if                                                    2e
          goto 1
c         end of loop                                                   1e
      end



      SUBROUTINE GET_FILTER_FILES( FILT_FILE, N_STREAMS )
c
c Read the names of the filter specification files -- one for each data stream.
c
c Arguments
c  character filt_file(*)*(*) # array of filter file names
c  integer   n_streams        # number of data streams
c
c Constants
c   MAX_TRIES=4
c
c Local variables
c  integer
c   file_len    # length of file name
c   i           # file name array index
c   istr        # stream number
c   item        # selected menu item
c   name_lu     # logical unit number of file of filter file names
c   n_display   # number of file name displayed
c   reserv_lu   # reserved logical unit number
c   jljstr      # function : left justify string
c   Jresp       #          : read integer
c   new_lun     #          : get new logical unit number
c   open_file   #          : open file
c  subroutine
c   Cresp
c   newscr
c
      include 'stdio.def'
      integer    MAX_TRIES
      parameter( MAX_TRIES=4 )
c
      integer file_len, i, istr, item, name_lu, n_display, n_streams,
     1        reserv_lu, jljstr, Jresp, new_lun, open_file
      character filt_file(*)*(*)
c     subroutine Cresp, newscr
      save istr, n_display, reserv_lu
      data istr, n_display, reserv_lu / 1, 5, 0 /
c
c     start of loop                                                       1b
1         call newscr()
          write( STD_OUT, 2 )
     1        ( i, filt_file(i), i = max( istr - n_display / 2, 1 ),
     2          min( istr + n_display / 2, n_streams ) )
2         format( 1x, i2, ') ', a )
          write( STD_OUT, 3 ) istr, istr, n_display
3         format( ' Define filter files' / '  0) Return' /
     1        '  1) Stream number          ', i2 /
     2        '  2) Next stream' /
     3        '  3) Filter file for stream ', i2 /
     4        '  4) Clear filter files' /
     5        '  5) Number of files displayed ', i2 /
     6        '  6) Read names from file  ', $ )
          item = Jresp( STD_IN, 0, 6, MAX_TRIES )
          if ( item .eq. 0 ) then                                         2b
              return
          else if ( item .eq. 1 ) then                                    2m
              write( STD_OUT, 4 )
4             format( ' Stream number ? ', $ )
              istr = Jresp( STD_IN, 1, n_streams, MAX_TRIES )
          else if ( item .eq. 2 ) then                                    2m
              istr = mod( istr, n_streams ) + 1
          else if ( item .eq. 3 ) then                                    2m
              write( STD_OUT, 5 ) istr
5             format( ' Name of filter file for stream ', i2, ' ? ', $ )
              call Cresp( STD_IN, filt_file(istr), MAX_TRIES )
              file_len = jljstr( filt_file(istr) )
              if ( 1 .eq. file_len .or. file_len .eq. 2 ) then            3b
                  read( filt_file(istr)(1:file_len), 6, ERR=7 ) i
6                 format( i )
                  filt_file(istr) = filt_file(i)
7                 continue
                  end if                                                  3e
              istr = mod( istr, n_streams ) + 1
          else if ( item .eq. 4 ) then                                    2m
              do 8 i = 1, n_streams                                       3b
                  filt_file(i) = ' '
8                 continue                                                3e
          else if ( item .eq. 5 ) then                                    2m
              write( STD_OUT, 9 )
9             format( ' Number of filter files displayed ? ', $ )
              n_display = Jresp( STD_IN, 3, min( 15, n_streams ),
     1             MAX_TRIES )
          else if ( item .eq. 6 ) then                                    2m
              if ( reserv_lu .eq. 0 )
     1            reserv_lu = new_lun()
              name_lu = open_file( reserv_lu, '?', 'old', 0 )
              if ( name_lu .gt. 0 ) then                                  3b
                  do 11 i = 1, n_streams                                  4b
                      read( name_lu, 10, ERR=12, END=12 ) filt_file(i)
10                    format( a )
11                    continue                                            4e
12                close( UNIT=name_lu )
                  end if                                                  3e
              end if                                                      2e
          goto 1
c         end of loop                                                     1e
      end



      SUBROUTINE SINGLE_POINT( IDATLU, METHOD, KINEM_VISC, COUNT_CLIP )
c
c         Single point correction
c
c Argument
c  integer idatlu     # logical unit no. of data file
c          method     # correction method (1=velocity, 2=Nusselt)
c  real    kinem_visc # kinematic viscosity (SI units)
c  integer count_clip # display calibration clip count
c
c Local variables
c   integer item      # selected item
c           fil_lu    # output file
c           lim_ret   #
c           row       # results table row number
c           table_ptr # pointer to next empty place in table
c           Jhw1_limit# function : control hot-wire calibration clipping
c           Jresp     # function : read integer
c           open_file # function opens file
c integer *2 iad_sample # A/D converter sample
c         JHW1_E2toAD # function convert voltage squared to A/D converter sample
c   real speed        # uncorrected air speed
c        y            # distance of probe from wall
c        y_table(TABLE_SIZE)   # table of result y values
c        corrected(TABLE_SIZE) # corrected velocity measurements
c        original(TABLE_SIZE)  # original velocity measurements
c        Qresp        # function : read real
c        QHW1_UtoE2   # function : Convert speed to volts_squared
c        QHW1_E2toU   # function : Convert volts_squared to speed
c   character Cyes*3  # function : NO/YES => ASCII

c
      include 'stdio.def'
      include 'prox_law.def'
c
      integer count_clip, idatlu, item, fil_lu, method, row, table_ptr,
     1        lim_ret, Jhw1_limit, Jresp, open_file
      integer *2 iad_sample, JHW1_E2toAD
      real y_table(TABLE_SIZE), corrected(TABLE_SIZE),
     1     original(TABLE_SIZE), kinem_visc, speed, y,
     2     QHW1_E2toU, QHW1_UtoE2, Qresp
      character Cyes*3
      external QHW1_E2toU, QHW1_UtoE2
      data table_ptr / 1 /
      data speed, y / 0.4, 0.4 /
c
c     start of loop
1         call newscr()
          write( STD_OUT, 2 ) speed, y
2         format( ' Select' / '   0) Exit' /
     1            '   1) Air speed          ', f9.5, ' m/s' /
     2            '   2) Distance from wall  ', f8.4, ' mm' /
     3            '   3) Calculate correction  ' /
     4            '   4) Delete last result' /
     5            '   5) Display all results' /
     6            '   6) Clear all results' /
     7            '   7) Switch debugging status' /
     8            '   8) Calibration clipping control ',$ )
          item = Jresp( STD_IN, 0, 8, MAX_TRIES )
c
c                     Return
c
          if ( item .eq. 0 ) then
              return
c
c                     Air speed
c
          else if ( item .eq. 1 ) then
              write( STD_OUT, 3 )
3             format( ' What is its value ', $ )
              speed = Qresp( STD_IN, 0.001, 100.0, MAX_TRIES  )
c
c                     Distance from wall
c
          else if ( item .eq. 2 ) then
              write( STD_OUT, 3 )
              y = Qresp( STD_IN, 0.001, 100.0, MAX_TRIES )
c
c                     Calculate correction
c
          else if ( item .eq. 3 .and.
     1              speed .gt. 0.0 .and. y .gt. 0.0 ) then
              if ( table_ptr .le. TABLE_SIZE ) then
                  original(table_ptr) = speed
                  y_table(table_ptr) = y
c
c                     BHATIA - velocity correction
                  if ( method .eq. 1 ) then
                      corrected(table_ptr) = speed
                      call sbhatia( corrected(table_ptr), 1, 1,
     1                    y / 1000.0, kinem_visc, 'prox_law.scr' )
c
c                     Nusselt correction
                  else if ( method .eq. 2 ) then
                      iad_sample = JHW1_E2toAD( QHW1_UtoE2( speed ) )
                      call prox_Nusselt( iad_sample,
     1                    corrected(table_ptr), 1, 1, y / 1000.0,
     2                    kinem_visc, QHW1_E2toU, QHW1_UtoE2,
     3                    'prox_law.scr' )
                  else
                      call print_err( 'BUG: Unknown correction method' )
                      end if
                  write( STD_OUT, 4 ) speed, y, corrected(table_ptr),
     1                corrected(table_ptr) - speed
4                 format( ' Measured air speed  = ', g14.6 /
     1                    ' Distance from wall  = ', f8.4, 'mm' /
     2                    ' Corrected air speed = ', g14.6 /
     3                    ' Difference          = ', g14.6 )
                  table_ptr = table_ptr + 1
c
c                   display calibration clip count
c
                  if ( count_clip .eq. YES ) then
                      lim_ret = jhw1_limit( STD_OUT, 0.0 )
                      lim_ret = jhw1_limit( -1, 0.0 )
                      call ilgwait()
                      end if
              else
                  call print_err( 'Results table is full' )
                  end if
c
c                     Delete last result
c
          else if ( item .eq. 4 ) then
              if ( table_ptr .gt. 1 ) then
                  table_ptr = table_ptr - 1
              else
                  table_ptr = 1
                  end if
c
c                     Display all results
c
          else if ( item .eq. 5 ) then
              if ( table_ptr .gt. 1 ) then
                  fil_lu = open_file( idatlu, '?', 'new', 0 )
                  if ( fil_lu .eq. 0 ) goto 1
                  write( fil_lu, 5 )
5     format( 11x, 'y       uncorrected   corrected     difference' )
                  do row = 1, table_ptr - 1
                      write( fil_lu, 6 ) row, y_table(row),
     1                    original(row), corrected(row),
     2                    original(row) - corrected(row)
6                     format( 1x, i2, 2x, 4g14.6 )
                      end do
                  if ( fil_lu .eq. STD_OUT ) then
                      call ilgwait()
                  else
                      close( UNIT=fil_lu )
                      end if
              else
                  call print_err( 'Empty results table' )
                  end if
c
c                     Clear all results
c
          else if ( item .eq. 6 ) then
              table_ptr = 1
c
c                     Switch debugging status
c
          else if ( item .eq. 7 ) then
              call prox_Nusselt( iad_sample, corrected(1), 1, -1,
     1                    y / 1000.0, kinem_visc,
     2                    QHW1_E2toU, QHW1_UtoE2, ' ' )
c
c                   Count calibration clipping
c
          else if ( item .eq. 8 ) then
c             repeat                                                    3b
20                call newscr()
                  write( STD_OUT, 21 ) Cyes( count_clip )
21                format( '   2) Count calibration clipping : ', a )
                  lim_ret = JHW1_LIMIT( -3, 0.0 )
                  if ( lim_ret .eq. -2 )
     1                count_clip = mod( count_clip + 1, 2 )
                  if ( lim_ret .ne. 0 ) goto 20
c                 until ( Exit from menu )                              3e
              end if
          goto 1
c         end of loop
      end



      LOGICAL FUNCTION READ_STREAM( IDATLU )
c
c Read a test data stream from file . Up to MAX_ROWS values can be read into
c a maximum of MAX_COLS data streams . The time (first column in the file)
c is loaded into TIME . The data from the remaining columns in the file is
c loaded into VELOCITY . The maximum number of values which can be loaded
c into VELOCITY is MAX_STREAM . After reading each N_STREAMS values from
c a line, a gap of N_STREAMS values is left blank . This makes provision
c for data processing .
c
c Argument
c  integer idatlu           # logical unit no. of data file
c
c Data file format :-
c   The data file contains a number of velocity time series records,
c each of which is stored as one column of data . The first column of
c data is time (in seconds) . A ';' marks the beginning of a comment
c field in the file . The comment field continues until the end of line.
c The first non-comment line in the file is read as a sequence of probe
c locations (in mm from the wall) . The expected number of velocity
c records is the same as the number of probe locations . Any line which
c contains insufficient data is rejected . Any surplus data on a line is
c ignored .
c
c Local variables
c  logical got_y  # have read the first line - of probe positions
c  integer i      # array index -
c    item         # item selected from menu
c    fil_lu       # logical unit of file
c    length       # length of input string
c    time_ptr     # pointer to next available space in TIME
c    n_words      # number of tokens in a string
c    vel_ptr      # pointer to next available space in VELOCITY
c    width        # number of data streams
c    jljstn       # function : left justify & count tokens
c    jresp        # function : read integer
c    str_length   # function : length of string
c    open_file    # function : opens data file for I/P or O/P
c  real
c    qresp        # function : read real value
c  character
c    line*80      # line read from file
c
      include 'stdio.def'
      include 'prox_law.def'
c
      logical got_y
      integer i, item, fil_lu, length, time_ptr, n_words, vel_ptr,
     1        idatlu, width, jljstn, jresp, open_file
      real qresp
      character line*512
c
      include 'prox_law.cmn'
c
c             initialize flags and pointers
c
      vel_ptr = 1
      time_ptr  = 1
      got_y   = .FALSE.
      read_stream = .FALSE.
c
c
c             open data file (and count length of file_name
c
      fil_lu = open_file( idatlu, '?', 'old', 0 )
      if ( fil_lu .eq. 0 ) return
c
c     start of loop
c
c             read line, count its length and number of tokens
c
1         read( fil_lu, 2, ERR=11, END=9 ) line
2         format( a )
          length = jljstn( line, n_words )
          if ( n_words .le. 0 .or. line(1:1) .eq. ';' .or.
     1        length .le. 0 ) goto 1
c
c             if don't have probe locations, interpret the line as probe
c             locations - check for too many, decide no. of data streams
c
          if ( .not. got_y ) then
              if ( n_words .gt. MAX_COLS ) then
                  call print_err( 'READ_STREAM: too many data streams' )
                  return
              else
                  read( line, *, ERR=13, END=15 ) ( y(i), i= 1,n_words )
                  width = n_words
                  got_y = .TRUE.
                  end if
c
c             skip if line does not have enough tokens
c
          else if ( n_words .le. width ) then
              write( STD_ERR, 4 )
4             format( ' READ_STREAM : line rejected - not enough data' )
c
c             'normal' termination if a data array is full
c
          else if ( time_ptr .gt. MAX_ROWS ) then
              write( STD_ERR, 5 ) MAX_ROWS
5             format( ' READ_STREAM : TIME array full at ', i5, ' rows')
              goto 9
          else if ( vel_ptr .gt. MAX_STREAM - width * 2 + 1 ) then
              write( STD_ERR, 6 ) MAX_STREAM
6             format( ' READ_STREAM : VELOCITY array full at ', i5,
     1                ' items' )
              goto 9
          else
c
c             decode time-velocity data
c
              if ( n_words .gt. width + 1 ) then
                  write( STD_ERR, 7 )
7                 format( ' READ_STREAM : surplus data ignored' )
                  end if
              read( line, *, ERR=13, END=15 ) time(time_ptr),
     1            ( velocity(i), i = vel_ptr, vel_ptr + width - 1 )
              vel_ptr = vel_ptr + width * 2
              if ( mod( time_ptr, 1000 ) .eq. 0 ) then
                  write( STD_OUT, 8 ) time_ptr
8                 format( 1x, i6 )
                  end if
              time_ptr = time_ptr + 1
              end if
          goto 1
c         end of loop
c
c               normal exit path
c
9     close( UNIT=fil_lu )
      read_stream = .TRUE.
      first_plot = .TRUE.
      n_rows = time_ptr - 1
      n_streams = width
      do 12 i = 1, n_streams
          n_data(i) = n_rows
12        continue
      write( STD_OUT, 10 ) n_rows, n_streams
10    format( ' READ_STREAM : Normal EOF ' /
     1    1x, i6, ' data points read from ', i2, ' data streams' )
      call ilgwait()
      return
c
c               TRAPS
c               *****
c
11    call print_err( ' READ_STREAM : ABORT : Cannot read file' )
      close( UNIT=fil_lu )
      return
c
13    call print_err( 'READ_STREAM : ABORT : Cannot decode line' )
      close( UNIT=fil_lu )
      return
15    stop ' READ_STREAM : bug'
      end



      SUBROUTINE PLOT_STREAM()
c
c    Plot the data from the currently selected data stream . If the stream
c has been corrected, the corrected stream is also plotted on the same
c axes .
c
c Local variables
c  integer i      # array index
c    item         # selected menu item
c    i_wind       # current window being processed
c    first_wind   # first data window to be plotted
c    last_wind    # last data window to be plotted
c    n_wind       # total number of windows over the data
c    vis_wind     # number of visible windows
c    wind_len     # number of data samples in each window
c    start        # time starting index for time
c    vel_ptr      # pointer to velocity value
c    ord_ptr      # pointer for ordinate assignment
c    Jresp        # integer : read integer
c  real offset    # vertical offset to corrected data
c    x_min, x_max # abscissa limits
c    y_min, y_max # ordinate limits
c    original(MAX_ROWS/4)   # uncorrected velocity measurements ordinates
c    altered(MAX_ROWS/4)    # corrected velocity measurements ordinates
c    Qresp                  # function : read real
c  character
c    Ctitle*16              # graph title
c    Capture_file*64        # capture file name
c
      include 'stdio.def'
      include 'prox_law.def'
c
      integer i, item, i_wind, first_wind, last_wind, n_wind, vis_wind,
     1    wind_len, start, vel_ptr, ord_ptr, Jresp
      real offset, x_min, x_max, y_min, y_max, Qresp
      character Ctitle*16, Capture_file*64
c
      include 'prox_law.cmn'
      real original(MAX_ROWS/4), altered(MAX_ROWS/4)
      data Capture_file / ' ' /
c
      if ( first_plot ) then
          n_wind     = ( n_data(stream) - 1 ) / 1000 + 1
          wind_len   = n_data(stream) / n_wind
          vis_wind   = min( 4, n_wind )
          first_wind = 1
          last_wind  = vis_wind
          offset = 0.1
          end if
      first_plot = .FALSE.
c
c     start of loop
c
c                 select plotting parameters
c
1         call newscr()
          write( STD_OUT, 2 ) first_wind, last_wind, vis_wind, n_wind,
     1                       wind_len, offset, stream, capture_file
2         format( ' Select' / '   0) Exit' /
     1            '   1) First visible window   ', i4 /
     2            '   2) Last visible window    ', i4 /
     3            '   3) No. of visible windows ', i4 /
     4            '   4) Total no. of windows   ', i4 /
     5            '   5) Window length         ', i5 /
     6            '   6) Corrected offset   ', f8.5 /
     7            '   7) Select stream      ', i2 /
     8            '   8) Capture file name  : ', a /
     9            '   9) Plot corrected & uncorrected speed' /
     a            '  10) Plot error   ', $ )
          item = Jresp( STD_IN, 0, 10, MAX_TRIES )
c
c                 return
c
          if ( item .eq. 0 ) then
              call ilgclos()
              return
c
c                 First visible window
c
          else if ( item .eq. 1 ) then
              write( STD_OUT, 3 )
3             format( ' What value ? ', $ )
              first_wind = Jresp( STD_IN, 1, n_wind - vis_wind + 1,
     1                            MAX_TRIES )
              last_wind = first_wind + vis_wind - 1
c
c                 Last visible window
c
          else if ( item .eq. 2 ) then
              write( STD_OUT, 3 )
              last_wind = Jresp( STD_IN, vis_wind, n_wind, MAX_TRIES )
              first_wind = last_wind - vis_wind + 1
c
c                 No. of visible windows
c
          else if ( item .eq. 3 ) then
              write( STD_OUT, 3 )
              vis_wind = Jresp( STD_IN, 1, n_wind, MAX_TRIES )
              last_wind = first_wind + vis_wind - 1
c
c                 Total no. of windows
c
          else if ( item .eq. 4 ) then
              write( STD_OUT, 3 )
              n_wind = Jresp( STD_IN,
     1             n_data(stream) / min(n_rows, MAX_ROWS/4),
     1             max( n_data(stream), MIN_SAMP ), MAX_TRIES )
              wind_len = n_data(stream) / n_wind
c
c                 Window length
c
          else if ( item .eq. 5 ) then
              write( STD_OUT, 3 )
              wind_len = Jresp( STD_IN, min( MIN_SAMP, n_data(stream) ),
     1            min( n_data(stream), MAX_ROWS / 4 ), MAX_TRIES )
              n_wind = n_data(stream) / wind_len
c
c                 Corrected offset
c
          else if ( item .eq. 6 ) then
              write( STD_OUT, 3 )
              offset = Qresp( STD_IN, -1.0, 1.0, MAX_TRIES )
c
c                 Select stream
c
          else if ( item .eq. 7 ) then
              write( STD_OUT, 3 )
              stream   = Jresp( STD_IN, 1, n_streams, MAX_TRIES )
c
c                 Specify capture file name
c
          else if ( item .eq. 8 ) then
              write( STD_OUT, 3 )
              call Cresp( STD_IN, Capture_file, MAX_TRIES )
c
c                 Plot corrected and uncorrected velocity each window
c                 ***************************************************
c
          else if ( item .eq. 9 ) then
              call ilgopen( Capture_file )
              call ilgpage( 'a4', 'wide', 1, vis_wind )
              do 6 i_wind = first_wind, last_wind
                  start = ( i_wind - 1 ) * wind_len
c
c                 find limits of data
c
                  x_min = time(start+1)
                  x_max = time(start+1)
                  vel_ptr = stream + start * n_streams * 2
                  ord_ptr = 1
                  y_min = velocity(vel_ptr)
                  y_max = velocity(vel_ptr)
                  do 5 i = start + 1, start + wind_len
                      x_min = min( x_min, time(i) )
                      x_max = max( x_max, time(i) )
                      original(ord_ptr) = velocity(vel_ptr)
                      y_min = min( y_min, original(ord_ptr) )
                      y_max = max( y_max, original(ord_ptr) )
                      if ( corrected(stream) ) then
                          altered(ord_ptr) =
     1                        velocity( vel_ptr + n_streams ) + offset
                          y_min = min( y_min, altered(ord_ptr) )
                          y_max = max( y_max, altered(ord_ptr) )
                          end if
                      vel_ptr = vel_ptr + n_streams * 2
                      ord_ptr = ord_ptr + 1
5                     continue
                  call ilgimit( x_min, y_min, x_max, y_max )
c
c                 plot the data
c
                  call ilgdata( time(start+1), original, 1,
     1                wind_len, 0 )
                  if ( corrected(stream) ) then
                      call ilgdata( time(start+1), altered, 1,
     1                    wind_len, 0 )
                      end if
6                 continue
              write( Ctitle, 7 ) y(stream)
7             format( 'y = ', f5.3, ' mm' )
              call ilgtitl( Ctitle )
              call ilgxlab( 'time (seconds)' )
c
c                 Plot error in each window
c                 *************************
c
          else if ( item .eq. 10 .and. corrected(stream) ) then
              call ilgopen( Capture_file )
              call ilgpage( 'a4', 'wide', 1, vis_wind )
              do 9 i_wind = first_wind, last_wind
                  start = ( i_wind - 1 ) * wind_len
c
c                 find limits of data
c
                  x_min = time(start+1)
                  x_max = time(start+1)
                  vel_ptr = stream + start * n_streams * 2
                  ord_ptr = 1
                  y_min = velocity(vel_ptr)-velocity(vel_ptr+n_streams)
                  y_max = velocity(vel_ptr)-velocity(vel_ptr+n_streams)
                  do 8 i = start + 1, start + wind_len
                      x_min = min( x_min, time(i) )
                      x_max = max( x_max, time(i) )
                      original( ord_ptr ) = velocity( vel_ptr ) -
     1                    velocity(vel_ptr + n_streams )
                      y_min = min( y_min, original(ord_ptr) )
                      y_max = max( y_max, original(ord_ptr) )
                      vel_ptr = vel_ptr + n_streams * 2
                      ord_ptr = ord_ptr + 1
8                     continue
                  call ilgimit( x_min, y_min, x_max, y_max )
c
c                 plot the data
c
                  call ilgdata( time(start+1), original, 1,
     1                wind_len, 0 )
9                 continue
              write( Ctitle, 7 ) y(stream)
              call ilgtitl( Ctitle )
              call ilgxlab( 'time (seconds)' )
              end if
c
c                 check parameters
c
          vis_wind = min( vis_wind, n_wind )
          first_wind = min( first_wind, n_wind )
          last_wind = min( last_wind, n_wind )
          goto 1
c         end of loop
c
      end



      SUBROUTINE MEAN_STREAM( IDATLU, KINEM_VISC, FRICTION_VEL )
c
c  Calculate, write and plot the statistics of all data streams .
c
c Argument
c  integer idatlu     # logical unit no. of data file
c  real kinem_visc    # INPUT  # kinematic viscosity
c       friction_vel  # INPUT  # friction_velocity
c
c Local variables
c  logical
c    got_means    # statistical moments have been calculated
c    some_corr    # at least one correction has been done
c    set_wind(0:MAX_MOMENT)  # window limits have been specified
c    Equery       # function : answer yes/no
c  integer
c    istr, nstr   # data stream number -- index
c    cstr         # corrected stream number -- index
c    imom,        # moment index
c    item         # item selected from menu
c    cur_moment   # currently selected moment
c    fil_lu       # logical unit number of the output file
c    n_moments   # maximum moment order
c    op_type      # output type : 1=simple, 2=non-dimensional
c    open_file    # function : open file
c    Jresp        # function : read integer
c  real
c    cor_ord      # temporary corrected ordinate value
c    x_min, y_min # lower left corner of plotting window
c    x_max, y_max # upper right corner of plotting window
c    ref_x, ref_y # data points for fixed reference lines on dimensionless mean
c    means(MAX_MOMENT,2*MAX_COLS) # array of statistical moments of distribution
c    abscissa(MAX_MOMENT)        # abscissa values - for plotting
c    ordinate(MAX_MOMENT)        # ordinate values - for plotting
c    xmin_wind(0:MAX_MOMENT)     # lower left window limit x
c    ymin_wind(0:MAX_MOMENT)     # lower left window limit y
c    xmax_wind(0:MAX_MOMENT)     # upper right window limit x
c    ymax_wind(0:MAX_MOMENT)     # upper right window limit y
c  character
c    capture_file*64  # name of capture file
c    Cdescript(8)*14  # description of each moment order
c    Cout_type(2)*15  # description of output type
c
      include 'stdio.def'
      include 'prox_law.def'
c
      logical got_means, some_corr, set_wind(0:MAX_MOMENT), Equery
      integer istr, imom, item, cstr, cur_moment, fil_lu, n_moments,
     1    idatlu, op_type, nstr, open_file, Jresp
      real kinem_visc, friction_vel, x_min, y_min, x_max, y_max,
     1    cor_ord, ref_x(6), ref_y(6),
     2    xmin_wind(0:MAX_MOMENT), xmax_wind(0:MAX_MOMENT),
     3    ymin_wind(0:MAX_MOMENT), ymax_wind(0:MAX_MOMENT)
      character capture_file*64, Cdescript(0:MAX_MOMENT)*14,
     1    Cout_type(2)*15
c
      include 'prox_law.cmn'
      real means(0:MAX_MOMENT,2*MAX_COLS),
     1     abscissa(MAX_COLS), ordinate(MAX_COLS)
c
      data ref_x / 0.0, 10.0, 0.0, 10.0, 0.0, 10.0 /
      data ref_y / 0.0, 10.0, 0.0,  9.0, 0.0,  8.0 /
      data Cdescript / 'turb intensity', 'mean', 'r.m.s',
     1    'skew ness', 'flat ness', 'superskewness', 'superflatness',
     2    'hyperskewness', 'hyperflatness' /
      data capture_file / ' ' /
      data Cout_type / 'simple', 'non-dimensional' /
      data op_type / 2 /
      data set_wind / 9 * .FALSE. /
c
      got_means  = .false.
      cur_moment = 1
      n_moments = 4
c     start of loop
1         call newscr()
          write( STD_OUT, 2 ) n_moments, Cdescript(cur_moment),
     1        capture_file, Cout_type(op_type)
2         format( ' Select' / '   0) Exit' /
     1            '   1) set maximum moment: ', i2 /
     2            '   2) set current moment: ', a /
     3            '   3) capture file  : ', a /
     4            '   4) output type   : ', a /
     5            '   5) set graph window limits ' /
     6            '   6) write moment profiles' /
     7            '   7) plot moment profile   ', $ )
          item = Jresp( STD_IN, 0, 7, MAX_TRIES )
          call ilgclos()
c
c                 RETURN
c                 ******
c
          if ( item .eq. 0 ) then
              return
c
c                 SET MAXIMUM MOMENT
c                 ******************
c
          else if ( item .eq. 1 ) then
              write( STD_OUT, 3 )
3             format( ' What value ? ', $ )
              n_moments = Jresp( STD_IN, 2, MAX_MOMENT, MAX_TRIES )
c
c                 SET CURRENT MOMENT
c                 ******************
c
          else if ( item .eq. 2 ) then
              write( STD_OUT, 9 ) ( imom, Cdescript(imom),
     1            imom = 0, n_moments )
9             format( ' select ', i1, ') ', a / ( 8x, i1, ') ', a ) )
              cur_moment = Jresp( STD_IN, 0, n_moments, MAX_TRIES )
c
c                 SET CAPTURE FILE
c                 ****************
c
          else if ( item .eq. 3 ) then
              write( STD_OUT, 3 )
              read( STD_IN, 4 ) capture_file
4             format( a )
c
c                 SET OUTPUT TYPE ... SIMPLE / NONDIMENSIONAL
c                 *******************************************
c
          else if ( item .eq. 4 ) then
              op_type = mod( op_type, 2 ) + 1
c
c                 SET/REMOVE GRAPH WINDOW LIMITS
c                 ******************************
c
          else if ( item .eq. 5 ) then
              set_wind(cur_moment) = .not. Equery( STD_IN,
     1            'Remove window limits ?' )
              if ( set_wind( cur_moment ) ) then
                  write( STD_OUT, 10 )
10                format( ' Lower left corner (coordinate) ? ', $ )
                  read ( STD_IN, *, ERR=27, END=29 )
     1                xmin_wind(cur_moment), ymin_wind(cur_moment)
                  write( STD_OUT, 11 )
11                format( ' Upper right corner (coordinate) ? ', $ )
                  read ( STD_IN, *, ERR=27, END=29 )
     1                xmax_wind(cur_moment), ymax_wind(cur_moment)
                  end if
c
c                 CALCULATE THE MEAN VALUES
c                 *************************
c
          else if ( item .eq. 6 .or. item .eq. 7 ) then
              if ( .not. got_means ) then
                  do 16 istr = 1, n_streams
                      means( 0, istr ) = -9.99999
                      means( 1, istr ) = -9.99999
                      means( 2, istr ) = -9.99999
                      do 12 imom = 3, MAX_MOMENT
                          means( imom, istr ) = -99.99
                          means( imom, istr + n_streams ) = -99.99
12                        continue
                      cstr = istr + n_streams
                      call smeang( velocity(istr), 2 * n_streams,
     1                    n_data(istr), means( 1, istr ), n_moments )
                      if ( means( 2, istr ) .gt. 0 ) then
                          if ( n_moments .gt. 1 ) then
                              means(2,istr) = sqrt( means(2,istr) )
                              means(0,istr) = means(2,istr) /
     1                            means(1,istr)
                              end if
                          do 13 imom = 3, n_moments
                              means(imom,istr) = means(imom,istr) /
     1                            means(2,istr) ** imom
13                            continue
                      else
                          write( STD_ERR, 14 ) istr
14     format( ' MEAN_STREAM : WARNING : rms = 0 for stream ', i3,
     1        ' Cannot calculate skewness etc.' )
                          end if
                      if ( corrected(istr) ) then
                          call smeang( velocity(cstr),
     1                        2 * n_streams, n_data(istr),
     2                        means(1,cstr), n_moments )
                          if ( means( 2, cstr ) .gt. 0 ) then
                              if ( n_moments .gt. 1 ) then
                                  means(2,cstr) = sqrt( means(2,cstr) )
                                  means(0,cstr) = means(2,cstr) /
     1                                means(1,cstr)
                                  end if
                              do 15 imom = 3, n_moments
                                  means(imom,cstr) = means(imom,cstr) /
     1                                means(2,cstr) ** imom
15                                continue
                          else
                              write( STD_OUT, 14 ) istr
                              end if
                          end if
16                    continue
                  got_means = .TRUE.
                  end if
c
c                 WRITE MOMENT PROFILES
c                 *********************
c
              if ( item .eq. 6 ) then
c
c                 open file
c
                  fil_lu = open_file( idatlu, '?', 'new', 0 )
                  if ( fil_lu .eq. 0 ) goto 1
c
c                 uncorrected header
c
                  write( fil_lu, 17 )  'UNCORRECTED MEANS',
     1                ( Cdescript(imom)(1:5), imom = 0, n_moments )
17                format( / 5x, a, / '    y ', 8( 4x, a5 ) )
                  write( fil_lu, 18 )
     1                ( Cdescript(imom)(6:9), imom = 0, n_moments )
18                format( '  (mm)', 8( 4x, a5 ) )
c
c                 uncorrected dimensioned values
c
                  if ( op_type .ne. 2 ) then
                      do 20 istr = 1, n_streams
                          write( fil_lu, 19 ) y(istr),
     1                        ( means(imom,istr), imom=0,n_moments )
19                        format( 1x, f7.3, 3f9.5, 2f9.3, 4f9.2 )
20                        continue
c
c                 uncorrected dimensionless values
c
                  else
                      do 21 istr = 1, n_streams
                          write( fil_lu, 19 )
     1  y(istr) * friction_vel / kinem_visc / 1000.0, means(0,istr),
     2  means(1,istr) / friction_vel,
     3  means(2,istr) / friction_vel,
     4  ( means( imom, istr ), imom = 3, n_moments )
21                        continue
                      end if
c
c                 corrected - header
c           don't print anything if no corrections have been done
c
                  some_corr = .FALSE.
                  do 22 istr = 1, n_streams
                      some_corr = some_corr .or. corrected(istr)
22                    continue
                  if ( some_corr ) then
                      write( fil_lu, 17 )  'CORRECTED MEANS',
     1                    ( Cdescript(imom)(1:5), imom = 0, n_moments )
                      write( fil_lu, 18 )
     1                    ( Cdescript(imom)(6:9), imom = 0, n_moments )
c
c                 corrected dimensioned values
c
                      if ( op_type .ne. 2 ) then
                          do 23 istr = 1, n_streams
                              if ( corrected(istr) ) then
                                  write( fil_lu, 19 )
     1                                y(istr), ( means(imom,istr+
     2                                n_streams), imom=0,n_moments )
                                  end if
23                            continue
c
c                 corrected dimensionless values
c
                      else
                          do 24 istr = 1, n_streams
                              nstr = istr + n_streams
                              write( fil_lu, 19 )
     1  y(istr) * friction_vel / kinem_visc / 1000.0, means(0,nstr),
     2  means(1,nstr) / friction_vel,
     3  means(2,nstr) / friction_vel,
     4  ( means( imom, nstr ), imom = 3, n_moments )
24                            continue
                          end if
                      end if
                  if ( fil_lu .eq. STD_OUT ) then
                      call ilgwait()
                  else
                      close( UNIT=fil_lu )
                      end if
c
c                   PLOT MOMENT PROFILE
c                   *******************
c
              else if ( item .eq. 7 ) then
                  call ilgopen( capture_file )
                  call ilgraph( 1, 1, 11, 11, 1 )
c
c                   find domain & range of data
c
                  if ( cur_moment .le. 2 ) then
                      x_min = 0.0
                      y_min = 0.0
                  else
                      x_min = 1.0e37
                      y_min = 1.0e37
                      end if
                  x_max = -1.0e37
                  y_max = -1.0e37
c
                  do 25 istr = 1, n_streams
                      abscissa(istr) = y(istr)
                      ordinate(istr) = means( cur_moment, istr )
                      if ( op_type .eq. 2 ) then
                          abscissa(istr) = abscissa(istr) * friction_vel
     1                                     / kinem_visc / 1000.0
                          if ( cur_moment .eq. 1 .or.
     1                         cur_moment .eq. 2 ) then
                              ordinate(istr) = ordinate(istr) /
     1                                         friction_vel
                              end if
                          end if
                      x_min = min( x_min, abscissa(istr) )
                      x_max = max( x_max, abscissa(istr) )
                      y_min = min( y_min, ordinate(istr) )
                      y_max = max( y_max, ordinate(istr) )
                      if ( corrected(istr) ) then
                          cor_ord = means( cur_moment, n_streams+ istr )
                          if ( op_type .eq. 2 .and. cor_ord .gt. 0.0
     1                        .and. ( cur_moment .eq. 1 .or.
     2                                cur_moment .eq. 2 ) ) then
                              cor_ord = cor_ord / friction_vel
                              end if
                          y_min = min( y_min, cor_ord )
                          y_max = max( y_max, cor_ord )
                          end if
25                    continue
c
c                   plot uncorrected data
c
                  if ( set_wind(cur_moment) ) then
                      call ilgwild( 'UU',
     1                    xmin_wind(cur_moment), ymin_wind(cur_moment),
     2                    xmax_wind(cur_moment), ymax_wind(cur_moment) )
                      call ilgimit(
     1                    xmin_wind(cur_moment), ymin_wind(cur_moment),
     2                    xmax_wind(cur_moment), ymax_wind(cur_moment) )
                  else
                      call ilgimit( x_min, y_min, x_max, y_max )
                      end if
                  call ilgline( 1, 0, 1 )
                  call ilgdata( abscissa, ordinate, 1, n_streams, 0 )
c
c                   plot corrected data
c
                  do 26 istr = 1, n_streams
                      ordinate(istr) =
     1                    means( cur_moment, n_streams + istr )
                      if ( op_type .eq. 2 .and. ( cur_moment .eq. 1
     1                        .or. cur_moment .eq. 2 ) ) then
                          ordinate(istr) = ordinate(istr) / friction_vel
                          end if
26                    continue
                  call ilgline( 3, 0, 1 )
                  call ilgdata( abscissa, ordinate, 1, n_streams, 0 )
                  call ilgxlab( 'distance from wall' )
                  call ilgylab( Cdescript(cur_moment) )
c
c                   plot reference lines on graph of dimensionless mean
c
                  if ( op_type .eq. 2 .and. cur_moment .eq. 1 ) then
                      call ilgline( 0, 1, 1 )
                      call ilgdata( ref_x, ref_y, 1, 6, 0 )
                      end if
              else
                  stop 'MEAN_STREAM - Bug'
                  end if
              end if
          cur_moment = min( cur_moment, n_moments )
          goto 1
c
c                   read error trap
c
27        call print_err( 'MEAN_STREAM : Cannot decode value(s)' )
          goto 1
c         end of loop
c
c                   end of file trap
c
29    stop 'MEAN_STREAM : ABORT'
      end



      SUBROUTINE ERROR_STATS( IDATLU, KINEM_VISC, FRICTION_VEL,
     1                        CORR_LAW, WORK )
c
c  Calculate, write and plot the statistics of all data streams .
c
c Argument
c  integer idatlu     # logical unit no. of data file
c  real kinem_visc    # INPUT  # kinematic viscosity
c       friction_vel  # INPUT  # friction_velocity
c       corr_law      # INPUT  # correction law
c  real work(*)       # work array
c
c Local variables
c  logical
c    got_means    # statistical moments have been calculated
c    set_wind(0:MAX_MOMENT)  # window limits have been specified
c    Equery       # function : yes/no answer
c  integer
c    cur_moment   # currently selected moment
c    fil_lu       # logical unit number of the output file
c    i            # data array index
c    istr,        # data stream number -- index
c    imom,        # moment index
c    item         # item selected from menu
c    n_moments   # maximum moment order
c    op_type      # output type : 1=simple, 2=non-dimensional
c    src1, src2   # source array pointers
c    open_file    # function : open file
c    Jresp        # function : read integer
c  real
c    abscissa(MAX_MOMENT)        # abscissa values - for plotting
c    means(MAX_MOMENT,MAX_COLS)  # array of statistical moments of distribution
c    ordinate(MAX_MOMENT)        # ordinate values - for plotting
c    xmin_wind(0:MAX_MOMENT),    # lower left window limit x
c    ymin_wind(0:MAX_MOMENT)     # lower left window limit y
c    xmax_wind(0:MAX_MOMENT),    # upper right window limit x
c    ymax_wind(0:MAX_MOMENT)     # upper right window limit y
c    x_min, y_min # lower left corner of plotting window
c    x_max, y_max # upper right corner of plotting window
c    temp_v_fact, temp_y_fact    # temporary storage for v_ & y_ factors
c    prox_law                    # function : evaluate correction function
c    QHW1_UtoE2                  # function : velocity to volts_squared
c  character
c    capture_file*64  # name of capture file
c    Cdescript(8)*14  # description of each moment order
c    Cout_type(2)*15  # description of output type
c
      include 'stdio.def'
      include 'wallprox.def'
      include 'prox_law.def'
      character Cno_method*32
      parameter( Cno_method = 'ERROR_STATS: BUG: Unknown method' )
c
      logical got_means, set_wind(0:MAX_MOMENT), Equery
      integer idatlu, cur_moment, fil_lu, i, istr, imom, item, n_diff,
     1        n_moments, op_type, src1, src2, open_file, Jresp
      real kinem_visc, friction_vel, corr_law(*), work(*),
     1     x_min, y_min, x_max, y_max, temp_v_fact, temp_y_fact,
     2     xmin_wind(0:MAX_MOMENT), xmax_wind(0:MAX_MOMENT),
     3     ymin_wind(0:MAX_MOMENT), ymax_wind(0:MAX_MOMENT),
     4     prox_law, QHW1_UtoE2
      real means(MAX_MOMENT,MAX_COLS),
     1     abscissa(MAX_COLS), ordinate(MAX_COLS),
     2     diff_sum, diff_min, diff_max, diff_fact
      character capture_file*64, Cdescript(8)*14, Cout_type(2)*15
c
c
      include 'prox_law.cmn'
c
      save Cdescript, capture_file, Cout_type, op_type, set_wind,
     1     diff_min, diff_max
      data Cdescript / 'mean', 'r.m.s', 'skew ness', 'flat ness',
     1    'superskewness', 'superflatness', 'hyperskewness',
     2    'hyperflatness' /
      data capture_file / ' ' /
      data Cout_type / 'simple', 'non-dimensional' /
      data op_type / 2 /
      data set_wind / 9 * .FALSE. /
      data diff_min, diff_max / 0.0, 3.0 /
c
      got_means  = .false.
      cur_moment = 1
      n_moments = 4
c     start of loop                                                      1b
1         call newscr()
          write( STD_OUT, 2 ) n_moments, Cdescript(cur_moment),
     1        capture_file, Cout_type(op_type), diff_min, diff_max
2         format( ' Select' / '   0) Exit' /
     1            '   1) maximum moment: ', i2 /
     2            '   2) current moment: ', a /
     3            '   3) capture file  : ', a /
     4            '   4) output type   : ', a /
     5            '   5) set graph window limits ' /
     6            '   6) write moment profiles' /
     7            '   7) plot moment profile' /
     8            '   8) Comparison limits   ', 2f7.3, ' ', $ )
          item = Jresp( STD_IN, 0, 8, MAX_TRIES )
          call ilgclos()
c
c                 RETURN
c                 ******
c
          if ( item .eq. 0 ) then                                        2b
              return
c
c                 SET MAXIMUM MOMENT
c                 ******************
c
          else if ( item .eq. 1 ) then                                   2m
              write( STD_OUT, 3 )
3             format( ' What value ? ', $ )
              n_moments = Jresp( STD_IN, 2, MAX_MOMENT, MAX_TRIES )
c
c                 SET CURRENT MOMENT
c                 ******************
c
          else if ( item .eq. 2 ) then                                   2m
              write( STD_OUT, 4 ) ( i, Cdescript(i), i = 1, n_moments )
4             format( ' select ', i1, ') ', a / ( 8x, i1, ') ', a ) )
              cur_moment = Jresp( STD_IN, 1, n_moments, MAX_TRIES )
c
c                 SET CAPTURE FILE
c                 ****************
c
          else if ( item .eq. 3 ) then                                   2m
              write( STD_OUT, 3 )
              read( STD_IN, 5 ) capture_file
5             format( a )
c
c                 SET OUTPUT TYPE ... SIMPLE / NONDIMENSIONAL
c                 *******************************************
c
          else if ( item .eq. 4 ) then                                   2m
              op_type = 3 - op_type
c
c                 SET/REMOVE GRAPH WINDOW LIMITS
c                 ******************************
c
          else if ( item .eq. 5 ) then
              set_wind(cur_moment) = .not. Equery( STD_IN,
     1            'Remove window limits ?' )
              if ( set_wind( cur_moment ) ) then
                  write( STD_OUT, 6 )
6                 format( ' Lower left corner (coordinate) ? ', $ )
                  read ( STD_IN, *, ERR=23, END=24 )
     1                xmin_wind(cur_moment), ymin_wind(cur_moment)
                  write( STD_OUT, 7 )
7                 format( ' Upper right corner (coordinate) ? ', $ )
                  read ( STD_IN, *, ERR=23, END=24 )
     1                xmax_wind(cur_moment), ymax_wind(cur_moment)
                  end if
c
c                 CALCULATE THE MEAN VALUES
c                 *************************
c
          else if ( item .eq. 6 .or. item .eq. 7 ) then                  2m
              if ( .not. got_means ) then                                3b
                  do 14 istr = 1, n_streams                              4b
c
c                 initialize values of means
                      means(1,istr) = -9.99999
                      means(2,istr) = -9.99999
                      do 8 imom = 3, MAX_MOMENT                          5b
                          means(imom,istr) = -99.99
8                         continue                                       5e
                      if ( corrected(istr) ) then                        5b
c
c                 calculate error stream
                          if ( method .eq. VELOC_METHOD ) then           6b
                              src1 = istr
                              src2 = istr + n_streams
                              do 9 i = 1, n_data(istr)                   7b
                                  work(i) = velocity(src1) -
     1                                velocity(src2)
                                  src1 = src1 + 2 * n_streams
                                  src2 = src2 + 2 * n_streams
9                                 continue                               7e
                          else if ( method .eq. HEAT_METHOD ) then       6m
                              src1 = istr
                              src2 = istr + n_streams
                              do 10 i = 1, n_data(istr)                  7b
                                  work(i) = y(istr) / 1000.0 *
     1                                ( QHW1_UtoE2( velocity(src1) ) -
     2                                  QHW1_UtoE2( velocity(src2) ) )
                                  src1 = src1 + 2 * n_streams
                                  src2 = src2 + 2 * n_streams
10                                continue                               7e
                          else if ( method .eq. MOD_H_METHOD ) then      6m
                              src1 = istr
                              src2 = istr + n_streams
                              do 11 i = 1, n_data(istr)                  7b
                                  work(i) =
     1   ( QHW1_UtoE2( velocity(src1) ) - QHW1_UtoE2( velocity(src2) ) )
     2   * ( y(istr) / 1000.0 ) ** ( 1.0 - HW_CALIB_EXPON_N ) *
     3   friction_vel ** ( 1.0 - 2.0 * HW_CALIB_EXPON_N )
                                  src1 = src1 + 2 * n_streams
                                  src2 = src2 + 2 * n_streams
11                                continue                               7e
                          else                                           6m
                              call print_err( Cno_method )
                              end if                                     6e
c
c                 calculate moments
                          call smeang( work, 1, n_data(istr),
     1                                 means(1,istr), n_moments )
                          if ( means(2,istr) .gt. 0 ) then               6b
                              if ( n_moments .gt. 1 )
     1                            means(2,istr) = sqrt( means(2,istr) )
                              do 12 imom = 3, n_moments                 7b
                                  means(imom,istr) = means(imom,istr) /
     1                                means(2,istr) ** imom
12                                continue                               7e
                          else                                           6m
                              write( STD_ERR, 13 ) istr
13    format( ' MEAN_STREAM : WARNING : rms = 0 for stream ', i3,
     1        ' Cannot calculate skewness etc.' )
                              end if                                     6e
                          end if                                         5e
14                    continue                                           4e
                  got_means = .TRUE.
                  end if                                                 3e
c
c                 WRITE MOMENT PROFILES
c                 *********************
c
              if ( item .eq. 6 ) then                                    3b
c
c                 open file
c
                  fil_lu = open_file( idatlu, '?', 'new', 0 )
                  if ( fil_lu .eq. 0 ) goto 1
c
c                 uncorrected header
c
                  write( fil_lu, 15 )  ' ERROR STATISTICS',
     1                ( Cdescript(imom)(1:5), imom = 1, n_moments )
15                format( / 5x, a, / '    y ', 8( 4x, a5 ) )
                  write( fil_lu, 16 )
     1                ( Cdescript(imom)(6:9), imom = 1, n_moments )
16                format( '  (mm)', 8( 4x, a5 ) )
c
c                 uncorrected dimensioned values
c
                  if ( op_type .ne. 2 ) then                             4b
                      do 18 istr = 1, n_streams                          5b
                          write( fil_lu, 17 ) y(istr),
     1                        ( means(imom,istr), imom=1,n_moments )
17                        format( 1x, f7.3, 2f9.5, 2f9.3, 4f9.2 )
18                        continue                                       5e
c
c                 uncorrected dimensionless values
c
                  else if ( method .eq. VELOC_METHOD ) then              4m
                      do 19 istr = 1, n_streams                          5b
                          write( fil_lu, 17 )
     1  y(istr) * friction_vel / kinem_visc / 1000.0,
     2  means(1,istr) / friction_vel,
     3  means(2,istr) / friction_vel,
     4  ( means( imom, istr ), imom = 3, n_moments )
19                        continue                                       5e
                  else if ( method .eq. HEAT_METHOD .or.
     1                      method .eq. MOD_H_METHOD ) then              4m
                      do 21 istr = 1, n_streams                          5b
                          write( fil_lu, 20 )
     1  y(istr) * friction_vel / kinem_visc / 1000.0,
     2  means(1,istr), means(2,istr),
     3  ( means( imom, istr ), imom = 3, n_moments )
20                        format( 1x, f7.3, 2g11.4e1, 2f9.3, 4f9.2 )
21                        continue                                       5e
                  else                                                   4m
                      call print_err( Cno_method )
                      end if                                             4e
                  if ( fil_lu .eq. STD_OUT ) then                        4b
                      call ilgwait()
                  else                                                   4m
                      close( UNIT=fil_lu )
                      end if                                             4e
c
c                   PLOT MOMENT PROFILE
c                   *******************
c
              else if ( item .eq. 7 ) then                               3m
                  call ilgopen( capture_file )
                  call ilgraph( 1, 1, 11, 11, 1 )
c
c                   find domain & range of data
c
                  if ( cur_moment .le. 2 ) then                          4b
                      x_min = 0.0
                      y_min = 0.0
                  else                                                   4m
                      x_min = 1.0e37
                      y_min = 1.0e37
                      end if                                             4e
                  x_max = -1.0e37
                  y_max = -1.0e37
c
                  do 22 istr = 1, n_streams                              4b
                      if ( corrected(istr) ) then                        5b
                          abscissa(istr) = y(istr)
                          ordinate(istr) = means( cur_moment, istr )
                          if ( op_type .eq. 2 ) then                     6b
                              abscissa(istr) = abscissa(istr) *
     1                            friction_vel / kinem_visc / 1000.0
                              if ( method .eq. VELOC_METHOD .and.
     1                             cur_moment .le. 2 ) then              7b
                                  ordinate(istr) = ordinate(istr) /
     1                                friction_vel
                                  end if                                 7e
                              end if                                     6e
                          x_min = min( x_min, abscissa(istr) )
                          x_max = max( x_max, abscissa(istr) )
                          y_min = min( y_min, ordinate(istr) )
                          y_max = max( y_max, ordinate(istr) )
                          end if                                         5e
22                    continue                                           4e
c
c                   plot data
c
                  if ( set_wind(cur_moment) ) then
                      call ilgwild( 'UU',
     1                    xmin_wind(cur_moment), ymin_wind(cur_moment),
     2                    xmax_wind(cur_moment), ymax_wind(cur_moment) )
                      call ilgimit(
     1                    xmin_wind(cur_moment), ymin_wind(cur_moment),
     2                    xmax_wind(cur_moment), ymax_wind(cur_moment) )
                  else
                      call ilgimit( x_min, y_min, x_max, y_max )
                      end if
                  call ilgline( 1, 0, 1 )
                  call ilgdata( abscissa, ordinate, 1, n_streams, 0 )
c
c                   plot mean correction law ( using v_ & y_ factors of 1 )
c
                  if ( op_type .eq. 2 .and. cur_moment .eq. 1 ) then     4b
                      temp_v_fact = corr_law(Hv_fact)
                      temp_y_fact = corr_law(Hy_fact)
                      corr_law(Hv_fact) = 1.0
                      corr_law(Hy_fact) = 1.0
                      call plot_corr_law( idatlu, corr_law, .false.,
     1                                    friction_vel )
                      diff_sum = 0.0
                      n_diff = 0
                      if ( int( corr_law(Htype) + 0.5 ) .eq. 7 ) then    5b
                          diff_fact = friction_vel **
     1                                ( HW_CALIB_EXPON_N - 1.0 )
                      else                                               5m
                          diff_fact = 1.0
                          end if                                         5e
                      do 25  istr = 1, n_streams                         5b
                          if ( diff_min .le. abscissa(istr) .and.
     1                         abscissa(istr) .le. diff_max ) then       6b
                              diff_sum = diff_sum +
     2                            ( ordinate(istr) - diff_fact *
     3                              prox_law( abscissa(istr), corr_law )
     4                            ) **2
                              n_diff = n_diff + 1
                              end if                                     6e
25                        continue                                       5e
                      if ( n_diff .gt. 0 ) then                          5b
                          write( STD_OUT, 26 )
     1                        sqrt( diff_sum / real( n_diff ) )
26                        format( ' Mean square difference = ', g12.5 )
                          call ilgwait()
                          end if
                      corr_law(Hv_fact) = temp_v_fact
                      corr_law(Hy_fact) = temp_y_fact
                      end if                                             4e
              else                                                       3m
                  stop 'ERROR_STATISTICS - Bug'
                  end if                                                 3e
          else if ( item .eq. 8 ) then                                   2m
              write( STD_OUT, 27 )
27            format( ' Comparison limits (min,max)? ', $ )
              read( STD_IN, *, ERR=23, END=24 ) diff_min, diff_max
              if ( diff_min .gt. diff_max )
     1            call swap4( diff_min, diff_max )
              end if                                                     2e
          cur_moment = min( cur_moment, n_moments )
          goto 1
c
c                   read error trap
c
23        call print_err( 'ERROR_STATS : Cannot decode value(s)' )
          goto 1
c         end of loop
c
c                   end of file trap
c
24    stop 'ERROR_STATS : ABORT'
      end



      SUBROUTINE WRITE_STREAM( IDATLU )
c
c Writes a corrected test data stream to file . The time is written as
c the first column in the table . The corrected velocity data is written
c as additional columns . Data which has not been corrected is left as
c blank columns in the table .
c
c Argument
c  integer idatlu           # logical unit no. of data file
c
c Data file format :-
c   The data file contains a number of velocity time series records,
c each of which is stored as one column of data . The first column of
c data is time (in seconds) . A ';' marks the beginning of a comment
c field in the file . The comment field continues until the end of line.
c The first non-comment line in the file is read as a sequence of probe
c locations (in mm from the wall) . The expected number of velocity
c records is the same as the number of probe locations . Any line which
c contains insufficient data is rejected . Any surplus data on a line is
c ignored .
c
c Local variables
c  integer
c    fil_lu       # logical unit number of output file
c    col, row     # row and column number of the output table
c    row_ptr      # pointer to row in velocity table
c    vel_ptr      # pointer to velocity data
c    str_ptr      # pointer to output string
c    last_char    # pointer to last printed character in line
c    max_str      # last stream to be written
c    min_str      # first stream to be written
c    Jresp        # function : read integer
c    open_file    # function : open file
c  character
c    line*80      # line written to file
c
      include 'stdio.def'
      include 'prox_law.def'
c
      integer fil_lu, col, row, vel_ptr, str_ptr, last_char,
     1        idatlu, row_ptr, max_str, min_str, Jresp, open_file
      character line*80
c
      include 'prox_law.cmn'
c
c
      if ( n_streams .gt. 4 ) then
          write( STD_OUT, 1 )
1         format( ' A maximum of 4 successive streams will be written.'
     1            / ' Which is the first ? ', $ )
          min_str = Jresp( STD_IN, 1, n_streams, MAX_TRIES )
      else
          min_str = 1
          end if
c
c             open file
c
      fil_lu = open_file( idatlu, '?', 'new', 0 )
      if ( fil_lu .eq. 0 ) return
c
c             write comment
c
      if ( fil_lu .ne. STD_OUT ) then
          write( STD_OUT, 2 )
2         format( ' Enter comment (end with .)' )
c         start of loop
3             read( STD_IN, 4, ERR=14, END=20 ) line
4             format( a )
              if ( line .eq. '.' ) goto 6
              write( fil_lu, 5, ERR=16 ) line
5             format( ' ; ', a )
              goto 3
          end if
c
c             write probe locations
c
6     max_str = min( min_str + 3, n_streams )
      write( fil_lu, 7, ERR=16 ) (y(col), col = min_str, max_str )
7     format( ' probe locations (mm)' / ( 1x, 4f16.5 ) )
      write( fil_lu, 8, ERR=16 )
8     format( '  time       alternate measured and corrected velocity' )
      row_ptr = 0
      do 13 row = 1, n_rows
c
c             clear the line
c
          line = ' '
c
c             encode time
c
          write( line(2:8), 9, ERR=18 ) time(row)
9         format( f7.4 )
c
c             encode velocity values
c
          str_ptr = -6
          last_char = 1
          vel_ptr = row_ptr + min_str
          do 11  col = min_str, max_str
              str_ptr = str_ptr + 16
              write( line(str_ptr:str_ptr+6), 10, ERR=18 )
     1            velocity(vel_ptr)
10             format( f7.5 )
              if ( corrected(col) ) then
                  write( line(str_ptr+8:str_ptr+14), 10, ERR=18 )
     1                velocity(vel_ptr+n_streams)
                  last_char = str_ptr + 14
              else
                  last_char = str_ptr + 6
                  end if
              vel_ptr = vel_ptr + 1
11            continue
c
c               write line
c
          write( fil_lu, 12, ERR=13 ) line(:last_char)
12        format( a )
          row_ptr = row_ptr + 2 * n_streams
13        continue
      if ( fil_lu .eq. STD_OUT )then
          call ilgwait()
      else
          close( UNIT=fil_lu )
          end if
      return
c
c               traps
c
14    call print_err( ' Cannot read input' )
      return
c
16    call print_err( 'Cannot write to file' )
      return
c
18    call print_err( ' Cannot encode value' )
      return
c
20    stop ' ABORT'
      end



      SUBROUTINE SMEANG( X, INCREM, N_POINTS, M, MAX_MOM )
c
c   Calculates the 1st ( mean ) to kth moments of N_POINTS samples in the array
c X, taking every INCREMth value. All but the 1st moment ( mean ) are centred
c about the sample mean . The maximum and minimum values of MAX_MOM are 14 and
c 1 . N.B. There are two one passes of the data.
c
c    Arguments
c real    x(*),     # data array
c         m(*),     # array of moment results
c integer increm,   # data array indexing increment
c         n_points  # number of data values used
c         max_mom   # maximum moment required
c
c    Variables
c real    qmean     # function : calculates mean
c integer i,j       # looping indices
c double precision
c         r, rk     # temporay products
c         total(2:14)
c
      include 'stdio.def'
c
      real             x(*), m(*), qmean
      double precision r, rk, total(2:14)
      integer          increm, n_points, i, j, max_mom
c
c
c               Check parameters
c
      if( n_points .le. 0 ) then                                          1b
          write( STD_ERR, 1 ) 'N_POINTS', n_points
1         format( ' SMEANG : ABORT : Illegal argument ', a, ' = ', i12 )
      else if ( increm .le. 0 ) then                                      1m
          write( STD_ERR, 1 ) 'INCREM', n_points
      else if ( max_mom .lt. 1 .or. max_mom .gt. 14 ) then                1m
          write( STD_ERR, 1 ) 'MAX_MOM',max_mom
      else                                                                1m
c
c               Calculate mean
c
          m(1) = qmean( x, increm, n_points )
c
c               Calculate sums of (x-mean)**i : i=1..k
c
          if ( max_mom .gt. 1 ) then                                      2b
              do 2 i = 2, max_mom                                         3b
                  total(i) = 0.0
2                 continue                                                3e
              do 4 j = 1, ( n_points - 1 ) * increm + 1, increm           3b
                  r = ( x(j) - m(1) )
                  rk = r
                  do 3 i = 2, max_mom                                     4b
                      rk = rk * r
                      total(i) = total(i) + rk
3                     continue                                            4e
4                 continue                                                3e
c
c               Calculate moments
c
              do 5 i = 2, max_mom                                         3b
                  m(i) = sngl( total(i) ) / float( n_points )
5                 continue
              m(2) = m(2) * float( n_points ) / float( n_points - 1 )
              end if                                                      2e
          end if                                                          1e
      return
      end



