c
c    Copyright (C) 1997  Peter Vernon Lanspeary
c
c    This program is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 2 of the License, or
c    (at your option) any later version.
c
c    This program is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program; if not, write to the Free Software
c    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
c


c
c SOME ADDITIONAL ROUTINES USED FOR PROCESSING BOUNDARY-LAYER DATA
c
c
c    Labelled COMMON blocks  / Ablylib1 / for QUTAU1
c                            / Ablylib3 / for QDELST, QTHETA, QDELEN, SCHKBL,
c                                             SDUCTR
c                            / Ablylib4 / for QDELNN, SCHKBL
c                            / Ablylib5 / for SBHATIA
c


      REAL FUNCTION QBLASIUS( X, Y, U1, NU )
c
c  Calculate the fluid velocity at a distance Y from the wall at a distance
c  X from the leading edge of a flate plate in a zero pressure gradient flow .
c  The boundary layer on the plate is laminar .
c
c  Arguments
c   real x,         # distance from the leading edge of the plate
c        y,         # distance from the surface of the plate
c        U1,        # free stream fluid speed
c        nu,        # kinematic viscosity of the fluid
c
c Constants
c  integer *2 Hntable=18
c
c Variables
c  real zeta(Hntable)
c
      implicit none
      integer *2 Hntable
      parameter( Hntable=18 )
c
      real x, y, U1, nu, zeta, zetas(Hntable), Uratios(Hntable), qlrpl1
      data zetas /   0.0,    0.1,    0.2,    0.3,    0.4,    0.5,
     1               0.6,    0.8,    1.0,    1.2,    1.4,    1.6,
     2               1.8,    2.0,    2.5,    3.0,    3.5,    4.0 /
      data Uratios / 0.0000, 0.0664, 0.1328, 0.1989, 0.2647, 0.3298,
     1               0.3938, 0.5168, 0.6298, 0.7290, 0.8115, 0.8761,
     2               0.9233, 0.9555, 0.9916, 0.9990, 0.9999, 1.0000 /
c
      if ( x .le. 0.0 .or. y .lt. 0.0 .or. U1 .le. 0.0 .or.
     1     nu .le. 0.0 ) then
        qblasius = -1.0
      else
        zeta = y / 2.0 * sqrt( U1 / nu / x )
        if ( zeta .ge. Zetas(Hntable) ) then
          qblasius = U1
        else
          qblasius = U1 * qlrpl1( zetas, Uratios, 1, Hntable, zeta )
          end if
        end if
      return
      end


      REAL FUNCTION QTHWAITES( M, CTYPE )
c
c  Returns the value of the
c    1)  boundary layer shape parameter QTHWAITES = H, or
c    2)  Thwaites' wall shear stress parameter QTHWAITES = L,
c  for a given input value of Thwaites's pressure gradient parameter M,
c  by linear interpolation of the values in the table given by Duncan,Thom
c  and Young on page 275, or of the Curle-Skan Improved Thwaites functions from
c  page 47 of `The Laminar Boundary Layer Equations' by N. Curle (1962) .
c
c    If the pressure gradient parameter M  is less than -0.25, the returned
c  value of H is 2.0, and the returned value of L is 0.50 .
c    If the pressure gradient parameter M is greater than 0.082 for Thwaites'
c  table the returned value of H is 3.7, or 0.090 for the Curle-Skan table the
c  returned value of H is 3.55 and the returned value of L is 0.0 .
c    An error message is printed in either of these cases .
c
c Arguments
c  real M                      # Thwaites pressure gradient parameter
c  character CTYPE*(*)         # result type .
c    The first character of CTYPE is the result type, and if it is
c      1) 'H' or 'h', the shape parameter H is returned,
c      2) 'L' or 'l', Thwaites wall shear stress parameter is returned .
c      3) anything else, the returned result is -1  .
c    If CTYPE has a second character, that second character specifies the
c    table to be used . If it is
c      1) 'T' or 't' the table of Thwaites suggested values is used .
c      2) 'C' or 'c' the Curle-Skan table of improved values is used .
c      3) anything else of CTYPE has only 1 character, the Curle-Skan table
c         is used .
c
c Constants
c  Hvdulu = 6                  # vdu logical unit number
c  Hmm    = 1                  # 1st row of table - for `m' values
c  Hll    = 2                  # 2nd row of table - for `l' values
c  HHH    = 3                  # 3rd row of table - for `H' values
c  HCStable = 26               # number of entries in the Curle-Skan table
c  HTtable  = 19               # number of entries in the Thwaites table
c  Hntable = HCStable + HTtable# total size of table
c Variables
c  real Mtable,                # pointer to `m' value in table
c       Ltable,                # pointer to `l' value in table
c       Htable,                # pointer to `H' value in table
c       table(HHH,Hntable)     # table of m, l and H values
c       qlrpl1,                # function : linear interpolation
c
      implicit none
      integer *2 Hvdulu, Hmm, Hll, HHH, HCStable, HTtable, Hntable
      parameter( Hvdulu=6, Hmm=1, Hll=2, HHH=3, HCStable=26, HTtable=19,
     1  Hntable=HCStable+HTtable )
c
      INTEGER I
      integer istart, ntable
      real table(3,Hntable), M, qlrpl1
      character Ctype*(*)
c
c       Curle-Skan table of improved Thwaites functions
c
      data table /
     1  0.090, 0.000, 3.55,   0.088, 0.015, 3.49,   0.086, 0.027, 3.44,
     2  0.084, 0.038, 3.39,   0.080, 0.056, 3.30,   0.076, 0.072, 3.22,
     3  0.072, 0.085, 3.15,   0.068, 0.095, 3.09,   0.064, 0.104, 3.04,
     4  0.060, 0.113, 2.99,   0.056, 0.122, 2.94,   0.048, 0.138, 2.87,
     5  0.040, 0.153, 2.81,   0.032, 0.168, 2.75,   0.016, 0.195, 2.67,
     6  0.000, 0.220, 2.61,  -0.016, 0.244, 2.55,  -0.032, 0.268, 2.49,
     7 -0.048, 0.291, 2.44,  -0.064, 0.313, 2.39,  -0.080, 0.333, 2.34,
     8 -0.100, 0.359, 2.28,  -0.120, 0.382, 2.23,  -0.140, 0.404, 2.18,
     9 -0.200, 0.463, 2.07,  -0.250, 0.500, 2.00,
c
c       Thwaites' suggested values
c
     1  0.082, 0.000, 3.70,   0.0816,0.016, 3.66,   0.0808,0.030, 3.61,
     2  0.080, 0.039, 3.58,   0.078, 0.055, 3.47,   0.074, 0.076, 3.30,
     3  0.070, 0.089, 3.17,   0.064, 0.104, 3.05,   0.056, 0.122, 2.94,
     4  0.048, 0.138, 2.87,   0.032, 0.168, 2.75,   0.016, 0.195, 2.67,
     5  0.000, 0.220, 2.61,  -0.032, 0.268, 2.49,  -0.064, 0.313, 2.39,
     6 -0.100, 0.359, 2.28,  -0.140, 0.404, 2.18,  -0.200, 0.463, 2.07,
     7 -0.250, 0.500, 2.00 /
c
      istart = 0
      ntable = HCStable
      if ( len( Ctype ) .gt. 1 ) then                                     1b
        if ( Ctype(2:2) .eq. 'T' .or. Ctype(2:2) .eq. 't' ) then          2b
          istart = HCStable
          ntable = HTtable
          end if                                                          2e
        end if                                                            1e
      if ( Ctype(1:1) .eq. 'l' .or. Ctype(1:1) .eq. 'L' ) then            1b
        if ( m .lt. table( Hmm, istart + ntable ) ) then                  2b
          qthwaites = table( Hll, istart + ntable )
          write( Hvdulu, 1 ) m
1         format( ' QTHWAITES : WARNING : ' / ' Pressure gradient ',
     1      'parameter ', g13.6, ' out of range of table ' )
        else if ( m .gt. table( Hmm, istart + 1 ) ) then                  2m
          qthwaites = table( Hll, istart + 1 )
          write( Hvdulu, 1 ) m
        else                                                              2m
          qthwaites = qlrpl1( table( Hmm, istart + 1 ),
     1      table( Hll, istart + 1 ), 3, ntable, M )
          end if                                                          2e
      else if ( Ctype(1:1) .eq. 'h' .or. Ctype(1:1) .eq. 'H' ) then       1m
        if ( m .lt. table( Hmm, istart + ntable ) ) then                  2b
          qthwaites = table( HHH, istart + ntable )
          write( Hvdulu, 1 ) m
        else if ( m .gt. table( Hmm, istart + 1 ) ) then                  2m
          qthwaites = table( HHH, istart + 1 )
          write( Hvdulu, 1 ) m
        else                                                              2m
          qthwaites = qlrpl1( table( Hmm, istart + 1 ),
     1      table( HHH, istart + 1 ), 3, ntable, M )
          end if                                                          2e
      else                                                                1m
        qthwaites = -1
        end if                                                            1e
      return
      end


      REAL FUNCTION QWYNG_F1( K_L, ETA_L )
c
c     Evaluate the signal attenuation of a single hot wire probe normal
c to a free stream due to space-averaging of small-scale turbulence over
c the length of the sensor .
c    This function evaluates the function shown in Fig. 1 of
c Wyngaard J.C. (1968) :
c    Measurement of small-scale turbulence structure with hot-wires,
c    Journal of Physics E (~Scientific Instruments~),} Vol. 1, pp 1105--1108.
c
c Arguments
c   real K_L,   # non-dimensional wavenumber of the signal
c               # = 2 * PI * Frequency / Free_stream_speed * Sensor_length
c        ETA_L  # wire length parameter
c               # = Kolmogorov length / Sensor_length
c
c Method : polynomials have been fitted to the reciprocals of each of the 4
c    curves shown by Wyngaard . The first step is to evaluate all of these
c    polynomials at the required wavenumber K_L . These polynomial values
c    are then used as data for 4 point (cubic) Lagrange interpolation to
c    evaluate the signal attenuation for the particular ETA/L . For the
c    Lagrange interpolation, the polynomial is of  ( 1 - ETA/L ) **2 .
c
c variables                    # polynomials
c                              # order eta/l rms error
c    real eta_l_1(5),          #   4    1    0.0019
c         eta_l_35(7),         #   6   0.35  0.0038
c         eta_l_01(7),         #   6   0.1   0.0069
c         eta_l_00(7)          #   6   0.0   0.0201
c
      implicit none
      integer Hvdulu
      parameter( Hvdulu=6 )
c
      real k_l, eta_l, eta_sqd(4), f(4), qpoly, qlagr_itp,
     1     eta_l_1(5), eta_l_35(7), eta_l_01(7), eta_l_00(7)
c    2     corr, mserr, work(36)
c
      data eta_l_1  / 0.9967170,  0.02949991, 967.5710e-6, -355.9060e-6,
     1              19.42390e-6 /
      data eta_l_35 /  1.000531,   0.02689499, 0.06644019,  -0.02665243,
     1              4.953558e-3, -434.7581e-6, 14.51247e-6 /
      data eta_l_01 / 0.9924484,   0.1150190,  0.06743427,  -0.01175561,
     1              279.9788e-6, 94.54528e-6,-6.150445e-6 /
      data eta_l_00 / 0.9974347,   0.1008902,   0.2058961,  -0.03651981,
     1              2.823407e-3,-51.67374e-6,-1.957242e-6 /
c the given values of ETA/L are
c                      0.0, 0.1,  0.35,   1.0
      data eta_sqd   / 1.0, 0.81, 0.4225, 0.0 /
c
      if ( k_l .lt. 0 .or. k_l .gt. 10 ) then
        qwyng_f1 = 1.0
        write( Hvdulu, 1 )
1       format( ' WYNG_F1 : ABORT :',
     1    ' Non_dimensional wavenumber must be between 0.0 and 10.0' )
      else if ( eta_l .lt. 0 .or. eta_l .gt. 1.0 ) then
        qwyng_f1 = 1.0
        write( Hvdulu, 2 )
2       format( ' WYNG_F1 : ABORT :',
     1    ' Sensor length parameter must be between 0.0 and 1.0' )
      else
        f(1) = 1.0 / qpoly( k_l, eta_l_00, 6 )
        f(2) = 1.0 / qpoly( k_l, eta_l_01, 6 )
        f(3) = 1.0 / qpoly( k_l, eta_l_35, 6 )
        f(4) = 1.0 / qpoly( k_l, eta_l_1, 4 )
        qwyng_f1 = qlagr_itp( eta_sqd, f, 4, ( 1.0 - eta_l ) **2  )
c
c  we leave this code in place because Lagrange interpolation is extremely
c  sensitive to error and if some monkey changes the data, a more robust method
c  may be required
c      call spyreg( eta_sqd, f, 1, 4, 5, corr, mserr, work )
c      qwyng_f1 = qpoly( ( 1.0 - eta_l ) **2, work, 3 )
        end if
      return
      end


