c
c    Copyright (C) 1997  Peter Vernon Lanspeary
c
c    This program is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 2 of the License, or
c    (at your option) any later version.
c
c    This program is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program; if not, write to the Free Software
c    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
c


      PROGRAM RESCUE
c
c Read velocity boundary layer data from file, determine the coefficients of a
c law-of-the-wall/law-of-the-wake model and rewrite the input data in non-
c dimensional form.
c
c Constants
c  MAX_DATA = 100   # maximum number of data values in a velocity profile
c  MAX_TRIES = 4    # maximum number of data-entry attempts
c
c Variables
c  integer
c    in_lu          # logical unit number of input file
c    i_profile      # profile number
c    n_data         # number of data points in each profile
c    out_len        # length of output file name
c    item           # item selected from menu
c    out_lu         # logical unit number of output file
c    iargc          # function : number of program arguments
c    Jresp          #          : integer response
c    new_lun        #          : get new logical unit number
c    open_file      #          : open a file
c    str_length     #          : length of a string
c  logical
c    done_analysis  # TRUE if analysis has been done on current input data
c    result_written # TRUE if results are written to output file
c    got_data       # TRUE if input data is read and available
c    ok             # temporary
c  real
c    free_stream    # free stream velocity in m/s
c    kinem_visc     # kinematic viscosity of the fluid in sq.m/s
c    wake_strength  # Coles' PI, wake strength in multiples of friction velocity
c    delta_plus     # Coles' delat scaled in wall units
c    frict_vel      # boundary layer friction velocity in m/s
c  velocity profile data
c    U_data(MAX_DATA)  # mean velocity
c    y_data(MAX_DATA)  # distance from the wall
c    rms_data(MAX_DATA)# r.m.s. velocity fluctuation
c    Qu_tau         # function : find friction velocity etc
c  character
c    err_msg*64     # displayed error message
c    in_file*64     # name of input file
c    out_file*64    # name of output file
c    done_str*8     # function : returns "done"/"not done"
c
c
      include 'stdio.def'
      integer MAX_DATA, MAX_TRIES
      parameter( MAX_DATA=100, MAX_TRIES=4 )
c
      integer in_lu, item, i_profile, n_data, out_len, out_lu,
     1        iargc, Jresp, new_lun, open_file, str_length
      logical done_analysis, result_written, got_data, got_error, ok,
     1        Equery, read_profile, write_results
      real    free_stream, kinem_visc, wake_strength, delta_plus,
     1        frict_vel, U_data(MAX_DATA), y_data(MAX_DATA),
     2        rms_data(MAX_DATA), Qu_tau
      character err_msg*64, in_file*64, out_file*64, done_str*8
c
c       open files
c
      if ( iargc() .lt. 2 ) then
          write( STD_ERR, 1 )
1         format( ' FIX_316 : ABORT : Need 2 args' /
     1            '   FIX_316 <input_file> <output file>' )
      else
          call getarg( 1, in_file )
          call getarg( 2, out_file )
          in_lu = open_file( new_lun(), in_file, 'old', 0 )
          out_lu = open_file( new_lun(), out_file, 'new', 0 )
          if ( in_lu .gt. 0 .and. out_lu .gt. 0 ) then
c
              n_data    = 25
              i_profile = 0
              out_len = str_length( out_file )
c             start of loop
2                 call newscr()
                  if ( got_error ) then
                      write( STD_OUT, 5 ) err_msg
5                     format( ' ** ERROR ** ', a )
                      end if
                  write( STD_OUT, 3 ) out_file(:out_len), i_profile,
     1                free_stream, kinem_visc, i_profile + 1,
     2                done_str(done_analysis), done_str(result_written),
     3                n_data
3     format( ' Number of profiles read from ', a, ' = ', i2 /
     1        ' Free stream speed   = ', f9.6, ' m/s' /
     2        ' Kinematic viscosity = ', g11.5e1, ' m^2/s' /
     3        ' Select' / '  0) Exit' /
     4        '  1) Read profile ', i2 /
     5        '  2) Analysis :', a /
     6        '  3) Write results :', a /
     7        '  4) Display results' /
     8        '  5) Rewind data file' /
     9        '  6) Number of data points in profile = ', i3 )
                  item = Jresp( STD_IN, 0, 6, MAX_TRIES )
c
c       EXIT
c       ****
                  if ( item .eq. 0 ) then
                      close( UNIT=in_lu )
                      close( UNIT=out_lu )
                      stop
c
c       READ PROFILE
c       ************
                  else if ( item .eq. 1 ) then
                      ok = .TRUE.
                      if ( i_profile .gt. 0 .and. .not. done_analysis )
     1                    ok = Equery( STD_IN,
     2                         ' Analysis not done; Are you sure ?' )
                      if ( i_profile .gt. 0 .and. .not. result_written )
     1                    ok = Equery( STD_IN,
     2                         ' Results not written; Are you sure ?' )
                      if ( ok ) then
                          got_data = read_profile( in_lu, free_stream,
     1                        kinem_visc, y_data, U_data, rms_data,
     2                        n_data, err_msg )
                          done_analysis = .FALSE.
                          result_written = .FALSE.
                          got_error = .not. got_data
                          if ( got_data ) then
                              i_profile = i_profile + 1
                              end if
                          end if
c
c       ANALYSIS
c       ********
                  else if ( item .eq. 2 ) then
                      if ( got_data ) then
                          frict_vel = Qu_tau(U_data,  y_data, 1, n_data,
     1                        free_stream, kinem_visc,
     2                        wake_strength, delta_plus )
                          done_analysis = frict_vel .gt. 0.0
                          got_error = .not. done_analysis
                          if ( got_error ) err_msg = 'ABORTED analysis'
                      else
                          got_error = .TRUE.
                          err_msg = 'Data not available for analysis'
                          end if
c
c       WRITE RESULTS
c       *************
                  else if ( item .eq. 3 ) then
                      if ( done_analysis ) then
                          result_written = write_results( out_lu,
     1                        free_stream, kinem_visc, frict_vel,
     2                        wake_strength, delta_plus, y_data, U_data,
     3                        rms_data, n_data, err_msg )
                          got_error = .not. result_written
                      else
                          got_error = .TRUE.
                          err_msg= 'Cannot write: results not available'
                          end if
c
c       DISPLAY RESULTS
c       ***************
                  else if ( item .eq. 4 ) then
                      if ( done_analysis ) then
                          ok = write_results( STD_OUT, free_stream,
     1                             kinem_visc, frict_vel, wake_strength,
     2                             delta_plus, y_data, U_data, rms_data,
     3                             n_data, err_msg )
                          got_error = .not. result_written
                      else
                          got_error = .TRUE.
                          err_msg= 'Cannot write: results not available'
                          end if
c
c       REWIND DATA FILE
c       ****************
                  else if ( item .eq. 5 ) then
                      rewind( UNIT=in_lu )
                      i_profile = 0
c
c       NUMBER OF DATA POINTS IN NEXT PROFILE
c       *************************************
                  else if ( item .eq. 6 ) then
                      write( STD_OUT, 4 )
4                     format( ' Number of data points in profile ? ' )
                      n_data = Jresp( STD_IN, 2, MAX_DATA, MAX_TRIES )
                      end if
                  goto 2
c                 end of loop
              end if
          end if
      end



      CHARACTER *8 FUNCTION DONE_STR( DONE )
      include 'stdio.def'
      logical done
      if ( done ) then
          done_str = 'done'
      else
          done_str = 'not done'
          end if
      return
      end


      LOGICAL FUNCTION READ_PROFILE( IN_LU, FREE_STREAM, KINEM_VISC,
     1                       Y_DATA, U_DATA, RMS_DATA, N_DATA, ERR_MSG )
c
c Read FREE_STREAM, KINEM_VISC and data for one velocity profile from
c input file.
c
      include 'stdio.def'
      integer i_line, in_lu, i, n_data, read_line
      real    free_stream, kinem_visc,
     1        y_data(*), u_data(*), rms_data(*)
      character err_msg*(*), buffer*80
c
      read_profile = .FALSE.
      i_line = read_line( in_lu, buffer )
      if ( i_line .lt. 0 ) then
          write( err_msg, 1 ) - i_line
1         format( ' Unexpected EOF in line ', i3, ' of file' )
          return
          end if
      read( buffer, *, ERR=3 ) free_stream, kinem_visc
      do 2 i = 1, n_data
          i_line = read_line( in_lu,buffer )
          if ( i_line .lt. 0 ) then
              write( err_msg, 1 ) - i_line
              return
              end if
          read( buffer, *, ERR=3 ) y_data(i), u_data(i), rms_data(i)
          write( STD_OUT, 5 ) i, y_data(i), u_data(i), rms_data(i)
5         format( 1x, i2, 1x, f8.4, f7.4, 1x, f7.4 )
2         continue
      write( STD_OUT, 6 )
6     format( ' Press <RETURN> to continue' )
      read( STD_IN, 7 )
7     format()
      err_msg = ' '
      read_profile = .TRUE.
      return
3     write( err_msg, 4 ) i_line
4     format( ' Cannot decode line ', i3, ' of file' )
      return
      end


      INTEGER FUNCTION READ_LINE( IN_LU, BUFFER )
c  Read a line from the input file until a non-comment line is found
c  Ignores blank lines.
      include 'stdio.def'
      integer i, in_lu, i_line
      character buffer*(*)
      save i_line
      data i_line / 0 /
c
1     i_line = i_line + 1
      read( in_lu, 2, ERR=3, END=4 ) buffer
2     format( a )
      do 3 i = 1, len( buffer )
          if ( buffer(i:i) .ne. ' ' ) then
              if ( buffer(i:i) .eq. ';' ) goto 1
              read_line = i_line
              return
              end if
3         continue
      goto 1
4     read_line = - i_line
      return
      end


      LOGICAL FUNCTION WRITE_RESULTS( OUT_LU, FREE_STREAM,
     1                 KINEM_VISC, FRICT_VEL, WAKE_STRENGTH, DELTA_PLUS,
     2                 Y_DATA, U_DATA, RMS_DATA, N_DATA, ERR_MSG )
      include 'stdio.def'
      real KARMEN
      parameter( KARMEN=0.41 )
      integer out_lu, i, n_data
      real    free_stream, kinem_visc, frict_vel, wake_strength,
     1        delta_plus, y_data(*), u_data(*), rms_data(*)
      character err_msg*(*)
c
      write( out_lu, 1, ERR=6 ) free_stream, kinem_visc, frict_vel,
     1    wake_strength, 2 * wake_strength / KARMEN, delta_plus
1     format(
     1    ' ; Coefficients for model of boundary-layer mean velocity' /
     2    ' ;  free stream speed   = ', f8.5 /
     3    ' ;  kinematic viscosity = ', g12.5 /
     4    ' ;  friction velocity   = ', f8.6 /
     5    ' ;  wake-function PI    = ', f8.5 /
     6    ' ;        wake strength = ', f8.5 /
     7    ' ;  wake-function delta = ', f8.3 /
     8    ' ;' /
     9    ' ; Nondimensional velocity distribution' /
     a    ' ; y^+      U^+    u''^+' )
      do 5 i = 1, n_data
          write( out_lu, 3, ERR=6 )
     1        y_data(i) * frict_vel / kinem_visc / 1000.0,
     2        u_data(i) / frict_vel, rms_data(i) / frict_vel
3         format( 1x, f9.3, f8.4, f8.5 )
          if ( out_lu .eq. STD_OUT ) then
              if ( mod( i, 20 ) .eq. 10 .and. i .ne. n_data ) then
                  write ( STD_OUT, 4 )
4                 format( ' Press <RETURN> to continue' )
                  read( STD_IN, 7 )
7                 format()
                  end if
              end if
5         continue
      if ( out_lu .eq. STD_OUT ) then
          write ( STD_OUT, 4 )
          read( STD_IN, 7 )
          end if
      write_results = .TRUE.
      err_msg = ' '
      return
6     write_results = .FALSE.
      err_msg = 'Cannot write results to file'
      return
      end

