c
c    Copyright (C) 1997  Peter Vernon Lanspeary
c
c    This program is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 2 of the License, or
c    (at your option) any later version.
c
c    This program is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program; if not, write to the Free Software
c    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
c


c           *****************************************************
c           ***                                               ***
c           ***    P.V.Lanspeary private FORTRAN 77 library   ***
c           ***                                               ***
c           ***        f77_sig : Signal processing            ***
c           ***                                               ***
c           *****************************************************
c  
      LOGICAL FUNCTION EFIR_READ( FILTER_FILE, MAX_LENGTH,
     1                            FILTER_COEFFS, FILTER_LEN )
c
c    Reads the specification of a moving average filter from the file
c FILTER_FILE. The filter specification consists of
c   a) FILTER_COEFFS  -- moving average filter coefficients
c                        listed in order of increasing delay
c   b) FILTER_LEN     -- number of moving average filter coefficients
c                        ( which can be no more than MAX_LENGTH )
c The specification is read from the file into a buffer, and is then copied
c into the output arguments FILTER_COEFFS and FILTER_LEN. If FILTER_FILE='?',
c the file name is obtained from the user. If the same file is given on two
c successive calls or FILTER_FILE=' ', the specification is copied from the
c buffer and the file is not read again.
c
c The FILTER_FILE is a sequential access ASCII file. The first non-blank line
c begins with FILTER_LEN. The values FILTER_COEFFS begin on the next non-blank
c line. There is no provision for comments in the file.
c
c Arguments
c  character        filter_file*(*)   # INPUT  # name of specification file
c  integer          max_length        # INPUT  # maximum number of filter coeffs
c  double precision filter_coeffs(*)  # OUTPUT # filter coefficients
c  integer          filter_len        # OUTPUT # number of filter coefficients
c
c Constants
c  HARD_MAX_LEN = 256
c  MAX_NAME_LEN = 64
c  ID_ABORT     = ' EFIR_READ : ABORT : '
c
c Local variables
c  integer
c   i             # array index
c   name_len      # length of file name
c   reserv_lu     # reserved logical unit number allocated from new_lun
c   spec_len      # Spec buffer : number of filter coeffs
c   spec_lu       # logical unit no. of spec file
c   new_lun       # function : get new logical unit number
c   open_file     # function : opens file
c   str_length    # function : length of string
c  logical
c   spec_got      # Spec buffer : filter spec is available
c  double precision
c   spec_coeffs(HARD_MAX_LEN) # Spec buffer : filter coefficients
c  character
c   spec_file*MAX_NAME_LEN    # current name of specification file
c   prev_file*MAX_NAME_LEN    # previous name of specification file
c  subroutine
c   Cresp         # read character string
c
      include 'stdio.def'
      integer    HARD_MAX_LEN,     MAX_NAME_LEN
      character  ID_ABORT*21
      parameter( HARD_MAX_LEN=256, MAX_NAME_LEN=64,
     1           ID_ABORT=' EFIR_READ : ABORT : ' )
c
      integer max_length, filter_len, i, name_len, reserv_lu, spec_len,
     1        spec_lu, new_lun, open_file, str_length
      logical spec_got
      double precision filter_coeffs(*), spec_coeffs(HARD_MAX_LEN)
      character        filter_file*(*), spec_file*(MAX_NAME_LEN),
     1                 prev_file*(MAX_NAME_LEN)
c
      save reserv_lu, spec_len, spec_got, spec_coeffs, prev_file
      data reserv_lu, spec_len / 0, 0 /
      data spec_got / .FALSE. /
      data spec_coeffs / HARD_MAX_LEN * 0.0d0 /
      data prev_file / '?' /
c
      Efir_read = .FALSE.
      if ( reserv_lu .eq. 0 ) then                                        1b
          reserv_lu = new_lun()
          end if                                                          1e
c
c       get file name from user if given a '?'
c
      if ( filter_file .eq. '?' ) then                                    1b
          write( STD_OUT, 1 )
1         format( ' Name of MA (FIR) filter specification file ? ', $ )
          call Cresp( STD_IN, spec_file, 4 )
      else                                                                1m
          spec_file = filter_file
          end if                                                          1e
      if ( spec_file .ne. prev_file .or. prev_file .eq. '?' ) then        1b
c
c       open filter file
c
          spec_lu = open_file( reserv_lu, spec_file, 'old', 0 )
          name_len  = max( 1, str_length( spec_file ) )
          if ( spec_lu .le. 0 ) then                                      2b
              write( STD_ERR, 2 ) id_abort, spec_file(:name_len)
2             format( a, ' Cannot open filter file ', a )
          else                                                            2m
c
c       read filter new spec from file into the buffer
c
              spec_got = .FALSE.
              read( spec_lu, *, ERR=7 ) spec_len
              if ( spec_len .lt. 0 .or.
     1             spec_len .gt. HARD_MAX_LEN ) then                      3b
                  write( STD_ERR, 3 )
     1                id_abort, spec_len, HARD_MAX_LEN
3                 format( a, 'Illegal filter length ', i12,
     1                    ' Maximum is ', i4 )
                  close( UNIT=spec_lu )
              else                                                        3m
                  read( spec_lu, *, ERR=7 )
     1                    ( spec_coeffs(i), i = 1, spec_len )
c
c       the set of new filter coefficients is available
c
                  close( UNIT=spec_lu )
                  prev_file = spec_file
                  spec_got  = .TRUE.
                  if ( spec_len .gt. max_length ) then                    4b
                      write( STD_ERR, 3 )
     1                    id_abort, spec_len, max_length
                  else                                                    4m
                      do 4 i = 1, spec_len                                5b
                          filter_coeffs(i) = spec_coeffs(i)
4                         continue                                        5e
                      filter_len = spec_len
                      efir_read = .TRUE.
                      end if                                              4e
                  end if                                                  3e
              end if                                                      2e
c
c           using previously obtained filter coefficients
c
      else if ( spec_got ) then                                           1m
          if ( spec_len .gt. max_length ) then                            2b
              write( STD_ERR, 3 ) id_abort, spec_len, max_length
          else                                                            2m
              do 5 i = 1, spec_len                                        3b
                  filter_coeffs(i) = spec_coeffs(i)
5                 continue                                                3e
              filter_len = spec_len
              efir_read = .TRUE.
              end if                                                      2e
      else                                                                1m
          write( STD_ERR, 6 ) ID_ABORT
6         format( a, 'Filter spec unavailable' )
          end if                                                          1e
      return
c
7     write( STD_ERR, 8 )  ID_ABORT, filter_file(:name_len)
8     format( a, 'Cannot read filter file ', a )
      return
      end




      INTEGER FUNCTION IFIR_FILT( IN_DATA, OUT_DATA, INCREM, N_POINTS,
     1                            FILTER_COEFFS, FILTER_LEN, METHOD )
c
c     Function -- Finite Impulse Response (moving average) FILTer
c                 of INTEGER *2 data
c Filter the input signal IN_DATA using the FIR filter defined by
c FILTER_COEFFS and FILTER_LEN. The result is placed in OUT_DATA.
c The returned value of IFIR_FILT is the number of valid data samples in
c OUT_DATA. The returned value is IFIR_FILT=0 if an error occurs.
c
c Arguments
c  integer *2 in_data(*) # array of input data
c             out_data(*)# array of output data
c  integer    increm     # data indexing increment
c             n_points   # number of data points used
c  double precision
c             filter_coeffs  # filter coeffients, ie filter impulse response
c  integer    filter_len # number of filter coefficients
c             method     # Filtering method :
c                     # 0, 4, 8  = Assume Non-linear phase filter
c                     # 1, 5, 9  = Assume Linear phase filter
c                     # 2, 6, 10 = Measure phase linearity
c                     # 0, 1, 2  = Assume data not supplied is zero; i.e.
c                     #            IN_DATA(i)=0 for i<=0. Output signal
c                     #            contains starting transient.
c                     # 4, 5, 6  = Assume data not supplied is IN_DATA(1); i.e.
c                     #            IN_DATA(i)=IN_DATA(1) for i<=0. Output
c                     #            signal contains starting transient.
c                     # 8, 9, 10 = Starting transient is ignored, output data
c                     #            begins with steady response, the number of
c                     #            data samples is reduced to
c                     #            N_POINTS - FILTER_LEN + 1
c
c
c If the filter has a linear phase response, a saving in calculation time can
c be made because the filter impulse response has mirror image symmetry about
c its central value . Ie. h(i)=h(m+1-i). All calculation is
c double precision.
c
c
c constant
c   TINY_COEFF=1.0e-12
c Variables
c  integer
c   first_steady      # data array index of first steady state
c   half_transient    # data index increment for half transient length
c   i                 # leading edge index for data array
c   ii
c   is_linear         # phase of the filter is linear NO, YES or QUERY
c   i_mirror          # trailing edge index for data array
c   j                 # summation index increment for data array
c   jj                # summation index increment for filter coefficients
c   last_data         # index of last data array element
c   middle_data       # index increment of mid-data element in convolution
c  logical
c   odd_length        # number of filter coefficients is odd
c   linear_phase      # filter is linear phase
c  double_precision
c   assumed_data      # assumed value of data prior to signal
c   filter_sum        # sum of filter coefficients
c   middle_coeff      # middle coefficient of odd-length filter
c   sum               # convolution sum
c
c
c
      include 'stdio.def'
      double precision DP_MIN_INT2,        DP_MAX_INT2,
     1                 TINY_COEFF
      parameter(       DP_MIN_INT2=-32768, DP_MAX_INT2=32767,
     1                 TINY_COEFF=1.0e-12 )
c
      integer increm, n_points, filter_len, method, first_steady,
     1        half_transient, i, ii, is_linear, i_mirror, j, jj,
     2        last_data, middle_data
      logical odd_length, linear_phase
      integer *2 in_data(*), out_data(*)
      double precision assumed_data, filter_sum, filter_coeffs(*),
     1        middle_coeff, sum
c
c
      IFIR_filt = 0
      if ( increm .le. 0 ) then                                           1b
          write( STD_ERR, 1 ) 'INCREM', increm
1         format( ' IFIR_FILT : ABORT : Illegal ', a, ' = ', i12 )
      else if ( n_points .le. 1 ) then                                    1m
          write( STD_ERR, 1 ) 'N_POINTS', n_points
      else if ( filter_len .le. 0 ) then                                  1m
          write( STD_ERR, 1 ) 'FILTER_LEN', filter_len
      else if ( method .lt. 0 .or. method .eq. 3 .or.
     1          method .eq. 7 .or. method .gt. 10 ) then                  1m
          write( STD_ERR, 1 ) 'METHOD', method
      else                                                                1m
c
c       determine if linear phase filter
c
          is_linear = mod( method, 4 )
          if ( is_linear .eq. QUERY ) then                                2b
              linear_phase = .TRUE.
              do 2 j = 1, filter_len / 2                                  3b
                  linear_phase = linear_phase .and. TINY_COEFF .gt.
     1                abs( filter_coeffs(j) /
     2                     filter_coeffs(filter_len+1-j) - 1.0 )
2                 continue                                                3e
          else if ( is_linear .eq. YES ) then                             2m
              linear_phase = .TRUE.
          else if ( is_linear .eq. NO ) then                              2m
              linear_phase = .FALSE.
              end if                                                      2e
          last_data = ( n_points - 1 ) * increm + 1
          first_steady = ( filter_len - 1 ) * increm + 1
c
c  ********* steady state response linear phase filter
c
          if ( linear_phase ) then                                        2b
              odd_length = mod( filter_len, 2 ) .gt. 0
              if ( odd_length ) then                                      3b
                  middle_coeff = filter_coeffs( filter_len / 2 + 1 )
                  end if                                                  3e
              half_transient = ( filter_len / 2 - 1 ) * increm
              middle_data = filter_len / 2 * increm
              do 4 i = last_data, first_steady, - increm                  3b
                  if ( odd_length ) then                                  4b
                      sum = middle_coeff *
     1                    dble( in_data( i - middle_data ) )
                  else                                                    4m
                      sum = 0.0d0
                      end if                                              4e
                  jj = 1
                  i_mirror = i - first_steady - 1
                  do  3 j = 0, half_transient, increm                     4b
                      sum = sum + filter_coeffs(jj) *
     1                    ( dble( in_data(i-j) ) +
     2                      dble( in_data(i_mirror+j) ) )
                      jj = jj + 1
3                     continue                                            4e
                  out_data(i) = max( DP_MIN_INT2,
     1                min( sign( abs( sum ) + 0.5d0, sum ),
     2                     DP_MAX_INT2 ) )
4                 continue                                                3e
          else                                                            2m
c
c    ********* steady state response non-linear phase filter
c
              do 6 i = last_data, first_steady, -increm                   3b
                  sum = filter_coeffs(1) * dble( in_data(i) )
                  jj = 2
                  do 5 j = increm, first_steady - 1, increm               4b
                      sum = sum +
     1                      filter_coeffs(jj) * dble( in_data(i-j) )
                      jj = jj + 1
5                     continue                                            4e
                  out_data(i) = max( DP_MIN_INT2,
     1                min( sign( abs( sum ) + 0.5d0, sum ),
     2                     DP_MAX_INT2 ) )
6                 continue                                                3e
              end if                                                      2e
          IFIR_filt = n_points
          if ( filter_len .gt. 1 ) then                                   2b
c
c    ******** transient response for linear and non-linear phase filter
c             Assume unknown data is zero or IN_DATA(1)
c
              if ( 0 .le. method .and. method .le. 2 .or.
     1             4 .le. method .and. method .le. 6 ) then               3b
                  if ( method / 4 .eq. 0 ) then                           4b
                      assumed_data = 0.0d0
                  else                                                    4m
                      assumed_data = dble( in_data(1) )
                      end if                                              4e
                  ii = filter_len
                  filter_sum = 0.0
                  do 8 i = first_steady - increm, 1, -increm              4b
                      sum = filter_coeffs(1) * dble( in_data(i) )
                      jj = 2
                      do 7 j = increm, i - 1, increm                      5b
                          sum = sum +
     1                          filter_coeffs(jj) * dble( in_data(i-j) )
                          jj = jj + 1
7                         continue                                        5e
                      filter_sum = filter_sum + filter_coeffs(ii)
                      ii = ii - 1
                      sum = sum + filter_sum * assumed_data
                      out_data(i) = max( DP_MIN_INT2,
     1                    min( sign( abs( sum ) + 0.5d0, sum ),
     2                         DP_MAX_INT2 ) )
8                     continue                                            4e
c
c    ******** copy steady state response to start of output array
c             Do it at the end because if this stage is incorporated
c             into the data processing, and OUT_DATA is the same
c             array as IN_DATA, the data is corrupted.
c
              else if ( 8 .le. method .and. method .le. 10 ) then         3m
                  j = 1
                  do 9 i = first_steady, last_data, increm                3b
                      out_data( j ) = out_data( i )
                      j = j + increm
9                     continue
                  IFIR_filt = n_points - filter_len + 1
                  end if                                                  3e
              end if                                                      2e
          end if                                                          1e
c
      return
      end



      INTEGER FUNCTION JFIR_FILT( IN_DATA, OUT_DATA, INCREM, N_POINTS,
     1                            FILTER_COEFFS, FILTER_LEN, METHOD )
c
c     Function -- Finite Impulse Response (moving average) FILTer
c Filter the input signal IN_DATA using the FIR filter defined by
c FILTER_COEFFS and FILTER_LEN. The result is placed in OUT_DATA.
c The returned value of JFIR_FILT is the number of valid data samples in
c OUT_DATA. The returned value is JFIR_FILT=0 if an error occurs.
c
c Arguments
c  real    in_data(*) # array of input data
c          out_data(*)# array of output data
c  integer increm     # data indexing increment
c          n_points   # number of data points used
c  double precision
c          filter_coeffs  # filter coeffients, ie filter impulse response
c  integer filter_len # number of filter coefficients
c          method     # Filtering method :
c                     # 0, 4, 8  = Assume Non-linear phase filter
c                     # 1, 5, 9  = Assume Linear phase filter
c                     # 2, 6, 10 = Measure phase linearity
c                     # 0, 1, 2  = Assume data not supplied is zero; i.e.
c                     #            IN_DATA(i)=0 for i<=0. Output signal
c                     #            contains starting transient.
c                     # 4, 5, 6  = Assume data not supplied is IN_DATA(1); i.e.
c                     #            IN_DATA(i)=IN_DATA(1) for i<=0. Output
c                     #            signal contains starting transient.
c                     # 8, 9, 10 = Starting transient is ignored, output data
c                     #            begins with steady response, the number of
c                     #            data samples is reduced to
c                     #            N_POINTS - FILTER_LEN + 1
c
c
c If the filter has a linear phase response, a saving in calculation time can
c be made because the filter impulse response has mirror image symmetry about
c its central value . Ie. h(i)=h(m+1-i). All floating point calculation is
c double precision.
c
c
c constant
c   TINY_COEFF=1.0e-12
c Variables
c  integer
c   first_steady      # data array index of first steady state
c   half_transient    # data index increment for half transient length
c   i                 # leading edge index for data array
c   ii
c   is_linear         # phase of the filter is linear NO, YES or QUERY
c   i_mirror          # trailing edge index for data array
c   j                 # summation index increment for data array
c   jj                # summation index increment for filter coefficients
c   last_data         # index of last data array element
c   middle_data       # index increment of mid-data element in convolution
c  logical
c   odd_length        # number of filter coefficients is odd
c   linear_phase      # filter is linear phase
c  double_precision
c   assumed_data      # assumed value of data prior to signal
c   filter_sum        # sum of filter coefficients
c   middle_coeff      # middle coefficient of odd-length filter
c   sum               # convolution sum
c
c
c
      include 'stdio.def'
      double precision TINY_COEFF
      parameter(       TINY_COEFF=1.0e-12 )
c
      integer increm, n_points, filter_len, method, first_steady,
     1        half_transient, i, ii, is_linear, i_mirror, j, jj,
     2        last_data, middle_data
      logical odd_length, linear_phase
      real    in_data(*), out_data(*)
      double precision assumed_data, filter_sum, filter_coeffs(*),
     1        middle_coeff, sum
c
c
      JFIR_filt = 0
      if ( increm .le. 0 ) then                                           1b
          write( STD_ERR, 1 ) 'INCREM', increm
1         format( ' SFIR_FILT : ABORT : Illegal ', a, ' = ', i12 )
      else if ( n_points .le. 1 ) then                                    1m
          write( STD_ERR, 1 ) 'N_POINTS', n_points
      else if ( filter_len .le. 0 ) then                                  1m
          write( STD_ERR, 1 ) 'FILTER_LEN', filter_len
      else if ( method .lt. 0 .or. method .eq. 3 .or.
     1          method .eq. 7 .or. method .gt. 10 ) then                  1m
          write( STD_ERR, 1 ) 'METHOD', method
      else                                                                1m
c
c       determine if linear phase filter
c
          is_linear = mod( method, 4 )
          if ( is_linear .eq. QUERY ) then                                2b
              linear_phase = .TRUE.
              do 2 j = 1, filter_len / 2                                  3b
                  linear_phase = linear_phase .and. TINY_COEFF .gt.
     1                abs( filter_coeffs(j) /
     2                     filter_coeffs(filter_len+1-j) - 1.0 )
2                 continue                                                3e
          else if ( is_linear .eq. YES ) then                             2m
              linear_phase = .TRUE.
          else if ( is_linear .eq. NO ) then                              2m
              linear_phase = .FALSE.
              end if                                                      2e
          last_data = ( n_points - 1 ) * increm + 1
          first_steady = ( filter_len - 1 ) * increm + 1
c
c  ********* steady state response linear phase filter
c
          if ( linear_phase ) then                                        2b
              odd_length = mod( filter_len, 2 ) .gt. 0
              if ( odd_length ) then                                      3b
                  middle_coeff = filter_coeffs( filter_len / 2 + 1 )
                  end if                                                  3e
              half_transient = ( filter_len / 2 - 1 ) * increm
              middle_data = filter_len / 2 * increm
              do 4 i = last_data, first_steady, - increm                  3b
                  if ( odd_length ) then                                  4b
                      sum = middle_coeff * in_data( i - middle_data  )
                  else                                                    4m
                      sum = 0.0d0
                      end if                                              4e
                  jj = 1
                  i_mirror = i - first_steady - 1
                  do  3 j = 0, half_transient, increm                     4b
                      sum = sum + filter_coeffs(jj) *
     1                    ( in_data(i-j) + in_data(i_mirror+j) )
                      jj = jj + 1
3                     continue                                            4e
                  out_data(i) = sum
4                 continue                                                3e
          else                                                            2m
c
c    ********* steady state response non-linear phase filter
c
              do 6 i = last_data, first_steady, -increm                   3b
                  sum = filter_coeffs(1) * dble( in_data(i) )
                  jj = 2
                  do 5 j = increm, first_steady - 1, increm               4b
                      sum = sum +
     1                      filter_coeffs(jj) * dble( in_data(i-j) )
                      jj = jj + 1
5                     continue                                            4e
                  out_data(i) = sum
6                 continue                                                3e
              end if                                                      2e
          JFIR_filt = n_points
          if ( filter_len .gt. 1 ) then                                   2b
c
c    ******** transient response for linear and non-linear phase filter
c             Assume unknown data is zero or IN_DATA(1)
c
              if ( 0 .le. method .and. method .le. 2 .or.
     1             4 .le. method .and. method .le. 6 ) then               3b
                  if ( method / 4 .eq. 0 ) then                           4b
                      assumed_data = 0.0d0
                  else                                                    4m
                      assumed_data = in_data(1)
                      end if                                              4e
                  ii = filter_len
                  filter_sum = 0.0
                  do 8 i = first_steady - increm, 1, -increm              4b
                      sum = filter_coeffs(1) * dble( in_data(i) )
                      jj = 2
                      do 7 j = increm, i - 1, increm                      5b
                          sum = sum +
     1                          filter_coeffs(jj) * dble( in_data(i-j) )
                          jj = jj + 1
7                         continue                                        5e
                      filter_sum = filter_sum + filter_coeffs(ii)
                      ii = ii - 1
                      out_data(i) = sum + filter_sum * assumed_data
8                     continue                                            4e
c
c    ******** copy steady state response to start of output array
c             Do it at the end because if this stage is incorporated
c             into the data processing, and OUT_DATA is the same
c             array as IN_DATA, the data is corrupted.
c
              else if ( 8 .le. method .and. method .le. 10 ) then         3m
                  j = 1
                  do 9 i = first_steady, last_data, increm                3b
                      out_data( j ) = out_data( i )
                      j = j + increm
9                     continue
                  JFIR_filt = n_points - filter_len + 1
                  end if                                                  3e
              end if                                                      2e
          end if                                                          1e
c
      return
      end



      REAL FUNCTION QARPSD( FREQ )
c
c         evaluate AutoRegressive Power Spectral Density at a single point
c
c Argument
c   real freq          # frequency for which PSD is calculated (Hz)     input
c
c variables
c   integer *4 j       # loop index
c   real c             # non-dimensional radian frequency
c
      implicit none
      real pi
      parameter( pi=3.1415926 )
c
      integer *4 j
      real freq, c
c
c          COMMON BLOCK DECLARATION
c
c real  autozz                 # zero time offset value of AUTO
c       auto(100)              # impulse response autocorrelation coefficients
c       rdelt                  # time increment for time series data
c       rxvar                  # variance of the time series data
c integer *4 im                # number of AR coefficients
c
      real autozz, auto(100), rdelt, rxvar
      integer *4 im
      common /Af77lib5/ autozz, auto, rdelt, rxvar, im
c
c         power spectral density
c
      c = 2.0 * pi * rdelt * freq
      qarpsd = autozz / 2.0
      do j = 1, im
        qarpsd = qarpsd + auto(j) * cos( j * c )
        end do
      qarpsd = abs( rxvar * rdelt / qarpsd / 2.0 )
      return
      end



      SUBROUTINE SACORR( X, NINCR, NPTS, ACORR, INCORR, NLAGS )
c
c       Calculates NLAGS autocorrelations from NPTS values taken at intervals
c of NINCR from the data array X . The indexing increment for the auto-
c correlation array ACORR is INCORR . ACORR(1) is the autocorrelation for the
c first non-zero lag . The zero lag autocorrelation is unity .
c
c         ARGUMENTS
c    type
c  real *4    x(nincr*(npts-1)+1)      # data array                     input
c integer *4  nincr                    # data indexing increment        input
c             npts                     # number of data points          input
c  real *4    acorr(incorr*(nlags-1)+1)# autocorrelation values         output
c integer *4  incorr                   # ACORR indexing increment       input
c             nlags                    # number of autocorrelation lags input
c variables
c  double precision rvar               # aggregate for variance
c                   rmean              # aggregate for mean
c  real xmean                          # mean of the data
c       xvar                           # signal variance
c  integer *4 nlast                    # index of last data value
c             i                        # data index
c             j                        # autocorrelation index
c             k                        # lag index
c
      implicit none
      integer *4  nincr, incorr, nlags, npts, npts1, nlast, i, j, k
      real *4 x(*), xmean, acorr(*), xvar
      double precision rvar, rmean
c
c           check arguments
c
      if ( nincr .le. 0 .or. npts .le. 0 .or. incorr .le. 0 .or.
     1  nlags .le. 0 ) return
      nlast = nincr * ( npts - 2 ) + 1
      k = nincr
c
c           variance
c
      rvar = 0.0d0
      rmean = 0.0d0
      do i = 1, nlast + nincr, nincr
        rmean = rmean + x(i)
        end do
      xmean = rmean / npts
      do i = 1, nlast + nincr, nincr
        rvar = rvar + ( x(i) - xmean ) **2
        end do
      xvar = sngl( rvar )
c
c           autocorrelations
c
      do j = 1, incorr * ( nlags - 1 ) + 1, incorr
        rvar  = 0.0
        do i = 1, nlast - k, nincr
          rvar = rvar + dble( ( x(i) - xmean ) * ( x(i+k) - xmean ))
          end do
        k = k + nincr
        acorr(j) = sngl( rvar ) / xvar
        end do
      return
      end


      SUBROUTINE SARINZ( ARCOEF, M, DELT, XVAR )
c
c   Initialize for single point AutoRegressive Power Spectral Density values
c       Calculate the impulse response autocorrelation coefficients so that a
c single single argument function can calculate point on the PSD curve .
c
c Arguments
c  TYPE          NAME             DESCRIPTION                            I/O
c  real         arcoef(m)      # Autoregression coefficients            input
c  integer *4   m              # number of autoregression coefficients  input
c  real         Delt           # time increment for time series data    input
c               xvar           # variance of the time series data       input
c
c variables
c  integer *4 i, j             # looping indices
c
      implicit none
      integer *4 m, i, j
      real arcoef(m), Delt, xvar
c
c          COMMON BLOCK DECLARATION
c
c real  autozz                 # zero time offset value of AUTO
c       auto(100)              # impulse response autocorrelation coefficients
c       rdelt                  # time increment for time series data
c       rxvar                  # variance of the time series data
c integer *4 im                # number of AR coefficients
      real autozz, auto(100), rdelt, rxvar
      integer *4 im
      common /Af77lib5/ autozz, auto, rdelt, rxvar, im
c
c         error check arguments
c
      if ( m .le. 0 .or. delt .le. 0.0 .or. xvar .le. 0.0 ) return
c
c         calculate autocorrelation coefficients for filter's impulse response
c
      rdelt = delt
      rxvar = xvar
      im = m
      autozz = 1.0
      do i = 1, im
        auto(i) = - arcoef(i)
        autozz = autozz + arcoef(i) **2
        do j = 1, im - i
          auto(i) = auto(i) + arcoef(j) * arcoef(j+i)
          end do
        end do
      return
      end



      SUBROUTINE SARPSD( ARCOEF, M, DELT, XVAR, PSD, FINCR, NFREQ,
     1  AUTO, IERR )
c
c               AutoRegressive Power Spectral Density
c       Calculate some value of the power spectral density, given
c M estimates ARCOEF of the autoregression coefficients . The frequency
c increment and lowest frequency value are FINCR . NFREQ frequency values
c are calculated .
c
c Arguments
c  TYPE          NAME             DESCRIPTION                            I/O
c  real         arcoef(m)      # Autoregression coefficients            input
c  integer *4   m              # number of autoregression coefficients  input
c  real         Delt           # time increment for time series data    input
c               xvar           # variance of the time series data       input
c               psd(nfreq)     # power spectral density array           output
c               fincr          # required frequency increment in PSD    input
c  integer *4   nfreq          # number of points calculated for PSD    input
c  real         auto(m)        # impulse response autocorrelation       output
c                              coefficients (doesn't include zero time value)
c  integer *4   ierr           # error flag  - see below                output
c
c constant
c  pi
c variables
c  real c                      # non dimensional radian frequency increment
c       autozz                 # zero time offset value of AUTO
c  integer *4 i, j             # looping indices
c
c    ERROR FLAGS
c       value              problem                       action
c     ierr = 0      none                               evaluate PSD
c     ierr = 129   m ,Delt, xvar, fincr or nfreq <= 0  PSD not evaluated
c     ierr = 130  negative value generated in PSD     take absolute value of PSD
c
      implicit none
      real pi
      parameter( pi=3.1415926 )
c
      integer *4 m, nfreq, i, j, ierr
      real arcoef(m), Delt, xvar, psd(nfreq), fincr, auto(m), c, autozz
c
c         error check arguments
c
      ierr = 0
      if ( m .le. 0 .or. nfreq .le. 0 .or. delt .le. 0.0 .or. xvar .le.
     1    0.0 .or. fincr .le. 0.0 ) then
        ierr = 129
        return
        endif
c
c         calculate autocorrelation coefficients for filter's impulse response
c
      autozz = 1.0
      do i = 1, m
        auto(i) = - arcoef(i)
        autozz = autozz + arcoef(i) **2
        do j = 1, m - i
          auto(i) = auto(i) + arcoef(j) * arcoef(j+i)
          end do
        end do
c
c         power spectral density
c
      c = 2.0 * pi * delt * fincr
      do i = 1, nfreq
        psd(i) = autozz / 2.0
        do j = 1, m
          psd(i) = psd(i) + auto(j) * cos( i * j * c )
          end do
        if ( psd(i) .le. 0.0 ) ierr = 130
        psd(i) = abs( xvar * delt / psd(i) /2.0 )
        end do
      return
      end


      SUBROUTINE SAUTO( X, NINCR, NPTS, AUTO, INC, NLAGS,
     1                  IERR, IWK, ZWK )
c
c       Calculates NLAGS autocorrelations from NPTS values taken at intervals
c of NINCR from the data array X . The indexing increment for the auto-
c correlation array AUTO is INC . AUTO(1) is the autocorrelation for the
c zero lag .
c
c         ARGUMENTS
c    type
c  real *4    x(nincr*(npts-1)+1)      # data array                     input
c integer *4  nincr                    # data indexing increment        input
c             npts                     # number of data points          input
c             auto(inc*(nlags-1)+1)    # autocorrelation values         output
c integer *4  inc                      # ACORR indexing increment       input
c             nlags                    # number of autocorrelation lags input
c             ierr                     # error flag = 0  -no errors-    output
c             iwk(*)                   # work array for FFT             ----
c complex     zwk(*)                   # work array for FFT             ----
c
c  ERROR FLAG
c ierr = 0    no errors
c      = 129  size or increment argument <= 0
c      = 130  too many lags for number of data points or too few lags
c      = 131  not enough data for 2**m point FFT
c
c minimum sizes of array arguments
c     define   len = 2 **m >= 2 * max( minlags, maxlags )
c   x     -   nincr * ( npts - 1 ) + 1
c   auto  -   ( nlags -1 ) * nincr + 1
c   iwk   -   len+1
c   zwk   -   len * 3 +1
c
c variables
c  integer *4 i                # looping index
c             n1               # number of data samples used
c  real xvar                   # normalizing factor
c       qvar                   # function : variance
c
      implicit none
      integer *2 Hvdulu
      parameter( Hvdulu=6 )
c
      real x(*), auto(*), xvar, qvar
      integer *4  nincr, npts, inc, nlags, i, ierr, iwk(*), n1
      complex zwk(*)
c
c         calculate negative lag correlations
c
      call sfconv( x, x, nincr, npts, auto, inc, nlags, n1,
     1             ierr, iwk, zwk )
c
c         normalize the result
c
      xvar = qvar( x, nincr, n1 ) * n1
      if ( xvar .gt. 0.0 ) then
        do i = ( nlags - 1 ) * inc + 1, 1, -inc
          auto(i) = auto(i) / xvar
          end do
      else
        write( Hvdulu, 1 )
1       format( ' SAUTO: WARNING: Non-positive variance - no scaling' )
        end if
      return
      end



      SUBROUTINE SBURG( X, NINCR, NPTS, XMEAN, FORW, BACK,ARCOEF, OLDAR,
     1  M, MMAX, CRITERR )
c
c       Calculates the Autoregressive linear prediction coefficients of the
c data array X . NPTS values from X are used, starting at X(1) , and using
c every NINCRth point . The user must supply the mean of the data XMEAN, the
c maximum filter order Mmax, and the Error criterion CRITERR to be used .
c
c       METHOD
c    The method , based on Burg's analysis , is a lattice method with the
c reflection coefficients being the harmonic mean of the forward and backward
c reflection coefficients .
c
c
c Arguments
c  TYPE   NAME                   DESCRIPTION                             I/O
c  real x(nincr*(npts-1)+1)    # data array                             input
c  integer *4 nincr            # data indexing increment                input
c             npts             # number of data points used             input
c  real xmean                  # mean value of data                     input
c       forw(npts-1)           # forward lattice filter residuals       output
c       back(npts-1)           # backward lattice filter residuals      output
c       arcoef(mmax)           # highest (mth) order AR coefficients    output
c       oldar(mmax-1)          # (m-1)th order AR coefficients          output
c       mmax                   # maximum filter order                   input
c       m                      # number of coeffients in ARCOEF         output
c  character criterr*3         # specifies error criterion to be used   input
c
      implicit none
      integer Hvdulu
      parameter( Hvdulu=6 )
c
      real x(*), xmean, xvar, forw(*), back(*), arcoef(*), oldar(*),
     1  respow, oldfpe, fpe, oldaic, aic, rcovfb, rvarfb
      integer *4 nincr, npts, mmax, nlast, i, m
      character criterr*3
c
c         initial forward and backward residuals , residual power
c         and variance & covariance of the residuals for calculating 1st
c         reflection coeff.
c
      if ( npts - m .le. 1 ) then
          write( Hvdulu, 998 ) m+2
998       format( ' SBURG : ABORT : Not  enough data.'
     1            ' Need at least ', i7, ' data points.' )
          return
          end if
      nlast = nincr * ( npts - 1 ) + 1
      rcovfb = ( x(nlast) - xmean ) * ( x( nlast - nincr ) -xmean )
      forw(1) = x(1) - xmean
      xvar = forw(1) **2 / 2.0
      back(npts) = x(nlast) - xmean
      do i = 2, npts - 1
        forw(i) = x( nincr * ( i - 1 ) + 1 ) - xmean
        back( i-1 ) = forw(i)
        xvar = xvar + forw(i) **2
        rcovfb = rcovfb + forw(i-1) * forw(i)
        end do
      rvarfb = xvar - ( x(nlast) - xmean ) **2 / 2.0
c
c         first reflection coefficient
c
      if ( abs( rvarfb ) .lt. 1.0e-35 ) then
          write( Hvdulu, 999 )
999       format( ' SBURG : ABORT : Data is constant.'
     1            ' No result calculated' )
          return
          end if
      arcoef(1) = rcovfb / rvarfb
c
c         initial residual power , FPE & AIC errors
c
      respow = 1.0 - arcoef(1) **2
      fpe = respow * float( npts + 2 ) / float( npts - 2 )
      if ( respow .gt. 0.0 ) then
        aic = log( respow ) + 2.0 / float(npts)
      else
        aic = -1.0e6
        end if
c
c         main recursion loop
c
      do m = 2, mmax
c
c         reserve the old AR coefficients
c
        do i = 1, m - 1
          oldar(i) = arcoef(i)
          end do
c
c         next lattice filter stage residuals
c
        do i = 1, npts - m
          forw(i) = forw(i) - arcoef( m-1 ) * back(i)
          back(i) = back(i+1) - arcoef( m-1 ) * forw(i+1)
          end do
c
c         mth reflection coefficient
c
        rcovfb = 0.0
        rvarfb = 0.0
        do i = 1, npts - m
          rcovfb = rcovfb + forw(i) * back(i)
          rvarfb = rvarfb + forw(i) **2 + back(i) **2
          end do
        arcoef(m) = 2.0 * rcovfb / rvarfb
c
c         residual power , Akaike FPE & AIC
c
        respow = respow * ( 1.0 - arcoef(m) **2 )
        oldfpe = fpe
        fpe = respow * float( npts + m + 1 ) / float( npts - m - 1 )
        oldaic = aic
        if ( respow .gt. 0.0 ) then
          aic = log( respow ) + 2.0 * float(m) / float(npts)
        else
          aic = -1.0e6
          end if
c
c         Levinson's recursion
c
        do i = 1, m-1
          arcoef(i) = oldar(i) - arcoef(m) * oldar( m-i )
          end do
c
c         test for end of recursion
c
        if ( ( criterr .eq. 'fpe' .or. criterr .eq. 'FPE' ) .and.
     1       fpe .ge. oldfpe .or. ( criterr .eq. 'aic' .or.
     2       criterr .eq. 'AIC' ) .and. aic .ge. oldaic )
     3    return
        end do
c end of loop
      m = m - 1
      return
      end



      SUBROUTINE SCOHER( PSDX, PSDY, INC1, XPSD, COHER, INC2, M )
c
c         COHERence function squared : from cross psd etc
c    Calculates the square of the coherence function of a system with input
c x(t) and output y(t) . Each INC1th point in PSDX and PSDY are the power
c spectral density of x(t)and y(t) respectively, and each INC2th point in
c XPSD is the cross power spectral density of x(t) & y(t) . The transfer
c function is put into each INC2th point in XFER . 2**M data points are taken
c from each array .
c
c    ARGUMENTS
c real       psdx(n1)          # power spectral density of x(t)         input
c            psdy(n1)          # power spectral density of y(t)         input
c integer *4 inc1              # indexing increment for PSDX            input
c complex    xpsd(n2)          # cross power spectral density of x & y  input
c real       coher(n2)         # coherence squared function of system   output
c integer *4 inc2              # indexing increment for XPSD & XFER     input
c            m                 # 2**m = n0. of points in each spectrum  input
c              where n1 = inc1 * ( 2 **m - 1 ) + 1
c                    n2 = inc2 * ( 2 **m - 1 ) + 1
c variables
c  integer *4 i                # index for PSDX & PSDY
c             j                # index for COHER
c             j1               # index for XPSD
c             inc22            # = 2 * inc2
c  real r                      # product of 2 PSDs
c
      implicit none
      integer inc1, inc2, m, i, j, j1, inc22
      real psdx(*), psdy(*), xpsd(*), coher(*), r
c
c           check arguments & initialize
c
      if ( inc1 .le. 0 .or. inc2 .le. 0 .or. m .le. 1 ) return
      j = 1
      j1 = 1
      inc22 = inc2 * 2
c
      do i = 1, inc1 * ( 2 ** m - 1 ) + 1, inc1
        r = psdx(i) * psdy(i)
        if ( r .le. 0 ) then
          coher(j) = 0.0
        else
          coher(j) = xpsd( j1) **2 /  psdx(i) / psdy(i)
          end if
        j = j + inc2
        j1 = j1 + inc22
        end do
      return
      end



      SUBROUTINE SCROSS( X, Y, NINCR, NPTS, CROSS, INC, MINLAG, MAXLAG,
     1  IERR, IWK, ZWK )
c
c     Calculates  MINLAG negative lag correlation values and MAXLAG positive
c lag correlation values  from NPTS values in the arrays X and Y taken at
c intervals of NINCR from the data arrays X and Y . The indexing increment
c for the  autocorrelation array CROSS is INC    .
c
c ARGUMENTS
c  real *4    x(*)                     # data array                     input
c             y(*)                     # data array                     input
c integer *4  nincr                    # data indexing increment        input
c             npts                     # number of data points          input
c real        cross(*)                 # cross correlation values  output
c integer *4  inc                      # ACORR indexing increment       input
c             minlag                   # minimum lag                    input
c             maxlag                   # maximum lag                    input
c             ierr                     # error flag = 0  -no errors-    output
c             iwk(*)                   # work array for FFT             ----
c complex     zwk(*)                   # work array for FFT             ----
c
c minimum sizes of array arguments
c     define   len = 2 **m >= 2 * max( minlags, maxlags )
c   x & y -   nincr * ( npts - 1 ) + 1
c   cross -   minlag + maxlag - 1
c   iwk   -   len+1
c   zwk   -   len * 3/2 +1
c
c  ERROR FLAG
c ierr = 0    no errors
c      = 129  size or increment argument <= 0
c      = 130  MINLAG too big for number of data points or too small
c      = 131  MINLAG : not enough data for 2**m point FFT
c      = 136  MAXLAG too big for number of data points or too small
c      = 140  MAXLAG : not enough data for 2**m point FFT
c      = 138  - 136 + 130
c      = 139  - 136 + 131
c      = 142  - 130 + 140
c      = 143  - 131 + 140
c variables
c  real xyvar                  # product of x & y rms
c       qvar                   # function : variance
c  integer *4 i                # looping index
c             ilast            # index of last correlation value
c             m                # 2**m = length of DFT
c             ni               # backward looping index
c             n1, n2           # number of samples used in each convolution
c
      implicit none
      real *4 x(*), y(*), cross(*), xyvar, qvar
      integer *4  nincr, npts, inc, minlag, maxlag, ierr, iwk(*), i,
     1  ilast, ni, ierr1, n1, n2
      complex zwk(*)
c
c
c         calculate negative lag correlations
      call sfconv( y, x, nincr, npts, cross, inc, minlag, n1, ierr,
     1  iwk, zwk )
      ni = minlag
      do i = 1, minlag / 2
        call swap4( cross(i), cross(ni) )
        ni = ni - 1
        end do
c
c         calculate positive lag correlations
c
      call sfconv( x, y, nincr, npts, cross(minlag), inc, maxlag, n2,
     1  ierr1, iwk, zwk )
      if ( ierr .gt. 129 ) ierr = ior( iand( ierr1, 3 ) * 4, ierr )
c
c         normalize the result
c

      xyvar = sqrt( qvar( x, nincr, npts ) * qvar( y, nincr, npts ) )
      ilast = ( minlag - 1 ) * inc + 1
      do i = 1, ilast, inc
        cross(i) = cross(i) / xyvar / n1
        end do
      do i = ilast + inc,  ( minlag + maxlag - 2 ) * inc +1
        cross(i) = cross(i) / xyvar / n2
        end do
      return
      end




      SUBROUTINE SFCONV( X, Y, NINCR, NPTS, CONV, INC, NLAGS, NUSED,
     1  IERR, IWK, ZWK )
c
c       Fast CONVolution
c  Evaluates the convolution of two data arrays by the FFT method
c
c  The steps are
c    a) select m s.t. len = 2**(m+1) >= nlags
c    b) divide the data array into npts/len2 segments
c    c) calculate len point DFT of each ith segment
c       1)   Xi=DFT{ xi(1:len2), double pad with zeros }
c       2)   Yi=DFT{ yi(1:len) }
c    d) aggregate Z = Z + Xi.conjg(Yi)
c    e) calculate inverse DFT of Z
c    f) convolution = real(Z)
c
c arguments
c  real x(*),           # first , reference data stream       input
c       y(*),           # second data stream                  input
c  integer *4 nincr     # data indexing increment             input
c             npts      # number of data points               input
c  real conv(*)         # convolution result array            output
c  integer *4 inc       # result array indexing increment     input
c             nlags     # number of convolution lags          input
c                            ( including zero lag )
c             nused     # number of data samples used         output
c             ierr      # error flag                          output
c  integer *4 iwk(*)    # work array for FFT                  ----
c  complex zwk(*)       # work array for FFT                  ----
c
c minimum sizes of array arguments
c     define   len = 2 **m >= nlags
c   x & y -   nincr * ( npts - 1 ) + 1
c   xcorr -   ( nlags - 1 ) * inc + 1
c   iwk   -   m + 2
c   zwk   -   len * 3 + 2
c
c constant
c   pi = 3.1415926
c variables
c integer *4 i, k       # looping index
c            j          # data index
c            jlast      # data index of first data item of the last segment
c            jnlast     # data index of first data item of the 2nd last segment
c            ilast      # index of last convolution value
c            len        # length of DFT
c            len2       # len / 2
c            len4       # len / 4
c            m          # log base 2 of number of correlation lags required
c            mm         # log base 2 of DFT length
c            nhalf      # data indexing increment for loading 2nd 1/2 of FFT
c            ni         # backward running index for unscrambling DFT
c            nk         # backward running index for Z
c            nlast      # index of the last data value
c real xmean, ymean     # mean value of data in a segment
c      qmean            # function : find mean value
c complex a             # DFT of first data stream
c         b             # DFT of second data stream
c         d             # angle increment
c         w             # rotation multiplier
c
      implicit none
      real *4 pi
      parameter( pi=3.1415926 )
c
      integer *4 i, k, j, jlast, jnlast, ilast, len, len2, len4, m, mm,
     1  nhalf, ni, nk, nlast, nincr, npts, inc, nlags, iwk(*), ierr,
     2  nused
      real xmean, ymean, qmean, x(*), y(*), conv(*)
      complex a, b, d, w, zwk(*)
c
      if ( nincr .le. 0 .or. npts .le. 0 .or. inc .le. 0 ) then
        ierr = 129
        return
      else if ( nlags  .gt. npts .or. nlags .lt. 2 ) then
        ierr = 130
        return
        end if
c
c           initialize
c
      m = int( log( real( nlags - 1 ) ) / log( 2.0 ) ) + 1
      if ( 2 ** m .gt. npts ) then
        ierr = 131
        return
        end if
c
      ierr = 0
c
c
c           initialize
c
      m = max( int( log( real( nlags - 1 ) ) / log(2.0) ) + 1, 2 )
      mm = m + 1
      len = 2 ** mm
      len2 = len / 2
      len4 = len2 / 2
      j = 1
      nhalf = nincr * len2
      jlast = ( npts - len2 ) * nincr + 1
      jnlast = jlast - len2 * nincr
c
c           zero the accumulator array
c
      do i = len + 2, len + len2 + 2
        zwk(i) = 0.0
        end do
c
c           start FFT averaging loop
c
      nused = 0
c      xmean = qmean( x, nincr, npts / len2 * len2 )
c      ymean = qmean( y, nincr, npts / len2 * len2 )
      do while ( j .le. jlast )
c
c         load the FFT
c
c         if not the last segment
c
        xmean = qmean( x(j), nincr, len2 )
        if ( j .le. jnlast ) then
          ymean = qmean( y(j), nincr, len )
          do i = 1, len2
            zwk(i) = cmplx( x(j) - xmean, y(j) - ymean )
            zwk(len2+i) = cmplx( 0.0, y( nhalf+j ) - ymean )
            j = j + nincr
            end do
        else
c
c         if the last segment
c
          ymean = qmean( y(j), nincr, len2 )
          do i = 1, len2
            zwk(i) = cmplx( x(j) - xmean, y(j) - ymean )
            zwk(len2+i) = cmplx( 0.0, 0.0 )
            j = j + nincr
            end do
          end if
        nused = nused + len2
c
c           do the "len" point FFT
c
        call fft2c( zwk, mm, iwk )
c
c           unscramble the transform
c             & accumulate the DFT of the convolution
c         2**m >= nlags . the DFTs of X & Y have 2**m complex values
c         The aggregated result Z = Z + X.Y*  is put into
c         zwk(len+2)...zwk(len+2+len2)
c
        k = len + 2
        ni = len + 1
        zwk( ni ) = zwk(1)
        do i = 1, len2+1
          a = ( zwk(i) + conjg( zwk(ni) ) ) * 0.5
          b = zwk(ni) - conjg( zwk(i))
          b = cmplx( aimag(b), real(b) ) * 0.5
          zwk(k) = zwk(k) + a * conjg( b )
          k = k + 1
          ni = ni - 1
          end do
        end do
c end of segmenting loop
c
c       load the inverse fft
c
c  k = forward stepping Z index
c  nk= backward stepping Z index
c  i = forward stepping index of DFT load array
c  ni= backward stepping index of DFT load array
c
      w = ( 1.0, 0.0 )
      d = cexp( cmplx( 0.0, pi / len2 ) )
      ni = len2 + 1
      k = len + 2
      nk = len2 + k
      zwk(k) = 0.0
      zwk(nk) = 0.0
      do i = 1, len4 + 1
        a = zwk(k) + conjg(zwk(nk))
        b = ( zwk(k) + conjg(zwk(nk)) ) * w
        b = cmplx( aimag(b), real(b) )
        zwk(i) = a - conjg(b)
        zwk(ni)= conjg(a) + b
        w = w * d
        ni = ni - 1
        k = k + 1
        nk = nk - 1
        end do
c
c       do the "len/2" point inverse FFT
c
      call fft2c( zwk, m, iwk )
c
c       put the result into the convolution array
c
      k = 1
      do i = 1, inc * ( nlags - 1 ) + 1, 2 * inc
        conv(i) = real( zwk(k) ) / len
        conv(inc+i) = aimag( zwk(k) ) / len
        k = k + 1
        end do
c
      return
      end



      SUBROUTINE SFFPSD( X, NINCR, NPTS, DELT, M, CWIND,
     1    RPSD, IPSD, DELF, ZWK, IWK, IERR )
c
c       Power Spectral Density by the FFT - periodogram method
c    A total of up to NPTS may be taken at intervals of NINCR from the data
c array X . The data is split into segments (ensembles), each of 2**M points
c The mean value is removed from each segment .
c An FFT is done on each segment and the ensemble average of the FFT squared
c is calculated . This ensemble average is the power spectral density (PSD)
c estimate .
c    The the sampling time between each
c NINCRth data sample , and the psd array increment IPSD must be supplied as
c arguments .
c    Frequency increment DELF, psd estimates RPSD and error flag IERR are
c output arguments . The psd estimate is scaled so that the area under the psd
c curve is equal to the signal variance . Using a time window MAY destroy this
c equality .
c    The argument CWIND is a 3 way switch for selecting
c a)  FFT with
c  1) Half integer frequencies - a 2**m point spectrum, half integer frequency
c    values are statistically independent but arithmetically dependent of the
c    integer frequency values . This is required if the autocorrelation is to
c    be calculated by inverse DFT .
c  2) Integer frequencies - a 2**(M-1) point spectrum with statistically
c    independent values is calculated .
c b) a time domain window on the data
c  1) No (ie rectangular) window - this results in leakage from spectral peaks
c    into "sidebands" , but produces the best possible frequency resolution .
c  2) Hanning (cosine) window - reduces leakage, but at the cost of some
c    widening of spectral peaks ( ie loss of frequency resolution ) .
c     The cosine window is calculated from an approximation derived from the
c    Runge-Kutta-Nystrom 4th order solution of y"=-y . This is calculation of
c    cos(x) (3*) is quicker than by truncated Taylor series (9*), but can still
c    be time consuming if the number of ensembles is large . The maximum errors
c    of the approximation are .053(M=2), .0027(M=3), .00015(M=4), and <10ppm for
c    4<M<21 . Initial conditions are x(0)=0, y(0)=1, y'(0)=0. The approximation
c    is : x(n) = x(n) + h  where  h = 2*pi/(2**M+1)
c         y(n) = (1-h**2+h**4/24) * y(n) + (h-h**3/6) * y'(n)
c         y'(n) = (h-h**3/6+h**5/96) * y(n) + (1-h**2/2+h**4/24) * y'(n)
c    The window amplitude is  sqrt(2/3) * (1-cos(ht) ).
c c) an autocorrelation (lag) window
c  1) None
c  2) Tukey (Hanning) lag window - on the autocorrelation to suppress bad
c    values at large lag values . This smooths the psd estimate and is
c    recommended for half integer frequency spectra .
c
c       ARGUMENTS
c         let "len = 2**m"
c   TYPE      NAME                 DESCRIPTION                           I/O
c  real x( nincr*(npts-1)+1 )  # data array                             input
c  integer *4 nincr            # data indexing increment                input
c             npts             # number of data points available        input
c  real delt                   # time increment for time series data    input
c  integer *4 M                # lenw = 2**M is the number of data      input
c                                points in each FFT
c  character cwind *(*)        # window type -                          input
c  real rpsd(*)                # power spectral density array           output
c  integer *4 ipsd             # indexing increment for PSD array       input
c  delf                        # frequency increment                    output
c  complex zwk(len+1)          # work array for the periodogram         ----
c  integer *4 iwk(m+1)         # work array for FFT                     ----
c             ierr             # error flag = 0    no errors            output
c
c    WINDOW TYPES
c  Code Letter    Action                               Reason
c    H or h      Hanning time ( cosine taper ) window   reduce leakage
c    A or a      Tukey autocorrelation window           smoothing
c    Z or z      Double ZERO pads the data to give      better "resolution"
c                half integer frequencies               & for autocorrelation
c anything else  No special windows, no zero padding
c
c    N.B A spectrum with integer frequencies only has 2**(m-1) spectral density
c        estimates . In this case "zwk" need have only size "len/2" and "iwk"
c        size "m" . A spectrum with half integer frequencies has 2**m spectral
c        density estimates
c
c    ERROR FLAGS
c  error flag = 0    no errors
c             = 129  argument <= 0
c             = 130  insuffient data
c             = 131  zero or -ve PSD value
c constants
c  pi
c  sqrt23 = sqrt(2./3.)
c
c variables
c  integer *4 i, k             # loop indices
c             j, nj            # data index
c             ilast            # index of last PSD estimate
c             len              # length of FFT = 2 ** M
c             len2             # 2 * len
c             lend             # length of data sample / 2
c             mm               # 2**m = size of fft
c             nfft             # number of FFTs calculated
c             ni               # backward running index
c             nincr2           # 2 * nincr
c  complex a, b                # temporary - for unscrambling the FFT result
c          d                   # angle increment in unscramble convolution
c          w                   # is angle of rotation in unscramble convolution
c  real h                      # window function angle increment
c       s, s1, s2              # temporary vars for smoothing
c       y,yy, yd               # values of & derivative of time window function
c       xmean                  # mean value of data in segment
c       qmean                  # function : find mean value
c
c  logical *2 fint             # integer frequencies only flag
c             ftwind           # time window used flag
c  logical *4 ein              # function : is CSTR in CSET ?
c
      implicit none
      real pi, sqrt23
      parameter( pi = 3.1415926, sqrt23 = 0.81649658 )
c
      real x(*), xmean, delt, rpsd(*), delf, s, s1, s2, y, yy, yd, h,
     1  qmean
      integer *4 iwk(*), nincr, npts, m, ipsd, ierr, len, len2,
     1  nfft, i, j, k, ni, ilast, lend, mm, nj, nincr2
      complex zwk(*), a, b, w, d
      character *(*) cwind
      logical *2 fint, ftwind
      logical *4 ein
c
c           check arguments
c
      if ( nincr .le. 0 .or. npts .le. 1 .or. delt .le. 0.0 .or.
     1    m .le. 1 .or. ipsd .le. 0 ) then
        ierr = 129
        return
        end if
c
c           initialize
c
      mm = m
      len = 2 ** m
      lend = len / 2
      if ( len .gt. npts ) then
        ierr = 130
        return
        end if
      fint = .not. ein( 'Z', Cwind ) .and. .not. ein( 'z', Cwind )
      if ( fint ) then
        len = len / 2
        mm = mm - 1
        end if
      len2 = len / 2
      d = cexp( cmplx( 0.0, pi / len ) )
      delf = 0.5 / delt / float( len )
      ierr = 0
      nfft = 0
      nincr2 = 2 * nincr
      j = 1
c
c           initialize for time window
c
      ftwind = ein( 'h', Cwind ) .or. ein( 'H', Cwind )
      if ( ftwind ) then
        h = 2.0 * pi / ( 2 * lend + 1 )
        s = 1.0 - h **2 / 2.0 * ( 1.0 - h **2 / 12.0 )
        s1 = h * ( 1.0 - h **2 / 6.0 )
        s2 = - s1 - h **5 / 96.0
        end if
c
c           zero the psd array
c
      do i = 1, ipsd * ( len - 1 ) + 1, ipsd
        rpsd(i) = 0
        end do
c
c           start FFT averaging loop
c
      do while ( j .le. nincr * ( npts - lend * 2 ) + 1 )
c
c           find the mean value of data in the segment
c
        xmean = qmean( x(j), nincr, lend * 2 )
c
c           time window & load the FFT array
c
        nfft = nfft + 1
        if ( ftwind ) then
          y = -sqrt23
          yd = 0.0
          ni = lend
          nj = j + nincr * ( 2 * lend - 1 )
          do i = 1, lend / 2
            yy = s * y + s1 * yd
            yd = s2 * y + s * yd
            y = yy
            yy = s * y + s1 * yd
            yd = s2 * y + s * yd
            zwk(i) = cmplx( ( x(j) - xmean ) * ( sqrt23 + y ),
     1                      ( x(j+nincr) - xmean ) * ( sqrt23 + yy ) )
            zwk(ni) = cmplx( ( x(nj-nincr) - xmean ) * ( sqrt23 + yy ),
     1                      ( x(nj) - xmean ) * ( sqrt23 + y ) )
            y = yy
            j = j + nincr2
            nj = nj - nincr2
            ni = ni - 1
            end do
          j = j + lend * nincr
        else
c
c         load the FFT - no window
c
          do i = 1, lend
            zwk(i) = cmplx( x(j) - xmean, x(j+nincr) - xmean )
            j = j + nincr2
            end do
          end if
c
c           pad the 2nd half of FFT array with zeros
c
        if ( .not. fint ) then
          do i = lend + 1, 2 * lend
            zwk(i) = ( 0.0, 0.0 )
            end do
          end if
c
c           do the "len" point FFT
c
        call fft2c( zwk, mm, iwk )
c
c           unscramble the transform ( with half integer frequencies )
c
        ni = len + 1
        w = 1.0
        zwk( ni ) = zwk(1)
        do i = 1, len2+1
          a = ( zwk(i) + conjg( zwk(ni) ) ) * 0.5
          b = - conjg( zwk(i)) + zwk(ni)
          b = cmplx( aimag(b), real(b) ) * w * 0.5
          zwk(i) = a + b
          zwk(ni) = conjg( a - b )
          w = w * d
          ni = ni - 1
          end do
c
c           accumulate the PSD
c
        k = 1
        do i = 1, len
          rpsd(k) = rpsd(k) + real( zwk(i) ) **2 + aimag( zwk(i) ) **2
          k = k + ipsd
          end do
        end do
c end loop
c
c           apply autocorrelation window
c
      ilast = ipsd * ( len - 1 ) + 1
      if ( ein( 'A', Cwind ) .or. ein( 'a', Cwind ) ) then
        s2 = 0.0
        s1 = ( rpsd(1) + rpsd(1+ipsd) ) / 4.0
        rpsd(1) = s1
        do i = 1 + ipsd, ilast - ipsd, ipsd
          s = ( rpsd(i+ipsd) + rpsd(i) ) / 4.0
          rpsd(i) = s + rpsd( i-ipsd ) - s2
          s2 = s1
          s1 = s
          end do
        rpsd(ilast) = s
        end if
c
c           normalize the result
c
      s = delt / float( nfft * lend )
      do i = 1, ilast, ipsd
        if ( rpsd(i) .le. 0.0 )
     1    ierr = 131
        rpsd(i) = abs( rpsd(i) ) * s
        end do
      return
      end


      SUBROUTINE SLOGSP( SPECLIN, FINSTART, FININC, INCIP, NPTSIP,
     1  FROCTAVE, INCOP, NPTSOP, FREQ, SPECLOG )
c
c       Routine to convert a spectrum with constant frequency increment
c to a spectrum with a logarithmic frequency increment . E.g. convert
c a spectrum generated by an FFT algorithm into an octave spectrum .
c
c arguments
c  real SPECLIN(*)   # input spectral values with constant frequency     INPUT
c                    #   increment
c       FINSTART     # centre frequency of first band of input spectrum  INPUT
c       FININC       # frequency increment of input spectrum             INPUT
c  integer *4 INCIP  # array indexing increment for input spectrum       INPUT
c             NPTSIP # number of values in the input spectrum            INPUT
c
c  real FROCTAVE     # bandwidth of output spectrum in fractions of an   INPUT
c                    # octave : 1.0=octave, 3.0=third octave
c  integer *4 INCOP  # array indexing increment for the output arrays    INPUT
c             NPTSOP # number of values in the output arrays             OUTPUT
c  real FREQ(*)      # list of log increment frequencies                 OUTPUT
c       SPECLOG(*)   # spectral values with log frequency increment      OUTPUT
c
c constants
c             fact             # actual mean octave increment
c             fref             # refernce freqency for log frequency values
c Variables
c  integer *4 iin              # input ( linear ) spectrum index
c             iout             # output (log) spectrum index
c             lastin           # index of last point in input spectrum
c             lastout          # index of last point in output spectrum
c             nstart           # number of log freq. bands between start
c                              # of the spectrum and the reference frequency
c             factn            # log frequency bandwidth or half bandwidth
c             fbegin           # start frequency of log spectrum
c             fbip             # lower edge of I/P frequency band
c             fledge           # lower edge of log frequency band
c             fuedge           # upper edge of log frequency band
c             fmaxin           # maximum input frequency
c             fnext            # centre frequency of next log band
c             power            # aggregate power over current log band
c
      implicit none
      real fref, fact
      parameter( fref=15.8546, fact=1.995262314968 )
c
      integer *4 iin, incip, incop, iout, lastin, lastout, nptsip,
     1   nptsop, nstart
      real factn, fbegin, fbip, fininc, finstart, fledge, fmaxin, fnext,
     1   froctave, fuedge, power, freq(*), speclin(*), speclog(*)
c
c            display all non-array arguments
c
c           check arguments
c
      if ( finstart .lt. 0.0 .or. fininc .le. 0.0 .or. incip .le. 0 .or.
     1    nptsip .le. 0 .or. 1000.0.lt. froctave .or.
     2    froctave .lt. 0.001 .or. incop .le. 0 ) return
c
      fmaxin = ( nptsip - 1 ) * fininc + finstart
      fbegin = finstart
      factn = fact ** ( 1.0 / froctave )
      if ( finstart / fininc .lt. 0.5 )
     1  fbegin = fbegin + fininc
      nstart = froctave / 0.3 * alog10( fbegin / fref )
      if ( nstart .ge. 0 )
     1  nstart = nstart + 1
c
c           calculate list of log increment frequencies
c
      fnext = fref * fact ** ( nstart / froctave )
      nptsop = 0
      iout = 1
      do while( fnext .lt. fmaxin )
        nptsop = nptsop + 1
        freq(iout) = fnext
        iout = iout + incop
        fnext = factn * fnext
        end do
c
c           CONVERT SPECTRUM TO LOG FREQUENCY INCREMENT
c
      fbip = finstart - fininc / 2.0
      factn = sqrt( factn )
      fledge = freq(1) / factn
      lastin = ( nptsip - 1 ) * incip + 1
      lastout = ( nptsop - 1 ) * incop + 1
      iin = 1
c
c           move frequency pointer up to beginning of O/P spectrum
c
      do while( fledge .ge. fbip + fininc )
        fbip = fbip + fininc
        iin = iin + incip
        end do
c
c           loop for O/P spectrum
c
      do iout = 1, lastout, incop
        power = 0.0
        fuedge = freq(iout) * factn
c
c           if the O/P log band includes a point in the I/P spectrum
c
        if ( fledge .lt. fbip + fininc .and.
     1             fbip + fininc .lt. fuedge ) then
c
c           lower edge of log band
c
          if ( fledge .lt. fbip + fininc .and. iin .le. lastin ) then
            fbip = fbip + fininc
            power = power + ( fbip - fledge ) * speclin(iin)
            iin = iin + incip
            end if
c
c           middle segment
c
          if ( fledge .le. fbip .and. fbip + fininc .le. fuedge ) then
            do while( fbip + fininc .le. fuedge .and. iin .le. lastin )
              power = power + fininc * speclin(iin)
              iin = iin + incip
              fbip = fbip + fininc
              end do
            end if
c
c           upper edge of log band
c
          if ( iin .le. lastin .and. fbip .lt. fuedge ) then
            power = power + ( fuedge - fbip ) * speclin(iin)
            end if
c
c           calculate power spectral density
c
          speclog(iout) = power / ( fuedge -fledge )
        else
          speclog(iout) = speclin(iin)
          end if
        fledge = fuedge
        end do
c
      return
      end


      SUBROUTINE SMAFRA( X, NINCR, NPTS, H, M )
c
c     Subroutine Moving Average Filter - Replace data - Array filter spec
c       Pass the signal in the array X through a Moving Average FILTER .
c The input signal is replaced by the result . The M filter coefficients
c are specified in the array H .
c
c arguments
c  real X(*)             # data array
c  integer *4 NINCR      # data indexing increment
c             NPTS       # number of data points used
c  real H(*)             # filter coeffients, ie filter impulse response
c  integer *4 M          # number of filter coefficients
c
c    If the filter has a linear phase response, a saving in calculation
c time can be made because the filter impulse response has mirror image
c symmetry about its central value . Ie. h(i)=h(m-i)
c
c Variables
c  integer *4 i          # leading edge index for data array
c             k          # index for filter coefficients
c             l          # trailing edge index for data array
c             m1         # m - 1
c             nlast      # index of last data item
c             mlast      # index of mth data item
c             m21        # limit of inner looping array index
c             m12        # mid span data array index
c             j          # span index
c  logical *2 flinph     # linear phase filter
c  real h12              # middle filter coefficient
c
      implicit none
      integer *4 i, k, l, m1, mlast, m21, m12, j, m, nincr, npts, nlast
      logical *2 flinph
      real x(*), h(*), h12
c
      if ( nincr .le. 0 .or. npts .le. 1 .or. m .le. 0 )
     1  return
      nlast = ( npts - 1 ) * nincr + 1
      mlast = ( m - 1 ) * nincr + 1
      flinph  = m .ne. 1
      do i = 1, m / 2
        flinph = flinph .and. abs( h(i) / h(m+1-i) - 1.0 ) .lt. 0.000003
        end do
c
c  ********* linear phase filter
c
      if ( flinph ) then
        m21 = ( m / 2 - 1 ) * nincr
        m1 = ( m - 1 ) * nincr
c
c           odd order filter
c
        if ( mod( m, 2 ) .ne. 0 ) then
          m12 = ( m - 1 ) / 2
          h12 = h( (m+1)/2 )
          do i = nlast, mlast,  -nincr
            l = i - m1
            k = 2
            x(i) = h(1) * ( x(i) + x(l) ) + h12 * x(i-m12)
            do j = nincr, m21, nincr
              x(i) = x(i) + h(k) * ( x(i-j) + x(l+j) )
              k = k + 1
              end do
            end do
c
c           even order filter
c
        else
          do i = nlast, mlast, -nincr
            l = i - m1
            k = 2
            x(i) = h(1) * ( x(i) + x(l) )
            do j = nincr, m21, nincr
              x(i) = x(i) + h(k) * ( x(i-j) + x(l+j) )
              k = k + 1
              end do
            end do
          end if
c
c  ********* non-linear phase filter
c
      else
        m21 = ( m - 1 ) * nincr
        do i = nlast, mlast, -nincr
          x(i) = h(1) * x(i)
          k = 2
          do j = nincr, m21, nincr
            x(i) = x(i) + h(k) * x(i-j)
            k = k + 1
            end do
          end do
        end if
c
c           tail
c
      do i = mlast - nincr, 1, -nincr
        x(i) = h(1) * x(i)
        k = 2
        do j = nincr, i - 1, nincr
          x(i) = x(i) + x(i-j) * h(k)
          k = k + 1
          end do
        end do
      return
      end


      SUBROUTINE SMAFRF( X, NINCR, NPTS, H, M, CFFILE )
c
c     Subroutine Moving Average Filter - Replace data - File filter spec
c       Pass the signal in the array X through a Moving Average FILTER .
c The input signal is replaced by the result . The filter coefficients
c are specified in the formatted file CFFILE .
c    If you only want to read the filter coefficients , put NPTS <= 1 .
c    If a file is already open or the file name is repeated , the file
c is not read a second time .
c    The format of the filter specification file is the same as that written
c by MATRIXX in ``fsave''ing the array containing the filter coefficients .
c
c arguments
c  real X(*)             # data array
c  integer *4 NINCR      # data indexing increment
c             NPTS       # number of data points used
c  real H(*)             # filter coefficients
c       M                # number of filter coefficients
c  character CFFILE*(*)  # name of filter specifier file
c
c Constants
c  Hkbdlu =5      # keyboard
c  Hvdulu =6      # v.d.u.
c
c Variables
c  logical fopen  # filter spec file is opened & read
c  real h(128)    # filter coefficients, ie filter impulse response
c  integer i      # looping index
c          m      # number of filter coefficients
c          ispclu # logical unit no. for filter specification file
c          open_file     # function : opens file
c          new_lun       #          : get new logical unit number
c  character  Cifile*64  # filter spec file name
c
      implicit none
      integer Hkbdlu, Hvdulu
      parameter( Hkbdlu=5, Hvdulu=6 )
c
      real h(*), x(*)
      integer nincr, npts, i, m, ispclu, open_file, new_lun
      character Cifile*64, Cffile*(*)
      logical fopen
      common / Af77lib7 / ispclu, fopen, Cifile
      data ispclu / 0 /
      data fopen  / .false. /
      data Cifile / ' ' /
c
c           open filter specification file
c
      if ( .not. fopen .or. Cffile .ne. ' ' .and.
     1        Cffile .ne. Cifile ) then
          if ( Cffile .eq. '?' ) then
              write( Hvdulu, 1 )
1             format( ' MA (FIR) filter spec ', $ )
          else
              Cifile = Cffile
              end if
          if ( ispclu .eq. 0 ) then
              ispclu = new_lun()
              end if
          if ( 0 .eq. open_file( ispclu, Cffile, 'old', 0 ) ) then
              write( Hvdulu, 2 )
2             format( ' SMAFRF : Cannot open spec file' )
              return
              end if
          end if
      fopen = .TRUE.
c
c           read filter specification file
c
      read( ispclu, 3, ERR=4 ) m
3     format( // 15x, i5 )
      read( ispclu, *, ERR=4 ) ( h(i), i = 1, min( m, 128 ) )
      close( UNIT=ispclu )
c
c           apply the filter
c
      call smafra( x, nincr, npts, h, m )
      return
c
c           trap
c
4     write( Hvdulu, 5 )
5     format( ' Cannot read spec file' )
      return
      end


      SUBROUTINE SMVMEAN( IN, ININCR, NPTS, M, OUT, OUTINCR )
c
c Moving average filter.
c   A window of M samples is passed along the data array IN . The data samples
c within the window are averaged and placed in the OUT array . If M is odd,
c the time delay of the filter is zero . If M is even, the time delay of the
c filter is one half the sampling interval . At the ends of the data record
c the effective length of the filter decreases to about M/2 because the centre
c of the window starts at the first element of the array and finishes at the
c last element of the array .
c
c Arguments
c  real    in(*)    # unfiltered input data stream  of length inincr*(npts-1)+1
c  integer inincr   # indexing increment for the input data stream
c          npts     # number of data points
c          m        # length of moving average filter
c  real    out(*)   # filtered output data stream of length outincr*(npts-1)+1
c  integer outincr  # indexing increment for the output data stream
c
c Variables
c  integer m_new       # corrected average length
c          m_on_2      # m_new/2
c          first,last  # limits of the windoe on the input data stream
c          part1       # location of end of part 1 on the output data stream
c          part3       # location of start of part 3 on the output data stream
c          i           # output data stream index
c  double precision sum # moving agregate
c
      implicit none
      integer inincr, npts, m, outincr
      real    in(*), out(*)
c
      integer m_new, m_on_2, first, last, part1, part3, i
      double precision sum
c
c         check arguments
c
      if ( inincr .le. 0 .or. npts .le. 0 .or. m .le. 0 .or.
     1      outincr .le. 0 ) return
      if ( m .gt. npts ) then
        m_new = npts
      else
        m_new = m
        end if
c
c         initial locations of the ends of the window
c
      m_on_2 = ( m_new - 1 ) / 2
      first  = ( m_on_2 - m_new ) * inincr + 1
      last   =  m_on_2 * inincr + 1
c
c The input data is filtered in three parts .
c   1) starting from the first sample, the length of the average increases to M
c   2) apply a moving average of length of M
c   3) at the end of the data, decrease the length of the average
c
c       FIRST PART
c
c       location of end of first part
c
      part1 = ( m_new - m_on_2 - 1 ) * outincr + 1
c
c       sum the input data over the initial window length
c
      sum = 0.0
      do i = 1, last - inincr, inincr
        sum = sum + in(i)
        end do
c
c       average of increasing length
c
      do i = 1, part1, outincr
        sum = sum + in(last)
        out(i) = sum / last
        first = first + inincr
        last = last + inincr
        end do
c
c       SECOND PART
c
c       location of beginning of third part
c
      part3 = ( npts - m_on_2 - 1 ) * outincr + 1
c
c       average of length M over the middle of the data
c
      do i = part1 + outincr, part3, outincr
        sum = sum + in(last) - in(first)
        out(i) = sum / m
        first = first + inincr
        last = last + inincr
        end do
c
c       THIRD PART
c
c       average of decreasing length
c
      do i = part3 + outincr, ( npts - 1 ) * outincr + 1, outincr
        sum = sum - in(first)
        out(i) = sum / ( npts - first )
        first = first + inincr
        last = last + inincr
        end do
      return
      end


      SUBROUTINE SPDF1( X, NINCR, NPTS, RPDF, IPDF, NPDF, XMIN, DELX )
c
c       Calculates the Probability density function of the data in X .
c The result , RPDF is normalized so that its integral from XMIN to XMAX
c is unity .
c
c    Arguments
c  real x(nincr*(npts-1)+1)    # data array                             input
c  integer *4 nincr            # data indexing increment                input
c             npts             # number of data points used             input
c  real rpdf(ipdf*(npdf-1)+1)  # probability density                    output
c  integer *4 ipdf             # probability density indexing increment input
c             npdf             # number of values in PDF                input
c  real xmin                   # minimum data value                     output
c       delx                   # data (class) interval in PDF           output
c
c variables
c  integer *4 nlast            # last data array index
c             ibin             # pdf bin which is incremented
c             i                # looping index
c  real rbinw                  # pdf bin width
c       rnorm                  # pdf increment size
c       xmax                   # data maximum
c
      implicit none
      integer *4 nincr, ipdf, nlast, ibin, npts, npdf, i
      real x(*), rpdf(*), xmin, delx, rbinw, rnorm, xmax
c
      nlast = nincr * ( npts - 1 ) + 1
c
c         find maximum & minimum of the data
c
      xmax = x(1)
      xmin = x(1)
      do i = 1 + nincr, nlast, nincr
        xmin = min( xmin, x(i) )
        xmax = max( xmax, x(i) )
        end do
c
c         calculate PDF bin width & offset, PDF increment etc
c
      delx = ( xmax - xmin ) / float( npdf )
      rbinw = delx * 1.00001 / float( ipdf )
      rnorm = 1.0 / delx / float(npts)
c
c         zero the PDF
c
      do i = 1, ipdf * ( npdf - 1 ) + 1
        rpdf(i) = 0.0
        end do
c
c         accumulate the PDF
c
      do i = 1, nlast, nincr
        ibin = ( x(i) - xmin ) / rbinw + 1
        rpdf(ibin) = rpdf(ibin) + rnorm
        end do
c
      return
      end


      SUBROUTINE SXCORR( X, Y, NINCR, NPTS, XMEAN, YMEAN,
     1  XCORR, INCORR, MINLAG, MAXLAG )
c
c     Calculates  MINLAG to  MAXLAG cross  correlation values  from NPTS
c values in the arrays X and Y taken at intervals of NINCR from the data
c array X  . The indexing increment for the  autocorrelation array ACORR
c is INCORR .
c
c ARGUMENTS
c  real *4    x(nincr*(npts-1)+1)      # data array                     input
c             y(nincr*(npts-1)+1)      # data array                     input
c integer *4  nincr                    # data indexing increment        input
c             npts                     # number of data points          input
c  real *4    xmean                    # mean of the x data             input
c             ymean                    # mean of the y data             input
c             xcorr(incorr*(maxlag-minlag+1)# cross correlation values  output
c integer *4  incorr                   # ACORR indexing increment       input
c             minlag                   # minimum lag                    input
c             maxlag                   # maximum lag                    input
c variables
c  double precision raggr              # aggregate
c  real xvar                           # signal variance
c  integer *4 nlast                    # index of last data value
c             i                        # data index
c             j                        # autocorrelation index
c             k                        # lag index
c
      implicit none
      integer *2 Hvdulu
      parameter( Hvdulu=6 )
c
      integer *4 nincr, npts, incorr, minlag, maxlag, nlast, k, i, j
      real *4 x(*), y(*), xmean, ymean, xyvar, xcorr(*)
      real *8 raggrx, raggry
c
c           check arguments
c
      if ( nincr .le. 0 .or. npts .le. 0 .or. incorr .le. 0 .or.
     1    abs(minlag) .gt. npts .or. abs(maxlag) .gt. npts .or.
     2    minlag .ge. maxlag )
     3  return
      nlast = nincr * ( npts - 1 ) + 1
      k = nincr * minlag
c
c           variance
c
      raggrx = 0.0d0
      raggry = 0.0d0
      do i = 1, nlast, nincr
        raggrx = raggrx + ( x(i) - xmean ) **2
        raggry = raggry + ( y(i) - ymean ) **2
        end do
      xyvar = sqrt( raggrx * raggry )
      if ( xyvar .eq. 0.0 ) then
        write( Hvdulu, 1 )
1       format( ' SXCORR : Unvarying signal' )
        xyvar = 1.0
        end if
c
c           cross correlations
c
      do j = 1, ( maxlag - minlag ) * incorr + 1, incorr
        raggrx = 0.0d0
        do i = max( 1-k, 1 ), min( nlast, nlast-k ), nincr
          raggrx = raggrx + dble( ( x(i) - xmean ) * ( y(i+k) - ymean ))
          end do
        k = k + nincr
        xcorr(j) = sngl( raggrx ) / xyvar
        end do
      return
      end


      SUBROUTINE SXFER( PSDX, INC1, XPSD, XFER, INC2, M )
c
c         TransFER function : from cross psd
c    Calculates the transfer function of a system with input x(t) and output
c y(t) . Each INC1th point in PSDX is the power spectral density of x(t), and
c each INC2th point in XPSD is the cross power spectral density of x(t) and
c y(t) . The transfer function is put into each INC2th point in XFER .
c 2**M data points are taken from each array .
c
c    ARGUMENTS
c real       psdx(n1)          # power spectral density of x(t)         input
c integer *4 inc1              # indexing increment for PSDX            input
c complex    xpsd(n2)          # cross power spectral density of x & y  input
c            xfer(n2)          # transfer function of system            output
c integer *4 inc2              # indexing increment for XPSD & XFER     input
c            m                 # 2**m = n0. of points in each spectrum  input
c              where n1 = inc1 * ( 2 **m - 1 ) + 1
c                    n2 = inc2 * ( 2 **m - 1 ) + 1
c
c variables
c  integer *4 i                # index for PSDX
c             j                # index for XPSD & XFER
c             inc22            # = 2 * inc2
c
      implicit none
      integer inc1, inc2, m, i, j, inc22
      real psdx(*), xpsd(*), xfer(*)
c
c           check arguments & initialize
c
      if ( inc1 .le. 0 .or. inc2 .le. 0 .or. m .le. 1 ) return
      j = 1
      inc22 = inc2 * 2
c
      do i = 1, inc1 * ( 2 ** m - 1 ) + 1, inc1
        if ( psdx(i) .le. 0 ) then
          xfer(j) = 0.0
        else
          xfer(j) = xpsd(j) / psdx(i)
          end if
        xfer(j+1) = xpsd(j+1)
        j = j + inc22
        end do
      return
      end



      SUBROUTINE SXYPSD( X, Y, NINCR, NPTS, DELT, M, CDO,
     1  PSDX, PSDY, XPSD, INC1, INC2, DELF, ZWK, IWK, IERR )
c
c       Cross Power Spectral Density by the FFT - periodogram method
c    A total of up to NPTS may be taken at intervals of NINCR from the data
c arrays X & Y. Each data array is split into segments (ensembles), each of
c  2**M points . The mean value is removed from each segment .
c An FFT is done on each segment and the ensemble average of the FFTs squared
c is calculated . These ensemble averages are the power spectral density (PSD)
c estimates . The cross spectral density estimate is
c FFT( x ) * conjg( FFT( y ) )
c    The the sampling time between each NINCRth data sample, the psd array
c indexing increment INC1, and the cross psd indexing increment INC2 must be
c supplied as arguments .
c    Frequency increment DELF, psd estimates PSDX & PSDY, cross psd estimate
c XPSD and error flag IERR are output arguments . The psd estimate is scaled
c so that the area under the psd curve is equal to the signal variance . The
c cross psd is scaled using the same factor . Using a time window MAY destroy
c this equality .
c    The argument CDO is a switch for selecting
c a)  FFT with
c  1) Half integer frequencies - a 2**m point spectrum, half integer frequency
c    values are statistically independent but arithmetically dependent of the
c    integer frequency values . This is required if the autocorrelation is to
c    be calculated by inverse DFT .
c  2) Integer frequencies - a 2**(M-1) point spectrum with statistically
c    independent values is calculated .
c b) a time domain window on the data
c  1) No (ie rectangular) window - this results in leakage from spectral peaks
c    into "sidebands" , but produces the best possible frequency resolution .
c  2) Hanning (cosine) window - reduces leakage, but at the cost of some
c    widening of spectral peaks ( ie loss of frequency resolution ) .
c     The cosine window is calculated from an approximation derived from the
c    Runge-Kutta-Nystrom 4th order solution of y"=-y . This is calculation of
c    cos(x) (3*) is quicker than by truncated Taylor series (9*), but can still
c    be time consuming if the number of ensembles is large . The maximum errors
c    of the approximation are .053(M=2), .0027(M=3), .00015(M=4), and <5ppm for
c    4<M<18 . Initial conditions are x(0)=0, y(0)=1, y'(0)=0. The approximation
c    is : x(n) = x(n) + h  where  h = 2*pi/(2**M+1)
c         y(n) = (1-h**2+h**4/24) * y(n) + (h-h**3/6) * y'(n)
c         y'(n) = (h-h**3/6+h**5/96) * y(n) + (1-h**2/2+h**4/24) * y'(n)
c    The window amplitude is  sqrt(2/3) * (1-cos(ht) ).
c c) an autocorrelation (lag) window
c  1) None
c  2) Tukey (Hanning) lag window - on the autocorrelation to suppress bad
c    values at large lag values . This smooths the psd estimate and is
c    recommended for half integer frequency spectra .
c d) the output arrays to be filled . Any of the following :-
c  1) None
c  2) PSDX - power spectral density of the X data array
c  3) PSDY - power spectral density of the Y data array
c  4) XPSD - cross power spectral density  of X and Y .
c
c       ARGUMENTS
c         let "len = 2**m"
c   TYPE      NAME                 DESCRIPTION                           I/O
c  real x( nincr*(npts-1)+1 )  # 1st data array                         input
c  real y( nincr*(npts-1)+1 )  # 2nd data array                         input
c  integer *4 nincr            # data indexing increment                input
c             npts             # number of data points available        input
c  real delt                   # time increment for time series data    input
c  integer *4 M                # lenw = 2**M is the number of data      input
c                                points in each FFT
c  character *1 cdo(*)         # what to do                             input
c  real psdx(*)                # power spectral density of x            output
c       psdy(*)                # power spectral density of y            output
c  complex xpsd(*)             # cross power spectral density of x & y  output
c  integer *4 inc1             # indexing increment for PSDX and PSDY   input
c             inc2             # indexing increment for XPSD            input
c  delf                        # frequency increment                    output
c  complex zwk(len+1)          # work array for the periodogram         ----
c  integer *4 iwk(m+1)         # work array for FFT                     ----
c             ierr             # error flag = 0    no errors            output
c
c    WHAT TO DO
c      CDO is a character string containing any number of code letters
c  which control what is done .
c Code Letter   Action
c   x or X    power spectral density of x
c   y or Y    power spectral density of y
c   s or S    cross power spectral density of x & y
c   h or H    Hanning time ( cosine taper ) window   reduce leakage
c   a or A    Tukey autocorrelation window           smoothing
c   z or Z    Double ZERO pads the data to give      better "resolution"
c              half integer frequencies               & for autocorrelation
c anything else  No special windows, no zero padding
c
c    N.B A spectrum with integer frequencies only has 2**(m-1) spectral density
c        estimates . In this case "zwk" need have only size "len/2+1" and "iwk"
c        size "m" . A spectrum with half integer frequencies has 2**m spectral
c        density estimates
c           If a result is not required ( and not requested by CDO ), the
c        corresponding argument in the call to SXYPSD need only be a scalar .
c
c    ERROR FLAGS
c  error flag = 0    no errors
c             = 129  argument <= 0
c             = 130  nothing to do
c             = 131  insuffient data
c             = 137  zero or -ve PSDX value
c             = 138  zero or -ve PSDY value
c             = 140  zero XPSD value
c constants
c  pi
c  sqrt23 = sqrt(2./3.)
c
c variables
c  integer *4 i, k1, k2        # loop indices
c             j, nj            # data index
c             ilast1           # index of last PSD estimate
c             ilast2           # index of last XPSD estimate
c             nlast            # index of last data value
c             len              # length of FFT = 2 ** M
c             len2             # len / 2
c             lend             # length of data sample
c             lend2            # lend / 2
c             mm               # 2**m = size of fft
c             nfft             # number of FFTs calculated
c             ni               # backward running index
c             nincr2           # 2 * nincr
c  complex a, b                # temporary - for unscrambling the FFT result
c          d                   # angle increment in unscramble convolution
c          w                   # is angle of rotation in unscramble convolution
c  real h                      # window function angle increment
c       s, s1, s2              # temporary vars for smoothing
c       z,zz, zd               # values of & derivative of time window function
c       xmean                  # mean value of data in segment
c       qmean                  # function : find mean value
c
c  logical *2 fhint            # half integer frequencies flag
c             ftwind           # time window used flag
c             fpsdx            # psd of X is required
c             fpsdy            # psd of Y is required
c             fxpsd            # cross psd is required
c  logical *4 ein              # function : is Cstr contained in Cset
c
      implicit none
      real pi, sqrt23
      parameter( pi = 3.1415926, sqrt23 = 0.81649658 )
c
      character cdo *(*)
      complex xpsd(*), zwk(*), a, b, d, w
      real x(*), y(*), psdx(*), psdy(*), delt, delf, h, s, s1, s2,
     1  z, zz, zd, xmean, ymean, qmean
      integer *4 nincr, npts, M, inc1, inc2, iwk(*), ierr, i, k1, k2,
     1  j, nj, ilast1, ilast2, len, len2, lend, mm, nfft, ni, nincr2,
     2  nlast, lend2
      logical *2 fhint, ftwind, fpsdx, fpsdy, fxpsd
      logical *4 ein
c
c           check arguments
c
      if ( nincr .le. 0 .or. npts .le. 1 .or. delt .le. 0.0 .or.
     1    m .le. 1 .or. inc1 .le. 0 .or. inc2 .le. 0 ) then
        ierr = 129
        return
        end if
c
c           initialize
c
      fhint = ein( 'z', cdo ) .or. ein( 'Z', cdo )
      lend = 2 ** m
      lend2 = lend / 2
      mm = m
      len = lend
      if ( fhint ) then
        len = len * 2
        mm = mm + 1
        end if
      len2 = len / 2
c    1  ' DELT,CDO ', g13.6, a )
      if ( lend .gt. npts ) then
        ierr = 130
        return
        end if
c
      d = cexp( cmplx( 0.0, pi / len ) )
      delf = 1.0 / delt / float( len )
      ierr = 0
      nfft = 0
      nincr2 = 2 * nincr
      j = 1
      nlast = nincr * ( npts - 1 ) + 1
      ilast1 = inc1 * ( len2 - 1 ) + 1
      ilast2 = inc2 * ( len2 - 1 ) + 1
      fpsdx = ein( 'x', cdo ) .or. ein( 'X', cdo )
      fpsdy = ein( 'y', cdo ) .or. ein( 'Y', cdo )
      fxpsd = ein( 's', cdo ) .or. ein( 'S', cdo )
c
c           initialize for time window
c
      ftwind = ein( 'h', cdo ) .or. ein( 'H', cdo )
      if ( ftwind ) then
        h = 2.0 * pi / ( lend + 1 )
        s = 1.0 - h **2 / 2.0 * ( 1.0 - h **2 / 12.0 )
        s1 = h * ( 1.0 - h **2 / 6.0 )
        s2 = - s1 - h **5 / 96.0
        end if
c
c           zero the psd arrays
c
      if ( fpsdx ) then
        do i = 1, inc1 * ( len2 - 1 ) + 1, inc1
          psdx(i) = 0.0
          end do
        end if
      if ( fpsdy ) then
        do i = 1, inc1 * ( len2 - 1 ) + 1, inc1
          psdy(i) = 0.0
          end do
        end if
      if ( fxpsd ) then
        do i = 1, inc2 * ( len2 - 1 ) + 1, inc2
          xpsd(i) = ( 0.0, 0.0 )
          end do
        end if
c
c           start FFT averaging loop
c
      do while ( j .le. nincr * ( npts - lend ) + 1 )
c
c           find the mean value of data in the segment
c
        xmean = qmean( x(j), nincr, lend )
        ymean = qmean( y(j), nincr, lend )
c
c           time window & load the FFT array
c
        nfft = nfft + 1
        if ( ftwind ) then
          z = -sqrt23
          zd = 0.0
          ni = lend
          nj = j + nincr * ( lend - 1 )
          do i = 1, lend2
            zz = s * z + s1 * zd
            zd = s2 * z + s * zd
            z = zz
            h = sqrt23 + z
            zwk(i) = cmplx( ( x(j) - xmean ) * h,
     1                      ( y(j) - ymean ) * h )
            zwk(ni) = cmplx( ( x(nj) - xmean ) * h,
     1                       ( y(nj) - ymean ) * h )
            j = j + nincr
            nj = nj - nincr
            ni = ni - 1
            end do
          j = j + lend2 * nincr
        else
c
c         load the FFT - no window
c
          do i = 1, lend
            zwk(i) = cmplx( x(j) - xmean, y(j) - ymean )
            j = j + nincr
            end do
          end if
c
c           pad the 2nd half of FFT array with zeros
c
        if ( fhint ) then
          do i = lend + 1, 2 * lend
            zwk(i) = ( 0.0, 0.0 )
            end do
          end if
c
c           do the "len" point FFT
c
        call fft2c( zwk, mm, iwk )
c
c           unscramble the transform ( with half integer frequencies )
c             & accumulate the PSD
c
        k1 = 1
        k2 = 1
        ni = len + 1
        w = 1.0
        zwk( len2 ) = zwk(1)
        do i = 1, len2+1
          a = ( zwk(i) + conjg( zwk(ni) ) ) * 0.5
          b = - conjg( zwk(i)) + zwk(ni)
          b = cmplx( aimag(b), real(b) ) * 0.5
          if ( fpsdx )
     1      psdx(k1) = psdx(k1) + real( a ) **2 + aimag( a ) **2
          if ( fpsdy )
     1      psdy(k1) = psdy(k1) + real( b ) **2 + aimag( b ) **2
          if ( fxpsd )
     1      xpsd(k2) = xpsd(k2) + a * conjg( b )
          k1 = k1 + inc1
          k2 = k2 + inc2
          ni = ni - 1
          end do
        end do
c end loop
c
c           apply autocorrelation window to each spectral estimate
c
      if ( ein( 'a', cdo ) .or. ein( 'A', cdo ) ) then
        if ( fpsdx ) then
          s2 = 0.0
          s1 = ( psdx(1) + psdx(1+inc1) ) / 4.0
          psdx(1) = s1
          do i = 1 + inc1, ilast1 - inc1, inc1
            s = ( psdx(i+inc1) + psdx(i) ) / 4.0
            psdx(i) = s + psdx( i-inc1 ) - s2
            s2 = s1
            s1 = s
            end do
          psdx(ilast1) = s
          end if
        if ( fpsdy ) then
          s2 = 0.0
          s1 = ( psdy(1) + psdy(1+inc1) ) / 4.0
          psdy(1) = s1
          do i = 1 + inc1, ilast1 - inc1, inc1
            s = ( psdy(i+inc1) + psdy(i) ) / 4.0
            psdy(i) = s + psdy( i-inc1 ) - s2
            s2 = s1
            s1 = s
            end do
          psdy(ilast1) = s
          end if
        if ( fxpsd ) then
          s2 = 0.0
          s1 = ( xpsd(1) + xpsd(1+inc1) ) / 4.0
          xpsd(1) = s1
          do i = 1 + inc2, ilast2 - inc2, inc2
            s = ( xpsd(i+inc2) + xpsd(i) ) / 4.0
            xpsd(i) = s + xpsd( i-inc2 ) - s2
            s2 = s1
            s1 = s
            end do
          xpsd(ilast2) = s
          end if
        end if
c
c           normalize the results
c
      s = delt / float( nfft * lend )
      if ( fpsdx ) then
        do i = 1, ilast1, inc1
          if ( psdx(i) .le. 0.0 )
     1      ierr = 137
          psdx(i) = abs( psdx(i) ) * s
          end do
        end if
      if ( fpsdy ) then
        do i = 1, ilast1, inc1
          if ( psdy(i) .le. 0.0 )
     1      ierr = ior( ierr, 138 )
          psdy(i) = abs( psdy(i) ) * s
          end do
        end if
      if ( fxpsd ) then
        do i = 1, ilast2, inc2
          if ( xpsd(i) .eq. ( 0.0, 0.0 ) ) then
            ierr = ior( ierr, 140 )
          else
            xpsd(i) = cmplx( cabs( xpsd(i) ) * s,
     1                     atan2( aimag( xpsd(i) ), real( xpsd(i) ) ) )
            end if
          end do
        end if
      return
      end





      SUBROUTINE FFT2C( ARRAY, DFT_LEN, IDUMMY)
c
c Discrete Fourier Transform of Complex data array of length 2**M
c This routine replaces FFT2C in IMSL. Obtained from Numerical Recipes FOUR1.
c This code has not been brought up to operational standard.
c
c Arguments
c  real    array(*)   # data in this array is replaced by its DFT
c                     # the calling routine should treat this as a complex
c                     # array of length 2**m
c  integer dft_len    # 2**m = length of complex data array
c  integer idummy     # IMSL requires a work array here
c
c Constant
c   ISIGN = 1   for forward transformation
c         =-1   for backward transformation
c
      include 'stdio.def'
      integer ISIGN
      double precision TWO_PI
      parameter( ISIGN=1, TWO_PI=6.28318530717959d0 )
c
      double precision wr, wi, wpr, wpi, wtemp, theta
      real    array(*), tempr, tempi
      integer idummy, dft_len, i, j, m, n, nn, istep, mmax
c
      if ( dft_len .le. 0 .or. dft_len .gt. 16 ) then
          write( STD_ERR, 99 ) dft_len
99        format( ' FFT2C : ABORT : Illegal DFT length = ', i12 )
          return
          end if
c
      nn = 2 ** dft_len
      n = 2 * nn
      j = 1
c
c       Bit reversal
c
      do 11 i = 1, n, 2
          if ( j .gt. i ) then
              tempr = array(j)
              tempi = array(j+1)
              array(j)   = array(i)
              array(j+1) = array(i+1)
              array(i)   = tempr
              array(i+1) = tempi
              end  if
          m = n / 2
1         if ( ( m .ge. 2 ) .and. ( j .gt. m ) ) then
              j = j - m
              m = m / 2
              goto 1
              end if
          j = j + m
11        continue
      mmax = 2
c
c       Danielson-Lanczos algorithm
c
2     if ( n .gt. mmax ) then
          istep = 2 * mmax
          theta = TWO_PI / ( isign * mmax )
          wpr   = -2.d0 * dsin( 0.5d0 * theta ) **2
          wpi   = dsin(theta)
          wr    = 1.d0
          wi    = 0.d0
          do 13 m = 1, mmax, 2
              do 12 i = m, n, istep
                  j = i + mmax
                  tempr = sngl(wr) * array(j) - sngl(wi) * array(j+1)
                  tempi = sngl(wr) * array(j+1) + sngl(wi) * array(j)
                  array(j)   = array(i)   - tempr
                  array(j+1) = array(i+1) - tempi
                  array(i)   = array(i)   + tempr
                  array(i+1) = array(i+1) + tempi
12                continue
c
c       trignometric recurrence
c
              wtemp = wr
              wr = wr * wpr - wi * wpi + wr
              wi = wi * wpr + wtemp * wpi + wi
13           continue
          mmax = istep
          go to 2
          end if
      return
      end

