c
c    Copyright (C) 1997  Peter Vernon Lanspeary
c
c    This program is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 2 of the License, or
c    (at your option) any later version.
c
c    This program is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program; if not, write to the Free Software
c    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
c


c
c    DIG_HIST
c
c Processes a data stream obtained from digitized graph -- which 
c is obtained, for example, from DIGZCV/DIGTAB.
c
c compile :  f77 -o dig_hist -O dig_hist.f ~/lib/sun4/f77lib.a
c
c     The input data is assumed to take the form of two columns of numbers
c with comment lines acting as record separators. Each record is processed
c separately. Separating comments are copied to the output file.
c First the abscissae (first column) of the input are sorted into ascending
c order. A moving average (uniform weight) filter is then applied to the data
c to provide smoothing. A new data stream is then interpolated from this
c sorted and filtered data and written to the output .
c
c Synopsis :
c   dig_hist [-filter [filter_length]]
c            [-start  [start_abscissa]]
c            [-finish [finish_abscissa]]
c            [-size   [OP_record_length]
c            [-error]
c            [-help   [help_file]]
c            [input_file]
c            [output_file]
c
c   filter_length     length of the smoothing filter,
c                      min = 1, max = 10% of input record (default = 5 or 2%)
c   start_abscissa    first abscissa value of output data record
c                      min & default = minimum input, max = max input
c   finish_abscissa   last abscissa value of output data record
c                      min = minimum input, max & default = max input
c   OP_record_length  length of output data record
c                      min = 1, max = MAX_LEN (default = DEF_OP_LEN )
c                      min = 1, max = MAX_POLY, default = 3
c   help_file         help is written to this file (default = stderr)
c   input_file        default = stdin
c   output_file       default = stdout
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc            [-order  [interpolation_order]]
cc interpolation_order polynomial order of the interpolation
c  interp_order       # order of polynomial interpolation
c     1                pol_inter( record(1,ABSCISSA), record(1,FILTERED),
c     2                           1, IP_len, interp_order, HUNT_CLOSEST,
c     3                           OP_abscissa, interp_err )
c
c          if ( interp_order .eq. 0 ) then
c              interp_order = DEF_ORDER
c          else
c              interp_order = max( min( interp_order, MAX_POLY ), 1 )
c              end if
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      PROGRAM DIG_HIST
c
c integer constants
c   ABSCISSA     = 1      data array column no. for abscissa values
c   ORDINATE     = 2                                ordinate
c   FILTERED     = 3                                filtered
c   DEF_FILT_LEN = 5      default length of moving average filter
c   DEF_OP_LEN   = 100    default length of output data record
c   DEF_ORDER    = 3      default order of polynomial interpolation
c   HUNT_CLOSEST = 3      flag for polynomial interpolation
c   MAX_LEN      = 16384  maximum input and output record length
c   MAX_POLY     = 10     maximum order of interpolation polynomial
c
      include 'stdio.def'
      integer ABSCISSA, ORDINATE, FILTERED,
     1        DEF_FILT_LEN, DEF_OP_LEN, DEF_ORDER,
     2        HUNT_CLOSEST, MAX_LEN, MAX_POLY
      parameter ( ABSCISSA=1, ORDINATE=2, FILTERED=3,
     1    DEF_FILT_LEN = 5, DEF_OP_LEN=100, DEF_ORDER=3,
     2    HUNT_CLOSEST=3, MAX_LEN=16384, MAX_POLY=10 )
c
c logical variables
c  write_error        # write interpolation error as 3rd column of output file
c  parse_args         # parses program arguments and opens files
c  read_record        # reads a record from the input file
c
c integer variables
c  error_lu           # logical unit number for error messages
c  filter_len         # length of moving average filter
c  len_filter
c  input_lu           # logical unit number of input file
c  IP_len             # length of input data record
c  OP_len             # length of output data record
c  output_lu          # logical unit number of output file
c
c real variables
c  actual_finish      # final output data abscissa
c  actual_increm      # output data abscissa increment
c  actual_start       # initial output data abscissa
c  interp_err         # estimated interpolation error
c  OP_abscissa        # output data abscissa value
c  OP_finish          # required final output data abscissa
c  OP_ordinate        # output data ordinate value
c  OP_start           # required initial output data abscissa
c  pol_inter          # function : polynomial interpolation
c  record(MAX_LEN,3)  # storage for the input and filtered data
c
      logical write_error, parse_args, read_record
      integer error_lu, filter_len, input_lu, IP_len,
     1        OP_len, output_lu, len_filter
      real actual_finish, actual_increm, actual_start, interp_err,
     1     OP_abscissa, OP_finish, OP_ordinate, OP_start,
     2     qlrpl1, record(MAX_LEN,3)
c
c         set default values
c
      input_lu   = -1
      output_lu  = -1
      error_lu   = -1
c
c         Get input file name & ouput file name from arg list and open the
c         files . Also get averaging length and other args .
c
      if ( parse_args( filter_len, OP_start, OP_finish, OP_len,
     1             write_error, input_lu, output_lu, error_lu ) ) then
c
c         check the values from PARSE_ARGS
c
          if ( OP_len .eq. 0 ) then
              OP_len = DEF_OP_LEN
          else
              OP_len = max( min( OP_len, MAX_LEN ), 1 )
              end if
          if ( input_lu .lt. 0 .or. output_lu .lt. 0 .or.
     1         error_lu .lt. 0 ) then
              write( STD_ERR, 1 )
1             format( ' BUG : logical unit not assigned in PARSE_ARGS' )
              stop
              end if
c
c         Read one data record
c
c         Start of main loop
2         if ( read_record( input_lu, output_lu,
     1                      record(1,ABSCISSA), record(1,ORDINATE),
     2                      IP_len, MAX_LEN ) ) then
c
c         Sort abscissae into ascending order
c
              call sortsh( record, IP_len, 1, MAX_LEN, 1, 3 )
c
c         Apply averaging filter to ordinate values
c         Use default filter length if necessary .
c
              if ( filter_len .eq. 0 ) then
                  len_filter = max( ( IP_len * 3 ) / 100, DEF_FILT_LEN )
              else
                  len_filter = filter_len
                  end if
              call SMVMEAN( record(1,ORDINATE), 1, IP_len, len_filter,
     1                      record(1,FILTERED), 1 )
c
c         Compress out any duplicated abscissa values
c
              call comp_dup( record, IP_len, IP_len, 1, MAX_LEN, 3,
     1            0.2e-3 * ( 1.0 - record(1,ABSCISSA) ) /
     2                             record(IP_len,ABSCISSA) )

c
c             check the output data parameters
c
              actual_start  = max( OP_start,  record(1,ABSCISSA) )
              actual_finish = max( OP_finish, record(IP_len,ABSCISSA) )
              actual_increm = ( actual_finish - actual_start ) /
     1                        real( max( OP_len - 1, 1 ) )
c
c         Interpolate new data from specified abscissa starting point
c         end point and increment, write data to output file .
c
              do 4 OP_abscissa = actual_start, actual_finish +
     1                           actual_increm / 2.0, actual_increm
                  OP_ordinate = qlrpl1( record(1,ABSCISSA),
     1                record(1,FILTERED), 1, IP_len, OP_abscissa )
                  if ( write_error ) then
                      write( output_lu, 3 ) OP_abscissa, OP_ordinate,
     1                                      interp_err
                  else
                      write( output_lu, 3 ) OP_abscissa, OP_ordinate
                      end if
3                     format( 1x, 3g13.5 )
4                 continue
              goto 2
              end if
c             end of main loop
      else
          call GIVE_HELP( error_lu, MAX_LEN, DEF_OP_LEN, MAX_POLY )
          end if
      stop
      end



      LOGICAL FUNCTION PARSE_ARGS( FILTER_LEN, OP_START, OP_FINISH,
     1        OP_LEN, WRITE_ERROR,
     2        INPUT_LU, OUTPUT_LU, ERROR_LU )
c
c    Parse the program arguments and open the necessary files . The
c program arguments, which are all optional, are
c
c   dig_hist [-filter [filter_length]]
c            [-start  [start_abscissa]]
c            [-finish [finish_abscissa]]
c            [-size   [OP_rec_length]
c            [-error]
c            [-help   [help_file]]
c            [input_file]
c            [output_file]
c
c filter_length   length of the smoothing filter, (default = 0)
c start_abscissa  first abscissa value of output data record (default = 0)
c finish_abscissa last abscissa value of output data record (default = 0)
c OP_rec_length   length of output data record      (default=0)
c interpol_order  polynomial order of the interpolation
c                 min = 1, max = MAX_POLY, default = 3
c help_file       help is written to this file (default = stderr)
c input_file      default = stdin
c output_file     default = stdout
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc            [-order  [interpol_order]]
c      interp_order = 0
c          interp_order # order of interpolation polynomial
cc
cc                 order of polynomial interpolation
cc
c                  else if ( Carg_str .eq. '-order' ) then                4m
c                      if ( arg_no .lt. total_args ) then                 5b
c                          call getarg( arg_no + 1, Carg_str )
c                          read( Carg_str, *, ERR=3 ) interp_order
c                          arg_no = arg_no + 1
c                          end if                                         5e
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c Arguments  -- all are OUTPUT
c  integer filter_len   # length of moving average filter
c          op_len       # length of output data record
c          input_lu     # logical unit number of input file
c          output_lu    # logical unit number of output file
c          error_lu     # logical unit number of error file
c  real op_start        # initial abscissa value of output data
c       op_finish       # final abscissa value of output data
c  logical write_error  # write the interpolation error a 3rd column of data
c
c logical variables
c  got_input_file       # have opened input file
c  got_output_file      # have opened output file
c
c integer variables
c  arg_no               # current argument number
c  arg_len              # number of characters in the argument
c  total_args           # total number of arguments
c  iargc                # function : argument count
c  new_lun              #          : find new logical unit number
c  open_file            #          : open a file
c  str_length           #          : length of a string
c
c character string variables
c  Carg_str*80          # argument
c
      include 'stdio.def'
      integer filter_len, op_len,
     1        input_lu, output_lu, error_lu
      real    op_start, op_finish
      logical write_error
c
      logical got_input_file, got_output_file
      integer arg_no, total_args, iargc, new_lun, open_file, str_length
      character Carg_str*80
c
c         set the default values
c
      filter_len   = 0
      op_len       = 0
      input_lu     = STD_IN
      output_lu    = STD_OUT
      error_lu     = STD_ERR
      op_start     = 0.0
      op_finish    = 0.0
      got_input_file = .FALSE.
      got_output_file = .FALSE.
      parse_args   = .TRUE.
c
c         Examine each argument in turn
c
c     start of loop
      arg_no = 1
      total_args = iargc()
1     if ( arg_no .le. total_args ) then                                 1b
          call getarg( arg_no, Carg_str )
          call scase( Carg_str, Carg_str, 'l' )
c
c         deal with non-file arguments
c
          if ( Carg_str(1:1) .eq. '-' ) then                             2b
              if ( str_length( Carg_str ) .gt. 1 ) then                  3b
c
c                 length of moving average filter
c                                                                        4b
                  if ( Carg_str .eq. '-filter' ) then                    5b
                      if ( arg_no .lt. total_args ) then
                          call getarg( arg_no + 1, Carg_str )
                          read( Carg_str, *, ERR=3 ) filter_len
                          arg_no = arg_no + 1
                          end if                                         5e
c
c                 starting value of output record abscissa
c
                  else if ( Carg_str .eq. '-start' ) then                4m
                      if ( arg_no .lt. total_args ) then                 5b
                          call getarg( arg_no + 1, Carg_str )
                          read( Carg_str, *, ERR=3 ) op_start
                          arg_no = arg_no + 1
                          end if                                         5e
c
c                 final value of output record abscissa
c
                  else if ( Carg_str .eq. '-finish' ) then               4m
                      if ( arg_no .lt. total_args ) then                 5b
                          call getarg( arg_no + 1, Carg_str )
                          read( Carg_str, *, ERR=3 ) op_finish
                          arg_no = arg_no + 1                            5e
                          end if
c
c                 size of output record abscissa
c
                  else if ( Carg_str .eq. '-size' ) then                 4m
                      if ( arg_no .lt. total_args ) then                 5b
                          call getarg( arg_no + 1, Carg_str )
                          read( Carg_str, *, ERR=3 ) OP_len
                          arg_no = arg_no + 1
                          end if                                         5e
c
c                 write the interpolation error as a third column
c
                  else if ( Carg_str .eq. '-error' ) then                4m
                      write_error = .TRUE.
c
c                 help
c
                  else if ( Carg_str .eq. '-help' ) then                 4m
                      if ( arg_no .lt. total_args ) then                 5b
                          call getarg( arg_no + 1, Carg_str )
                          if ( Carg_str(1:1) .ne. '-' ) then             6b
                              error_lu = open_file( new_lun(),
     1                                             Carg_str, 'new', 0 )
                              if ( error_lu .eq. 0 ) then                7b
                                  write( STD_ERR, 2 ) Carg_str
2                                 format( ' Cannot open file ', a )
                                  parse_args = .FALSE.
                                  error_lu = STD_ERR
                                  end if                                 7e
                              arg_no = arg_no + 1
                              end if                                     6e
                          end if                                         5e
                      parse_args = .FALSE.
                      end if                                             4e
                  end if                                                 3e
c
c                 first non-flag argument is the input file name
c
          else if ( .not. got_input_file ) then                          2m
              input_lu = open_file( new_lun(), Carg_str, 'old', 0 )
              if ( input_lu .eq. 0 ) then                                3b
                  write( STD_ERR, 2 ) Carg_str
                  parse_args = .FALSE.
                  input_lu = STD_IN
              else                                                       3m
                  got_input_file = .TRUE.
                  end if                                                 3e
c
c                 second non-flag argument is the output file name
c
          else if ( .not. got_output_file ) then                         2m
              output_lu = open_file( new_lun(), Carg_str, 'new', 0 )
              if ( output_lu .eq. 0 ) then                               3b
                  write( STD_ERR, 2 ) Carg_str
                  parse_args = .FALSE.
                  output_lu = STD_OUT
              else                                                       3m
                  got_output_file = .TRUE.
                  end if                                                 3e
              end if                                                     2e
          arg_no = arg_no + 1
          goto 1
c
c             handling of error reading the numerical value after an argument
c
3         write( STD_ERR, 4 )
4         format( ' Cannot decode numeric argument - assuming default' )
          goto 1
          end if                                                         1e
c         end of loop
      return
      end


      LOGICAL FUNCTION READ_RECORD( INPUT_LU, OUTPUT_LU, X, Y, NPTS,
     1    MAX_PTS )
c
c  Reads a single record from the data file with logical unit number
c  INPUT_LU . Initial lines beginning with ';' are copied to the file
c  with logical unit number OUTLU . The following lines are interpreted
c  as (X,Y) coordinate pairs until the next comment line is encountered .
c  This new comment line is stored in a buffer until READ_RECORD is used
c  again, and it is then copied to the output file .
c
c Arguments
c  integer input_lu     # logical unit number of input file   INPUT
c          output_lu    # logical unit number of output file  INPUT
c  real    x(*),        # abscissa values                    OUTPUT
c          y(*)         # ordinate values                    OUTPUT
c  integer npts,        # number of input data coordinates   OUTPUT
c          max_pts      # maximum allowed size of data record INPUT
c
c logical variables
c   buffer_empty        # comment storage buffer is empty
c integer variables
c   columns             # number of items in a non-comment line
c   line_len            # length of a comment line
c   jljstn              # function : left justify & count columns
c   str_length          #          : length of a string
c character variables
c   Cbuffer *80         # comment buffer
c   Cline *80           # line read from input file
c
      include 'stdio.def'
      integer input_lu, output_lu, npts, max_pts
      real x(*), y(*)
c
      logical buffer_empty
      integer columns, line_len, jljstn, str_length
      character Cbuffer*81, Cline*81
      data buffer_empty / .TRUE. /
c
c             if the buffer is not empty flush it to output file
c
      npts = 0
      if ( .not. buffer_empty ) then
          write( output_lu, 1, ERR=10 ) Cbuffer(1:str_length(Cbuffer))
1         format( 1x, a )
          buffer_empty = .TRUE.
          end if
c
c             copy initial comments to output file and skip blank lines
c
c     start of loop
2         read( input_lu, 3, ERR=8, END=7 ) Cline
3         format( a )
          line_len = jljstn( Cline, columns )
          if ( line_len .le. 0 ) then
              goto 2
          else if ( Cline(1:1) .eq. ';' ) then
              write( output_lu, 1, ERR=10 ) Cline(1:line_len)
              goto 2
              end if
c         end of loop
c
c             read the data until the next comment is found
c
c     start of loop
4         if ( jljstn( Cline, columns ) .le. 0 ) then
              read( input_lu, 3, ERR=8, END=7 ) Cline
              goto 4
              end if
          if ( Cline(1:1) .ne. ';' ) then
              if ( columns .le. 2 ) then
                  read( Cline, *, ERR=5 ) x(npts+1), y(npts+1)
                  npts = npts + 1
                  if ( npts .gt. max_pts ) goto 7
                  end if
              read( input_lu, 3, ERR=8, END=7 ) Cline
              goto 4
c
5             write( STD_ERR, 6 )
6             format( ' WARNING: Cannot decode input data; skipped' )
              read( input_lu, 3, ERR=8, END=7 ) Cline
              goto 4
              end if
c         end of loop
c
c             fall through to here if a comment line is found
c
      Cbuffer = Cline
      buffer_empty = .FALSE.
7     read_record = npts .gt. 0
      return
c
c             trap : error reading input file
c
8     write( STD_ERR, 9 )
9     format( ' ABORT: Cannot read input file' )
      read_record = .FALSE.
      return
c
c             trap : error writing to output file
c
10    write( STD_ERR, 11 )
11    format( ' ABORT: Cannot write to output file' )
      read_record = .FALSE.
      return
      end


      SUBROUTINE COMP_DUP( ARRAY, IN_ROWS, OUT_ROWS, COMP_COL,
     1                     ROWS, COLUMNS, TOLERANCE )
c
c    COMPress out DUPlicate rows
c Compresses out the rows of the ARRAY, which contain a value in the
c column COMP_COL duplicating the previous row .
c
c Arguments
c  real    array(rows,columns) # IN / OUT # data array
c  integer in_rows             # IN # number of rows to be checked
c          out_rows           # OUT # number of rows remaining after compression
c          comp_col            # IN # column to be compressed
c          rows, columns       # IN # size of array
c  real tolerance              # IN # minimum difference for 2 values to be
c                                   # considered different
c                                   # tolerance = ABS( x - y) /
c                                   #             ( ABS( x ) + ABS( y ) )
c integer variables
c  col_index      # column index used for copying rows
c  dest_index     # row index of the destination row
c  ref_index      # row index of the reference row
c  src_index      # row index of the source row
c real variables
c  tol            # absolute value of tolerance
c
      include 'stdio.def'
      integer columns, col_index, comp_col, dest_index, in_rows,
     1        out_rows, ref_index, rows, src_index
      real array(rows,columns), tol, tolerance
c
c         check the arguments
c
      tol = abs( tolerance )
      if ( rows    .lt. 2 .or. columns .lt. 1    .or.
     1     in_rows .lt. 2 .or. in_rows .gt. rows )
     2    return
c
c         The source row is compared with the reference row . If there
c         is a significant difference, the source row is copied to the
c         destination row which immediately follows the reference row .
c
      ref_index  = 1
      dest_index = ref_index + 1
      do 2 src_index = 2, in_rows
          if ( abs( array(src_index,comp_col) -
     1              array(ref_index,comp_col) ) /
     2         ( abs( array(src_index,comp_col) ) +
     3           abs( array(ref_index,comp_col) ) ) .gt. tol ) then
c
c         Don't copy if the source row is the same as the destination row
c
              if ( src_index .ne. dest_index ) then
                  do 1 col_index = 1, columns
                      array(dest_index,col_index) =
     1                    array(src_index,col_index)
1                     continue
                  end if
              ref_index = dest_index
              dest_index = dest_index + 1
              end if
2          continue
      out_rows = ref_index
      return
      end


      REAL FUNCTION QLRPL1( X, Y, NINCR, NPTS, X0 )
c
c       LineaR inTerPoLation from tabulated data .
c   Function to linearly interpolate ordinate y0, given a table of (x,y)
c coordinates in the arrays X & Y and an abicissa value x0 . Starting from the
c first elments of X & Y , and with an indexing  increment of NINCR , NPTS
c coordinate pairs are used from X & Y .
c The array X is searched until adjacent values x(l) & x(r) straddle x0 .
c   N.B. ELEMENTS OF X MUST BE IN STRICTLY ASCENDING ORDER OR STRICTLY
c        DESCENDING ORDER
c
c       Arguments
c real  x(*),y(*),# arrays of tabulated (x,y) coordinates
c       x0        # target for which a y0 value is required
c integer nincr,  # data array indexing increment
c         npts    # number of data points used
c
c  Variables
c integer n       # size of arrays
c         l       # left hand indices of the search interval
c logical pass1   # .true. until an attempt to extrapolate is detected
c
c
      include 'stdio.def'
      real x(*), y(*), x0
      integer nincr, npts, n, l, Jbsrch
c
c       Find the array interval
c
      if( npts .le. 0 .or. nincr .le. 0 ) return
      n = 1 + ( npts - 1 ) * nincr
      l = max( 1, min( Jbsrch( x, nincr, npts, x0 ), n - nincr ) )
c
c       Interpolate
c
      qlrpl1 = ( y(l+nincr) - y(l) ) / ( x(l+nincr) - x(l) ) *
     1           ( x0 - x(l) ) + y(l)
      return
      end



      SUBROUTINE GIVE_HELP( HELP_LU, MAX_LEN, DEF_OP_LEN, MAX_POLY )
c
c  Write help message to the file with logical unit number HELP_LU .
c
c  help_lu   = logical unit number of output file
c  max_len    = maximum input and output record lengths
c  def_op_len = default length of output record
c  max_poly   = maximum order of polynomial interpolation
c
      include 'stdio.def'
      integer help_lu, max_len, def_op_len, max_poly
      write( help_lu, 1, ERR=4 )
1     format( '    DIG_HIST' //
     1 ' Processes a data stream obtained from digitized graph .' //
     2 '   The input data is assumed to take the form of two columns' /
     3 ' of numbers with comment lines acting as record separators .' /
     4 ' Each record is processed separately. Separating comments are' /
     5 ' copied to the output file. Comments begin with a semicolon .'//
     6 ' First, the abscissae (first column) of the input are sorted' /
     7 ' into ascending order. A moving average (uniform weight)' /
     8 ' filter is then applied to the data to provide smoothing. A' /
     9 ' new data stream is then interpolated from this sorted and' /
     a ' filtered data and written to the output .' / )

      write( help_lu, 2, ERR=4 )
2     format( '   Synopsis :' / ' dig_hist [-filter [filter_length]]' /
     1 10x, '[-start  [start_abscissa]]' /
     2 10x, '[-finish [finish_abscissa]]' /
     3 10x, '[-size   [OP_record_length]' /
c     4 10x, '[-order  [interpolation_order]]' /
     5 10x, '[-error] -- write interpolation error as a third column' /
     6 10x, '[-help   [help_file]]' /
     7 10x, '[input_file]' / 10x, '[output_file]' //
     8 ' filter_length     length of the smoothing filter' /
     9 19x, 'min = 1, max = 3% of input record (default = 5 or 2%)' /
     a ' start_abscissa    first abscissa value of output data record' /
     b 19x, 'min & default = minimum input, max = max input' /
     c ' finish_abscissa   last abscissa value of output data record' )

      write( help_lu, 3, ERR=4 ) max_len, def_op_len, max_poly, max_len
3     format( 19x, 'min = minimum input, max & default = max input' /
     1 ' OP_record_length  length of output data record' /
     2 19x, 'min = 1, max = ', i5, '(default = ', i5, ' )' /
c     3 ' interpolation     polynomial order of the interpolation' /
     4 19x, 'min = 1, max = ', i2,', default = 3' /
     5 ' help_file         help is written to this (default = stderr)' /
     6 ' input_file        default = stdin' /
     7 ' output_file       default = stdout' /
     8 3x, i5, ' = maximum input record length' )
      return
c
4     write( STD_ERR, 5 )
5     format( 'ABORT : Cannot write to help file' )
      return
      end
